   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                    DRIVE MODULE                     */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _DRIVE_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_

#include "setup.h"

#if DEFRULE_CONSTRUCT

#include "constant.h"
#include "symbol.h"
#include "clipsmem.h"
#include "expressn.h"
#include "evaluatn.h"
#include "network.h"
#include "facts.h"
#include "match.h"
#include "engine.h"
#include "reteutil.h"
#include "utility.h"
#include "drive.h"
#include "router.h"
#include "retract.h"
#include "generate.h"
#include "lgcldpnd.h"
   
/***************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/***************************************/

#if ANSI_COMPILER  
   static VOID                    PPDrive(struct partialMatch *,struct partialMatch *,struct joinNode *);
   static VOID                    PNRDrive(struct joinNode *,struct partialMatch *);
   static VOID                    EmptyDrive(struct joinNode *,struct partialMatch *);
   static VOID                    JoinNetErrorMessage(struct joinNode *);
#else
   static VOID                    PPDrive();
   static VOID                    PNRDrive();
   static VOID                    EmptyDrive();
   static VOID                    JoinNetErrorMessage();
#endif

/****************************************/
/* GLOBAL INTERNAL VARIABLE DEFINITIONS */
/****************************************/

   globle BOOLEAN              IncrementalResetFlag = CLIPS_FALSE;
   globle BOOLEAN              IncrementalReset = CLIPS_TRUE;

/********************************************************************************/
/* Drive: Primary routine for driving a partial match through the join network. */
/********************************************************************************/
globle VOID Drive(binds,join,enterDirection,useCopy)
  struct partialMatch *binds;
  struct joinNode *join;
  int enterDirection;
  int useCopy;
  {
   struct partialMatch *lhsBinds = NULL, *rhsBinds = NULL;
   struct partialMatch *comparePMs, *newBinds;
   int exprResult;

   /*=========================================================*/
   /* If an incremental reset is being performed and the join */
   /* is not part of the network to be reset, then return.    */
   /*=========================================================*/
   
   if (IncrementalResetFlag && (join->initialize == CLIPS_FALSE)) 
     {
      if (useCopy) ReturnPartialMatch(binds);
      return;
     }
   
   /*==========================================*/
   /* Handle the storage of the partial match. */
   /*==========================================*/
   
   if (enterDirection == LHS)
     {
      /*=========================================================*/
      /* If the associated LHS pattern is a not CE, then we need */
      /* an additional field in the partial match to keep track  */
      /* of the pseudo fact if one is created.                   */
      /*=========================================================*/
      
      if (join->LHSPatternIsNegated)
        {
         newBinds = AddSingleMatch(binds,NULL);
         newBinds->notOriginf = CLIPS_TRUE;
         newBinds->counterf = CLIPS_TRUE;
         newBinds->bcount--;    /* Don't include the not origin slot. */
         if (useCopy) ReturnPartialMatch(binds);
         binds = newBinds;
        }
        
      /*=======================================================*/
      /* If it's the last join, we need an additional field to */
      /* point to the activation associated with this match.   */
      /*=======================================================*/
      
      else if (join->lastJoin)
        {
         newBinds = AddSingleMatch(binds,NULL);
         newBinds->activationf = CLIPS_TRUE;
         newBinds->bcount--;    /* Don't include the activation slot. */
         if (useCopy) ReturnPartialMatch(binds);
         binds = newBinds;
        }
        
      /*======================================*/
      /* Otherwise make a copy of the partial */
      /* match if it can't be used directly.  */
      /*======================================*/
      
      else if (useCopy == CLIPS_FALSE)
       { binds = CopyPartialMatch(binds); }
       
      /*====================================================*/
      /* Store the partial match in the join's beta memory. */
      /*====================================================*/
      
      binds->next = join->beta;
      join->beta = binds;
     }
   
   /*=======================================================*/
   /* If the link to next join is null, then the terminator */
   /* join has been reached and the rule is ready to fire.  */
   /* This join is connected to the list of actions for the */
   /* rule about to fire.  Add the rule to the agenda.      */
   /*=======================================================*/
    
   if (join->lastJoin == CLIPS_TRUE)
     {
      AddActivation(join,binds);
      return;
     }

   /*==================================================*/
   /* Use a special routine if this is the first join. */
   /*==================================================*/

   if (join->firstJoin)
     {
      EmptyDrive(join,binds);
      return;
     }

   /*============================*/
   /* Initialize some variables. */
   /*============================*/
      
   if (enterDirection == LHS)
     {
      comparePMs = join->entry_pat->alpha;
      lhsBinds = binds;
     }
   else if (enterDirection == RHS)
     {
      comparePMs = join->beta;
      rhsBinds = binds;
     }
   else
     {
      CLIPSSystemError("DRIVE",1);
      ExitCLIPS(5);
     }
     
   /*===================================================*/
   /* Compare each set of binds on the opposite side of */
   /* the join with the set of binds that entered this  */
   /* join.  If the binds don't mismatch, then perform  */
   /* the appropriate action for the logic of the join. */
   /*===================================================*/
      
   while (comparePMs != NULL)
     {
      exprResult = CLIPS_FALSE;
      
      if (enterDirection == RHS)
        { lhsBinds = comparePMs; }
      else
        { rhsBinds = comparePMs; } 

      /*======================================================*/
      /* Evaluate the join expression for this combination of */
      /* rhs and lhs bindings.                                */
      /*======================================================*/
     
      if (join->eval == NULL)
        { exprResult = CLIPS_TRUE; }
      else
        { 
         exprResult = EvaluateJoinExpression(join->eval,lhsBinds,rhsBinds,join); 
         if (EvaluationError)
           {  
            if (join->LHSPatternIsNegated) exprResult = CLIPS_TRUE;
            SetEvaluationError(CLIPS_FALSE);
           }
        }

      /*====================================================*/
      /* If the join expression evaluated to true (i.e.     */ 
      /* there were no conflicts between variable bindings, */
      /* all tests were satisfied, etc.), then perform the  */
      /* appropriate action given the logic of this join.   */
      /*====================================================*/
 
      if (exprResult != CLIPS_FALSE)
        {
         if (join->LHSPatternIsNegated == CLIPS_FALSE)
           { PPDrive(lhsBinds,rhsBinds,join); }
         else if (enterDirection == RHS)
           { PNRDrive(join,comparePMs); }
         else if (enterDirection == LHS)
           { binds->binds[binds->bcount] = (struct alphaMatch *) (((long int) binds->binds[binds->bcount]) + 1); }
        }
 
      comparePMs = comparePMs->next;
     } 
     
   /*======================================================*/
   /* If a join with a positive lhs and a negative rhs was */
   /* entered from the lhs side of the join, and the join  */
   /* test failed for all sets of matches for the new      */
   /* bindings on the lhs side (the counter on the lhs is  */
   /* set to zero), then the lhs bindings should be send   */
   /* down to the joins below along with an NID marker     */
   /* that represents the instance of the not pattern that */
   /* was satisfied.                                       */
   /*======================================================*/

   if ((join->LHSPatternIsNegated == CLIPS_TRUE) &&
       (enterDirection == LHS) &&
       (((long int) binds->binds[binds->bcount]) == 0L))
     { PNLDrive(join,binds); }
   
   return; 
  }

/****************************************************************/
/* EvaluateJoinExpression: Evaluates join expressions. Performs */
/*   a faster evaluation for join expressions than if           */
/*   EvaluateExpression were used directly. Function calls to   */
/*   (eq_vars), (neq_vars), /and/, and /or/ are evaluated       */
/*   directly. All other function calls are evaluated using     */
/*   EvaluateExpression.                                        */
/****************************************************************/
globle BOOLEAN EvaluateJoinExpression(jn_test,lbinds,rbinds,joinPtr)
  struct expr *jn_test;
  struct partialMatch *lbinds, *rbinds;
  struct joinNode *joinPtr;
  {
   int p1, e1, p2, e2;
   DATA_OBJECT vresult;
   struct element *elem_a, *elem_b;
   struct fact *fact_ptr1, *fact_ptr2;
   int a_extent, b_extent;
   struct multifieldMarker *marks1, *marks2; 
   int and_logic, result = CLIPS_TRUE;
   union
     {
      VOID *theValue;
      struct 
         {
          unsigned int pass : 1;
          unsigned int fail : 1;
          unsigned int int1 : 7;
          unsigned int int2 : 8;
          unsigned int int3 : 7;
         } theInts;
     } hack;
     
   if (jn_test == NULL) return(CLIPS_TRUE);
   
   if (jn_test->value == PTR_AND)
     {
      and_logic = CLIPS_TRUE;
      jn_test = jn_test->arg_list;
     }
   else if (jn_test->value == PTR_OR)
     {
      and_logic = CLIPS_FALSE;
      jn_test = jn_test->arg_list;
     }
   else
     { and_logic = CLIPS_TRUE; }
     
   while (jn_test != NULL)
     {
   /*==============================================================*/
   /* Evaluate expressions eq_vars and neq_vars.  Calls to these   */
   /* functions are expressed in the format:                       */
   /*   (eq_vars  <pattern m> <element n> <pattern p> <element r>) */
   /*   (neq_vars <pattern m> <element n> <pattern p> <element r>) */
   /* Function eq_vars returns true (1.0) if the nth element of    */
   /* the mth pattern is equal to the rth element of the pth       */
   /* pattern.  Function neq_vars acts similarly to eq_vars except */
   /* the return value is negated.                                 */
   /*==============================================================*/

   if (jn_test->type == SCALL_COMP_VARS)
     {
      /*==============================================*/
      /* Determine which elements are to be extracted */
      /* out of which patterns.                       */
      /*==============================================*/                  
      
      hack.theValue = jn_test->value;
      
      p1 = joinPtr->depth - 1;
      e1 = (int) hack.theInts.int1;
      p2 = ((int) hack.theInts.int2) - 1;
      e2 = (int) hack.theInts.int3;

      /*==============================================*/
      /* Extract the fact pointer and list of segment */
      /* markers for the first pattern.               */
      /*==============================================*/

      if ((p1 > 0) || (lbinds == NULL))
        { 
         fact_ptr1 = rbinds->binds[0]->origin;
         marks1 = rbinds->binds[0]->marker; 
        }
      else
        {
         fact_ptr1 = lbinds->binds[p1]->origin;
         marks1 = lbinds->binds[p1]->marker; 
        }

      /*==============================================*/
      /* Extract the fact pointer and list of segment */
      /* markers for the second pattern.              */
      /*==============================================*/

      if (p2 == p1)
        { 
         fact_ptr2 = fact_ptr1;
         marks2 = marks1;
        }
      else
        {
         fact_ptr2 = lbinds->binds[p2]->origin;
         marks2 = lbinds->binds[p2]->marker;
        }

      /*=============================================*/
      /* Determine the beginning point and extent of */
      /* the comparision for the first pattern.      */
      /*=============================================*/

      a_extent = 1;

      if (marks1 != NULL)
        { 
         if (marks1->element < e1)
           { 
            e1 = AdjustFieldPosition(marks1,e1,&a_extent);
           }
         else if (marks1->element == e1)
           { a_extent = (marks1->end - marks1->start) + 1; }
        }

      /*=============================================*/
      /* Determine the beginning point and extent of */
      /* the comparision for the second pattern.     */
      /*=============================================*/

      b_extent =  1;

      if (marks2 != NULL)
        { 
         if (marks2->element < e2)
           { 
            e2 = AdjustFieldPosition(marks2,e2,&b_extent);
           }
         else if (marks2->element == e2)
           { b_extent = (marks2->end - marks2->start) + 1; }
        }
  
      /*===============================================*/
      /* If the extents are not the same, then the     */
      /* comparision fails (e.g. two segment variables */
      /* of different length.                          */
      /*===============================================*/

      if (a_extent != b_extent)
        { result = (int) hack.theInts.fail; }

      /*=========================================*/
      /* If the extents are both zero, then the  */
      /* comparision succeeds (e.g. two segment  */
      /* variables of length zero).              */
      /*=========================================*/

      else if (a_extent == 0)
        { result = (int) hack.theInts.pass; }

      /*==============================================*/
      /* Compare the two variables element by element */
      /* for the length of the extent.                */
      /*==============================================*/
     
      else
        {
         e1--;
         elem_a = &fact_ptr1->atoms[e1];
         e2--;
         elem_b = &fact_ptr2->atoms[e2];

         result = -1;
         while (result < 0)
           {
            if (elem_a->type != elem_b->type) result = (int) hack.theInts.fail;
            else if (elem_a->value != elem_b->value) result = (int) hack.theInts.fail;

            if (result < 0)
              {
               a_extent--;

               if (a_extent == 0)
                 { result = (int) hack.theInts.pass; }
               else
                 {
                  elem_a++;
                  elem_b++;
                 }
              }
           }   
        }
     }

   /*=========================================================*/
   /* Evaluate or expressions expressed in the format:        */
   /*   (or <expression 1> <expression 2> ... <expression n>) */
   /* Returns true (1.0) if any of the expression are true,   */
   /* otherwise returns CLIPS_FALSE (0.0).                    */
   /*=========================================================*/

   else if (jn_test->value == PTR_OR)
     { 
      result = CLIPS_FALSE;
      if (EvaluateJoinExpression(jn_test,lbinds,rbinds,joinPtr) == CLIPS_TRUE)
        { 
         if (EvaluationError) return(CLIPS_FALSE);
         result = CLIPS_TRUE; 
        } 
      else if (EvaluationError) 
        { return(CLIPS_FALSE); }     
     }

   /*==========================================================*/
   /* Evaluate and expressions expressed in the format:        */
   /*   (and <expression 1> <expression 2> ... <expression n>) */
   /* Returns CLIPS_FALSE (0) if any of the expression are     */
   /* CLIPS_FALSE, otherwise returns true (1).                 */
   /*==========================================================*/

   else if (jn_test->value == PTR_AND)
     {
      result = CLIPS_TRUE;
      if (EvaluateJoinExpression(jn_test,lbinds,rbinds,joinPtr) == CLIPS_FALSE)
        { 
         if (EvaluationError) return(CLIPS_FALSE);
         result = CLIPS_FALSE;
        } 
      else if (EvaluationError) 
        { return(CLIPS_FALSE); }
     }

   /*==========================================================*/
   /* Evaluate all other expressions using EvaluateExpression. */
   /*==========================================================*/

   else
     {
      struct partialMatch *old_lhsBinds = NULL;
      struct partialMatch *old_rhsBinds = NULL;
   
      old_lhsBinds = GlobalLHSBinds;
      old_rhsBinds = GlobalRHSBinds;  
      GlobalLHSBinds = lbinds;
      GlobalRHSBinds = rbinds;

      EvaluateExpression(jn_test,&vresult);
      
      GlobalLHSBinds = old_lhsBinds;
      GlobalRHSBinds = old_rhsBinds;
      
      if (EvaluationError) 
        {
         JoinNetErrorMessage(joinPtr);
         return(CLIPS_FALSE);
        }
      
      if ((vresult.value == (VOID *) CLIPSFalseSymbol) && (vresult.type == SYMBOL))
        { result = CLIPS_FALSE; }
      else
        { result = CLIPS_TRUE; }
     }
     
      if ((and_logic == CLIPS_TRUE) && (result == CLIPS_FALSE)) return(CLIPS_FALSE);
      else if ((and_logic == CLIPS_FALSE) && (result == CLIPS_TRUE)) return(CLIPS_TRUE);
      
      jn_test = jn_test->next_arg;
     }
     
   return(result);
  }

/*******************************************************************/
/* PPDrive: Handles the merging of an alpha memory partial match   */
/*   with a beta memory partial match for a join that has positive */
/*   LHS entry and positive RHS entry. The partial matches being   */
/*   merged have previously been checked to determine that they    */
/*   satisify the constraints for the join. Once merged, the new   */
/*   partial match is sent to each child join of the join from     */
/*   which the merge took place.                                   */
/*******************************************************************/
static VOID PPDrive(lhsBinds,rhsBinds,join)
  struct partialMatch *lhsBinds, *rhsBinds;
  struct joinNode *join;
  {
   struct partialMatch *linker;
   struct joinNode *listOfJoins;

   /*==================================================*/
   /* Merge the alpha and beta memory partial matches. */
   /*==================================================*/

   linker = MergePartialMatches(lhsBinds,rhsBinds);
   
   /*================================================*/
   /* Send the new partial match to all child joins. */
   /*================================================*/

   listOfJoins = join->next_level;
   while (listOfJoins->right_drive_node != NULL)
     {
      Drive(linker,listOfJoins,LHS,CLIPS_FALSE);
      listOfJoins = listOfJoins->right_drive_node;
     }

   Drive(linker,listOfJoins,LHS,CLIPS_TRUE);

   return;
  }

/**********************************************************************/
/* PNRDrive: Handles the entry of a partial match from the RHS of a   */
/*   join that has positive LHS entry and negative RHS entry (meaning */
/*   the conditional element associated with this join is a not       */
/*   conditional element). Entry of the alpha memory partial match    */
/*   will cause the count value of the associated beta memory partial */
/*   match to be incremented. This in turn may cause partial matches  */
/*   associated with the beta memory partial match to be removed from */
/*   the network.                                                     */
/**********************************************************************/
static VOID PNRDrive(join,lhsBinds)
  struct joinNode *join;
  struct partialMatch *lhsBinds;
  {
   FACT_ID notID;
   struct joinNode *listOfJoins;

   /*=================================================================*/ 
   /* The partial matches contained in the beta memory of a positive  */
   /* negative join have a "counter" attached to them which indicates */
   /* the number alpha memory partial matches which are consistent    */
   /* with them. If no alpha memory partial matches are consistent,   */
   /* then the not CE has been satisfied, the counter will have a     */
   /* value of < 0 (which indicates the id associated with the not CE */
   /* and that there are no consistent bindings), and the LHS         */
   /* bindings will have been sent down to the next join.  This       */
   /* routine handles the case in which a partial match entered the   */
   /* negative RHS entry of the join and were consistent with the     */
   /* LHS partial matches. If the counter for the LHS partial match   */
   /* is less than zero, then all partial matches with that "not" id  */
   /* value must be removed from descendent joins. If the counter for */
   /* the partial match is greater than zero, then the counter only   */
   /* needs to be incremented.                                        */
   /*=================================================================*/ 
        
   if (lhsBinds->counterf == CLIPS_FALSE)
     {
      /*===========================================================*/
      /* The counter flag was FALSE. This means that a pointer to  */
      /* the pseudo-fact matching the not CE is stored directly in */
      /* the partial match. Determine the ID of this pseudo-fact   */
      /* and remove all partial matches from descendent joins that */
      /* contain the ID.                                           */
      /*===========================================================*/
      
      notID = lhsBinds->binds[lhsBinds->bcount]->whoset;

      listOfJoins = join->next_level;
      while (listOfJoins != NULL)
        {
         PosEntryRetract(LHS,listOfJoins,notID,join->depth-1);
         listOfJoins = listOfJoins->right_drive_node; 
        } 
        
      /*========================================*/
      /* Put the pseudo-fact on a garbage list. */
      /*========================================*/
      
      lhsBinds->binds[lhsBinds->bcount]->aorigin = GarbageAlphaMatches;
      GarbageAlphaMatches = lhsBinds->binds[lhsBinds->bcount];
      
      /*===========================================================*/
      /* Set the partial match's count value to 1. Set the counter */
      /* flag to TRUE to indicate that the counter is active.      */
      /*===========================================================*/
      
      lhsBinds->binds[lhsBinds->bcount] = (struct alphaMatch *) 1L;
      lhsBinds->counterf = CLIPS_TRUE;
     }
   else
     { 
      /*=============================================================*/
      /* The counter flag was TRUE. This means that count value is   */
      /* active and greater than zero. Simply increment the counter. */
      /*=============================================================*/
      
      lhsBinds->binds[lhsBinds->bcount] = 
        (struct alphaMatch *) (((long int) lhsBinds->binds[lhsBinds->bcount]) + 1);
     }
         
  }
  
/********************************************************************/
/* PNLDrive: Handles the entry of a partial match from the LHS of a */
/*   join that has positive LHS entry and negative RHS entry        */
/*   (meaning the conditional element associated with this join is  */
/*   a not conditional element). An new partial match is created by */
/*   combining the match from the beta memory with a "pseudo"       */
/*   partial match corresponding to the facts which didn't match    */
/*   the not CE. Once merged, the new partial match is sent to each */
/*   child join of the join from which the merge took place.        */
/********************************************************************/
globle VOID PNLDrive(join,binds)
  struct joinNode *join;
  struct partialMatch *binds;
  {
   struct joinNode *listOfJoins;
   struct partialMatch *linker;
   int joinExpr;
   struct alphaMatch *tempAlpha;
   long int newNotID;
   
   /*=============================================================*/
   /* Check the secondary join expression. If this evaluates to   */
   /* FALSE then the alpha partial match is not consistent with   */
   /* the beta partial match and no new partial match is created. */
   /*=============================================================*/
   
   if (join->not_eval != NULL)
     { 
      joinExpr = EvaluateJoinExpression(join->not_eval,binds,NULL,join);
      EvaluationError = CLIPS_FALSE;
      if (joinExpr == CLIPS_FALSE) return;
     }
        
    /*=======================================================*/
    /* Create a pseudo-fact representing the facts which did */
    /* not match the not CE associated with this join.       */
    /*=======================================================*/  
    
    newNotID = IncrementPseudoFactIndex();
    tempAlpha = get_struct(alphaMatch);
    tempAlpha->aorigin = tempAlpha;
    tempAlpha->whoset = newNotID;
    tempAlpha->origin = NULL;
    tempAlpha->marker = NULL;
    
    /*===============================================*/
    /* Store the pointer to the pseudo-fact directly */
    /* in the beta memory partial match.             */
    /*===============================================*/
    
    binds->counterf = CLIPS_FALSE; 
    binds->binds[binds->bcount] = tempAlpha;
    
    /*=======================================================*/
    /* Create a new partial match by merging the beta memory */
    /* partial match with the pseudo-fact partial match.     */
    /*=======================================================*/
    
    linker = AddSingleMatch(binds,tempAlpha);
      
    /*========================================================*/
    /* Send the merged partial match to all descendent joins. */
    /*========================================================*/

    listOfJoins = join->next_level;
    while (listOfJoins->right_drive_node != NULL)
      {
       Drive(linker,listOfJoins,LHS,CLIPS_FALSE);
       listOfJoins = listOfJoins->right_drive_node;
      }

    Drive(linker,listOfJoins,LHS,CLIPS_TRUE);
   }

/***************************************************************/
/* EmptyDrive: Handles the entry of a alpha memory partial     */
/*   match from the RHS of a join that is the first join of    */
/*   a rule (i.e. a join that cannot be entered from the LHS). */
/*   Both positive and negative RHS join entry are handled.    */  
/***************************************************************/
static VOID EmptyDrive(join,rhsBinds)
  struct joinNode *join;
  struct partialMatch *rhsBinds;
  {
   struct partialMatch *linker;
   struct joinNode *listOfJoins;
   int joinExpr;

   /*======================================================*/
   /* Determine if the alpha memory partial match satifies */
   /* the primary join expression. If it doesn't then no   */
   /* further action is taken.                             */
   /*======================================================*/

   if (join->eval != NULL)
     {
      joinExpr = EvaluateJoinExpression(join->eval,NULL,rhsBinds,join);
      EvaluationError = CLIPS_FALSE;
      if (joinExpr == CLIPS_FALSE) return;
     }
   
   /*==========================================================*/
   /* If the join's RHS entry is associated with a not CE      */
   /* (negative entry), then remove partial matches associated */
   /* with this join from descendent joins if necessary.       */
   /*==========================================================*/

   if (join->LHSPatternIsNegated == CLIPS_TRUE)
     {
      /*======================================================*/
      /* If the join's count value is greater than zero, then */
      /* just increment by one because some other fact has    */
      /* already prevented the not CE from being satisfied.   */
      /*======================================================*/
      
      if (join->id > 0)
        {
         join->id++;
         return;
        }

      /*=================================================*/
      /* Get rid of the pseudo-fact partial match stored */
      /* in the beta memory of the join.                 */
      /*=================================================*/
      
      if (join->beta == NULL)
        {
         if (IncrementalReset == CLIPS_TRUE)
           {
            CLIPSSystemError("DRIVE",2);
            ExitCLIPS(5);
           }
        }
      else
        {
         join->beta->next = GarbagePartialMatches;
         GarbagePartialMatches = join->beta;
         join->beta = NULL;
        }
      
      /*=========================================================*/
      /* Clear all partial matches from the beta memories of all */
      /* descendent joins since all descendent partial matches   */
      /* are dependent upon the single partial match created by  */
      /* this join.                                              */
      /*=========================================================*/
         
      listOfJoins = join->next_level;
      while (listOfJoins != NULL)
        {
         ClearLowerBetaMemory(listOfJoins);
         listOfJoins = listOfJoins->right_drive_node;
        }

      join->id = 1;
      return;
     }

   /*=========================================================*/
   /* If the join's RHS entry is associated with a pattern CE */
   /* (positive entry), then copy the alpha memory partial    */
   /* match and send it to all child joins.                   */
   /*=========================================================*/

   linker = CopyPartialMatch(rhsBinds);

   listOfJoins = join->next_level;
   while (listOfJoins->right_drive_node != NULL)
     {
      Drive(linker,listOfJoins,LHS,CLIPS_FALSE);
      listOfJoins = listOfJoins->right_drive_node;
     }

   Drive(linker,listOfJoins,LHS,CLIPS_TRUE);
  }

/*******************************************************************/
/* ClearLowerBetaMemory: Removes all partial matches from the beta */
/*   memory of a join and all joins which descend from that join.  */
/*******************************************************************/
globle VOID ClearLowerBetaMemory(node)
  struct joinNode *node;
  {
   struct partialMatch *listOfPMs, *tempPM;
   struct joinNode *listOfJoins;

   /*====================================================*/
   /* Remove beta memory partial matches from this join. */
   /*====================================================*/

   listOfPMs = node->beta;
   node->beta = NULL;
   while (listOfPMs != NULL)
     { 
      tempPM = listOfPMs->next;
          
      /*=================================================*/
      /* If the partial match is linked to an activation */
      /* on the agenda, then remove the link.            */
      /*=================================================*/
      
      if ((listOfPMs->activationf) ? 
          (listOfPMs->binds[listOfPMs->bcount] != NULL) : FALSE)
        { RemoveActivation((struct activation *) listOfPMs->binds[listOfPMs->bcount],CLIPS_TRUE,CLIPS_FALSE); }
        
      /*===========================================================*/
      /* If facts are logically dependent upon this partial match, */
      /* then check to see if the fact should be retracted.        */
      /*===========================================================*/
      
      if (listOfPMs->dependents != NULL)
        { AddToDependencyList(listOfPMs); }
        
      /*============================================================*/
      /* If the partial match has a pseudo-fact associated with it, */
      /* then place the partial match on a garbage list because     */
      /* other data structures may be pointing to the pseudo-fact.  */
      /* Otherwise, remove the partial match.                       */
      /*============================================================*/
      
      if (listOfPMs->notOriginf)
        {
         listOfPMs->next = GarbagePartialMatches;
         GarbagePartialMatches = listOfPMs;
        }
      else
        { ReturnPartialMatch(listOfPMs); }
      
      listOfPMs = tempPM;
     }

   /*========================================*/
   /* Recursively remove beta memory partial */
   /* matches from descendent joins.         */
   /*========================================*/

   listOfJoins = node->next_level;
   while (listOfJoins != NULL)
     {
      ClearLowerBetaMemory(listOfJoins);
      listOfJoins = listOfJoins->right_drive_node;
     }
  }
 
/********************************************************************/
/* JoinNetErrorMessage: Prints an informational message indicating  */
/*   which join of a rule generated an error when a join expression */
/*   was being evaluated.                                           */
/********************************************************************/
static VOID JoinNetErrorMessage(joinPtr)
  struct joinNode *joinPtr;
  {
   char buffer[60];
   
   PrintCLIPS(WERROR,"\nThis error occurred in the join network\n");
   
   sprintf(buffer,"   Problem resides in join #%d in rule(s):\n",joinPtr->depth);
   PrintCLIPS(WERROR,buffer);
   
   TraceErrorToRule(joinPtr->next_level,"      ");
   PrintCLIPS(WERROR,"\n");
  }
  
/*****************************************************/
/* GetIncrementalReset: Returns the current value of */
/*   the IncrementalReset flag.                      */
/*****************************************************/
globle BOOLEAN GetIncrementalReset()
  { return(IncrementalReset); }
  
/**************************************************/
/* SetIncrementalReset: Sets the current value of */
/*   the IncrementalReset flag.                   */
/**************************************************/
globle BOOLEAN SetIncrementalReset(value)
  int value;
  { 
   int ov;
   
   ov = IncrementalReset;
   if (GetNextDefrule(NULL) != NULL) return(-1);
   IncrementalReset = value;
   return(ov);
  }

#endif
