
   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                 RULE COMPILER MODULE                */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*      Barry Cameron                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#include "setup.h"

#if DEFRULE_CONSTRUCT && (! RUN_TIME)

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "symbol.h"
#include "constant.h"
#include "expressn.h"
#include "evaluatn.h"
#include "utility.h"
#include "reteutil.h"
#include "build.h"
#include "engine.h"

#if CONSTRUCT_COMPILER
#include "conscomp.h"
#endif

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
#include "bload.h"
#include "bsave.h"
#include "facts.h"
#include "match.h"
#include "reteutil.h"
#include "network.h"
#include "clipsmem.h"

#define PATTERNS 0
#define JOINS    1

#define FindPatternPointer(patPtr) ((patPtr == NULL) ? -1L : patPtr->bsaveID)
#define FindJoinPointer(joinPtr) ((joinPtr == NULL) ? -1L :  joinPtr->bsaveID)

#define PatternPointer(x,i) ((struct patternNode *) (((long) i == -1L) ? NULL : \
                                   &x[(long) i]))
#define JoinPointer(x,i) ((struct joinNode *) (((long) i == -1L) ? NULL : \
                                   &x[(long) i]))
#endif

/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/
   
#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   static unsigned long                NumberOfDefrules;
   static unsigned long                NumberOfPatternPointers;
   static unsigned long                NumberOfPatterns;
   static unsigned long                NumberOfJoins;
   static struct defrule HUGE_ADDR    *DefruleArray;
   static struct patptr HUGE_ADDR     *PatPtrArray;
   static struct patternNode HUGE_ADDR   *PatternArray;
   static struct joinNode HUGE_ADDR   *JoinArray;
#endif  

/***************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/***************************************/

#if ANSI_COMPILER  
   VOID                           DefruleBinarySetup(void);
#if CONSTRUCT_COMPILER
   static int                     DefrulesToCode(char *,int,FILE *,int,int);
   static VOID                    InitDefrulesCode(FILE *,int,int);
#endif
#if BLOAD_AND_BSAVE
   static VOID                    BsaveFind(void);
   static VOID                    BsaveExpressions(FILE *);
   static VOID                    BsavePatternExpressions(FILE *);
   static VOID                    BsaveJoinExpressions(FILE *);
   static VOID                    BsaveDefrules(FILE *);
   static VOID                    BsavePatterns(FILE *);
   static VOID                    BsaveJoins(FILE *);
   static VOID                    FindPatternItems(struct patternNode *);
   static VOID                    FindJoinItems(struct joinNode *,int);
#endif
#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   static VOID                    BloadDefrules(void);
   static VOID                    ClearBload(void);
   static int                     ClearBloadReady(void);
#endif
#else
   VOID                           DefruleBinarySetup();
#if CONSTRUCT_COMPILER
   static int                     DefrulesToCode();
   static VOID                    InitDefrulesCode();
#endif
#if BLOAD_AND_BSAVE
   static VOID                    BsaveFind();
   static VOID                    BsaveExpressions();
   static VOID                    BsavePatternExpressions();
   static VOID                    BsaveJoinExpressions();
   static VOID                    BsaveDefrules();
   static VOID                    BsavePatterns();
   static VOID                    BsaveJoins();
   static VOID                    FindPatternItems();
   static VOID                    FindJoinItems();
#endif
#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   static VOID                    BloadDefrules();
   static VOID                    ClearBload();
   static int                     ClearBloadReady();
#endif
#endif

#if CONSTRUCT_COMPILER

/**********************************************************/
/* InitDefrulesCode:    */
/**********************************************************/
static VOID InitDefrulesCode(initFP,imageID,maxIndices)
  FILE *initFP;
  int imageID;
  int maxIndices;
  {
   if (GetNextDefrule(NULL) != NULL)
     { 
      int topPattern;
      
      topPattern = (int) NetworkPointer()->bsaveID;
      
      fprintf(initFP,"   SetListOfDefrules(drl%d_1);\n",imageID);
      fprintf(initFP,"   SetNetworkPointer(&pnd%d_%d[%d]);\n",
                       imageID,(topPattern / maxIndices) + 1,
                                 topPattern % maxIndices); 
      fprintf(initFP,"   ResetDeployedRuleImage();\n");
     }
   else
     { 
      fprintf(initFP,"   SetListOfDefrules(NULL);\n");
      fprintf(initFP,"   SetNetworkPointer(NULL);\n");
     }
  }
     
/**********************************************************/
/* DefrulesToCode:  Produces all the defrule definitions. */
/**********************************************************/
static int DefrulesToCode(fileName,fileID,headerFP,imageID,maxIndices)
  char *fileName;
  int fileID;
  FILE *headerFP;
  int imageID;
  int maxIndices;
  {
   struct defrule *dfPtr, *tempRule;
   FILE *fp;
   int version = 1;
   int arrayVersion = 1;
   int i = 1;
   int newHeader = CLIPS_TRUE;
   unsigned long pcount = 0, jcount = 0;
   int ppCount = 0;
   int ruleCount = 0, nextRule = 0;
   struct patptr *ppPtr;
   struct patternNode *pnPtr;
   struct joinNode *jnPtr;
   
   /*==================*/
   /* Create the file. */
   /*==================*/
   
   fprintf(headerFP,"#include \"defrule.h\"\n");
   if (GetNextDefrule(NULL) == NULL) return(-1);
   if ((fp = NewCFile(fileName,fileID,version)) == NULL) return(-1);
   fprintf(headerFP,"#include \"network.h\"\n");
   fprintf(headerFP,"#include \"reteutil.h\"\n");
   
   TagRuleNetwork(&pcount,&jcount);
   
   /*==============================*/
   /* Dump the defrule structures. */
   /*==============================*/
   
   dfPtr = (struct defrule *) GetNextDefrule(NULL);
     
   while (dfPtr != NULL)
     {
      if (newHeader)
        {
         fprintf(fp,"struct defrule drl%d_%d[] = {\n",imageID,arrayVersion);
         fprintf(headerFP,"extern struct defrule drl%d_%d[];\n",imageID,arrayVersion);
         newHeader = CLIPS_FALSE;
        }
     
      ruleCount++;
      fprintf(fp,"{");                                /* name */
      PrintSymbolReference(fp,dfPtr->name); 
      fprintf(fp,",NULL,%d,%d,0,",dfPtr->salience,dfPtr->complexity);  
      
      ExpressionToCode(fp,dfPtr->dynamicSalience);    /* dynamicSalience */
      fprintf(fp,",");
      ExpressionToCode(fp,dfPtr->actions);            /* actions */
      
      /* logicalJoin */
      
      if (dfPtr->logicalJoin == NULL)
        { fprintf(fp,",NULL,"); }
      else
        {
         fprintf(fp,",&jnd%d_%d[%d],",
                     imageID,(((int) dfPtr->logicalJoin->bsaveID) / maxIndices) + 1,
                              ((int) dfPtr->logicalJoin->bsaveID) % maxIndices); 
        }
        
      /* pats */

      if (dfPtr->pats == NULL)
        { fprintf(fp,"NULL,"); }
      else
        {
         fprintf(fp,"&ppt%d_%d[%d],",imageID, 
                                       (ppCount / maxIndices) + 1,
                                       ppCount % maxIndices); 
        }

      for (ppPtr = dfPtr->pats; ppPtr != NULL ; ppPtr = ppPtr->next) ppCount++;

      /* disjunct */

      if (dfPtr->disjunct == NULL)
        { fprintf(fp,"NULL,"); }
      else
        { 
         fprintf(fp,"&drl%d_%d[%d],",imageID,
                                       (ruleCount / maxIndices) + 1,
                                       ruleCount % maxIndices);
        }

      /* next */
      
      if (dfPtr->next == NULL)
        { fprintf(fp,"NULL}"); }
      else
        { 
         tempRule = dfPtr->disjunct;
         nextRule = 0;
         while (tempRule != NULL)
           {
            nextRule++;
            tempRule = tempRule->disjunct;
           }
           
         fprintf(fp,"&drl%d_%d[%d]}",imageID,
                                       ((ruleCount + nextRule) / maxIndices) + 1,
                                        (ruleCount + nextRule) % maxIndices);
        }
                                       
      i++;
      if (dfPtr->disjunct != NULL) dfPtr = dfPtr->disjunct;
      else dfPtr = dfPtr->next;
      
      if ((i > maxIndices) || (dfPtr == NULL))
        {
         fprintf(fp,"};\n");
         fclose(fp);
         i = 1;
         version++;
         arrayVersion++;
         if (dfPtr != NULL)
           { 
            if ((fp = NewCFile(fileName,fileID,version)) == NULL) return(0);
            newHeader = CLIPS_TRUE;
           }
        }
      else
        { fprintf(fp,",\n"); }
     }
     
   /*=============================*/
   /* Dump the patptr structures. */
   /*=============================*/
      
   if ((fp = NewCFile(fileName,fileID,version)) == NULL) return(-1);
   newHeader = CLIPS_TRUE;
   
   arrayVersion = 1;
   i = 1;
   
   dfPtr = (struct defrule *) GetNextDefrule(NULL);
   ppPtr = dfPtr->pats;
     
   while (ppPtr != NULL)
     {
      if (newHeader)
        {
         fprintf(fp,"struct patptr ppt%d_%d[] = {\n",imageID,arrayVersion);
         fprintf(headerFP,"extern struct patptr ppt%d_%d[];\n",imageID,arrayVersion);
         newHeader = CLIPS_FALSE;
        }
     
      if (ppPtr->pptr == NULL)
        { fprintf(fp,"{NULL,"); }
      else
        {
         fprintf(fp,"{&pnd%d_%d[%d],",
                    imageID,(((int) ppPtr->pptr->bsaveID) / maxIndices) + 1,
                            ((int) ppPtr->pptr->bsaveID) % maxIndices);
        }
        
      if (ppPtr->jptr == NULL)
        { fprintf(fp,"NULL,"); }
      else
        {
         fprintf(fp,"&jnd%d_%d[%d],",
                     imageID,(((int) ppPtr->jptr->bsaveID) / maxIndices) + 1,
                              ((int) ppPtr->jptr->bsaveID) % maxIndices); 
        }
                           
                           
      if (ppPtr->next == NULL)
        { fprintf(fp,"NULL}"); }
      else
        { fprintf(fp,"&ppt%d_%d[%d]}",imageID,(i / maxIndices) + 1,i % maxIndices); }
        
      i++;
      ppPtr = ppPtr->next;
      while ((ppPtr == NULL) && (dfPtr != NULL))
        {
         if (dfPtr->disjunct != NULL) dfPtr = dfPtr->disjunct;
         else dfPtr = dfPtr->next;
         if (dfPtr != NULL) ppPtr = dfPtr->pats;
        }
        
      if ((i > maxIndices) || (ppPtr == NULL))
        {
         fprintf(fp,"};\n");
         fclose(fp);
         i = 1;
         version++;
         arrayVersion++;
         if (ppPtr != NULL)
           { 
            if ((fp = NewCFile(fileName,fileID,version)) == NULL) return(0);
            newHeader = CLIPS_TRUE;
           }
        }
      else
        { fprintf(fp,",\n"); }
     }
     
   /*===================================*/
   /* Dump the pattern node structures. */
   /*===================================*/
      
   if ((fp = NewCFile(fileName,fileID,version)) == NULL) return(-1);
   newHeader = CLIPS_TRUE;
   
   arrayVersion = 1;
   i = 1;
   
   dfPtr = (struct defrule *) GetNextDefrule(NULL);
   ppPtr = dfPtr->pats;
   pnPtr = ppPtr->pptr;
     
   while (pnPtr != NULL)
     {
      if (newHeader)
        {
         fprintf(fp,"struct patternNode pnd%d_%d[] = {\n",imageID,arrayVersion);
         fprintf(headerFP,"extern struct patternNode pnd%d_%d[];\n",imageID,arrayVersion);
         newHeader = CLIPS_FALSE;
        }
     
      pnPtr->marked = 0;
      fprintf(fp,"{%d,%d,%d,0,0,0,0,",
                 pnPtr->singlefieldNode,
                 pnPtr->multifieldNode,
                 pnPtr->stopNode);
         
      ExpressionToCode(fp,pnPtr->eval); /* eval */
         
      fprintf(fp,",NULL,");             /* alpha */
                  
      if (pnPtr->entry_join == NULL)
        { fprintf(fp,"NULL,"); }
      else
        {
         fprintf(fp,"&jnd%d_%d[%d],",
                 imageID,(((int) pnPtr->entry_join->bsaveID) / maxIndices) + 1,
                         ((int) pnPtr->entry_join->bsaveID) % maxIndices);
        }
                 
      if (pnPtr->next_level == NULL)
        { fprintf(fp,"NULL,"); }
      else
        {
         fprintf(fp,"&pnd%d_%d[%d],",
                    imageID,(((int) pnPtr->next_level->bsaveID) / maxIndices) + 1,
                            ((int) pnPtr->next_level->bsaveID) % maxIndices);
        }
               
      if (pnPtr->last_level == NULL)
        { fprintf(fp,"NULL,"); }
      else
        {
         fprintf(fp,"&pnd%d_%d[%d],",
              imageID,(((int) pnPtr->last_level->bsaveID) / maxIndices) + 1,
                   ((int) pnPtr->last_level->bsaveID) % maxIndices);
        }
      
      if (pnPtr->left_node == NULL)
        { fprintf(fp,"NULL,"); }
      else
        {
         fprintf(fp,"&pnd%d_%d[%d],",
              imageID,(((int) pnPtr->left_node->bsaveID) / maxIndices) + 1,
                   ((int) pnPtr->left_node->bsaveID) % maxIndices);
        }
        
      if (pnPtr->right_node == NULL)
        { fprintf(fp,"NULL}"); }
      else
        {
         fprintf(fp,"&pnd%d_%d[%d]}",
              imageID,(((int) pnPtr->right_node->bsaveID) / maxIndices) + 1,
                   ((int) pnPtr->right_node->bsaveID) % maxIndices);
        }
            
      i++;
        
      pnPtr = pnPtr->last_level;
      while (((pnPtr == NULL) ? CLIPS_TRUE : (pnPtr->marked == 0)) && (dfPtr != NULL))
        {
         if (ppPtr->next != NULL)
           { 
            ppPtr = ppPtr->next;
            pnPtr = ppPtr->pptr;
           }
         else if (dfPtr->disjunct != NULL) 
           {
            dfPtr = dfPtr->disjunct;
            ppPtr = dfPtr->pats;
            pnPtr = ppPtr->pptr;
           }
         else 
           {
            dfPtr = dfPtr->next;
            if (dfPtr != NULL) 
              {
               ppPtr = dfPtr->pats;
               pnPtr = ppPtr->pptr;
              }
           }
        }
        
      if ((i > maxIndices) || (pnPtr == NULL))
        {
         fprintf(fp,"};\n");
         fclose(fp);
         i = 1;
         version++;
         arrayVersion++;
         if (pnPtr != NULL)
           { 
            if ((fp = NewCFile(fileName,fileID,version)) == NULL) return(0);
            newHeader = CLIPS_TRUE;
           }
        }
      else if (pnPtr != NULL)
        { fprintf(fp,",\n"); }
     }
     
   /*================================*/
   /* Dump the join node structures. */
   /*================================*/
      
   if ((fp = NewCFile(fileName,fileID,version)) == NULL) return(-1);
   newHeader = CLIPS_TRUE;
   
   arrayVersion = 1;
   i = 1;
   
   dfPtr = (struct defrule *) GetNextDefrule(NULL);
   ppPtr = dfPtr->pats;
   jnPtr = ppPtr->jptr;
     
   while (jnPtr != NULL)
     {
      if (newHeader)
        {
         fprintf(fp,"#define EP (struct expr *)\n\n");
         fprintf(fp,"struct joinNode jnd%d_%d[] = {\n",imageID,arrayVersion);
         fprintf(headerFP,"extern struct joinNode jnd%d_%d[];\n",imageID,arrayVersion);
         newHeader = CLIPS_FALSE;
        }
     
      jnPtr->marked = 0;
      fprintf(fp,"{%d,%d,%d,0,0,0,NULL,",
                 jnPtr->firstJoin,
                 jnPtr->lastJoin,
                 jnPtr->LHSPatternIsNegated);
                 
      /* eval */
      
      if (jnPtr->eval == NULL)
        { fprintf(fp,"NULL,"); }
      else
        {
         if (jnPtr->lastJoin == CLIPS_FALSE)
           { 
            ExpressionToCode(fp,jnPtr->eval);  
            fprintf(fp,",");
           }
         else
           {
            int ruleIndex;
            
            ruleIndex = GetDisjunctIndex((RULEPTR) jnPtr->eval);
            fprintf(fp,"EP &drl%d_%d[%d],",imageID,
                                       (ruleIndex / maxIndices) + 1,
                                       ruleIndex % maxIndices);
           
           }
        }     
      
      /* not_eval */
      
      ExpressionToCode(fp,jnPtr->not_eval);  
      
      /* ID, depth */
      fprintf(fp,",0,%d,",jnPtr->depth);             
                  
      if (jnPtr->entry_pat == NULL)
        { fprintf(fp,"NULL,"); }
      else
        {
         fprintf(fp,"&pnd%d_%d[%d],",
                 imageID,(((int) jnPtr->entry_pat->bsaveID) / maxIndices) + 1,
                         ((int) jnPtr->entry_pat->bsaveID) % maxIndices);
        }
                 
      if (jnPtr->next_level == NULL)
        { fprintf(fp,"NULL,"); }
      else
        {
         fprintf(fp,"&jnd%d_%d[%d],",
                    imageID,(((int) jnPtr->next_level->bsaveID) / maxIndices) + 1,
                            ((int) jnPtr->next_level->bsaveID) % maxIndices);
        }
               
      if (jnPtr->last_level == NULL)
        { fprintf(fp,"NULL,"); }
      else
        {
         fprintf(fp,"&jnd%d_%d[%d],",
              imageID,(((int) jnPtr->last_level->bsaveID) / maxIndices) + 1,
                   ((int) jnPtr->last_level->bsaveID) % maxIndices);
        }
      
      if (jnPtr->right_drive_node == NULL)
        { fprintf(fp,"NULL,"); }
      else
        {
         fprintf(fp,"&jnd%d_%d[%d],",
              imageID,(((int) jnPtr->right_drive_node->bsaveID) / maxIndices) + 1,
                   ((int) jnPtr->right_drive_node->bsaveID) % maxIndices);
        }
        
      if (jnPtr->right_match_node == NULL)
        { fprintf(fp,"NULL}"); }
      else
        {
         fprintf(fp,"&jnd%d_%d[%d]}",
              imageID,(((int) jnPtr->right_match_node->bsaveID) / maxIndices) + 1,
                   ((int) jnPtr->right_match_node->bsaveID) % maxIndices);
        }
            
      i++;
      jnPtr = NULL;
      
      while (((jnPtr == NULL) ? CLIPS_TRUE : (jnPtr->marked == 0)) && (dfPtr != NULL))
        {
         if (ppPtr->next != NULL)
           { 
            ppPtr = ppPtr->next;
            jnPtr = ppPtr->jptr;
           }
         else if (dfPtr->disjunct != NULL) 
           {
            dfPtr = dfPtr->disjunct;
            ppPtr = dfPtr->pats;
            jnPtr = ppPtr->jptr;
           }
         else 
           {
            dfPtr = dfPtr->next;
            if (dfPtr != NULL) 
              {
               ppPtr = dfPtr->pats;
               jnPtr = ppPtr->jptr;
              }
           }
        }
        
      if ((i > maxIndices) || (jnPtr == NULL))
        {
         fprintf(fp,"};\n");
         fclose(fp);
         i = 1;
         version++;
         arrayVersion++;
         if (jnPtr != NULL)
           { 
            if ((fp = NewCFile(fileName,fileID,version)) == NULL) return(0);
            newHeader = CLIPS_TRUE;
           }
        }
      else if (jnPtr != NULL)
        { fprintf(fp,",\n"); }
     }
     
   return(1);
  }
  
#endif

#if BLOAD_AND_BSAVE 

/**********************************/
/* BsaveFind:                     */
/**********************************/
static VOID BsaveFind()
  {
   struct defrule *rPtr, *tPtr;
   
   rPtr = (struct defrule *) GetNextDefrule(NULL);
   while (rPtr != NULL)
     {
      rPtr->name->neededSymbol = CLIPS_TRUE;
      ExpressionCount += ExpressionSize(rPtr->dynamicSalience);
      MarkNeededItems(rPtr->dynamicSalience);
      tPtr = rPtr;
      while (tPtr != NULL)
        {
         ExpressionCount += ExpressionSize(tPtr->actions);
         MarkNeededItems(tPtr->actions);
         tPtr = tPtr->disjunct;
        }
        
      rPtr = (struct defrule *) GetNextDefrule(rPtr);
     } 
      
   FindPatternItems(NetworkPointer());
  }
  
/**********************************/
/* BsaveExpressions:              */
/**********************************/
static VOID BsaveExpressions(fp)
  FILE *fp;
  {
   struct defrule *rPtr, *tPtr;
   
   rPtr = (struct defrule *) GetNextDefrule(NULL);
   while (rPtr != NULL)
     {
      BsaveExpression(rPtr->dynamicSalience,fp);
      tPtr = rPtr;
      while (tPtr != NULL)
        {
         BsaveExpression(tPtr->actions,fp);
         tPtr = tPtr->disjunct;
        }
      rPtr = (struct defrule *) GetNextDefrule(rPtr);
     } 
      
   MarkRuleNetwork(1);
   BsavePatternExpressions(fp);
                          
   BsaveJoinExpressions(fp);
  }

/****************************************************/
/* BsavePatternExpressions: Saves expressions found */
/*   in the pattern network to the binary file.     */
/****************************************************/
static VOID BsavePatternExpressions(fp)
  FILE *fp;
  {
   struct defrule *rulePtr;
   struct patptr *ppPtr;
   struct patternNode *pnPtr;
   
   rulePtr = (struct defrule *) GetNextDefrule(NULL);
   
   while (rulePtr != NULL)
     {
      ppPtr = rulePtr->pats;
      while (ppPtr != NULL)
        {
         pnPtr = ppPtr->pptr;
         while (pnPtr != NULL)
           {
            if (pnPtr->marked)
              { 
               pnPtr->marked = 0;
               if (pnPtr->eval != NULL) BsaveExpression(pnPtr->eval,fp);
               pnPtr = pnPtr->last_level; 
              }
            else
              { pnPtr = NULL; }
           }
         
         ppPtr = ppPtr->next;
        }
     
     if (rulePtr->disjunct != NULL) rulePtr = rulePtr->disjunct;
     else rulePtr = rulePtr->next;
    }
  }
  
/**************************************************/
/* BsaveJoinExpressions:  Saves expressions found */
/*   in the join network to the binary file.      */                                            
/**************************************************/
static VOID BsaveJoinExpressions(fp)
  FILE *fp;
  {
   struct defrule *rulePtr;
   struct patptr *ppPtr;
   struct joinNode *joinPtr;
   
   rulePtr = (struct defrule *) GetNextDefrule(NULL);
   
   while (rulePtr != NULL)
     {
      ppPtr = rulePtr->pats;
      while (ppPtr != NULL)
        {
         joinPtr = ppPtr->jptr;
         if (joinPtr->marked)
           { 
            joinPtr->marked = 0;
            if ((joinPtr->eval != NULL) && (joinPtr->lastJoin == CLIPS_FALSE))
              { BsaveExpression(joinPtr->eval,fp); }
           
            if (joinPtr->not_eval != NULL)
              { BsaveExpression(joinPtr->not_eval,fp); }
           }
         
         ppPtr = ppPtr->next;
        }
     
     if (rulePtr->disjunct != NULL) rulePtr = rulePtr->disjunct;
     else rulePtr = rulePtr->next;
    }
  }

/*****************************************************/
/* BsaveDefrules: Writes out all defrule structures */ 
/*   to the binary file                             */
/*****************************************************/
static VOID BsaveDefrules(fp)
  FILE *fp;
  {
   unsigned long int count = 0, space;
   struct defrule *rPtr, *tPtr;
   struct patptr *pjPtr, tempPatPtr;
   unsigned long int numberOfPatternPointers = 0;
   struct defrule tempDefrule;
   unsigned long int PatternCount, JoinCount, topPattern;
   int first;
   
   /*===================================*/
   /* Write out the number of defrules. */
   /*===================================*/
   
   rPtr = (struct defrule *) GetNextDefrule(NULL);
   while (rPtr != NULL)
     {
      tPtr = rPtr;
      while (tPtr != NULL)
        {
         count++;
         pjPtr = tPtr->pats;
         while (pjPtr != NULL)
           {
            numberOfPatternPointers++;
            pjPtr = pjPtr->next;
           }
         tPtr = tPtr->disjunct;
        }
        
      rPtr = (struct defrule *) GetNextDefrule(rPtr);
     } 
     
   PatternCount = 0;
   JoinCount = 0;
   
   TagRuleNetwork(&PatternCount,&JoinCount);
   
   space = (count * sizeof(struct defrule)) + 
           (numberOfPatternPointers * sizeof(struct patptr)) +
           (PatternCount * sizeof(struct patternNode)) +
           (JoinCount * sizeof(struct joinNode)) +
           (4 * sizeof(unsigned long));
   GenWrite(&space,(unsigned long) sizeof(unsigned long int),fp);
   GenWrite(&count,(unsigned long) sizeof(unsigned long int),fp);
   GenWrite(&numberOfPatternPointers,(unsigned long) sizeof(unsigned long int),fp);
   GenWrite(&PatternCount,(unsigned long) sizeof(long int),fp);
   GenWrite(&JoinCount,(unsigned long) sizeof(long int),fp);
   if (PatternCount == 0) topPattern = 0;
   else topPattern = NetworkPointer()->bsaveID;
   GenWrite(&topPattern,(unsigned long) sizeof(long int),fp);
   
   /*=========================*/
   /* Write out each defrule. */
   /*=========================*/
   
   rPtr = (struct defrule *) GetNextDefrule(NULL);
   while (rPtr != NULL)
     {
      tPtr = rPtr;
      first = CLIPS_TRUE;
      while (tPtr != NULL)
        {
         tempDefrule.name = (SYMBOL_HN *) tPtr->name->bucket;
         tempDefrule.pp_form = NULL;
         tempDefrule.salience = tPtr->salience;
         tempDefrule.complexity = tPtr->complexity;
         tempDefrule.afterBreakpoint = 0;
         if ((tPtr->dynamicSalience != NULL) && (first))
           { 
            tempDefrule.dynamicSalience = (struct expr *) ExpressionCount;
            ExpressionCount += ExpressionSize(tPtr->dynamicSalience); 
           }
         else
           { tempDefrule.dynamicSalience = (struct expr *) -1L; }
           
         if (tPtr->actions != NULL)
           { 
            tempDefrule.actions = (struct expr *) ExpressionCount;
            ExpressionCount += ExpressionSize(tPtr->actions); 
           }
         else
           { tempDefrule.actions = (struct expr *) -1L; }
      
         tempDefrule.logicalJoin = (struct joinNode *) FindJoinPointer(tPtr->logicalJoin);
         
         if (tPtr->pats != NULL) tempDefrule.pats = NULL;
         else tempDefrule.pats = (struct patptr *) -1L;
      
         if (tPtr->disjunct != NULL) tempDefrule.disjunct = NULL;
         else tempDefrule.disjunct = (struct defrule *) -1L;
         
         if (tPtr->next != NULL) tempDefrule.next = NULL;
         else tempDefrule.next = (struct defrule *) -1L;
      
         GenWrite(&tempDefrule,(unsigned long) sizeof(struct defrule),fp);
         first = CLIPS_FALSE;
         tPtr = tPtr->disjunct;
        }
         
      rPtr = (struct defrule *) GetNextDefrule(rPtr);
     }
     
   /*=============================*/
   /* Write out pattern pointers. */
   /*=============================*/
   
   rPtr = (struct defrule *) GetNextDefrule(NULL);
   while (rPtr != NULL)
     {
      tPtr = rPtr;
      while (tPtr != NULL)
        {
         pjPtr = tPtr->pats;
         while (pjPtr != NULL)
           {
            tempPatPtr.pptr = (struct patternNode *) FindPatternPointer(pjPtr->pptr);
            tempPatPtr.jptr = (struct joinNode *) FindJoinPointer(pjPtr->jptr);
         
            if (pjPtr->next != NULL) tempPatPtr.next = NULL;
            else tempPatPtr.next = (struct patptr *) -1L;
         
            GenWrite(&tempPatPtr,(unsigned long) sizeof(struct patptr),fp);
            pjPtr = pjPtr->next;
           }
         tPtr = tPtr->disjunct;
        }
           
      rPtr = (struct defrule *) GetNextDefrule(rPtr);
     }

   /*=============================*/
   /* Write out the Rete Network. */
   /*=============================*/
   
   BsavePatterns(fp);
   BsaveJoins(fp);
  }
  
/************************************************************************/
/* BsavePatterns: Saves the pattern node structures to the binary file. */
/************************************************************************/
static VOID BsavePatterns(fp)
  FILE *fp;
  {
   struct defrule *rulePtr;
   struct patptr *ppPtr;
   struct patternNode *pnPtr;
   struct patternNode tempNode;
   
   rulePtr = (struct defrule *) GetNextDefrule(NULL);
   
   while (rulePtr != NULL)
     {
      ppPtr = rulePtr->pats;
      while (ppPtr != NULL)
        {
         pnPtr = ppPtr->pptr;
         while (pnPtr != NULL)
           {
            if (pnPtr->marked)
              { 
               pnPtr->marked = 0;
               tempNode.singlefieldNode = pnPtr->singlefieldNode;
               tempNode.multifieldNode = pnPtr->multifieldNode;
               tempNode.stopNode = pnPtr->stopNode;
               tempNode.blocked = 0;
               tempNode.initialize = 0;
               tempNode.alpha = NULL;
               if (pnPtr->eval != NULL)
                 { 
                  tempNode.eval = (struct expr *) ExpressionCount;
                  ExpressionCount += ExpressionSize(pnPtr->eval); 
                 }
               else
                 { tempNode.eval = (struct expr *) -1L; }
        
               tempNode.entry_join = (struct joinNode *) FindJoinPointer(pnPtr->entry_join);
               tempNode.next_level = (struct patternNode *) 
                            FindPatternPointer(pnPtr->next_level);
               tempNode.right_node = (struct patternNode *) 
                            FindPatternPointer(pnPtr->right_node);
               tempNode.left_node = (struct patternNode *) 
                      FindPatternPointer(pnPtr->left_node);
               tempNode.last_level = (struct patternNode *) 
                            FindPatternPointer(pnPtr->last_level);
        
               GenWrite(&tempNode,(unsigned long) sizeof(struct patternNode),fp);
               
               pnPtr = pnPtr->last_level;
              }
            else
              { pnPtr = NULL; }
           }
         
         ppPtr = ppPtr->next;
        }
     
     if (rulePtr->disjunct != NULL) rulePtr = rulePtr->disjunct;
     else rulePtr = rulePtr->next;
    }
  }
  
/*************************************************************/
/* BsaveJoins: Saves the join structures to the binary file. */
/*************************************************************/
static VOID BsaveJoins(fp)
  FILE *fp;
  {
   struct defrule *rulePtr;
   struct patptr *ppPtr;
   struct joinNode *joinPtr;
   struct joinNode tempJoin;
   
   rulePtr = (struct defrule *) GetNextDefrule(NULL);
   
   while (rulePtr != NULL)
     {
      ppPtr = rulePtr->pats;
      while (ppPtr != NULL)
        {
         joinPtr = ppPtr->jptr;
         if (joinPtr->marked)
           { 
            joinPtr->marked = 0;
            genmemcpy(&tempJoin,joinPtr,(unsigned long) sizeof(struct joinNode));
            tempJoin.beta = NULL;
            tempJoin.id = 0;
            tempJoin.initialize = 0;
            tempJoin.entry_pat = (struct patternNode *)
                              FindPatternPointer(joinPtr->entry_pat);
            tempJoin.last_level = (struct joinNode *)
                         FindJoinPointer(joinPtr->last_level);
            tempJoin.next_level = (struct joinNode *)
                         FindJoinPointer(joinPtr->next_level);
            tempJoin.right_match_node = (struct joinNode *)
                         FindJoinPointer(joinPtr->right_match_node);
            tempJoin.right_drive_node = (struct joinNode *)
                         FindJoinPointer(joinPtr->right_drive_node);
            if (joinPtr->eval != NULL)
              { 
               if (joinPtr->lastJoin == CLIPS_FALSE)
                 {
                  tempJoin.eval = (struct expr *) ExpressionCount;
                  ExpressionCount += ExpressionSize(joinPtr->eval);
                 }
               else
                 {
                  tempJoin.eval = (struct expr *) 
                            GetDisjunctIndex((struct defrule *) joinPtr->eval);
                 } 
              }
            else
              { tempJoin.eval = (struct expr *) -1L; }
           
            if (joinPtr->not_eval != NULL)
              { 
               tempJoin.not_eval = (struct expr *) ExpressionCount;
               ExpressionCount += ExpressionSize(joinPtr->not_eval); 
              }
            else
              { tempJoin.not_eval = (struct expr *) -1L; }
   
            GenWrite(&tempJoin,(unsigned long) sizeof(struct joinNode),fp);
           }
         
         ppPtr = ppPtr->next;
        }
     
     if (rulePtr->disjunct != NULL) rulePtr = rulePtr->disjunct;
     else rulePtr = rulePtr->next;
    }
  }
 
/***************************************************************************/
/* FindPatternItems: Counts the number of expression structures needed     */
/*   to store all expressions in the pattern network. Keeps track of which */
/*   functions are needed to evaluate these expressions. Call routines to  */
/*   perform the same function for the join network.                       */
/***************************************************************************/
static VOID FindPatternItems(pat_ptr)
  struct patternNode *pat_ptr;
  {
   while (pat_ptr != NULL)
     {
      if (pat_ptr->eval != NULL)
        { 
         ExpressionCount += ExpressionSize(pat_ptr->eval);
         MarkNeededItems(pat_ptr->eval); 
        }
        
      if (pat_ptr->next_level == NULL)
        { FindJoinItems(pat_ptr->entry_join,CLIPS_FALSE); }
      else
        { FindPatternItems(pat_ptr->next_level); }
        
      pat_ptr = pat_ptr->right_node;
     }
  }
  
/************************************************************************/
/* FindJoinItems: Counts the number of expression structures needed     */
/*   to store all expressions in the join network. Keeps track of which */
/*   functions are needed to evaluate these expressions.                */
/************************************************************************/
static VOID FindJoinItems(joinPtr,flag)
  struct joinNode *joinPtr;
  int flag;
  {
   while (joinPtr != NULL)
     {
      if (joinPtr->firstJoin || flag)
        {
         if ((joinPtr->eval != NULL) && (joinPtr->lastJoin == CLIPS_FALSE))
           { 
            ExpressionCount += ExpressionSize(joinPtr->eval);
            MarkNeededItems(joinPtr->eval); 
           }
           
         if (joinPtr->not_eval != NULL)
           { 
            ExpressionCount += ExpressionSize(joinPtr->not_eval);
            MarkNeededItems(joinPtr->not_eval); 
           }
           
         FindJoinItems(joinPtr->next_level,CLIPS_TRUE);
        }
        
      if (flag) joinPtr = joinPtr->right_drive_node;
      else joinPtr = joinPtr->right_match_node;
     }
  }
#endif  

#if (BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE)
/*******************************************************/
/* BloadDefrules: Loads in the join and pattern network */
/*   needed for this binary image.                     */
/*******************************************************/
static VOID BloadDefrules()
  {
   unsigned long space, i, j; 
   long int index;  
   unsigned long int lastDefrule;
   int first;
   unsigned long int topPattern;
   
   /*====================================================*/
   /* Determine the number of structures in the network. */
   /*====================================================*/
  
   GenRead(&space,(unsigned long) sizeof(unsigned long int));
   GenRead(&NumberOfDefrules,(unsigned long) sizeof(unsigned long int));
   GenRead(&NumberOfPatternPointers,(unsigned long) sizeof(unsigned long int));
   GenRead(&NumberOfPatterns,(unsigned long) sizeof(long int));
   GenRead(&NumberOfJoins,(unsigned long) sizeof(long int));
   GenRead(&topPattern,(unsigned long) sizeof(long int));
   
   if (NumberOfDefrules == 0)
     {
      DefruleArray = NULL;
      PatPtrArray = NULL;
      PatternArray = NULL;
      JoinArray = NULL;
      return;
     }
         
   /*===================*/
   /* Load the network. */
   /*===================*/
     
   space = NumberOfDefrules * sizeof(struct defrule);
   DefruleArray = (struct defrule *) genlongalloc(space);
   GenRead((VOID *) DefruleArray,space);
   
   space = NumberOfPatternPointers * sizeof(struct patptr);
   PatPtrArray = (struct patptr *) genlongalloc(space);
   GenRead((VOID *) PatPtrArray,space);
   
   space = NumberOfPatterns * sizeof(struct patternNode);
   PatternArray = (struct patternNode *) genlongalloc(space);
   GenRead((VOID *) PatternArray,space);
   
   space = NumberOfJoins * sizeof(struct joinNode);
   JoinArray = (struct joinNode *) genlongalloc(space);
   GenRead((VOID *) JoinArray,space);
   
   /*======================================*/
   /* Reset defrule and patptr structures. */
   /*======================================*/
   
   first = CLIPS_TRUE;
   lastDefrule = -1;
   j = 0;
   for (i = 0; i < NumberOfDefrules; i++)
     {
      if ((lastDefrule != -1) && (first == CLIPS_TRUE))
        { 
         while (lastDefrule != i)
           { 
            DefruleArray[lastDefrule].next = (struct defrule *) &DefruleArray[i]; 
            lastDefrule++;
           }
        }
        
      if (first == CLIPS_TRUE) lastDefrule = i;
      
      if (first == CLIPS_TRUE)
        {
         DefruleArray[i].name = SymbolPointer(DefruleArray[i].name);
         IncrementCount(DefruleArray[i].name);
        }
      else
        { DefruleArray[i].name = DefruleArray[lastDefrule].name; }
  
      if (first == CLIPS_TRUE)
        {
         DefruleArray[i].dynamicSalience = 
                  ExpressionPointer(DefruleArray[i].dynamicSalience);
        }
      else
        { DefruleArray[i].dynamicSalience = DefruleArray[lastDefrule].dynamicSalience; }
        
      DefruleArray[i].actions = ExpressionPointer(DefruleArray[i].actions);

      DefruleArray[i].logicalJoin = JoinPointer(JoinArray,DefruleArray[i].logicalJoin);
      
      if (((long int) DefruleArray[i].disjunct) == -1L)
        {
         DefruleArray[i].disjunct = NULL;
         DefruleArray[i].next = NULL;
         first = CLIPS_TRUE;
        }
      else
        { 
         DefruleArray[i].disjunct = (struct defrule *) &DefruleArray[i + 1];
         DefruleArray[i].next = NULL;
         first = CLIPS_FALSE; 
         }
        
      DefruleArray[i].pats = (struct patptr *) &PatPtrArray[j];
        
      do 
        {
         index = (long int) PatPtrArray[j].next;
         
         if (index != -1L) 
           PatPtrArray[j].next = (struct patptr *) &PatPtrArray[j + 1];
         else PatPtrArray[j].next = NULL;
         
         PatPtrArray[j].pptr = PatternPointer(PatternArray,PatPtrArray[j].pptr);
         PatPtrArray[j].jptr = JoinPointer(JoinArray,PatPtrArray[j].jptr);
         j++;
        }
      while (index != -1L);
     }
   
   /*=================================*/
   /* Reset pattern network pointers. */
   /*=================================*/
   
   for (i = 0; i < NumberOfPatterns; i++)
     {
      PatternArray[i].right_node = PatternPointer(PatternArray,PatternArray[i].right_node);
      PatternArray[i].next_level = PatternPointer(PatternArray,PatternArray[i].next_level);
      PatternArray[i].last_level = PatternPointer(PatternArray,PatternArray[i].last_level);
      PatternArray[i].left_node  = PatternPointer(PatternArray,PatternArray[i].left_node);
      PatternArray[i].eval       = ExpressionPointer(PatternArray[i].eval);
      PatternArray[i].entry_join = JoinPointer(JoinArray,PatternArray[i].entry_join);
      PatternArray[i].alpha      = NULL;
     }
     
   /*==============================*/
   /* Reset join network pointers. */
   /*==============================*/
   
   for (i = 0; i < NumberOfJoins; i++)
     {
      if (JoinArray[i].lastJoin == CLIPS_TRUE)
        {
         /* This is a pointer to a defrule structure. */
         JoinArray[i].eval = (struct expr *) 
                             (DefruleArray + ((long int) JoinArray[i].eval)); 
        }
      else
        { JoinArray[i].eval = ExpressionPointer(JoinArray[i].eval); }
        
      JoinArray[i].not_eval = ExpressionPointer(JoinArray[i].not_eval);
      JoinArray[i].next_level = JoinPointer(JoinArray,
                                            JoinArray[i].next_level);
      JoinArray[i].last_level = JoinPointer(JoinArray,
                                            JoinArray[i].last_level);
      JoinArray[i].entry_pat = PatternPointer(PatternArray,
                                              JoinArray[i].entry_pat);
      JoinArray[i].right_drive_node = JoinPointer(JoinArray,
                                                  JoinArray[i].right_drive_node);
      JoinArray[i].right_match_node = JoinPointer(JoinArray,
                                                  JoinArray[i].right_match_node);
     }

   /*=============================================*/
   /* Set the network pointer to the new network. */
   /*=============================================*/
   
   SetListOfDefrules((VOID *) DefruleArray);
   SetNetworkPointer((VOID *) PatternPointer(PatternArray,topPattern));
  }
  
/****************************************************************/
/* ClearBload:                                                  */
/****************************************************************/
static VOID ClearBload()
  {
   unsigned long int space, i;
   struct fact *theFact;
   struct patternMatch *theMatch, *tempMatch;
   
   /*=========================================*/
   /* Remove all activations from the agenda. */
   /*=========================================*/
   
   RemoveAllActivations();
   
   /*================================*/
   /* Remove all partial matches in  */
   /* the join and pattern networks. */
   /*================================*/
   
   for (i = 0; i < NumberOfPatterns; i++)
     { FlushAlphaBetaMemory(PatternArray[i].alpha); }
   
   for (i = 0; i < NumberOfJoins; i++)
     { FlushAlphaBetaMemory(JoinArray[i].beta); }
   
   /*====================================*/
   /* Return all of the match structures */
   /* allocated with the facts.          */
   /*====================================*/
   
   theFact = (struct fact *) GetNextFact(NULL);
   while (theFact != NULL)
     {
      theMatch = (struct patternMatch *) theFact->list;
      theFact->list = NULL;
      while (theMatch != NULL)
        {
         tempMatch = theMatch->next;
         rtn_struct(patternMatch,theMatch);
         theMatch = tempMatch;
        }
        
      theFact = (struct fact *) GetNextFact((VOID *) theFact);
     }
       
   /*================================================*/
   /* Decrement the symbol count for each rule name. */
   /*================================================*/
    
   for (i = 0; i < NumberOfDefrules; i++)
     { 
      if (DefruleArray[i].disjunct == NULL)
        { DecrementSymbolCount(DefruleArray[i].name); }
     }
     
   /*==================================================*/
   /* Return the space allocated for the bload arrays. */
   /*==================================================*/
   
   space = NumberOfDefrules * sizeof(struct defrule);
   if (space != 0) genlongfree((VOID *) DefruleArray,space); 
   
   space = NumberOfPatternPointers * sizeof(struct patptr);
   if (space != 0) genlongfree((VOID *) PatPtrArray,space); 
   
   space = NumberOfPatterns * sizeof(struct patternNode);
   if (space != 0) genlongfree((VOID *) PatternArray,space); 
   
   space = NumberOfJoins * sizeof(struct joinNode);
   if (space != 0) genlongfree((VOID *) JoinArray,space);
   
   /*====================================*/
   /* Set defrule pointer lists to NULL. */
   /*====================================*/
   
   SetListOfDefrules(NULL);
   SetNetworkPointer(NULL);
  }  
  
/****************************************************************/
/* ClearBloadReady:                                              */
/****************************************************************/
static int ClearBloadReady()
  {
   if (ExecutingRule != NULL) return(CLIPS_FALSE);
   
   return(CLIPS_TRUE);
  }  

#endif

/****************************************************************/
/* DefruleBinarySetup:                                          */
/****************************************************************/
globle VOID DefruleBinarySetup()
  {
  
#if BLOAD_AND_BSAVE || BLOAD
   AddBeforeBloadFunction("defrule",ClearDefrules,0);
#endif

#if BLOAD_AND_BSAVE || BLOAD || BLOAD_ONLY
   AddAfterBloadFunction("defrule",ResetDeployedRuleImage,0);
   AddClearBloadReadyFunction("defrule",ClearBloadReady,0);
#endif
#if BLOAD_AND_BSAVE
   AddBinaryItem("defrule",0,BsaveFind,BsaveExpressions,BsaveDefrules,
                             BloadDefrules,NULL,ClearBload);
#endif
#if BLOAD || BLOAD_ONLY
   AddBinaryItem("defrule",0,NULL,NULL,NULL,BloadDefrules,NULL,ClearBload);
#endif

#if CONSTRUCT_COMPILER
   AddCodeGeneratorItem("defrules",10,NULL,InitDefrulesCode,DefrulesToCode);
#endif
  }
  
#endif


