   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                    ENGINE MODULE                    */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*      Bebe Ly                                              */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _ENGINE_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "setup.h"

#if DEFRULE_CONSTRUCT

#define _DEFRULE_STRUCTURES_

#include "constant.h"
#include "symbol.h"
#include "clipsmem.h"
#include "expressn.h"
#include "evaluatn.h"
#include "constrct.h"
#include "network.h"
#include "facts.h"
#include "match.h"
#include "drive.h"
#include "retract.h"
#include "defrule.h"
#include "router.h"
#include "utility.h"
#include "reteutil.h"
#include "sysdep.h"
#include "intrfile.h"
#include "engine.h"
#include "sysprime.h"

/*************************/
/* STRUCTURE DEFINITIONS */
/*************************/
  
struct runFunction
  {
   char *name;
#if ANSI_COMPILER
   VOID (*ip)(void);
#else
   VOID (*ip)();
#endif
   int priority;
   struct runFunction *next;
  };
  
/***************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/***************************************/

#if ANSI_COMPILER  
   static VOID                    PlaceActivation(struct activation *);
   static struct activation      *PlaceDepthActivation(struct activation *);
   static struct activation      *PlaceBreadthActivation(struct activation *);
   static struct activation      *PlaceLEXActivation(struct activation *);
   static struct activation      *PlaceMEAActivation(struct activation *);
   static struct activation      *PlaceComplexityActivation(struct activation *);
   static struct activation      *PlaceSimplicityActivation(struct activation *);
   static struct activation      *PlaceRandomActivation(struct activation *);
   static struct partialMatch    *SortBindings(struct partialMatch *); 
   static int                     CompareBindings(struct activation *,struct activation *);
#else
   static VOID                    PlaceActivation();
   static struct activation      *PlaceDepthActivation();
   static struct activation      *PlaceBreadthActivation();
   static struct activation      *PlaceLEXActivation();
   static struct activation      *PlaceMEAActivation();
   static struct activation      *PlaceComplexityActivation();
   static struct activation      *PlaceSimplicityActivation();
   static struct activation      *PlaceRandomActivation();
   static struct partialMatch    *SortBindings(); 
   static int                     CompareBindings();
#endif

/****************************************/
/* GLOBAL INTERNAL VARIABLE DEFINITIONS */
/****************************************/

   globle struct defrule      *ExecutingRule = NULL;
   globle BOOLEAN              HaltRules = CLIPS_FALSE;
   globle BOOLEAN              DeletedFiringRule = CLIPS_FALSE;
   globle struct joinNode     *TheLogicalJoin = NULL; 
   globle int                  AgendaChanged = CLIPS_FALSE;
   
/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/

   static struct activation   *Agenda = NULL;
   static char                *RuleFiring = NULL;   
   static struct runFunction  *ListOfRunFunctions = NULL; 
   static int                  Strategy = DEFAULT_STRATEGY;  
   static long int             NumberOfActivations = 0;
   static BOOLEAN              SalienceEvaluation = WHEN_DEFINED;
   static unsigned long int    CurrentTimetag = 0;
#if DEBUGGING_FUNCTIONS
   static int                  WatchActivations = OFF; 
   static int                  WatchStatistics = OFF;
#endif
   
/**************************************************************/
/* RunCLIPS: Begins execution of rules.  If run limit is less */
/*   than  zero, then rules will be executed until the agenda */
/*   is empty.  If run limit is greater than zero, then rules */
/*   will be executed until either the agenda is empty, the   */
/*   run limit has been reached, or a rule execution error    */
/*   has occurred.  Returns the number of rules fired.        */
/**************************************************************/
globle long int RunCLIPS(runLimit)
  long int runLimit;
  {
   struct expr *exprPtr;
   struct activation *ruleToFire;
   DATA_OBJECT_PTR oldBindList;
   long int rulesFired = 0;
   DATA_OBJECT result;
   struct runFunction *exec_ptr;
#if DEBUGGING_FUNCTIONS
   long int maxActivations = 0, sumActivations = 0, maxFacts = 0, sumFacts = 0;
   double end_time = 0.0, start_time = 0.0;   
   char print_space[60];
   long int tempValue;
#endif
   unsigned int i,svBreakContext,svReturnContext;
   static int alreadyRunning = CLIPS_FALSE;
   
   /*=====================================================*/
   /* Make sure the run command is not already executing. */
   /*=====================================================*/
   
   if (alreadyRunning) return(0);
   alreadyRunning = CLIPS_TRUE;
   
   /*================================*/
   /* Set up statistics information. */
   /*================================*/

#if DEBUGGING_FUNCTIONS   
   if (WatchStatistics)
     { 
      maxFacts = GetNumberOfFacts();
      sumFacts = maxFacts;
      maxActivations = GetNumberOfActivations();
      sumActivations = maxActivations;
      start_time = gentime();
     }
#endif
     
   /*=============================*/
   /* Set up execution variables. */
   /*=============================*/
   
   if (CurrentEvaluationDepth == 0) SetHaltExecution(CLIPS_FALSE);
   HaltRules = CLIPS_FALSE;
   oldBindList = BindList;
   BindList = NULL;
   svBreakContext = BreakContext;
   svReturnContext = ReturnContext;
   BreakContext = ReturnContext = CLIPS_FALSE;
   
   /*=====================================================*/
   /* Fire rules until the agenda is empty, the run limit */
   /* has been reached, or a rule execution error occurs. */
   /*=====================================================*/
   
   while ((Agenda != NULL) && 
          (runLimit != 0) &&
          (HaltExecution == CLIPS_FALSE) &&
          (HaltRules == CLIPS_FALSE))
     {
        
      /*==========================*/
      /* Bookkeeping and Tracing. */
      /*==========================*/
       
      RuleFiring = ValueToString(Agenda->theRule->name);
      rulesFired++;
      if (runLimit > 0) { runLimit--; }

#if DEBUGGING_FUNCTIONS   
      if (GetRulesWatch() == ON)
        { 
         sprintf(print_space,"FIRE %4ld ",rulesFired);
         PrintCLIPS(WTRACE,print_space);
         PrintCLIPS(WTRACE,RuleFiring);
         PrintCLIPS(WTRACE,": ");
         PrintPartialMatch(WTRACE,Agenda->basis);
         PrintCLIPS(WTRACE,"\n");
        }

      if (CRSVTraceActive())
        {
         sprintf(print_space,"F    %-4ld   ",rulesFired);
         PrintCLIPS("wcrsv_tr",print_space);
         PrintCLIPS("wcrsv_tr",RuleFiring);
         PrintCLIPS("wcrsv_tr","\n");
        }
#endif
   
      /*=====================================================*/
      /* Detach the activation from the Agenda. Remember the */
      /* currently executing rule (so the deletion routines  */
      /* won't delete the actions in the event the rule is   */
      /* deleted).                                           */
      /*=====================================================*/
    
      AgendaChanged = CLIPS_TRUE;
      ExecutingRule = Agenda->theRule;
      ruleToFire = Agenda;  
      Agenda = Agenda->next;     
      if (Agenda != NULL) Agenda->prev = NULL;
      
      /*=================================================*/
      /* Remove the link between the activation and the  */
      /* completed match for the rule. Set the busy flag */
      /* for the completed match to TRUE (so the match   */
      /* upon which our RHS variables are dependent is   */
      /* not deleted while our rule is firing). Set up   */
      /* the global pointers to the completed match for  */
      /* routines which do variable extractions.         */
      /*=================================================*/
      
      ruleToFire->basis->binds[ruleToFire->basis->bcount] = NULL;
      ruleToFire->basis->busy = CLIPS_TRUE;

      GlobalLHSBinds = ruleToFire->basis;
      GlobalRHSBinds = NULL;
      
      /*===========================================================*/
      /* Increment the count for each of the facts associated with */
      /* the rule activation so that the facts cannot be deleted   */
      /* by garbage collection while the rule is executing.        */
      /*===========================================================*/
      
      for (i = 0; i < ruleToFire->basis->bcount; i++)
        {
         if (ruleToFire->basis->binds[i]->whoset >= 0)
           { ruleToFire->basis->binds[i]->origin->count++; }
        } 
      
      /*====================================================*/
      /* Execute the rule's right hand side actions. If the */
      /* rule has logical patterns, set up the pointer to   */
      /* the rules logical join so the assert command will  */
      /* attach the appropriate dependencies to the facts.  */
      /*====================================================*/
      
      TheLogicalJoin = ruleToFire->theRule->logicalJoin;
      CurrentEvaluationDepth++;
      exprPtr = ruleToFire->theRule->actions->arg_list;
      SetEvaluationError(CLIPS_FALSE);
      while ((exprPtr != NULL) && (HaltExecution == CLIPS_FALSE))
        {
         EvaluateExpression(exprPtr,&result);
         exprPtr = exprPtr->next_arg;
        }
      SetEvaluationError(CLIPS_FALSE);
      CurrentEvaluationDepth--;
      TheLogicalJoin = NULL;

#if DEBUGGING_FUNCTIONS
      if ((HaltExecution) || (HaltRules && (GetRulesWatch() == ON)))
#else
      if ((HaltExecution) || (HaltRules))                    
#endif
          
        {
         PrintCLIPS(WERROR,"\nExecution halted during the RHS actions\n");
         PrintCLIPS(WERROR,"   of rule ");
         PrintCLIPS(WERROR,RuleFiring);
         PrintCLIPS(WERROR,"\n");
        }
        
      ruleToFire->basis->busy = CLIPS_FALSE;
     
      /*===========================================*/
      /* Decrement the count for each of the facts */
      /* associated with the rule activation.      */
      /*===========================================*/
      
      for (i = 0; i < ruleToFire->basis->bcount; i++)
        {
         if (ruleToFire->basis->binds[i]->whoset >= 0)
           { ruleToFire->basis->binds[i]->origin->count--; }
        } 
      
      /*==================================================*/
      /* If the rule was deleted while firing, return its */
      /* corresponding defrule structure.                 */
      /*==================================================*/

#if ! RUN_TIME      
      if (DeletedFiringRule)
        {
         if (ExecutingRule != NULL)
           {
            struct defrule *temp;
            
            temp = ExecutingRule;
            ExecutingRule = NULL; /* Must be NULL for next call to work. */
            ReturnDefrule(temp);
           }
           
         DeletedFiringRule = CLIPS_FALSE;
        }
#endif
        
      /*========================================*/
      /* Return the agenda node to free memory. */
      /*========================================*/
      
      RemoveActivation(ruleToFire,CLIPS_FALSE,CLIPS_FALSE);

      /*============================================*/
      /* Remove retracted facts, ephemeral symbols, */
      /* variable bindings, and temporary segments. */
      /*============================================*/
       
      FlushGarbagePartialMatches();
      RemoveOldFacts();
      PeriodicCleanup(CLIPS_FALSE,CLIPS_TRUE); 
      if (BindList != NULL) 
        {
         ReturnValues(BindList);
         BindList = NULL;
        }
      
      /*==========================*/
      /* Keep up with statistics. */
      /*==========================*/

#if DEBUGGING_FUNCTIONS         
      if (WatchStatistics)
        { 
         tempValue = GetNumberOfFacts();
         if (tempValue > maxFacts) maxFacts = tempValue;
         sumFacts += tempValue;
         tempValue = GetNumberOfActivations();
         if (tempValue > maxActivations) maxActivations = tempValue;
         sumActivations += tempValue;
        }
#endif

      /*==================================*/
      /* Update saliences if appropriate. */
      /*==================================*/
      
      if (SalienceEvaluation == EVERY_CYCLE) RefreshAgenda();
      
      /*=============================================*/
      /* Execute exec list after performing actions. */
      /*=============================================*/
      
      exec_ptr = ListOfRunFunctions;
      while (exec_ptr != NULL)
        {
         (*exec_ptr->ip)();
         exec_ptr = exec_ptr->next;
        }
      
      /*========================*/
      /* Check for breakpoints. */
      /*========================*/
        
      if (Agenda != NULL)
        { 
         if (Agenda->theRule->afterBreakpoint) 
           {
            HaltRules = CLIPS_TRUE;
            PrintCLIPS(WDIALOG,"Breaking on rule ");
            PrintCLIPS(WDIALOG,ValueToString(Agenda->theRule->name));
            PrintCLIPS(WDIALOG,"\n"); 
           }
        }
     }
   
   if (runLimit == rulesFired)
     { PrintCLIPS(WDIALOG,"rule firing limit reached\n"); }
     
   /*==============================*/
   /* Restore execution variables. */
   /*==============================*/
   
   BreakContext = svBreakContext;
   ReturnContext = svReturnContext;
   ExecutingRule = NULL;
   HaltRules = CLIPS_FALSE;
   BindList = oldBindList;
   RuleFiring = NULL;
   
   /*=================================================*/
   /* Print out statistics if they are being watched. */
   /*=================================================*/

#if DEBUGGING_FUNCTIONS         
   if (WatchStatistics)
     {
      char print_space[60];
      
      end_time = gentime();

      PrintLongInteger(WDIALOG,rulesFired);
      PrintCLIPS(WDIALOG," rules fired");

      if (start_time != end_time)
        { 
         PrintCLIPS(WDIALOG,"        Run time is ");
         PrintFloat(WDIALOG,end_time - start_time);
         PrintCLIPS(WDIALOG," seconds\n");
         PrintFloat(WDIALOG,(double) rulesFired / (end_time - start_time));
         PrintCLIPS(WDIALOG," rules per second\n");
        }  
      else
        { PrintCLIPS(WDIALOG,"\n"); }
        
      sprintf(print_space,"%ld mean number of facts (%ld maximum)\n",
                          (long) (((double) sumFacts / (rulesFired + 1)) + 0.5), 
                          maxFacts);
      PrintCLIPS(WDIALOG,print_space);
      sprintf(print_space,"%ld mean number of activations (%ld maximum)\n",
                          (long) (((double) sumActivations / (rulesFired + 1)) + 0.5), 
                          maxActivations);
      PrintCLIPS(WDIALOG,print_space);
     }
#endif
     
   /*===================================*/
   /* Return the number of rules fired. */
   /*===================================*/
   
   alreadyRunning = CLIPS_FALSE;
   return(rulesFired);
  }

/***************************************************************/
/* AddActivation: Adds a rule activation to the agenda.  This */
/*   function is called when all the patterns on the LHS of a  */
/*   rule have been satisfied.                                 */
/***************************************************************/
globle VOID AddActivation(join,binds)
  struct joinNode *join;
  struct partialMatch *binds;
  {
   struct activation *new_act;

   /*======================================================*/
   /* Create the activation.  Store the join connected to  */ 
   /* the actions in the rhs and the list of fact bindings */
   /* in the lhs.                                          */
   /*======================================================*/

   new_act = get_struct(activation);
   new_act->theRule = (struct defrule *) join->eval;
   new_act->basis = binds;
   new_act->sortedBasis = NULL;
   new_act->salience = EvaluateSalience(join->eval);
   new_act->randomID = genrand();
   new_act->timetag = CurrentTimetag++;
   new_act->prev = NULL;
   new_act->next = NULL;
   
   NumberOfActivations++;
   
   /*=======================================================*/
   /* Point the partial match to the activation to complete */
   /* the link between the join network and the agenda.     */
   /*=======================================================*/
   
   binds->binds[binds->bcount] = (struct alphaMatch *) new_act;

   /*====================================================*/
   /* If activations are being watch, display a message. */
   /*====================================================*/

#if DEBUGGING_FUNCTIONS     
   if (WatchActivations == ON)
     { 
      PrintCLIPS(WTRACE,"==> Activation ");
      PrintActivation(WTRACE,new_act); 
      PrintCLIPS(WTRACE,"\n");
     }
       
   if (CRSVTraceActive())
     {
      PrintCLIPS("wcrsv_tr","AC   ");
      PrintCRSVActivation("wcrsv_tr",new_act);
      PrintCLIPS("wcrsv_tr","\n");
     }
#endif

    /*=====================================*/
    /* Place the activation in the agenda. */
    /*=====================================*/
    
    PlaceActivation(new_act);
   }
   
/****************************************************/
/* PlaceActivation:                          */
/****************************************************/
static VOID PlaceActivation(new_act)
  struct activation *new_act;
  {
   struct activation *place_after = NULL;
   
   AgendaChanged = CLIPS_TRUE;

   /*=============================================*/
   /* Determine the location where the activation */
   /* should be placed in the agenda.             */
   /*==============================================*/

   if (Agenda != NULL) switch (Strategy)
     {
      case DEPTH_STRATEGY:
        place_after = PlaceDepthActivation(new_act);
        break;
        
      case BREADTH_STRATEGY:
        place_after = PlaceBreadthActivation(new_act);
        break;
        
      case LEX_STRATEGY:
        place_after = PlaceLEXActivation(new_act);
        break;
        
      case MEA_STRATEGY:
        place_after = PlaceMEAActivation(new_act);
        break;
        
      case COMPLEXITY_STRATEGY:
        place_after = PlaceComplexityActivation(new_act);
        break;
           
      case SIMPLICITY_STRATEGY:
        place_after = PlaceSimplicityActivation(new_act);
        break;
           
      case RANDOM_STRATEGY:
        place_after = PlaceRandomActivation(new_act);
        break;
     }
        
   /*==============================================================*/
   /* Place the activation at the appropriate place in the agenda. */
   /*==============================================================*/
   
   if (place_after == NULL)               
     {
      new_act->next = Agenda; 
      Agenda = new_act;
      if (new_act->next != NULL) new_act->next->prev = new_act;
     }
   else           
     {
      new_act->next = place_after->next;
      new_act->prev = place_after;
      place_after->next = new_act; 
      if (new_act->next != NULL)
        { new_act->next->prev = new_act; }
     }
  }
  
/*********************************************************************/
/* PlaceDepthActivation: Determines the location in the agenda where */
/*    a new activation should be placed for the depth strategy.      */
/*********************************************************************/
static struct activation *PlaceDepthActivation(new_act)
  struct activation *new_act;
  {
   int salience;
   unsigned long timetag;
   struct activation *actPtr, *last_act;
       
   salience = new_act->salience;
   timetag = new_act->timetag;
   actPtr = (struct activation *) GetNextActivation(NULL); 
   last_act = NULL;

   while (actPtr != NULL)
     {
      if (actPtr->salience > salience)
        {
         last_act = actPtr;
         actPtr = actPtr->next;
        }
      else if (actPtr->salience < salience)
        { return(last_act); }
      else if (timetag < actPtr->timetag)
        {
         last_act = actPtr;
         actPtr = actPtr->next;
        }
      else
        { return(last_act); }
     }

   return(last_act);
  }
  
/***********************************************************************/
/* PlaceBreadthActivation: Determines the location in the agenda where */
/*    a new activation should be placed for the breadth strategy.      */
/***********************************************************************/
static struct activation *PlaceBreadthActivation(new_act)
  struct activation *new_act;
  {
   int salience;
   unsigned long timetag;
   struct activation *actPtr, *last_act;
       
   salience = new_act->salience;
   timetag = new_act->timetag;
   actPtr = (struct activation *) GetNextActivation(NULL); 
   last_act = NULL;

   while (actPtr != NULL)
     {
      if (actPtr->salience > salience)
        {
         last_act = actPtr;
         actPtr = actPtr->next;
        }
      else if (actPtr->salience < salience)
        { return(last_act); }
      else if (timetag > actPtr->timetag)
        {
         last_act = actPtr;
         actPtr = actPtr->next;
        }
      else
       {  return(last_act); }
     }

   return(last_act);
  }
  
/********************************************************************/
/* PlaceLEXActivation:  Determines the location in the agenda where */
/*    a new activation should be placed for the LEX strategy.       */ 
/********************************************************************/
static struct activation *PlaceLEXActivation(new_act)
  struct activation *new_act;
  {
   int salience;
   unsigned long timetag;
   struct activation *actPtr, *last_act;
   int flag;
   
   /*===============================================*/
   /* Sort the fact identifiers for the activation. */
   /*===============================================*/
   
   if (new_act->sortedBasis == NULL)
     { new_act->sortedBasis = SortBindings(new_act->basis); }
         
   timetag = new_act->timetag;
   salience = new_act->salience;
   actPtr = (struct activation *) GetNextActivation(NULL); 
   last_act = NULL;
   flag = CLIPS_FALSE;

   while (actPtr != NULL)
     {
      if (actPtr->salience > salience)
        {
         last_act = actPtr;
         actPtr = actPtr->next;
        }
      else if (actPtr->salience < salience)
        { return(last_act); }
      else
        {     
         flag = CompareBindings(actPtr,new_act);
         
         if (flag == LESS_THAN)
           {
            last_act = actPtr;
            actPtr = actPtr->next;
           }
         else if (flag == GREATER_THAN)
           { return(last_act); }
         else /* flag == EQUAL */
           {
            if (timetag > actPtr->timetag)
              {
               last_act = actPtr;
               actPtr = actPtr->next;
              }
            else
             { return(last_act); }
           }
        }
     }

   return(last_act);
  }
     
/*******************************************************************/
/* PlaceMEAActivation: Determines the location in the agenda where */
/*    a new activation should be placed for the MEA strategy.      */
/*******************************************************************/
static struct activation *PlaceMEAActivation(new_act)
  struct activation *new_act;
  {
   int salience;
   unsigned long timetag;
   struct activation *actPtr, *last_act;
   int flag;
   long int cWhoset, oWhoset;
      
   if (new_act->sortedBasis == NULL)
     { new_act->sortedBasis = SortBindings(new_act->basis); }
    
   timetag = new_act->timetag;
   salience = new_act->salience;
   actPtr = (struct activation *) GetNextActivation(NULL); 
   last_act = NULL;
   flag = CLIPS_FALSE;

   while (actPtr != NULL)
     {
      if (actPtr->salience > salience)
        {
         last_act = actPtr;
         actPtr = actPtr->next;
        }
      else if (actPtr->salience < salience)
        { return(last_act); }
      else
        {
         cWhoset = new_act->basis->binds[0]->whoset;
         oWhoset = actPtr->basis->binds[0]->whoset;
         if (oWhoset < cWhoset)
           { 
            if (cWhoset > 0) flag = GREATER_THAN; 
            else flag = LESS_THAN;
           }
         else if (oWhoset > cWhoset)
           { 
            if (oWhoset > 0) flag = LESS_THAN; 
            else flag = GREATER_THAN;
           }
         else
           { flag = CompareBindings(actPtr,new_act); }
         
         if (flag == LESS_THAN)
           {
            last_act = actPtr;
            actPtr = actPtr->next;
           }
         else if (flag == GREATER_THAN)
           { return(last_act); }
         else /* flag == EQUAL */
           {
            if (timetag > actPtr->timetag)
              {
               last_act = actPtr;
               actPtr = actPtr->next;
              }
            else
             { return(last_act); }
           }
        }
     }

   return(last_act);
  }
 
/*********************************************************************/
/* PlaceComplexityActivation: Determines the location in the agenda  */
/*    where a new activation should be placed for the complexity     */
/*    strategy. Returns a pointer to the activation  after which the */
/*    new activation should be placed (or NULL if the activation     */
/*    should be placed at the beginning of the agenda.               */
/*********************************************************************/
static struct activation *PlaceComplexityActivation(new_act)
  struct activation *new_act;
  {
   int salience, complexity;
   unsigned long timetag;
   struct activation *actPtr, *last_act;
       
   /*========================================*/
   /* Set up initial information for search. */
   /*========================================*/
   
   timetag = new_act->timetag;
   salience = new_act->salience;
   complexity = new_act->theRule->complexity;
   actPtr = (struct activation *) GetNextActivation(NULL); 
   last_act = NULL;

   /*====================================================*/
   /* Search through the agenda for the insertion point. */
   /*====================================================*/
   
   while (actPtr != NULL)
     {
      if (actPtr->salience > salience)
        {
         last_act = actPtr;
         actPtr = actPtr->next;
        }
      else if (actPtr->salience < salience)
        { return(last_act); }
      else if (complexity < actPtr->theRule->complexity)
        {
         last_act = actPtr;
         actPtr = actPtr->next;
        }
      else if (complexity > actPtr->theRule->complexity)
        { return(last_act); }
      else if (timetag > actPtr->timetag)
        {
         last_act = actPtr;
         actPtr = actPtr->next;
        }
      else
        { return(last_act); }
     }

   /*===========================================*/
   /* Return the insertion point in the agenda. */
   /*===========================================*/
   
   return(last_act);
  }
  
/*******************************************/
/* PlaceSimplicityActivation:              */
/*******************************************/
static struct activation *PlaceSimplicityActivation(new_act)
  struct activation *new_act;
  {
   int salience, complexity;
   unsigned long timetag;
   struct activation *actPtr, *last_act;
       
   timetag = new_act->timetag;
   salience = new_act->salience;
   complexity = new_act->theRule->complexity;
   actPtr = (struct activation *) GetNextActivation(NULL); 
   last_act = NULL;

   while (actPtr != NULL)
     {
      if (actPtr->salience > salience)
        {
         last_act = actPtr;
         actPtr = actPtr->next;
        }
      else if (actPtr->salience < salience)
        { return(last_act); }
      else if (complexity > actPtr->theRule->complexity)
        {
         last_act = actPtr;
         actPtr = actPtr->next;
        }
      else if (complexity < actPtr->theRule->complexity)
        { return(last_act); }
      else if (timetag > actPtr->timetag)
        {
         last_act = actPtr;
         actPtr = actPtr->next;
        }
      else
       { return(last_act); }
     }

   return(last_act);
  }
  
/*******************************************/
/* PlaceRandomActivation:                  */
/*******************************************/
static struct activation *PlaceRandomActivation(new_act)
  struct activation *new_act;
  {
   int salience, randomID;
   unsigned long timetag;
   struct activation *actPtr, *last_act;
       
   timetag = new_act->timetag;
   salience = new_act->salience;
   randomID = new_act->randomID;
   actPtr = (struct activation *) GetNextActivation(NULL); 
   last_act = NULL;

   while (actPtr != NULL)
     {
      if (actPtr->salience > salience)
        {
         last_act = actPtr;
         actPtr = actPtr->next;
        }
      else if (actPtr->salience < salience)
        { return(last_act); }
      else if (randomID > actPtr->randomID)
        {
         last_act = actPtr;
         actPtr = actPtr->next;
        }
      else if (randomID < actPtr->randomID)
       { return(last_act); }
      else if (timetag > actPtr->timetag)
        {
         last_act = actPtr;
         actPtr = actPtr->next;
        }
      else
       { return(last_act); }
     }

   return(last_act);
  }

/*******************************************************************/
/* SortBindings: Copies a list of fact identifiers and sorts them. */
/*******************************************************************/
static struct partialMatch *SortBindings(binds)
  struct partialMatch *binds;
  {
   struct partialMatch *nbinds;
   struct alphaMatch *temp;
   int flag, j, k;
   
   /*=================*/
   /* Copy the array. */
   /*=================*/
   
   nbinds = CopyPartialMatch(binds);
      
   /*=================*/
   /* Sort the array. */
   /*=================*/
   
   flag = CLIPS_TRUE;
   k = binds->bcount - 1;
   while (flag == CLIPS_TRUE)
     {
      flag = CLIPS_FALSE;
      for (j = 0 ; j < k ; j++)
        {
         if (nbinds->binds[j]->whoset < nbinds->binds[j + 1]->whoset)
           { 
            temp = nbinds->binds[j];
            
            nbinds->binds[j] = nbinds->binds[j+1];
              
            nbinds->binds[j+1] = temp;

            flag = CLIPS_TRUE;
           }
        }
      k--;
     }
   
   /*===================*/
   /* Return the array. */
   /*===================*/
   
   return(nbinds);
  }
  
/************************************************************************/
/* CompareBindings: Compares two lists of (sorted) fact identifiers and */
/*   determines if one is "lexigraphically" greater than the other.     */ 
/*   This comparison is done for the LEX and MEA strategies.            */ 
/************************************************************************/
static int CompareBindings(actPtr,new_act)
  struct activation *actPtr, *new_act;
  {
   int cCount, oCount, mCount, flag, i;
   int complexity;
  
   complexity = new_act->theRule->complexity;
   if (actPtr->sortedBasis == NULL)
     { actPtr->sortedBasis = SortBindings(actPtr->basis); }
        
   cCount = new_act->sortedBasis->bcount;
   oCount = actPtr->sortedBasis->bcount;
   if (oCount > cCount) mCount = cCount;
   else mCount = oCount;
         
   flag = EQUAL;
   for (i = 0 ; (i < mCount) && (flag == EQUAL) ; i++)
     {
      if ((actPtr->sortedBasis->binds[i]->whoset > 0) || 
          (new_act->sortedBasis->binds[i]->whoset > 0))
        { 
         if (new_act->sortedBasis->binds[i]->whoset < 
             actPtr->sortedBasis->binds[i]->whoset) 
           { flag = LESS_THAN; }
         else if (new_act->sortedBasis->binds[i]->whoset > 
                  actPtr->sortedBasis->binds[i]->whoset) 
           { flag = GREATER_THAN; }
        }
      else
        { 
         if (new_act->sortedBasis->binds[i]->whoset > 
             actPtr->sortedBasis->binds[i]->whoset) 
           { flag = LESS_THAN; }
         else if (new_act->sortedBasis->binds[i]->whoset < 
                  actPtr->sortedBasis->binds[i]->whoset) 
           { flag = GREATER_THAN; }
        }
     }
        
   if (flag == EQUAL)
     {
      if (cCount < oCount) flag = LESS_THAN;
      else if (oCount > cCount) flag = GREATER_THAN;
     }
     
   if (flag == EQUAL)
     {
      if (complexity < actPtr->theRule->complexity) flag = LESS_THAN;
      else if (complexity > actPtr->theRule->complexity) flag = GREATER_THAN;
     }
     
   return(flag);
  }       

/********************************************************************/
/* ClearRuleFromAgenda: Clears the agenda of a specified rule name. */
/********************************************************************/
globle VOID ClearRuleFromAgenda(ruleName)
  SYMBOL_HN *ruleName;
  {
   struct activation *agenda_ptr, *agenda_next;

   agenda_ptr = Agenda;

   /*==============================================*/
   /* Loop through every activation on the agenda. */
   /*==============================================*/

   while (agenda_ptr != NULL)
     {
      agenda_next = agenda_ptr->next;
      if (agenda_ptr->theRule->name == ruleName)
        { RemoveActivation(agenda_ptr,CLIPS_TRUE,CLIPS_TRUE); }
      agenda_ptr = agenda_next;
     }
  }
  
/***************************************************/
/* GetNextActivation:                              */
/***************************************************/
globle VOID *GetNextActivation(actPtr)
  VOID *actPtr;
  {
   if (actPtr == NULL)
     { return((VOID *) Agenda); }
   else
     { return((VOID *) (((struct activation *) actPtr)->next)); }
  }

/***************************************************/
/* GetActivationName:                             */
/***************************************************/
globle char *GetActivationName(actPtr)
  VOID *actPtr;
  { return(ValueToString(((struct activation *) actPtr)->theRule->name)); }
  
/***************************************************/
/* ActivationBasis:                             */
/***************************************************/
globle struct partialMatch *ActivationBasis(actPtr)
  VOID *actPtr;
  { return(((struct activation *) actPtr)->basis); }
  
/***************************************************/
/* GetActivationSalience:                             */
/***************************************************/
globle int GetActivationSalience(actPtr)
  VOID *actPtr;
  { return(((struct activation *) actPtr)->salience); }
  
/***************************************************/
/* SetActivationSalience:                             */
/***************************************************/
globle int SetActivationSalience(actPtr,value)
  VOID *actPtr;
  int value;
  { 
   int temp;
   
   temp = ((struct activation *) actPtr)->salience;
   ((struct activation *) actPtr)->salience = value;
   return(temp); 
  }

/***************************************************/
/* GetActivationPPForm:                            */
/***************************************************/
globle VOID GetActivationPPForm(buffer,bufferLength,theActivation)
  char *buffer;
  int bufferLength;
  VOID *theActivation;
  {
   OpenStringDestination("ActPPForm",buffer,bufferLength);
   PrintActivation("ActPPForm",(struct activation *) theActivation);
   CloseStringDestination("ActPPForm");
  }
  
/*************************************************/
/* MoveActivationToTop: Moves the specified activation */
/*   to the top of the agenda.                   */
/*************************************************/
globle BOOLEAN MoveActivationToTop(theActivation)
  VOID *theActivation;
  {
   struct activation *tmpPtr, *prevPtr;
   
   tmpPtr = (struct activation *) theActivation;
   prevPtr = tmpPtr->prev;
   
   if (tmpPtr != Agenda)
     {
      prevPtr->next = tmpPtr->next;
      if (tmpPtr->next != NULL) tmpPtr->next->prev = prevPtr;
      
      tmpPtr->next = Agenda;
      Agenda->prev = tmpPtr;
      tmpPtr->prev = NULL;
      Agenda = tmpPtr;
     }
     
   AgendaChanged = CLIPS_TRUE;
   return(CLIPS_TRUE);
  }
  
/******************************************************/
/* DeleteActivation: Removes the specified activation */
/*   from the agenda.                                 */
/******************************************************/
globle BOOLEAN DeleteActivation(theActivation)
  VOID *theActivation;
  {
   if (theActivation == NULL) RemoveAllActivations();
   else RemoveActivation((struct activation *) theActivation,CLIPS_TRUE,CLIPS_TRUE); 
   
   return(CLIPS_TRUE);
  }
  
/***************************************************/
/* PrintActivation:                         */
/***************************************************/
globle VOID PrintActivation(logicalName,actPtr)
  char *logicalName;
  struct activation *actPtr;
  {
   char print_space[20];
   
   sprintf(print_space,"%-6d ",actPtr->salience);
   PrintCLIPS(logicalName,print_space);
   PrintCLIPS(logicalName,ValueToString(actPtr->theRule->name));
   PrintCLIPS(logicalName,": ");
   PrintPartialMatch(logicalName,actPtr->basis);
  }
  
/***************************************************/
/* PrintCRSVActivation:                          */
/***************************************************/
globle VOID PrintCRSVActivation(logicalName,actPtr)
  char *logicalName;
  struct activation *actPtr;
  {
   char print_space[20];
   struct partialMatch *list;   
   short int i;

   sprintf(print_space,"%-6d ",actPtr->salience);
   PrintCLIPS(logicalName,print_space);
   PrintCLIPS(logicalName,ValueToString(actPtr->theRule->name));
   PrintCLIPS(logicalName,": ");
   
   list = actPtr->basis;
   for (i = 0; i < list->bcount; i++)
     {
      if (list->binds[i]->whoset >= 0)
        {
         sprintf(print_space,"%ld ",list->binds[i]->whoset);
         PrintCLIPS(logicalName,print_space);
        }
     }
  }  
  
/***************************************************/
/* ListAgenda:                                     */
/***************************************************/
globle VOID ListAgenda()
  {
   VOID *agenda_ptr;
   long count = 0;
  
   agenda_ptr = GetNextActivation(NULL);
   while (agenda_ptr != NULL)
     {       
      if (GetHaltExecution() == CLIPS_TRUE) return;
      PrintActivation(WDISPLAY,agenda_ptr);
      PrintCLIPS(WDISPLAY,"\n");
      agenda_ptr = GetNextActivation(agenda_ptr);
      count++;
     }
           
   PrintTally(WDISPLAY,count,"activation","activations");
  }
  
/*************************/
/* AddRunFunction:    */
/*************************/
globle BOOLEAN AddRunFunction(name,functionPtr,priority)
  char *name;
  VOID (*functionPtr)();
  int priority;
  {
   struct runFunction *newPtr, *currentPtr, *lastPtr = NULL;

   newPtr = get_struct(runFunction);

   newPtr->name = name;
   newPtr->ip = functionPtr;
   newPtr->priority = priority;
   
   if (ListOfRunFunctions == NULL)
     {
      newPtr->next = NULL;
      ListOfRunFunctions = newPtr;
      return(1);
     }
     
   currentPtr = ListOfRunFunctions;
   while ((currentPtr != NULL) ? (priority < currentPtr->priority) : CLIPS_FALSE)
     {
      lastPtr = currentPtr; 
      currentPtr = currentPtr->next;
     }

   if (lastPtr == NULL)
     { 
      newPtr->next = ListOfRunFunctions;
      ListOfRunFunctions = newPtr;
     }
   else
     {
      newPtr->next = currentPtr;
      lastPtr->next = newPtr;
     }
   
   return(1);
  }
 
/****************************/
/* RemoveRunFunction:    */
/****************************/
globle BOOLEAN RemoveRunFunction(name)
  char *name;
  {
   struct runFunction *currentPtr, *lastPtr;

   lastPtr = NULL;
   currentPtr = ListOfRunFunctions;
   
   while (currentPtr != NULL)
     {
      if (strcmp(name,currentPtr->name) == 0)
        {
         if (lastPtr == NULL)
           { ListOfRunFunctions = currentPtr->next; }
         else
           { lastPtr->next = currentPtr->next; }
         rtn_struct(runFunction,currentPtr);
         return(1);
        }
      lastPtr = currentPtr;
      currentPtr = currentPtr->next;
     }
     
   return(0);
  }

/***************************************************************/
/* RemoveActivation:                     */
/***************************************************************/
globle VOID RemoveActivation(actPtr,update_agenda,update_links)
  struct activation *actPtr;
  int update_agenda, update_links;
  {   
   /*=================================*/
   /* Update the agenda if necessary. */
   /*=================================*/
   
   if (update_agenda == CLIPS_TRUE)
     {
      if (actPtr->prev == NULL)
        {
         Agenda = Agenda->next;  
         if (Agenda != NULL) Agenda->prev = NULL; 
        }
      else
        {
         actPtr->prev->next = actPtr->next;
         if (actPtr->next != NULL) 
           { actPtr->next->prev = actPtr->prev; }
        }
    
    
      /*===============================================*/
      /* Indicate removal of activation if activations */
      /* are being watched.                            */
      /*===============================================*/

#if DEBUGGING_FUNCTIONS 
      if (WatchActivations == ON)
        { 
         PrintCLIPS(WTRACE,"<== Activation ");
         PrintActivation(WTRACE,actPtr); 
         PrintCLIPS(WTRACE,"\n");
        }
    
      if (CRSVTraceActive())
        {
         char print_space[20];
         PrintCLIPS("wcrsv_tr","D    ");
         sprintf(print_space,"%-6d ",actPtr->salience);
         PrintCLIPS("wcrsv_tr",print_space);
         PrintCLIPS("wcrsv_tr",ValueToString(actPtr->theRule->name));
         PrintCLIPS("wcrsv_tr","\n");
        } 
#endif  
        
      AgendaChanged = CLIPS_TRUE;
     }
 
   /*============================================*/
   /* Update join and agenda links if necessary. */
   /*============================================*/
   
   if ((update_links == CLIPS_TRUE) && (actPtr->basis != NULL))
     { actPtr->basis->binds[actPtr->basis->bcount] = NULL; }
     
   /*========================*/
   /* Return the activation. */
   /*========================*/

   NumberOfActivations--;   
   
   if (actPtr->sortedBasis != NULL)
     { ReturnPartialMatch(actPtr->sortedBasis); }
     
   rtn_struct(activation,actPtr); 
   
   if (Agenda == NULL) CurrentTimetag = 0;
  }

/******************************************************************/
/* RemoveAllActivations: Removes all activations from the agenda. */
/******************************************************************/
globle VOID RemoveAllActivations()
  {
   struct activation *tempPtr, *actPtr;
   
   actPtr = Agenda;
   while (actPtr != NULL)
     { 
      tempPtr = actPtr->next;
      RemoveActivation(actPtr,CLIPS_TRUE,CLIPS_TRUE); 
      actPtr = tempPtr;
     }
  }

/***************************************************/
/* GetAgendaChanged:                               */
/***************************************************/
globle int GetAgendaChanged()
  { return(AgendaChanged); }
  
/***************************************************/
/* SetAgendaChanged:                              */
/***************************************************/
globle VOID SetAgendaChanged(value)
  int value;
  { 
   AgendaChanged = value;
  }
  
/***************************************************************/
/* SetStrategy:                     */
/***************************************************************/
globle int SetStrategy(value)
  int value;
  {
   int old_strategy;
   
   old_strategy = Strategy;
   Strategy = value;
   
   if (old_strategy != Strategy) ReorderAgenda();
     
   return(old_strategy);
  }
    
/*******************************************************/
/* ReorderAgenda: Completely reorders the agenda based */
/*   on the current conflict resolution strategy.      */
/*******************************************************/
globle VOID ReorderAgenda()
  {
   struct activation *actPtr, *tempPtr;
   
   actPtr = Agenda;
   Agenda = NULL;
   
   while (actPtr != NULL)
     {
      tempPtr = actPtr->next;
      actPtr->next = NULL;
      actPtr->prev = NULL;
      PlaceActivation(actPtr);
      actPtr = tempPtr;
     }
  }
  
/*************************************************************/
/* RefreshAgenda: Recomputes the salience values for all     */
/*   activations on the agenda and then reorders the agenda. */
/*************************************************************/
globle VOID RefreshAgenda()
  {
   struct activation *actPtr;
   BOOLEAN oldValue;
   
   actPtr = Agenda;
   
   oldValue = GetSalienceEvaluation();
   SetSalienceEvaluation(WHEN_ACTIVATED);
   
   while (actPtr != NULL)
     {
      actPtr->salience = EvaluateSalience(actPtr->theRule);
      actPtr = actPtr->next;
     }
     
   ReorderAgenda();
   
   SetSalienceEvaluation(oldValue);
  }
  
/***************************************************************/
/* GetStrategy:                     */
/***************************************************************/
globle int GetStrategy()
  {
   return(Strategy);
  }

/*************************************************************/
/* GetRuleFiring:                                            */
/*************************************************************/
globle char *GetRuleFiring()
  {
   return(RuleFiring);
  }
  
/***************************************************/
/* GetNumberOfActivations:                         */
/***************************************************/
globle long int GetNumberOfActivations()
  { return(NumberOfActivations); }
  
/***************************************************/
/* InitializeEngine:                         */
/***************************************************/
globle VOID InitializeEngine()
  {
#if DEBUGGING_FUNCTIONS 
   AddWatchItem("activations",&WatchActivations,40);
   AddWatchItem("statistics",&WatchStatistics,20);
#endif
  }

/************************************************************/
/* AddBreakpoint:                                          */
/************************************************************/
globle VOID AddBreakpoint(theRule)
  VOID *theRule;
  { 
   struct defrule *thePtr;
   
   thePtr = (struct defrule *) theRule;
   while (thePtr != NULL)
     {
      thePtr->afterBreakpoint = 1; 
      thePtr = thePtr->disjunct;
     }
  }
  
/************************************************************/
/* RemoveBreakpoint:                                      */
/************************************************************/
globle BOOLEAN RemoveBreakpoint(theRule)
  VOID *theRule;
  {
   struct defrule *thePtr;
   int rv = CLIPS_FALSE;
   
   thePtr = (struct defrule *) theRule;
   while (thePtr != NULL)
     {
      if (thePtr->afterBreakpoint == 1)
        {
         thePtr->afterBreakpoint = 0;
         rv = CLIPS_TRUE;
        }
      thePtr = thePtr->disjunct;
     }
     
   return(rv);
  }
  
/************************************************************/
/* RemoveAllBreakpoints:                                       */
/************************************************************/
globle VOID RemoveAllBreakpoints()
  {
   VOID *theRule = NULL;
   
   while ((theRule = GetNextDefrule(theRule)) != NULL)
     { RemoveBreakpoint(theRule); }
  }

/************************************************************/
/* ListBreakpoints:                                             */
/************************************************************/
globle VOID ListBreakpoints()
  {
   VOID *theRule = NULL;
   
   while ((theRule = GetNextDefrule(theRule)) != NULL)
     { 
      if (((struct defrule *) theRule)->afterBreakpoint == 1)  
        {
         PrintCLIPS(WDISPLAY,GetDefruleName(theRule));
         PrintCLIPS(WDISPLAY,"\n");
        }
     }
   }
   
/*************************/
/* DefruleHasBreakpoint:    */
/*************************/
globle BOOLEAN DefruleHasBreakpoint(theRule)
  VOID *theRule;
  {
   return(((struct defrule *) theRule)->afterBreakpoint);
  }
  
/***************************************************/
/* GetSalienceEvaluation:                              */
/***************************************************/
globle BOOLEAN GetSalienceEvaluation()
  { return(SalienceEvaluation); }
  
/***************************************************/
/* SetSalienceEvaluation:                              */
/***************************************************/
globle BOOLEAN SetSalienceEvaluation(value)
  int value;
  { 
   int ov;
   
   ov = SalienceEvaluation;
   SalienceEvaluation = value;
   return(ov);
  } 
  
#endif

