   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*               CLIPS Version 5.10  07/17/91          */
   /*                                                     */
   /*                                                     */
   /*******************************************************/

/*************************************************************/
/* Purpose: Binary Load/Save Functions for Generic Functions */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
/* =========================================
   *****************************************
               EXTERNAL DEFINITIONS
   =========================================
   ***************************************** */
#include "setup.h"

#if DEFGENERIC_CONSTRUCT && (BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE)

#include "constant.h"
#include "clipsmem.h"
#include "bload.h"
#include "bsave.h"

#if OBJECT_SYSTEM
#include "objbin.h"
#endif

#include "genrcfun.h"

#define _GENRCBIN_SOURCE_
#include "genrcbin.h"

#include "router.h"

/* =========================================
   *****************************************
                   CONSTANTS
   =========================================
   ***************************************** */
   
/* =========================================
   *****************************************
               MACROS AND TYPES
   =========================================
   ***************************************** */
#define MethodPointer(i) ((((long) (i)) == -1L) ? NULL : \
                          (METHOD *) &methodArray[(long) (i)])
#define RestrictionPointer(i) ((((long) (i)) == -1L) ? NULL : \
                               (RESTRICTION *) &restrictionArray[(long) (i)])
#define TypePointer(i) ((((long) (i)) == -1L) ? NULL : \
                        (VOID **) &typeArray[(long) (i)])

/* =========================================
   *****************************************
      INTERNALLY VISIBLE FUNCTION HEADERS
   =========================================
   ***************************************** */
#if ANSI_COMPILER

#if BLOAD_AND_BSAVE

static VOID BsaveGenericsFind(void);
static VOID BsaveGenericsExpressions(FILE *);
static VOID BsaveGenerics(FILE *);

#endif

static int BloadGenericsReady(void);
static VOID BloadGenerics(void);
static VOID RefreshGenerics(void);
static VOID ClearBloadGenerics(void);

#else

#if BLOAD_AND_BSAVE

static VOID BsaveGenericsFind();
static VOID BsaveGenericsExpressions();
static VOID BsaveGenerics();

#endif

static int BloadGenericsReady();
static VOID BloadGenerics();
static VOID RefreshGenerics();
static VOID ClearBloadGenerics();

#endif
      
/* =========================================
   *****************************************
      EXTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */

globle GENERIC_FUNC HUGE_ADDR *defgenericArray = NULL;

/* =========================================
   *****************************************
      INTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */

static long GenericCount = 0L,
            MethodCount = 0L,
            RestrictionCount = 0L,
            TypeCount = 0L;   

static METHOD HUGE_ADDR *methodArray = NULL;
static RESTRICTION HUGE_ADDR *restrictionArray = NULL;
static VOID * HUGE_ADDR *typeArray = NULL;

/* =========================================
   *****************************************
          EXTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
   
/***********************************************************
  NAME         : SetupGenericsBload
  DESCRIPTION  : Initializes data structures and
                   routines for binary loads of
                   generic function constructs
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Routines defined and structures initialized
  NOTES        : None
 ***********************************************************/
globle VOID SetupGenericsBload()
  {
#if BLOAD_AND_BSAVE || BLOAD
   AddBeforeBloadFunction("methods",(VOID (*)(VOID_ARG)) ClearDefmethods,-500);
   AddBeforeBloadFunction("generic functions",(VOID (*)(VOID_ARG)) ClearDefgenerics,-2000);
#endif
#if BLOAD_AND_BSAVE || BLOAD || BLOAD_ONLY
   AddBloadReadyFunction("generic functions",BloadGenericsReady,0);
   AddClearBloadReadyFunction("generic functions",BloadGenericsReady,0);
#endif
#if BLOAD_AND_BSAVE
   AddBinaryItem("generic functions",0,BsaveGenericsFind,BsaveGenericsExpressions,
                              BsaveGenerics,BloadGenerics,RefreshGenerics,
                              ClearBloadGenerics);
#endif
#if BLOAD || BLOAD_ONLY
   AddBinaryItem("generic functions",0,NULL,NULL,NULL,BloadGenerics,RefreshGenerics,
                              ClearBloadGenerics);
#endif
  }
   
/* =========================================
   *****************************************
          INTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */

#if BLOAD_AND_BSAVE

/***************************************************************************
  NAME         : BsaveGenericsFind
  DESCRIPTION  : For all generic functions and their
                   methods, this routine marks all 
                   the needed symbols and system functions.
                 Also, it also counts the number of
                   expression structures needed.
                 Also, counts total number of generics, methods,
                   restrictions and types.
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : ExpressionCount (a global from BSAVE.C) is incremented
                   for every expression needed
                 Symbols and system function are marked in their structures
  NOTES        : Also sets bsaveIndex for each generic function (assumes
                   generic functions will be bsaved in order of binary list)
 ***************************************************************************/
static VOID BsaveGenericsFind()
  {
   GENERIC_FUNC *gfunc;
   register unsigned i,j;
   METHOD *meth;
   RESTRICTION *rptr;
   
   GenericCount = 0L;
   MethodCount = 0L;
   RestrictionCount = 0L;
   TypeCount = 0L;
   for (gfunc = GenericList ; gfunc != NULL ; gfunc = gfunc->nxt)
     {
      gfunc->bsaveIndex = GenericCount++;
      gfunc->name->neededSymbol = CLIPS_TRUE;
      if (gfunc->sysfunc != NULL)
        gfunc->sysfunc->bsaveIndex = CLIPS_TRUE;
      MethodCount += gfunc->mcnt;
      for (i = 0 ; i < gfunc->mcnt ; i++)
        {
         meth = &gfunc->methods[i];
         ExpressionCount += ExpressionSize(meth->actions);
         MarkNeededItems(meth->actions);
         RestrictionCount += meth->minRestrictions;
         for (j = 0 ; j < meth->minRestrictions ; j++)
           {
            rptr = &meth->restrictions[j];
            ExpressionCount += ExpressionSize(rptr->query);
            MarkNeededItems(rptr->query);
            TypeCount += rptr->tcnt;
           }
        }
     }
  }

/***************************************************
  NAME         : BsaveGenericsExpressions
  DESCRIPTION  : Writes out all expressions needed
                   by generic functions
  INPUTS       : The file pointer of the binary file
  RETURNS      : Nothing useful
  SIDE EFFECTS : File updated
  NOTES        : None
 ***************************************************/
static VOID BsaveGenericsExpressions(fp)
  FILE *fp;
  {
   GENERIC_FUNC *gfunc;
   register unsigned i,j;
   METHOD *meth;
   
   for (gfunc = GenericList ; gfunc != NULL ; gfunc = gfunc->nxt)
     {
      /* ================================================================
         Important to save all expressions for methods before any
         expressions for restrictions, since methods will be stored first
         ================================================================ */
      for (i = 0 ; i < gfunc->mcnt ; i++)
        {
         meth = &gfunc->methods[i];
         BsaveExpression(meth->actions,fp);
        }
      for (i = 0 ; i < gfunc->mcnt ; i++)
        {
         meth = &gfunc->methods[i];
         for (j = 0 ; j < meth->minRestrictions ; j++)
           BsaveExpression(meth->restrictions[j].query,fp);
        }
     }
  }

/*************************************************************************************
  NAME         : BsaveGenerics
  DESCRIPTION  : Writes out generic function in binary format
                 Space required (unsigned long)
                 Number of generics (long)
                 All generic headers (sizeof(GENERIC_FUNC) * Number of generics)
                 All methods (sizeof(METHOD) * Number of methods)
                 All method restrictions (sizeof(RESTRICTION) * Number of restrictions)
                 All restriction type arrays (sizeof(VOID *) * # of types)
  INPUTS       : File pointer of binary file
  RETURNS      : Nothing useful
  SIDE EFFECTS : Binary file adjusted
  NOTES        : None
 *************************************************************************************/
static VOID BsaveGenerics(fp)
  FILE *fp;
  {
   GENERIC_FUNC *gfunc;
   register unsigned i,j;
   METHOD *meth;
   RESTRICTION *rptr;
   unsigned long space;
   GENERIC_FUNC dummy_generic;
   METHOD dummy_method;
   RESTRICTION dummy_restriction;
#if OBJECT_SYSTEM
   register unsigned k;
   VOID *dummy_type;
#endif

   /* =======================================================================
      Space is: Sum over all structures(sizeof(structure) * structure-cnt)) +
                4 longs to write out counts for structures
      ======================================================================= */
   if (GenericCount == 0L)
     {
      space = (unsigned long) GenericCount;
      GenWrite((VOID *) &space,(unsigned long) sizeof(unsigned long),fp);
      return;
     }
     
   space = ((unsigned long) GenericCount * (unsigned long) sizeof(GENERIC_FUNC)) +
           ((unsigned long) MethodCount * (unsigned long) sizeof(METHOD)) +
           ((unsigned long) RestrictionCount * (unsigned long) sizeof(RESTRICTION)) +
           ((unsigned long) TypeCount * (unsigned long) sizeof(VOID *)) +
           (unsigned long) (sizeof(long) * 4);
                   
   /* ===================================================================
      Write out the total amount of space required and the number of each
      structure required:  headers, methods, restrictions, types
      =================================================================== */         
   GenWrite((VOID *) &space,(unsigned long) sizeof(unsigned long),fp);
   GenWrite((VOID *) &GenericCount,(unsigned long) sizeof(long),fp);
   GenWrite((VOID *) &MethodCount,(unsigned long) sizeof(long),fp);
   GenWrite((VOID *) &RestrictionCount,(unsigned long) sizeof(long),fp);
   GenWrite((VOID *) &TypeCount,(unsigned long) sizeof(long),fp);
   
   /* ======================================
      Write out the generic function headers
      ====================================== */
   MethodCount = 0L;
   for (gfunc = GenericList ; gfunc != NULL ; gfunc = gfunc->nxt)
     {
      CopyMemory(GENERIC_FUNC,1,&dummy_generic,gfunc);
      dummy_generic.name = (SYMBOL_HN *) gfunc->name->bucket;
      dummy_generic.busy = 0;
      if (gfunc->methods != NULL)
        {
         dummy_generic.methods = (METHOD *) MethodCount;
         MethodCount += (long) gfunc->mcnt;
        }
      else
        dummy_generic.methods = (METHOD *) -1;
      if (gfunc->sysfunc != NULL)
        dummy_generic.sysfunc = (struct FunctionDefinition *) gfunc->sysfunc->bsaveIndex;
      else
        dummy_generic.sysfunc = (struct FunctionDefinition *) -1;
      if (gfunc->prv != NULL)
        dummy_generic.prv = (GENERIC_FUNC *) gfunc->prv->bsaveIndex;
      else
        dummy_generic.prv = (GENERIC_FUNC *) -1;
      if (gfunc->nxt != NULL)
        dummy_generic.nxt = (GENERIC_FUNC *) gfunc->nxt->bsaveIndex;
      else
        dummy_generic.nxt = (GENERIC_FUNC *) -1;
      GenWrite((VOID *) &dummy_generic,(unsigned long) sizeof(GENERIC_FUNC),fp);
     }
     
   /* =====================
      Write out the methods
      ===================== */
   RestrictionCount = 0L;
   for (gfunc = GenericList ; gfunc != NULL ; gfunc = gfunc->nxt)
     {
      for (i = 0 ; i < gfunc->mcnt ; i++)
        {
         meth = &gfunc->methods[i];
         CopyMemory(METHOD,1,&dummy_method,meth);
         dummy_method.busy = 0;
         if (meth->restrictions != NULL)
           {
            dummy_method.restrictions = (RESTRICTION *) RestrictionCount;
            RestrictionCount += (long) meth->minRestrictions;
           }
         else
           dummy_method.restrictions = (RESTRICTION *) -1;
         if (meth->actions != NULL)
           {
            dummy_method.actions = (EXPRESSION *) ExpressionCount;
            ExpressionCount += ExpressionSize(meth->actions);
           }
         else
           dummy_method.actions = (EXPRESSION *) -1;
         GenWrite((VOID *) &dummy_method,(unsigned long) sizeof(METHOD),fp);
        }
     }
   
   /* =================================
      Write out the method restrictions
      ================================= */
   TypeCount = 0L;
   for (gfunc = GenericList ; gfunc != NULL ; gfunc = gfunc->nxt)
     {
      for (i = 0 ; i < gfunc->mcnt ; i++)
        {
         for (j = 0 ; j < gfunc->methods[i].minRestrictions ; j++)
           {
            rptr = &gfunc->methods[i].restrictions[j];
            CopyMemory(RESTRICTION,1,&dummy_restriction,rptr);
            if (rptr->types != NULL)
              {
               dummy_restriction.types = (VOID **) TypeCount;
               TypeCount += (long) rptr->tcnt;
              }
            else
              dummy_restriction.types = (VOID **) -1;
            if (rptr->query != NULL)
              {
               dummy_restriction.query = (EXPRESSION *) ExpressionCount;
               ExpressionCount += ExpressionSize(rptr->query);
              }
            else
              dummy_restriction.query = (EXPRESSION *) -1;
            GenWrite((VOID *) &dummy_restriction,(unsigned long) sizeof(RESTRICTION),fp);
           }
        }
     }
     
   /* =============================================================
      Finally, write out the type lists for the method restrictions
      ============================================================= */
   for (gfunc = GenericList ; gfunc != NULL ; gfunc = gfunc->nxt)
     {
      for (i = 0 ; i < gfunc->mcnt ; i++)
        {
         for (j = 0 ; j < gfunc->methods[i].minRestrictions ; j++)
           {
            rptr = &gfunc->methods[i].restrictions[j];
#if OBJECT_SYSTEM
            for (k = 0 ; k < rptr->tcnt ; k++)
              {
               dummy_type = (VOID *) ((CLASS_TYPE *) rptr->types[k])->bsaveIndex;
               GenWrite(&dummy_type,(unsigned long) sizeof(VOID *),fp);
              }
#else
            GenWrite((VOID *) rptr->types,
                     (unsigned long) (rptr->tcnt * sizeof(VOID *)),fp);
#endif
           }
        }
     }
   GenericCount = 0L;
  }

#endif

/***************************************************
  NAME         : BloadGenericsReady
  DESCRIPTION  : Checks on timing of bload
  INPUTS       : None
  RETURNS      : TRUE if it's OK to proceed with a
                  bload, FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : Assumes Before Bload function
                   already called
 ***************************************************/
static int BloadGenericsReady()
  {
   GENERIC_FUNC *gfunc;
   
   for (gfunc = GenericList ; gfunc != NULL ; gfunc = gfunc->nxt)
     if (gfunc->busy > 0)
       return(FALSE);
   return(TRUE);
  }

/***********************************************************************
  NAME         : BloadGenerics
  DESCRIPTION  : This routine reads generic function information from
                 a binary file in five chunks:
                 Count array
                   0 - Generic header count
                   1 - Method count
                   2 - Method restriction count
                   3 - Restriction type count
                 Generic-header array
                 Method array
                 Method restrictions array
                 Restriction types array
  INPUTS       : Notthing
  RETURNS      : Nothing useful
  SIDE EFFECTS : Arrays allocated and set
  NOTES        : This routine makes no attempt to reset any pointers
                   within the structures
                 Bload fails if there are still generics in the system!!
 ***********************************************************************/
static VOID BloadGenerics()
  {
   unsigned long space;
   long counts[4];
   
   GenRead((VOID *) &space,(unsigned long) sizeof(unsigned long));
   if (space == 0L)
     return;
   GenRead((VOID *) counts,(unsigned long) (sizeof(long) * 4));
   GenericCount = counts[0];
   MethodCount = counts[1];
   RestrictionCount = counts[2];
   TypeCount = counts[3];
   if (GenericCount != 0L)
     {
      space = (unsigned long) (sizeof(GENERIC_FUNC) * GenericCount);
      defgenericArray = (GENERIC_FUNC HUGE_ADDR *) genlongalloc(space);
      GenRead((VOID *) defgenericArray,space);
     }
   else
     return;
   if (MethodCount != 0L)
     {
      space = (unsigned long) (sizeof(METHOD) * MethodCount);
      methodArray = (METHOD HUGE_ADDR *) genlongalloc(space);
      GenRead((VOID *) methodArray,space);
     }
   else
     return;
   if (RestrictionCount != 0L)
     {
      space = (unsigned long) (sizeof(RESTRICTION) * RestrictionCount);
      restrictionArray = (RESTRICTION HUGE_ADDR *) genlongalloc(space);
      GenRead((VOID *) restrictionArray,space);
     }
   else
     return;
   if (TypeCount != 0L)
     {
      space = (unsigned long) (sizeof(VOID *) * TypeCount);
      typeArray = (VOID * HUGE_ADDR *) genlongalloc(space);
      GenRead((VOID *) typeArray,space);
     }
  }

/***************************************************************
  NAME         : RefreshGenerics
  DESCRIPTION  : This routine moves through the generic function
                   binary arrays updating pointers
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Pointers reset from array indices
  NOTES        : Assumes all loading is finished
 **************************************************************/
static VOID RefreshGenerics()
  {
   register long i;
   
   if (GenericCount == 0L)
     return;
   for (i = 0L ; i < GenericCount ; i++)
     {
      defgenericArray[i].name = SymbolPointer(defgenericArray[i].name);
      IncrementCount(defgenericArray[i].name);
      defgenericArray[i].methods = MethodPointer(defgenericArray[i].methods);
      defgenericArray[i].sysfunc = FunctionPointer(defgenericArray[i].sysfunc);
      defgenericArray[i].pp_form = NULL;
      defgenericArray[i].nxt = GenericPointer(defgenericArray[i].nxt);
      defgenericArray[i].prv = GenericPointer(defgenericArray[i].prv);
     }
   SetGenericList((GENERIC_FUNC *) &defgenericArray[0]);
   
   for (i = 0L ; i < MethodCount ; i++)
     {
      methodArray[i].restrictions = RestrictionPointer(methodArray[i].restrictions);
      methodArray[i].actions = ExpressionPointer(methodArray[i].actions);
      methodArray[i].pp_form = NULL;
     }
     
   for (i = 0L ; i < RestrictionCount ; i++)
     {
      restrictionArray[i].types = TypePointer(restrictionArray[i].types);
      restrictionArray[i].query = ExpressionPointer(restrictionArray[i].query);
     }

#if OBJECT_SYSTEM
   for (i = 0L ; i < TypeCount ; i++)
     typeArray[i] = (VOID *) ClassPointer(typeArray[i]);
#endif
  }

/***************************************************************
  NAME         : ClearBloadGenerics
  DESCRIPTION  : Release all binary-loaded generic function
                   structure arrays
                 Resets generic function list to NULL
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Memory cleared
  NOTES        : Generic function name symbol counts decremented
 ***************************************************************/
static VOID ClearBloadGenerics()
  {
   register long i;
   unsigned long space;
   
   SetGenericList(NULL);
   for (i = 0 ; i < GenericCount ; i++)
     DecrementSymbolCount(defgenericArray[i].name);
     
   space = (unsigned long) (sizeof(GENERIC_FUNC) * GenericCount);
   if (space == 0L)
     return;
   genlongfree((VOID *) defgenericArray,space);
   defgenericArray = NULL;
   GenericCount = 0L;
   
   space = (unsigned long) (sizeof(METHOD) * MethodCount);
   if (space == 0L)
     return;
   genlongfree((VOID *) methodArray,space);
   methodArray = NULL;
   MethodCount = 0L;
   
   space = (unsigned long) (sizeof(RESTRICTION) * RestrictionCount);
   if (space == 0L)
     return;
   genlongfree((VOID *) restrictionArray,space);
   restrictionArray = NULL;
   RestrictionCount = 0L;
   
   space = (unsigned long) (sizeof(VOID *) * TypeCount);
   if (space == 0L)
     return;
   genlongfree((VOID *) typeArray,space);
   typeArray = NULL;
   TypeCount = 0L;
  }

#endif

/***************************************************
  NAME         : 
  DESCRIPTION  : 
  INPUTS       : 
  RETURNS      : 
  SIDE EFFECTS : 
  NOTES        : 
 ***************************************************/
