   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*               CLIPS Version 5.10  07/17/91          */
   /*                                                     */
   /*                                                     */
   /*******************************************************/

/*************************************************************/
/* Purpose: Generic Function Construct Compiler Code         */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
/* =========================================
   *****************************************
               EXTERNAL DEFINITIONS
   =========================================
   ***************************************** */
#include "setup.h"

#if DEFGENERIC_CONSTRUCT && CONSTRUCT_COMPILER && (! RUN_TIME)

#include "constant.h"
#include "genrcfun.h"
#include "conscomp.h"

#if OBJECT_SYSTEM
#include "objcmp.h"
#endif

#define _GENRCCMP_SOURCE_
#include "genrccmp.h"

/* =========================================
   *****************************************
                   CONSTANTS
   =========================================
   ***************************************** */
   
/* =========================================
   *****************************************
               MACROS AND TYPES
   =========================================
   ***************************************** */
   
/* =========================================
   *****************************************
      INTERNALLY VISIBLE FUNCTION HEADERS
   =========================================
   ***************************************** */
#if ANSI_COMPILER
static int GenericsToCode(char *,int,FILE *,int,int);
static VOID InitGenericsCode(FILE *,int,int);
static VOID ReadyGenericsForCode(void);
#else
static int GenericsToCode();
static VOID InitGenericsCode();
static VOID ReadyGenericsForCode();
#endif
      
/* =========================================
   *****************************************
          EXTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
   
/***************************************************
  NAME         : SetupGenericsCompiler
  DESCRIPTION  : Initializes the construct compiler
                   item for generic functions
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Code generator item initialized
  NOTES        : None
 ***************************************************/
globle VOID SetupGenericsCompiler()
  {
   AddCodeGeneratorItem("generics",0,ReadyGenericsForCode,
                        InitGenericsCode,GenericsToCode);
  }
  
   
/* =========================================
   *****************************************
          INTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
   
/*******************************************************
  NAME         : GenericsToCode
  DESCRIPTION  : Writes out static array code for
                   generic functions, methods, etc.
  INPUTS       : 1) The base name of the construct set
                 2) The base id for this construct
                 3) The file pointer for the header file
                 4) The base id for the construct set
                 5) The max number of indices allowed
                    in an array
  RETURNS      : -1 if no generic functions, 0 on errors,
                  1 if generic functions written
  SIDE EFFECTS : Code written to files
  NOTES        : None
 *******************************************************/
static int GenericsToCode(fileName,fileID,headerFP,imageID,maxIndices)
  char *fileName;
  int fileID;
  FILE *headerFP;
  int imageID;
  int maxIndices;
  {
   GENERIC_FUNC *gfunc;
   METHOD *meth;
   RESTRICTION *rptr;
   FILE *fp = NULL;
   int i,version,partition,ti,tpart,newHeader;
   unsigned mi,ri,typi;
   unsigned long count;
   
   fprintf(headerFP,"#include \"genrcfun.h\"\n");
   if (GenericList == NULL)
     return(-1);
   
   /* ================================
      Write out the header definitions
      ================================ */
   i = 1;
   version = 1;
   partition = 1;
   newHeader = CLIPS_TRUE;
   ti = 1;
   tpart = 1;
   for (gfunc = GenericList ; gfunc != NULL ; gfunc = gfunc->nxt)
     {
      if (newHeader)
        {
         if ((fp = NewCFile(fileName,fileID,version)) == NULL)
           return(0);
         fprintf(fp,"GENERIC_FUNC gfn%d_%d[] = {\n{",imageID,partition);
         fprintf(headerFP,"extern GENERIC_FUNC gfn%d_%d[];\n",
                 imageID,partition);
         version++;
         partition++;
         newHeader = CLIPS_FALSE;
        }
      else
        fprintf(fp,",\n{");
      PrintSymbolReference(fp,gfunc->name);
      fprintf(fp,",0,");
      if (gfunc->methods != NULL)
        {
         fprintf(fp,"&mth%d_%d[%d],",imageID,tpart,ti-1);
         ti += gfunc->mcnt;
         if (ti > maxIndices)
           {
            tpart++;
            ti = 1;
           }
        }
      else
        fprintf(fp,"NULL,");
      PrintFunctionReference(fp,gfunc->sysfunc);
      fprintf(fp,",%u,0,NULL,0L,",gfunc->mcnt);
      PrintGenericFunctionReference(fp,gfunc->nxt);
      fprintf(fp,",");
      PrintGenericFunctionReference(fp,gfunc->prv);
      fprintf(fp,"}");
      i++;
      if (i > maxIndices)
        {
         fprintf(fp,"\n};\n");
         fclose(fp);
         fp = NULL;
         i = 1;
         newHeader = CLIPS_TRUE;
        }
     }
   if (fp != NULL)
     {
      fprintf(fp,"\n};\n");
      fclose(fp);
     }
   
   /* =========================================================================
      Write out the method definitions - if need to split up array, make sure
      splits only occur on generic function boundaries because generic function
      methods are accessed as a contiguous array
      ========================================================================= */
   i = 1;
   partition = 1;
   tpart = 1;
   ti = 1;
   newHeader = CLIPS_TRUE;
   for (gfunc = GenericList ; gfunc != NULL ; gfunc = gfunc->nxt)
     {
      for (mi = 0 ; mi < gfunc->mcnt ; mi++)
        {
         meth = &gfunc->methods[mi];
         if (newHeader)
           {
            if ((fp = NewCFile(fileName,fileID,version)) == NULL)
              return(0);
            fprintf(fp,"METHOD mth%d_%d[] = {\n{",imageID,partition);
            fprintf(headerFP,"extern METHOD mth%d_%d[];\n",
                    imageID,partition);
            version++;
            partition++;
            newHeader = CLIPS_FALSE;
           }
         else
           fprintf(fp,",\n{");
         fprintf(fp,"%u,0,%d,%d,",meth->index,
                 meth->minRestrictions,meth->maxRestrictions);
         if (meth->restrictions != NULL)
           {
            fprintf(fp,"&rst%d_%d[%d],",imageID,tpart,ti-1);
            ti += meth->minRestrictions;
            if (ti > maxIndices)
              {
               tpart++;
               ti = 1;
              }
           }
         else
           fprintf(fp,"NULL,");
         ExpressionToCode(fp,meth->actions);
         fprintf(fp,",NULL}");
         i++;
        }
      if (i > maxIndices)
        {
         fprintf(fp,"\n};\n");
         fclose(fp);
         fp = NULL;
         i = 1;
         newHeader = CLIPS_TRUE;
        }
     }
   if (fp != NULL)
     {
      fprintf(fp,"\n};\n");
      fclose(fp);
     }

   /* =========================================================================
      Write out the method restriction definitions - if need to split up array,
      make sure splits only occur on method boundaries because method
      restrictions are accessed as a contiguous array
      ========================================================================= */
   i = 1;
   partition = 1;
   count = 0L;
   newHeader = CLIPS_TRUE;
   for (gfunc = GenericList ; gfunc != NULL ; gfunc = gfunc->nxt)
     {
      for (mi = 0 ; mi < gfunc->mcnt ; mi++)
        {
         meth = &gfunc->methods[mi];
         for (ri = 0 ; ri < meth->minRestrictions ; ri++)
           {
            rptr = &meth->restrictions[ri];
            if (newHeader)
              {
               if ((fp = NewCFile(fileName,fileID,version)) == NULL)
                 return(0);
               fprintf(fp,"RESTRICTION rst%d_%d[] = {\n{",imageID,partition);
               fprintf(headerFP,"extern RESTRICTION rst%d_%d[];\n",
                       imageID,partition);
               version++;
               partition++;
               newHeader = CLIPS_FALSE;
              }
            else
              fprintf(fp,",\n{");
            fprintf(fp,"%u,",rptr->tcnt);
            if (rptr->types != NULL)
              {
               fprintf(fp,"&typ%d[%lu],",imageID,count);
               count += rptr->tcnt;
              }
            else
              fprintf(fp,"NULL,");
            ExpressionToCode(fp,rptr->query);
            fprintf(fp,"}");
            i++;
           }
         if (i > maxIndices)
           {
            fprintf(fp,"\n};\n");
            fclose(fp);
            fp = NULL;
            i = 1;
            newHeader = CLIPS_TRUE;
           }
        }
     }
   if (fp != NULL)
     {
      fprintf(fp,"\n};\n");
      fclose(fp);
     }

   /* ===========================================================
      Write out all the restriction types in one contiguous array
      Don't anticipate this array growing to unmanageable size
      so no need to split it up.
      =========================================================== */
   newHeader = CLIPS_TRUE;
   fp = NULL;
   for (gfunc = GenericList ; gfunc != NULL ; gfunc = gfunc->nxt)
     {
      for (mi = 0 ; mi < gfunc->mcnt ; mi++)
        {
         meth = &gfunc->methods[mi];
         for (ri = 0 ; ri < meth->minRestrictions ; ri++)
           {
            rptr = &meth->restrictions[ri];
            for (typi = 0 ; typi < rptr->tcnt ; typi++)
              {
               if (newHeader)
                 {
                  if ((fp = NewCFile(fileName,fileID,version)) == NULL)
                    return(0);
                  fprintf(fp,"VOID *typ%d[] = {\n",imageID);
                  fprintf(headerFP,"extern VOID *typ%d[];\n",imageID);
                  newHeader = CLIPS_FALSE;
                 }
               else
                 fprintf(fp,",\n");
#if OBJECT_SYSTEM               
               PrintClassReference(fp,(CLASS_TYPE *) rptr->types[typi],
                                   imageID,maxIndices);
#else
               fprintf(fp,"%d",(int) rptr->types[typi]);
#endif               
              }
           }
        }
     }
   if (fp != NULL)
     {
      fprintf(fp,"\n};\n");
      fclose(fp);
     }
     
   return(1);
  }
  
/*******************************************************
  NAME         : InitGenericsCode
  DESCRIPTION  : Writes out initialization code for
                   generic functions
  INPUTS       : 1) The initialization code file pointer
                 2) The construct set image id
                 3) The max number of indices allowed in
                    an array for this construct set
  RETURNS      : Nothing useful
  SIDE EFFECTS : Writes out initialization code
  NOTES        : None
 *******************************************************/
static VOID InitGenericsCode(initFP,imageID,maxIndices)
  FILE *initFP;
  int imageID;
  int maxIndices;
  {
#if MAC_MPW
#pragma unused(maxIndices)
#endif
   if (GenericList != NULL)
     fprintf(initFP,"   SetGenericList(gfn%d_1);\n",imageID);
   else
     fprintf(initFP,"   SetGenericList(NULL);\n");
  }

/***************************************************
  NAME         : ReadyGenericsForCode
  DESCRIPTION  : Sets index of generic-functions
                   for use in compiled expressions
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : BsaveIndices set
  NOTES        : None
 ***************************************************/
static VOID ReadyGenericsForCode()
  {
   GENERIC_FUNC *gfunc;
   register long i;
   
   for (gfunc = GenericList , i = 0L ; gfunc != NULL ; 
        gfunc = gfunc->nxt , i++)
     gfunc->bsaveIndex = i;
  }

#endif

/***************************************************
  NAME         : 
  DESCRIPTION  : 
  INPUTS       : 
  RETURNS      : 
  SIDE EFFECTS : 
  NOTES        : 
 ***************************************************/
