   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*               CLIPS Version 5.10  07/17/91          */
   /*                                                     */
   /*                                                     */
   /*******************************************************/

/*************************************************************/
/* Purpose: Query Functions for Objects                      */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
/* =========================================
   *****************************************
               EXTERNAL DEFINITIONS
   =========================================
   ***************************************** */
#include "setup.h"

#if OBJECT_SYSTEM

#include "classfun.h"
#include "clipsmem.h"
#include "constant.h"
#include "engine.h"
#include "evaluatn.h"
#include "facts.h"
#include "generate.h"
#include "insfun.h"
#include "router.h"
#include "scanner.h"
#include "spclform.h"
#include "utility.h"

#define _INSQUERY_SOURCE_
#include "insquery.h"

/* =========================================
   *****************************************
                   CONSTANTS
   =========================================
   ***************************************** */
#define INSTANCE_SLOT_REF ':'

/* =========================================
   *****************************************
               MACROS AND TYPES
   =========================================
   ***************************************** */
typedef struct query_class
  {
   CLASS_TYPE *cls;
   struct query_class *chain,*nxt;
  } QUERY_CLASS;

typedef struct query_core
  {
   INSTANCE_TYPE **solns;
   EXPRESSION *query,*action;
   INSTANCE_TYPE **soln_set,**soln_bottom;
   int soln_size,soln_cnt;
   DATA_OBJECT *result;
  } QUERY_CORE;
  
typedef struct query_stack
  {
   QUERY_CORE *core;
   struct query_stack *nxt;
  } QUERY_STACK;
  
/* =========================================
   *****************************************
      INTERNALLY VISIBLE FUNCTION HEADERS
   =========================================
   ***************************************** */
#if ANSI_COMPILER
#if ! RUN_TIME
static EXPRESSION *ParseQueryNoAction(EXPRESSION *,char *);
static EXPRESSION *ParseQueryAction(EXPRESSION *,char *);
static EXPRESSION *ParseQueryRestrictions(EXPRESSION *,char *,struct token *);
static int ParseQueryTestExpression(EXPRESSION *,char *);
static int ParseQueryActionExpression(EXPRESSION *,char *,EXPRESSION *);
static VOID ReplaceInstanceVariables(EXPRESSION *,EXPRESSION *,int,int);
static VOID ReplaceSlotReference(EXPRESSION *,EXPRESSION *,
                                 struct FunctionDefinition *,int);
static int IsQueryFunction(EXPRESSION *);
#endif
static VOID PushQueryCore(void);
static VOID PopQueryCore(void);
static QUERY_CORE *FindQueryCore(int);
static QUERY_CLASS *DetermineQueryClasses(EXPRESSION *,char *,int *);
static QUERY_CLASS *FormChain(char *,DATA_OBJECT *);
static VOID DeleteQueryClasses(QUERY_CLASS *);
static int TestForFirstInChain(QUERY_CLASS *,int);
static int TestForFirstInstanceInClass(int,CLASS_TYPE *,QUERY_CLASS *,int);
static VOID TestEntireChain(QUERY_CLASS *,int);
static VOID TestEntireClass(int,CLASS_TYPE *,QUERY_CLASS *,int);
static VOID AddSolution(void);
#else
#if ! RUN_TIME
static EXPRESSION *ParseQueryNoAction();
static EXPRESSION *ParseQueryAction();
static EXPRESSION *ParseQueryRestrictions();
static int ParseQueryTestExpression();
static int ParseQueryActionExpression();
static VOID ReplaceInstanceVariables();
static VOID ReplaceSlotReference();
static int IsQueryFunction();
#endif
static VOID PushQueryCore();
static VOID PopQueryCore();
static QUERY_CORE *FindQueryCore();
static QUERY_CLASS *DetermineQueryClasses();
static QUERY_CLASS *FormChain();
static VOID DeleteQueryClasses();
static int TestForFirstInChain();
static int TestForFirstInstanceInClass();
static VOID TestEntireChain();
static VOID TestEntireClass();
static VOID AddSolution();
#endif
      
/* =========================================
   *****************************************
      EXTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */
globle SYMBOL_HN *QUERY_DELIMETER_SYMBOL = NULL;
   
/* =========================================
   *****************************************
      INTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */
static QUERY_CORE *QueryCore = NULL;
static QUERY_STACK *QueryCoreStack = NULL;
static int AbortQuery = FALSE;
static unsigned char CTID = 0;

/* =========================================
   *****************************************
          EXTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
/****************************************************
  NAME         : SetupQuery
  DESCRIPTION  : Initializes instance query CLIPS
                   functions and parsers
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Sets up kernel functions and parsers
  NOTES        : None
 ****************************************************/
globle VOID SetupQuery()
  {
#if ! RUN_TIME
   QUERY_DELIMETER_SYMBOL = (SYMBOL_HN *) AddSymbol(QUERY_DELIMETER_STRING);
   IncrementSymbolCount(QUERY_DELIMETER_SYMBOL);

   DefineFunction("(query-instance)",'u',
                  PTIF GetQueryInstance,"GetQueryInstance");

   DefineFunction("(query-instance-slot)",'u',
                  PTIF GetQueryInstanceSlot,"GetQueryInstanceSlot");

   DefineFunction("any-instancep",'b',PTIF AnyInstances,"AnyInstances");
   AddFunctionParser("any-instancep",ParseQueryNoAction);

   DefineFunction("find-instance",'u',
                  PTIF QueryFindInstance,"QueryFindInstance");
   AddFunctionParser("find-instance",ParseQueryNoAction);

   DefineFunction("find-all-instances",'u',
                  PTIF QueryFindAllInstances,"QueryFindAllInstances");
   AddFunctionParser("find-all-instances",ParseQueryNoAction);

   DefineFunction("do-for-instance",'u',
                  PTIF QueryDoForInstance,"QueryDoForInstance");
   AddFunctionParser("do-for-instance",ParseQueryAction);

   DefineFunction("do-for-all-instances",'u',
                  PTIF QueryDoForAllInstances,"QueryDoForAllInstances");
   AddFunctionParser("do-for-all-instances",ParseQueryAction);

   DefineFunction("delayed-do-for-all-instances",'u',
                  PTIF DelayedQueryDoForAllInstances,
                  "DelayedQueryDoForAllInstances");
   AddFunctionParser("delayed-do-for-all-instances",ParseQueryAction);
#endif
  }
  
/***************************************************
  NAME         : GetTraversalID
  DESCRIPTION  : Returns a unique integer ID for a
                  traversal into the class hierarchy
  INPUTS       : None
  RETURNS      : The id, or -1 if none available
  SIDE EFFECTS : EvaluationError set when no ids
                   available
  NOTES        : Used for recursive traversals of
                  class hierarchy to assure that a
                  class is only visited once
 ***************************************************/
globle int GetTraversalID()
  {
   register CLASS_TYPE *cls;
   
   if (CTID >= MAX_TRAVERSALS)
     {
      PrintCLIPS(WERROR,"Maximum number of simultaneous class hierarchy traversals\nexceeded ");
      PrintLongInteger(WERROR,(long) MAX_TRAVERSALS);
      PrintCLIPS(WERROR,".  Reduce complexity of nested instance-set queries.\n");
      SetEvaluationError(TRUE);
      return(-1);
     }
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     clearbit(cls->tvids[CTID / BITS_PER_BYTE],CTID % BITS_PER_BYTE);
   return(CTID++);
  }
  
/***************************************************
  NAME         : ReleaseTraversalID
  DESCRIPTION  : Releases an ID for future use
                 Also clears id from all classes
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Old ID released for later reuse
  NOTES        : Releases ID returned by most recent
                   call to GetTraversalID()
 ***************************************************/
globle VOID ReleaseTraversalID()
  {
   CTID--;
  }
  
/***********************************************************
  NAME         : GetQueryInstance
  DESCRIPTION  : Internal function for referring to instance
                    array on instance-queries
  INPUTS       : The caller's result buffer
  RETURNS      : Nothing useful
  SIDE EFFECTS : Caller's result buffer set appropriately
  NOTES        : CLIPS Syntax : ((query-instance) <index>)
 ***********************************************************/
globle VOID GetQueryInstance(result)
  DATA_OBJECT *result;
  {
   QUERY_CORE *core;
   
   core = FindQueryCore(DOPToInteger(GetFirstArgument()));
   result->type = INSTANCE_NAME;
   result->value = (VOID *) 
     core->solns[DOPToInteger(GetFirstArgument()->next_arg)]->name;
  }
  
/***************************************************************************
  NAME         : GetQueryInstanceSlot
  DESCRIPTION  : Internal function for referring to slots of instances in
                    instance array on instance-queries
  INPUTS       : The caller's result buffer
  RETURNS      : Nothing useful
  SIDE EFFECTS : Caller's result buffer set appropriately
  NOTES        : CLIPS Syntax : ((query-instance-slot) <index> <slot-name>)
 **************************************************************************/
globle VOID GetQueryInstanceSlot(result)
  DATA_OBJECT *result;
  {
   INSTANCE_TYPE *ins;
   INSTANCE_SLOT *sp;
   DATA_OBJECT temp;
   QUERY_CORE *core;
   
   result->type = SYMBOL;
   result->value = (VOID *) CLIPSFalseSymbol;
   
   core = FindQueryCore(DOPToInteger(GetFirstArgument()));
   ins = core->solns[DOPToInteger(GetFirstArgument()->next_arg)];
   EvaluateExpression(GetFirstArgument()->next_arg->next_arg,&temp);
   if (temp.type != SYMBOL)
     {
      ExpectedTypeError("get",1,"symbol");
      SetEvaluationError(TRUE);
      return;
     }
   sp = FindInstanceSlot(ins,(SYMBOL_HN *) temp.value);
   if (sp == NULL)
     {
      SlotExistError(ValueToString(temp.value),"instance-set query");
      return;
     }
   if (*(sp->valaddr) != NULL)
     CopyDataObject(result,*(sp->valaddr));
   else
     {
      PrintCLIPS(WERROR,ValueToString(sp->desc->name));
      PrintCLIPS(WERROR," slot unbound in instance ");
      PrintCLIPS(WERROR,ValueToString(ins->name));
      PrintCLIPS(WERROR,".\n");
      SetEvaluationError(TRUE);
     }
  }
  
/* =============================================================================
   =============================================================================
   Following are the instance query functions :
   
     any-instancep         : Determines if any instances satisfy the query
     find-instance         : Finds first (set of) instance(s) which satisfies
                               the query and stores it in a multi-field
     find-all-instances    : Finds all (sets of) instances which satisfy the
                               the query and stores them in a multi-field
     do-for-instance       : Executes a given action for the first (set of)
                               instance(s) which satisfy the query
     do-for-all-instances  : Executes an action for all instances which satisfy
                               the query as they are found
     delayed-do-for-all-instances : Same as above - except that the list of instances
                               which satisfy the query is formed before any
                               actions are executed
                               
     Instance candidate search algorithm :
     
     All permutations of first restriction class instances with other
       restriction class instances (Rightmost are varied first)
     All permutations of first restriction class's subclasses' instances with
       other restriction class instances.
     And  so on...
     
     For any one class, instances are examined in the order they were defined
     
     Example :
     (defclass a (is-a standard-user))
     (defclass b (is-a standard-user))
     (defclass c (is-a standard-user))
     (defclass d (is-a a b))
     (make-instance a1 of a)
     (make-instance a2 of a)
     (make-instance b1 of b)
     (make-instance b2 of b)
     (make-instance c1 of c)
     (make-instance c2 of c)
     (make-instance d1 of d)
     (make-instance d2 of d)
     
     (any-instancep ((?a a b) (?b c)) <query>)
     
     The permutations (?a ?b) would be examined in the following order :
     
     (a1 c1),(a1 c2),(a2 c1),(a2 c2),(d1 c1),(d1 c2),(d2 c1),(d2 c2),
     (b1 c1),(b1 c2),(b2 c1),(b2 c2),(d1 c1),(d1 c2),(d2 c1),(d2 c2)
     
     Notice the duplication because d is a subclass of both and a and b.
   =============================================================================
   ============================================================================= */
   
/******************************************************************************
  NAME         : AnyInstances
  DESCRIPTION  : Determines if there any existing instances which satisfy
                   the query
  INPUTS       : None
  RETURNS      : TRUE if the query is satisfied, FALSE otherwise
  SIDE EFFECTS : The query class-expressions are evaluated once,
                   and the query boolean-expression is evaluated
                   zero or more times (depending on instance restrictions
                   and how early the expression evaulates to TRUE - if at all).
  NOTES        : CLIPS Syntax : See ParseQueryNoAction()
 ******************************************************************************/
globle BOOLEAN AnyInstances()
  {
   QUERY_CLASS *qclasses;
   int rcnt;
   int TestResult;
   
   qclasses = DetermineQueryClasses(GetFirstArgument()->next_arg,
                                      "any-instancep",&rcnt);
   if (qclasses == NULL)
     return(FALSE);
   PushQueryCore();
   QueryCore = get_struct(query_core);
   QueryCore->solns = (INSTANCE_TYPE **) gm2((int) (sizeof(INSTANCE_TYPE *) * rcnt));
   QueryCore->query = GetFirstArgument();
   TestResult = TestForFirstInChain(qclasses,0);
   AbortQuery = FALSE;
   rm((VOID *) QueryCore->solns,(int) (sizeof(INSTANCE_TYPE *) * rcnt));
   rtn_struct(query_core,QueryCore);
   PopQueryCore();
   DeleteQueryClasses(qclasses);
   return(TestResult);
  }
  
/******************************************************************************
  NAME         : QueryFindInstance
  DESCRIPTION  : Finds the first set of instances which satisfy the query and
                   stores their names in the user's multi-field variable
  INPUTS       : Caller's result buffer
  RETURNS      : TRUE if the query is satisfied, FALSE otherwise
  SIDE EFFECTS : The query class-expressions are evaluated once,
                   and the query boolean-expression is evaluated
                   zero or more times (depending on instance restrictions
                   and how early the expression evaulates to TRUE - if at all).
  NOTES        : CLIPS Syntax : See ParseQueryNoAction()
 ******************************************************************************/
globle VOID QueryFindInstance(result)
  DATA_OBJECT *result;
  {
   QUERY_CLASS *qclasses;
   int rcnt,i;
   
   result->type = MULTIFIELD;
   result->begin = 0;
   result->end = -1;
   qclasses = DetermineQueryClasses(GetFirstArgument()->next_arg,
                                      "find-instance",&rcnt);
   if (qclasses == NULL)
     {
      result->value = (VOID *) CreateMultifield(0);
      return;
     }
   PushQueryCore();
   QueryCore = get_struct(query_core);
   QueryCore->solns = (INSTANCE_TYPE **) 
                      gm2((int) (sizeof(INSTANCE_TYPE *) * rcnt));
   QueryCore->query = GetFirstArgument();
   if (TestForFirstInChain(qclasses,0) == TRUE)
     {
      result->value = (VOID *) CreateMultifield(rcnt);
      result->end = rcnt-1;
      for (i = 0 ; i < rcnt ; i++)
        {
         ((struct fact *) result->value)->atoms[i].type = INSTANCE_NAME;
         ((struct fact *) result->value)->atoms[i].value = 
                          (VOID *) QueryCore->solns[i]->name;
        }
     }
   else
      result->value = (VOID *) CreateMultifield(0);
   AbortQuery = FALSE;
   rm((VOID *) QueryCore->solns,(int) (sizeof(INSTANCE_TYPE *) * rcnt));
   rtn_struct(query_core,QueryCore);
   PopQueryCore();
   DeleteQueryClasses(qclasses);
  }
  
/******************************************************************************
  NAME         : QueryFindAllInstances
  DESCRIPTION  : Finds all sets of instances which satisfy the query and
                   stores their names in the user's multi-field variable
                   
                 The sets are stored sequentially :
                 
                   Number of sets = (Multi-field length) / (Set length)
                 
                 The first set is if the first (set length) atoms of the
                   multi-field variable, and so on.
  INPUTS       : Caller's result buffer
  RETURNS      : Nothing useful
  SIDE EFFECTS : The query class-expressions are evaluated once,
                   and the query boolean-expression is evaluated
                   once for every instance set.
  NOTES        : CLIPS Syntax : See ParseQueryNoAction()
 ******************************************************************************/
globle VOID QueryFindAllInstances(result)
  DATA_OBJECT *result;
  {
   QUERY_CLASS *qclasses;
   int rcnt;
   register int i,j;
   
   result->type = MULTIFIELD;
   result->begin = 0;
   result->end = -1;
   qclasses = DetermineQueryClasses(GetFirstArgument()->next_arg,
                                      "find-all-instances",&rcnt);
   if (qclasses == NULL)
     {
      result->value = (VOID *) CreateMultifield(0);
      return;
     }
   PushQueryCore();
   QueryCore = get_struct(query_core);
   QueryCore->solns = (INSTANCE_TYPE **) gm2((int) (sizeof(INSTANCE_TYPE *) * rcnt));
   QueryCore->query = GetFirstArgument();
   QueryCore->action = NULL;
   QueryCore->soln_set = NULL;
   QueryCore->soln_size = rcnt;
   QueryCore->soln_cnt = 0;
   TestEntireChain(qclasses,0);
   AbortQuery = FALSE;
   result->value = (VOID *) CreateMultifield(QueryCore->soln_cnt * rcnt);
   if (QueryCore->soln_cnt > 0)
     {
      while (QueryCore->soln_set != NULL)
        {
         for (i = 0 , j = result->end+1 ; i < rcnt ; i++ , j++)
           {
            ((struct fact *) result->value)->atoms[j].type = INSTANCE_NAME;
            ((struct fact *) result->value)->atoms[j].value = 
                             (VOID *) QueryCore->soln_set[i]->name;
           }
         result->end = j-1;
         QueryCore->soln_bottom = QueryCore->soln_set;
         QueryCore->soln_set = (INSTANCE_TYPE **) QueryCore->soln_set[rcnt];
         rm((VOID *) QueryCore->soln_bottom,
            (int) (sizeof(INSTANCE_TYPE *) * (rcnt+1)));
        }
     }
   rm((VOID *) QueryCore->solns,(int) (sizeof(INSTANCE_TYPE *) * rcnt));
   rtn_struct(query_core,QueryCore);
   PopQueryCore();
   DeleteQueryClasses(qclasses);
  }
  
/******************************************************************************
  NAME         : QueryDoForInstance
  DESCRIPTION  : Finds the first set of instances which satisfy the query and
                   executes a user-action with that set
  INPUTS       : None
  RETURNS      : Caller's result buffer
  SIDE EFFECTS : The query class-expressions are evaluated once,
                   and the query boolean-expression is evaluated
                   zero or more times (depending on instance restrictions
                   and how early the expression evaulates to TRUE - if at all).
                   Also the action expression is executed zero or once.
                 Caller's result buffer holds result of user-action
  NOTES        : CLIPS Syntax : See ParseQueryAction()
 ******************************************************************************/
globle VOID QueryDoForInstance(result)
  DATA_OBJECT *result;
  {
   QUERY_CLASS *qclasses;
   int rcnt;

   result->type = SYMBOL;
   result->value = (VOID *) CLIPSFalseSymbol;
   qclasses = DetermineQueryClasses(GetFirstArgument()->next_arg->next_arg,
                                      "do-for-instance",&rcnt);
   if (qclasses == NULL)
     return;
   PushQueryCore();
   QueryCore = get_struct(query_core);
   QueryCore->solns = (INSTANCE_TYPE **) gm2((int) (sizeof(INSTANCE_TYPE *) * rcnt));
   QueryCore->query = GetFirstArgument();
   QueryCore->action = GetFirstArgument()->next_arg;
   if (TestForFirstInChain(qclasses,0) == TRUE)
     EvaluateExpression(QueryCore->action,result);
   AbortQuery = FALSE;
   rm((VOID *) QueryCore->solns,(int) (sizeof(INSTANCE_TYPE *) * rcnt));
   rtn_struct(query_core,QueryCore);
   PopQueryCore();
   DeleteQueryClasses(qclasses);
  }
  
/******************************************************************************
  NAME         : QueryDoForAllInstances
  DESCRIPTION  : Finds all sets of instances which satisfy the query and
                   executes a user-function for each set as it is found
  INPUTS       : Caller's result buffer
  RETURNS      : Nothing useful
  SIDE EFFECTS : The query class-expressions are evaluated once,
                   and the query boolean-expression is evaluated
                   once for every instance set.  Also, the action is
                   executed for every instance set.
                 Caller's result buffer holds result of last action executed.
  NOTES        : CLIPS Syntax : See ParseQueryAction()
 ******************************************************************************/
globle VOID QueryDoForAllInstances(result)
  DATA_OBJECT *result;
  {
   QUERY_CLASS *qclasses;
   int rcnt;

   result->type = SYMBOL;
   result->value = (VOID *) CLIPSFalseSymbol;
   qclasses = DetermineQueryClasses(GetFirstArgument()->next_arg->next_arg,
                                      "do-for-all-instances",&rcnt);
   if (qclasses == NULL)
     return;
   PushQueryCore();
   QueryCore = get_struct(query_core);
   QueryCore->solns = (INSTANCE_TYPE **) gm2((int) (sizeof(INSTANCE_TYPE *) * rcnt));
   QueryCore->query = GetFirstArgument();
   QueryCore->action = GetFirstArgument()->next_arg;
   QueryCore->result = result;
   TestEntireChain(qclasses,0);
   AbortQuery = FALSE;
   rm((VOID *) QueryCore->solns,(int) (sizeof(INSTANCE_TYPE *) * rcnt));
   rtn_struct(query_core,QueryCore);
   PopQueryCore();
   DeleteQueryClasses(qclasses);
  }
  
/******************************************************************************
  NAME         : DelayedQueryDoForAllInstances
  DESCRIPTION  : Finds all sets of instances which satisfy the query and
                   and exceutes a user-action for each set
                   
                 This function differs from QueryDoForAllInstances() in
                   that it forms the complete list of query satisfactions
                   BEFORE executing any actions. 
  INPUTS       : Caller's result buffer
  RETURNS      : TRUE if the query is satisfied, FALSE otherwise
  SIDE EFFECTS : The query class-expressions are evaluated once,
                   and the query boolean-expression is evaluated
                   once for every instance set.  The action is executed
                   for evry query satisfaction.
                 Caller's result buffer holds result of last action executed.
  NOTES        : CLIPS Syntax : See ParseQueryNoAction()
 ******************************************************************************/
globle VOID DelayedQueryDoForAllInstances(result)
  DATA_OBJECT *result;
  {
   QUERY_CLASS *qclasses;
   int rcnt;
   register int i;
   
   result->type = SYMBOL;
   result->value = (VOID *) CLIPSFalseSymbol;
   qclasses = DetermineQueryClasses(GetFirstArgument()->next_arg->next_arg,
                                      "delayed-do-for-all-instances",&rcnt);
   if (qclasses == NULL)
     return;
   PushQueryCore();
   QueryCore = get_struct(query_core);
   QueryCore->solns = (INSTANCE_TYPE **) gm2((int) (sizeof(INSTANCE_TYPE *) * rcnt));
   QueryCore->query = GetFirstArgument();
   QueryCore->action = NULL;
   QueryCore->soln_set = NULL;
   QueryCore->soln_size = rcnt;
   QueryCore->soln_cnt = 0;
   TestEntireChain(qclasses,0);
   AbortQuery = FALSE;
   if (QueryCore->soln_cnt > 0)
     {
      QueryCore->action = GetFirstArgument()->next_arg;
      while (QueryCore->soln_set != NULL)
        {
         for (i = 0 ; i < rcnt ; i++)
           QueryCore->solns[i] = QueryCore->soln_set[i];
         QueryCore->soln_bottom = QueryCore->soln_set;
         QueryCore->soln_set = (INSTANCE_TYPE **) QueryCore->soln_set[rcnt];
         rm((VOID *) QueryCore->soln_bottom,(int) (sizeof(INSTANCE_TYPE *) * (rcnt+1)));
         EvaluateExpression(QueryCore->action,result);
         if (HaltExecution == TRUE)
           {
            while (QueryCore->soln_set != NULL)
              {
               QueryCore->soln_bottom = QueryCore->soln_set;
               QueryCore->soln_set = (INSTANCE_TYPE **) QueryCore->soln_set[rcnt];
               rm((VOID *) QueryCore->soln_bottom,
                  (int) (sizeof(INSTANCE_TYPE *) * (rcnt+1)));
              }
            break;
           }
        }
     }
   rm((VOID *) QueryCore->solns,(int) (sizeof(INSTANCE_TYPE *) * rcnt));
   rtn_struct(query_core,QueryCore);
   PopQueryCore();
   DeleteQueryClasses(qclasses);
  }
  
/* =========================================
   *****************************************
          INTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
   
#if ! RUN_TIME

/***********************************************************************
  NAME         : ParseQueryNoAction
  DESCRIPTION  : Parses the following functions :
                   (any-instancep)
                   (find-first-instance)
                   (find-all-instances)
  INPUTS       : 1) The address of the top node of the query function
                 2) The logical name of the input
  RETURNS      : The completed expression chain, or NULL on errors
  SIDE EFFECTS : The expression chain is extended, or the "top" node
                 is deleted on errors
  NOTES        : CLIPS Syntax : 
  
                 (<function> <query-block>)
                 
                 <query-block>  :== (<instance-var>+) <query-expression>
                 <instance-var> :== (<var-name> <class-name>+)
                 
                 Parses into following form :
                 
                 <query-function>
                      |
                      V
                 <query-expression>  ->
                              
                 <class-1a> -> <class-1b> -> (QDS) ->
                 
                 <class-2a> -> <class-2b> -> (QDS) -> ... 
 ***********************************************************************/
static EXPRESSION *ParseQueryNoAction(top,read_source)
  EXPRESSION *top;
  char *read_source;
  {
   EXPRESSION *insvars;
   struct token inp_tkn;

   insvars = ParseQueryRestrictions(top,read_source,&inp_tkn);
   if (insvars == NULL)
     return(NULL);
   IncrementIndentDepth(3);
   PPCRAndIndent();
   if (ParseQueryTestExpression(top,read_source) == FALSE)
     {
      DecrementIndentDepth(3);
      ReturnExpression(insvars);
      return(NULL);
     }
   DecrementIndentDepth(3);
   GetToken(read_source,&inp_tkn);
   if (GetType(inp_tkn) != RPAREN)      
     {
      SyntaxErrorMessage("instance-set query function");
      ReturnExpression(top);
      ReturnExpression(insvars);
      SetEvaluationError(TRUE);
      return(NULL);
     }
   ReplaceInstanceVariables(insvars,top->arg_list,TRUE,0);
   ReturnExpression(insvars);
   return(top);
  }
  
/***********************************************************************
  NAME         : ParseQueryAction
  DESCRIPTION  : Parses the following functions :
                   (do-for-first-instance)
                   (do-for-all-instances)
                   (delayed-do-for-all-instances)
  INPUTS       : 1) The address of the top node of the query function
                 2) The logical name of the input
  RETURNS      : The completed expression chain, or NULL on errors
  SIDE EFFECTS : The expression chain is extended, or the "top" node
                 is deleted on errors
  NOTES        : CLIPS Syntax : 
  
                 (<function> <query-block> <query-action>)
                 
                 <query-block>  :== (<instance-var>+) <query-expression>
                 <instance-var> :== (<var-name> <class-name>+)
                 
                 Parses into following form :
                 
                 <query-function>
                      |
                      V
                 <query-expression> -> <query-action>  ->
                              
                 <class-1a> -> <class-1b> -> (QDS) ->
                 
                 <class-2a> -> <class-2b> -> (QDS) -> ... 
 ***********************************************************************/
static EXPRESSION *ParseQueryAction(top,read_source)
  EXPRESSION *top;
  char *read_source;
  {
   EXPRESSION *insvars;
   struct token inp_tkn;
   
   insvars = ParseQueryRestrictions(top,read_source,&inp_tkn);
   if (insvars == NULL)
     return(NULL);
   IncrementIndentDepth(3);
   PPCRAndIndent();
   if (ParseQueryTestExpression(top,read_source) == FALSE)
     {
      DecrementIndentDepth(3);
      ReturnExpression(insvars);
      return(NULL);
     }
   PPCRAndIndent();
   if (ParseQueryActionExpression(top,read_source,insvars) == FALSE)
     {
      DecrementIndentDepth(3);
      ReturnExpression(insvars);
      return(NULL);
     }
   DecrementIndentDepth(3);
   GetToken(read_source,&inp_tkn);
   if (GetType(inp_tkn) != RPAREN)      
     {
      SyntaxErrorMessage("instance-set query function");
      ReturnExpression(top);
      ReturnExpression(insvars);
      SetEvaluationError(TRUE);
      return(NULL);
     }
   ReplaceInstanceVariables(insvars,top->arg_list,TRUE,0);
   ReplaceInstanceVariables(insvars,top->arg_list->next_arg,FALSE,0);
   ReturnExpression(insvars);
   return(top);
  }

/***************************************************************
  NAME         : ParseQueryRestrictions
  DESCRIPTION  : Parses the class restrictions for a query
  INPUTS       : 1) The top node of the query expression
                 2) The logical name of the input
                 3) Caller's token buffer
  RETURNS      : The instance-variable expressions
  SIDE EFFECTS : Entire query expression deleted on errors
                 Nodes allocated for restrictions and instance
                   variable expressions
                 Class restrictions attached to query-expression
                   as arguments
  NOTES        : Expects top != NULL
 ***************************************************************/
static EXPRESSION *ParseQueryRestrictions(top,read_source,inp_tkn)
  EXPRESSION *top;
  char *read_source;
  struct token *inp_tkn;
  {
   EXPRESSION *insvars = NULL,*linsvars = NULL,
              *cexp,*lcexp,*tmp,*last_one = NULL;
   int error = FALSE;

   SavePPBuffer(" ");
   GetToken(read_source,inp_tkn);
   if (inp_tkn->type != LPAREN)
     {
      SyntaxErrorMessage("instance-set query function");
      ReturnExpression(top);
      SetEvaluationError(TRUE);
      return(NULL);
     }
   GetToken(read_source,inp_tkn);
   if (inp_tkn->type != LPAREN)
     {
      SyntaxErrorMessage("instance-set query function");
      ReturnExpression(top);
      SetEvaluationError(TRUE);
      return(NULL);
     }
   while (inp_tkn->type == LPAREN)
     {
      GetToken(read_source,inp_tkn);
      if (inp_tkn->type != BWORD)
        {
         SyntaxErrorMessage("instance-set query function");
         ReturnExpression(top);
         ReturnExpression(insvars);
         SetEvaluationError(TRUE);
         return(NULL);
        }
      tmp = insvars;
      while (tmp != NULL)
        {
         if (tmp->value == inp_tkn->value)
           {
            PrintCLIPS(WERROR,"Duplicate instance member variable name in function ");
            PrintCLIPS(WERROR,ValueToString(ExpressionFunctionCallName(top)));
            PrintCLIPS(WERROR,".\n");
            ReturnExpression(top);
            ReturnExpression(insvars);
            SetEvaluationError(TRUE);
            return(NULL);
           }
         tmp = tmp->next_arg;
        }
      tmp = GenConstant(BWORD,inp_tkn->value);
      if (insvars == NULL)
        insvars = tmp;
      else
        linsvars->next_arg = tmp;
      linsvars = tmp;
      SavePPBuffer(" ");
      cexp = ArgumentParse(read_source,&error);
      if (error == TRUE)
        {
         ReturnExpression(top);
         ReturnExpression(insvars);
         return(NULL);
        }
      if (cexp == NULL)
        {
         SyntaxErrorMessage("instance-set query function");
         ReturnExpression(top);
         ReturnExpression(insvars);
         SetEvaluationError(TRUE);
         return(NULL);
        }
      lcexp = cexp;
      SavePPBuffer(" ");
      while ((tmp = ArgumentParse(read_source,&error)) != NULL)
        {
         if (error == TRUE)
           {
            ReturnExpression(cexp);
            ReturnExpression(top);
            ReturnExpression(insvars);
            return(NULL);
           }
         lcexp->next_arg = tmp;
         lcexp = tmp;
         SavePPBuffer(" ");
        }
      PPBackup();
      PPBackup();
      SavePPBuffer(")");
      tmp = GenConstant(SYMBOL,(VOID *) QUERY_DELIMETER_SYMBOL);
      lcexp->next_arg = tmp;
      lcexp = tmp;
      if (top->arg_list == NULL)
        top->arg_list = cexp;
      else
        last_one->next_arg = cexp;
      last_one = lcexp;
      SavePPBuffer(" ");
      GetToken(read_source,inp_tkn);
     }
   if (inp_tkn->type != RPAREN)
     {
      SyntaxErrorMessage("instance-set query function");
      ReturnExpression(top);
      ReturnExpression(insvars);
      SetEvaluationError(TRUE);
      return(NULL);
     }
   PPBackup();
   PPBackup();
   SavePPBuffer(")");
   return(insvars);
  }
  
/*************************************************************
  NAME         : ParseQueryTestExpression
  DESCRIPTION  : Parses the test-expression for a query
  INPUTS       : 1) The top node of the query expression
                 2) The logical name of the input
  RETURNS      : TRUE if all OK, FALSE otherwise
  SIDE EFFECTS : Entire query-expression deleted on errors
                 Nodes allocated for new expression
                 Test shoved in front of class-restrictions on
                    query argument list
  NOTES        : Expects top != NULL
 *************************************************************/
static int ParseQueryTestExpression(top,read_source)
  EXPRESSION *top;
  char *read_source;
  {
   EXPRESSION *qtest;
   int error;
   struct BindInfo *oldBindList;
   
   error = FALSE;
   oldBindList = GetParsedBindNames();
   SetParsedBindNames(NULL);
   qtest = ArgumentParse(read_source,&error);
   if (error == TRUE)
     {
      SetParsedBindNames(oldBindList);
      ReturnExpression(top);
      return(FALSE);
     }
   if (qtest == NULL)
     {
      SetParsedBindNames(oldBindList);
      SyntaxErrorMessage("instance-set query function");
      ReturnExpression(top);
      SetEvaluationError(TRUE);
      return(FALSE);
     }
   qtest->next_arg = top->arg_list;
   top->arg_list = qtest;
   if (ParsedBindNamesEmpty() == CLIPS_FALSE)
     {
      ClearParsedBindNames();
      SetParsedBindNames(oldBindList);
      PrintCLIPS(WERROR,"Binds are not allowed in instance-set query in function ");
      PrintCLIPS(WERROR,ValueToString(ExpressionFunctionCallName(top)));
      PrintCLIPS(WERROR,".\n");
      ReturnExpression(top);
      SetEvaluationError(TRUE);
      return(FALSE);
     }
   SetParsedBindNames(oldBindList);
   return(TRUE);
  }
  
/*************************************************************
  NAME         : ParseQueryActionExpression
  DESCRIPTION  : Parses the action-expression for a query
  INPUTS       : 1) The top node of the query expression
                 2) The logical name of the input
                 3) List of query parameters
  RETURNS      : TRUE if all OK, FALSE otherwise
  SIDE EFFECTS : Entire query-expression deleted on errors
                 Nodes allocated for new expression
                 Action shoved in front of class-restrictions
                    and in back of test-expression on query
                    argument list
  NOTES        : Expects top != NULL && top->arg_list != NULL
 *************************************************************/
static int ParseQueryActionExpression(top,read_source,insvars)
  EXPRESSION *top;
  char *read_source;
  EXPRESSION *insvars;
  {
   EXPRESSION *qaction,*tinsvar;
   int error;
   struct BindInfo *oldBindList,*newBindList,*prev;
   
   error = FALSE;
   oldBindList = GetParsedBindNames();
   SetParsedBindNames(NULL);
   qaction = ArgumentParse(read_source,&error);
   if (error == TRUE)
     {
      SetParsedBindNames(oldBindList);
      ReturnExpression(top);
      return(FALSE);
     }
   if (qaction == NULL)
     {
      SetParsedBindNames(oldBindList);
      SyntaxErrorMessage("instance-set query function");
      ReturnExpression(top);
      SetEvaluationError(TRUE);
      return(FALSE);
     }
   qaction->next_arg = top->arg_list->next_arg;
   top->arg_list->next_arg = qaction;
   newBindList = GetParsedBindNames();
   prev = NULL;
   while (newBindList != NULL)
     {
      tinsvar = insvars;
      while (tinsvar != NULL)
        {
         if (tinsvar->value == (VOID *) newBindList->name)
           {
            ClearParsedBindNames();
            SetParsedBindNames(oldBindList);
            PrintCLIPS(WERROR,"Cannot rebind instance-set member variable ");
            PrintCLIPS(WERROR,ValueToString(tinsvar->value));
            PrintCLIPS(WERROR," in function ");
            PrintCLIPS(WERROR,ValueToString(ExpressionFunctionCallName(top)));
            PrintCLIPS(WERROR,".\n");
            ReturnExpression(top);
            SetEvaluationError(TRUE);
            return(FALSE);
           }
         tinsvar = tinsvar->next_arg;
        }
      prev = newBindList;
      newBindList = newBindList->next;
     }
   if (prev == NULL)
     SetParsedBindNames(oldBindList);
   else
     prev->next = oldBindList;
   return(TRUE);
  }
  
/***********************************************************************************
  NAME         : ReplaceInstanceVariables
  DESCRIPTION  : Replaces all references to instance-variables within an
                   instance query-function with function calls to query-instance
                   (which references the instance array at run-time)
  INPUTS       : 1) The instance-variable list
                 2) A boolean expression containing variable references
                 3) A flag indicating whether to allow slot references of the type
                    <instance-query-variable>:<slot-name> for direct slot access
                    or not
                 4) Nesting depth of query functions
  RETURNS      : Nothing useful
  SIDE EFFECTS : If a BWORD node is found and is on the list of instance
                   variables, it is replaced with a query-instance function call.
  NOTES        : Other BWORD(S) are left alone for replacement by other
                   parsers.  This implies that a user may use defgeneric,
                   defrule, and defmessage-handler variables within a query-function
                   where they do not conflict with instance-variable names.
 ***********************************************************************************/
static VOID ReplaceInstanceVariables(vlist,bexp,sdirect,ndepth)
  EXPRESSION *vlist,*bexp;
  int sdirect,ndepth;
  {
   EXPRESSION *eptr;
   struct FunctionDefinition *rindx_func,*rslot_func;
   int posn;
   
   rindx_func = FindFunction("(query-instance)");
   rslot_func = FindFunction("(query-instance-slot)");
   while (bexp != NULL)
     {
      if (bexp->type == BWORD)
        {
         eptr = vlist;
         posn = 0;
         while ((eptr != NULL) ? (eptr->value != bexp->value) : FALSE)
           {
            eptr = eptr->next_arg;
            posn++;
           }
         if (eptr != NULL)
           {
            bexp->type = FCALL;
            bexp->value = (VOID *) rindx_func;
            eptr = GenConstant(INTEGER,(VOID *) AddLong((long) ndepth));
            eptr->next_arg = GenConstant(INTEGER,(VOID *) AddLong((long) posn));
            bexp->arg_list = eptr;
           }
         else if (sdirect == TRUE)
           ReplaceSlotReference(vlist,bexp,rslot_func,ndepth);
        }
      if (bexp->arg_list != NULL)
        {
         if (IsQueryFunction(bexp))
           ReplaceInstanceVariables(vlist,bexp->arg_list,sdirect,ndepth+1);
         else
           ReplaceInstanceVariables(vlist,bexp->arg_list,sdirect,ndepth);
        }
      bexp = bexp->next_arg;
     }
  }

/*************************************************************************
  NAME         : ReplaceSlotReference
  DESCRIPTION  : Replaces instance-set query function variable
                   references of the form: <instance-variable>:<slot-name>
                   with function calls to get these instance-slots at run
                   time
  INPUTS       : 1) The instance-set variable list
                 2) The expression containing the variable
                 3) The address of the instance slot access function
                 4) Nesting depth of query functions
  RETURNS      : Nothing useful
  SIDE EFFECTS : If the variable is a slot reference, then it is replaced
                   with the appropriate function-call.
  NOTES        : None
 *************************************************************************/
static VOID ReplaceSlotReference(vlist,exp,func,ndepth)
  EXPRESSION *vlist,*exp;
  struct FunctionDefinition *func;
  int ndepth;
  {
   int len,posn,oldpp;
   register int i;
   register char *str;
   EXPRESSION *eptr;
   struct token itkn;
   
   str = ValueToString(exp->value);
   len =  strlen(str);
   if (len < 3)
     return;
   for (i = len-2 ; i >= 1 ; i--)
     {
      if ((str[i] == INSTANCE_SLOT_REF) ? (i > 0) : FALSE)
        {
         eptr = vlist;
         posn = 0;
         while ((eptr == NULL) ? FALSE : 
                (strncmp(ValueToString(eptr->value),str,i) != 0))
           {
            eptr = eptr->next_arg;
            posn++;
           }
         if (eptr != NULL)
           {
            OpenStringSource("query-var",str+i+1,0);
            oldpp = GetPPBufferStatus();
            SetPPBufferStatus(OFF);
            GetToken("query-var",&itkn);
            SetPPBufferStatus(oldpp);
            CloseStringSource("query-var");
            exp->type = FCALL;
            exp->value = (VOID *) func;
            exp->arg_list = GenConstant(INTEGER,(VOID *) AddLong((long) ndepth));
            exp->arg_list->next_arg = 
              GenConstant(INTEGER,(VOID *) AddLong((long) posn));
            exp->arg_list->next_arg->next_arg = GenConstant(itkn.type,itkn.value);
            break;
           }
        }
     }
  }

/********************************************************************
  NAME         : IsQueryFunction
  DESCRIPTION  : Determines if an expression is a query function call
  INPUTS       : The expression
  RETURNS      : TRUE if query function call, FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : None
 ********************************************************************/
static int IsQueryFunction(exp)
  EXPRESSION *exp;
  {
   int (*fptr)(VOID_ARG);
   
   if (exp->type != FCALL)
     return(FALSE);
   fptr = (int (*)(VOID_ARG)) ExpressionFunctionPointer(exp);
   if (fptr == (int (*)(VOID_ARG)) AnyInstances)
     return(TRUE);
   if (fptr == (int (*)(VOID_ARG)) QueryFindInstance)
     return(TRUE);
   if (fptr == (int (*)(VOID_ARG)) QueryFindAllInstances)
     return(TRUE);
   if (fptr == (int (*)(VOID_ARG)) QueryDoForInstance)
     return(TRUE);
   if (fptr == (int (*)(VOID_ARG)) QueryDoForAllInstances)
     return(TRUE);
   if (fptr == (int (*)(VOID_ARG)) DelayedQueryDoForAllInstances)
     return(TRUE);
   return(FALSE);
  }
  
#endif

/*******************************************************
  NAME         : PushQueryCore
  DESCRIPTION  : Pushes the current QueryCore onto stack
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Allocates new stack node and changes
                   QueryCoreStack
  NOTES        : None
 *******************************************************/
static VOID PushQueryCore()
  {
   QUERY_STACK *qptr;
   
   qptr = get_struct(query_stack);
   qptr->core = QueryCore;
   qptr->nxt = QueryCoreStack;
   QueryCoreStack = qptr;
  }
  
/******************************************************
  NAME         : PopQueryCore
  DESCRIPTION  : Pops top of QueryCore stack and
                   restores QueryCore to this core
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Stack node deallocated, QueryCoreStack
                   changed and QueryCore reset
  NOTES        : Assumes stack is not empty
 ******************************************************/
static VOID PopQueryCore()
  {
   QUERY_STACK *qptr;
   
   QueryCore = QueryCoreStack->core;
   qptr = QueryCoreStack;
   QueryCoreStack = QueryCoreStack->nxt;
   rtn_struct(query_stack,qptr);
  }
  
/***************************************************
  NAME         : FindQueryCore
  DESCRIPTION  : Looks up a QueryCore Stack Frame
                    Depth 0 is current frame
                    1 is next deepest, etc.
  INPUTS       : Depth
  RETURNS      : Address of query core stack frame
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
static QUERY_CORE *FindQueryCore(depth)
  int depth;
  {
   QUERY_STACK *qptr;
   
   if (depth == 0)
     return(QueryCore);
   qptr = QueryCoreStack;
   while (depth > 1)
     {
      qptr = qptr->nxt;
      depth--;
     }
   return(qptr->core);
  }
  
/**********************************************************
  NAME         : DetermineQueryClasses
  DESCRIPTION  : Builds a list of classes to be used in
                   instance queries - uses parse form.
  INPUTS       : 1) The parse class expression chain
                 2) The name of the function being executed
                 3) Caller's buffer for restriction count
                    (# of separate lists)
  RETURNS      : The query list, or NULL on errors
  SIDE EFFECTS : Memory allocated for list
                 Busy count incremented for all classes
  NOTES        : Each restriction is linked by nxt pointer,
                   multiple classes in a restriction are
                   linked by the chain pointer.
                 Rcnt caller's buffer is set to reflect the
                   total number of chains
                 Assumes cexp is not NULL and that each
                   restriction chain is terminated with
                   the QUERY_DELIMITER_SYMBOL "(QDS)"
 **********************************************************/
static QUERY_CLASS *DetermineQueryClasses(cexp,func,rcnt)
  EXPRESSION *cexp;
  char *func;
  int *rcnt;
  {
   QUERY_CLASS *clist = NULL,*cnxt = NULL,*cchain = NULL,*tmp;
   int new_list = FALSE;
   DATA_OBJECT temp;
   
   *rcnt = 0;
   while (cexp != NULL)
     {
      if (EvaluateExpression(cexp,&temp) == TRUE)
        {
         DeleteQueryClasses(clist);
         return(NULL);
        }
      if ((temp.type == SYMBOL) && (temp.value == (VOID *) QUERY_DELIMETER_SYMBOL))
        {
         new_list = TRUE;
         (*rcnt)++;
        }
      else if ((tmp = FormChain(func,&temp)) != NULL)
        {
         if (clist == NULL)
           clist = cnxt = cchain = tmp;
         else if (new_list == TRUE)
           {
            new_list = FALSE;
            cnxt->nxt = tmp;
            cnxt = cchain = tmp;
           }
         else
           cchain->chain = tmp;
         while (cchain->chain != NULL)
           cchain = cchain->chain;
        }
      else
        {
         SyntaxErrorMessage("instance-set query class restrictions");
         DeleteQueryClasses(clist);
         SetEvaluationError(TRUE);
         return(NULL);
        }
      cexp = cexp->next_arg;
     }
   return(clist);
  }
  
/*************************************************************
  NAME         : FormChain
  DESCRIPTION  : Builds a list of classes to be used in
                   instance queries - uses parse form.
  INPUTS       : 1) Name of calling function for error msgs
                 2) Data object - must be a symbol or a
                      multifield value containing all symbols
                 The symbols must be names of existing classes
  RETURNS      : The query chain, or NULL on errors
  SIDE EFFECTS : Memory allocated for chain
                 Busy count incremented for all classes
  NOTES        : None
 *************************************************************/
static QUERY_CLASS *FormChain(func,val)
  char *func;
  DATA_OBJECT *val;
  {
   CLASS_TYPE *cls;
   QUERY_CLASS *head,*bot,*tmp;
   register int i,end;
   
   if (val->type == SYMBOL)
     {
      cls = FindDefclassBySymbol((SYMBOL_HN *) val->value);
      if (cls == NULL)
        {
         ClassExistError(func,ValueToString(val->value));
         return(NULL);
        }
      cls->busy++;
      head = get_struct(query_class);
      head->cls = cls;
      head->chain = NULL;
      head->nxt = NULL;
      return(head);
     }
   if (val->type == MULTIFIELD)
     {
      head = bot = NULL;
      end = GetpDOEnd(val);
      for (i = GetpDOBegin(val) ; i <= end ; i++)
        {
         if (GetMFType(val->value,i) != SYMBOL)
           {
            DeleteQueryClasses(head);
            return(NULL);
           }
         cls = FindDefclassBySymbol((SYMBOL_HN *) GetMFValue(val->value,i));
         if (cls == NULL)
           {
            ClassExistError(func,ValueToString(GetMFValue(val->value,i)));
            DeleteQueryClasses(head);
            return(NULL);
           }
         cls->busy++;
         tmp = get_struct(query_class);
         tmp->cls = cls;
         tmp->chain = NULL;
         tmp->nxt = NULL;
         if (head == NULL)
           head = tmp;
         else
           bot->chain = tmp;
         bot = tmp;
        }
      return(head);
     }
   return(NULL);
  }

/******************************************************
  NAME         : DeleteQueryClasses
  DESCRIPTION  : Deletes a query class-list
  INPUTS       : The query list address
  RETURNS      : Nothing useful
  SIDE EFFECTS : Nodes deallocated
                 Busy count decremented for all classes
  NOTES        : None
 ******************************************************/
static VOID DeleteQueryClasses(qlist)
  QUERY_CLASS *qlist;
  {
   QUERY_CLASS *tmp;
   
   while (qlist != NULL)
     {
      while (qlist->chain != NULL)
        {
         tmp = qlist->chain;
         qlist->chain = qlist->chain->chain;
         tmp->cls->busy--;
         rtn_struct(query_class,tmp);
        }
      tmp = qlist;
      qlist = qlist->nxt;
      tmp->cls->busy--;
      rtn_struct(query_class,tmp);
     }
  }
  
/************************************************************
  NAME         : TestForFirstInChain
  DESCRIPTION  : Processes all classes in a restriction chain
                   until success or done
  INPUTS       : 1) The current chain
                 2) The index of the chain restriction
                     (e.g. the 4th query-variable)
  RETURNS      : TRUE if query succeeds, FALSE otherwise
  SIDE EFFECTS : Sets current restriction class
                 Instance variable values set
  NOTES        : None
 ************************************************************/
static int TestForFirstInChain(qchain,indx)
  QUERY_CLASS *qchain;
  int indx;
  {
   QUERY_CLASS *qptr;
   int id;
   
   AbortQuery = TRUE;
   for (qptr = qchain ; qptr != NULL ; qptr = qptr->chain)
     {
      AbortQuery = FALSE;
      if ((id = GetTraversalID()) == -1)
        return(FALSE);
      if (TestForFirstInstanceInClass(id,qptr->cls,qchain,indx) == TRUE)
        {
         ReleaseTraversalID();
         return(TRUE);
        }
      ReleaseTraversalID();
      if ((HaltExecution == TRUE) || (AbortQuery == TRUE))
        return(FALSE);
     }
   return(FALSE);
  }

/*****************************************************************
  NAME         : TestForFirstInstanceInClass
  DESCRIPTION  : Processes all instances in a class and then
                   all subclasses of a class until success or done
  INPUTS       : 1) Visitation traversal id
                 2) The class
                 3) The current class restriction chain
                 4) The index of the current restriction
  RETURNS      : TRUE if query succeeds, FALSE otherwise
  SIDE EFFECTS : Instance variable values set
  NOTES        : None
 *****************************************************************/
static int TestForFirstInstanceInClass(id,cls,qchain,indx)
  int id;
  CLASS_TYPE *cls;
  QUERY_CLASS *qchain;
  int indx;
  {
   CLASS_LINK *cptr;
   INSTANCE_TYPE *ins;
   DATA_OBJECT temp;
   int svmaintain;
   
   if (TestTraversalID(cls->tvids,id))
     return(FALSE);
   SetTraversalID(cls->tvids,id);
   svmaintain = MaintainGarbageInstances;
   MaintainGarbageInstances = TRUE;
   ins = cls->instances;
   while (ins != NULL)
     {
      QueryCore->solns[indx] = ins;
      if (qchain->nxt != NULL)
        {
         if (TestForFirstInChain(qchain->nxt,indx+1) == TRUE)
           break;
         if ((HaltExecution == TRUE) || (AbortQuery == TRUE))
           break;
        }
      else
        {
         EvaluateExpression(QueryCore->query,&temp);
         if (HaltExecution == TRUE)
           break;
         if ((temp.type != SYMBOL) ? TRUE : 
             (temp.value != (VOID *) CLIPSFalseSymbol))
           break;
        }
      ins = ins->nxt_cls;
      while ((ins != NULL) ? (ins->garbage == 1) : FALSE)
        ins = ins->nxt_cls;
     }
   MaintainGarbageInstances = svmaintain;
   if (ins != NULL)
     return(((HaltExecution == TRUE) || (AbortQuery == TRUE)) ? FALSE : TRUE);
   for (cptr = cls->sublink ; cptr != NULL ; cptr = cptr->nxt)
     {
      if (TestForFirstInstanceInClass(id,cptr->cls,qchain,indx) == TRUE)
        return(TRUE);
      if ((HaltExecution == TRUE) || (AbortQuery == TRUE))
        return(FALSE);
     }
   return(FALSE);
  }
  
/************************************************************
  NAME         : TestEntireChain
  DESCRIPTION  : Processes all classes in a restriction chain
                   until done
  INPUTS       : 1) The current chain
                 2) The index of the chain restriction
                     (i.e. the 4th query-variable)
  RETURNS      : Nothing useful
  SIDE EFFECTS : Sets current restriction class
                 Query instance variables set
                 Solution sets stored in global list
  NOTES        : None
 ************************************************************/
static VOID TestEntireChain(qchain,indx)
  QUERY_CLASS *qchain;
  int indx;
  {
   QUERY_CLASS *qptr;
   int id;
   
   AbortQuery = TRUE;
   for (qptr = qchain ; qptr != NULL ; qptr = qptr->chain)
     {
      AbortQuery = FALSE;
      if ((id = GetTraversalID()) == -1)
        return;
      TestEntireClass(id,qptr->cls,qchain,indx);
      ReleaseTraversalID();
      if ((HaltExecution == TRUE) || (AbortQuery == TRUE))
        return;
     }
  }

/*****************************************************************
  NAME         : TestEntireClass
  DESCRIPTION  : Processes all instances in a class and then
                   all subclasses of a class until done
  INPUTS       : 1) Visitation traversal id
                 2) The class
                 3) The current class restriction chain
                 4) The index of the current restriction
  RETURNS      : TRUE if query succeeds, FALSE otherwise
  SIDE EFFECTS : Instance variable values set
                 Solution sets stored in global list
  NOTES        : None
 *****************************************************************/
static VOID TestEntireClass(id,cls,qchain,indx)
  int id;
  CLASS_TYPE *cls;
  QUERY_CLASS *qchain;
  int indx;  
  {
   CLASS_LINK *cptr;
   INSTANCE_TYPE *ins;
   DATA_OBJECT temp;
   int svmaintain;
   
   if (TestTraversalID(cls->tvids,id))
     return;
   SetTraversalID(cls->tvids,id);
   svmaintain = MaintainGarbageInstances;
   MaintainGarbageInstances = TRUE;
   ins = cls->instances;
   while (ins != NULL)
     {
      QueryCore->solns[indx] = ins;
      if (qchain->nxt != NULL)
        {
         TestEntireChain(qchain->nxt,indx+1);
         if ((HaltExecution == TRUE) || (AbortQuery == TRUE))
           break;
        }
      else
        {
         EvaluateExpression(QueryCore->query,&temp);
         if (HaltExecution == TRUE) 
           break;
         if ((temp.type != SYMBOL) ? TRUE : 
             (temp.value != (VOID *) CLIPSFalseSymbol))
           {
            if (QueryCore->action != NULL)
              {
               EvaluateExpression(QueryCore->action,QueryCore->result);
               if (HaltExecution == TRUE)
                 break;
              }
            else
              AddSolution();
           }
        }
      ins = ins->nxt_cls;
      while ((ins != NULL) ? (ins->garbage == 1) : FALSE)
        ins = ins->nxt_cls;
     }
   MaintainGarbageInstances = svmaintain;
   if (ins != NULL)
     return;
   for (cptr = cls->sublink ; cptr != NULL ; cptr = cptr->nxt)
     {
      TestEntireClass(id,cptr->cls,qchain,indx);
      if ((HaltExecution == TRUE) || (AbortQuery == TRUE))
        return;
     }
  }
  
/***************************************************************************
  NAME         : AddSolution
  DESCRIPTION  : Adds the current instance set to a global list of
                   solutions
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Global list and count updated
  NOTES        : Solutions are stored as sequential arrays of INSTANCE_TYPE *
                   nodes where the last soln element points to another array
                   (if any) 
                 Assumes sizeof(INSTANCE_TYPE **) == sizeof(INSTANCE_TYPE *)
 ***************************************************************************/
static VOID AddSolution()
  {
   INSTANCE_TYPE **new_soln;
   register int i;
   
   new_soln = (INSTANCE_TYPE **) 
              gm2((int) (sizeof(INSTANCE_TYPE *) * (QueryCore->soln_size+1)));
   for (i = 0 ; i < QueryCore->soln_size ; i++)
     new_soln[i] = QueryCore->solns[i];
   new_soln[i] = NULL;
   if (QueryCore->soln_set == NULL)
     QueryCore->soln_set = new_soln;
   else
     QueryCore->soln_bottom[QueryCore->soln_size] = (INSTANCE_TYPE *) new_soln;
   QueryCore->soln_bottom = new_soln;
   QueryCore->soln_cnt++;
  }
  
#endif

/***************************************************
  NAME         : 
  DESCRIPTION  : 
  INPUTS       : 
  RETURNS      : 
  SIDE EFFECTS : 
  NOTES        : 
 ***************************************************/


