   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*               CLIPS Version 5.10  07/17/91          */
   /*                                                     */
   /*                                                     */
   /*******************************************************/

/*************************************************************/
/* Purpose: Object System Construct Compiler Code            */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
/* =========================================
   *****************************************
               EXTERNAL DEFINITIONS
   =========================================
   ***************************************** */
#include "setup.h"

#if OBJECT_SYSTEM && CONSTRUCT_COMPILER && (! RUN_TIME)

#include "constant.h"
#include "classfun.h"
#include "defins.h"
#include "conscomp.h"

#define _OBJCMP_SOURCE_
#include "objcmp.h"

/* =========================================
   *****************************************
                   CONSTANTS
   =========================================
   ***************************************** */
   
/* =========================================
   *****************************************
               MACROS AND TYPES
   =========================================
   ***************************************** */
   
/* =========================================
   *****************************************
      INTERNALLY VISIBLE FUNCTION HEADERS
   =========================================
   ***************************************** */
#if ANSI_COMPILER
static int ObjectsToCode(char *,int,FILE *,int,int);
static int DefinstancesToCode(char *,int,FILE *,int,int,int *);
static int ClassHashTableToCode(char *,int,FILE *,int,int,int *);
static int ClassesToCode(char *,int,FILE *,int,int,int *);
static int LinksToCode(char *,int,FILE *,int,int,int *);
static int SlotsToCode(char *,int,FILE *,int,int,int *);
static int TemplateSlotsToCode(char *,int,FILE *,int,int *);
static int OrderedSlotsToCode(char *,int,FILE *,int,int *);
static int HandlersToCode(char *,int,FILE *,int,int,int *);
static int OrderedHandlersToCode(char *,int,FILE *,int,int *);
static VOID InitObjectsCode(FILE *,int,int);
static VOID ReadyObjectsForCode(void);
#else
static int ObjectsToCode();
static int DefinstancesToCode();
static int ClassHashTableToCode();
static int ClassesToCode();
static int LinksToCode();
static int SlotsToCode();
static int TemplateSlotsToCode();
static int OrderedSlotsToCode();
static int OrderedHandlersToCode();
static int HandlersToCode();
static VOID InitObjectsCode();
static VOID ReadyObjectsForCode();
#endif
      
/* =========================================
   *****************************************
          EXTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
   
/***************************************************
  NAME         : SetupObjectsCompiler
  DESCRIPTION  : Initializes the construct compiler
                   item for generic functions
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Code generator item initialized
  NOTES        : None
 ***************************************************/
globle VOID SetupObjectsCompiler()
  {
   AddCodeGeneratorItem("objects",0,ReadyObjectsForCode,
                        InitObjectsCode,ObjectsToCode);
  }
  
   
/*********************************************************
  NAME         : PrintClassReference
  DESCRIPTION  : Writes out a reference to the class array
  INPUTS       : 1) Output file pointer
                 2) Class address
                 3) Construct set image id
                 4) The maximum number of indices allowed
                    in an array
  RETURNS      : Nothing useful
  SIDE EFFECTS : Writes out class array reference to file
  NOTES        : None
 *********************************************************/
globle VOID PrintClassReference(fp,cls,imageID,maxIndices)
  FILE *fp;
  CLASS_TYPE *cls;
  int imageID,maxIndices;
  {
   if (cls == NULL)
     fprintf(fp,"NULL");
   else
     fprintf(fp,"&cls%d_%d[%d]",imageID,
                                (int) ((cls->bsaveIndex / maxIndices) + 1),
                                (int) (cls->bsaveIndex % maxIndices)); 
  }
  
/* =========================================
   *****************************************
          INTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
   
/*************************************************************
  NAME         : ObjectsToCode
  DESCRIPTION  : Writes out static array code for
                   classes, message-handlers, definstances
                   and associated structures
  INPUTS       : 1) The base name of the construct set
                 2) The base id for this construct
                 3) The file pointer for the header file
                 4) The base id for the construct set
                 5) The max number of indices allowed
                    in an array
  RETURNS      : -1 if no classes or definstances, 0 on errors,
                  1 if object system structures written
  SIDE EFFECTS : Code written to files
  NOTES        : None
 *************************************************************/
static int ObjectsToCode(fileName,fileID,headerFP,imageID,maxIndices)
  char *fileName;
  int fileID;
  FILE *headerFP;
  int imageID;
  int maxIndices;
  {
   int version = 1,rtn;
   
   fprintf(headerFP,"#include \"classcom.h\"\n");
   rtn = DefinstancesToCode(fileName,fileID,headerFP,imageID,maxIndices,&version);
   if (rtn == CLIPS_FALSE)
     return(0);
   rtn = ClassHashTableToCode(fileName,fileID,headerFP,imageID,maxIndices,&version);
   if (rtn == CLIPS_FALSE)
     return(0);
   rtn = ClassesToCode(fileName,fileID,headerFP,imageID,maxIndices,&version);
   if (rtn == CLIPS_FALSE)
     return(0);
   rtn = LinksToCode(fileName,fileID,headerFP,imageID,maxIndices,&version);
   if (rtn == CLIPS_FALSE)
     return(0);
   rtn = SlotsToCode(fileName,fileID,headerFP,imageID,maxIndices,&version);
   if (rtn == CLIPS_FALSE)
     return(0);
   rtn = TemplateSlotsToCode(fileName,fileID,headerFP,imageID,&version);
   if (rtn == CLIPS_FALSE)
     return(0);
   rtn = OrderedSlotsToCode(fileName,fileID,headerFP,imageID,&version);
   if (rtn == CLIPS_FALSE)
     return(0);
   rtn = HandlersToCode(fileName,fileID,headerFP,imageID,maxIndices,&version);
   if (rtn == CLIPS_FALSE)
     return(0);
   rtn = OrderedHandlersToCode(fileName,fileID,headerFP,imageID,&version);
   if (rtn == CLIPS_FALSE)
     return(0);
   return(1);
  }
  
/************************************************************
  NAME         : DefinstancesToCode
  DESCRIPTION  : Writes out definstances
  INPUTS       : 1) Header file pointer
                 2) Output file pointer
                 3) The construct set image id
                 4) The max # of allowed indices
                 5) Caller's version number buffer
  RETURNS      : CLIPS_TRUE if all OK, CLIPS_FALSE otherwise
  SIDE EFFECTS : Definstances Written
  NOTES        : None
 ***********************************************************/
static int DefinstancesToCode(fileName,fileID,headerFP,imageID,maxIndices,version)
  char *fileName;
  FILE *headerFP;
  int fileID,imageID,maxIndices,*version;
  {
   DEFINSTANCES *dptr;
   FILE *fp = NULL;
   int i = 1,partition = 1,newHeader = CLIPS_TRUE;
   
   for (dptr = DefinstancesList ; dptr != NULL ; dptr = dptr->nxt)
     {
      if (newHeader)
        {
         if ((fp = NewCFile(fileName,fileID,*version)) == NULL)
           return(CLIPS_FALSE);
         fprintf(fp,"DEFINSTANCES dfi%d_%d[] = {\n{",imageID,partition);
         fprintf(headerFP,"extern DEFINSTANCES dfi%d_%d[];\n",
                 imageID,partition);
         (*version)++;
         partition++;
         newHeader = CLIPS_FALSE;
        }
      else
        fprintf(fp,",\n{");
      fprintf(fp,"0,");
      PrintSymbolReference(fp,dptr->name);
      fprintf(fp,",");
      ExpressionToCode(fp,dptr->mkinstance);
      fprintf(fp,",NULL,");
      if (dptr->prv == NULL)
        fprintf(fp,"NULL,");
      else
        fprintf(fp,"&dfi%d_%d[%d],",imageID,
                                  (int) ((dptr->prv->bsaveIndex / maxIndices) + 1),
                                  (int) (dptr->prv->bsaveIndex % maxIndices));
      if (dptr->nxt == NULL)
        fprintf(fp,"NULL,0L}");
      else
        fprintf(fp,"&dfi%d_%d[%d],0L}",imageID,
                                  (int) ((dptr->nxt->bsaveIndex / maxIndices) + 1),
                                  (int) (dptr->nxt->bsaveIndex % maxIndices));
      i++;
      if (i > maxIndices)
        {
         fprintf(fp,"\n};\n");
         fclose(fp);
         fp = NULL;
         i = 1;
         newHeader = CLIPS_TRUE;
        }
     }
   if (fp != NULL)
     {
      fprintf(fp,"\n};\n");
      fclose(fp);
     }
   return(CLIPS_TRUE);
  }
  
/************************************************************
  NAME         : ClassHashTableToCode
  DESCRIPTION  : Writes out class hash table
  INPUTS       : 1) Header file pointer
                 2) Output file pointer
                 3) The construct set image id
                 4) The max # of allowed indices
                 5) Caller's version number buffer
  RETURNS      : CLIPS_TRUE if all OK, CLIPS_FALSE otherwise
  SIDE EFFECTS : Class Hash Table Written
  NOTES        : None
 ***********************************************************/
static int ClassHashTableToCode(fileName,fileID,headerFP,imageID,maxIndices,version)
  char *fileName;
  FILE *headerFP;
  int fileID,imageID,maxIndices,*version;
  {
   register int i;
   FILE *fp;
   
   if ((fp = NewCFile(fileName,fileID,*version)) == NULL)
     return(CLIPS_FALSE);
   (*version)++;
   fprintf(headerFP,"extern CLASS_TYPE *cht%d[];\n",imageID);
   fprintf(fp,"CLASS_TYPE *cht%d[] = {\n",imageID);
   for (i = 0 ; i < CLASS_TABLE_HASH_SIZE ; i++)
     {
      PrintClassReference(fp,ClassTable[i],imageID,maxIndices);
      if (i < (CLASS_TABLE_HASH_SIZE-1))
        fprintf(fp,",\n");
      else
        fprintf(fp,"\n");
     }
   fprintf(fp,"};\n");
   fclose(fp);
   return(CLIPS_TRUE);
  }
  
/************************************************************
  NAME         : ClassesToCode
  DESCRIPTION  : Writes out classes
  INPUTS       : 1) Header file pointer
                 2) Output file pointer
                 3) The construct set image id
                 4) The max # of allowed indices
                 5) Caller's version number buffer
  RETURNS      : CLIPS_TRUE if all OK, CLIPS_FALSE otherwise
  SIDE EFFECTS : Classes Written
  NOTES        : None
 ***********************************************************/
static int ClassesToCode(fileName,fileID,headerFP,imageID,maxIndices,version)
  char *fileName;
  FILE *headerFP;
  int fileID,imageID,maxIndices,*version;
  {
   CLASS_TYPE *cls;
   FILE *fp = NULL;
   int i = 1,partition = 1,newHeader = CLIPS_TRUE,
       sloti = 1,slotpart = 1,
       hndli = 1,hndlpart = 1,
       clscnt,tvds_saved = CLIPS_FALSE,clsi;
   unsigned long linki = 0L,tmpli = 0L,ohndli = 0L;
   CLASS_LINK *sptr;
   
   clscnt = 0;
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     clscnt++;
   for (cls = ClassList , clsi = 0 ; cls != NULL ; cls = cls->nxt_lst , clsi++)
     {
      if (newHeader)
        {
         if ((fp = NewCFile(fileName,fileID,*version)) == NULL)
           return(CLIPS_FALSE);
         if (tvds_saved == CLIPS_FALSE)
           {
            tvds_saved = CLIPS_TRUE;
            fprintf(headerFP,"extern char tvds%d[];\n",imageID);
            fprintf(fp,"char tvds%d[%d];\n\n",imageID,(int) (clscnt * TRAVERSAL_BYTES));
           }
         fprintf(fp,"CLASS_TYPE cls%d_%d[] = {\n{",imageID,partition);
         fprintf(headerFP,"extern CLASS_TYPE cls%d_%d[];\n",
                 imageID,partition);
         (*version)++;
         partition++;
         newHeader = CLIPS_FALSE;
        }
      else
        fprintf(fp,",\n{");
      fprintf(fp,"1,%u,%u,%u,%u,0,%u,0,",cls->system,cls->abstract,cls->primitive,
              cls->primitive_type,cls->hashval);
      PrintSymbolReference(fp,cls->name);
      fprintf(fp,",");
      if (cls->superlink != NULL)
        {
         fprintf(fp,"&clk%d_%d[%d],",imageID,
                                     (int) ((linki / maxIndices) + 1),
                                     (int) (linki % maxIndices)); 
         for (sptr = cls->superlink ; sptr != NULL ; sptr = sptr->nxt)
           linki++;
        }
      else
        fprintf(fp,"NULL,");
      if (cls->sublink != NULL)
        {
         fprintf(fp,"&clk%d_%d[%d],",imageID,
                                     (int) ((linki / maxIndices) + 1),
                                     (int) (linki % maxIndices)); 
         for (sptr = cls->sublink ; sptr != NULL ; sptr = sptr->nxt)
           linki++;
        }
      else
        fprintf(fp,"NULL,");
      if (cls->precedence != NULL)
        {
         fprintf(fp,"&clk%d_%d[%d],\n  ",imageID,
                                     (int) ((linki / maxIndices) + 1),
                                     (int) (linki % maxIndices)); 
         for (sptr = cls->precedence ; sptr != NULL ; sptr = sptr->nxt)
           linki++;
        }
      else
        fprintf(fp,"NULL,\n  ");
      if (cls->slots != NULL)
        {
         fprintf(fp,"&slt%d_%d[%d],",imageID,slotpart,sloti-1);
         sloti += cls->slot_cnt;
         if (sloti > maxIndices)
           {
            sloti = 1;
            slotpart++;
           }
        }
      else
        fprintf(fp,"NULL,");
      if (cls->itemplate != NULL)
        {
         fprintf(fp,"&tsl%d[%lu],&msl%d[%lu],",imageID,tmpli,imageID,tmpli);
         tmpli += (unsigned long) cls->islot_cnt;
        }
      else
        fprintf(fp,"NULL,NULL,");
      fprintf(fp,"%u,%u,NULL,",cls->slot_cnt,cls->islot_cnt);
      if (cls->handlers != NULL)
        {
         fprintf(fp,"&hnd%d_%d[%d],&mhd%d[%lu],",imageID,hndlpart,hndli-1,
                                                 imageID,ohndli);
         hndli += cls->handler_cnt;
         if (hndli > maxIndices)
           {
            hndli = 1;
            hndlpart++;
           }
         ohndli += cls->handler_cnt;
        }
      else
        fprintf(fp,"NULL,NULL,");
      fprintf(fp,"%u,",cls->handler_cnt);
      PrintClassReference(fp,cls->prv_hsh,imageID,maxIndices);
      fprintf(fp,",");
      PrintClassReference(fp,cls->nxt_hsh,imageID,maxIndices);
      fprintf(fp,",");
      PrintClassReference(fp,cls->prv_lst,imageID,maxIndices);
      fprintf(fp,",");
      PrintClassReference(fp,cls->nxt_lst,imageID,maxIndices);
      fprintf(fp,",0L,NULL,&tvds%d[%d]}",imageID,(int) (clsi * TRAVERSAL_BYTES));
      i++;
      if (i > maxIndices)
        {
         fprintf(fp,"\n};\n");
         fclose(fp);
         fp = NULL;
         i = 1;
         newHeader = CLIPS_TRUE;
        }
     }
   if (fp != NULL)
     {
      fprintf(fp,"\n};\n");
      fclose(fp);
     }
   return(CLIPS_TRUE);
  }
  
/************************************************************
  NAME         : LinksToCode
  DESCRIPTION  : Writes out class links
  INPUTS       : 1) Header file pointer
                 2) Output file pointer
                 3) The construct set image id
                 4) The max # of allowed indices
                 5) Caller's version number buffer
  RETURNS      : CLIPS_TRUE if all OK, CLIPS_FALSE otherwise
  SIDE EFFECTS : Class links written
  NOTES        : None
 ***********************************************************/
static int LinksToCode(fileName,fileID,headerFP,imageID,maxIndices,version)
  char *fileName;
  FILE *headerFP;
  int fileID,imageID,maxIndices,*version;
  {
   CLASS_TYPE *cls;
   FILE *fp = NULL;
   int i = 1,partition = 1,newHeader = CLIPS_TRUE;
   CLASS_LINK *sptr,*sv1,*sv2;
   
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      sv1 = sv2 = NULL;
      if (cls->superlink != NULL)
        {
         sv1 = cls->superlink;
         while (sv1->nxt != NULL)
           sv1 = sv1->nxt;
        }
      if (cls->sublink != NULL)
        {
         sv2 = cls->sublink;
         while (sv2->nxt != NULL)
           sv2 = sv2->nxt;
         sv2->nxt = cls->precedence;
        }
      if (sv1 != NULL)
        {
         sptr = cls->superlink;
         if (sv2 != NULL)
           sv1->nxt = cls->sublink;
         else
           sv1->nxt = cls->precedence;
        }
      else if (sv2 != NULL)
        sptr = cls->sublink;
      else
        sptr = cls->precedence;
      for ( ; sptr != NULL ; sptr = sptr->nxt)
        {
         if (newHeader)
           {
            if ((fp = NewCFile(fileName,fileID,*version)) == NULL)
              return(CLIPS_FALSE);
            fprintf(fp,"CLASS_LINK clk%d_%d[] = {\n{",imageID,partition);
            fprintf(headerFP,"extern CLASS_LINK clk%d_%d[];\n",
                    imageID,partition);
            (*version)++;
            partition++;
            newHeader = CLIPS_FALSE;
           }
         else
           fprintf(fp,",\n{");
         PrintClassReference(fp,sptr->cls,imageID,maxIndices);
         i++;
         if ((sptr->nxt == NULL) || (sptr == sv1) || (sptr == sv2))
           fprintf(fp,",NULL}");
         else
           {
            if (i > maxIndices)
              fprintf(fp,",&clk%d_%d[0]}",imageID,partition);
            else
              fprintf(fp,",&clk%d_%d[%d]}",imageID,partition-1,i-1);
           }
         if (i > maxIndices)
           {
            fprintf(fp,"\n};\n");
            fclose(fp);
            fp = NULL;
            i = 1;
            newHeader = CLIPS_TRUE;
           }
        }
      if (sv1 != NULL)
        sv1->nxt = NULL;
      if (sv2 != NULL)
        sv2->nxt = NULL;
     }
   if (fp != NULL)
     {
      fprintf(fp,"\n};\n");
      fclose(fp);
     }
   return(CLIPS_TRUE);
  }
  
/****************************************************************
  NAME         : SlotsToCode
  DESCRIPTION  : Writes out class slots
  INPUTS       : 1) Header file pointer
                 2) Output file pointer
                 3) The construct set image id
                 4) The max # of allowed indices
                 5) Caller's version number buffer
  RETURNS      : CLIPS_TRUE if all OK, CLIPS_FALSE otherwise
  SIDE EFFECTS : Class slots Written
                 cls and name fields of each slot overwritten
                   with partition and index for later reference.
                   This info needs to be restored before the
                   compilation process is complete. (Done in
                   TemplateSlotsToCode().)
  NOTES        : None
 ***************************************************************/
static int SlotsToCode(fileName,fileID,headerFP,imageID,maxIndices,version)
  char *fileName;
  FILE *headerFP;
  int fileID,imageID,maxIndices,*version;
  {
   CLASS_TYPE *cls;
   FILE *fp = NULL;
   int i = 1,partition = 1,newHeader = CLIPS_TRUE;
   unsigned si;
   SLOT_DESC *sp;
   
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      for (si = 0 ; si < cls->slot_cnt ; si++)
        {
         sp = &cls->slots[si];
         if (newHeader)
           {
            if ((fp = NewCFile(fileName,fileID,*version)) == NULL)
              return(CLIPS_FALSE);
            fprintf(fp,"SLOT_DESC slt%d_%d[] = {\n{",imageID,partition);
            fprintf(headerFP,"extern SLOT_DESC slt%d_%d[];\n",
                    imageID,partition);
            (*version)++;
            partition++;
            newHeader = CLIPS_FALSE;
           }
         else
           fprintf(fp,",\n{");
         fprintf(fp,"0,0,0,0,%u,%u,%u,%u,%u,%u,0,%u,",
                 sp->shared,sp->multiple,sp->composite,sp->noinherit,
                 sp->nowrite,sp->initonly,sp->dynamic);
         PrintClassReference(fp,sp->cls,imageID,maxIndices);
         fprintf(fp,",\n   ");
         PrintSymbolReference(fp,sp->name);
         fprintf(fp,",");
         PrintSymbolReference(fp,sp->gethandler);
         fprintf(fp,",");
         PrintSymbolReference(fp,sp->puthandler);
         fprintf(fp,",");
         ExpressionToCode(fp,sp->defexp);
         fprintf(fp,",NULL,0}");
         i++;
         sp->cls = (CLASS_TYPE *) (partition-1);
         sp->name = (SYMBOL_HN *) (i-2);
        }
      if (i > maxIndices)
        {
         fprintf(fp,"\n};\n");
         fclose(fp);
         fp = NULL;
         i = 1;
         newHeader = CLIPS_TRUE;
        }
     }
   if (fp != NULL)
     {
      fprintf(fp,"\n};\n");
      fclose(fp);
     }
   return(CLIPS_TRUE);
  }
  
/************************************************************
  NAME         : TemplateSlotsToCode
  DESCRIPTION  : Writes out class instance templates
  INPUTS       : 1) Header file pointer
                 2) Output file pointer
                 3) The construct set image id
                 4) Caller's version number buffer
  RETURNS      : CLIPS_TRUE if all OK, CLIPS_FALSE otherwise
  SIDE EFFECTS : Class instance templates Written
  NOTES        : None
 ***********************************************************/
static int TemplateSlotsToCode(fileName,fileID,headerFP,imageID,version)
  char *fileName;
  FILE *headerFP;
  int fileID,imageID,*version;
  {
   CLASS_TYPE *cls;
   FILE *fp = NULL;
   int newHeader = CLIPS_TRUE;
   unsigned si;
   SLOT_DESC *sp;
   
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      for (si = 0 ; si < cls->islot_cnt ; si++)
        {
         sp = cls->itemplate[si];
         if (newHeader)
           {
            if ((fp = NewCFile(fileName,fileID,*version)) == NULL)
              return(CLIPS_FALSE);
            fprintf(fp,"SLOT_DESC *tsl%d[] = {\n",imageID);
            fprintf(headerFP,"extern SLOT_DESC *tsl%d[];\n",imageID);
            (*version)++;
            newHeader = CLIPS_FALSE;
           }
         else
           fprintf(fp,",\n");
         fprintf(fp,"&slt%d_%d[%d]",imageID,(int) sp->cls,(int) sp->name);
        }
     }
   if (fp != NULL)
     {
      fprintf(fp,"\n};\n");
      fclose(fp);
     }
     
   /* ======================================
      Restore class and name fields of slots
      ====================================== */
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      for (si = 0 ; si < cls->slot_cnt ; si++)
        {
         sp = &cls->slots[si];
         sp->cls = cls;
         sp->name = sp->gethandler->relatedSymbol;
        }
     }
   return(CLIPS_TRUE);
  }
  
/************************************************************
  NAME         : OrderedSlotsToCode
  DESCRIPTION  : Writes out class instance ordered slot maps
  INPUTS       : 1) Header file pointer
                 2) Output file pointer
                 3) The construct set image id
                 4) Caller's version number buffer
  RETURNS      : CLIPS_TRUE if all OK, CLIPS_FALSE otherwise
  SIDE EFFECTS : Class instance templates written
  NOTES        : None
 ***********************************************************/
static int OrderedSlotsToCode(fileName,fileID,headerFP,imageID,version)
  char *fileName;
  FILE *headerFP;
  int fileID,imageID,*version;
  {
   CLASS_TYPE *cls;
   FILE *fp = NULL;
   int newHeader = CLIPS_TRUE;
   unsigned si;
   
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      for (si = 0 ; si < cls->islot_cnt ; si++)
        {
         if (newHeader)
           {
            if ((fp = NewCFile(fileName,fileID,*version)) == NULL)
              return(CLIPS_FALSE);
            fprintf(fp,"unsigned msl%d[] = {\n",imageID);
            fprintf(headerFP,"extern unsigned msl%d[];\n",imageID);
            (*version)++;
            newHeader = CLIPS_FALSE;
           }
         else
           fprintf(fp,",\n");
         fprintf(fp,"%u",cls->sordered[si]);
        }
     }
   if (fp != NULL)
     {
      fprintf(fp,"\n};\n");
      fclose(fp);
     }
   return(CLIPS_TRUE);
  }
  
/****************************************************************
  NAME         : HandlersToCode
  DESCRIPTION  : Writes out class message-handlers
  INPUTS       : 1) Header file pointer
                 2) Output file pointer
                 3) The construct set image id
                 4) The max # of allowed indices
                 5) Caller's version number buffer
  RETURNS      : CLIPS_TRUE if all OK, CLIPS_FALSE otherwise
  SIDE EFFECTS : Class message-handlers Written
  NOTES        : None
 ***************************************************************/
static int HandlersToCode(fileName,fileID,headerFP,imageID,maxIndices,version)
  char *fileName;
  FILE *headerFP;
  int fileID,imageID,maxIndices,*version;
  {
   CLASS_TYPE *cls;
   FILE *fp = NULL;
   int i = 1,partition = 1,newHeader = CLIPS_TRUE;
   unsigned mi;
   HANDLER *hnd;
   
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      for (mi = 0 ; mi < cls->handler_cnt ; mi++)
        {
         hnd = &cls->handlers[mi];
         if (newHeader)
           {
            if ((fp = NewCFile(fileName,fileID,*version)) == NULL)
              return(CLIPS_FALSE);
            fprintf(fp,"HANDLER hnd%d_%d[] = {\n{",imageID,partition);
            fprintf(headerFP,"extern HANDLER hnd%d_%d[];\n",
                    imageID,partition);
            (*version)++;
            partition++;
            newHeader = CLIPS_FALSE;
           }
         else
           fprintf(fp,",\n{");
         fprintf(fp,"%u,%u,0,0,",hnd->system,hnd->type);
         PrintSymbolReference(fp,hnd->name);
         fprintf(fp,",");
         PrintClassReference(fp,hnd->cls,imageID,maxIndices);
         fprintf(fp,",%d,%d,",hnd->minParams,hnd->maxParams);
         ExpressionToCode(fp,hnd->actions);
         fprintf(fp,",NULL}");
        }
      i += cls->handler_cnt;
      if (i > maxIndices)
        {
         fprintf(fp,"\n};\n");
         fclose(fp);
         fp = NULL;
         i = 1;
         newHeader = CLIPS_TRUE;
        }
     }
   if (fp != NULL)
     {
      fprintf(fp,"\n};\n");
      fclose(fp);
     }
   return(CLIPS_TRUE);
  }
  
/************************************************************
  NAME         : OrderedHandlersToCode
  DESCRIPTION  : Writes out class message-handler order maps
  INPUTS       : 1) Output file name
                 2) Output file id
                 3) Header file pointer
                 4) The construct set image id
                 5) Caller's version number buffer
  RETURNS      : CLIPS_TRUE if all OK, CLIPS_FALSE otherwise
  SIDE EFFECTS : Class handler maps written
  NOTES        : None
 ***********************************************************/
static int OrderedHandlersToCode(fileName,fileID,headerFP,imageID,version)
  char *fileName;
  FILE *headerFP;
  int fileID,imageID,*version;
  {
   CLASS_TYPE *cls;
   FILE *fp = NULL;
   int newHeader = CLIPS_TRUE;
   unsigned hi;
   
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      for (hi = 0 ; hi < cls->handler_cnt ; hi++)
        {
         if (newHeader)
           {
            if ((fp = NewCFile(fileName,fileID,*version)) == NULL)
              return(CLIPS_FALSE);
            fprintf(fp,"unsigned mhd%d[] = {\n",imageID);
            fprintf(headerFP,"extern unsigned mhd%d[];\n",imageID);
            (*version)++;
            newHeader = CLIPS_FALSE;
           }
         else
           fprintf(fp,",\n");
         fprintf(fp,"%u",cls->hordered[hi]);
        }
     }
   if (fp != NULL)
     {
      fprintf(fp,"\n};\n");
      fclose(fp);
     }
   return(CLIPS_TRUE);
  }
  
/*******************************************************
  NAME         : InitObjectsCode
  DESCRIPTION  : Writes out initialization code for
                   generic functions
  INPUTS       : 1) The initialization code file pointer
                 2) The construct set image id
                 3) The max number of indices allowed in
                    an array for this construct set
  RETURNS      : Nothing useful
  SIDE EFFECTS : Writes out initialization code
  NOTES        : None
 *******************************************************/
static VOID InitObjectsCode(initFP,imageID,maxIndices)
  FILE *initFP;
  int imageID;
  int maxIndices;
  {
#if MAC_MPW
#pragma unused(maxIndices)
#endif
   fprintf(initFP,"   ObjectsRunTimeInitialize(");
   if (ClassList != NULL)
     fprintf(initFP,"cls%d_1,",imageID);
   else
     fprintf(initFP,"NULL,");
   if (DefinstancesList != NULL)
     fprintf(initFP,"dfi%d_1,",imageID);
   else
     fprintf(initFP,"NULL,");
   fprintf(initFP,"cht%d);\n",imageID);
  }

/******************************************************
  NAME         : ReadyObjectsForCode
  DESCRIPTION  : Sets index of classes and definstances
                   for use in compiled expressions
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : BsaveIndices set
  NOTES        : None
 ******************************************************/
static VOID ReadyObjectsForCode()
  {
   CLASS_TYPE *cls;
   DEFINSTANCES *dfi;
   register long i;
   
   for (cls = ClassList , i = 0L ; cls != NULL ; 
        cls = cls->nxt_lst , i++)
     cls->bsaveIndex = i;
   for (dfi = DefinstancesList , i = 0L ; dfi != NULL ; 
        dfi = dfi->nxt , i++)
     dfi->bsaveIndex = i;
  }

#endif

/***************************************************
  NAME         : 
  DESCRIPTION  : 
  INPUTS       : 
  RETURNS      : 
  SIDE EFFECTS : 
  NOTES        : 
 ***************************************************/
