#include <stdio.h>
#include "crsv.h"

/*
 *****************************************************************
 *  This file contains the functions used for final analysis of
 *  the knowledge base. 
 *****************************************************************
 */

/* ===========  Functions defined Externally but used here  =========== */

/* -----------------------------
 *  From the file: CRSVMEM.C
 * -----------------------------
 */

extern COUNTER   *AllocCounter();
extern IL_PTR     AllocIntegerList();
extern void       GenFree();

/* -----------------------------
 *  From the file: CRSVPRN.C
 * -----------------------------
 */

extern void    SendMessage();

/* -----------------------------
 *  From the file: CRSVMISC.C
 * -----------------------------
 */

extern int     IsWordInList();

/* ----------------------------
 *  From the file: CRSVTMP.C
 * ----------------------------
 */

extern int     CountTemplales();
extern DT_PTR  FindDeftemplate();

/* -----------------------------
 *  From the file: CRSVTR.C
 * -----------------------------
 */

extern void    CheckDynamicRules();
extern void    CheckDynamicRelations();
extern void    PrintDetailRelations();

/* ===========  Functions defined here for Global use  ================ */

void   analysis();


/* ===========  Functions defined here for internal use  ============== */

void     CheckRHSLoops();
void     FindLoopInformation();
void     PrintLoopWarning();
void     CountLoops();
int      CheckMatchAssertions();
void     FindSalienceInformation();
void     PrintSalienceWarnings();
void     CountSaliences();
void     CountSalienceValues();
DEF_EXT *CheckDefexternalAssertions();



/* ===========  Variables defined here for Global use  ================ */


/* ===========  Variables defined Externally but used here  =========== */

/* -----------------------------
 *  From the file: CRSV.C
 * -----------------------------
 */

extern DEF_EXT     *DefexternalHead;
extern DT_PTR       DeftemplateHead;
extern int          NUM_ERRORS;
extern int          NUM_WARNINGS;


/* ===========  Variables defined here for internal use  ============== */

#define WARNING_MIN_RULES    10

#define MAX_LOOPS            50
#define WARNING_LOOPS        10

#define MAX_SALIENCE         50
#define WARNING_SALIENCE     25
#define MAX_SALIENCE_LEVELS   7

IL_PTR  salience_head;

/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  analysis()
 *
 *  PURPOSE:  This function is the primary entry point for a
 *            series of functions which attempt to analyze the
 *            collected info on a CLIPS knowledge base.
 *
 *  INPUTS :  Eight arguments,
 *             A pointer to the deffacts list (FACT_BLOCK *)
 *             A pointer to the rules list (RULE *)
 *             A pointer to the relations list (RELATION *)
 *             A pointer to the list of external functions
 *               (WORD_LIST *)
 *             A pointer to the list of rules found in the
 *             the dribble file (ACT_RULE *)
 *             A pointer to the list of relations found in the
 *             the dribble file (ACT_RELATION *)
 *             A pointer to the list of defined relations
 *               (DEF_REL *)
 *             A pointer to the list of defined external functions
 *               (DEF_EXT *)
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  This function should merely be used as the 
 *            starting point. Actual analysis is done in
 *            other functions.
 * -------------------------------------------------------------
 */
 
/*ARGSUSED*/ /* Make lint happy. */
void analysis(MyFactBlockHead, MyRuleHead, MyRelationHead, MyexternalFunctionHead, MyActivatedRuleHead, 
              MyActivatedRelationHead, MyDefrelationHead, MyDefexternalHead)
FACT_BLOCK    *MyFactBlockHead;
RULE          *MyRuleHead;
RELATION      *MyRelationHead;
WORD_LIST     *MyexternalFunctionHead;
ACT_RULE      *MyActivatedRuleHead;
ACT_RELATION  *MyActivatedRelationHead;
DEF_REL       *MyDefrelationHead;
DEF_EXT       *MyDefexternalHead;
{
   int num_rules, num_rels;

   SendMessage
      ("\n\n===================== STATIC ANALYSIS SUMMARY ====================\n",NO);

   num_rules = CheckSalience(MyRuleHead);

   CheckRHSLoops(MyRuleHead);

   num_rels  = CheckMatchAssertions(MyRelationHead);


   IF(ANALYZE_TRACE IS_ON) THEN
      SendMessage
         ("\n\n===================== DYNAMIC ANALYSIS SUMMARY ===================\n",NO);

      CheckDynamicRules(MyRuleHead, MyActivatedRuleHead);
      CheckDynamicRelations(MyRelationHead, MyDefrelationHead, MyActivatedRelationHead);
      IF(VERBOSE IS_ON) THEN
        SendMessage("\n======================================================\n",NO);
        PrintDetailRelations(MyActivatedRelationHead);
      END_IF
   END_IF

   SendMessage
      ("\n\n==================================================================",NO);

   sprintf(MsgBuf, "\n\nTotal number of rules ......... %d",num_rules);
   SendMessage(MsgBuf,NO);

   sprintf(MsgBuf," \nTotal number of templates ..... %d",
           CountTemplales(DeftemplateHead));
   SendMessage(MsgBuf,NO);

   sprintf(MsgBuf, "\nTotal number of relations ..... %d",num_rels);
   SendMessage(MsgBuf,NO);

   IF (NUM_ERRORS > 0) THEN
     sprintf(MsgBuf, "\n\nTotal number of Errors ........ %d",NUM_ERRORS);
     SendMessage(MsgBuf,NO);
   END_IF

   IF ((CHECK_STYLE IS_ON) AND (NUM_WARNINGS > 0)) THEN
     sprintf(MsgBuf, "\nTotal number of Warnings ...... %d",NUM_WARNINGS);
     SendMessage(MsgBuf,NO);
   END_IF
}

/* ======================================================================= */
/*                        CHECK SALIENCE FUNCTIONS                         */
/* ======================================================================= */

/*
 * -------------------------------------------------------------
 *  NAME   :  CheckSalience()
 *
 *  PURPOSE:  This function examines all the rules in the
 *            KB for overuse of salience and prints a 
 *            warning if one is necessary.
 *
 *  INPUTS :  A pointer to the rules list (RULE *)
 *
 *  RETURNS:  An integer, the number of rules in the KB.
 *
 *  NOTES  :  
 * -------------------------------------------------------------
 */
 
static int CheckSalience(node)
RULE *node;
{ 
   struct counter     *count_ptr;
   int rtn;

   count_ptr = AllocCounter();
   
   salience_head = NULL;

   FindSalienceInformation(node, count_ptr);
   PrintSalienceWarnings(count_ptr);

   rtn = count_ptr->prime_count;
   GenFree((char *)count_ptr, sizeof(struct counter));
   return(rtn);
}

/* ======================================================================= */

/*
 * -------------------------------------------------------------
 *  NAME   :  FindSalienceInformation()
 *
 *  PURPOSE:  This is a recursive function which processes
 *            the binary tree of rules.
 *
 *  INPUTS :  A pointer to the rules list (RULE *), and
 *            a pointer to a structure for count info.
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  Actual processing of salience info is done in 
 *            CountSaliences function.
 * -------------------------------------------------------------
 */
 
static void FindSalienceInformation(node, count)
RULE  *node;
struct counter *count;
{
   IF(node NEQ NULL) THEN
     FindSalienceInformation(node->lft_rule, count);
     CountSaliences(node, count);
     FindSalienceInformation(node->rht_rule, count);
   END_IF
}

/* ======================================================================= */

/*
 * -------------------------------------------------------------
 *  NAME   :  CountSaliences()
 *
 *  PURPOSE:  This function counts the number of times
 *            salience is used.
 *
 *  INPUTS :  A pointer to the current rule (RULE *), and
 *            a pointer to a structure for count info.
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  
 * -------------------------------------------------------------
 */
 
static void CountSaliences(temp, count)
RULE *temp;
struct counter *count;
{
   count->prime_count++;

   IF(temp->salience_set EQ YES) THEN
      count->count_A++;
      CountSalienceValues(temp->salience_val, count);
   END_IF
}

/* ======================================================================= */

/*
 * -------------------------------------------------------------
 *  NAME   :  CountSalienceValues()
 *
 *  PURPOSE:  This function counts the number of times
 *            a specific salience value is used.
 *
 *  INPUTS :  The current salience value (int), and
 *            a pointer to a structure for count info.
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  Stores salience values in a number list.
 * -------------------------------------------------------------
 */
 
static void CountSalienceValues(number, count)
int number;
struct counter *count;
{
   IL_PTR temp;

   temp  = salience_head;

   while(temp NEQ NULL) DO
      IF(temp->number EQ (long)number) THEN
        temp->times_used++;
        return;
      END_IF

      temp = temp->next_number;
   END_WHILE

   temp              = AllocIntegerList();
   temp->number      = (long)number;
   temp->next_number = salience_head;
   temp->times_used  = 1;
   salience_head     = temp;
   count->count_B++;
}

/* ======================================================================= */

/*
 * -------------------------------------------------------------
 *  NAME   :  PrintSalienceWarnings()
 *
 *  PURPOSE:  This function is called after all salience info
 *            has been gathered and decides whether or not
 *            to print warnings.
 *
 *  INPUTS :  A pointer to the structure with count info (COUNTER *).
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  Uses global #define flags to determine when
 *            salience has been overused.
 * -------------------------------------------------------------
 */
 
static void PrintSalienceWarnings(count)
struct counter *count;
{
   float temp1;

   IF((CHECK_STYLE IS_ON) AND (count->prime_count NEQ 0)) THEN
      temp1 = (float) count->count_A * 100.0 / (float) count->prime_count;

     IF(((temp1 > WARNING_SALIENCE) AND 
          (count->prime_count > WARNING_MIN_RULES)) OR
         (temp1 > MAX_SALIENCE)) THEN
         sprintf(MsgBuf,
            "\nWarning: A high percentage of rules (%4.2f%%) have explicit salience values",
            temp1);
         SendMessage(MsgBuf,NO);
      END_IF
      
      IF(count->count_B > MAX_SALIENCE_LEVELS) THEN
         sprintf(MsgBuf, 
         "\nWarning: A large number of different salience values used (%d). ",
          count->count_B);
         SendMessage(MsgBuf,NO);
      END_IF
   END_IF
}

/* ======================================================================= */
/*                        CHECK IF/WHILE USAGE                             */
/* ======================================================================= */

/*
 * -------------------------------------------------------------
 *  NAME   :  CheckRHSLoops()
 *
 *  PURPOSE:  This function examines all the rules in the
 *            KB for overuse of ifs or whiles and prints a 
 *            warning if one is necessary.
 *
 *  INPUTS :  A pointer to the rules list (RULE *)
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  
 * -------------------------------------------------------------
 */
 
static void CheckRHSLoops(node)
RULE *node;
{ 
   struct counter *count_ptr;

   count_ptr = AllocCounter();

   FindLoopInformation(node, count_ptr);
   PrintLoopWarning(count_ptr);

   GenFree((char *)count_ptr, sizeof(struct counter));
}

/* ======================================================================= */

/*
 * -------------------------------------------------------------
 *  NAME   :  FindLoopInformation()
 *
 *  PURPOSE:  This is a recursive function which processes
 *            the binary tree of rules.
 *
 *  INPUTS :  A pointer to the rules list (RULE *), and
 *            a pointer to a structure for count info.
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  Actual processing of if/while info is done in 
 *            CountLoops function.
 * -------------------------------------------------------------
 */
 
static void FindLoopInformation(node, count)
RULE  *node;
struct counter *count;
{
   IF(node NEQ NULL) THEN
     FindLoopInformation(node->lft_rule, count);
     CountLoops(node, count);
     FindLoopInformation(node->rht_rule, count);
   END_IF
}

/* ======================================================================= */

/*
 * -------------------------------------------------------------
 *  NAME   :  CountLoops()
 *
 *  PURPOSE:  This function counts the number of times
 *            if or while loops are used.
 *
 *  INPUTS :  A pointer to the current rule (RULE *), and
 *            a pointer to a structure for count info.
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  
 * -------------------------------------------------------------
 */
 
static void CountLoops(temp, count)
RULE *temp;
struct counter *count;
{
   count->prime_count++;

   IF(temp->num_ifs > 0) THEN
      count->count_A++;
   END_IF

   IF(temp->num_whiles > 0) THEN
      count->count_B++;
   END_IF
}

/* ======================================================================= */

/*
 * -------------------------------------------------------------
 *  NAME   :  PrintLoopWarnings()
 *
 *  PURPOSE:  This function is called after all if/while info
 *            has been gathered and decides whether or not
 *            to print warnings.
 *
 *  INPUTS :  A pointer to the structure with count info (COUNTER *).
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  Uses global #define flags to determine when
 *            if/while has been overused.
 * -------------------------------------------------------------
 */
 
static void PrintLoopWarning(count)
struct counter *count;
{
   float perc_ifs, perc_whiles;

   IF((CHECK_STYLE IS_ON) AND (count->prime_count NEQ 0)) THEN
      perc_ifs    = ((float) count->count_A/(float) count->prime_count)*100.0;
      perc_whiles = ((float) count->count_B/(float) count->prime_count)*100.0;

      /*==============*/
      /*  Check If's  */
      /*==============*/

      IF(((perc_ifs > WARNING_LOOPS) AND 
         (count->prime_count > WARNING_MIN_RULES)) OR
         (perc_ifs > MAX_LOOPS)) THEN
         sprintf(MsgBuf,
            "\nWarning: If statements used in %4.1f%% of the rules", perc_ifs);
         SendMessage(MsgBuf,NO);
      END_IF

      /*=================*/
      /*  Check while's  */
      /*=================*/

      IF(((perc_whiles > WARNING_LOOPS) AND 
         (count->prime_count > WARNING_MIN_RULES)) OR
         (perc_whiles > MAX_LOOPS)) THEN
         sprintf(MsgBuf, 
                "\nWarning: While statements used in %4.1f%% of the rules",
               perc_whiles);
         SendMessage(MsgBuf,NO);
      END_IF
   END_IF
}

/* ======================================================================= */
/*                    CHECK FOR UNASSERTED RELATIONS                       */
/* ======================================================================= */

/*
 * -------------------------------------------------------------
 *  NAME   :  CheckMatchAssertions()
 *
 *  PURPOSE:  This function scans the list of relations to
 *            make sure all of them are asserted.
 *
 *  INPUTS :  A pointer to the relations list (RELATION *)
 *
 *  RETURNS:  The number of relations used.
 *
 *  NOTES  :
 * -------------------------------------------------------------
 */
 
static int CheckMatchAssertions(node)
RELATION *node;
{ 
   IF(node NEQ NULL) THEN
      int left = CheckMatchAssertions(node->lft_rel);

      IF(CHECK_STYLE IS_ON) THEN
         IF((node->assert_list EQ NULL) AND (node->rule_list NEQ NULL) AND
            (node->fb_list EQ NULL) AND
            (CheckDefexternalAssertions(node->name, DefexternalHead) EQ NULL) AND
            (strcmp(node->name, "initial-fact") NEQ 0)) THEN
            sprintf(MsgBuf, "\nWarning: Relation %.40s is never been asserted!",
               node->name);
            SendMessage(MsgBuf,NO);
         END_IF
      END_IF

      IF ((FindDeftemplate(DeftemplateHead,node->name)) 
       OR (    (node->rule_list == NULL)
           AND (node->assert_list == NULL)
           AND (node->retract_list == NULL)))THEN
        return (left + CheckMatchAssertions(node->rht_rel));
      ELSE
        return (left + 1 + CheckMatchAssertions(node->rht_rel));
      END_IF
   ELSE
      return (0);
   END_IF
}


/* ======================================================================= */

/*
 * -------------------------------------------------------------
 *  NAME   :  CheckDefexternalAssertions()
 *
 *  PURPOSE:  This function checks to see if the given 
 *            relation is listed as being asserted by a
 *            an external function.
 *
 *  INPUTS :  A pointer to the name of the current relation  
 *            (char *), and a pointer to the external function
 *            definition.
 *
 *  RETURNS:  A pointer to the external function definition, 
 *            or NULL if relation is not asserted.
 *
 *  NOTES  :  Uses the global def_external list pointer.
 *            Currently Stubbed out!
 * -------------------------------------------------------------
 */
 
static DEF_EXT *CheckDefexternalAssertions(rel_name, node)
char *rel_name;
DEF_EXT *node;
{
   DEF_EXT *rtn;
   int      found;

   IF(node NEQ NULL) THEN
      rtn = CheckDefexternalAssertions(rel_name, node->lft_def_ext);
      IF(rtn EQ NULL) THEN
         found = IsWordInList(node->assert_list, rel_name);
         IF(found EQ YES) THEN
            return(node);
         ELSE
            rtn = CheckDefexternalAssertions(rel_name, node->rht_def_ext);
            return(rtn);
         END_IF
      ELSE
         return(rtn);
      END_IF
   END_IF

   return(NULL);
}

