#include <stdio.h>
#include "crsv.h"

/********************************************************************
 *  This file contains all the functions used to process defrelation
 *  structures and store or retreive defrelation information.
 ********************************************************************
 */

/* ===========  Functions defined Externally but used here  =========== */

/* -----------------------------
 *  From the file: CRSVMEM.C
 * -----------------------------
 */

extern DR_PTR     AllocDefrelation();
extern void       FreeDefrelation();
extern DF_PTR     AllocDeffield();
extern WL_PTR     AllocWordList();
extern void       FreeWordList();
extern void       FreeDeffield();
extern WL_PTR     FreeWordAndGetNext();
extern FL_PTR     FreeFloatAndGetNext();
extern IL_PTR     FreeIntegerAndGetNext();

/* -----------------------------
 *  From the file: CRSVTKN.C
 * -----------------------------
 */

extern int      GetToken();

/* -----------------------------
 *  From the file: CRSVPRNT.C
 * -----------------------------
 */

extern void     ErrorMessage();
extern void     SendMessage();

/* -----------------------------
 *  From the file: CRSVMISC.C
 * -----------------------------
 */

extern int      FindRhtParen();
extern WL_PTR   AddToWordList();
extern IL_PTR	AddToIntegerList();
extern FL_PTR	AddToFloatList();
extern int      IsWordInList();
extern int      IsIntegerInList();
extern int      IsFloatInList();

/* -----------------------------
 *  From the file: CRSVPROC.C
 * -----------------------------
 */

extern RELATION  *FindRelation();

/* -----------------------------
 *  From the system
 * -----------------------------
 */

extern double atof();


/* ===========  Functions defined here for Global use  ================ */

int      ProcessDefrelation();
DEF_FLD *CopyFieldInfo();
void     CheckFieldInfo();
void     CheckVarFieldInfo();
void     CheckMinNumOfFields();


/* ===========  Functions defined here for internal use  ============== */

DEF_REL   *FindCreateDefrelation();
DR_PTR     CreateDefrelation();
DR_PTR     InsertDefrelation();
DEF_REL   *FindDefrelation();
int        ProcessDeffield();
DEF_FLD   *FindDeffield();
int        ProcessFieldType();
int        ProcessFieldWordList();
int        ProcessFieldStringList();
int        ProcessFieldInstanceList();
int        ProcessFieldNumberList();
int        ProcessFieldIntegerList();
int        ProcessFieldFloatList();
int        ProcessFieldRange();
int        CompareFieldTypes();
int        CompareFieldWordLists();
int        CompareFieldStringLists();
int        CompareFieldInstanceLists();
int        CompareFieldNumbers();
int        CompareIntegerListWithRange();
int        CompareFloatListWithRange();
int        CompareFieldRanges();
int        InIntegerRange();
int        InFloatRange();
void       DuplicateDefrelation();


/* ===========  Variables defined Externally but used here  =========== */

/* -----------------------------
 *  From the file: CRSV.C
 * -----------------------------
 */

extern char        *CurrentObjectName;         /* Name of current object   */
extern FILE        *CurrentFile;             /* current file pointer  */
extern char        *CurrentFileName;        /* Name of the current file */

extern DEF_REL     *DefrelationHead;      /* First Element in Defrelation list */
extern RELATION    *RelationHead;          /* First Element in relation list */

/* -----------------------------
 *  From the file: CRSVTKN.C
 * -----------------------------
 */

extern double  FLOATNUMBER;                  /* GetToken number return  */
extern long    INTNUMBER;
extern char   *TKNWORD;                    /* GetToken string storage */


/* ===========  Variables defined here for Global use  ================ */

int RelationCreatedBy = USER_MADE;

/* ===========  Variables defined here for internal use  ============== */

DEF_REL *CurrentDefrelation;
int      IsATemplate;


/* ======================================================================= */
/*                  FUNCTIONS TO PROCESS DEFRELATIONS                      */
/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : ProcessDefrelation()
 *
 *  PURPOSE: This function reads and stores in the
 *           appropriate data structures all of the
 *           information defined in a defrelation
 *           construct.
 *
 *  INPUTS:  None.
 *
 *  RETURNS: An integer, -99 (RESTART) if it got
 *           hopelessly confused while trying to
 *           read the defrelation, a -1 if it found
 *           some kind of syntax error, otherwise, a
 *           1 (OK).
 *
 *  NOTES:   The global symbol, DefrelationHead, points at the newly
 *           processed symbol.
 * ------------------------------------------------------------------
 */

int ProcessDefrelation()
{
   int        token;
   DEF_REL   *CurDefrelation;

   /*============================*/
   /* Check for defrelation name */
   /*============================*/

   token = GetToken(CurrentFile);

   IF((token NEQ WORD)AND(token NEQ STRING) AND 
      (token NEQ INTEGER) AND (token NEQ FLOAT)AND
      (token NEQ INSTANCE)) THEN
      ErrorMessage(ERROR, "Expected a name following defrelation keyword!");
      return(RESTART);
   END_IF

   CurrentObjectName  = TKNWORD;
   IF (RelationCreatedBy EQ TEMPLATE_MADE) THEN
     IsATemplate = 1;
   ELSE
     IsATemplate = 0;
   END_IF


   /*======================================*/
   /* Get pointer to defrelation structure */
   /*======================================*/
   
   if(( CurDefrelation = FindDefrelation(DefrelationHead,CurrentObjectName)) != NULL)
     {
       if(CurDefrelation->created_by == USER_MADE) {
         ErrorMessage(ERROR,"Duplicate defrelation, ignoring new definition\n");
         return(RESTART);}
       else if(CurDefrelation->created_by == TEMPLATE_MADE)
        {
          ErrorMessage(ERROR,"Same name with deftemplate, ignoring defrelation\n");
          return(RESTART);
        }
       else{
         FreeDeffield(CurDefrelation->fields);
         CurDefrelation->fields = NULL;
         CurDefrelation->file = CurrentFileName;
         CurDefrelation->created_by = RelationCreatedBy;
         }
     }
    else
      CurDefrelation = CurrentDefrelation = CreateDefrelation(CurrentObjectName);
  
   CurDefrelation->created_by = RelationCreatedBy;


   /*==============================================================*/
   /* If Verbose flag is on and relation was not already processed */
   /* as a Template, then print out "Processing defrelation ..."   */
   /*==============================================================*/



   IF((VERBOSE IS_ON) AND (!IsATemplate)) THEN
     sprintf(MsgBuf,"\n\nProcessing defrelation: %.40s.",CurrentObjectName);
     SendMessage(MsgBuf,NO);
   END_IF

   /*============================================*/
   /* Check if relation has been previously used */
   /*============================================*/

   IF((FindRelation(RelationHead, CurrentObjectName)) AND (!IsATemplate)) THEN
      sprintf(MsgBuf, "Relation %.40s defined after being used.",
         CurrentObjectName);
      ErrorMessage(WARNING, MsgBuf);
   END_IF

   /*===================================*/
   /* Begin processing defrelation info */
   /*===================================*/

   token = GetToken(CurrentFile);

   IF(token NEQ LPAREN) THEN
      IF (!IsATemplate) THEN
         ErrorMessage(ERROR, "Expecting a left paren inside a defrelation.");
         return(RESTART);
      END_IF
   END_IF


   /*=================================*/
   /* Check for ART style defrelation */
   /*=================================*/

   IF((token EQ BWORD) OR (token EQ BWORDS)) THEN
      FreeDefrelation(CurDefrelation);
      return(RESTART);                      /* Return error, not restart */
   END_IF

   /*===================================*/
   /* Read and process defrelation info */
   /*===================================*/

   while(token NEQ RPAREN) DO

      /*======================================*/
      /* Find opening paren, and get key word */
      /*======================================*/

      IF(token NEQ LPAREN) THEN
         IF (!IsATemplate){
           ErrorMessage(ERROR, "Expecting a left paren inside a defrelation.");
           FreeDefrelation(CurDefrelation);
           return(RESTART);}
      END_IF

      token = GetToken(CurrentFile);

      IF((token EQ WORD) AND
         (strcmp("min-number-of-fields", TKNWORD) EQ 0)) THEN

         /*==============================*/
         /* Process min-number-of-fields */
         /*==============================*/

         token = GetToken(CurrentFile);
         IF((token EQ BWORD) AND (strcmp("VARIABLE", TKNWORD) EQ 0)) THEN
            CurDefrelation->min_fields = 0;
         ELSE_IF(token NEQ INTEGER)
            ErrorMessage(ERROR,
                          "Expecting an integer for field length definition.");
            FreeDefrelation(CurDefrelation);
            return(RESTART);
         ELSE
            CurDefrelation->min_fields = (int)INTNUMBER;
         END_IF
         if(FindRhtParen() != OK){
           FreeDefrelation(CurDefrelation);
           return(RESTART);}

      ELSE_IF((token EQ WORD) AND
         (strcmp("max-number-of-fields", TKNWORD) EQ 0))

         /*==============================*/
         /* Process max-number-of-fields */
         /*==============================*/

         token = GetToken(CurrentFile);
         IF((token EQ BWORD) AND (strcmp("VARIABLE", TKNWORD) EQ 0)) THEN
            CurDefrelation->max_fields = 10000;
         ELSE_IF(token NEQ INTEGER)
            ErrorMessage(ERROR,
                          "Expecting a number for field length definition.");
            FreeDefrelation(CurDefrelation);
            return(RESTART);
         ELSE
            CurDefrelation->max_fields = (int)INTNUMBER;
         END_IF
         if(FindRhtParen() != OK){
           FreeDefrelation(CurDefrelation);
           return(RESTART);}

      ELSE_IF((token EQ WORD) AND
         (strcmp("field", TKNWORD) EQ 0))

         /*============================*/
         /* Process a field definition */
         /*============================*/

         if(ProcessDeffield(CurDefrelation) != OK)
         {
           FreeDefrelation(CurDefrelation);
           return(RESTART);
          }

      ELSE

         /*========================*/
         /* Unrecognized key word! */
         /*========================*/

         sprintf(MsgBuf,
            "Unrecognized attribute: %.40s inside a defrelation.",
            TKNWORD);
         ErrorMessage(ERROR, MsgBuf);
         FreeDefrelation(CurDefrelation);
         return(RESTART);
      END_IF

      token = GetToken(CurrentFile);
   END_WHILE
  if((CurDefrelation->lft_def_rel == NULL) AND (CurDefrelation->rht_def_rel == NULL))
     DefrelationHead = InsertDefrelation(DefrelationHead,CurDefrelation);
  return(OK);
}

/* ======================================================================= */
/* ------------------------------------------------------------------
 *  NAME   :  InsertDefrelation()
 *
 *  PURPOSE: This function inserts a defrelation in to the banary tree list.
 *
 *  INPUTS : a pointer pointing to the defrelation  being inserted
 *         : a poiter pointing to the root of the list
 *
 *  RETURNS: new root of the list
 *
 *  Note:    
 * ------------------------------------------------------------------
 */ 
 DR_PTR InsertDefrelation(list,node)
 DR_PTR list,node;
 {
    DR_PTR DefRelPtr = NULL,DefRelParent = NULL, NewDefRel;
    int rtn;
   
   if (list == NULL)
     return(node);
   DefRelPtr = list;
   while(DefRelPtr != NULL)
    {
       rtn = strcmp(node->name,DefRelPtr->name);
       if(rtn < 0)
        {
          DefRelParent = DefRelPtr;
          DefRelPtr = DefRelPtr->lft_def_rel;   
        }
       else if(rtn > 0)
        {
          DefRelParent = DefRelPtr;
          DefRelPtr = DefRelPtr->rht_def_rel;   
        }
       else{
         ErrorMessage(ERROR,"CRSV error in InsertDefrelation, Please report this error\n");
         return(list);}
    }
   if(rtn < 0)
      DefRelParent->lft_def_rel = node;
   else if(rtn > 0)
      DefRelParent->rht_def_rel = node;
   else
      ErrorMessage(ERROR,"CRSV error in InsertDefrelation, Please report this error\n");
   return(list);
 }

/* ------------------------------------------------------------------
 *  NAME   : FindCreateDefrelation()
 *
 *  PURPOSE: This function searches through a list of
 *           defrelations to find a specific name. If it doesn't
 *           find an entry with that name, it creates one.
 *
 *  INPUTS : A pointer to the list of defrelations (DEF_REL *)
 *           and a pointer to the name to search for (char *)
 *
 *  RETURNS: If the name is found, it returns a pointer to
 *           that structure. If it isn't found it returns a
 *           pointer to the newly created structure.
 * ------------------------------------------------------------------
 */

DEF_REL *FindCreateDefrelation(node, name)
DR_PTR  node;
char   *name;
{
   int rtn;

   IF(node EQ NULL) THEN                       /* A new defrelation! */
      node        = AllocDefrelation();
      node->name  = name;
      node->file  = CurrentFileName;
      CurrentDefrelation  = node;
   ELSE
      rtn = strcmp(name, node->name);
      IF(rtn < 0) THEN
         node->lft_def_rel = FindCreateDefrelation(node->lft_def_rel, name);
      ELSE_IF(rtn > 0)
         node->rht_def_rel = FindCreateDefrelation(node->rht_def_rel, name);
      ELSE
         if(node->created_by != TEMPLATE_MADE)
           DuplicateDefrelation(name,"relation");
         node->max_fields       = 10000;
         node->min_fields       = 0;
         node->file             = CurrentFileName;
         CurrentDefrelation     = node;
         FreeDeffield(node->fields);
         node->fields           = NULL;
      END_IF
   END_IF

   return(node);
}
/* ------------------------------------------------------------------
 *  NAME   : CreateDefrelation()
 *
 *  PURPOSE: This function creates a defrelation node.
 *
 *  INPUTS : a pointer to the name 
 *
 *  RETURNS: a new node
 * ------------------------------------------------------------------
 */
DEF_REL *CreateDefrelation(name)
char   *name;
{
   DEF_REL *node;                      /* A new defrelation! */
      
      node        = AllocDefrelation();
      node->name  = name;
      node->file  = CurrentFileName;
   return(node);
}

/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : DuplicateDefrelation()
 *
 *  PURPOSE: This function merely prints an error message if
 *           a duplicate relation definition is found.
 *
 *  INPUTS : A pointer to the relation name.
 *
 *  RETURNS: Nothing
 *
 *  NOTES  : This function is called to avoid putting a large
 *           string buffer on the stack many times.
 * ------------------------------------------------------------------
 */

static void DuplicateDefrelation(name,definition)
char *name,*definition;
{
   sprintf(MsgBuf, "Redefining %s %.40s, erasing old info.",definition,name);
   ErrorMessage(WARNING, MsgBuf);

   return;
}

/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : FindDefrelation()
 *
 *  PURPOSE: This function searches through a list of
 *           defrelations to find a specific name.
 *
 *  INPUTS : A pointer to the list of defrelations (DEF_REL *)
 *           and a pointer to the name to search for (char *)
 *
 *  RETURNS: If the name is found, it returns a pointer to
 *           that structure. If it isn't found it returns NULL.
 * ------------------------------------------------------------------
 */

DEF_REL *FindDefrelation(node, name)
DR_PTR  node;
char   *name;
{
   int rtn;

   IF(node EQ NULL) THEN                       /* A new defrelation! */
      return(NULL);
   ELSE
      rtn = strcmp(name, node->name);
      IF(rtn < 0) THEN
         return(FindDefrelation(node->lft_def_rel, name));
      ELSE_IF(rtn > 0)
         return(FindDefrelation(node->rht_def_rel, name));
      ELSE
         return(node);                         /* An old defrelation! */
      END_IF
   END_IF
}

/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : ProcessDeffield()
 *
 *  PURPOSE: This function reads and stores in the
 *           appropriate data structures all of the
 *           information about fields from a defrelation.
 *
 *  INPUTS:  A pointer to the current defrelation.
 *
 *  RETURNS: An integer, either OK or ERROR.
 *
 *  NOTES:   None.
 * ------------------------------------------------------------------
 *                      Pseudo Code (may not be complete)
 *
 *   get token
 *   IF token is not a number THEN
 *      send error message
 *      find right paren
 *      return ERROR
 *   ELSE_IF token is < 2 OR > max
 *      send error message
 *      find right paren
 *      return ERROR
 *   END_IF
 *   find_def_field()   -(Check for duplicate field info, create new struct)
 *   IF field was found THEN
 *      send error message
 *      reset values
 *   ELSE
 *      create a new list entry
 *      reset list pointers
 *   END_IF
 *   get token
 *   WHILE token is not a ')' DO    -(Switch on field keywords)
 *      IF token is not a '(' THEN
 *         send error message
 *         find right paren
 *         return ERROR
 *      END_IF
 *      get token
 *
 *      CASE "type"
 *         ProcessFieldType()
 *      CASE "allowed-strings"
 *         ProcessFieldStringList()
 *      CASE "allowed-words"
 *         ProcessFieldWordList()
 *      CASE "allow-instance"
 *         process_instance_list()
 *      CASE "allowed-numbers"
 *         ProcessFieldNumberList()
 *      CASE "allowed-integers"
 *         ProcessFieldIntegerList()
 *      CASE "allowed-floats"
 *		   ProcessFieldFloatList()
 *      CASE "range"
 *         ProcessFieldRange()
 *      CASE anything else
 *         send error message
 *         find right paren
 *      END CASE
 *      get token
 *   END_WHILE
 *   return OK
 * ------------------------------------------------------------------
 */

static int ProcessDeffield(CurDefrelation)
DEF_REL *CurDefrelation;
{
   int token;
   DEF_FLD *CurDeffield;
   short SetValue;

   token = GetToken(CurrentFile);

   /*=======================*/
   /* Check position number */
   /*=======================*/

   IF(token NEQ INTEGER) THEN
      ErrorMessage(ERROR, "Field keyword must be followed by an integer.");
      return(RESTART);
   ELSE_IF((INTNUMBER < 2) OR (INTNUMBER > CurDefrelation->max_fields))
      sprintf(MsgBuf,"Field position, %d, is too big or too small.",
                  INTNUMBER);
      ErrorMessage(ERROR, MsgBuf);
      return(RESTART);
   END_IF

   /*==============================================*/
   /* Check if field postion has already been used */
   /*==============================================*/

   CurDeffield = FindDeffield((int)INTNUMBER, CurDefrelation->fields);

   IF(CurDeffield NEQ NULL) THEN
      ErrorMessage(ERROR, "Duplicate field definition\n");
      return(RESTART);
   
   ELSE
      CurDeffield             = AllocDeffield();
      CurDeffield->position   = (int)INTNUMBER;
   END_IF


   /*=======================================*/
   /* Field found, start examining keywords */
   /*=======================================*/

   token = GetToken(CurrentFile);
   SetValue = FALSE;


   while(token NEQ RPAREN) DO

      /*======================================*/
      /* Find opening paren, and get key word */
      /*======================================*/

      IF(token NEQ LPAREN) THEN
         IF (!IsATemplate)
           ErrorMessage(ERROR,
                 "Expecting a left paren inside field definition.");
         return(RESTART);
      END_IF

      token = GetToken(CurrentFile);
      /*========================*/
      /* Process field keywords */
      /*========================*/

      if(ProcessFieldKeyword(CurDeffield,token,&SetValue,CurDefrelation->created_by) != OK)
       {
          FreeDeffield(CurDeffield);
          return(RESTART);
       }
      token = GetToken(CurrentFile);
   END_WHILE
   
   if(CheckDefField(CurDeffield) != OK)
    {
      FreeDeffield(CurDeffield);
      return(RESTART);
    }
   CurDeffield->next_field = CurDefrelation->fields;
   CurDefrelation->fields     = CurDeffield;
   return(OK);
}

/*
 * ---------------------------------------------------
 * ProcessFieldKeyword
 *  Process the attribute found in the field definition
 * Input: - definition of a field
 *        - flag when allowed-values attribute is used 
 *          in the field definition
 * Return RESTART if found syntax error
 * ---------------------------------------------------
 */

ProcessFieldKeyword(CurDeffield,token,SetValue,CreatedBy)
DEF_FLD *CurDeffield;
int token;
short *SetValue,CreatedBy;
{

      IF((token EQ WORD) AND (strcmp("type", TKNWORD) EQ 0)) THEN
         if(ProcessFieldType(CurDeffield) != OK)
           return(RESTART);
      ELSE_IF((token EQ WORD) AND (strcmp("allowed-strings", TKNWORD) EQ 0))
         IF(CurDeffield->possible_strings != NULL)THEN
           ErrorMessage(ERROR,"The allowed-strings attribute has already been defined\n");
           return(RESTART);
         ELSE
           if(ProcessFieldStringList(CurDeffield) != OK)
             return(RESTART);
         END_IF
      ELSE_IF((token EQ WORD) AND ((strcmp("allowed-words", TKNWORD) EQ 0) ||
                                   (strcmp("allowed-symbols", TKNWORD) EQ 0)))
         IF(CurDeffield->possible_words != NULL)THEN
           ErrorMessage(ERROR,"The allowed-symbols/allowed_words attribute has already been defined\n");
           return(RESTART);
         ELSE
           if(ProcessFieldWordList(CurDeffield) != OK)
             return(RESTART);
         END_IF
      ELSE_IF((token EQ WORD) AND (strcmp("allowed-instances", TKNWORD) EQ 0))
         IF(CurDeffield->possible_instances != NULL)THEN
           ErrorMessage(ERROR,"The allowed-instances attribute has already been defined\n");
           return(RESTART);
         ELSE
           if(ProcessFieldInstanceList(CurDeffield) != OK)
             return(RESTART);
         END_IF
      ELSE_IF((token EQ WORD) AND (strcmp("allowed-numbers", TKNWORD) EQ 0))
         IF((CurDeffield->possible_integers != NULL) ||(CurDeffield->possible_floats != NULL)) THEN
           ErrorMessage(ERROR,"The allowed-numbers attribute has already been defined\n");
           return(RESTART);
         ELSE
           if(ProcessFieldNumberList(CurDeffield) != OK)
             return(RESTART);
         END_IF
      ELSE_IF((token EQ WORD) AND (strcmp("allowed-integers", TKNWORD) EQ 0))
         IF(CurDeffield->possible_integers != NULL)THEN
           ErrorMessage(ERROR,"The allowed-integers attribute has already been defined\n");
           return(RESTART);
         ELSE
           if(ProcessFieldIntegerList(CurDeffield) != OK)
             return(RESTART);
         END_IF
      ELSE_IF((token EQ WORD) AND (strcmp("allowed-floats", TKNWORD) EQ 0))
         IF(CurDeffield->possible_floats != NULL)THEN
           ErrorMessage(ERROR,"The allowed-floats attribute has already been defined\n");
           return(RESTART);
         ELSE
           if(ProcessFieldFloatList(CurDeffield) != OK)
              return(RESTART);;
         END_IF
      ELSE_IF((token EQ WORD) AND (strcmp("range", TKNWORD) EQ 0))
         IF(CurDeffield->set_int_min OR CurDeffield->set_int_max OR
            CurDeffield->set_float_min OR CurDeffield->set_float_max)THEN
           ErrorMessage(ERROR,"The range attribute has already been defined\n");
           return(RESTART);
         ELSE
            if(ProcessFieldRange(CurDeffield) != OK)
              return(RESTART);;
         END_IF
      ELSE_IF((token EQ WORD) AND (strcmp("allowed-values",TKNWORD) EQ 0))
         IF(!(*SetValue))THEN
           *SetValue = TRUE;
           if(ProcessFieldValues(CurDeffield) != OK)
             return(RESTART);;
         ELSE
            ErrorMessage(MsgBuf,"Allowed-values attributes has already been defined\n");
            return(RESTART);
         END_IF
      ELSE
         sprintf(MsgBuf, "Unrecognized key word: %.40s inside a %s.",
                       TKNWORD, "field definition");
         ErrorMessage(ERROR, MsgBuf);
         return(RESTART);
      END_IF
   return(OK);
}


/*-------------------------------------------------------------
 *  CheckDefField
 *  Check for any contradiction in the definition of a field
 *  in a defrelation or detemplate
 *-------------------------------------------------------------
 */

CheckDefField(CurDeffield)
DEF_FLD *CurDeffield;
{
  double min = -1000000.0, max = 1000000.0;
  IF((CurDeffield->possible_words != NULL) AND (CurDeffield->allow_word IS_NO)) THEN
     ErrorMessage(ERROR,"the allowed-words attribute conflicts with attribute type\n");
     FreeWordList(CurDeffield->possible_words);
     CurDeffield->possible_words = NULL;
     return(RESTART);
   END_IF
   IF((CurDeffield->possible_strings != NULL) AND (CurDeffield->allow_string IS_NO))THEN
     ErrorMessage(ERROR,"the allowed-strings attribute conflicts with attribute type\n");
     FreeWordList(CurDeffield->possible_strings);
     CurDeffield->possible_strings = NULL;
     return(RESTART);
   END_IF
   IF((CurDeffield->possible_instances != NULL) AND (CurDeffield->allow_instance IS_NO))THEN
     ErrorMessage(ERROR,"the allowed-instances attribute conflicts with attribute type\n");
     FreeWordList(CurDeffield->possible_instances);
     CurDeffield->possible_instances = NULL;
     return(RESTART);
   END_IF
   IF((CurDeffield->possible_integers != NULL) AND (CurDeffield->allow_integer IS_NO))THEN
     ErrorMessage(ERROR,"the allowed-integers/numbers atribute conflicts with type attribute\n");
     FreeIntegerList(CurDeffield->possible_integers);
     CurDeffield->possible_integers = NULL;
     return(RESTART);
   END_IF
   IF((CurDeffield->possible_floats != NULL) AND (CurDeffield->allow_float IS_NO))THEN
     ErrorMessage(ERROR,"the allowed-floats/numbers atribute conflicts with type attribute\n");
     FreeIntegerList(CurDeffield->possible_integers);
     CurDeffield->possible_integers = NULL;
     return(RESTART);
   END_IF
   IF(((CurDeffield->possible_integers)OR(CurDeffield->possible_floats != NULL))
    AND ((CurDeffield->set_int_min IS_YES) OR(CurDeffield->set_int_max IS_YES)OR
    (CurDeffield->set_float_min IS_YES) OR (CurDeffield->set_float_max IS_YES)))THEN
     ErrorMessage(ERROR,"allowed-integers/floats/numbers cannot be used \n");
     SendMessage("       in conjunction with the range attribute\n",NO);
     CurDeffield->set_int_min  = NO;
     CurDeffield->set_int_max  = NO;
     return(RESTART);
   END_IF
   IF(((CurDeffield->set_int_min)||(CurDeffield->set_int_max)) && 
      (CurDeffield->allow_integer IS_NO))THEN
     ErrorMessage(ERROR,"Range attribute conflicts with type attribute\n");
     CurDeffield->set_int_min = NO;
     CurDeffield->set_int_max = NO; 
     return(RESTART);
   END_IF   
   IF(((CurDeffield->set_float_min)||(CurDeffield->set_float_max)) && 
      (CurDeffield->allow_float IS_NO))THEN
     ErrorMessage(ERROR,"Range attribute conflicts with type attribute\n");
     CurDeffield->set_float_min = NO;
     CurDeffield->set_float_max = NO; 
     return(RESTART); 
   END_IF
   IF(CurDeffield->set_int_min) THEN
     min = (double)CurDeffield->int_min_range;
   ELSE_IF (CurDeffield->set_float_min)
     min = CurDeffield->float_min_range;
   END_IF
   IF(CurDeffield->set_int_max)THEN
     max = (double)CurDeffield->int_max_range;
   ELSE_IF(CurDeffield->set_float_max)
     max = CurDeffield->float_max_range;
   END_IF
   if(max < min) {
     ErrorMessage(ERROR,
       "Minimum value is larger than maximum value in range attribute\n");
       return(RESTART);
       }
  return(OK);
     
}

/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : FindDeffield()
 *
 *  PURPOSE: This function searches through a list of
 *           field definitions to find a specific field number.
 *
 *  INPUTS : A pointer to the list of field definitions (DF_PTR *)
 *           and a field number to search for (int).
 *
 *  RETURNS: If the number is found, it returns a pointer to
 *           that structure. If it isn't found, it returns NULL.
 * ------------------------------------------------------------------
 */

DF_PTR FindDeffield(number, list)
int number;
DF_PTR   list;

{
   DF_PTR temp;

   temp = list;
   while(temp NEQ NULL) DO
      IF(number EQ temp->position) THEN
         return(temp);
      END_IF

      temp = temp->next_field;
   END_WHILE

   return(NULL);
}


/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : ProcessFieldType()
 *
 *  PURPOSE: This function reads and inteprets the
 *           allowable data type for a field
 *
 *  INPUTS:  A pointer to the current field definition (DF_PTR *).
 *
 *  RETURNS: nothing.
 *
 *  NOTES:   None.
 * ------------------------------------------------------------------
 *                      Pseudo Code (may not be complete)
 *
 *   get token
 *   WHILE token isn't a ')' DO
 *      IF token is ?VARIABLE THEN
 *         set all allowed types to YES
 *         find right paren
 *         return OK
 *      ELSE_IF token is "NUMBER" THEN
 *         set allowed number flag
 *      ELSE_IF token is "WORD"  or "SYMBOL" THEN
 *         set allowed word flag
 *      ELSE_IF token is "STRING" THEN
 *         set allowed string flag
 *      ELSE
 *         send error message
 *      END_IF
 *      get token
 *   END_WHILE
 *   IF no flags are set THEN
 *      send error message
 *      return ERROR
 *   END_IF
 *   return OK
 * ------------------------------------------------------------------
 */

int ProcessFieldType(CurDeffield)
DF_PTR CurDeffield;
{
   int token;
   short SetInt    = NO;
   short SetFloat  = NO;
   short SetWord   = NO;
   short SetString = NO;
   short SetInstance = NO;
   short SetAddress  = NO;

   token = GetToken(CurrentFile);
   IF(token == RPAREN)THEN
     ErrorMessage(ERROR,"Allowed type attribute values are\n");
     SendMessage("WORD,STRING,NUMBER,INTEGER,FLOAT,INSTANCE, EXTERNAL-ADDRESS\n");
     return(RESTART);
   END_IF
   while(token NEQ RPAREN) DO
      IF((token EQ BWORD) AND (strcmp(TKNWORD,"VARIABLE") EQ 0)) THEN
         CurDeffield->allow_word   = YES;
         CurDeffield->allow_integer = YES;
         CurDeffield->allow_float = YES;
         CurDeffield->allow_string = YES;
         CurDeffield->allow_instance = YES;
         CurDeffield->allow_address = YES;
         if(FindRhtParen() != OK) 
           return(RESTART);
         return(OK);
      ELSE_IF(strcmp(TKNWORD,"NUMBER") EQ 0)
         SetInt = YES;
         SetFloat = YES;
      ELSE_IF (strcmp(TKNWORD,"INTEGER") EQ 0)
         SetInt = YES;
      ELSE_IF (strcmp(TKNWORD,"FLOAT") EQ 0)
         SetFloat = YES;
      ELSE_IF(strcmp(TKNWORD,"STRING") EQ 0)
         SetString = YES;
      ELSE_IF((strcmp(TKNWORD,"WORD") EQ 0) OR
              (strcmp(TKNWORD,"SYMBOL") EQ 0))
         SetWord = YES;
      ELSE_IF(strcmp(TKNWORD,"LEXEME") EQ 0)
         SetWord = YES;
         SetString = YES;
      ELSE_IF(strcmp(TKNWORD,"INSTANCE") EQ 0)
         SetInstance = YES;
      ELSE_IF(strcmp(TKNWORD,"EXTERNAL-ADDRESS") EQ 0)
         SetAddress = YES;
      ELSE
         ErrorMessage(ERROR, "Unknown field or argument type.");
         return(RESTART);
      END_IF

      token = GetToken(CurrentFile);
   END_WHILE

   CurDeffield->allow_word     = SetWord;
   CurDeffield->allow_string   = SetString;
   CurDeffield->allow_float    = SetFloat;
   CurDeffield->allow_integer  = SetInt;
   CurDeffield->allow_instance = SetInstance;
   CurDeffield->allow_address  = SetAddress;
   return(OK);
}


/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : ProcessFieldWordList()
 *
 *  PURPOSE: This function reads a list of
 *           allowable words for a field
 *
 *  INPUTS:  A pointer to the current field definition (DF_PTR *).
 *
 *  RETURNS: Nothing.
 *
 *  NOTES:   None.
 * ------------------------------------------------------------------
 *                      Pseudo Code (may not be complete)
 *
 *   get token
 *   WHILE token isn't a ')' DO
 *      IF token is ?VARIABLE THEN
 *         set word list to NULL
 *         find right paren
 *         return OK
 *      ELSE_IF token is not a "WORD" THEN
 *         send error message
 *      ELSE
 *         add word to word list
 *      END_IF
 *      get token
 *   END_WHILE
 *   return OK
 * ------------------------------------------------------------------
 */

int ProcessFieldWordList(CurDeffield)
DF_PTR CurDeffield;
{
   int token;

   token = GetToken(CurrentFile);
   if(token == RPAREN)
    {
      ErrorMessage(ERROR,"Allowed-words/symbols attributes must have some value(s)\n");
      return(RESTART);
    }
   /*============================*/
   /* Read list of allowed words */
   /*============================*/

   while(token NEQ RPAREN) DO

      /*===============================*/
      /* If ?VARIABLE used, clear list */
      /*===============================*/

      IF((token EQ BWORD) AND (strcmp(TKNWORD,"VARIABLE") EQ 0)) THEN
         IF(CurDeffield->possible_words != NULL)THEN
            ErrorMessage(ERROR,"?VARIABLE must be by itself in the allowed-words attribute\n");
            FreeWordList(CurDeffield->possible_words);
            CurDeffield->possible_words = NULL;
            return(RESTART);
          ELSE
            token = GetToken(CurrentFile);
            IF(token != RPAREN)THEN
              ErrorMessage(ERROR,"?VARIABLE must be by itself in the allowed-words attribute\n");
              return(RESTART);
            END_IF
            return(OK);
          END_IF

      /*===========================*/
      /* If not a WORD, then error */
      /*===========================*/

      ELSE_IF(token NEQ WORD)
         sprintf(MsgBuf, "%s is of wrong type. Expecting a SYMBOL.",TKNWORD);
         ErrorMessage(ERROR, MsgBuf);
         return(RESTART);
      /*========================*/
      /* Else, add word to list */
      /*========================*/

      ELSE
          CurDeffield->possible_words =
                      AddToWordList(CurDeffield->possible_words,
                                       TKNWORD, WORD,YES);
      END_IF
   token = GetToken(CurrentFile);
   END_WHILE

   return(OK);
}


/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : ProcessFieldStringList()
 *
 *  PURPOSE: This function reads a list of
 *           allowable strings for a field
 *
 *  INPUTS:  A pointer to the current field definition (DF_PTR *).
 *
 *  RETURNS: Nothing.
 *
 *  NOTES:   None.
 * ------------------------------------------------------------------
 */

int ProcessFieldStringList(CurDeffield)
DF_PTR CurDeffield;
{
   int token;

   token = GetToken(CurrentFile);

   if(token == RPAREN)
    {
      ErrorMessage(ERROR,"Allowed-strings attributes must have some value(s)\n");
      return(RESTART);
    }
   
   /*============================*/
   /* Read list of allowed words */
   /*============================*/

   while(token NEQ RPAREN) DO

      /*===============================*/
      /* If ?VARIABLE used, clear list */
      /*===============================*/
      IF((token EQ BWORD) AND (strcmp(TKNWORD,"VARIABLE") EQ 0)) THEN
         IF(CurDeffield->possible_strings != NULL)THEN
            ErrorMessage(ERROR,"?VARIABLE must be by itself in the allowed-strings attribute\n");
            FreeWordList(CurDeffield->possible_strings);
            CurDeffield->possible_strings = NULL;
            return(RESTART);
          ELSE
            token = GetToken(CurrentFile);
            IF(token != RPAREN)THEN
              ErrorMessage(ERROR,"?VARIABLE must be by itself in the allowed-strings attribute\n");
              return(RESTART);
            END_IF
            return(OK);
          END_IF
      /*=============================*/
      /* If not a STRING, then error */
      /*=============================*/

      ELSE_IF(token NEQ STRING)
         sprintf(MsgBuf, "%s is of wrong type. Expecting a STRING.",TKNWORD);
         ErrorMessage(ERROR,MsgBuf);
         return(RESTART);
      /*==========================*/
      /* Else, add string to list */
      /*==========================*/

      ELSE
          CurDeffield->possible_strings =
                           AddToWordList(CurDeffield->possible_strings,
                                            TKNWORD, STRING,YES);
      END_IF
   token = GetToken(CurrentFile);
   END_WHILE

   return(OK);
}

/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : ProcessFieldInstanceList()
 *
 *  PURPOSE: This function reads a list of
 *           allowable words for a field
 *
 *  INPUTS:  A pointer to the current field definition (DF_PTR *).
 *
 *  RETURNS: Nothing.
 *
 *  NOTES:   None.
 * ------------------------------------------------------------------
 *                      Pseudo Code (may not be complete)
 *
 *   get token
 *   WHILE token isn't a ')' DO
 *      IF token is ?VARIABLE THEN
 *         set word list to NULL
 *         find right paren
 *         return OK
 *      ELSE_IF token is not a "WORD" THEN
 *         send error message
 *      ELSE
 *         add word to word list
 *      END_IF
 *      get token
 *   END_WHILE
 *   return OK
 * ------------------------------------------------------------------
 */

int ProcessFieldInstanceList(CurDeffield)
DF_PTR CurDeffield;
{
   int token;

   token = GetToken(CurrentFile);
   if(token == RPAREN){
     ErrorMessage(ERROR,"Allowed-instances attributes must have some value(s)\n");
     return(RESTART);}
   /*================================*/
   /* Read list of allowed instances */
   /*================================*/

   while(token NEQ RPAREN) DO

      /*===============================*/
      /* If ?VARIABLE used, clear list */
      /*===============================*/

      IF((token EQ BWORD) AND (strcmp(TKNWORD,"VARIABLE") EQ 0)) THEN
         IF(CurDeffield->possible_instances != NULL)THEN
            ErrorMessage(ERROR,"?VARIABLE must be by itself in the allowed-instances attribute\n");
            FreeWordList(CurDeffield->possible_instances);
            CurDeffield->possible_instances = NULL;
            return(RESTART);
          ELSE
            token = GetToken(CurrentFile);
            IF(token != RPAREN)THEN
              ErrorMessage(ERROR,"?VARIABLE must be by itself in the allowed-instances attribute\n");
              return(RESTART);
            END_IF
            return(OK);
          END_IF

      /*===========================*/
      /* an INSTANCE               */
      /*===========================*/

      ELSE_IF(token NEQ INSTANCE)
         sprintf(MsgBuf, "%s is of wrong type. Expecting an INSTANCE.",TKNWORD);
         ErrorMessage(ERROR, MsgBuf);
         return(RESTART);
      /*========================*/
      /* Else, add word to list */
      /*========================*/

      ELSE
          CurDeffield->possible_instances =
                      AddToWordList(CurDeffield->possible_instances,
                                       TKNWORD, INSTANCE,YES);
      END_IF
   token = GetToken(CurrentFile);
   END_WHILE

   return(OK);
}


/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : ProcessFieldNumberList()
 *
 *  PURPOSE: This function reads a list of
 *           allowable numbers for a field
 *
 *  INPUTS:  A pointer to the current field definition (DF_PTR *).
 *
 *  RETURNS: nothing.
 *
 *  NOTES:   None.
 * ------------------------------------------------------------------
 */

int ProcessFieldNumberList(CurDeffield)
DF_PTR CurDeffield;
{
   int token;

   token = GetToken(CurrentFile);
   if(token == RPAREN){
     ErrorMessage(ERROR,"Allowed-numbers attributes must have some value(s)\n");
     return(RESTART);}
   /*============================*/
   /* Read list of allowed words */
   /*============================*/

   while(token NEQ RPAREN) DO

      /*===============================*/
      /* If ?VARIABLE used, clear list */
      /*===============================*/

      IF((token EQ BWORD) AND (strcmp(TKNWORD,"VARIABLE") EQ 0)) THEN
         IF((CurDeffield->possible_integers != NULL) OR
           (CurDeffield->possible_floats != NULL)) THEN
           ErrorMessage(ERROR,"?VARIABLE must be by itself in the allowed-numbers attribute\n");
           FreeIntegerList(CurDeffield->possible_integers);
           FreeFloatList(CurDeffield->possible_floats);
           CurDeffield->possible_integers = NULL;
           CurDeffield->possible_floats = NULL;
           return(RESTART);
         ELSE
           token = GetToken(CurrentFile);
           IF(token != RPAREN) THEN
             ErrorMessage(ERROR,"?VARIABLE must be by itself in the allowed-numbers attribute\n");
             return(RESTART);
           END_IF
           return(OK);
         END_IF
      /*=============================*/
      /* If not a NUMBER, then error */
      /*=============================*/

      ELSE_IF((token NEQ FLOAT) AND(token NEQ INTEGER))
         sprintf(MsgBuf, "%s is of wrong type. Expecting a FLOAT or an INTEGER.",TKNWORD);
         ErrorMessage(ERROR,MsgBuf);
         return(RESTART);
      /*==========================*/
      /* Else, add number to list */
      /*==========================*/

      ELSE_IF (token EQ FLOAT)
         CurDeffield->possible_floats =
                          AddToFloatList(CurDeffield->possible_floats,
                                             FLOATNUMBER,YES);
      ELSE_IF (token EQ INTEGER)
         CurDeffield->possible_integers =
                          AddToIntegerList(CurDeffield->possible_integers,
                                             INTNUMBER,YES);
      END_IF
   token = GetToken(CurrentFile);
   END_WHILE

   return(OK);
}


/* ------------------------------------------------------------------
 *  NAME   : ProcessFieldIntegerList()
 *
 *  PURPOSE: This function reads a list of
 *           allowable numbers for a field
 *
 *  INPUTS:  A pointer to the current field definition (DF_PTR *).
 *
 *  RETURNS: nothing.
 *
 *  NOTES:   None.
 * ------------------------------------------------------------------
 */

int ProcessFieldIntegerList(CurDeffield)
DF_PTR CurDeffield;
{
   int token;

   token = GetToken(CurrentFile);
   if(token == RPAREN){
     ErrorMessage(ERROR,"Allowed-integers attributes must have some value(s)\n");
     return(RESTART);}
   /*============================*/
   /* Read list of allowed words */
   /*============================*/

   while(token NEQ RPAREN) DO

      /*===============================*/
      /* If ?VARIABLE used, clear list */
      /*===============================*/

      IF((token EQ BWORD) AND (strcmp(TKNWORD,"VARIABLE") EQ 0)) THEN
         IF(CurDeffield->possible_integers != NULL)THEN
           ErrorMessage(ERROR,"?VARIABLE must be by itself in the allowed-integers attribute\n");
           FreeIntegerList(CurDeffield->possible_integers);
           CurDeffield->possible_integers = NULL;
           return(RESTART);
         ELSE
           token = GetToken(CurrentFile);
           IF(token != RPAREN) THEN
             ErrorMessage(ERROR,"?VARIABLE must be by itself in the allowed-integers attribute\n");
             return(RESTART);
           END_IF
           return(OK);
         END_IF
      ELSE_IF(token NEQ INTEGER)
         sprintf(MsgBuf, "%s is of wrong type. Expecting an INTEGER.",TKNWORD);
         ErrorMessage(ERROR,MsgBuf);
         return(RESTART);
      /*==========================*/
      /* Else, add number to list */
      /*==========================*/
      ELSE
         CurDeffield->possible_integers =
                          AddToIntegerList(CurDeffield->possible_integers,
                                             INTNUMBER,YES);
      END_IF
   token = GetToken(CurrentFile);
   END_WHILE

   return(OK);
}


/* ------------------------------------------------------------------
 *  NAME   : ProcessFieldFloatList()
 *
 *  PURPOSE: This function reads a list of
 *           allowable numbers for a field
 *
 *  INPUTS:  A pointer to the current field definition (DF_PTR *).
 *
 *  RETURNS: nothing.
 *
 *  NOTES:   None.
 * ------------------------------------------------------------------
 */

int ProcessFieldFloatList(CurDeffield)
DF_PTR CurDeffield;
{
   int token;

   token = GetToken(CurrentFile);
   if(token == RPAREN){
     ErrorMessage(ERROR,"Allowed-floats attributes must have some value(s)\n");
     return(RESTART);}
   /*============================*/
   /* Read list of allowed words */
   /*============================*/

   while(token NEQ RPAREN) DO

      /*===============================*/
      /* If ?VARIABLE used, clear list */
      /*===============================*/

      IF((token EQ BWORD) AND (strcmp(TKNWORD,"VARIABLE") EQ 0)) THEN
         IF(CurDeffield->possible_floats != NULL) THEN
           ErrorMessage(ERROR,"?VARIABLE must be by itself in the allowed-floats attribute\n");
           FreeFloatList(CurDeffield->possible_floats);
           CurDeffield->possible_floats = NULL;
           return(RESTART);
         ELSE
           token = GetToken(CurrentFile);
           IF(token != RPAREN) THEN
             ErrorMessage(ERROR,"?VARIABLE must be by itself in the allowed-floats attribute\n");
             return(RESTART);
           END_IF
           return(OK);
         END_IF
      /*=============================*/
      /* If not a NUMBER, then error */
      /*=============================*/

      ELSE_IF(token NEQ FLOAT)
         sprintf(MsgBuf, "%s is of wrong type. Expecting a FLOAT.",TKNWORD);
         ErrorMessage(ERROR,MsgBuf);
         return(RESTART);
      /*==========================*/
      /* Else, add number to list */
      /*==========================*/

      ELSE
         CurDeffield->possible_floats =
                          AddToFloatList(CurDeffield->possible_floats,
                                             FLOATNUMBER,YES);
      END_IF
   token = GetToken(CurrentFile);
   END_WHILE

   return(OK);
}

/* -----------------------------------------------------------------
 * ProcessFieldValues
 *  This funtion processes the allowed-values attribute
 * -----------------------------------------------------------------
*/
ProcessFieldValues(CurDeffield)
DF_PTR CurDeffield;
{
  int token;
  
  token = GetToken(CurrentFile);
   if(token == RPAREN){
     ErrorMessage(ERROR,"Allowed-values attribute must have some value(s)\n");
     return(RESTART);}
   if((CurDeffield->possible_words != NULL) OR
      (CurDeffield->possible_strings != NULL) OR
      (CurDeffield->possible_integers != NULL) OR
      (CurDeffield->possible_floats != NULL) OR
      (CurDeffield->possible_instances != NULL))
     {
      ErrorMessage(ERROR,"Allowed-values attribute cannot be used in conjunction\n");
      SendMessage("      with others value retriction attribute\n",NO);
      return(RESTART);
     }
  WHILE(token != RPAREN OR token != STOP) DO
    switch(token)
     {
       case WORD:
         CurDeffield->possible_words = 
              AddToWordList(CurDeffield->possible_words,TKNWORD,WORD,YES);
       break;
       case STRING:
         CurDeffield->possible_strings =
                AddToWordList(CurDeffield->possible_strings,TKNWORD,STRING,YES);
       break;
       case INTEGER:
         CurDeffield->possible_integers = 
                AddToIntegerList(CurDeffield->possible_integers,INTNUMBER,YES);
       break;
       case FLOAT:
         CurDeffield->possible_floats =
                AddToFloatList(CurDeffield->possible_floats,FLOATNUMBER,YES);
       break;
       case INSTANCE:
         CurDeffield->possible_instances = 
                AddToWordList(CurDeffield->possible_instances,TKNWORD,INSTANCE,YES);
       break;
       default: 
         ErrorMessage(ERROR,"Unrecognized token in the values list\n");
         return(RESTART);
       
     }
  
    token = GetToken(CurrentFile);
  END_WHILE
  
  return(OK);
}
/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : ProcessFieldRange()
 *
 *  PURPOSE: This function reads the min and max range
 *           values a number can have.
 *
 *  INPUTS:  A pointer to the current field definition (DF_PTR *).
 *
 *  RETURNS: Nothing.
 *
 *  NOTES:   None.
 * ------------------------------------------------------------------
 */

int ProcessFieldRange(CurDeffield)
DF_PTR CurDeffield;
{
   int token;

  

   token = GetToken(CurrentFile);

   /*===========================================*/
   /* Process MIN, If ?VARIABLE used, clear min */
   /*===========================================*/

   IF((token EQ BWORD) AND (strcmp(TKNWORD,"VARIABLE") EQ 0)) THEN
      CurDeffield->set_int_min = NO;
      CurDeffield->set_float_min = NO;
   /*=========================*/
   /* If a RPAREN, then error */
   /*=========================*/

   ELSE_IF(token EQ RPAREN)
      ErrorMessage(ERROR, "Expecting a minimum value to define range.");
      return(RESTART);

   /*=============================*/
   /* If not a NUMBER, then error */
   /*=============================*/

   ELSE_IF((token NEQ INTEGER) AND (token NEQ FLOAT))
      ErrorMessage(ERROR, "Expecting a number for min value to define range.");
      return(RESTART);
   ELSE_IF (token EQ INTEGER)
      CurDeffield->set_int_min   = YES;
      CurDeffield->int_min_range = INTNUMBER;
   ELSE_IF (token EQ FLOAT)
      CurDeffield->set_float_min   = YES;
      CurDeffield->float_min_range = FLOATNUMBER;
   END_IF

   token = GetToken(CurrentFile);

   /*===========================================*/
   /* Process MAX, If ?VARIABLE used, clear max */
   /*===========================================*/

   IF((token EQ BWORD) AND (strcmp(TKNWORD,"VARIABLE") EQ 0)) THEN
      CurDeffield->set_int_max = NO;
      CurDeffield->set_float_max = NO;
   /*=========================*/
   /* If a RPAREN, then error */
   /*=========================*/

   ELSE_IF(token EQ RPAREN)
      ErrorMessage(ERROR, "Expecting maximum value to define range.");
      return(RESTART);

   /*=============================*/
   /* If not a NUMBER, then error */
   /*=============================*/

   ELSE_IF((token NEQ INTEGER) AND (token NEQ FLOAT))
      ErrorMessage(ERROR, "Expecting a number for max value to define range.");
      return(RESTART);
   ELSE_IF (token EQ INTEGER)
      CurDeffield->set_int_max   = YES;
      CurDeffield->int_max_range = INTNUMBER;
   ELSE_IF (token EQ FLOAT)
      CurDeffield->set_float_max   = YES;
      CurDeffield->float_max_range = FLOATNUMBER;
   END_IF
   if(GetToken(CurrentFile) !=  RPAREN)
    {
      ErrorMessage(ERROR,"Expect a ')' after maximun value in range attribute\n");
      return(RESTART);
    }
   return(OK);
}

/* ------------------------------------------------------------------
 *  NAME   : CopyFieldInfo()
 *
 *  PURPOSE: This function copies the field info from the current
 *           field into a new def_field structure and returns a
 *           pointer to that structure.
 *
 *  INPUTS:  Two arguments,
 *           A pointer to the relation name, (char *)
 *           the position in the pattern, (int)
 *
 *  RETURNS: A pointer to the new structure (DEF_FLD *)
 *
 *  NOTES:   Does not check field number here!
 * ------------------------------------------------------------------
 */

DEF_FLD *CopyFieldInfo(RelName,FieldPosition)
   char *RelName;
   int   FieldPosition;
{
   DEF_FLD *NewDeffield,*CurField;
   DEF_REL *rel;
   WL_PTR   wl_ptr;
   IL_PTR   il_ptr;
   FL_PTR   fl_ptr;

   /*===========================*/
   /* Find defrelation (if any) */
   /*===========================*/

   rel = FindDefrelation(DefrelationHead, RelName);

   /*=======================================*/
   /* No defrelation, flag error and return */
   /*=======================================*/

   IF(rel EQ NULL) THEN
      DefrelationHead = FindCreateDefrelation(DefrelationHead, RelName);
      IF (strcmp(RelName, "initial-fact") NEQ 0) THEN
         sprintf(MsgBuf, "Relation <%.40s> used prior to its definition.",
            RelName);
         ErrorMessage(WARNING, MsgBuf);
      END_IF
      return(NULL);
   END_IF

   CurField = FindDeffield(FieldPosition, rel->fields);
   IF(CurField EQ NULL) THEN

      /*================================*/
      /* This field not defined, return */
      /*================================*/

      return(NULL);
   END_IF

   NewDeffield = AllocDeffield();

   /*================*/
   /* Copy type info */
   /*================*/

   NewDeffield->allow_word   = CurField->allow_word;
   NewDeffield->allow_string = CurField->allow_string;
   NewDeffield->allow_integer = CurField->allow_integer;
   NewDeffield->allow_float = CurField->allow_float;
   NewDeffield->allow_address = CurField->allow_address;
   NewDeffield->allow_instance = CurField->allow_instance;
  
   /*================*/
   /* Copy word list */
   /*================*/

   for (wl_ptr = CurField->possible_words; wl_ptr NEQ NULL;
         wl_ptr = wl_ptr->next_word) DO
      NewDeffield->possible_words =
         AddToWordList(NewDeffield->possible_words, wl_ptr->word, WORD,NO);
   END_FOR

   /*==================*/
   /* Copy string list */
   /*==================*/

   for (wl_ptr = CurField->possible_strings; wl_ptr NEQ NULL;
         wl_ptr = wl_ptr->next_word) DO
      NewDeffield->possible_strings =
         AddToWordList(NewDeffield->possible_strings, wl_ptr->word, STRING,
         NO);
   END_FOR

   /*==================*/
   /* Copy number list */
   /*==================*/

   for (il_ptr = CurField->possible_integers; il_ptr NEQ NULL;
         il_ptr = il_ptr->next_number) DO
      NewDeffield->possible_integers =
         AddToIntegerList(NewDeffield->possible_integers, il_ptr->number, NO);
   END_FOR

   for (fl_ptr = CurField->possible_floats; fl_ptr NEQ NULL;
         fl_ptr = fl_ptr->next_number) DO
      NewDeffield->possible_floats =
         AddToFloatList(NewDeffield->possible_floats, fl_ptr->number, NO);
   END_FOR
   /*=================*/
   /* Copy range info */
   /*=================*/

   NewDeffield->set_int_max     = CurField->set_int_max;
   NewDeffield->set_int_min     = CurField->set_int_min;
   NewDeffield->set_float_max   = CurField->set_float_max;
   NewDeffield->set_float_min   = CurField->set_float_min;
   NewDeffield->float_max_range = CurField->float_max_range;
   NewDeffield->float_min_range = CurField->float_min_range;
   NewDeffield->int_max_range   = CurField->int_max_range;
   NewDeffield->int_min_range   = CurField->int_min_range;
   return(NewDeffield);
}


/* ======================================================================= */
/*                  FUNCTIONS TO CHECK DEFRELATION INFO                    */
/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : CheckFieldInfo()
 *
 *  PURPOSE: This function compares info from relations
 *           used in a rule or deffact to the defrelation
 *           paramaters and prints appropriate error
 *           messages.
 *
 *  INPUTS:  Four arguments,
 *           A pointer to the relation name, (char *)
 *           the type of the value, WORD, STRING, NUMBER,(int)
 *           the position in the pattern, (int)
 *           and the literal value of the field (char *).
 *
 *  RETURNS: Nothing.
 *
 *  NOTES:   Does not check field number here!
 * ------------------------------------------------------------------
 */

void CheckFieldInfo(name, type, position, value)
char *name, *value;
int   type,  position;
{
   DEF_REL  *def_rel;
   DEF_FLD  *fld;
   double    float_number;
   long		 int_number;
   char     *def_form;

   /*======================================================*/
   /*  if value is CRSV_NIL, which is made up CRSV exit    */
   /*======================================================*/
   
   if(strcmp("CRSV_NIL",value) == 0)
     return;
   
   /*===========================*/
   /* Find defrelation (if any) */
   /*===========================*/

   def_rel = FindDefrelation(DefrelationHead, name);

   /*=======================================*/
   /* No defrelation, flag error and return */
   /*=======================================*/

   IF(def_rel EQ NULL) THEN
      DefrelationHead = FindCreateDefrelation(DefrelationHead, name);
      IF (strcmp(name, "initial-fact") NEQ 0) THEN
         sprintf(MsgBuf, "Relation <%.40s> used prior to its definition.",
            name);
         ErrorMessage(WARNING, MsgBuf);
      END_IF
      return;
   END_IF

   def_form = def_rel->created_by NEQ TEMPLATE_MADE ? "defrelation" : 
      "deftemplate";

   /*==================*/
   /* Check field info */
   /*==================*/

   fld = FindDeffield(position, def_rel->fields);
   IF(fld EQ NULL) THEN

      /*================================*/
      /* This field not defined, return */
      /*================================*/

      return;
   ELSE

      /*=====================================*/
      /* Check field type and literal values */
      /*=====================================*/

      switch (type) {

         case WORD:
              /*=====================*/
              /* Check WORDS allowed */
              /*=====================*/

              IF(fld->allow_word NEQ YES) THEN
                 sprintf(MsgBuf,
                    "Used a SYMBOL in field %d, which isn't allowed by %s %s.",
                    position, def_form,name);
                 ErrorMessage(ERROR, MsgBuf);


              /*=================*/
              /* Check WORD list */
              /*=================*/

              ELSE_IF (fld->possible_words EQ NULL OR
                 IsWordInList(fld->possible_words, value) IS_NO)

                 /* ------------------------------------------------
                  *  If LITERAL is on, the word will be added to the
                  *  ALLOWED_WORDS list.  If VERBOSE is on, a message
                  *  will be printed which will tell the user that
                  *  this is being done.
                  * ------------------------------------------------
                  */

                 IF (LITERAL IS_ON) THEN

                    fld->possible_words = AddToWordList(fld->possible_words,
                         value, WORD, YES);

                    IF (VERBOSE IS_ON) THEN
                      sprintf(MsgBuf,
                        "Adding <%.40s> to ALLOWED-WORDS list of %s field %d.",
                          value, def_rel->name, fld->position);
                       ErrorMessage(WARNING, MsgBuf);
                    END_IF

                 /* ----------------------------------------------------------
                  *  ELSE_IF The permitted list was not NULL (i.e. ?VARIABLE)
                  *  AND check style messages are turned on
                  *  Generate a style warning message
                  * ----------------------------------------------------------
                  */

                   ELSE_IF (fld->possible_words NEQ NULL)
                    sprintf(MsgBuf,
                     "Used a SYMBOL <%.40s> in field %d of %s %s",value,position,
                        name,"that is not in the ALLOWED-WORDS list.");
                    ErrorMessage(WARNING, MsgBuf);
                 END_IF

              END_IF
              break;

         case STRING:
              /*=======================*/
              /* Check STRINGS allowed */
              /*=======================*/

              IF(fld->allow_string NEQ YES) THEN
                 sprintf(MsgBuf,
                    "Used a STRING in field %d, which isn't allowed by %s %s.",
                         position, def_form,name);
                 ErrorMessage(ERROR,MsgBuf);

              /*====================*/
              /* Check STRINGS list */
              /*====================*/

              ELSE_IF(fld->possible_strings EQ NULL OR
                 IsWordInList(fld->possible_strings, value) IS_NO)

                 /* ---------------------------------------
                  *  Add string to WORD list if LITERAL
                  *  flag turned on.
                  * ---------------------------------------
                  */

                 IF (LITERAL IS_ON) THEN

                    fld->possible_strings = 
                       AddToWordList(fld->possible_strings,value,STRING,NO);

                    IF (VERBOSE IS_ON) THEN
                      sprintf(MsgBuf,
                      "Adding <%.40s> to ALLOWED-STRINGS list of %s field %d.",
                         value, def_rel->name, fld->position);
                        ErrorMessage(WARNING,MsgBuf);
                    END_IF

                 /* ------------------------------
                  *  ELSE_IF The permitted list was not NULL (i.e. ?VARIABLE)
                  *  AND check style messages are turned on
                  *  Generate a style warning message
                  * ------------------------------
                  */

                 ELSE_IF (fld->possible_strings NEQ NULL)
                    sprintf(MsgBuf,
                       "Used a STRING (%.40s) in field %d that is not in the ALLOWED_STRINGS list.",
                       value, position);
                    ErrorMessage(ERROR, MsgBuf);
                 END_IF

              END_IF
              break;
		 case INSTANCE:
              /*=======================*/
              /* Check STRINGS allowed */
              /*=======================*/

              IF(fld->allow_instance NEQ YES) THEN
                 sprintf(MsgBuf,
                    "Used an INSTANCE (%s) in field %d, which isn't allowed by %s %s",
                         value,position, def_form,name);
                 ErrorMessage(ERROR,MsgBuf);

              /*====================*/
              /* Check STRINGS list */
              /*====================*/

              ELSE_IF(fld->possible_instances EQ NULL OR
                 IsWordInList(fld->possible_instances, value) IS_NO)

                 /* ---------------------------------------
                  *  Add string to WORD list if LITERAL
                  *  flag turned on.
                  * ---------------------------------------
                  */

                 IF (LITERAL IS_ON) THEN

                    fld->possible_instances = 
                       AddToWordList(fld->possible_instances,value,STRING,NO);

                    IF (VERBOSE IS_ON) THEN
                      sprintf(MsgBuf,
                      "Adding <%.40s> to ALLOWED-INSTANCES list of %s field %d.",
                         value, def_rel->name, fld->position);
                        ErrorMessage(WARNING,MsgBuf);
                    END_IF

                 /* --------------------------------------------------------
                  *  ELSE_IF The permitted list was not NULL (i.e. ?VARIABLE)
                  *  AND check style messages are turned on
                  *  Generate a style warning message
                  * --------------------------------------------------------
                  */

                 ELSE_IF (fld->possible_instances NEQ NULL)
                    sprintf(MsgBuf,
                       "Used an INSTANCE (%s) in field %d that is not in the ALLOWED_INSATNCES list.",
                       value, position);
                    ErrorMessage(ERROR, MsgBuf);
                 END_IF

              END_IF
              break;
         case INTEGER:
              /*=======================*/
              /* Check NUMBERS allowed */
              /*=======================*/

              IF(fld->allow_integer NEQ YES) THEN
                 sprintf(MsgBuf,
                   "Used an INTEGER (%s) in field %d that is not allowed by %s %s.",
                    value,position, def_form,name);
                 ErrorMessage(ERROR,MsgBuf);
              ELSE
                 /*====================*/
                 /* Check NUMBERS list */
                 /*====================*/

                 int_number = atol(value);
                 IF (fld->possible_integers EQ NULL OR 
                    IsIntegerInList(fld->possible_integers, int_number) IS_NO)
                    THEN

                    /* ---------------------------------------
                     *  Add number to NUMBERS list if LITERAL
                     *  flag turned on.
                     * ---------------------------------------
                     */

                    IF (LITERAL IS_ON) THEN

                      fld->possible_integers =
                           AddToIntegerList(fld->possible_integers,int_number,NO);

                      IF (VERBOSE IS_ON) THEN
                         sprintf(MsgBuf,
                          "Adding <%.40s> to the ALLOWED-INTEGERS list of %s field %d.",
                              value, def_rel->name, fld->position);
                         ErrorMessage(WARNING,MsgBuf);
                      END_IF
                    /* ------------------------------
                     *  ELSE_IF The permitted list was not NULL(i.e. ?VARIABLE)
                     *  AND check style messages are turned on
                     *  Generate a style warning message
                     * ------------------------------
                     */

                    ELSE_IF (fld->possible_integers NEQ NULL)
                       sprintf(MsgBuf,
                          "Used an INTEGER (%s) in field %d that is not in the ALLOWED_INTEGERS list.",
                          value,position);
                       ErrorMessage(ERROR, MsgBuf);
                    END_IF
                 END_IF
                 /*=====================*/
                 /* Check INTEGERS range */
                 /*=====================*/

                 IF(fld->set_int_max EQ YES) THEN
                    IF(int_number > fld->int_max_range) THEN
                       sprintf(MsgBuf,
                          "Used a value (%d) in field %d greater than range max (%d).",
                          int_number, position, fld->int_max_range);
                       ErrorMessage(ERROR, MsgBuf);
                    END_IF
                 END_IF

                 IF(fld->set_int_min EQ YES) THEN
                    IF(int_number < fld->int_min_range) THEN
                       sprintf(MsgBuf,
                          "Used a value (%d) in field %d smaller than range min (%d).",
                          int_number, position, fld->int_min_range);
                       ErrorMessage(ERROR, MsgBuf);
                    END_IF
                 END_IF
              END_IF

              break;
	     case FLOAT:
              /*=======================*/
              /* Check FLOATS allowed */
              /*=======================*/

              IF(fld->allow_float NEQ YES) THEN
                 sprintf(MsgBuf,
                    "Used a FLOAT (%s) in field %d that is not allowed by %s %s.",
                    value,position, def_form,name);
                 ErrorMessage(ERROR,MsgBuf);
              ELSE
                 /*====================*/
                 /* Check FLOATS list */
                 /*====================*/

                 float_number = atof(value);
                 IF (fld->possible_floats EQ NULL OR 
                    IsFloatInList(fld->possible_floats, float_number) IS_NO)
                    THEN

                    /* ---------------------------------------
                     *  Add a FLOAT to FLOATS list if LITERAL
                     *  flag turned on.
                     * ---------------------------------------
                     */

                    IF (LITERAL IS_ON) THEN

                      fld->possible_floats =
                           AddToFloatList(fld->possible_floats,float_number,NO);

                      IF (VERBOSE IS_ON) THEN
                         sprintf(MsgBuf,
                          "Adding <%.40s> to the ALLOWED-FLOATS list of %s field %d.",
                              value, def_rel->name, fld->position);
                         ErrorMessage(WARNING,MsgBuf);
                      END_IF
    
                    /* ------------------------------
                     *  ELSE_IF The permitted list was not NULL(i.e. ?VARIABLE)
                     *  AND check style messages are turned on
                     *  Generate a style warning message
                     * ------------------------------
                     */

                    ELSE_IF (fld->possible_floats NEQ NULL)
                       sprintf(MsgBuf,
                          "Used a FLOAT (%s) in field %d that is not in the ALLOWED_FLOATS list.",
                          value,position);
                       ErrorMessage(ERROR, MsgBuf);
                    END_IF
                 END_IF

                 /*=====================*/
                 /* Check FLOATS range */
                 /*=====================*/

                 IF(fld->set_float_max EQ YES) THEN
                    IF(float_number > fld->float_max_range) THEN
                       sprintf(MsgBuf,
                          "Used a value (%f) in field %d greater than range max (%f).",
                          float_number, position, fld->float_max_range);
                       ErrorMessage(ERROR, MsgBuf);
                    END_IF
                 END_IF

                 IF(fld->set_float_min EQ YES) THEN
                    IF(float_number < fld->float_min_range) THEN
                       sprintf(MsgBuf,
                          "Used a value (%f) in field %d smaller than range min (%f).",
                          float_number, position, fld->float_min_range);
                       ErrorMessage(ERROR, MsgBuf);
                    END_IF
                 END_IF
              END_IF

              break;
         }
   END_IF
}

/* ------------------------------------------------------------------
 *  NAME   : CheckVarFieldInfo()
 *
 *  PURPOSE: This function checks to see that a variable is used
 *           in a field whose defrelation info is consistent with
 *           the info from fields in which the variable was used
 *           previously.
 *
 *  INPUTS:  Four arguments,
 *           A pointer to the relation name, (char *)
 *           a pointer to the variable name, (char *),
 *           a pointer to the structure containing the cumulative
 *              field info for that variable (DEF_FLD *),
 *           and the position in the pattern, (int).
 *
 *  RETURNS: Nothing.
 *
 *  NOTES:   Does not check field number here!
 * ------------------------------------------------------------------
 */

void CheckVarFieldInfo(rule_side,RelName, var_name, var_fld, position)
   char    *RelName, *var_name;
   DEF_FLD *var_fld;
   int      rule_side, position;
{
   DEF_REL  *def_rel;
   DEF_FLD  *fld;
   int       found_common;
   int       num_in_common,tot_num_in_common = 0;

   /*===========================*/
   /* Find defrelation (if any) */
   /*===========================*/

   def_rel = FindDefrelation(DefrelationHead, RelName);

   /*=======================================*/
   /* No defrelation, flag error and return */
   /*=======================================*/

   IF(def_rel EQ NULL) THEN
      DefrelationHead = FindCreateDefrelation(DefrelationHead, RelName);
      IF (strcmp(RelName, "initial-fact") NEQ 0) THEN
         sprintf(MsgBuf, "Relation <%.40s> used prior to its definition.",
            RelName);
         ErrorMessage(WARNING, MsgBuf);
      END_IF
      return;
   END_IF

   /*==================*/
   /* Check field info */
   /*==================*/

   fld = FindDeffield(position, def_rel->fields);
   IF(fld EQ NULL) THEN

      /*================================*/
      /* This field not defined, return */
      /*================================*/

      return;
   END_IF

   /*============================*/
   /* Compare data types allowed */
   /*============================*/

   found_common = CompareFieldTypes(var_fld,fld);
   IF(found_common IS_NO) THEN
     IF(rule_side == LHS) THEN
      sprintf(MsgBuf,
         "Due to the inconsistency of the field definitions, this rule MAY NOT fire. Variable <%.40s> has incompatible type definitions.",
         var_name);
      ErrorMessage(WARNING, MsgBuf);
     ELSE
      sprintf(MsgBuf,"variable <%.40s> has incompatible type definitions.",
              var_name);
      ErrorMessage(ERROR,MsgBuf);
     END_IF

   /*===========================================================*/
   /* If data types are not exclusive, look at lists and ranges */
   /*===========================================================*/
    
    ELSE_IF(var_fld->allow_address != TRUE)

      /*====================*/
      /* Compare word lists */
      /*====================*/

          num_in_common = CompareFieldWordLists(var_name,var_fld,fld);

       /* ------------------------------------------------------
        *  If there are no words in common and LITERAL is Off,
        *  then print out an error message, because if LITERAL
        *  is on, we are adding values to the allowed lists
        *  as the file is parsed and there is no need to display
        *  error messages that will be cleared up by further
        *  parsing.
        * -------------------------------------------------------
        */

       IF ((num_in_common == 0) AND (LITERAL IS_OFF)) THEN
          sprintf(MsgBuf,
          "Variable <%.40s> used in fields whose ALLOWED_WORDS definitions have no common values.", 
              var_name);
          ErrorMessage(ERROR, MsgBuf);
       ELSE_IF (num_in_common > 0 )
          tot_num_in_common += num_in_common;
       END_IF

      /*======================*/
      /* Compare string lists */
      /*======================*/

      num_in_common = CompareFieldStringLists(var_name,var_fld,fld);
      IF ((num_in_common == 0) AND (LITERAL IS_OFF)) THEN
          sprintf(MsgBuf,
          "Variable <%.40s> used in fields whose ALLOWED_STRINGS definitions have no common values.", 
              var_name);
          ErrorMessage(ERROR, MsgBuf);
      ELSE_IF (num_in_common > 0)
         tot_num_in_common += num_in_common;
      END_IF

      /*========================*/
      /* Compare instance lists */
      /*========================*/
      
      num_in_common = CompareFieldInstanceLists(var_name,var_fld,fld);
      IF ((num_in_common == 0) AND (LITERAL IS_OFF)) THEN
          sprintf(MsgBuf,
          "Variable <%.40s> used in fields whose ALLOWED_INSTANCES definitions have no common values.", 
              var_name);
          ErrorMessage(ERROR, MsgBuf);
      ELSE_IF (num_in_common > 0)
         tot_num_in_common += num_in_common;
      END_IF
      /*=================*/
      /* Compare numbers */
      /*=================*/

     num_in_common = CompareFieldNumbers(var_name,var_fld,fld);

     if (num_in_common > 0)
       tot_num_in_common += num_in_common;
      /*=====================================================*/
      /* If no common values found in ANY of the categories, */
      /*    print error message                              */
      /*=====================================================*/

      IF(tot_num_in_common EQ 0) THEN
       IF(rule_side == LHS) THEN
         sprintf(MsgBuf,
            "Due to the inconsistency of the field definitions, this rule MAY NOT fire.  Variable <%.40s> has no common allowed values.",
            var_name);
         ErrorMessage(WARNING, MsgBuf);
       END_IF

      /*============================================*/
      /* Otherwise, if only one common value found, */
      /*    print warning message                   */
      /*============================================*/

      ELSE_IF(tot_num_in_common EQ 1)
         sprintf(MsgBuf, "Field definitions allow variable <%.40s> only one possible value.",
                 var_name);
         ErrorMessage(WARNING,MsgBuf);
      END_IF
   END_IF
}



/* ------------------------------------------------------------------
 *  NAME   : CompareFieldTypes()
 *
 *  PURPOSE: This function checks the types allowed in the current
 *           field against those used in previous fields.
 *
 *  INPUTS:  Two arguments,
 *           A pointer to the cumulative field info (DEF_FLD *),
 *           and a pointer to the current field info (DEF_FLD *).
 *
 *  RETURNS: YES if found a common type,
 *           NO  if didn't.
 * ------------------------------------------------------------------
 */

int CompareFieldTypes(var_fld,CurField)
   DEF_FLD *var_fld,*CurField;
{
   int found_common = NO;

   /*===========================*/
   /* Check if word types match */
   /*===========================*/

   IF(var_fld->allow_word IS_YES) THEN
      IF(CurField->allow_word IS_YES) THEN
         found_common = YES;
      ELSE
         var_fld->allow_word     = NO;
         FreeWordList(var_fld->possible_words);
         var_fld->possible_words = NULL;
      END_IF
   END_IF

   /*=============================*/
   /* Check if string types match */
   /*=============================*/

   IF(var_fld->allow_string IS_YES) THEN
      IF(CurField->allow_string IS_YES) THEN
         found_common = YES;
      ELSE
         var_fld->allow_string     = NO;
         FreeWordList(var_fld->possible_strings);
         var_fld->possible_strings = NULL;
      END_IF
   END_IF

   /*=============================*/
   /* Check if number types match */
   /*=============================*/

   IF(var_fld->allow_integer IS_YES) THEN
      IF(CurField->allow_integer IS_YES) THEN
         found_common = YES;
      ELSE
         var_fld->allow_integer     = NO;
         FreeIntegerList(var_fld->possible_integers);
         var_fld->possible_integers = NULL;
      END_IF
   END_IF
   IF(var_fld->allow_float IS_YES) THEN
      IF(CurField->allow_float IS_YES) THEN
         found_common = YES;
      ELSE
         var_fld->allow_float     = NO;
         FreeFloatList(var_fld->possible_floats);
         var_fld->possible_floats = NULL;
      END_IF
   END_IF
   IF(var_fld->allow_address IS_YES) THEN
     IF(CurField->allow_address IS_YES) THEN
         found_common = YES;
     ELSE
       var_fld->allow_address = NO;
     END_IF
   END_IF
   IF((var_fld->allow_float == NO) && (var_fld->allow_integer == NO))THEN
         var_fld->set_int_min          = NO;
         var_fld->set_int_max          = NO;
         var_fld->set_float_min        = NO;
         var_fld->set_float_max        = NO;
         var_fld->int_min_range        = -1000;
         var_fld->int_max_range        = 1000;
         var_fld->float_min_range      = -1000.0;
         var_fld->float_max_range      = 1000.0;
   END_IF
   return(found_common);
}

/* ------------------------------------------------------------------
 *  NAME   : CompareFieldWordLists()
 *
 *  PURPOSE: This function compares the list of allowed words in the
 *           current field against the cumulative word list from
 *           previous fields.
 *
 *  INPUTS:  Three arguments,
 *           The variable's name (char *),
 *           a pointer to the cumulative field info (DEF_FLD *),
 *           and a pointer to the current field info (DEF_FLD *).
 *
 *  RETURNS: 0 if no common values,
 *           1000 (no count) if either word list is NULL,
 *           or the number of common words if neither is NULL.
 * ------------------------------------------------------------------
 */

int CompareFieldWordLists(var_name,var_fld,CurField)
   char    *var_name;
   DEF_FLD *var_fld,*CurField;
{
   WL_PTR var_wl, cur_wl, old_word = NULL;
   int    num_in_common, found_word;

   /*=========================================*/
   /* If no words allowed, return -1, no error */
   /*=========================================*/

   IF(var_fld->allow_word IS_NO) THEN
      return(-1);
   END_IF

   var_wl = var_fld->possible_words;
   cur_wl = CurField->possible_words;

   /*==================================================================*/
   /* If no word list specified by current field, no need to calculate */
   /*    overlap                                                       */
   /*==================================================================*/

   IF(cur_wl EQ NULL) THEN
      return(1000);
   END_IF

   /*=========================================================*/
   /* Otherwise, if no word list specified by previous fields */
   /*=========================================================*/

   IF(var_wl EQ NULL) THEN

      /*===============================================================*/
      /* Copy current field's word list, no need to count common words */
      /*===============================================================*/

      while(cur_wl NEQ NULL) DO
         var_wl = AddToWordList(var_wl, cur_wl->word, WORD,NO);
         cur_wl = cur_wl->next_word;
      END_WHILE
      var_fld->possible_words = var_wl;
      return(1000);
   END_IF

   /*==============================*/
   /* If neither word list is NULL */
   /*==============================*/

   num_in_common = 0;

   while (var_wl NEQ NULL) DO
      found_word = NO;
      cur_wl     = CurField->possible_words;
      while (cur_wl NEQ NULL) DO

         /*===================*/
         /* Compare each word */
         /*===================*/

         IF (strcmp(cur_wl->word,var_wl->word) EQ 0) THEN
            found_word = YES;
            break;
         END_IF
         cur_wl = cur_wl->next_word;
      END_WHILE

      /*==================================================================*/
      /* If a word was in a previous field, but not in the current field, */
      /*    delete it from the variable's cumulative word list            */
      /*==================================================================*/

      IF(found_word IS_NO) THEN
         var_wl = FreeWordAndGetNext(&var_fld->possible_words,
                                          var_wl,old_word);

      /*======================================*/
      /* Otherwise, count it as a common word */
      /*======================================*/

      ELSE
         ++num_in_common;
         old_word = var_wl;
         var_wl   = var_wl->next_word;
      END_IF
   END_WHILE

   return(num_in_common);
}

/* ------------------------------------------------------------------
 *  NAME   : CompareFieldStringLists()
 *
 *  PURPOSE: This function compares the list of allowed strings in
 *           the current field against the cumulative string list
 *           from previous fields.
 *
 *  INPUTS:  Three arguments,
 *           The variable's name (char *),
 *           a pointer to the cumulative field info (DEF_FLD *),
 *           and a pointer to the current field info (DEF_FLD *).
 *
 *  RETURNS: 0 if no common values,
 *           1000 (no count) if either string list is NULL,
 *           or the number of common strings if neither is NULL.
 * ------------------------------------------------------------------
 */

int CompareFieldStringLists(var_name,var_fld,CurField)
   char    *var_name;
   DEF_FLD *var_fld,*CurField;
{
   WL_PTR var_wl, cur_wl, old_string = NULL;
   int    num_in_common, found_string;

   /*============================================*/
   /* If no strings allowed, return -1, no error */
   /*============================================*/

   IF(var_fld->allow_string IS_NO) THEN
      return(-1);
   END_IF

   var_wl = var_fld->possible_strings;
   cur_wl = CurField->possible_strings;

   /*====================================================================*/
   /* If no string list specified by current field, no need to calculate */
   /*    overlap                                                         */
   /*====================================================================*/

   IF(cur_wl EQ NULL) THEN
      return(1000);
   END_IF

   /*===========================================================*/
   /* Otherwise, if no string list specified by previous fields */
   /*===========================================================*/

   IF(var_wl EQ NULL) THEN

      /*===================================================================*/
      /* Copy current field's string list, no need to count common strings */
      /*===================================================================*/

      while(cur_wl NEQ NULL) DO
         var_wl = AddToWordList(var_wl, cur_wl->word, STRING,NO);
         cur_wl = cur_wl->next_word;
      END_WHILE

      var_fld->possible_strings = var_wl;
      return(1000);
   END_IF

   /*================================*/
   /* If neither string list is NULL */
   /*================================*/

   num_in_common = 0;

   while (var_wl NEQ NULL) DO
      found_string = NO;
      cur_wl       = CurField->possible_strings;
      while (cur_wl NEQ NULL) DO

         /*=====================*/
         /* Compare each string */
         /*=====================*/

         IF (strcmp(cur_wl->word,var_wl->word) EQ 0) THEN
            found_string = YES;
            break;
         END_IF
         cur_wl = cur_wl->next_word;
      END_WHILE

      /*====================================================================*/
      /* If a string was in a previous field, but not in the current field, */
      /*    delete it from the variable's cumulative string list            */
      /*====================================================================*/

      IF(found_string IS_NO) THEN
         var_wl = FreeWordAndGetNext(&var_fld->possible_strings,
                                          var_wl,old_string);

      /*========================================*/
      /* Otherwise, count it as a common string */
      /*========================================*/

      ELSE
         ++num_in_common;
         old_string = var_wl;
         var_wl     = var_wl->next_word;
      END_IF
   END_WHILE

   return(num_in_common);
}


/* ------------------------------------------------------------------
 *  NAME   : CompareFieldInstanceLists()
 *
 *  PURPOSE: This function compares the list of allowed instances in
 *           the current field against the cumulative instance list
 *           from previous fields.
 *
 *  INPUTS:  Three arguments,
 *           The variable's name (char *),
 *           a pointer to the cumulative field info (DEF_FLD *),
 *           and a pointer to the current field info (DEF_FLD *).
 *
 *  RETURNS: 0 if no common values,
 *           1000 (no count) if either instance list is NULL,
 *           or the number of common instances if neither is NULL.
 * ------------------------------------------------------------------
 */

int CompareFieldInstanceLists(var_name,var_fld,CurField)
   char    *var_name;
   DEF_FLD *var_fld,*CurField;
{
   WL_PTR var_wl, cur_wl, old_instance = NULL;
   int    num_in_common, found_instance;

   /*============================================*/
   /* If no instances allowed, return -1, no error */
   /*============================================*/

   IF(var_fld->allow_instance IS_NO) THEN
      return(-1);
   END_IF

   var_wl = var_fld->possible_instances;
   cur_wl = CurField->possible_instances;

   /*====================================================================*/
   /* If no instance list specified by current field, no need to calculate */
   /*    overlap                                                         */
   /*====================================================================*/

   IF(cur_wl EQ NULL) THEN
      return(1000);
   END_IF

   /*===========================================================*/
   /* Otherwise, if no instance list specified by previous fields */
   /*===========================================================*/

   IF(var_wl EQ NULL) THEN

      /*===================================================================*/
      /* Copy current field's instance list, no need to count common instances */
      /*===================================================================*/

      while(cur_wl NEQ NULL) DO
         var_wl = AddToWordList(var_wl, cur_wl->word, STRING,NO);
         cur_wl = cur_wl->next_word;
      END_WHILE

      var_fld->possible_instances = var_wl;
      return(1000);
   END_IF

   /*================================*/
   /* If neither instance list is NULL */
   /*================================*/

   num_in_common = 0;

   while (var_wl NEQ NULL) DO
      found_instance = NO;
      cur_wl       = CurField->possible_instances;
      while (cur_wl NEQ NULL) DO

         /*=====================*/
         /* Compare each instance */
         /*=====================*/

         IF (strcmp(cur_wl->word,var_wl->word) EQ 0) THEN
            found_instance = YES;
            break;
         END_IF
         cur_wl = cur_wl->next_word;
      END_WHILE

      /*====================================================================*/
      /* If a instance was in a previous field, but not in the current field, */
      /*    delete it from the variable's cumulative instance list            */
      /*====================================================================*/

      IF(found_instance IS_NO) THEN
         var_wl = FreeWordAndGetNext(&var_fld->possible_instances,
                                          var_wl,old_instance);

      /*========================================*/
      /* Otherwise, count it as a common instance */
      /*========================================*/

      ELSE
         ++num_in_common;
         old_instance = var_wl;
         var_wl     = var_wl->next_word;
      END_IF
   END_WHILE

   return(num_in_common);
}

/* ------------------------------------------------------------------
 *  NAME   : CompareFieldNumbers()
 *
 *  PURPOSE: This function compares the numeric information from
 *           the current field against the cumulative numeric info
 *           from previous fields.
 *
 *  INPUTS:  Three arguments,
 *           The variable's name (char *),
 *           a pointer to the cumulative field info (DEF_FLD *),
 *           and a pointer to the current field info (DEF_FLD *).
 *
 *  RETURNS: 0 if no common values,
 *           1000 (no count) if counting was inappropriate,
 *           or the number of common numeric values.
 * ------------------------------------------------------------------
 */

int CompareFieldNumbers(var_name,var_fld,CurField)
   char    *var_name;
   DEF_FLD *var_fld,*CurField;
{
   int  num_in_common = 0;

   /*==============================================*/
   /* If no numbers allowed, return -1, no error   */
   /*==============================================*/

   IF((var_fld->allow_integer IS_NO) && (var_fld->allow_float IS_NO))THEN
      return(-1);
   END_IF

   /*================================================*/
   /* If no range set, compare only the number lists */
   /*================================================*/

   IF((var_fld->set_int_min IS_NO) AND (var_fld->set_int_max IS_NO) AND
      (CurField->set_int_min IS_NO) AND (CurField->set_int_max IS_NO) AND
      (var_fld->set_float_min IS_NO) AND (var_fld->set_float_max IS_NO) AND
      (CurField->set_float_min IS_NO) AND (CurField->set_float_max IS_NO)) THEN
      if(var_fld->allow_integer IS_YES)
         num_in_common = compare_fld_integer_lists(var_fld, CurField);
      if(var_fld->allow_float IS_YES)
         num_in_common += compare_fld_float_lists(var_fld, CurField);
      /*======================*/
      /* If no common values, */
      /*    print warning     */
      /*======================*/
      IF(num_in_common EQ 0) THEN
         sprintf(MsgBuf,
            "Variable <%.40s> used in fields whose ALLOWED_INTEGERS/FLOATS/NUMBERS definitions have no common values.",
            var_name);
         ErrorMessage(ERROR, MsgBuf);
      END_IF
   /*=======================================================*/
   /* Otherwise, if no number list, compare only the ranges */
   /*=======================================================*/

   ELSE_IF((var_fld->possible_integers EQ NULL) AND
           (CurField->possible_integers EQ NULL) AND
           (var_fld->possible_floats EQ NULL) AND
           (CurField->possible_floats EQ NULL))
      num_in_common = CompareFieldRanges(var_fld, CurField);

      /*==================*/
      /* If no overlap,   */
      /*    print warning */
      /*==================*/

      IF(num_in_common EQ 0) THEN
         sprintf(MsgBuf,
            "Variable <%.40s> used in fields/arguments whose RANGE \
definitions do not overlap.", var_name);
         ErrorMessage(ERROR, MsgBuf);
      END_IF

   /*=========================================*/
   /* Otherwise, compare number list to range */
   /*=========================================*/

   ELSE
      num_in_common = CompareIntegerListWithRange(var_fld, CurField);
      num_in_common += CompareFloatListWithRange(var_fld, CurField);
   END_IF

   return(num_in_common);
}


/* ------------------------------------------------------------------
 *  NAME   : compare_fld_integer_lists()
 *
 *  PURPOSE: This function compares the list of allowed integers in
 *           the current field against the cumulative integer list
 *           from previous fields.
 *
 *  INPUTS:  Two arguments,
 *           A pointer to the cumulative field info (DEF_FLD *),
 *           and a pointer to the current field info (DEF_FLD *).
 *
 *  RETURNS: 0 if no common values,
 *           1000 (no count) if either number list is NULL,
 *           or the number of common numbers if neither is NULL.
 * ------------------------------------------------------------------
 */

int compare_fld_integer_lists(var_fld,CurField)
   DEF_FLD *var_fld,*CurField;
{
   IL_PTR var_nl, cur_nl, old_num = NULL;
   int    num_in_common, found_number;

   var_nl = var_fld->possible_integers;
   cur_nl = CurField->possible_integers;

   /*============================================*/
   /* If current field doesn't have number list, */
   /*    don't reset, don't count common values  */
   /*============================================*/

   IF(cur_nl EQ NULL) THEN
      return(1000);
   END_IF

   /*=================================================*/
   /* Else if previous fields don't have number list, */
   /*    copy current field's number list,            */
   /*    don't count common values                    */
   /*=================================================*/

   IF(var_nl EQ NULL) THEN
      while(cur_nl NEQ NULL) DO
         var_nl = AddToIntegerList(var_nl, cur_nl->number,NO);
         cur_nl = cur_nl->next_number;
      END_WHILE

      var_fld->possible_integers = var_nl;
      return(1000);
   END_IF

   /*=========================*/
   /* If neither list is NULL */
   /*=========================*/

   num_in_common = 0;

   while (var_nl NEQ NULL) DO
      found_number = NO;
      cur_nl = CurField->possible_integers;
      while (cur_nl NEQ NULL) DO

         /*=====================*/
         /* Compare each number */
         /*=====================*/

         IF (cur_nl->number EQ var_nl->number) THEN
            found_number = YES;
            break;
         END_IF
         cur_nl = cur_nl->next_number;
      END_WHILE

      /*====================================================================*/
      /* If a number was in a previous field, but not in the current field, */
      /*    delete it from the variable's cumulative number list            */
      /*====================================================================*/

      IF(found_number IS_NO) THEN
         var_nl = FreeIntegerAndGetNext(&var_fld->possible_integers,
                                          var_nl,old_num);

      /*========================================*/
      /* Otherwise, count it as a common number */
      /*========================================*/

      ELSE
         ++num_in_common;
         old_num = var_nl;
         var_nl  = var_nl->next_number;
      END_IF
   END_WHILE

   return(num_in_common);
}

/* ------------------------------------------------------------------
 *  NAME   : compare_fld_integer_lists()
 *
 *  PURPOSE: This function compares the list of allowed integers in
 *           the current field against the cumulative integer list
 *           from previous fields.
 *
 *  INPUTS:  Two arguments,
 *           A pointer to the cumulative field info (DEF_FLD *),
 *           and a pointer to the current field info (DEF_FLD *).
 *
 *  RETURNS: 0 if no common values,
 *           1000 (no count) if either number list is NULL,
 *           or the number of common numbers if neither is NULL.
 * ------------------------------------------------------------------
 */

int compare_fld_float_lists(var_fld,CurField)
   DEF_FLD *var_fld,*CurField;
{
   FL_PTR var_nl, cur_nl, old_num = NULL;
   int    num_in_common, found_number;

   var_nl = var_fld->possible_floats;
   cur_nl = CurField->possible_floats;

   /*============================================*/
   /* If current field doesn't have number list, */
   /*    don't reset, don't count common values  */
   /*============================================*/

   IF(cur_nl EQ NULL) THEN
      return(1000);
   END_IF

   /*=================================================*/
   /* Else if previous fields don't have number list, */
   /*    copy current field's number list,            */
   /*    don't count common values                    */
   /*=================================================*/

   IF(var_nl EQ NULL) THEN
      while(cur_nl NEQ NULL) DO
         var_nl = AddToFloatList(var_nl, cur_nl->number,NO);
         cur_nl = cur_nl->next_number;
      END_WHILE

      var_fld->possible_floats = var_nl;
      return(1000);
   END_IF

   /*=========================*/
   /* If neither list is NULL */
   /*=========================*/

   num_in_common = 0;

   while (var_nl NEQ NULL) DO
      found_number = NO;
      cur_nl = CurField->possible_floats;
      while (cur_nl NEQ NULL) DO

         /*=====================*/
         /* Compare each number */
         /*=====================*/

         IF (cur_nl->number EQ var_nl->number) THEN
            found_number = YES;
            break;
         END_IF
         cur_nl = cur_nl->next_number;
      END_WHILE

      /*====================================================================*/
      /* If a number was in a previous field, but not in the current field, */
      /*    delete it from the variable's cumulative number list            */
      /*====================================================================*/

      IF(found_number IS_NO) THEN
         var_nl = FreeFloatAndGetNext(&var_fld->possible_floats,
                                          var_nl,old_num);

      /*========================================*/
      /* Otherwise, count it as a common number */
      /*========================================*/

      ELSE
         ++num_in_common;
         old_num = var_nl;
         var_nl  = var_nl->next_number;
      END_IF
   END_WHILE

   return(num_in_common);
}
/* ------------------------------------------------------------------
 *  NAME   : CompareFieldRanges()
 *
 *  PURPOSE: This function compares the range of allowed numbers in
 *           the current field against the cumulative number range
 *           from previous fields.
 *
 *  INPUTS:  Two arguments,
 *           A pointer to the cumulative field info (DEF_FLD *),
 *           and a pointer to the current field info (DEF_FLD *).
 *
 *  RETURNS: 0 if ranges don't overlap,
 *           1 if max_range = min_range,
 *           1000 (no count) if ranges overlap by more than 1.
 * ------------------------------------------------------------------
 */

int CompareFieldRanges(var_fld,CurField)
   DEF_FLD *var_fld, *CurField;
{
   int num_common = 1000;
   /*==================================================*/
   /* reset the min range to be the greater of the two */
   /*==================================================*/

   IF(CurField->set_int_min IS_YES) THEN
      IF(var_fld->set_int_min IS_NO) THEN
         var_fld->set_int_min   = YES;
         var_fld->int_min_range = CurField->int_min_range;
      ELSE
         IF(var_fld->int_min_range < CurField->int_min_range) THEN
            var_fld->int_min_range = CurField->int_min_range;
         END_IF
      END_IF
   END_IF
   IF(CurField->set_float_min IS_YES) THEN
      IF(var_fld->set_float_min IS_NO) THEN
         var_fld->set_float_min   = YES;
         var_fld->float_min_range = CurField->float_min_range;
      ELSE
         IF(var_fld->float_min_range < CurField->float_min_range)THEN
            var_fld->float_min_range = CurField->float_min_range;
         END_IF
      END_IF
   END_IF

   /*=================================================*/
   /* reset the max range to be the lesser of the two */
   /*=================================================*/

   IF(CurField->set_int_max IS_YES) THEN
      IF(var_fld->set_int_max IS_NO) THEN
         var_fld->set_int_max   = YES;
         var_fld->int_max_range = CurField->int_max_range;
      ELSE
         IF(var_fld->int_max_range > CurField->int_max_range) THEN
            var_fld->int_max_range = CurField->int_max_range;
         END_IF
      END_IF
   END_IF
   IF(CurField->set_float_max IS_YES) THEN
      IF(var_fld->set_float_max IS_NO) THEN
         var_fld->set_float_max   = YES;
         var_fld->float_max_range = CurField->float_max_range;
      ELSE
         IF(var_fld->float_max_range > CurField->float_max_range)THEN
            var_fld->float_max_range = CurField->float_max_range;
         END_IF
      END_IF
   END_IF

   /*=============================================================*/
   /* If both range limits have been set, check the range overlap */
   /*=============================================================*/

   IF((var_fld->set_int_min IS_ON) AND (var_fld->set_int_max IS_ON)) THEN
     
      /*================================================*/
      /* If no overlap, reset range values and return 0 */
      /*================================================*/
      
      IF(var_fld->int_max_range < var_fld->int_min_range) THEN
         var_fld->int_max_range = -1000;
         var_fld->int_min_range = 1000;
         num_common = 0;
      ELSE_IF(var_fld->int_max_range EQ var_fld->int_min_range)
        num_common = 1;
      END_IF
   END_IF
   IF((var_fld->set_float_min IS_ON) AND (var_fld->set_float_max IS_ON))THEN
      
      /*================================================*/
      /* If no overlap, reset range values and return 0 */
      /*================================================*/
      
      IF(var_fld->float_max_range < var_fld->float_min_range) THEN
         var_fld->float_max_range = -1000.0;
         var_fld->float_min_range = 1000.0;
         if(num_common != 1)
            num_common = 0;
      ELSE_IF(var_fld->float_max_range EQ var_fld->float_min_range)
        num_common = 1;
      END_IF    
   END_IF

   /*================================*/
   /* Otherwise, overlap is infinite */
   /*================================*/

   return(num_common);
}

/* --------------------------------------------------------------------
 *  NAME   : CompareIntegerListWithRange()
 *
 *  PURPOSE: This function is called when there is a mix of number
 *           list and range.  It checks to see that the numbers in
 *           number list are all within the range.  The resulting
 *           intersection is a number list.
 *
 *  INPUTS:  Two arguments,
 *           A pointer to the cumulative field info (DEF_FLD *),
 *           and a pointer to the current field info (DEF_FLD *).
 *
 *  RETURNS: 0 if no numbers from the number list are inside the range,
 *           1000 (no count) if the range has not been specified,
 *           the count of numbers in the range otherwise.
 * --------------------------------------------------------------------
 */

int CompareIntegerListWithRange(var_fld, CurField)
   DEF_FLD *var_fld, *CurField;
{
   IL_PTR nl_ptr,old_num = NULL;
   int    num_in_common = 0;

   /*=================================================*/
   /* If it is the var field that has the number list */
   /*=================================================*/

   IF (var_fld->possible_integers NEQ NULL)THEN

      /*==============================================================*/
      /* If the current field has no range, don't count common values */
      /*==============================================================*/

      IF((CurField->set_int_min IS_NO) AND (CurField->set_int_max IS_NO)) THEN
         return(1000);
      END_IF

      /*=================================================================*/
      /* Otherwise, check to make sure each number in the list is within */
      /*    the range, and count the common values                       */
      /*=================================================================*/

      nl_ptr = var_fld->possible_integers;
      while(nl_ptr NEQ NULL) DO
         IF(InIntegerRange(nl_ptr->number,CurField)) THEN
            ++num_in_common;
            old_num = nl_ptr;
            nl_ptr  = nl_ptr->next_number;
         ELSE
            nl_ptr = FreeIntegerAndGetNext(&var_fld->possible_integers,
                                              nl_ptr,old_num);
         END_IF
      END_WHILE

   /*================================================================*/
   /* Otherwise, if it is the current field that has the number list */
   /*================================================================*/

   ELSE
      nl_ptr = CurField->possible_integers;

      /*===============================================================*/
      /* If the var field has no range, copy the list, but don't count */
      /*    common values                                              */
      /*===============================================================*/

      IF((var_fld->set_int_min IS_NO) AND (var_fld->set_int_max IS_NO)) THEN
         while(nl_ptr NEQ NULL) DO
            var_fld->possible_integers =
               AddToIntegerList(var_fld->possible_integers,nl_ptr->number,NO);
            nl_ptr = nl_ptr->next_number;
         END_WHILE
         return(1000);

      /*===================================================================*/
      /* Otherwise, copy only the numbers within the range, and count them */
      /*===================================================================*/

      ELSE
         while(nl_ptr NEQ NULL) DO
            IF(InIntegerRange(nl_ptr->number,var_fld)) THEN
               ++num_in_common;
               var_fld->possible_integers =
                  AddToIntegerList(var_fld->possible_integers,nl_ptr->number,
                  NO);
           END_IF
            nl_ptr = nl_ptr->next_number;
         END_WHILE

         var_fld->set_int_min   = NO;
         var_fld->set_int_max   = NO;
         var_fld->int_min_range = -1000;
         var_fld->int_max_range = 1000;
      END_IF
   END_IF

   return(num_in_common);
}

/* --------------------------------------------------------------------
 *  NAME   : CompareFloatListWithRange()
 *
 *  PURPOSE: This function is called when there is a mix of number
 *           list and range.  It checks to see that the numbers in
 *           number list are all within the range.  The resulting
 *           intersection is a number list.
 *
 *  INPUTS:  Two arguments,
 *           A pointer to the cumulative field info (DEF_FLD *),
 *           and a pointer to the current field info (DEF_FLD *).
 *
 *  RETURNS: 0 if no numbers from the number list are inside the range,
 *           1000 (no count) if the range has not been specified,
 *           the count of numbers in the range otherwise.
 * --------------------------------------------------------------------
 */

int CompareFloatListWithRange(var_fld, CurField)
   DEF_FLD *var_fld, *CurField;
{
   FL_PTR nl_ptr,old_num = NULL;
   int    num_in_common = 0;

   /*=================================================*/
   /* If it is the var field that has the number list */
   /*=================================================*/

   IF (var_fld->possible_floats NEQ NULL)THEN

      /*==============================================================*/
      /* If the current field has no range, don't count common values */
      /*==============================================================*/

      IF((CurField->set_float_min IS_NO) AND (CurField->set_float_max IS_NO)) THEN
         return(1000);
      END_IF

      /*=================================================================*/
      /* Otherwise, check to make sure each number in the list is within */
      /*    the range, and count the common values                       */
      /*=================================================================*/

      nl_ptr = var_fld->possible_floats;
      while(nl_ptr NEQ NULL) DO
         IF(InFloatRange(nl_ptr->number,CurField)) THEN
            ++num_in_common;
            old_num = nl_ptr;
            nl_ptr  = nl_ptr->next_number;
         ELSE
            nl_ptr = FreeFloatAndGetNext(&var_fld->possible_floats,
                                              nl_ptr,old_num);
         END_IF
      END_WHILE

   /*================================================================*/
   /* Otherwise, if it is the current field that has the number list */
   /*================================================================*/

   ELSE
      nl_ptr = CurField->possible_floats;

      /*===============================================================*/
      /* If the var field has no range, copy the list, but don't count */
      /*    common values                                              */
      /*===============================================================*/

      IF((var_fld->set_float_min IS_NO) AND (var_fld->set_float_max IS_NO)) THEN
         while(nl_ptr NEQ NULL) DO
            var_fld->possible_floats =
               AddToFloatList(var_fld->possible_floats,nl_ptr->number,NO);
            nl_ptr = nl_ptr->next_number;
         END_WHILE
         return(1000);

      /*===================================================================*/
      /* Otherwise, copy only the numbers within the range, and count them */
      /*===================================================================*/

      ELSE
         while(nl_ptr NEQ NULL) DO
            IF(InFloatRange(nl_ptr->number,var_fld)) THEN
               ++num_in_common;
               var_fld->possible_floats =
                  AddToFloatList(var_fld->possible_floats,nl_ptr->number,
                  NO);
           END_IF
            nl_ptr = nl_ptr->next_number;
         END_WHILE

         var_fld->set_float_min   = NO;
         var_fld->set_float_max   = NO;
         var_fld->float_min_range = -1000;
         var_fld->float_max_range = 1000;
      END_IF
   END_IF

   return(num_in_common);
}

/* --------------------------------------------------------------------
 *  NAME   : InIntegerRange()
 *
 *  PURPOSE: This function determines whether or not a number is within
 *           the given field's range of allowed numbers.
 *
 *  INPUTS:  Two arguments,
 *           The number (int),
 *           and a pointer to the field info (DEF_FLD *).
 *
 *  RETURNS: TRUE if the number is inside the range,
 *           and FALSE otherwise.
 * --------------------------------------------------------------------
 */

int InIntegerRange(number,fld_ptr)
   int    number;
   DEF_FLD *fld_ptr;
{
   IF((fld_ptr->set_int_min IS_YES) AND (number < fld_ptr->int_min_range)) THEN
      return(FALSE);
   END_IF

   IF((fld_ptr->set_int_max IS_YES) AND (number > fld_ptr->int_max_range)) THEN
      return(FALSE);
   END_IF

   return(TRUE);
}

/* --------------------------------------------------------------------
 *  NAME   : InFloatRange()
 *
 *  PURPOSE: This function determines whether or not a number is within
 *           the given field's range of allowed numbers.
 *
 *  INPUTS:  Two arguments,
 *           The number (int),
 *           and a pointer to the field info (DEF_FLD *).
 *
 *  RETURNS: TRUE if the number is inside the range,
 *           and FALSE otherwise.
 * --------------------------------------------------------------------
 */

int InFloatRange(number,fld_ptr)
   float    number;
   DEF_FLD *fld_ptr;
{
   IF((fld_ptr->set_float_min IS_YES) AND (number < fld_ptr->float_min_range)) THEN
      return(FALSE);
   END_IF

   IF((fld_ptr->set_float_max IS_YES) AND (number > fld_ptr->float_max_range)) THEN
      return(FALSE);
   END_IF

   return(TRUE);
}

/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : check_num_fields()
 *
 *  PURPOSE: This function compares the number of fields
 *           used in a rule or deffact to the defrelation
 *           paramaters and prints appropriate error
 *           messages (if needed).
 *
 *  INPUTS:  Two arguments,
 *           A pointer to the relation name, (char *)
 *           the number of fields in the pattern (int).
 *
 *  RETURNS: Nothing.
 *
 *  NOTES:   The number of fields paramater is negative if a
 *           multiple field variable was used in the pattern.
 * ------------------------------------------------------------------
 */

void CheckMinNumOfFields(name, org_fields,assert_flag)
char *name;
int   org_fields, assert_flag;
{
   DEF_REL  *def_rel;
   int       num_fields, USED_MULTI;

   /*===========================*/
   /* Find defrelation (if any) */
   /*===========================*/

   def_rel = FindDefrelation(DefrelationHead, name);

   /*=======================================*/
   /* No defrelation, flag error and return */
   /*=======================================*/

   IF(def_rel EQ NULL) THEN
      DefrelationHead = FindCreateDefrelation(DefrelationHead, name);
      IF (strcmp(name, "initial-fact") NEQ 0) THEN
         sprintf(MsgBuf, "Relation <%.40s> used prior to its definition.",
            name);
         ErrorMessage(WARNING, MsgBuf);
      END_IF
      return;
   END_IF

   /*===========================*/
   /* Check for multi-field use */
   /*===========================*/

   IF(org_fields < 0) THEN
      num_fields = abs(org_fields);
      USED_MULTI = YES;
   ELSE
      num_fields = org_fields;
      USED_MULTI = NO;
   END_IF

   /*==================*/
   /* Check min fields */
   /*==================*/

   IF((def_rel->min_fields > num_fields) AND (USED_MULTI IS_NO)) THEN
      sprintf(MsgBuf,"Fewer fields used than defrelation <%.40s> requires.",
               name);
      ErrorMessage(ERROR, MsgBuf);
   END_IF

   /*==================*/
   /* Check max fields */
   /*==================*/

   IF((def_rel->max_fields < num_fields) AND (USED_MULTI IS_NO)) THEN
      sprintf(MsgBuf,"more fields used than defrelation <%.40s> allows.",
               name);
      ErrorMessage(ERROR, MsgBuf);
   /*ELSE_IF((def_rel->max_fields EQ num_fields) AND
           (USED_MULTI IS_YES))
      IF (!assert_flag) THEN
        ErrorMessage(WARNING, "Unecessary $? variable used.");
      END_IF*/
   END_IF
}
