#include <stdio.h>
#include "crsv.h"

/* ==================================================================== */
/*   This file contains functions that allow CRSV to store information  */
/*   about nested-if levels so that it may give intelligent error       */
/*   messages when multiple retracts are encountered.  The first few    */
/*   functions are general to nested if's, else's, and while's, and the */
/*   last few are specific to retracts. The stack (and its information!)*/
/*   can ONLY be accessed through these functions.                      */
/* ==================================================================== */


/* ===========  Functions defined Externally but used here  =========== */

/* -----------------------------
 *  From the file: CRSVMEM.C
 * -----------------------------
 */

extern void  MemoryError();
extern char *GenAlloc();
extern void  GenFree();


/* ===========  Functions defined here for Global use  ================ */

void ResetStack();
void Push();
void Pop();
int  InsideIf();
int  InsideWhile();
int  NumberOfNestLevels();
int  DescribePrevious();
void MarkRetraction();


/* ===========  Functions defined here for internal use  ============== */

STK_ELT *NewElement();
int      FindElementInStack();


/* ===========  Variables defined here for internal use  ============== */

STK_PTR  TopOfStack = NULL;
int      NewestLevel = 0;

/* ==================================================================== */
/*                IF, ELSE, and WHILE STACK OPERATORS                   */
/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  ResetStack
 *
 *  PURPOSE:  This function initializes the stack by setting
 *            TopOfStack to NULL and resetting the global
 *            level-counter to zero. This function should be
 *            called each time a new rule is being processed.
 *
 *  INPUTS:   None.
 *
 *  RETURNS:  Nothing.
 * -------------------------------------------------------------
 */

void ResetStack()
{
   while(TopOfStack NEQ NULL) DO
      Pop();
   END_WHILE

   NewestLevel = 0;
}

/*
 * -------------------------------------------------------------
 *  NAME   :  NewElement
 *
 *  PURPOSE:  This function allocates the memory needed by an
 *            individual stack element. It does NOT initialize
 *            any of the new element's fields.
 *
 *  INPUTS:   None.
 *
 *  RETURNS:  a pointer to the new element.
 * -------------------------------------------------------------
 */

static STK_ELT *NewElement()
{
   STK_ELT *NewElt;

   if ((NewElt = (STK_ELT *) GenAlloc(sizeof(STK_ELT),"NewElement")) ==
      NULL)
      MemoryError();

   return(NewElt);
}

/*
 * -------------------------------------------------------------
 *  NAME   :  Push
 *
 *  PURPOSE:  This function Pushes a new element on the stack
 *            and assigns it a level and a type. The element
 *            then becomes TopOfStack.
 *
 *  INPUTS:   The element's type (if, else, or while).
 *
 *  RETURNS:  Nothing.
 * -------------------------------------------------------------
 */

void Push(elt_type)
   int elt_type;
{
   STK_ELT *NewElt;

   NewElt = NewElement();

   if (elt_type != _else_)
   {
      NewElt->level      = ++NewestLevel;
      NewElt->next_level = 0;
   }

   /*======================================================================*/
   /* NOTE: When an else is encountered, its companion-if is TopOfStack  */
   /*======================================================================*/

   else if (NewestLevel != 0)
   {
      NewElt->level      = TopOfStack->level;
      NewElt->next_level = NewestLevel + 1;
   }

   NewElt->type = elt_type;
   NewElt->next = TopOfStack;
   TopOfStack  = NewElt;
}

/*
 * -------------------------------------------------------------
 *  NAME   :  Pop
 *
 *  PURPOSE:  This function Pops an element off the stack. If
 *            the element (TopOfStack) is an else, it Pops off
 *            the next element also, which is its companion if.
 *            It resets the TopOfStack accordingly.
 *
 *  INPUTS:   None.
 *
 *  RETURNS:  Nothing.
 * -------------------------------------------------------------
 */

void Pop()
{
   STK_ELT *OldElt;

   if (TopOfStack != NULL)
   {
      if (TopOfStack->type == _else_)
      {
         /*==========================*/
         /* Pop off an extra element */
         /*==========================*/

         OldElt = TopOfStack;
         TopOfStack = TopOfStack->next;
         GenFree((char *)OldElt, sizeof(STK_ELT));
      }
      OldElt = TopOfStack;
      TopOfStack = TopOfStack->next;
      GenFree((char *)OldElt, sizeof(STK_ELT));
   }
}

/* ==================================================================== */
/*                      STACK INFORMATION FUNCTIONS                     */
/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  InsideIf()
 *
 *  PURPOSE:  This function detects whether the current level is
 *            an if block.
 *
 *  INPUTS:   None.
 *
 *  RETURNS:  YES if it is, NO if it isn't.
 * -------------------------------------------------------------
 */

int InsideIf()
{
   if (TopOfStack == NULL)
      return(NO);

   if (TopOfStack->type == _if_)
      return(YES);
   else
      return(NO);
}

/*
 * --------------------------------------------------------------
 *  NAME   :  InsideWhile()
 *
 *  PURPOSE:  This function detects whether the current level is
 *            inside a while block.  NOTE: This is different from
 *            InsideIf() in that it checks the whole stack, not
 *            just the current level!
 *
 *  INPUTS:   None.
 *
 *  RETURNS:  YES if it is, NO if it isn't.
 * --------------------------------------------------------------
 */

int InsideWhile()
{
   int CurLevel;
   STK_PTR temp;

   IF(TopOfStack NEQ NULL) THEN
      CurLevel = TopOfStack->level;
   ELSE
      CurLevel = 0;
   END_IF

   temp = TopOfStack;

   while(temp != NULL)
   {
      if ((temp->type == _while_) && (CurLevel >= temp->level))
         return(YES);
      temp = temp->next;
   }
   return(NO);
}

/*
 * --------------------------------------------------------------
 *  NAME   :  NumberOfNestLevels()
 *
 *  PURPOSE:  This function counts how many blocks of type
 *            block_type have been nested..
 *
 *  INPUTS:   The block type (an integer).
 *
 *  RETURNS:  The number of nested levels.
 * --------------------------------------------------------------
 */

int NumberOfNestLevels(BlockType)
   int BlockType;
{
   STK_PTR temp;
   int level_count = 0;

   for (temp = TopOfStack; temp != NULL; temp = temp->next)
      if (temp->type == BlockType)
         ++level_count;

   return(level_count);
}

/*
 * -------------------------------------------------------------
 *  NAME   :  FindElementInStack
 *
 *  PURPOSE:  This function searches the stack to determine
 *            whether it contains an element with the given
 *            level. It returns upon finding the first occurence.
 *
 *  INPUTS:   The level to be located.
 *
 *  RETURNS:  YES if it is there, NO if it isn't.
 * -------------------------------------------------------------
 */

static int FindElementInStack(EltLevel)
   int EltLevel;
{
   STK_PTR temp;

   if (TopOfStack == NULL)
      return(NO);

   temp = TopOfStack->next;

   while(temp != NULL)
   {
      if (temp->level == EltLevel)
         return(YES);
      temp = temp->next;
   }
   return(NO);
}

/* ==================================================================== */
/*                      RETRACT-SPECIFIC FUNCTIONS                      */
/* ==================================================================== */

/*
 * ---------------------------------------------------------------
 *  NAME   :  DescribePrevious
 *
 *  PURPOSE:  This function classifies a previously encountered
 *            retraction variable as seen from the current level.
 *
 *  INPUTS:   A previously retracted retraction variable.
 *
 *  RETURNS:  An integer that classifies the retraction variable.
 * ---------------------------------------------------------------
 */

int DescribePrevious(PrevRetVar)
   VAR_PTR PrevRetVar;
{
   STK_ELT curr;
   int     PrevLevel = PrevRetVar->level_retracted;
   int     PrevType  = PrevRetVar->block_type;
   int     found;
   int     rtn = OK_RETRACT;

   if (TopOfStack == NULL)
   {
      curr.level = 0;
      curr.type  = 0;
      found      = NO;
   }
   else
   {
      curr.level      = TopOfStack->level;
      curr.type       = TopOfStack->type;
      if (curr.type == _else_)
         curr.next_level = TopOfStack->next_level;
      found           = FindElementInStack(PrevLevel);
   }

   /*=====================================================================*/
   /*  If previous level was outside an if, current level doesn't matter  */
   /*=====================================================================*/

   if (PrevLevel == OUTSIDE_IF)
      rtn = OUTSIDE_IF;

   /*==================================================================*/
   /*  If the previous level was entered before the current level was  */
   /*==================================================================*/

   else if (PrevLevel < curr.level)
   {
      /*==============================================*/
      /*  If not on stack, must be separate if-block  */
      /*==============================================*/

      if (!found)
         rtn = SEPARATE;

      /*==========================================*/
      /*  Otherwise, must be superlevel if_block  */
      /*==========================================*/

      else
         rtn = SUPERLEVEL;
   }

   /*=================================================================*/
   /*  If the previous level was entered after the current level was  */
   /*=================================================================*/

   else if (PrevLevel > curr.level)
   {
      /*=======================================================*/
      /*  If the current type wasn't an else, must be sublevel */
      /*=======================================================*/

      if (curr.type != _else_)
         rtn = SUBLEVEL;

      /*===============================================================*/
      /*  Otherwise, if the previous level was within the else block,  */
      /*     it must be a sublevel                                     */
      /*===============================================================*/

      else if (PrevLevel >= curr.next_level)
         rtn = SUBLEVEL;

      /*=============================================================*/
      /*  Otherwise, it's a sublevel of companion-if and must be OK  */
      /*=============================================================*/

      else
         rtn = OK_RETRACT;
   }

   /*==========================================================*/
   /*  If the previous level is the same as the current level  */
   /*==========================================================*/

   else if (curr.level == PrevLevel)
   {
      /*=====================================================*/
      /*  If the types are the same, then must be same block */
      /*=====================================================*/

      if (curr.type == PrevType)
         rtn = SAME;

      /*===================================*/
      /*  Otherwise, must be companion-if  */
      /*===================================*/

      else
         rtn = COMPANION_IF;
   }

   return(rtn);
}

/*
 * -------------------------------------------------------------
 *  NAME   :  MarkRetraction
 *
 *  PURPOSE:  This function marks a retraction variable as
 *            retracted - sets it up for future encounters.
 *
 *  INPUTS:   A pointer to the retraction variable.
 *
 *  RETURNS:  Nothing.
 * -------------------------------------------------------------
 */

void MarkRetraction(RetVar)
   VAR_PTR RetVar;
{
   if (TopOfStack == NULL)
      RetVar->level_retracted = OUTSIDE_IF;
   else
   {
      RetVar->level_retracted = TopOfStack->level;
      RetVar->block_type      = TopOfStack->type;
   }
}
