#include <stdio.h>
#include "crsv.h"

/*****************************************************************
 *  This file contains the functions used to parse the CLIPS
 *  rule file. The functions are similar to the token parsing
 *  functions used in CLIPS, but there are changes. The primary
 *  entry point is through the GetToken function.
 *****************************************************************
 */


/* ===========  Functions defined Externally but used here  =========== */

/* -----------------------------
 *  From the system
 * -----------------------------
 */

extern double   atof();
extern long     atol();

/* ------------------------------
 * From CRSVPROC.C
 * ------------------------------
 */

extern PAT_TKN_PTR InsertTokenIntoPattern();
extern void        DequeueTokenFromPattern();

/* -----------------------------
 *  From the file: CRSVUSR.C
 * -----------------------------
 */

extern void UserStartComment();
extern void UserProcessComment();
extern void UserEndComment();

/* -----------------------------
 *  From the file: CRSVPRNT.C
 * -----------------------------
 */

extern void ErrorMessage();

/* -----------------------------
 *  From the file: CRSVMEM.C
 * -----------------------------
 */
 
extern TK_PTR AllocToken();
extern void FreeToken();
extern TMP_TKN_PTR AllocTempToken();
extern TMP_TKN_PTR FreeTempToken();

/* -----------------------------
 *  From the file: CRSVHASH.C
 * -----------------------------
 */
 
extern HASH  *AddSymbol();

/* -----------------------------
 *  From the file: CRSVMISC.C
 * -----------------------------
 */
 
extern TK_PTR DequeueToken();
extern void QueueToken ();


/* ===========  Functions defined here for Global use  ================ */

int       SetGetTokenQueue();
int       GetToken();
TK_PTR    GetTokenPtr();
void      UnGetTokenPtr();
void      SaveToken();
char      *StringSave();
void      FreeTokenInRelation();
void 	  StoreTokenInRelation();

/* ===========  Functions defined here for internal use  ============== */

int  FetchNum();
int  FetchVar();
int  FetchAny();


/* ===========  Variables defined here for Global use  ================ */

double FLOATNUMBER;                  /* GetToken float return  */
long   INTNUMBER;              /* GetToken integer return */
char  *TKNWORD;                      /* GetToken string return  */
HASH  *HASHWORD;                     /* GetToken hash table storage */


/* ===========  Variables defined Externally but used here  =========== */

/* -----------------------------
 *  From the file: CRSV.C
 * -----------------------------
 */

extern int   CurrentLineNumber;             /* Current line number   */
extern short LHSPattern;
extern PAT_LIST_PTR CurrentPattern;
extern PAT_TKN_PTR CurrentTokenInPattern;
extern FILE *CurrentFile;


/* ===========  Variables defined here for internal use  ============== */


static int     TokenSaved = NO;         /* flag for saved token */
static int     SvToken;
static long    SvInt;
static double  SvFloat;
static char   *SvWord;
static HASH   *SvHash;
static short   StartReadField = YES;
static short   StoreToken = NO;
static TMP_TKN_PTR CurrentTokenPtr = NULL;
static TMP_TKN_PTR TokenList = NULL;
static TN_PTR  UnGetTokens = NULL;


/* ===================  PARSER FUNCTIONS  ======================= */

/*
 * -------------------------------------------------------------
 *  NAME   :  SetGetTokenQueue
 *
 *  PURPOSE:  Permits substitution of the unget token queue.
 *
 *  INPUTS:   The new queue.
 *
 *  RETURNS:  An integer, currently OK.
 * -------------------------------------------------------------
 */ 

int SetGetTokenQueue (new_queue)
 TN_PTR new_queue;
 {
   UnGetTokens= new_queue;

   return (OK);
 }

/* ====================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  GetTokenPtr
 *
 *  PURPOSE:  This function scans the file for the next token.
 *
 *  INPUTS:   A single argument, a pointer to the current file.
 *
 *  RETURNS:  A pointer to a token structure containing all of
 *            the token attributes is returned.
 * -------------------------------------------------------------
 */ 

TK_PTR GetTokenPtr(infile)
 FILE *infile;
 {
   TK_PTR tknstruct        = AllocToken ();

   tknstruct->token        = GetToken (infile);
   tknstruct->tknword      = TKNWORD;
   tknstruct->hashword     = HASHWORD;
   tknstruct->intnumber    = INTNUMBER;
   tknstruct->floatnumber  = FLOATNUMBER;
   tknstruct->field        = CurrentTokenInPattern;
   tknstruct->print_rep    = tknstruct->tknword;

   return (tknstruct);
 }

/* ====================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  UnGetTokenPtr
 *
 *  PURPOSE:  Places a token (via TK_PTR) back onto the input
 *            buffer.
 *
 *  INPUTS :  The token to unget.
 *
 *  RETURNS:  Nothing
 *
 *  NOTE   :  Unlike the UNIX unget, an arbitrary number of tokens
 *            may be placed back on the input buffer.
 * -------------------------------------------------------------
 */ 

void UnGetTokenPtr (tknstruct)
 TK_PTR tknstruct;
 {
   QueueToken (&UnGetTokens, tknstruct);
 }

/* ====================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  GetAToken
 *
 *  PURPOSE:  This function scans the file for the next token.
 *            The type of the token (word, string, number, 
 *            variable, etc.) is returned.
 *            If the token has an associated string, then the
 *            value of this string hashed and stored in the global 
 *            variable, TKNWORD. If the token is a number, then 
 *            its value is stored in the global variable,
 *            INTNUMBER or FLOATNUMBER.
 *
 *  INPUTS:   A single argument, a pointer to the current file.
 *
 *  RETURNS:  An integer, the token ID number, or ERROR, if the
 *            token contained bad characters.
 * -------------------------------------------------------------
 */ 

int GetAToken(infile,CallFrmGettokenWEOL)
   FILE  *infile;
   int CallFrmGettokenWEOL;
{
   int inchar;
   int i,token;

   if ((StoreToken IS_YES) && (CurrentTokenPtr == NULL)) 
     CurrentTokenPtr = TokenList;
   if(CurrentTokenPtr != NULL)
    {
       token   = CurrentTokenPtr->type;
       TKNWORD = CurrentTokenPtr->word;
       if (token == INTEGER) 
         INTNUMBER = atol (TKNWORD);
       else if (token == FLOAT)
         FLOATNUMBER = atof(TKNWORD);
       CurrentTokenInPattern = CurrentTokenPtr->field;
       CurrentTokenPtr = CurrentTokenPtr->next;
       return(token);
    }
   if (TokenSaved IS_YES)
   {
      token = GetSaveToken();
      /* ================================*/
      /* This for inference trace option */ 
      /* ================================*/  
      if(LHSPattern)
         CurrentPattern->pattern = InsertTokenIntoPattern(CurrentPattern->pattern,
				TKNWORD,token);
      return(token);
   }
   
   IF (UnGetTokens NEQ NULL) THEN
      TK_PTR tknstruct = DequeueToken (&UnGetTokens);
      int dequeued_token = tknstruct->token;

      TKNWORD     = tknstruct->tknword;
      HASHWORD    = tknstruct->hashword;
      INTNUMBER   = tknstruct->intnumber;
      FLOATNUMBER = tknstruct->floatnumber;
      CurrentTokenInPattern = tknstruct->field;
      FreeToken (tknstruct);
      return (dequeued_token);
   END_IF

   /*=======================================*/
   /* Set Unknown default values for token. */
   /*=======================================*/

   token     = KUNKNOWN;
   TKNWORD   = "unknown";
   HASHWORD  = NULL;
   FLOATNUMBER = 0.0;
   INTNUMBER   = 0;

   /*==============================================*/
   /* Remove all white space before processing the */ 
   /* GetToken() request.                          */
   /*==============================================*/

   inchar = SaveGetc(infile);
   
   while ((inchar == ' ') || (inchar == '\f') ||
          (inchar == '\r') || (inchar == ';') || (inchar == '\t'))
   {
      /*=======================*/
      /* Remove comment lines. */
      /*=======================*/
 
      if (inchar == ';')               
      {
         UserStartComment(';');
         while (((inchar = SaveGetc(infile))!= '\n') &&
                 (inchar != '\r') && (inchar != EOF))
            UserProcessComment(inchar);
         if(inchar == '\n')
           CurrentLineNumber++;
         UserEndComment('\n');
      }
     
      inchar = SaveGetc(infile);
   }

      /*========================*/
      /* Increment line counter */
      /*========================*/
   if (inchar == '\n') 
     {
        token = EOL;
        TKNWORD = "\n";
        return(token);
      }
   /*==========================*/
   /* Process Symbolic Tokens. */
   /*==========================*/

   if (((inchar >= 'a') && (inchar <= 'z')) ||
       ((inchar >= 'A') && (inchar <= 'Z')))
   {
      UnSaveGetc(inchar,infile);
      if (FetchVar(infile,&TKNWORD,NULL) == ERROR)
         return (ERROR);
      token = WORD;
   }

   /*========================*/
   /* Process String Tokens. */
   /*========================*/

   else if (inchar == '"')                          
   {
      if (FetchAny(infile,&TKNWORD) == ERROR)
         return(ERROR);
      token = STRING;
   }

   /*===============================================*/
   /* Process Number Tokens beginning with a digit. */
   /*===============================================*/
   
   else if ((inchar >= '0') && (inchar <= '9')) 
   {
      UnSaveGetc(inchar,infile);
      if (FetchNum(infile, &TKNWORD, NO,CallFrmGettokenWEOL) == ERROR)
          token = WORD; 
      else
         token = GetIntOrFloat();
   }

   /*==================================================================*/
   /* Process Number Tokens beginning with a "-" and the "-" Operator. */
   /*==================================================================*/

   else if (inchar == '-')
   {
      inchar = SaveGetc(infile);
      if ((inchar == '.') ||        /* number or decimal follows, */
          ((inchar >= '0') &&       /* process as numerical token */
           (inchar <= '9')))
      {
         UnSaveGetc(inchar,infile);
         if (FetchNum(infile, &TKNWORD, YES,CallFrmGettokenWEOL) == ERROR)
            token = WORD;
         else
           token = GetIntOrFloat();
      }
      else                         /* otherwise, use it as the */
      {                            /* subtraction operator     */
         UnSaveGetc(inchar,infile);
         if (FetchVar(infile, &TKNWORD, "-") == ERROR)
            return (ERROR);
         token = WORD;
      }
   }

   /*==================================================================*/
   /* Process Number Tokens beginning with a "+" and the "+" Operator. */
   /*==================================================================*/

   else if (inchar == '+')
   {
      inchar = SaveGetc(infile);
      if ((inchar == '.') ||         /* number or decimal follows, */
          ((inchar >= '0') &&        /* process as numerical token */
           (inchar <= '9')))
      {
         UnSaveGetc(inchar,infile);
         if (FetchNum(infile, &TKNWORD, NO,CallFrmGettokenWEOL) == ERROR)
            token  = WORD;
         else
           token = GetIntOrFloat();
      }
      else                                /* otherwise, use as the */
      {                                   /* ADD operator.         */
         UnSaveGetc(inchar,infile);
         if (FetchVar(infile, &TKNWORD, "+") == ERROR)
            return (ERROR);
         token = WORD;         
      }
   }
   
   /*============================*/
   /* Process ? and ?var Tokens. */
   /*============================*/
   
   else if (inchar == '?')
   {
      inchar = SaveGetc(infile);
      if (((inchar >= 'a') && (inchar <= 'z')) ||
           ((inchar >= 'A') && (inchar <= 'Z')))
      {
         UnSaveGetc(inchar,infile);
         if (FetchVar(infile, &TKNWORD, NULL) == ERROR)
            return (ERROR);
         token = BWORD;
      }
      else if (inchar == '*')
      {
         UnSaveGetc(inchar,infile);
         if (FetchVar(infile, &TKNWORD, NULL) == ERROR)
            return (ERROR);
         if(TKNWORD[strlen(TKNWORD) - 1 ] == '*')
             token = GLOBAL;
         else
            token = BWORD;
      }
      else
      {
         TKNWORD = "?";
         token = SINGLE;
         UnSaveGetc(inchar,infile);
      }
   }

   /*==============================*/
   /* Process "&" and "&&" Tokens. */
   /*==============================*/

   else if (inchar == '&')
   {
         TKNWORD = "&";
         token   = LAND;
   }

   /*==============================*/
   /* Process $? and $?var Tokens. */
   /*==============================*/

   else if (inchar == '$')
   {
      if ((inchar = SaveGetc(infile)) == '?')
      {
         inchar = SaveGetc(infile);
         if (((inchar >= 'a') && (inchar <= 'z')) ||
             ((inchar >= 'A') && (inchar <= 'Z')))
         {
            UnSaveGetc(inchar,infile);
            if (FetchVar(infile, &TKNWORD, NULL) == ERROR)
               return (ERROR);
            token = BWORDS;
         }
         else
         {
            TKNWORD = "$?";
            token   = MULTIPLE;
            UnSaveGetc(inchar,infile);
         }
      }
      else 
      {
         UnSaveGetc(inchar,infile);
         if (FetchVar(infile, &TKNWORD, "$") == ERROR)
            return (ERROR);
         token = WORD;         
      }
   }

   /*==============================*/
   /* Process "|" and "||" Tokens. */
   /*==============================*/

   else if (inchar == '|') 
   {
         TKNWORD = "|";
         token   = LOR;
   }

   /*==============================*/
   /* Process "=" and "=>" Tokens. */
   /*==============================*/
 
   else if (inchar == '=') 
   {
      if ((inchar = SaveGetc(infile)) == '>')
      { 
         TKNWORD = "=>";
         token   = SEPARATOR;
      }
      else
      {
         UnSaveGetc(inchar,infile);
         if (FetchVar(infile, &TKNWORD, "=") == ERROR)
            return (ERROR);
         token = WORD;         
      }
   }

   /*==============================*/
   /* Process "<" and "<-" Tokens. */
   /*==============================*/

   else if (inchar == '<')
   {
      if ((inchar = SaveGetc(infile)) == '-')
      {
         TKNWORD = "<-";
         token   = BINDER;
      }
      else
      {
         UnSaveGetc(inchar,infile);
         if (FetchVar(infile, &TKNWORD, "<") == ERROR)
            return (ERROR);
         token = WORD;         
      }
   }
   
   /*===================================*/
   /* Process "(", ")", and "~" Tokens. */
   /*===================================*/

   else if (inchar == '(')                         
   { 
      TKNWORD = "(";
      token   = LPAREN;
   }
   else if (inchar == ')')                          
   {
      TKNWORD = ")";
      token   = RPAREN;
   }
   else if (inchar == '~') 
   { 
      TKNWORD = "~";
      token   = LNOT;
   }

   /*=====================================*/
   /* Process Instance token for COOL     */
   /*=====================================*/
   else if (inchar == '[')
     {
       UnSaveGetc(inchar,infile);
       if(FetchVar(infile,&TKNWORD,NULL) == ERROR)
         return(ERROR);
       if(TKNWORD[strlen(TKNWORD) - 1] == ']')
         token = INSTANCE;
       else
         token  = WORD;
     }
   /*============================*/
   /* Process End-of-File Token. */
   /*============================*/

   else if (inchar == EOF)
   {
      token   = STOP;
      TKNWORD = "stop"; 
   }

   /*=======================*/
   /* Process Other Tokens. */
   /*=======================*/

   else 
   {
      UnSaveGetc(inchar,infile);
      if (FetchVar(infile, &TKNWORD, NULL) == ERROR)
         return (ERROR);
      token = WORD;
   }

/* ================================*/
/* This for inference trace option */ 
/* ================================*/  
   if(LHSPattern)  /* HACKING!!! */
     CurrentPattern->pattern = InsertTokenIntoPattern(CurrentPattern->pattern,TKNWORD,token);
   return(token);
}

/***************************************************
 *  GETTOKEN:
 *  This function return the token except EOL
 ***************************************************
 */
GetToken(file)
FILE *file;
{
  int token;
  
  while((token = GetAToken(file,NO)) == EOL)
    CurrentLineNumber++;
  return(token);
}

/***************************************************
 *  GetTokenWithEndOfLine
 *   This function returns the token including
 *   EOL
 ***************************************************
 */
GetTokenWithEndOfLine(file)
FILE *file;

{
  int token;
  
  token = GetAToken(file,YES);
  if(token == EOL)
    CurrentLineNumber++;
  return(token);
}

/******************************************************
 *  GetIntOrFLoat
 ******************************************************
 */
GetIntOrFloat()
{
   int i;
   for (i = 0; i < strlen(TKNWORD) ;i++)
     {
        if((TKNWORD[i] == '.') || (TKNWORD[i] == 'e') ||
           (TKNWORD[i] == 'E'))
         {
            FLOATNUMBER = atof(TKNWORD);
             return(FLOAT);
          }
     }
   INTNUMBER = atol(TKNWORD);
   return(INTEGER);
}

/* ==================================================================== */

/* -------------------------------------------
 *  NAME   :    SaveToken
 *
 *  PURPOSE:    This function stores a token
 *              and it's associated value, 
 *              either word or number in a 
 *              buffer for reuse.
 *
 *  INPUTS:     Three arguments
 *   int    token  - The token number
 *   float  number - tokens numerical value
 *   char  *word   - tokens string value
 *
 *  RETURNS:     Nothing useful.
 * -------------------------------------------
 */

void SaveToken(token, word, intnumber,floatnumber)
int token;
long intnumber;
double floatnumber;
char *word;
{
   SvToken = token;
   SvInt   = intnumber;
   SvFloat = floatnumber;
   SvWord  = word;
   SvHash  = HASHWORD;
   if(LHSPattern)             /*   HACKING !!!! */
     DequeueTokenFromPattern(CurrentPattern->pattern);
   TokenSaved = YES;
}


/* ====================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  FetchVar
 *
 *  PURPOSE:  This function gets one legal variable name
 *            from the input file. A legal variable name
 *            consists of any character that is not a 
 *            delimeter.
 *
 *  INPUTS :  A pointer to the input file (FILE *).
 *            A pointer to the string to be saved (char **).
 *            A string containing the first letters of the
 *               string to be saved (char *).
 *
 *  RETURNS:  OK or ERROR.
 * -------------------------------------------------------------
 */

static int FetchVar(infile,tknword,beg_str)
   FILE *infile;
   char **tknword;
   char *beg_str;
{
   int count = 0;
   int inchar;
   char buf[WORDLENGTH];
   
   if (beg_str != NULL)
      while (beg_str[count] != EOS)
      {
         buf[count] = beg_str[count];
         ++count;
      }

   inchar = SaveGetc(infile);
   while ( (inchar != '<') && (inchar != '"') &&
           (inchar != '(') && (inchar != ')') &&
           (inchar != '&') && (inchar != '|') &&
           (inchar != ' ') && (inchar != ';') &&
           (inchar >= ' ') && (inchar < '~') )
   {
      buf[count++] = inchar;
      inchar = SaveGetc(infile);
   }

   buf[count] = EOS;

   /*===================================================*/
   /* Stuff last character back into buffer and give    */
   /* the string a permanent address through StringSave(). */
   /*===================================================*/

   UnSaveGetc(inchar,infile);
   *tknword = StringSave(buf);  
   return(OK);
}

/* ====================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  FetchAny
 *
 *  PURPOSE:  This function gets one string, from a beginning
 *            double quote to the ending double quote.
 *
 *  INPUTS :  A pointer to the input file (FILE *).
 *            A pointer to the string to be saved (char **).
 *
 *  RETURNS:  OK or ERROR.
 * -------------------------------------------------------------
 */

static int FetchAny(infile,tknword)
   FILE *infile;
   char **tknword;
{
   int count = 1;
   int inchar,lastchar;
   char buf[WORDLENGTH];

   buf[0] = '"';
   inchar = SaveGetc(infile);
   
   while (inchar != '"')
   {
      if (count > (WORDLENGTH - 2))
      {
         buf[count] = EOS;
         *tknword = StringSave(buf);
         
         sprintf(buf,"Maximum token length of %d exceeded\n",WORDLENGTH);
         ErrorMessage(ERROR, buf);
         return (ERROR);
      }

      if (inchar == EOF)
      {
         UnSaveGetc(inchar,infile);
         buf[count] = EOS;
         *tknword = StringSave(buf);
         
         ErrorMessage (ERROR,
            "End of file encountered while reading string\n");
         return (ERROR);
      }

/*
      if (inchar == '\\')
         inchar = SaveGetc(infile);
*/

      if (inchar == '\n')
         CurrentLineNumber++;

      buf[count++] = inchar;
      lastchar = inchar;
      inchar = SaveGetc(infile);
      if ((lastchar == '\\') && (inchar == '"')) {
        buf[count++] = inchar;
        inchar = SaveGetc(infile);
      }
        
   }

   /*=================================================*/
   /* Add an end-quote and an EOS to buf and give the */
   /* string a permanent address through StringSave().   */
   /*=================================================*/
   
   buf[count++] = '"';
   buf[count] = EOS;
   *tknword = StringSave(buf);
   return(OK);
}

/* ====================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  FetchNum
 *
 *  PURPOSE:  This function gets a number from the input
 *            file. The number may be preceeded by a '+'
 *            or a '-' sign, and may be in exponential form.
 *
 *  INPUTS :  A pointer to the input file (FILE *), a pointer
 *            to the string to be saved (char **), and a
 *            flag telling whether or not this number is
 *            negative (int).
 *
 *  RETURNS:  OK or ERR.
 *
 *  NOTE   :  Major difference from standard CLIPS token 
 *            reader, which returns floating point values.
 * -------------------------------------------------------------
 */

static int FetchNum(infile, tknword, is_neg,CallFrmGettokenWEOL)
   FILE *infile;
   char **tknword;
   int   is_neg,CallFrmGettokenWEOL;
{
   int count = 0;
   int inchar,phase;
   char buf[WORDLENGTH];

/* Phases:                    */

#     define INTEGER_PART 0
#     define DECIMAL_PART 1
#     define EXPONENT_BEGIN 2
#     define EXPONENT_VALUE 3

   if (is_neg IS_YES)
      buf[count++] = '-';

   inchar = SaveGetc(infile);
   phase  = INTEGER_PART;

   while ((inchar != '<') && (inchar != '"')&&
          (inchar != '(') && (inchar != ')') &&
          (inchar != '&') && (inchar != '|') &&
          (inchar != ' ') && (inchar != ';') &&
          (inchar >= ' ') && (inchar <= '~') &&
          (inchar != ','))
          
   {
      if ((inchar >= '0') && (inchar <= '9'))
      {
         if (phase == EXPONENT_BEGIN)
            phase = EXPONENT_VALUE;
      }
      else if (inchar == '.')
      {
         if (phase == INTEGER_PART)
            phase = DECIMAL_PART;
         else
         {
            buf[count++] = inchar;
            buf[count]   = EOS;
            /* read string 'til next delimeter */
            (void)FetchVar(infile,tknword,buf); 
            return (ERROR);
         }
      }
      else if ((inchar == 'E') || (inchar == 'e'))
      {
         if ((phase == INTEGER_PART) || (phase == DECIMAL_PART))
            phase = EXPONENT_BEGIN;
         else
         {
      
            buf[count++] = inchar;
            buf[count]   = EOS;
            /* read string 'til next delimeter */
            (void)FetchVar(infile,tknword,buf); 
            return (ERROR);
         }
      }
      else if ((inchar == '-') || (inchar == '+'))
      {
         if (phase == EXPONENT_BEGIN)
            phase = EXPONENT_VALUE;
         else
         {

            buf[count++] = inchar;
            buf[count]   = EOS;
            /* read string 'til next delimeter */
            (void)FetchVar(infile,tknword,buf); 
            return (ERROR);
         }
      }
      else
      {
         buf[count++] = inchar;
         buf[count]   = EOS;
         /* read string 'til next delimeter */
         (void)FetchVar(infile,tknword,buf);
         return (ERROR);
      }

      buf[count++] = inchar;
      inchar = SaveGetc(infile);
   }
   if((inchar == ',') && (!CallFrmGettokenWEOL))
     {
         buf[count++] = inchar;
         buf[count]   = EOS;
         /* read string 'til next delimeter */
         (void)FetchVar(infile,tknword,buf);
         return (ERROR);
      }

   buf[count] = EOS;

   /*=======================================*/
   /* Stuff last character back into buffer */
   /* and return the number.                */
   /*=======================================*/

   UnSaveGetc(inchar,infile);
   *tknword = StringSave(buf);  
   return(OK);
  }

/* ---------------------------------------------------------------------- */

static SaveGetc(file)
FILE *file;
{
   return(getc(file));
}

/* ---------------------------------------------------------------------- */

static UnSaveGetc(c,file)
char  c;
FILE *file;
{
   (void)ungetc(c, file);
}

/* ---------------------------------------------------------------------- */

static int GetSaveToken()
{
   TokenSaved  = NO;
   TKNWORD     = SvWord;
   HASHWORD    = SvHash;
   FLOATNUMBER = SvFloat;
   INTNUMBER   = SvInt;
   
   return(SvToken);
}

/* ---------------------------------------------------------------------- */

/* ------------------------------------------------
 *  NAME   :  StringSave
 *
 *  PURPOSE:  This functions takes a pointer to a
 *            string, and puts it the hash table
 *            in the appropriate way.
 *
 *  INPUTS:   A pointer to the original string.
 *
 *  RETURN:   A pointer to the new string.
 * ------------------------------------------------
 */

char *StringSave(s)
   char *s;
{
   HASHWORD = AddSymbol(s);

   return(HASHWORD->contents);
}

/* =================================================== */
/* Store the fields of the relation (s) for latter      */
/* processing. This is done because of the possibility */
/* of multiple relations.                              */
/* =================================================== */
   
void StoreTokenInRelation()
{
  int count = 1,token;
  TMP_TKN_PTR temp_token = NULL,tail_list = NULL;
  HASH *temp_hash;
   
   
   WHILE (count > 0) DO
      token = GetToken (CurrentFile);

      IF (token == RPAREN) THEN
         count--;
      ELSE_IF (token == LPAREN) 
         count++;
      ELSE_IF (token == STOP)
         count = 0;
      END_IF

      temp_token = AllocTempToken ();
      temp_hash = AddSymbol (TKNWORD);
      temp_token->word = temp_hash->contents;
      temp_token->type = token;           /* Type of the token */
      temp_token->field = CurrentTokenInPattern;

      IF (TokenList == NULL) THEN
         tail_list = temp_token;
         TokenList = temp_token;
      ELSE
         tail_list->next = temp_token;
         tail_list = tail_list->next;
      END_IF
   DONE
   StoreToken = TRUE;
   CurrentTokenPtr = TokenList;
}
/* ===================================================================
 *
 * ===================================================================
 */

void FreeTokenInRelation()
{
  FreeTempToken(TokenList);
  TokenList = NULL;
  StoreToken = FALSE;
  CurrentTokenPtr = NULL;
}