#include <stdio.h>
#include <string.h>
#include "crsv.h"

/********************************************************************
 *  This file contains all the functions used to process, store, or
 *   retreive deftemplate information.
 ********************************************************************
 */

/* ===========  Functions defined Externally but used here  =========== */

/* -----------------------------
 *  From the file: CRSVMEM.C
 * -----------------------------
 */
 
extern DT_PTR     AllocDeftemplate();
extern void       FreeDeftemplate();
extern TN_PTR     alloc_token_node();
extern DS_PTR     AllocDefslot();
extern FIELD_PTR  AllocField();
extern TK_PTR     AllocToken();
extern void       FreeToken();
extern char      *GenAlloc();
extern void       GenFree();

/* -----------------------------
 *  From the file: CRSVTKN.C
 * -----------------------------
 */

extern TK_PTR      GetTokenPtr();
extern void        UnGetTokenPtr();

/* -----------------------------
 *  From the file: CRSVHASH.C
 * -----------------------------
 */

extern HASH *AddSymbol();

/* -----------------------------
 *  From the file: CRSVPRNT.C
 * -----------------------------
 */

extern void     ErrorMessage();
extern void     SendMessage();

/* -----------------------------
 *  From the file: CRSVPROC.C
 * -----------------------------
 */

extern RELATION *FindRelation();

/* -----------------------------
 *  From the file: CRSVMISC.C
 * -----------------------------
 */

extern void     QueueToken();
extern TK_PTR   DequeueToken();

/* -----------------------------
 *  From the file: CRSVREL.C
 * -----------------------------
 */

extern int        ProcessDefrelation();

/* -----------------------------
 *  From the system
 * -----------------------------
 */

extern double atof();


/* ===========  Functions defined here for Global use  ================ */

int      ProcessDeftemplate();
int      CountTemplale();
char     *ReturnTokenType();
DT_PTR   FindDeftemplate();
DT_PTR   CreateDeftemplate();
DT_PTR   InsertDeftemplate();
DS_PTR   FindDefslotByPosition();
DS_PTR   FindDefslot();


/* ===========  Functions defined here for internal use  ============== */

static void       FreeTokensInSavedList();
static DS_PTR     CreateDefslot();
static void       AssignSlotPosition ();
static void       QueueVarList ();
static int        ProcessDefaultsInDeftemplate();
static int        ProcessDefTmpAttribute();
static int        CheckDefaultTypes();
static int        CheckDefault();
static int        CheckDefaultInt();
static int        CheckDefaultFloat();
static int        CheckDefaultValues();
static int        CheckDefaultRange();


/* ===========  Variables defined Externally but used here  =========== */

/* -----------------------------
 *  From the file: CRSV.C
 * -----------------------------
 */

extern int          VERBOSE;              /* Verbose printout flag */
extern char        *CurrentObjectName;         /* Name of current object   */
extern FILE        *CurrentFile;             /* current file pointer  */
extern char        *CurrentFileName;        /* Name of the current file */
extern DT_PTR       DeftemplateHead;      /* First Element in Deftemplate list */
extern RELATION    *RelationHead;          /* First element in relation list */
extern DR_PTR      *DefrelationHead;       /* First element in defrelation list */

/* -----------------------------
 *  From the file: CRSVREL.C
 * -----------------------------
 */

extern int RelationCreatedBy;             /* Source of the defrelation     */

/* ===========  Variables defined here for Global use  ================ */


/* ===========  Variables defined here for internal use  ============== */

static TN_PTR TokensInDefrelation = NULL;

/* ======================================================================= */
/*                  FUNCTIONS TO PROCESS DEFTEMPLATES                      */
/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : ProcessDeftemplate()
 *
 *  PURPOSE: This function reads and stores in the
 *           appropriate data structures all of the
 *           information defined in a deftemplate
 *           construct.
 *
 *  INPUTS:  None.
 *
 *  RETURNS: An integer, -99 (RESTART) if it got
 *           hopelessly confused while trying to
 *           read the deftemplate, a -1 if it found
 *           some kind of syntax error, otherwise, a
 *           1 (OK).
 *
 *  NOTES:   Process_deftemplate acts like a preprocessor.  It reads
 *           the tokens which comprise the deftemplate, creates the
 *           deftemplate structure containing template specific data,
 *           and constructs a defrelation, which is processed by
 *           ProcessDefrelation, for the data shared by deftemplate
 *           and defrelation.
 * ------------------------------------------------------------------
 */

#define AddTokenToDefrel(x) QueueToken(&TokensInDefrelation, x)

int ProcessDeftemplate()
{
   int rtn;
   DT_PTR cur_def_tmp = NULL;
   TK_PTR tknstruct;
   REL_PTR temp_rel;
   int found_multi_field = NO;
   TN_PTR WordsList = NULL,StringsList = NULL,
          IntegersList = NULL,FloatsList = NULL,
          InstancesList = NULL,ValuesList = NULL;
   double min,max;
   short set_max;
   short set_min;

   /*============================*/
   /* Check for deftemplate name */
   /*============================*/

   tknstruct = GetTokenPtr (CurrentFile);
   
   IF(tknstruct->token NEQ WORD) THEN
      IF (VERBOSE IS_ON) THEN
        ErrorMessage(ERROR, "Expected a name following deftemplate keyword, ");
        sprintf(MsgBuf,
           "found <%.40s> of type %.40s.",
           tknstruct->tknword,ReturnTokenType(tknstruct->token));
        SendMessage(MsgBuf,NO);
      ELSE
        ErrorMessage(ERROR, "Expected a name following deftemplate keyword");
      END_IF
      FreeToken (tknstruct);
      return(RESTART);
   END_IF
   CurrentObjectName  = tknstruct->tknword;

   IF (VERBOSE IS_ON) THEN
      sprintf(MsgBuf,"\n\nProcessing deftemplate: %.40s",CurrentObjectName);
      SendMessage(MsgBuf,NO);
   END_IF

   /*==========================================================*/
   /* Check if template has been previously used as a relation */
   /*==========================================================*/

   IF ((temp_rel = FindRelation(RelationHead, CurrentObjectName)) NEQ NULL)THEN
      if(temp_rel->is_template IS_NO){
      sprintf(MsgBuf,
        "Template <%.40s> being redefined, previously defined as a relation.",
         CurrentObjectName);
      ErrorMessage(ERROR, MsgBuf);
      FreeToken (tknstruct);
      return(RESTART);}
   END_IF
   
   IF(FindDeftemplate(DeftemplateHead,CurrentObjectName) != NULL) THEN
     sprintf(MsgBuf,
        "Template <%.40s> being redefined, second definition is ignored\n",
         CurrentObjectName);
      ErrorMessage(WARNING, MsgBuf);
      FreeToken (tknstruct);
      return(RESTART);
   END_IF
    

   /*==========================================*/
   /* Add the template name to the defrelation */
   /*==========================================*/

   AddTokenToDefrel (tknstruct);

   /*======================================*/
   /* Get pointer to deftemplate structure */
   /*======================================*/

   
    cur_def_tmp = CreateDeftemplate(CurrentObjectName);

   /*===================================*/
   /* Begin processing deftemplate info */
   /*===================================*/

   tknstruct = GetTokenPtr(CurrentFile);
   IF(tknstruct->token EQ STRING) THEN
     tknstruct = GetTokenPtr (CurrentFile);
   END_IF
   /*===================================*/
   /* Read and process deftemplate info */
   /*===================================*/

   WHILE (tknstruct->token NEQ RPAREN) DO
      TK_PTR keyword, field_name;
      int tkn_type = 0;

      /*======================================*/
      /* Find opening paren, and get key word */
      /*======================================*/

      IF(tknstruct->token NEQ LPAREN) THEN
        ErrorMessage(ERROR,"Missing a closed parenthesis!\n");
        FreeTokenAndDefTmp(tknstruct,cur_def_tmp);
        return(RESTART);
      END_IF

      keyword = GetTokenPtr (CurrentFile);

      /*============================================*/
      /* Process "field" and "multi-field" keywords */
      /*============================================*/

      IF((keyword->token EQ WORD) AND
         ((strcmp("field", keyword->tknword) IS_NIL) OR
          (strcmp ("multi-field", keyword->tknword) IS_NIL) OR
          (strcmp ("multifield",keyword->tknword) IS_NIL))) THEN
         int field_is_multi_valued = keyword->tknword[0] EQ 'm';

         /*===============================*/
         /* Process "multi-field" keyword */
         /*===============================*/

         IF (field_is_multi_valued) THEN
            keyword->hashword = AddSymbol("field");
            keyword->tknword = keyword->hashword->contents;

            IF (found_multi_field) THEN
               ErrorMessage (ERROR,
                  "Two multi-field slot definitions  ");
               FreeTokenAndDefTmp(keyword,cur_def_tmp);
               return(RESTART);
            ELSE
               found_multi_field = YES;
            END_IF
         END_IF

         /*====================*/
         /* Get the field name */
         /*====================*/

         field_name = GetTokenPtr (CurrentFile);

         IF (field_name->token NEQ WORD) THEN
            IF (VERBOSE IS_ON) THEN
              sprintf(MsgBuf,
                "Unrecognized slot name <%.40s>, Expected a SYMBOL.",
                 field_name->tknword);
            ELSE_IF (CHECK_STYLE IS_ON)
              sprintf (MsgBuf,
                 "Check appropriate syntax for deftemplate construct\n");
            END_IF
            ErrorMessage(ERROR, MsgBuf);
            FreeTokenAndDefTmp(field_name,cur_def_tmp);
            return(RESTART);
         ELSE_IF (FindDefslot (cur_def_tmp, field_name->tknword) NEQ NULL)
            sprintf(MsgBuf,"Multiple occurences of slot %s\n",field_name->tknword);
            ErrorMessage (ERROR,MsgBuf);
            FreeTokenAndDefTmp(field_name,cur_def_tmp);
            return(RESTART);
         ELSE
            DS_PTR cur_def_slot = CreateDefslot (cur_def_tmp,
               field_name->tknword);
            TN_PTR default_cur;

            /*===========================================================*/
            /* Place a left paren., "field" keyword, and field name into */
            /* the defrelation.                                          */
            /*===========================================================*/

            AddTokenToDefrel (tknstruct);
            AddTokenToDefrel (keyword);
            AddTokenToDefrel (field_name);

            /*=========================================*/
            /* Override default max/min element limits */
            /*=========================================*/

            IF (field_is_multi_valued IS_NO) THEN
               cur_def_slot->max_elements = 1;
               cur_def_slot->min_elements = 1;
            END_IF

            /*===================================*/
            /* Read what should be a left paren. */
            /*===================================*/

            tknstruct = GetTokenPtr(CurrentFile);

            /*============================*/
            /* Read and process slot info */
            /*============================*/

            tkn_type = 0;
            set_min = FALSE;
            set_max = FALSE;

            WHILE ((tknstruct->token NEQ RPAREN) && (tknstruct->token NEQ STOP)) DO

               /*============================================================*/
               /* Find opening paren, and get single field attribute keyword */
               /*============================================================*/

               IF(tknstruct->token NEQ LPAREN) THEN
                  ErrorMessage(ERROR,"Missing a closed parenthesis!\n");
                  FreeTokenAndDefTmp(tknstruct,cur_def_tmp);
                  return(RESTART);
               END_IF

               /*=================================*/
               /* Read the attribute keyword name */
               /*=================================*/

               keyword = GetTokenPtr (CurrentFile);

               /*===============================*/
               /* Process the default attribute */
               /*===============================*/

               IF((keyword->token EQ WORD) AND
                  (strcmp("default", keyword->tknword) IS_NIL)) THEN

                  FreeToken (tknstruct);
                  FreeToken (keyword);
                  if(ProcessDefaultsInDeftemplate(cur_def_slot,field_is_multi_valued) != OK)
                    {
                      FreeTokensInSavedList (&TokensInDefrelation);
                      FreeDeftemplate(cur_def_tmp);
                      return(RESTART);
                    }
               /*=====================================================*/
               /* Process multi-value keywords when multi-value field */
               /*=====================================================*/

               ELSE_IF((keyword->token EQ WORD) AND field_is_multi_valued AND
                  (strcmp("min-number-of-elements", keyword->tknword) IS_NIL OR
                  strcmp("max-number-of-elements", keyword->tknword) IS_NIL))
                  int *num_elements = (keyword->tknword[1] EQ 'a' ?
                     &cur_def_slot->max_elements :
                     &cur_def_slot->min_elements);

                  /*=========================================================*/
                  /* Read and discard tokens until the span value is reached */
                  /*=========================================================*/

                  FreeToken (tknstruct);
                  FreeToken (keyword);
                  tknstruct = GetTokenPtr (CurrentFile);

                  /*=======================================*/
                  /* Process the span keyword of ?VARIABLE */
                  /*=======================================*/

                  IF((tknstruct->token EQ BWORD) AND
                     strcmp("VARIABLE", tknstruct->tknword) IS_NIL) THEN
                     *num_elements = (keyword->tknword[1] EQ 'a' ? 9900 : 0);

                  /*=============================*/
                  /* Process a non-numeric value */
                  /*=============================*/

                  ELSE_IF(tknstruct->token NEQ INTEGER)
                     IF (VERBOSE IS_ON) THEN
                        ErrorMessage(ERROR, 
                        "Expected an integer or ?VARIABLE for min/max-number-of-elements attribute,\n ");
                        sprintf (MsgBuf, "      but found a %.40s",
                           ReturnTokenType(tknstruct->token));
                        SendMessage(MsgBuf,NO);
                     ELSE
                        ErrorMessage(ERROR,
                           "Expected an integer or VARIABLE for min/max-number-of-elements attribute,\n ");
                     END_IF
                     FreeTokenAndDefTmp(tknstruct,cur_def_tmp);
                     return(RESTART);
                  /*=========================*/
                  /* Process a numeric value */
                  /*=========================*/

                 ELSE
                     *num_elements = tknstruct->intnumber;
                     FreeToken(tknstruct);
                     tknstruct = GetTokenPtr(CurrentFile);
                     if(tknstruct->token != RPAREN)
                       {
                         ErrorMessage(ERROR,
                           "More than one values are not allowed in the min/max-number-of-elements attribute\n");
                         FreeTokenAndDefTmp(tknstruct,cur_def_tmp);
                         return(RESTART);
                       }
                     FreeToken(tknstruct);
                 END_IF

               /*========================================================*/
               /* Process an attribute keyword that is not recognized by */
               /* deftemplate                                            */
               /*========================================================*/


               ELSE

                  /*======================================================*/
                  /* Copy all tokens for the attribute to the defrelation */
                  /*======================================================*/

                  AddTokenToDefrel (tknstruct);
                  AddTokenToDefrel (keyword);

                  do {
                     tknstruct = GetTokenPtr (CurrentFile);
                     IF (ProcessDefTmpAttribute(keyword,tknstruct,&min,&max,
                           &set_min,&set_max,&tkn_type,&IntegersList,
                           &FloatsList,&InstancesList,&WordsList,&StringsList,&ValuesList) == RESTART) THEN
                        FreeTokenAndDefTmp(tknstruct,cur_def_tmp);
                        return(RESTART);
                     END_IF
                     AddTokenToDefrel (tknstruct);
                  } while (tknstruct->token NEQ RPAREN);
               END_IF

               tknstruct = GetTokenPtr (CurrentFile);
            END_WHILE

            AddTokenToDefrel (tknstruct);

           /*================================================*/
           /* Test default Values against user defined types */
           /*================================================*/

           default_cur = cur_def_slot->defaults;

           IF(tkn_type EQ 128) THEN
             IF (VERBOSE IS_ON) THEN
               sprintf(MsgBuf,
                  "No type specified for the type attribute in slot <%.40s>.",
                  cur_def_slot->name);
               ErrorMessage(ERROR,MsgBuf);
             ELSE
               ErrorMessage(ERROR,"No type specified for the type attribute.");
             END_IF
             FreeTokensInSavedList (&TokensInDefrelation);
             FreeDeftemplate(cur_def_tmp);
             return(RESTART);
           END_IF

           WHILE(default_cur) DO     /* OTHER THAN ?NONE */

             if(CheckDeftemplateDefault(default_cur,tkn_type,ValuesList,WordsList,StringsList,
                            InstancesList,IntegersList,FloatsList,min,max,set_min,set_max) != OK)
               {
                 FreeTokensInSavedList (&TokensInDefrelation);
                 FreeDeftemplate(cur_def_tmp);
                 return(RESTART);
               }
             default_cur = default_cur->next;

           END_WHILE

         END_IF


      /*=========================================================*/
      /* Process a keyword that is not recognized by deftemplate */
      /*=========================================================*/

      ELSE
         sprintf(MsgBuf,
            "Unrecognized attribute: %.40s inside a deftemplate",
            tknstruct->tknword);
         ErrorMessage(ERROR, MsgBuf);
         FreeTokensInSavedList (&TokensInDefrelation);
         FreeDeftemplate(cur_def_tmp);
         return(RESTART);
      END_IF

      tknstruct = GetTokenPtr (CurrentFile);
   END_WHILE

   /*===========================================*/
   /* Add the closing paren. to the defrelation */
   /*===========================================*/

   AddTokenToDefrel (tknstruct);

   /*==================================================================*/
   /* Scan slot descriptions to assign field position (the multi-field */
   /* slot is placed last)                                             */
   /*==================================================================*/

   AssignSlotPosition (cur_def_tmp);

   /*===========================================================*/
   /* Substitute field positions for field names in the implied */
   /* defrelation.                                              */
   /*===========================================================*/

   AssignFieldName(TokensInDefrelation,cur_def_tmp);

   /*======================================================*/
   /* Make the defrelation tokens be read next by GetToken */
   /*======================================================*/

   (void)SetGetTokenQueue (TokensInDefrelation);
   TokensInDefrelation = NULL;

   /*=========================================================*/
   /* Process the defrelation with its source set to template */
   /*=========================================================*/

   RelationCreatedBy = TEMPLATE_MADE;
   rtn = ProcessDefrelation();
   if(rtn != OK)
    FreeDeftemplate(cur_def_tmp);
   else
    DeftemplateHead = InsertDeftemplate(DeftemplateHead,cur_def_tmp);
   RelationCreatedBy = USER_MADE;

  return(rtn);
}
/* --------------------------------------------------------------
 * ProcessDefaultsInDeftemplate
 * This function process the default attribute in a deftemplate
 * --------------------------------------------------------------
 */

static int ProcessDefaultsInDeftemplate(CurDefslot,multiple)
DS_PTR CurDefslot;
int multiple;
{

  TK_PTR tknstruct;
  /*=============================================*/
  /* Handle error of multiple default attributes */
  /*=============================================*/

   IF (CurDefslot->defaults) THEN
     ErrorMessage (ERROR,"The default attribute has already been defined\n");
     return(RESTART);
   ELSE
  /*=====================================*/
  /* Read the first value in the default */
  /*=====================================*/

     tknstruct = GetTokenPtr (CurrentFile);

   /*=========================================*/
   /* Process the default value keyword ?NONE */
   /*=========================================*/

     IF (tknstruct->token EQ BWORD AND
         strcmp (tknstruct->tknword, "NONE") IS_NIL) THEN
        FreeToken (tknstruct);
        tknstruct = GetTokenPtr (CurrentFile);

   /*======================================*/
   /* Expect no values after ?NONE keyword */
   /*======================================*/

        IF (tknstruct->token NEQ RPAREN) THEN
          ErrorMessage (ERROR,"Expected \')\' after ?NONE default value");
          FreeToken(tknstruct);
          return(RESTART);
        END_IF
     ELSE_IF(tknstruct->token == RPAREN)
       IF(!multiple) THEN
          ErrorMessage(ERROR,"default for single field must have at least one value\n");
          FreeToken(tknstruct);
          return(RESTART);
       END_IF
  /*==============================*/
  /* Process default value tokens */
  /*==============================*/
                     
     ELSE
        do
           {
              IF(CheckDefaultTypes(tknstruct,CurDefslot,multiple) == RESTART)THEN
                 FreeToken(tknstruct);
                 return(RESTART);
              ELSE
                 QueueToken (&(CurDefslot->defaults), tknstruct);
              END_IF
              tknstruct = GetTokenPtr (CurrentFile);
           }while((multiple)&&(tknstruct->token NEQ RPAREN)
                                 &&(tknstruct->token NEQ STOP));

  /*===============================================*/
  /* Terminating loop without finding right paren. */
  /* indicates that multiple values were provided  */
  /* for a single value slot.                      */
  /*===============================================*/

         IF ((tknstruct->token NEQ RPAREN)&&(tknstruct->token NEQ STOP)) THEN
            ErrorMessage (ERROR,"Multiple default values not permitted in a ");
            SendMessage ("single value slot.",YES);
            FreeToken (tknstruct);
            return(RESTART);
         END_IF
         FreeToken (tknstruct);
     END_IF
     
   END_IF
   return(OK);             
 }

/* -------------------------------------------------------------
 * ProcessDefTmpAttribute
 * This function processes the attributes of a deftemplate and
 * puts the allowed values in the appropriate queues.
 * -------------------------------------------------------------
 */
 
static int ProcessDefTmpAttribute(keyword,tknstruct,min,max,set_min,set_max,tkn_type,IntegersList,
                           FloatsList,InstancesList,WordsList,StringsList,ValuesList)
  TK_PTR keyword,tknstruct;
  TN_PTR *WordsList,*StringsList,
          *IntegersList,*FloatsList,
          *InstancesList,*ValuesList;
  double *min,*max;
  short  *set_min,*set_max;
  int *tkn_type;
  {

    IF ((keyword->token EQ WORD) AND
        (strcmp("type",keyword->tknword) IS_NIL)) THEN
        *tkn_type |= 128;
        if(SetTypeFlag(tkn_type,tknstruct) != OK)
          return(RESTART);
    ELSE_IF((keyword->token EQ WORD) AND
            (strcmp("range",keyword->tknword) IS_NIL))
       IF(tknstruct->token == INTEGER)THEN
         if(*set_min == FALSE) 
          {
            *min = (double)tknstruct->intnumber;
            *set_min = TRUE;
          }
         else if (*set_max == FALSE)
          {
            *max = (double)tknstruct->intnumber;
            *set_max = TRUE;
          }
       ELSE_IF(tknstruct->token == FLOAT)
         if(*set_min == FALSE) 
          {
            *min = tknstruct->floatnumber;
            *set_min = TRUE;
          }
         else if (*set_max == FALSE)
          {
            *max = tknstruct->floatnumber;
            *set_max = TRUE;
          }
       END_IF
     ELSE_IF((keyword->token EQ WORD) AND
             (strcmp("allowed-words",keyword->tknword) IS_NIL))
         QueueToken(WordsList,tknstruct);
     ELSE_IF((keyword->token EQ WORD) AND
             (strcmp("allowed-strings",keyword->tknword) IS_NIL))
         QueueToken(StringsList,tknstruct);
     ELSE_IF((keyword->token EQ WORD) AND
             (strcmp("allowed-instances",keyword->tknword) IS_NIL))
         QueueToken(InstancesList,tknstruct);
     ELSE_IF((keyword->token EQ WORD) AND
             (strcmp("allowed-integers",keyword->tknword) IS_NIL))
         QueueToken(IntegersList,tknstruct);
     ELSE_IF((keyword->token EQ WORD) AND
             (strcmp("allowed-floats",keyword->tknword) IS_NIL))
         QueueToken(FloatsList,tknstruct);
     ELSE_IF((keyword->token EQ WORD) AND
             (strcmp("allowed-numbers",keyword->tknword) IS_NIL))
        if(tknstruct->token == INTEGER)
          QueueToken(IntegersList,tknstruct);
        else if(tknstruct->token == FLOAT)
          QueueToken(FloatsList,tknstruct);
     ELSE_IF((keyword->token EQ WORD) AND
        (strcmp("allowed-values",keyword->tknword) IS_NIL))
         QueueToken(ValuesList,tknstruct);
     END_IF
     return(OK);
  }

/* -------------------------------------------------------------
 * CheckDefaultTypes
 * This function processes the types of the default values
 * -------------------------------------------------------------
 */

static int CheckDefaultTypes(tknstruct,cur_def_slot,multiple)
TK_PTR tknstruct;
DS_PTR cur_def_slot;
int multiple;
{
   
  IF((tknstruct->token EQ BWORD) AND
      (strcmp(tknstruct->tknword,"NONE") == 0) AND
      (cur_def_slot->defaults != NULL)) THEN
      ErrorMessage(ERROR,"?NONE  cannot be used with other default values\n");
      return(RESTART);
   ELSE_IF ((tknstruct->token NEQ WORD) AND
      (tknstruct->token NEQ INTEGER) AND
      (tknstruct->token NEQ FLOAT) AND
      (tknstruct->token NEQ STRING))
      ErrorMessage(ERROR,"Default values must be primitive data types");
      return(RESTART);
   END_IF
   return(OK);
 }
/*
 * ------------------------------------------------------------
 * CheckDeftemplateDefault
 *  Check if the types and values of the defaults meeting the
 *  allowed types and allowed values lists
 * ------------------------------------------------------------
 */

int CheckDeftemplateDefault(default_cur,tkn_type,ValuesList,WordsList,StringsList,
                            InstancesList,IntegersList,FloatsList,min,max,set_min,set_max)
TN_PTR default_cur,ValuesList,WordsList,StringsList,
                            InstancesList,IntegersList,FloatsList;
int tkn_type;
short set_min,set_max;
double min,max;
{
     IF((((default_cur->token->token EQ WORD) AND
        ((tkn_type & 1) NEQ 1)) OR
        ((default_cur->token->token EQ STRING) AND
        ((tkn_type & 2) NEQ 2)) OR
        ((default_cur->token->token EQ INSTANCE) AND
        ((tkn_type & 4) NEQ 4)) OR
        ((default_cur->token->token EQ INTEGER) AND
        ((tkn_type & 8) NEQ 8)) OR
        ((default_cur->token->token EQ FLOAT) AND
        ((tkn_type & 16) NEQ 16))) AND 
        (((tkn_type & 64) NEQ 64) AND
         (tkn_type NEQ 0))) THEN
            sprintf(MsgBuf,"Default value <%.40s> not of proper type.",
                       default_cur->token->tknword);
            ErrorMessage(ERROR,MsgBuf);
            return(RESTART);
    ELSE_IF(ValuesList != NULL)
       if(CheckDefaultValues(default_cur,ValuesList) != OK)
         return(RESTART);
    ELSE_IF((default_cur->token->token EQ WORD) AND
            (WordsList != NULL))
       if(CheckDefault(default_cur,WordsList) != OK)
         return(RESTART);
    ELSE_IF((default_cur->token->token EQ STRING) AND
            (StringsList != NULL))
       if(CheckDefault(default_cur,StringsList) != OK)
         return(RESTART);
    ELSE_IF((default_cur->token->token EQ INSTANCE) AND
            (InstancesList != NULL))
       if(CheckDefault(default_cur,InstancesList) != OK)
          return(RESTART);
    ELSE_IF(default_cur->token->token EQ INTEGER)
       if(IntegersList != NULL)
         {
           if( CheckDefaultInt(default_cur,IntegersList) != OK)
             return(RESTART);
         }
       else if (set_min OR set_max){
          if(CheckDefaultRange(default_cur,min,max,set_min,set_max) != OK)
                   return(RESTART);}
    ELSE_IF(default_cur->token->token EQ FLOAT)
       if(FloatsList != NULL){
           if(CheckDefaultFloat(default_cur,FloatsList) != OK)
               return(RESTART);}
       else if (set_min OR set_max){
           if(CheckDefaultRange(default_cur,min,max,set_min,set_max) != OK)
                   return(RESTART);}
   END_IF
   return(OK);
}


/*
 * ------------------------------------------------
 * AssignFieldName
 *  This function assigns the field id for each field,
 *  after the deftemplate has been converted to 
 *  defrelation form
 * ------------------------------------------------
 */
 
 AssignFieldName(TokensInDefrelation,cur_def_tmp)
 TN_PTR TokensInDefrelation;
 DT_PTR cur_def_tmp;
 {
      int paren_level;
      TN_PTR tknnode = TokensInDefrelation->next;
      TK_PTR field_id;DS_PTR def_slot;

      /*=====================================================*/
      /* WHILE the closing right paren. has not been reached */
      /*=====================================================*/

      WHILE (tknnode->token->token NEQ RPAREN) DO
         paren_level = 1;

         tknnode = tknnode->next;

         /*==================================*/
         /* Process field id to field number */
         /*==================================*/

         IF (tknnode->token->token EQ WORD AND
            strcmp (tknnode->token->tknword, "field") IS_NIL) THEN
            field_id = tknnode->next->token;

            IF (field_id->token EQ WORD) THEN
               def_slot = FindDefslot (cur_def_tmp,
                  field_id->tknword);

               field_id->token = INTEGER;
               field_id->intnumber = def_slot->position;
            END_IF
         END_IF

         /*=====================================*/
         /* Chain forward to the closing paren. */
         /*=====================================*/

         WHILE (paren_level NEQ 0) DO
            IF (tknnode->token->token EQ LPAREN) THEN
               paren_level++;
            ELSE_IF (tknnode->token->token EQ RPAREN)
               paren_level--;
            END_IF

            tknnode = tknnode->next;
         DONE
      DONE
   }

/*================================================*/
/* SetTypeFlag:                                 */         
/* Set the possible types of slot values in the   */
/* variable tkn_type through a bit-wise operation */
/*================================================*/

int SetTypeFlag(tkn_type,tknstruct)
int *tkn_type;
TK_PTR tknstruct;
{
       IF(strcmp("WORD",tknstruct->tknword) IS_NIL) THEN
          if(((*tkn_type) & 1) == 1){
            ErrorMessage(ERROR,
            "WORD/SYMBOL must not be appeared in conjunction with another SYMBOL/WORD or ?VARIABLE\n");
            return(ERROR);}
          else
           *tkn_type |= 1;
       ELSE_IF(strcmp("SYMBOL",tknstruct->tknword) IS_NIL)
          if(((*tkn_type) & 1) == 1){
            ErrorMessage(ERROR,
            "WORD/SYMBOL must not be appeared in conjunction with another SYMBOL/WORD or ?VARIABLE\n");
             return(ERROR);}
          else
            *tkn_type |= 1;
       ELSE_IF(strcmp("STRING",tknstruct->tknword) IS_NIL)
          if(((*tkn_type) & 2) == 2){
            ErrorMessage(ERROR,
            "STRING must not be appeared in conjunction with another STRING or ?VARIABLE\n");
            return(ERROR);}
          else
            *tkn_type |= 2;
       ELSE_IF(strcmp("LEXEME",tknstruct->tknword) IS_NIL)
          if((((*tkn_type) & 1) == 1) OR (((*tkn_type) & 2) == 2)){
            ErrorMessage(ERROR,
            "LEXEME must not be appeared in conjunction with another LEXEME ");
            SendMessage("\n       or SYMBOL(WORD)/STRING or ?VARIABLE\n");
             return(ERROR);}
           else
            *tkn_type |= 3;
       ELSE_IF(strcmp("INSTANCE",tknstruct->tknword) IS_NIL)
          if(((*tkn_type) & 4) == 4){
            ErrorMessage(ERROR,
            "INSTANCE must not be appeared in conjunction with another INSTANCE or ?VARIABLE\n");
            return(ERROR);}
          else
          *tkn_type |= 4;
       ELSE_IF(strcmp("INTEGER",tknstruct->tknword) IS_NIL)
          if(((*tkn_type) & 8) == 8){
            ErrorMessage(ERROR,
            "INTEGER must not be appeared in conjunction with another INTEGER or ?VARIABLE\n");
            return(ERROR);}
          else
            *tkn_type |= 8;
       ELSE_IF(strcmp("FLOAT",tknstruct->tknword) IS_NIL)
          if(((*tkn_type) & 16) == 16){
            ErrorMessage(ERROR,"FLOAT must not be appeared in conjunction with another FLOAT or ?VARIABLE\n");
            return(ERROR);}
          else
          *tkn_type |= 16;
       ELSE_IF(strcmp("NUMBER",tknstruct->tknword) IS_NIL)
          if((((*tkn_type) & 8) == 8) OR (((*tkn_type) & 16) == 16)){
           ErrorMessage(ERROR,
           "NUMBER must not be appeared in conjunction with another NUMBER or FLOAT or INTEGER or ?VARIABLE\n");
           return(ERROR);}
          else
           *tkn_type |= 24;
       ELSE_IF(strcmp("EXTERNAL-ADDRESS",tknstruct->tknword) IS_NIL)
         if(((*tkn_type) & 32) == 32){
            ErrorMessage(ERROR,"EXTERNAL-ADDRESS must not be appeared in conjunction with another EXTERNAL-ADDRESS or ?VARIABLE\n");
            return(ERROR);}
          else
          *tkn_type |= 32;
       ELSE_IF((tknstruct->token == BWORD) &&
               (strcmp("VARIABLE",tknstruct->tknword) IS_NIL))
          IF((((*tkn_type)&1) == 1)||(((*tkn_type)&2) == 2)||
             (((*tkn_type)&4) == 4)||(((*tkn_type)&8) == 8)||
             (((*tkn_type)&16) == 16)||(((*tkn_type)&32) == 32)) THEN
                *tkn_type |= 64;
                sprintf(MsgBuf,
                    "?VARIABLE must be by itself in the type attribute\n");
                ErrorMessage(ERROR,MsgBuf);
                return(ERROR);
           ELSE
                *tkn_type |= 63;
           END_IF
       ELSE
           IF(tknstruct->token NEQ RPAREN) THEN
                *tkn_type |= 64;
                 IF (VERBOSE) THEN
                   sprintf(MsgBuf,
               "<%.40s> is not a valid type. Expected <WORD,SYMBOL,STRING,INSTANCE,INTEGER or FLOAT|?VARIABLE>.",
                                     tknstruct->tknword);
                  ELSE
                    sprintf(MsgBuf,"<%.40s> is not a valid type.",
                                     tknstruct->tknword);
                  END_IF
                  ErrorMessage(ERROR,MsgBuf);
                  return(ERROR);
           END_IF
       END_IF
       return(OK);
}
/* ------------------------------------------------- *
 * CheckDefaultInt                                   *
 *   This function check a default value against     *
 * the allowed-integers list.                        *
 * ------------------------------------------------- */

static int CheckDefaultInt(default_val,list)
TN_PTR list, default_val;
{
    TN_PTR list_ptr;
    
    if (list == NULL)
      return(OK);
    list_ptr = list;
    while(list_ptr != NULL)
       {
        if(list_ptr->token->intnumber == default_val->token->intnumber)
           break;
        list_ptr = list_ptr->next;
       }
    if(list_ptr == NULL)
      {
        ErrorMessage(ERROR,"Default value is not in the allowed List\n");
        return(ERROR);
      }
    return(OK);
}
/* -------------------------------------------------- *
 * CheckDefaultFloat                                  *
 *   This function check a default value against      *
 * the allowed-floats list.                           *
 * -------------------------------------------------- */

static int CheckDefaultFloat(default_val,list)
TN_PTR list, default_val;
{
    TN_PTR list_ptr;
    
    if (list == NULL)
      return(OK);
    list_ptr = list;
    while(list_ptr != NULL)
       {
        if(list_ptr->token->floatnumber == default_val->token->floatnumber)
           break;
        list_ptr = list_ptr->next;
       }
    if(list_ptr == NULL)
      {
        ErrorMessage(ERROR,"Default value is not in the allowed List\n");
        return(ERROR);
      }
    return(OK);
}

/* ------------------------------------------------- *
 * CheckDefault                                      *
 *   This function check a default value against     *
 * the allowed-strings and allowed-words lists.      *
 * ------------------------------------------------- */
static int CheckDefault(default_val,list)
TN_PTR list, default_val;
 {
    TN_PTR list_ptr;
    
    if (list == NULL)
      return(OK);
    list_ptr = list;
    while(list_ptr != NULL)
       {
        if(strcmp(list_ptr->token->tknword,default_val->token->tknword) == 0)
          break;
        list_ptr = list_ptr->next;
       }
    if(list_ptr == NULL)
      {
        ErrorMessage(ERROR,"Default value is not in the allowed List\n");
        return(ERROR);
      }
    return(OK);  
 }

/* ------------------------------------------------- *
 * CheckDefaultValues                                *
 *   This function check a default value against     *
 * the allowed-values list.                          *
 * ------------------------------------------------- */

static int CheckDefaultValues(default_val,list)
TN_PTR list, default_val;
 {
    TN_PTR list_ptr;
    short found = FALSE;
    
    if (list == NULL)
      return(OK);
    list_ptr = list;
    while((list_ptr != NULL) && (!found))
    {
      switch (default_val->token->token)
       {
          case INTEGER:if((list_ptr->token->token == INTEGER) && 
                     (list_ptr->token->intnumber == default_val->token->intnumber))
                      found  = TRUE;
                  break;
          case FLOAT  :if((list_ptr->token->token == FLOAT) && (list_ptr->token->floatnumber ==
                     default_val->token->floatnumber))
                      found  = TRUE;
                  break;
          case WORD   :if((list_ptr->token->token == WORD) && 
                    (strcmp(list_ptr->token->tknword,default_val->token->tknword) == 0))
                      found  = TRUE;
                  break;
          case STRING :if((list_ptr->token->token == STRING) && 
                    (strcmp(list_ptr->token->tknword,default_val->token->tknword) == 0))
                      found  = TRUE;
                  break;
          case INSTANCE:if((list_ptr->token->token == INSTANCE) && 
                    (strcmp(list_ptr->token->tknword,default_val->token->tknword) == 0))
                      found  = TRUE;
                  break;
          default:
                 ErrorMessage(ERROR,"Unrecognized token in default attribute\n");
                 return(ERROR);
                   
       
       }
      list_ptr = list_ptr->next;
    }
    if(!found){
       ErrorMessage(ERROR,"Default is not in the allowed-values attribute\n");
       return(ERROR);
       }
    return(OK);
 }
 
 
/* ------------------------------------------------- *
 * CheckDefaultRange                                 *
 *   This function check a default value against     *
 * the range values.                                 *
 * ------------------------------------------------- */
static int CheckDefaultRange(default_val,min,max,set_min,set_max)
TN_PTR default_val;
double min,max;
short set_min,set_max;
{
  if(default_val->token->token == INTEGER)
   {
     if(((set_min)&&((double)(default_val->token->intnumber) < min)) OR
        ((set_max)&&((double)(default_val->token->intnumber) > max))){
      ErrorMessage(ERROR,"Default is out of range\n");
      return(ERROR);}   
   }
  if(default_val->token->token == FLOAT)
   {
     if(((set_min) && (default_val->token->floatnumber < min)) OR
        ((set_max) &&(default_val->token->floatnumber > max))){
      ErrorMessage(ERROR,"Default is out of range\n"); 
      return(ERROR);}  
   }
  return(OK);
}

/* ======================================================================= */

/* -----------------------------------------------------------------
 *  NAME   : ReturnTokenType()
 *  PURPOSE: Returns the string representation associated with the
 *           token field of the structure token.  This function is
 *           to be used in the error traps especially when VERBOSE
 *           is turned on.
 *
 *  INPUTS : token number
 *
 *  RETURNS: string representation.
 *
 * -----------------------------------------------------------------
 */

char *ReturnTokenType(token)
int token;
{

  switch(token) {
    case LNOT      : return("~");
    case LAND      : return("&");
    case LOR       : return("|");
    case COAMP     : return("COAMP");
    case SEPARATOR : return("=>");
    case BINDER    : return("<-");
    case MULTIPLE  : return("$?");
    case LPAREN    : return("(");
    case RPAREN    : return(")");
    case SINGLE    : return("?");

    case BWORD     : return("?<var>");
    case BWORDS    : return("$?<var>");
    case OPERATOR  : return("OPERATOR");
    case KUNKNOWN  : return("KUNKNOWN");
    case FCALL     : return("FUNCTION CALL");
    case STOP      : return("END OF FILE");
    case PATTERN   : return("PATTERN");

    case WORD      : return("WORD");
    case POINTER   : return("POINTER");
    case INTEGER    : return("INTEGER");
    case FLOAT     : return("FLOAT");
    case STRING    : return("STRING");
  }
  return("Unknown Token Type");
}



/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : AssignSlotPosition()
 *
 *  PURPOSE: Determines the mapping of slot names into field positions.
 *
 *  INPUTS : A pointer to the template.
 *
 *  RETURNS: None
 *
 *  Note: Algorithm assumes that no more than one multi-valued slot
 *        will be found in the template.
 * ------------------------------------------------------------------
 */

static void AssignSlotPosition (template)
 DT_PTR template;
 {
   int position = 1;                        /* Field position */
   DS_PTR mf_ptr = NULL;                /* Pointer to multi-field slot */
   DS_PTR def_slot = template->slots;   /* Pointer to current slot */

   /*===================*/
   /* Process all slots */
   /*===================*/

   WHILE (def_slot) DO

      /*===================================================================*/
      /* Assign position to field slots and record location of multi-field */
      /*===================================================================*/

      IF (def_slot->min_elements NEQ 1 OR def_slot->max_elements NEQ 1) THEN
         mf_ptr = def_slot;
      ELSE
         def_slot->position = ++position;
      END_IF

      def_slot = def_slot->next_slot;
   DONE

   /*===========================================*/
   /* Assign field position to multi-field slot */
   /*===========================================*/

   IF (mf_ptr) THEN
      mf_ptr->position = ++position;
   END_IF

   /*====================================================*/
   /* Record the position of the last (by position) slot */
   /*====================================================*/

   template->max_slot_pos = position;
 }

/* ------------------------------------------------------------------
 *  NAME   : FreeTokensInSavedList()
 *
 *  PURPOSE: This function frees all of the tokens that have been queued
 *           to describe the implied defrelation structure.
 *
 *  INPUTS : None
 *
 *  RETURNS: None
 * ------------------------------------------------------------------
 */

static void FreeTokensInSavedList (list)
TN_PTR *list;
 {
   WHILE (*list) DO
      FreeToken (DequeueToken (list));
   DONE
 }
/* ------------------------------------------------------------------
 *  NAME   :  create_def_tmp()
 *
 *  PURPOSE: This function creates a deftemplate.
 *
 *  INPUTS : a pointer to the name of the deftemplate
 *
 *  RETURNS: poiter pointing to the new deftemplate
 *
 *  Note:    
 * ------------------------------------------------------------------
 */
 
 DT_PTR CreateDeftemplate(name)
 char *name;
 {
    DT_PTR  node;
    HASH *temp_hash;
    
    node        = AllocDeftemplate();
    temp_hash   = AddSymbol(name);
    node->name  = temp_hash->contents;
    temp_hash   = AddSymbol(CurrentFileName);
    node->file  = temp_hash->contents;
    return(node);
 }
 
/* ------------------------------------------------------------------
 *  NAME   :  InsertDeftemplate()
 *
 *  PURPOSE: This function inserts a deftemplate in to the banary tree list.
 *
 *  INPUTS : a pointer pointing to the deftemplate being inserted
 *         : a poiter pointing to the root of the list
 *
 *  RETURNS: new root of the list
 *
 *  Note:    
 * ------------------------------------------------------------------
 */ 
 DT_PTR InsertDeftemplate(list,node)
 DT_PTR list,node;
 {
    DT_PTR DefTempPtr = NULL,DefTempParent = NULL, NewDefTemp;
    int rtn;
   
   if (list == NULL)
     return(node);
   DefTempPtr = list;
   while(DefTempPtr != NULL)
    {
       rtn = strcmp(node->name,DefTempPtr->name);
       if(rtn < 0)
        {
          DefTempParent = DefTempPtr;
          DefTempPtr = DefTempPtr->lft_def_tmp;   
        }
       else if(rtn > 0)
        {
          DefTempParent = DefTempPtr;
          DefTempPtr = DefTempPtr->rht_def_tmp;   
        }
       else{
         ErrorMessage(ERROR,"CRSV error in InsertDefTemplate, Please report this error\n");
         return(list);}
    }
   if(rtn < 0)
      DefTempParent->lft_def_tmp = node;
   else if(rtn > 0)
      DefTempParent->rht_def_tmp = node;
   else
      ErrorMessage(ERROR,"CRSV error in InsertDefTemplate, Please report this error\n");
   return(list);
 }


/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : FindDeftemplate()
 *
 *  PURPOSE: This function searches through a tree of
 *           deftemplates to find a specific name.
 *
 *  INPUTS : A pointer to the tree of deftemplates (DT_PTR)
 *           and a pointer to the name to search for (char *)
 *
 *  RETURNS: If the name is found, it returns a pointer to
 *           that structure. If it isn't found it returns NULL.
 * ------------------------------------------------------------------
 */

DT_PTR FindDeftemplate(node, name)
DT_PTR node;
char *name;
{
   int rtn;

   /*=================*/
   /* Search complete */
   /*=================*/

   IF(node EQ NULL OR (rtn = strcmp (name, node->name)) IS_NIL) THEN
      return(node);

   /*=================*/
   /* Continue search */
   /*=================*/
   ELSE
      return(FindDeftemplate(( rtn < 0 ? node->lft_def_tmp : node->rht_def_tmp),
         name));
   END_IF
}


/* ******************************************  */

int CountTemplales(node)
DT_PTR node;
{
  if(node)
    return(1 + CountTemplales(node->lft_def_tmp) +
           CountTemplales(node->rht_def_tmp));
  else
    return(0);
}


/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : FindDefslot()
 *
 *  PURPOSE: This function searches through a deftemplate's slots to
 *           find a specific name.
 *
 *  INPUTS : A pointer to the deftemplate (DT_PTR)
 *           and a pointer to the name to search for (char *)
 *
 *  RETURNS: If the name is found, it returns a pointer to
 *           that structure. If it isn't found it returns NULL.
 * ------------------------------------------------------------------
 */

DS_PTR FindDefslot(template, name)
   DT_PTR template;
   char *name;
{
   DS_PTR slot = template->slots;

   WHILE (slot NEQ NULL AND strcmp(name, slot->name) NEQ 0) DO
      slot = slot->next_slot;
   DONE;

   return (slot);
}

/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : FindDefslotByPosition()
 *
 *  PURPOSE: This function searches through a deftemplate's slots to
 *           find a specific position.
 *
 *  INPUTS : A pointer to the deftemplate (DT_PTR)
 *           and the position of the slot to search for (int)
 *
 *  RETURNS: If the name is found, it returns a pointer to
 *           that structure. If it isn't found it returns NULL.
 * ------------------------------------------------------------------
 */

DS_PTR FindDefslotByPosition(template, position)
   DT_PTR template;
   int position;
{
   DS_PTR slot = template->slots;

   WHILE (slot NEQ NULL AND (position >= slot->position + slot->max_elements OR
      position < slot->position)) DO
      slot = slot->next_slot;
   DONE;

   return (slot);
}

/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : CreateDefslot()
 *
 *  PURPOSE: This function creates a specific named slot.
 *
 *  INPUTS : A pointer to the deftemplate (DEF_TMP *) whose slot is
 *           being created and a pointer to the name of the new slot.
 *
 *  RETURNS: The new slot structure.
 * ------------------------------------------------------------------
 */

static DS_PTR CreateDefslot(template, name)
   DT_PTR template;
   char *name;
{
   DS_PTR new_slot = AllocDefslot();
   DS_PTR *slot_ptr = &(template->slots);

   new_slot->name = name;
   new_slot->next_slot = NULL;

   WHILE (*slot_ptr) DO
      slot_ptr = &((*slot_ptr)->next_slot);
   DONE;

   *slot_ptr = new_slot;

   return (new_slot);
}

/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : (TemplateFormatToRelationFormat)
 *
 *  PURPOSE: This function converts a template pattern into a
 *           standard relation pattern.
 *
 *  INPUTS : A flag, assert_pattern, that is set to true when the
 *           pattern is being asserted and the name of the template.
 *
 *  RETURNS: None
 *
 *  Notes:   The resulting relation pattern is placed in the GetToken
 *           queue so that it will be read by the proper pattern
 *           analysis routines.
 * ------------------------------------------------------------------
 */

int  TemplateFormatToRelationFormat(name, assert_pattern,var_list)
   char *name;
   int assert_pattern;
   FIELD_PTR *var_list;
{
   DT_PTR template = FindDeftemplate (DeftemplateHead, name);
   TN_PTR *slots = (TN_PTR *)
      GenAlloc ((unsigned)(sizeof (TN_PTR) * (template->max_slot_pos + 1)),
                  "TemplateFormatToRelationFormat");
   TK_PTR tknstruct = GetTokenPtr (CurrentFile);
   int position,i,num_fields ,
       num_field_err, expect_lft_paren,
       last_token,left_paren;

   /*================================================================*/
   /* The following variables should have been declared in the local */
   /* blocks where they were used.  A compiler bug resulted in the   */
   /* declarations having to be scoped to the entire function.       */
   /*================================================================*/

   TN_PTR *tokens;
   DS_PTR slot_desc;
   TN_PTR *token_list_ptr;
   char *tempstr;

   /*============================*/
   /* Initialize the slots table */
   /*============================*/


    for(i = 0; i <  (template->max_slot_pos + 1) ; i++)
      slots[i] = NULL;

   /*====================================*/
   /* Process the fields of the template */
   /*====================================*/

   WHILE ((tknstruct->token NEQ RPAREN) && (tknstruct->token != STOP)) DO

      /*=============================================*/
      /* Process relation style syntax in a template */
      /*=============================================*/

      IF (tknstruct->token NEQ LPAREN) THEN
         ErrorMessage (ERROR,
              "Expected a '(' or ')' within template pattern.");
         UnGetTokenPtr (tknstruct);
         GenFree ((char *)slots, (unsigned)(sizeof (TN_PTR) * (template->max_slot_pos + 1)));
         return(RESTART);

      /*======================*/
      /* Process valid syntax */
      /*======================*/

      ELSE
         FreeToken (tknstruct);

         /*=======================*/
         /* Process the slot name */
         /*=======================*/

         tknstruct = GetTokenPtr (CurrentFile);
         slot_desc = FindDefslot (template, tknstruct->tknword);

         IF(slot_desc NEQ NULL) THEN

           FreeToken (tknstruct);

           /*==============================*/
           /* Locate the slot's value list */
           /*==============================*/

           token_list_ptr = &slots[slot_desc->position];

           /*==================================================*/
           /* Copy the slot's value tokens into the slot table */
           /*==================================================*/
           i = 1;
           num_fields = 0;
           num_field_err = FALSE;
           expect_lft_paren = FALSE;
           last_token = 0;
           left_paren = 0;
           WHILE(i) DO
             tknstruct = GetTokenPtr (CurrentFile);
             if (tknstruct->token EQ RPAREN) 
               i--;
             else if (tknstruct->token EQ LPAREN) 
               i++;
             else if (((tknstruct->token EQ BWORD) ||
			(tknstruct->token EQ BWORDS)) && (!assert_pattern))
               QueueVarList(var_list,tknstruct->tknword);
             else if(tknstruct->token EQ STOP)
               i = 0;
             IF(i)THEN
                 QueueToken (token_list_ptr, tknstruct);
                 IF(!num_field_err)THEN
	                 switch(tknstruct->token)
	                 {
	                   case WORD:
	                        IF((strcmp(tknstruct->tknword,":") != 0) &&
	                          (strcmp(tknstruct->tknword,"=") != 0) && (!left_paren))THEN
	                          num_fields++;
	                        END_IF
	                        break;
	                   case STRING:
	                   case INTEGER:
	                   case FLOAT:
	                   case GLOBAL:
	                   case INSTANCE:
	                   case SINGLE:
	                   case MULTIPLE:
	                   case BWORD: 
	                   case BWORDS:
	                       IF(!left_paren) THEN
	                         num_fields++;
	                       END_IF
	                       break;
	                   case LPAREN:
	                       if(!left_paren) 
	                          num_fields++;
	                       left_paren++;
	                       break;
	                   case RPAREN:
	                        left_paren--;
	                        break;
	                   case LAND:
	                   case LOR:
	                        if((!left_paren)&&(last_token != LAND)
	                           &&(last_token != LOR))
	                          num_fields--;
	                        break;
	                   case LNOT:
	                      /*  if((!left_paren)&&(last_token != LAND)
	                           &&(last_token != LOR))
	                          num_fields--;*/
	                        break;
	                   defaults:
	                        num_fields++;
	                        break;
	                   
	                 }
	                
                 END_IF
             END_IF
             last_token = tknstruct->token;
           END_WHILE
           IF((num_fields != 1) && (slot_desc->max_elements == 1) &&
	           (slot_desc->min_elements == 1))THEN
	         sprintf(MsgBuf,"Single field slot %s must contain a single field value\n",
	                      slot_desc->name);
	         ErrorMessage(ERROR,MsgBuf);
	         GenFree ((char *)slots, (unsigned)(sizeof (TN_PTR) * (template->max_slot_pos + 1)));
	         return(RESTART);
	       END_IF
           FreeToken (tknstruct);
         ELSE                          /* SLOT IS NOT DEFINED */
           IF(VERBOSE IS_ON) THEN
              sprintf(MsgBuf,
                 "Slot <%.40s> is not defined in deftemplate <%.40s>.",
                 tknstruct->tknword,template->name);
              ErrorMessage(ERROR,MsgBuf);
           ELSE
              ErrorMessage(ERROR,"Slot not defined in a deftemplate.");
           END_IF
           FreeToken (tknstruct);
           GenFree ((char *)slots, (unsigned)(sizeof (TN_PTR) * (template->max_slot_pos + 1)));
           return(RESTART);
         END_IF         
      END_IF
      tknstruct = GetTokenPtr (CurrentFile);
   DONE

   /*=============================================================*/
   /* Process the slot table into the equivalent relation pattern */
   /*=============================================================*/

   FOR (position = 2; position <= template->max_slot_pos; position++) DO
      tokens = &slots[position];

      /*=========================================*/
      /* Get the description of the current slot */
      /*=========================================*/

      slot_desc = FindDefslotByPosition (template, position);

      /*===========================================*/
      /* Process explicit slot value specification */
      /*===========================================*/

      IF (*tokens) THEN
         WHILE (*tokens) DO
            UnGetTokenPtr (DequeueToken (tokens));
         DONE

      /*=======================================*/
      /* Implied slot value (inside an assert) */
      /*=======================================*/

      ELSE_IF (assert_pattern)
         TN_PTR tknnode = slot_desc->defaults;

         IF (tknnode) THEN
            WHILE (tknnode) DO
               TK_PTR copy = AllocToken();
               char *str1 = (char *)tknnode->token , *str2 = (char *) copy;

               for (i = 0; i < sizeof(TOKEN) ;i++)
                 str2[i] = str1[i];

               UnGetTokenPtr (copy);

               tknnode = tknnode->next;
            DONE
         ELSE
            TK_PTR null_token = AllocToken();
      
            null_token->token = WORD;
            null_token->tknword = "CRSV_NIL";
            null_token->hashword = AddSymbol(null_token->tknword);

            UnGetTokenPtr (null_token);
         END_IF

      /*==============================================*/
      /* Implied slot value (a pattern to be matched) */
      /*==============================================*/

      ELSE
         TK_PTR wildcard = AllocToken();

         IF (slot_desc->min_elements EQ 1 AND
            slot_desc->max_elements EQ 1) THEN
            wildcard->tknword = "?";
            wildcard->token = SINGLE;
         ELSE
            wildcard->tknword = "$?";
            wildcard->token = MULTIPLE;
         END_IF

        UnGetTokenPtr (wildcard);
      END_IF
   DONE

   /*============================*/
   /* Process the closing paren. */
   /*============================*/

   UnGetTokenPtr (tknstruct);

   /*=====================*/
   /* Free the slot table */
   /*=====================*/

   GenFree ((char *)slots, (unsigned)(sizeof (TN_PTR) * (template->max_slot_pos + 1)));
   return(OK);
}

/* ======================================================================= */

static void QueueVarList(var_list,var)
FIELD_PTR *var_list;
char *var;
{
   FIELD_PTR node,temp;
  
   
   temp = *var_list;
   WHILE(temp != NULL)DO
       if(strcmp (var, temp->word) == 0)
         break;
       temp = temp->next;
   END_WHILE
   IF (temp == NULL)THEN
       node = AllocField();
       node->word = var;
       node->num = 1;
       node->next = *var_list;
       *var_list = node;
   ELSE
      temp->num += 1;
   END_IF
     
}

/* --------------------------------------------------------------
 * FreeTokenAndDefTmp
 * --------------------------------------------------------------
 */

FreeTokenAndDefTmp(tknstruct,cur_def_tmp)
TK_PTR tknstruct;
DT_PTR cur_def_tmp;
{
  FreeToken (tknstruct);
  FreeTokensInSavedList (&TokensInDefrelation);
  FreeDeftemplate(cur_def_tmp);
}