/*******************************************************************
 *      programer:                                                 *
 *                Huyen-anh (Bebe) Ly                              *
 *                Artificial Intelligence Section (MPAD)           *
 *                NASA/JSC   FM7, Houston, TX 77058                *
 *      Purpose :                                                  *
 *                This program is designed for dynamically         *
 *                checking an expert system, which is written      *
 *                in CLIPS.                                        *
 *      Description:                                               *
 *                One way to verify the dynamic information of     *
 *                an expert system is to look at the trace file    *
 *                of the expert system. This file records the      *
 *                behavior of an expert system during run time.    *
 *                This program is designed to scan the trace or    *
 *                dribble files,and use the defrelation and        *
 *                deftemplate forms to verify the facts.  The      *
 *                trace file and the definition forms are provided *
 *                by the user. This program also provides users    *
 *                information about the behavior of the rules, and *
 *                the related relations.                           *
 *******************************************************************/


#include <stdio.h>
#include "crsv.h"
#include <ctype.h>
#include <errno.h>



#define MAX_DEACTIVATE_RATIO   0.3      /* Max ratio of deactivated rules */




/* ======================================================================== */
/* ==============  Functions defined Externally but used here  ============ */
/* ======================================================================== */

/* -----------------------------
 *  From the SYSTEM
 * -----------------------------
 */

extern double          atof();
 
/* -----------------------------
 *  From the file: CRSVPRST.C
 * -----------------------------
 */

extern int  ProcessRelationActivation();
extern int  ProcessRuleActivation();
extern int  ProcessRelationDactivation();
extern int  ProcessRuleDactivation();
extern int  ProcessRuleFire();

/* -----------------------------
 *  From the file: CRSVPRNT.C
 * -----------------------------
 */

extern int            SendMessage();
extern void           ErrorMessage ();
 
/* -----------------------------
 *  From the file: CRSVPROC.C
 * -----------------------------
 */

extern DEF_REL        *FindDefrelation();
 

/* -----------------------------
 * From the file : CRSVTMP.C
 * -----------------------------
 */

extern DT_PTR FindDeftemplate();
extern DS_PTR FindDefslot();

/* -----------------------------
 * From the file : CRSVMEM.C
 * -----------------------------
 */
 extern FIELD_PTR AllocField();
 extern void FreeFieldValue();
 
 /* -----------------------------
 * From the file : CRSVHASH.C
 * -----------------------------
 */
 extern HASH *AddSymbol();

/* ------------------------------
 * From the file: CRVSTKN.C
 * ------------------------------
 */
 
 extern int GetTokenWithEndOfLine();

/* ======================================================================== */
/* ==============  Functions defined here for Internal used =============== */
/* ======================================================================== */

static int         CheckWord();
static int         CheckString();
static int         CheckNumber();
static int         CheckInteger();
static int         CheckFloat();
static int         CheckFieldsOfRelation();

static void        PrintRelation();
static void        PrintRuleList();
static void        PrintSumaryOfTrace();

/* ======================================================================== */
/* ==============  Functions defined here for globally used  ============== */
/* ======================================================================== */

void        EvaluationRelation();
ACT_REL_PTR FindActivatedRelation();
int         ProcessTraceFile();
void        CheckDynamicRules();
void        CheckDynamicRelations();
void        PrintDetailRelations();
char   		*getline();
int     	ScanTraceFile();
void    	CheckForDynUnassrtRels();
void    	CheckForDynUnactRels();
void 		EvaluateTemplate();
FIELD_PTR 	ConvertTemplateToRelation();
FIELD_PTR 	QueueFieldByPosition();

/* ======================================================================== */
/* ==============  Variables defined Externally but used here  ============ */
/* ======================================================================== */

extern ACT_R_PTR    ActivatedRuleHead;      /* First node of the activated rules */
                                            /* found in a trace file             */
extern ACT_REL_PTR  ActvatedRelationHead;   /* First node of the activated       */
                                            /* relation found in a trace file    */
extern DR_PTR       DefrelationHead;        /* First node of the definitions of  */
                                            /* the relations                     */
extern RULE        *RuleHead;               /* First node of the rule list       */
extern DT_PTR		DeftemplateHead;

extern FILE *CurrentFile;
extern char *TKNWORD;
extern char *CurrentObjectName;
extern int errno;

/* ======================================================================== */
/* ==============  Variables defined here for Internal use  =============== */
/* ======================================================================== */

ACT_REL_PTR LastActivatedRelation;       /* last activated relation               */
int MaxOccurences;             /* keep track with the maximum occurrences
                                   at one time of an activated relation  */


/* ======================================================================== */

 /*******************************************************
  *  ScanTraceFile()
  *     scan the trace file,a and call the appropriate
  *     function to process the trace file.
  *  Argument:
  *     fp : File pointer - Points to a current position
  *          of the trace file.
  *  Return : OK or ERROR
  *
  *******************************************************
  */
  static int ScanTraceFile()
  {
    int token;
    
    token = GetTokenWithEndOfLine(CurrentFile);
    if(strcmp(TKNWORD,"CRSV>") != 0)
      return(ERROR);
    token = GetTokenWithEndOfLine(CurrentFile);
    if(token != EOL)
      return(ERROR);
    WHILE((token = GetTokenWithEndOfLine(CurrentFile)) != STOP) DO
          IF(strcmp(TKNWORD,"AS") == 0)THEN
             if(ProcessRelationActivation() == ERROR)
               return(ERROR);
          ELSE_IF(strcmp(TKNWORD,"AC") == 0)
             if(ProcessRuleActivation() == ERROR)
               return(ERROR);
          ELSE_IF(strcmp(TKNWORD,"R") == 0)
             if(ProcessRelationDactivation() == ERROR)
               return(ERROR);
          ELSE_IF((strcmp(TKNWORD,"D") == 0))
             if(ProcessRuleDactivation() == ERROR)
               return(ERROR);
          ELSE_IF ((strcmp(TKNWORD,"F") == 0))
             if(ProcessRuleFire() == ERROR)
               return(ERROR);
          ELSE_IF(token != EOL)
            return(ERROR);
          END_IF
    DONE

    return(OK);
  }
  

/* ----------------------------------------------------
 *    Name     : CheckWord()
 *    Purpose  : Checks to see if the relation's field
 *               is a WORD and has been specified in the
 *               allowed-words list of the particular defrelation.
 *    Arguments: words - Pointer to the allowed-words list.
 *               fld_ptr - Pointer to a field of a relation.
 *    Returns  : ERROR if the field does not meet the
 *               specification;else returns OK.
 * --------------------------------------------------------
 */

 static int CheckWord(words,fld_ptr,fact_ptr,name)
 WL_PTR words;
 FIELD_PTR fld_ptr;
 SPE_FACT_PTR fact_ptr;
 char *name;
 {
    WL_PTR current_word;

    if(((fld_ptr->word[0] == '-')||
      (fld_ptr->word[0] == '+'))
      &&(isdigit(fld_ptr->word[1])))
      return(ERROR);
    if(isdigit(fld_ptr->word[0]))
      return(ERROR);
    if(fld_ptr->word[0] == '"')
      return(ERROR);
    IF(words!= NULL)THEN
       current_word = words;
       while(current_word != NULL)DO
           if(strcmp(current_word->word,fld_ptr->word) == 0)
              return(OK);
           else
             current_word = current_word->next_word;
       END_WHILE
       sprintf(MsgBuf,"In fact:  f-%d ", fact_ptr->fact_num);
       ErrorMessage(ERROR,MsgBuf);
       PrintRelation(fact_ptr,name);
       sprintf(MsgBuf,
      "\n        Used a SYMBOL <%.40s> in field %d that is not in the ALLOWED_WORDS list",
          fld_ptr->word,fld_ptr->num);
       SendMessage(MsgBuf,NO);
       return(OK);
    ELSE
       return(OK);
    END_WHILE
 }


/* ------------------------------------------------------------
 * Name        : CheckString()
 * Purpose     : Checks to see if the a relation's field
 *               is a STRING and specified in the allowed-strings
 *               list of the defrelation.
 * Arguments   : strings_ptr - Pointer to the allowed-strings.
 *               fld_ptr     - Pointer to a field of a relation.
 * Returns     : ERROR if the field does not meet the specification.
 *               OK if everything is ok.
 * -------------------------------------------------------------*/

 static int CheckString(strings_ptr,fld_ptr,fact_ptr,name)
 WL_PTR strings_ptr;
 FIELD_PTR fld_ptr;
 SPE_FACT_PTR fact_ptr;
 char *name;
 {
    WL_PTR current_string;

    if(fld_ptr->word[0] != '"')
      return(ERROR);
    IF(strings_ptr != NULL)THEN
       current_string = strings_ptr;
       WHILE(current_string != NULL)DO
             if(strcmp(current_string->word,fld_ptr->word) == 0)
                return(OK);
             else
                current_string = current_string->next_word;
       END_WHILE
       sprintf(MsgBuf,"In fact:  f-%d ", fact_ptr->fact_num);
       ErrorMessage(ERROR,MsgBuf);
       PrintRelation(fact_ptr,name);
       sprintf(MsgBuf,
      "\n         Used a STRING <%.40s> in field %d that is not in the ALLOWED_STRINGS list",
          fld_ptr->word,fld_ptr->num);
       SendMessage(MsgBuf,NO);
        return(OK);
    END_IF
    return(OK);
 }

/* ------------------------------------------------------------
 * Name        : CheckInteger()
 * Purpose     : Checks to see if the a relation's field
 *               is an INTEGER and specified in the allowed-numbers
 *               list of the defrelation.
 * Arguments   : integers_ptr - Pointer to the allowed-strings.
 *               fld_ptr     - Pointer to a field of a relation.
 * Returns     : ERROR if the field does not meet the specification.
 *               OK if everything is ok.
 * -------------------------------------------------------------*/

 static int CheckInteger(def_fld_ptr,act_fld_ptr,fact_ptr,name)
 DF_PTR def_fld_ptr;
 FIELD_PTR act_fld_ptr;
 SPE_FACT_PTR fact_ptr;
 char *name;
 {
    
    long num;
    
    num = atol(act_fld_ptr->word);
    if((num == 0) && (errno == ERANGE))
      return(ERROR);
    IF(def_fld_ptr->possible_integers!= NULL)THEN
       IF(IsIntegerInList(def_fld_ptr->possible_integers,num) == NO) THEN
             sprintf(MsgBuf,"In fact:  f-%d ", fact_ptr->fact_num);
             ErrorMessage(ERROR,MsgBuf);
             PrintRelation(fact_ptr,name);
             sprintf(MsgBuf,
             "\n         Used an INTEGER <%d> in field %d that is not in the ALLOWED-INTEGERS list",
               num,act_fld_ptr->num);
             SendMessage(MsgBuf,NO);
             return(OK);
        ELSE
         return (OK);
        END_IF
     END_IF

     IF(def_fld_ptr->set_int_max == YES) THEN           /* If the maximum value has been set */
        IF(num > def_fld_ptr->int_max_range) THEN    /* If the value is greater than the maximum */
              sprintf(MsgBuf,"%d is out of range. Max allowed value is %d",
                      num,def_fld_ptr->int_max_range);
              ErrorMessage(ERROR,MsgBuf);
              sprintf(MsgBuf,"\n         f-%d   ",fact_ptr->fact_num);
              SendMessage(MsgBuf,NO);
              PrintRelation(fact_ptr,name);
              return(OK);
        END_IF
     END_IF
     IF(def_fld_ptr->set_int_min == YES) THEN          /* If the minimum value has been set */
       IF(num < def_fld_ptr->int_min_range)  THEN   /* If the value is less than the minimum value */
            sprintf(MsgBuf,"%d is out of range. Min allowed value is %d",
                    num,def_fld_ptr->int_min_range);
            ErrorMessage(ERROR,MsgBuf);
            sprintf(MsgBuf,"\n         f-%d   ",fact_ptr->fact_num);
            SendMessage(MsgBuf,NO);
            PrintRelation(fact_ptr,name);
            return(OK);
       END_IF
     END_IF
     return(OK);
 }

/* ------------------------------------------------------------
 * Name        : CheckFloat()
 * Purpose     : Checks to see if the a relation's field
 *               is a FLOAT and specified in the allowed-floats
 *               list of the defrelation.
 * Arguments   : floats_ptr - Pointer to the allowed-floats.
 *               fld_ptr     - Pointer to a field of a relation.
 * Returns     : ERROR if the field does not meet the specification.
 *               OK if everything is ok.
 * -------------------------------------------------------------*/

 static int CheckFloat(def_fld_ptr,act_fld_ptr,fact_ptr,name)
 DF_PTR def_fld_ptr;
 FIELD_PTR act_fld_ptr;
 SPE_FACT_PTR fact_ptr;
 char *name;
 {
    
    double num;
    
    num = atof(act_fld_ptr->word);
    IF((num == 0) && (errno == ERANGE))
      return(ERROR);
    IF(def_fld_ptr->possible_floats != NULL)THEN
      IF(IsFloatInList(def_fld_ptr->possible_floats,num) == NO) THEN
         sprintf(MsgBuf,"In fact:  f-%d ", fact_ptr->fact_num);
         ErrorMessage(ERROR,MsgBuf);
         PrintRelation(fact_ptr,name);
         sprintf(MsgBuf,
             "\n         Used a FLOAT <%f> in field %d that is not in the ALLOWED-FLOATS list",
               num,act_fld_ptr->num);
         SendMessage(MsgBuf,NO);
         return(OK);
       ELSE
          return (OK);
       END_IF
     END_IF

     IF(def_fld_ptr->set_float_max == YES) THEN           /* If the maximum value has been set */
        IF(num > def_fld_ptr->float_max_range) THEN    /* If the value is greater than the maximum */

              sprintf(MsgBuf,"%f is out of range. Max allowed value is %f",
                      num,def_fld_ptr->float_max_range);
              ErrorMessage(ERROR,MsgBuf);
              sprintf(MsgBuf,"\n         f-%d   ",fact_ptr->fact_num);
              SendMessage(MsgBuf,NO);
              PrintRelation(fact_ptr,name);
              return(OK);
        END_IF
     END_IF
     IF(def_fld_ptr->set_float_min == YES) THEN   /* If the minimum value has been set */
       IF(num < def_fld_ptr->float_min_range) THEN    /* If the value is less than the minimum value */
         
            sprintf(MsgBuf,"%f is out of range. Min allowed value is %f",
                    num,def_fld_ptr->float_min_range);
            ErrorMessage(ERROR,MsgBuf);
            sprintf(MsgBuf,"\n         f-%d   ",fact_ptr->fact_num);
            SendMessage(MsgBuf,NO);
            PrintRelation(fact_ptr,name);
            return(OK);
       END_IF
     END_IF
     return(OK);
 }
/* -------------------------------------------------
 * Name        :  CheckNumber();
 *
 * Purpose     :  Checks to see if a relation's field
 *                is a number and specified
 *                in the allowed-numbers list.
 *
 * Arguments   :  def_fld_ptr - pointer to a specified
 *                field in defrelation.
 *                act_fld_ptr - pointer to a relation's
 *                field.
 *
 * Returns     :  ERROR if the field does not not meet
 *                the specification.
 *                OK if everything is ok.
 *                an error message is also generated
 *                when error is detected.
 * -------------------------------------------------*/

  static int CheckNumber(def_fld_ptr,act_fld_ptr,fact_ptr,name)
  DF_PTR def_fld_ptr;
  FIELD_PTR act_fld_ptr;
  SPE_FACT_PTR fact_ptr;
  char *name;
  {
     double num;

     if((act_fld_ptr->word[0] != '-')
        &&(act_fld_ptr->word[0] != '+')
        &&(!isdigit(act_fld_ptr->word[0])))
          return(ERROR);
     if(((act_fld_ptr->word[0] == '-')
        ||(act_fld_ptr->word[0] == '+'))
        &&(!isdigit(act_fld_ptr->word[1])))
          return(ERROR);

     num = atof(act_fld_ptr->word);

     /* If there is a retriction in the number (according to defrelation ).*/

     IF(def_fld_ptr->possible_integers != NULL)THEN
          if(IsIntegerInList(def_fld_ptr->possible_integers,(int)num) == YES)
             return(OK);
     END_IF
     IF(def_fld_ptr->possible_floats != NULL)THEN
       IF(IsFloatInList(def_fld_ptr->possible_floats,num) == NO)THEN
               sprintf(MsgBuf,"In fact:  f-%d ", fact_ptr->fact_num);
               ErrorMessage(ERROR,MsgBuf);
               PrintRelation(fact_ptr,name);
               sprintf(MsgBuf,
               "\n         Used a NUMBER <%f> in field %d that is not in the ALLOWED-NUMBERS list",
                num,act_fld_ptr->num);
               SendMessage(MsgBuf,NO);
               return(OK);
        END_IF
      END_IF
      IF(def_fld_ptr->set_float_max == YES) THEN           /* If the maximum value has been set */
        IF(num > def_fld_ptr->float_max_range) THEN    /* If the value is greater than the maximum */

              sprintf(MsgBuf,"%f is out of range. Max allowed value is %f",
                      num,def_fld_ptr->float_max_range);
              ErrorMessage(ERROR,MsgBuf);
              sprintf(MsgBuf,"\n         f-%d   ",fact_ptr->fact_num);
              SendMessage(MsgBuf,NO);
              PrintRelation(fact_ptr,name);
              return(OK);
        END_IF
      END_IF
      IF(def_fld_ptr->set_float_min == YES)  THEN         /* If the minimum value has been set */
        IF(num < def_fld_ptr->float_min_range) THEN    /* If the value is less than the minimum value */
            sprintf(MsgBuf,"%f is out of range. Min allowed value is %f",
                    num,def_fld_ptr->float_min_range);
            ErrorMessage(ERROR,MsgBuf);
            sprintf(MsgBuf,"\n         f-%d   ",fact_ptr->fact_num);
            SendMessage(MsgBuf,NO);
            PrintRelation(fact_ptr,name);
            return(OK);
         END_IF
      END_IF
     return(OK);
  }

/* ---------------------------------------------
  * Prints error message.
  * ---------------------------------------------
  */
 static void ErrorMessageInTrace(fld_ptr,fact_ptr,name)
 FIELD_PTR fld_ptr;
 SPE_FACT_PTR fact_ptr;
 char *name;
 {
     sprintf(MsgBuf,"<%.40s> is not allowed in field %d of",
                fld_ptr->word, fld_ptr->num);
     ErrorMessage(ERROR,MsgBuf);
     sprintf(MsgBuf,"\n        relation <%.40s>. Used in fact:",name);
     SendMessage(MsgBuf,NO);
     sprintf(MsgBuf,"\n        f-%d   ",fact_ptr->fact_num);
     SendMessage(MsgBuf,NO);
     PrintRelation(fact_ptr,name);
 }


/* ------------------------------------------------------------ *
 *  Checks each field of a relation to make sure it
 *  meet the specification of the defrelation.
 *  - fact_ptr : the fact that needs to be checked with definition.
 *  - node     : the node that contains the definition of the relation.
 *  - name     : name of the relation.
 * ------------------------------------------------------------ */

 static int CheckFieldsOfRelation(fact_ptr,node,name)
 SPE_FACT_PTR fact_ptr;
 DR_PTR node;
 char *name;

 {
   DF_PTR def_fld_ptr;
   FIELD_PTR act_fld_ptr;
   int i;

   IF(fact_ptr->number_field > 1)THEN
       act_fld_ptr = fact_ptr->field;
       WHILE(act_fld_ptr != NULL)DO
           i = act_fld_ptr->num;
           def_fld_ptr = node->fields;

           /* ----------------------------------------- *
            *   Check the specification of the  field. *
            * ----------------------------------------- */

           while((def_fld_ptr !=NULL)&&(def_fld_ptr->position != i))
             def_fld_ptr = def_fld_ptr->next_field;
           /* ----------------------------------------- *
            *  If not specified,assumed to be OK        *
            * ----------------------------------------- */

           IF(def_fld_ptr != NULL)THEN
              switch(act_fld_ptr->type)
              {
                case INTEGER:
                  if(def_fld_ptr->allow_integer IS_NO)
                    ErrorMessageInTrace(act_fld_ptr,fact_ptr,name);
                  break;
                case FLOAT:
                  if(def_fld_ptr->allow_float IS_NO)
                    ErrorMessageInTrace(act_fld_ptr,fact_ptr,name);
                  break;
                case WORD:
                  if(def_fld_ptr->allow_word IS_NO)
                    ErrorMessageInTrace(act_fld_ptr,fact_ptr,name);
                  break;
                case STRING:
                  if(def_fld_ptr->allow_string IS_NO)
                    ErrorMessageInTrace(act_fld_ptr,fact_ptr,name);
                  break;
                case INSTANCE:
                  if(def_fld_ptr->allow_instance IS_NO)
                    ErrorMessageInTrace(act_fld_ptr,fact_ptr,name);
                  break;
              } 
           END_IF 
         act_fld_ptr = act_fld_ptr->next;
      END_WHILE
   END_IF
     return (OK);
 }

/* ------------------------------------------------------- *
 * Evaluates each relation which was asserted during
 * the run time. Since defrelation is a binary tree structure
 * list, recursive is used.
 * Defrelation is provided by either the user or the CRSV.
 * If it is provided by CRVS the evaluation process
 *   will stop and a warning message will be generated.
 * If it is provided by the user the evaluation process
 *   will check each field to detect any conflict bettween
 *   the relation and its definition.
 * -------------------------------------------------------- */

void EvaluationRelation(fact_ptr,name)
SPE_FACT_PTR fact_ptr;
char *name;
{
   DR_PTR node;


   IF((node = FindDefrelation(DefrelationHead,name)) == NULL)THEN
     IF(strcmp(name,"initial-fact") != 0)THEN
            IF(fact_ptr->templ_slot == NULL) THEN
               sprintf(MsgBuf,"No definition for relation <%.40s> provided",
                    name);
               ErrorMessage(WARNING,MsgBuf);
            ELSE
                sprintf(MsgBuf,"No definition for template <%.40s> provided",
                    name);
               ErrorMessage(WARNING,MsgBuf);
            END_IF
      END_IF
    ELSE_IF(node->created_by == CRSV_MADE)
        sprintf(MsgBuf,"No definition for relation <%.40s> provided",name);
        ErrorMessage(WARNING,MsgBuf);
    ELSE
      IF(fact_ptr->number_field > node->max_fields)THEN
             sprintf(MsgBuf,
                     "Too many fields used in fact: \n         f-%d ",
                     fact_ptr->fact_num);
             ErrorMessage(ERROR,MsgBuf);
             PrintRelation(fact_ptr,name);
      ELSE_IF(fact_ptr->number_field < node->min_fields)
             sprintf(MsgBuf,
                  "Not enough fields used in fact: \n         f-%d ",
                  fact_ptr->fact_num);
             ErrorMessage(ERROR,MsgBuf);
             PrintRelation(fact_ptr,name);
      ELSE
           (void)CheckFieldsOfRelation(fact_ptr,node,name);
      END_IF
    END_IF
}


/* ------------------------------------------------------- *
 * Evaluates each template which was asserted during
 * the run time. 
 * Deftemplate is provided bythe user.This function
 *   will check each field to detect any conflict bettween
 *   the template  and its definition.
 * -------------------------------------------------------- */

void EvaluateTemplate(fact_ptr,name)
SPE_FACT_PTR fact_ptr;
char *name;

{
    FIELD_PTR fields = NULL;
    
    fields = ConvertTemplateToRelation(fact_ptr->templ_slot,name);
    fact_ptr->field = fields;
    EvaluationRelation(fact_ptr,name);
    fact_ptr->field = NULL;
    FreeFieldValue(fields);
}
/* =======================================================
 * ConvertTemplateToRelation
 * =======================================================
 */
FIELD_PTR ConvertTemplateToRelation(mytemplate,name)
MY_TMPLATE_PTR mytemplate;
char *name;

{
   MY_TMPLATE_PTR temp;
   FIELD_PTR temp_field = NULL, fields = NULL , afield = NULL;
   DS_PTR slot = NULL;
   DT_PTR d_template;
   HASH *temp_hash;
   int num = 0;
   
   temp = mytemplate;
   d_template = FindDeftemplate(DeftemplateHead,name);   
   IF(d_template == NULL)THEN
      sprintf(MsgBuf,"Deftemplate : %s was not provided \n",name);
      ErrorMessage(ERROR,MsgBuf);
      return(NULL);
   END_IF
   WHILE(temp != NULL)DO
      slot = FindDefslot(d_template,temp->name);
      IF (slot != NULL) THEN
        afield = temp->value_list;
        while(afield != NULL)DO
               temp_field = AllocField();
        	temp_field->num = slot->position + num++;
        	temp_hash = AddSymbol(afield->word);
        	temp_field->word = temp_hash->contents;
        	fields = QueueFieldByPosition(fields,temp_field);
        	afield = afield->next;
        END_WHILE
      ELSE
        sprintf(MsgBuf,
	    "Slot %s is not defined in deftemplate <%s> \n",
		temp->name,name);
        ErrorMessage(ERROR,MsgBuf);
      END_IF
      temp = temp->next;
   END_WHILE
  return(fields);
}
/* =======================================================
 * QueueFieldByPosition
 * =======================================================
 */

FIELD_PTR QueueFieldByPosition(list,field)
FIELD_PTR list,field;

{
   FIELD_PTR temp_fld1,temp_fld2 = NULL;
   
   IF (list == NULL)THEN
     return(field);
   ELSE
     temp_fld1 = list;
     WHILE(temp_fld1 != NULL)DO
         IF(temp_fld1->num < field->num)THEN
           temp_fld2 = temp_fld1;
           temp_fld1 = temp_fld1->next;
         ELSE
           break;
         END_IF
     END_WHILE
     IF(temp_fld2 != NULL)THEN
       temp_fld2->next = field;
       field->next = temp_fld1;
     ELSE
        field->next = temp_fld1;
        return(field);
     END_IF
   END_IF
   return(list);
}
/* ------------------------------------------------ *
 *  PrintRelation()
 *    Print a relation, includes the relation's name
 *         and its fields.
 *
 *  Argument:
 *     fact_ptr : pointer pointing to the relation
 *     name     : name of the relation.
 *
 *  Returns: nothing.
 * ------------------------------------------------ */

static void PrintRelation(fact_ptr,name)
SPE_FACT_PTR fact_ptr;
char *name;
{
     FIELD_PTR fld_ptr;
     MY_TMPLATE_PTR slot_ptr;

     sprintf(MsgBuf,"(%s",name);
     SendMessage(MsgBuf,NO);
      IF(fact_ptr->field != NULL) THEN
	     fld_ptr = fact_ptr->field;
	     while(fld_ptr!=NULL)
	       {
	         sprintf(MsgBuf,"  %s",fld_ptr->word);
	         SendMessage(MsgBuf,NO);
	         fld_ptr = fld_ptr->next;
	       }
	  ELSE_IF (fact_ptr->templ_slot != NULL)
	    slot_ptr = fact_ptr->templ_slot; 
	    while(slot_ptr != NULL)
	     {
	        sprintf(MsgBuf,"(%s ",slot_ptr->name);
	        SendMessage(MsgBuf,NO);
	        fld_ptr = slot_ptr->value_list;
	        while(fld_ptr != NULL)
	          {
	            sprintf(MsgBuf," %s",fld_ptr->word);
	            SendMessage(MsgBuf,NO);
	            fld_ptr = fld_ptr->next;
	          }
	         SendMessage(")",NO);
	        slot_ptr = slot_ptr->next;
	     }
	  END_IF
     SendMessage(")",NO);

}
 /* ---------------------------------------------------------------
  * PrintRuleList()
  *    This function prints all the rules that were activated or
  *    deactivated because of the assertion or retraction of
  *    the relation.
  * Arguments:
  *   - RULE_NAME_PTR  list;
  *           head of the rule list.
  * Return: Nothing;
  * ----------------------------------------------------------------
  */
  static void PrintRuleList(list)
  RULE_NAME_PTR list;
  {
    RULE_NAME_PTR  node;

    node = list;
    WHILE(node != NULL)DO
       IF(node->last_fact == YES)THEN
             sprintf(MsgBuf,"         <%.40s> ( %d times, last fact)\n",
                     node->name,node->counter);
             SendMessage(MsgBuf,NO);
       ELSE
             sprintf(MsgBuf,"         <%.40s> ( %d times)\n",
                     node->name,node->counter);
             SendMessage(MsgBuf,NO);
       END_IF
       node = node->next;
    END_WHILE

  }
 /* -------------------------------------------------------
 *  void PrintDetailRelations()
 *     This function will traverse through the defrelation list
 *     and specifically print all the important information
 *     about each relation during the process of trace file.
 *        
 *  Argument:
 *     Node : a node from defrelation list.
 *
 *  Returns: Nothing.
 * -------------------------------------------------------
 */

  void PrintDetailRelations(node)
  ACT_REL_PTR node;
  {
    SPE_FACT_PTR fact_ptr;

   if(node->left != NULL)
      PrintDetailRelations(node->left);
   sprintf(MsgBuf,"Relation                  : %.40s\n", node->name);
   SendMessage(MsgBuf,NO);
   sprintf(MsgBuf,"has been asserted         : %d times\n",
           node->num_assert);
   SendMessage(MsgBuf,NO);
   sprintf(MsgBuf,"Retracted                 : %d times\n",node->num_retract);
   SendMessage(MsgBuf,NO);
   sprintf(MsgBuf,"Max existing at one time  : %d \n",node->max_occurences);
   SendMessage(MsgBuf,NO);
   fact_ptr = node->spec_fact_list;

   while(fact_ptr != NULL)DO
      sprintf(MsgBuf,"   f-%d  ",fact_ptr->fact_num);
      SendMessage(MsgBuf,NO);
      PrintRelation(fact_ptr,node->name);
      IF((fact_ptr->num_activated_a == 0)&&(fact_ptr->num_deactivated_a == 0)&&
         (fact_ptr->num_activated_d == 0)&&(fact_ptr->num_deactivated_d == 0))THEN
           SendMessage(
           "\n       was not involved in Activating or Deactivating any rule\n",NO);
      ELSE
           IF(fact_ptr->num_activated_a > 0)THEN
               sprintf(MsgBuf, "\n   Involved in Activating by being asserted     : %d  times \n",
                  fact_ptr->num_activated_a);
               SendMessage(MsgBuf,NO);
               SendMessage("      Activated rules were:\n",NO);
               PrintRuleList(fact_ptr->activate_list_a);
           END_IF
           IF(fact_ptr->num_deactivated_a > 0)THEN
               sprintf(MsgBuf,"\n   Involved in Deactivating by being asserted   : %d  times \n",
                  fact_ptr->num_deactivated_a);
               SendMessage(MsgBuf,NO);
               SendMessage("      Deactivated rules were:\n",NO);
               PrintRuleList(fact_ptr->deactivate_list_a);
           END_IF
           IF(fact_ptr->num_activated_d > 0)THEN
               sprintf(MsgBuf,
               "\n   Involved in Activating by being retracted    : %d  times \n",
                       fact_ptr->num_activated_d);
               SendMessage(MsgBuf,NO);
               SendMessage("      Activated rules were:\n",NO);
               PrintRuleList(fact_ptr->activate_list_d);
           END_IF
           IF(fact_ptr->num_deactivated_d > 0)THEN
               sprintf(MsgBuf,
               "\n   Involved in Deactivating by being retracted  : %d  times \n",
                       fact_ptr->num_deactivated_d);
               SendMessage(MsgBuf,NO);
               SendMessage("      Deactivated rules were:\n",NO);
               PrintRuleList(fact_ptr->deactivate_list_d);
           END_IF
     END_IF
     SendMessage("\n",NO);
     fact_ptr = fact_ptr->next;
   END_WHILE

   SendMessage(" ------------------------------------------------------\n",NO);
   if(node->right != NULL)
      PrintDetailRelations(node->right);
}

 /* ----------------------------------------- *
  * This function returns the node            *
  * which has the same name as the provided   *
  * <name>.                                   *
  * ----------------------------------------- */

 ACT_REL_PTR FindActivatedRelation(node,name)
 ACT_REL_PTR node;
 char *name;
 {
     int rtn;

     if(node == NULL)
        return(NULL);

     rtn = strcmp(name,node->name);
     if(rtn < 0)
       return(FindActivatedRelation(node->left,name));
     else if (rtn > 0)
       return(FindActivatedRelation(node->right,name));
     else
       return(node);
 }

 /* ---------------------------------------------------- *
  *  evaluates each relation it finds in dribble file    *
  * ---------------------------------------------------- */

static void CheckForDynUnassrtRels(def_rel_ptr)
DR_PTR        def_rel_ptr;
{
   ACT_REL_PTR rel_ptr;
    
   IF(def_rel_ptr EQ NULL) THEN
      return;
   END_IF

   CheckForDynUnassrtRels(def_rel_ptr->lft_def_rel);
   
   IF(def_rel_ptr->created_by EQ USER_MADE) THEN
      rel_ptr = FindActivatedRelation(ActvatedRelationHead,def_rel_ptr->name);
      IF(rel_ptr EQ NULL) THEN
         sprintf(MsgBuf,"\nRelation <%.40s> was not asserted",
                 def_rel_ptr->name);
         SendMessage(MsgBuf,NO);
      END_IF
   END_IF
   
   CheckForDynUnassrtRels(def_rel_ptr->rht_def_rel);
}
  
/* ==========================================================
 * This function checks if the rules are activated
 * during execution. If they have been activated, it checks
 * the number of times they were deactivated.
 * ==========================================================
 */
  
void CheckForDynUnactRels(rule_ptr)
RULE          *rule_ptr;
{ 
   ACT_R_PTR rule;
   float     ratio;
   
   IF(rule_ptr EQ NULL) THEN
      return;
   END_IF

   CheckForDynUnactRels(rule_ptr->lft_rule);
   
   rule = ActivatedRuleHead;               /* Use the global variable */
   while(rule NEQ NULL) DO
      IF(strcmp(rule_ptr->name, rule->name) EQ 0) THEN
         break;
      END_IF
      rule = rule->next;
   END_WHILE
    
   IF(rule EQ NULL) THEN
      sprintf(MsgBuf,"\nRule <%.50s> was not activated",
                      rule_ptr->name);
      SendMessage(MsgBuf,NO);
   ELSE
      ratio = (float) rule->deactivate_num / (float) rule->activate_num;
      IF(ratio >= MAX_DEACTIVATE_RATIO) THEN
          sprintf(MsgBuf,"\nRule <%.37s> was deactivated %5.1f%% of the time",
                  rule_ptr->name, ratio * 100.0);
          SendMessage(MsgBuf,NO);
       END_IF
   END_IF
     
   CheckForDynUnactRels(rule_ptr->rht_rule);
   
   return;
  }

  /* -----------------------------------------
   * ProcessTraceFile()
   *   This function will  scan through the
   *   trace file ,is created by user;and
   *   store all information concerning about
   *   each relation,each rule ect...
   *
   * Argument :
   *    file_name : Name of trace file.
   *
   * return:
   * -----------------------------------------
   */

 int ProcessTraceFile(file_name)
 char *file_name;
 {
      CurrentObjectName = file_name;
      LastActivatedRelation = NULL;
      MaxOccurences = 0;
      IF(ScanTraceFile() == ERROR)THEN
          sprintf(MsgBuf,"There is an error in file : %s",file_name);
          ErrorMessage(ERROR,MsgBuf);
      END_IF

 }


 /* ============================================== *
  *  cross referencing relations found in in the   *
  *  trace file.                                   *
  * ============================================== *
  */

/*ARGSUSED*/ /* Make lint happy */
void CheckDynamicRelations(MyRelationHead, MyDefrelationHead, MyActvatedRelationHead)
RELATION      *MyRelationHead;
DEF_REL       *MyDefrelationHead;
ACT_RELATION  *MyActvatedRelationHead;
{
   CheckForDynUnassrtRels(MyDefrelationHead);
}

/*ARGSUSED*/ /* Make lint happy */
void CheckDynamicRules(MyRuleHead, MyActivatedRuleHead)
RULE          *MyRuleHead;
ACT_RULE      *MyActivatedRuleHead;
{
   CheckForDynUnactRels(MyRuleHead);
}
