#include <stdio.h>
#include <math.h>
#include <string.h>

#include "crsv.h"

/*==================================================
 *  This file contains all the functions used to 
 *  create a file of defrelation info.
 *==================================================
 */
 
/* ===========  Functions defined Externally but used here  =========== */

/* -------------------
 *  From the system
 * -------------------
 */

FILE *fopen();

/* -----------------------------
 *  From the file: CRSVMEM.C
 * -----------------------------
 */

extern COUNTER *AllocCounter();
extern char    *GenAlloc();
extern void     GenFree();

/* -----------------------------
 *  From the file: CRSVMISC.C
 * -----------------------------
 */

extern FIELD  *SortFieldList();
extern int    IsWordInList();
extern double IsFloatInList();


/* -----------------------------
 *  From the file: CRSVTMP.C
 * -----------------------------
 */

extern DS_PTR   FindDefslotByPosition();
extern DT_PTR   FindDeftemplate();


/* ------------------------
 *  From the file: CRSV.C
 * ------------------------
 */

extern DT_PTR   DeftemplateHead;
extern DR_PTR   DefrelationHead;


/* ---------------------------
 *  From the file: CRSVREL.C
 * ---------------------------
 */

extern DR_PTR   FindDefrelation();


/* ===========  Functions defined here for Global use  ================ */

void     CreateDefrelFile();
void     CreateDefextFile();


/* ===========  Functions defined here for internal use  ============== */

void   LoopThroughRelations();
void   OutputDefrelationInformation();
void   CountRRFields();
void   CountFBRFields();
void   ExamineFieldValues();
void   ExamineTemplateFieldValues();
void   PrintTemplateDefaults();
int    DetermineType();
void   OutputListItems();
void   OutputRelationFieldInformation();
void   DetermineRange();

void   LoopThroughFuntions();
void   OutputDefexternaIInformation();


/* ===========  Variables defined here for Global use  ================ */


/* ===========  Variables defined Externally but used here  =========== */

/* ===========  Variables defined here for internal use  ============== */

FILE *fp;


/* ========================================================================= */
/*                    CREATE DEFRELATION FUNCTIONS                           */
/* ========================================================================= */

/*
 * -------------------------------------------------------------
 *  NAME   : CreateDefrelFile()
 *
 *  PURPOSE: This function creates a file with defrelations info
 *           for each of the known defrelations. It makes a 'best'
 *           guess at some of the appropriate info.
 *
 *  INPUTS : Two arguments, a pointer to the head of the
 *           relations list (RELATION *), and a pointer to
 *           the name of the file for defrelation info (char *);
 *
 *  RETURNS: Nothing.
 * -------------------------------------------------------------
 */ 

void CreateDefrelFile(rel_list, file_name)
char      *file_name;
RELATION  *rel_list;
{

   /*========================*/
   /* Open defrelations file */
   /*========================*/

#if MAC_LSC
   GotoDefrelationsDirectory();
#endif

   fp = fopen(file_name, "w");
   IF(fp EQ NULL) THEN

      /*=================================================*/
      /* If file can't be opened, print error and return */
      /*=================================================*/

      sprintf(MsgBuf, "Unable to open file %.20s", file_name);
      ErrorMessage(ERROR, MsgBuf);
      return;
   END_IF
   
   /*=============================*/
   /* Loop through relations list */
   /*=============================*/

   LoopThroughRelations(rel_list);

   fprintf(fp,"\n");
   fclose(fp);
   
#if MAC_LSC
   SetDefrelationsFileCreator();
#endif
}

/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  LoopThroughRelations()
 *
 *  PURPOSE:  This is a recursive function which processes
 *            through all the nodes of the binary tree of
 *            relations.
 *
 *  INPUTS :  A pointer to the list of relations (REL_PTR)
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  Actual output of relations info is in the
 *            OutputDefrelationInformation function.
 * -------------------------------------------------------------
 */

static void LoopThroughRelations(node)
REL_PTR node;
{
   IF(node NEQ NULL) THEN
     LoopThroughRelations(node->lft_rel);
     OutputDefrelationInformation(node);
     LoopThroughRelations(node->rht_rel);
   END_IF
}

/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  OutputDefrelationInformation()
 *
 *  PURPOSE:  This function writes all the defrelation info
 *            to a file.
 *
 *  INPUTS :  A pointer to the current relation (REL_PTR)
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  
 * -------------------------------------------------------------
 */

static void OutputDefrelationInformation(node)
REL_PTR node;
{
   FIELD     *temp_fld;
   int        min = 1000;
   int        max = 0;
   DT_PTR     template;
   DR_PTR     defrelation;
   DF_PTR     def_field;
   register   i;

   IF(VERBOSE IS_ON) THEN
      sprintf(MsgBuf, "\nCreating defrelation for %.40s", node->name);
      SendMessage(MsgBuf,NO);
   END_IF

   template = FindDeftemplate(DeftemplateHead,node->name);
   defrelation = FindDefrelation(DefrelationHead,node->name);

   /* -------------------------------------------
    *  If LITERAL is on, then add the values of
    *  node (relation) to the defrelation 
    *  structure.
    * -------------------------------------------
    */

   IF(template) THEN
     fprintf(fp,"\n\n(deftemplate %s",node->name);
   ELSE

     /*================================================================*/
     /* Search various lists to determine min and max number of fields */
     /*================================================================*/

     CountRRFields(node->rule_list,    &min, &max);
     CountRRFields(node->assert_list,  &min, &max);
     CountFBRFields(node->fb_list,     &min, &max);

     IF (min NEQ max) THEN
        fprintf(fp,"\n\n;; ---------------------------------------------------");
        fprintf(fp,"\n;; NOTICE FOR RELATION: %s",node->name);
        fprintf(fp,"\n;; Due to the use of wildcards the information in the");
        fprintf(fp,"\n;; defrelation structure is not guaranteed accurate.");
        fprintf(fp,"\n;; ---------------------------------------------------");
     END_IF

     fprintf(fp,"\n\n(defrelation %s", node->name);
     fprintf(fp,"\n   (min-number-of-fields %d)", min);
     fprintf(fp,"\n   (max-number-of-fields %d)", max);
   END_IF

   /*============================================*/
   /* Process field list for specific field info */
   /*============================================*/
   
   temp_fld = node->field_list;
   IF(temp_fld NEQ NULL) THEN

      /*=========================*/
      /* Process known arguments */
      /*=========================*/
   
      node->field_list = SortFieldList(node->field_list);

      temp_fld = node->field_list;
      i = 2;
      WHILE (temp_fld NEQ NULL) DO

         WHILE (temp_fld->field_num > i) DO
            fprintf(fp,"\n   (field %d", i++);
            fprintf(fp,"\n      (type ?VARIABLE)");
            fprintf(fp,"\n      (allowed-words ?VARIABLE)");
            fprintf(fp,"\n      (allowed-strings ?VARIABLE)");
            fprintf(fp,"\n      (allowed-instances ?VARIABLE)");
            fprintf(fp,"\n      (allowed-numbers ?VARIABLE)");
            fprintf(fp,"\n      (range ?VARIABLE ?VARIABLE)");
            fprintf(fp,")");
         END_WHILE


         /* ------------------------------------------------
          *  If the relation was defined by a deftemplate,
          *  then parse the template values.  If LITERAL
          *  is ON then any values found in the parsing of
          *  the rules will be printed in the appropriate
          *  allowed list, else only the information found
          *  in the deftemplate structures will be printed.
          * ------------------------------------------------
          */

         IF(template) THEN
             ExamineTemplateFieldValues(temp_fld,template,defrelation);
             break;
         ELSE
           fprintf(fp,"\n   (field %d", temp_fld->field_num);      
         END_IF

         /* ---------------------------------------------------------
          *  In order to keep the allowed values for a field that
          *  were defined in a defrelation as well as those that
          *  are defined in a rule, the def_relation and relation
          *  lists must be parsed. (LITERAL option turned on)
          * ---------------------------------------------------------
          */

         IF ((defrelation) AND (!template)) THEN
            def_field = defrelation->fields;
            IF (def_field) THEN 
              WHILE (def_field) DO
                 IF (def_field->position EQ temp_fld->field_num) THEN
                   OutputRelationFieldInformation(def_field);
                 END_IF
                 def_field = def_field->next_field;
              END_WHILE
            ELSE
              ExamineFieldValues(temp_fld->lit_values);
            END_IF
         ELSE_IF((!defrelation)AND(!template))
	    ExamineFieldValues(temp_fld->lit_values);
         END_IF


         fprintf(fp,")");

         i        = temp_fld->field_num + 1;
         temp_fld = temp_fld->next_field;
      END_WHILE
      
      WHILE (i <= max) DO
         fprintf(fp,"\n   (field %d",i);
         IF ((defrelation) AND (!template)) THEN
           fprintf(fp,"\n      (type");
           def_field = defrelation->fields;
           IF (def_field) THEN
             WHILE (def_field) DO
               IF (def_field->position EQ i) THEN
                 IF (def_field->allow_word)
                    fprintf(fp," WORD");
                 IF (def_field->allow_string)
                    fprintf(fp," STRING");
                 IF (def_field->allow_instance)
                    fprintf(fp,"  INSTANCE");
                 IF (def_field->allow_integer)
                    fprintf(fp," INTEGER");
                 IF (def_field->allow_float)
                    fprintf(fp," FLOAT");
                 fprintf(fp,")");
                 IF (def_field->allow_word)
                    fprintf(fp,"\n      (allowed-words ?VARIABLE)");
                 IF (def_field->allow_string)
                    fprintf(fp,"\n      (allowed-strings ?VARIABLE)");
                 IF (def_field->allow_instance)
                    fprintf(fp,"\n      (allowed-instances ?VARIABLE)");
                 IF ((def_field->allow_integer) || (def_field->allow_float)) THEN
                    IF (def_field->set_int_max OR def_field->set_float_max
                        OR def_field->set_int_min OR def_field->set_float_min) THEN
                       fprintf(fp,"\n      (range");
                       IF (def_field->set_int_min AND def_field->set_float_min) THEN
                         if((float)def_field->set_int_min > def_field->set_float_min)
                            fprintf(fp," %5.2f",def_field->float_min_range);
                         else
                            fprintf(fp," %5d",def_field->int_min_range);
                       ELSE_IF(def_field->set_int_min)
                         fprintf(fp," %5d",def_field->int_min_range);
                       ELSE_IF(def_field->set_float_min)
                         fprintf(fp," %5.2f",def_field->float_min_range);
                       ELSE
                         fprintf(fp," ?VARIABLE");
                       END_IF
                       IF (def_field->set_int_max AND def_field->set_float_max) THEN
                         if((float)def_field->set_int_max > def_field->set_float_max)
                            fprintf(fp," %5.2f",def_field->float_max_range);
                         else
                            fprintf(fp," %5d",def_field->int_max_range);
                       ELSE_IF(def_field->set_int_max)
                         fprintf(fp," %5d",def_field->int_max_range);
                       ELSE_IF(def_field->set_float_max)
                         fprintf(fp," %5.2f",def_field->float_max_range);
                       ELSE
                         fprintf(fp," ?VARIABLE");
                       END_IF
                    ELSE
                       IF(def_field->allow_integer)THEN
                         fprintf(fp,"\n      (allowed-integers ?VARIABLE)");
                       END_IF
                       IF (def_field->allow_float) THEN
                         fprintf(fp,"\n      (allowed-floats ?VARIABLE)");
                       END_IF
                      fprintf(fp,"\n      (range ?VARIABLE ?VARIABLE)");
                    END_IF
                 END_IF
               END_IF
               def_field = def_field->next_field;
             END_WHILE
             fprintf(fp,")");
           ELSE
             fprintf(fp," ?VARIABLE)");
             fprintf(fp,"\n      (allowed-words ?VARIABLE)");
             fprintf(fp,"\n      (allowed-strings ?VARIABLE)");
             fprintf(fp,"\n      (allowed-instances ?VARIABLE)");
             fprintf(fp,"\n      (allowed-integers ?VARIABLE)");
             fprintf(fp,"\n      (allowed-floats ?VARIABLE)");
             fprintf(fp,"\n      (range ?VARIABLE ?VARIABLE)");
             fprintf(fp,")");
           END_IF
         ELSE
           fprintf(fp,"\n      (type ?VARIABLE)");
           fprintf(fp,"\n      (allowed-words ?VARIABLE)");
           fprintf(fp,"\n      (allowed-strings ?VARIABLE)");
           fprintf(fp,"\n      (allowed-instances ?VARIABLE)");
           fprintf(fp,"\n      (allowed-integers ?VARIABLE)");
           fprintf(fp,"\n      (allowed-floats ?VARIABLE)");
           fprintf(fp,"\n      (range ?VARIABLE ?VARIABLE)");
           fprintf(fp,")");
         END_IF
         i++;
      DONE

   ELSE_IF(min NEQ 1000 OR max NEQ 0)

      /*===========================*/
      /* Process unknown arguments */
      /*===========================*/
   
      for(i = 2; i <= max; i++) DO
         fprintf(fp,"\n   (field %d",i);
         IF ((defrelation) AND (!template)) THEN
           fprintf(fp,"\n      (type");
           def_field = defrelation->fields;
           IF (def_field) THEN
             WHILE (def_field) DO
               IF (def_field->position EQ i) THEN
                 IF (def_field->allow_word)
                    fprintf(fp," WORD");
                 IF (def_field->allow_string)
                    fprintf(fp," STRING");
                 IF (def_field->allow_instance)
                    fprintf(fp," INSTANCE");
                 IF (def_field->allow_integer)
                    fprintf(fp," INTEGER");
                 IF (def_field->allow_float)
                    fprintf(fp," FLOAT");
                 fprintf(fp,")");
                 IF (def_field->allow_word)
                    fprintf(fp,"\n      (allowed-words ?VARIABLE)");
                 IF (def_field->allow_string)
                    fprintf(fp,"\n      (allowed-strings ?VARIABLE)");
                 IF (def_field->allow_instance)
                    fprintf(fp,"\n      (allowed-instances ?VARIABLE)");
                 IF ((def_field->allow_integer) ||(def_field->allow_float))THEN
                    IF (def_field->set_int_max OR def_field->set_int_min
                       OR def_field->set_float_max OR def_field->set_float_min) THEN
                       fprintf(fp,"\n      (range");
                       IF (def_field->set_int_min AND def_field->set_float_min) THEN
                         if((float)def_field->set_int_min > def_field->set_float_min)
                            fprintf(fp," %5.2f",def_field->float_min_range);
                         else
                            fprintf(fp," %5d",def_field->int_min_range);
                       ELSE_IF(def_field->set_int_min)
                         fprintf(fp," %5d",def_field->int_min_range);
                       ELSE_IF(def_field->set_float_min)
                         fprintf(fp," %5.2f",def_field->float_min_range);
                       ELSE
                         fprintf(fp," ?VARIABLE");
                       END_IF
                       IF (def_field->set_int_max AND def_field->set_float_max) THEN
                         if((float)def_field->set_int_max > def_field->set_float_max)
                            fprintf(fp," %5.2f",def_field->float_max_range);
                         else
                            fprintf(fp," %5d",def_field->int_max_range);
                       ELSE_IF(def_field->set_int_max)
                         fprintf(fp," %5d",def_field->int_max_range);
                       ELSE_IF(def_field->set_float_max)
                         fprintf(fp," %5.2f",def_field->float_max_range);
                       ELSE
                         fprintf(fp," ?VARIABLE");
                       END_IF
                    ELSE
                      IF(def_field->allow_integer)
                         fprintf(fp,"\n      (allowed-integers ?VARIABLE)");
                      IF(def_field->allow_float)
                          fprintf(fp,"\n      (allowed-floats ?VARIABLE)");
                      fprintf(fp,"\n      (range ?VARIABLE ?VARIABLE)");
                    END_IF
                 END_IF
               END_IF
               def_field = def_field->next_field;
             END_WHILE
             fprintf(fp,")");
           ELSE
             fprintf(fp," ?VARIABLE)");
             fprintf(fp,"\n      (allowed-words ?VARIABLE)");
             fprintf(fp,"\n      (allowed-strings ?VARIABLE)");
             fprintf(fp,"\n      (allowed-instances ?VARIABLE)");
             fprintf(fp,"\n      (allowed-integers ?VARIABLE)");
             fprintf(fp,"\n      (allowed-floats ?VARIABLE)");
             fprintf(fp,"\n      (range ?VARIABLE ?VARIABLE)");
             fprintf(fp,")");
           END_IF
         ELSE
           fprintf(fp,"\n      (type ?VARIABLE)");
           fprintf(fp,"\n      (allowed-words ?VARIABLE)");
           fprintf(fp,"\n      (allowed-strings ?VARIABLE)");
           fprintf(fp,"\n      (allowed-instances ?VARIABLE)");
           fprintf(fp,"\n      (allowed-integers ?VARIABLE)");
           fprintf(fp,"\n      (allowed-floats ?VARIABLE)");
           fprintf(fp,"\n      (range ?VARIABLE ?VARIABLE)");
           fprintf(fp,")");
         END_IF
      END_FOR
   END_IF

   fprintf(fp,"\n   )");
}


/* ==================================================================== */

/*
 * ------------------------------------------------------------------
 *  NAME   :  OutputRelationFieldInformation()
 *
 *  PURPOSE:  This function is used to output defrelation info
 *            to a file specified with the -c option.  The
 *            function presently outputs allowed-number field
 *            values.  It takes the values from the def_relation
 *            list.  This function only works when there exists
 *            a defrelation statement for the current relation.
 *
 *  INPUTS :  A pointer to the field of the defrelation (DF_PTR)
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  This function only takes values from the def_relation
 *            list.  It will be expanded to take values from the
 *            relation list when no defrelations have been supplied.
 * ------------------------------------------------------------------
 */

static void OutputRelationFieldInformation(fields)
DF_PTR fields;
{
   IL_PTR num_list;
   FL_PTR fnum_list;
   WL_PTR wlist;
   int num_words = 0;
   int num_strings = 0;
   int num_instances = 0;
   int num_integers = 0;
   int num_floats = 0;

   /* ------------------------------
    *  Print out the field type(s)
    * ------------------------------
    */

   fprintf(fp,"\n      (type");

   IF (fields->allow_integer) THEN
     fprintf(fp," INTEGER");
     num_list = fields->possible_integers;
     WHILE (num_list) DO
       num_integers++;
       num_list = num_list->next_number;
     END_WHILE
   END_IF
   IF (fields->allow_float) THEN
     fprintf(fp," FLOAT");
     fnum_list = fields->possible_floats;
     WHILE (fnum_list) DO
       num_floats++;
       fnum_list = fnum_list->next_number;
     END_WHILE
   END_IF
   IF (fields->allow_word) THEN
     fprintf(fp," WORD");
     wlist = fields->possible_words;
     WHILE (wlist) DO
       num_words++;
       wlist = wlist->next_word;
     END_WHILE
   END_IF
   IF (fields->allow_string) THEN
     fprintf(fp," STRING");
     wlist = fields->possible_strings;
     WHILE (wlist) DO
       num_strings++;
       wlist = wlist->next_word;
     END_WHILE
   END_IF
   IF (fields->allow_instance) THEN
     fprintf(fp," INSTANCE");
     wlist = fields->possible_instances;
     WHILE (wlist) DO
       num_instances++;
       wlist = wlist->next_word;
     END_WHILE
   END_IF

   fprintf(fp,")");

   /* -------------------------------
    *  Print out the allowed values
    * -------------------------------
    */

   IF (fields->allow_string) THEN
      fprintf(fp,"\n      (allowed-strings");
      wlist = fields->possible_strings;
      IF (num_strings <= MAX_SINGLE_LIST) THEN
        IF (wlist) THEN
          WHILE(wlist) DO
            fprintf(fp," %s",wlist->word);
            wlist = wlist->next_word;
          END_WHILE
        ELSE
          fprintf(fp," ?VARIABLE");
        END_IF
      ELSE
        fprintf(fp," ?VARIABLE");
      END_IF
      fprintf(fp,")");
   END_IF

   IF (fields->allow_word) THEN
      fprintf(fp,"\n      (allowed-words");
      wlist = fields->possible_words;
      IF (num_words <= MAX_SINGLE_LIST) THEN
        IF (wlist) THEN
          WHILE(wlist) DO
            fprintf(fp," %.40s",wlist->word);
            wlist = wlist->next_word;
          END_WHILE
        ELSE
          fprintf(fp," ?VARIABLE");
        END_IF
      ELSE
        fprintf(fp," ?VARIABLE");
      END_IF
      fprintf(fp,")");
   END_IF
   IF (fields->allow_instance) THEN
      fprintf(fp,"\n      (allowed-instances");
      wlist = fields->possible_instances;
      IF (num_instances <= MAX_SINGLE_LIST) THEN
        IF (wlist) THEN
          WHILE(wlist) DO
            fprintf(fp," %s",wlist->word);
            wlist = wlist->next_word;
          END_WHILE
        ELSE
          fprintf(fp," ?VARIABLE");
        END_IF
      ELSE
        fprintf(fp," ?VARIABLE");
      END_IF
      fprintf(fp,")");
   END_IF
   IF ((fields->allow_integer) || (fields->allow_float))THEN
      IF (fields->set_int_min OR fields->set_int_max OR 
          fields->set_float_min OR fields->set_float_max) THEN
         fprintf(fp,"\n      (range");
         IF (fields->set_int_min AND fields->set_float_min) THEN
           if((float)fields->set_int_min > fields->set_float_min)
              fprintf(fp," %5.2f",fields->float_min_range);
           else
              fprintf(fp," %5d",fields->int_min_range);
         ELSE_IF(fields->set_int_min)
           fprintf(fp," %5d",fields->int_min_range);
         ELSE_IF(fields->set_float_min)
           fprintf(fp," %5.2f",fields->float_min_range);
         ELSE
           fprintf(fp," ?VARIABLE");
         END_IF
         IF (fields->set_int_max AND fields->set_float_max) THEN
           if((float)fields->set_int_max > fields->set_float_max)
              fprintf(fp," %5.2f",fields->float_max_range);
           else
              fprintf(fp," %5d",fields->int_max_range);
         ELSE_IF(fields->set_int_max)
           fprintf(fp," %5d",fields->int_max_range);
         ELSE_IF(fields->set_float_max)
           fprintf(fp," %5.2f",fields->float_max_range);
         ELSE
           fprintf(fp," ?VARIABLE");
         END_IF
      ELSE  /* no range specified */
        IF(fields->allow_integer) THEN
          fprintf(fp,"\n      (allowed-integers");
          num_list = fields->possible_integers;
          IF (num_list) THEN
            IF (num_integers <= MAX_SINGLE_LIST) THEN
              WHILE(num_list) DO
               fprintf(fp," %d",num_list->number);
               num_list = num_list->next_number;
              END_WHILE
              fprintf(fp,")");
             ELSE
              fprintf(fp," ?VARIABLE)");
             END_IF
           END_IF
         END_IF /* end of fields->allow-integer */
         IF(fields->allow_float) THEN
          fprintf(fp,"\n      (allowed-floats");
          fnum_list = fields->possible_floats;
          IF (fnum_list) THEN
            IF (num_floats <= MAX_SINGLE_LIST) THEN
              WHILE(fnum_list) DO
               fprintf(fp," %3.2f",fnum_list->number);
               fnum_list = fnum_list->next_number;
              END_WHILE
              fprintf(fp,")");
             ELSE
              fprintf(fp," ?VARIABLE)");
             END_IF
           END_IF /* End fnum_list */
          END_IF /* end field->allow_float */
         END_IF /* end no range specified */
        ELSE
          fprintf(fp,"\n      (allowed-integers");
          fprintf(fp," ?VARIABLE)");
          fprintf(fp,"\n      (allowed-floats");
          fprintf(fp," ?VARIABLE)");
          fprintf(fp,"\n      (range ?VARIABLE ?VARIABLE)");
        END_IF
       /* End field->allow_float || field->allow_integer */
}


/* =========================================================================
 * =========================================================================
 *                   Counting fields functions
 *
 *   These functions count the number of fields used for this relation in
 *   various rules, deffacts, and asserts.
 *   They currently arn't too smart, merely returning the smallest and
 *   biggest numbers found. They should be more intelligent about how
 *   many times that number of fields is used.
 * =========================================================================
 * =========================================================================
 */

/*
 * -------------------------------------------------------------
 *  NAME   :  CountRRFields()
 *
 *  PURPOSE:  This function searches a rule_relation list for
 *            the minimum and maximum number of fields used.
 *
 *  INPUTS :  A pointer to the rule_relation list (RR_PTR), and
 *            integer pointers to the min and max currently
 *            found (int *).
 *
 *  RETURNS:  Nothing, but it can modify the min/max pointers.
 *
 *  NOTES  :  Called many times for the various rr lists.
 * -------------------------------------------------------------
 */

static void CountRRFields(rr_list, min, max)
RR_PTR   rr_list;
int     *min, *max;
{
   RR_PTR  temp;
   
   temp = rr_list;
   while(temp NEQ NULL) DO
      IF(CHECK_DEBUG IS_ON) THEN
         sprintf(MsgBuf, "\nInside rr, : %5d, max: %5d, obj: %s (%5d)", 
                   *min, *max, temp->rule->name, temp->num_fields);
         SendMessage(MsgBuf,NO);
      END_IF
      IF(abs(temp->num_fields) < *min) THEN
         *min = abs(temp->num_fields);
      END_IF

      IF(abs(temp->num_fields) > *max) THEN
         *max = abs(temp->num_fields);
      END_IF

      temp = temp->next_rr;
   END_WHILE
}

/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  count_fbr_fileds()
 *
 *  PURPOSE:  This function searches a fact block structure for
 *            the minimum and maximum number of fields used.
 *
 *  INPUTS :  A pointer to the fact block list (FBR_PTR), and
 *            integer pointers to the min and max currently
 *            found (int *).
 *
 *  RETURNS:  Nothing, but it can modify the min/max pointers.
 *
 *  NOTES  :  
 * -------------------------------------------------------------
 */

static void CountFBRFields(fbr_list, min, max)
FBR_PTR  fbr_list;
int     *min, *max;
{
   FBR_PTR  temp;
   
   temp = fbr_list;
   while(temp NEQ NULL) DO
      IF(CHECK_DEBUG IS_ON) THEN
         sprintf(MsgBuf, "\nInside fbr, min: %5d, max: %5d, obj: %5d", 
                *min, *max, temp->num_fields);
         SendMessage(MsgBuf,NO);
      END_IF

      IF(abs(temp->num_fields) < *min) THEN
         *min = temp->num_fields;
      END_IF

      IF(abs(temp->num_fields) > *max) THEN
         *max = temp->num_fields;
      END_IF

      temp = temp->next_fbr;
   END_WHILE
}

/* ========================================================================= */

/*
 * -------------------------------------------------------------------
 *  NAME   :  ExamineFieldValues()
 *
 *  PURPOSE:  This function examines a word list to determine
 *            type, range and allowed values.
 *
 *  INPUTS :  A pointer to list of literal values used in
 *            this field.
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  This function is also somewhat stupid about what
 *            it does. Eventually need to implement some of the 
 *            decisions in a rule base to improve the quality of 
 *            the 'guesses' in areas such as type or allowed values.
 * -------------------------------------------------------------------
 */

void ExamineFieldValues(list)
WORD_LIST *list;
{
   COUNTER   *counts;
   int        length, num_words, num_strings,num_instances, num_integers, num_floats;

   counts      = AllocCounter();
   
   /*================================================*/
   /* Decide what type this field is and gather info */
   /* on number of times each type is used.          */
   /*================================================*/

   (void)DetermineType(list, counts);

   length       = counts->prime_count;
   num_words    = counts->count_A;
   num_strings  = counts->count_B;
   num_integers = counts->count_C;
   num_floats   = counts->count_D;
   num_instances = counts->count_E;
   
 /*  IF(CHECK_DEBUG IS_ON) THEN
      sprintf(MsgBuf, "\nType of list is %d and length is %d", type, length);
      SendMessage(MsgBuf,NO);
   END_IF
  */

   IF((num_words > 0) && (num_strings > 0) &&
      (num_integers > 0) && (num_floats > 0)
      && (num_instances > 0)) THEN
        fprintf(fp,"\n      (type ?VARIABLE)");
   ELSE_IF ((num_words > 0) || (num_strings > 0)||
      (num_integers > 0)|| (num_floats > 0)
      || (num_instances > 0)) 
        fprintf(fp,"\n      (type ");
        if(num_integers > 0)
           fprintf(fp," INTEGER");
        if(num_floats > 0)
           fprintf(fp," FLOAT");
        if(num_words > 0)
          fprintf(fp," WORD");
        if (num_strings > 0)
          fprintf(fp," STRING");
        if (num_instances > 0)
          fprintf(fp," INSTANCE");
        fprintf(fp, " )");
   ELSE
    fprintf(fp,"\n      (type ?VARIABLE)");
   END_IF
   IF(num_words > 0) THEN
      fprintf(fp,"\n      (allowed-words ");
      if(num_words <= MAX_SINGLE_LIST)
         OutputListItems(list,WORD);
      else
        fprintf(fp, " ?VARIABLE");
      fprintf(fp,")");
   END_IF
   IF (num_strings > 0) THEN
      fprintf(fp,"\n      (allowed-strings");
      if(num_strings <= MAX_SINGLE_LIST)
         OutputListItems(list,STRING);
      else
        fprintf(fp, " ?VARIABLE");
      fprintf(fp,")");
   END_IF
   IF (num_instances > 0) THEN
      fprintf(fp,"\n      (allowed-instances");
      if(num_strings <= MAX_SINGLE_LIST)
         OutputListItems(list,INSTANCE);
      else
        fprintf(fp, " ?VARIABLE");
      fprintf(fp,")");
   END_IF
   IF (num_integers > 0) THEN
      fprintf(fp,"\n      (allowed-integers");
      if(num_integers <= MAX_SINGLE_LIST)
         OutputListItems(list,INTEGER);
      else
        fprintf(fp, " ?VARIABLE");
      fprintf(fp,")");
   END_IF
   IF(num_floats > 0) THEN
      fprintf(fp,"\n      (allowed-floats");
      if(num_floats <= MAX_SINGLE_LIST)
         OutputListItems(list,FLOAT);
      else
        fprintf(fp, " ?VARIABLE");
      fprintf(fp,")");
   END_IF
   IF((num_integers > MAX_SINGLE_LIST) || (num_floats > MAX_SINGLE_LIST))
       DetermineRange(list);
   
   
}

/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  ExamineTemplateFieldValues()
 *
 *  PURPOSE:  This function will determine the type(s) of a 
 *            multi-field variable as well as all the allowed
 *            values for each type.
 *
 *  INPUTS :  A pointer to the field_list of a relation.
 *            A pointer to the template that created the relation.
 *            A pointer to the defrelation that holds all
 *              pertinent information.
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  What is printed depends on if LITERAL is On or Off.
 * -------------------------------------------------------------
 */

#define BLOCKSIZE 80
#define STRCAT(a,b) strcat (strcat (a, " "), b)

void ExamineTemplateFieldValues(flist,template,defrel)
FIELD *flist;
DT_PTR template;
DR_PTR defrel;
{
  WL_PTR  wlist;
  IL_PTR  nlist;
  FL_PTR  floatlist;
  DS_PTR  slot;
  DF_PTR  field;
  int     MULTI = 0;
  int     TYPE = 0;
  int     str_size = BLOCKSIZE;
  int     num_size = BLOCKSIZE;
  int	  fnum_size = BLOCKSIZE;
  int     word_size = BLOCKSIZE;
  char    *fnc_name = "examine_multi_field_values";
  char    *new_word_vals = GenAlloc(word_size,fnc_name);
  char    *new_str_vals  = GenAlloc(str_size,fnc_name);
  char    *new_num_vals  = GenAlloc(num_size,fnc_name);
  char    *new_fnum_vals = GenAlloc(fnum_size,fnc_name);
  char    fl_buf[20];

  /* ========================================================
   *  Allocate memory to create the lists for allowed words,
   *  allowed strings, and allowed numbers.  Allocate in BLOCKSIZE
   *  byte chunks.  When current chunk is exceeded, allocate
   *  new chunk which is BLOCKSIZE + old chunk size and free old
   *  chunk.
   * =========================================================
   */

  *new_word_vals = '\0';
  *new_str_vals = '\0';
  *new_num_vals = '\0';
  *new_fnum_vals = '\0';

  WHILE (flist) DO

    /* ==========================================================
     *  Process the template relations found in the rules first.
     *  If the type is a WORD then bit-wise OR a 1 with the local
     *  variable TYPE.  If type is a STRING OR a 2, a NUMBER OR
     *  a 4.  In this manner, the variable TYPE will be able to
     *  hold any/all of the types the multi-field can have.
     * ==========================================================
     */


    slot = FindDefslotByPosition(template,flist->field_num);
    IF(((slot->max_elements NEQ 1) OR (slot->min_elements NEQ 1)) AND
      (MULTI EQ 0)) THEN
      MULTI = 1;
      fprintf(fp,"\n   (multi-field %.40s",slot->name);
    ELSE_IF (MULTI NEQ 1)
      fprintf(fp,"\n   (field %.40s",slot->name);
    END_IF
      
    /* ===================================================
     *  The information to be printed will come from the 
     *  defrelation structure.  If LITERAL is On, then
     *  values were added to the allowed lists of def-
     *  relation during the parsing of the rules, else
     *  defrelation holds only that information found
     *  in the previously defined defrelation or def-
     *  template structures.
     * ====================================================
     */

    field = defrel->fields;
    IF (field) THEN
       WHILE (field) DO
          IF (field->position EQ flist->field_num) THEN
            IF (field->allow_word) THEN
              TYPE |= 1;
            ELSE_IF (field->allow_string)
              TYPE |= 2;
            ELSE_IF (field->allow_integer)
              TYPE |= 4;
            ELSE_IF (field->allow_float)
              TYPE |= 8;
            END_IF

            IF (field->possible_words) THEN
               wlist = field->possible_words;
               WHILE (wlist) DO
                 IF ((strlen(new_word_vals)+strlen(wlist->word)+1) >=
                          word_size) THEN
                    char *old_word_vals = new_word_vals;
                    new_word_vals = GenAlloc(word_size+BLOCKSIZE,fnc_name);
                    *new_word_vals = '\0';
                    strcpy (new_word_vals, old_word_vals);
                    GenFree(old_word_vals,word_size);
                    word_size += BLOCKSIZE;
                  END_IF
                  STRCAT (new_word_vals, wlist->word);
                  wlist = wlist->next_word;
               END_WHILE
            END_IF
            IF (field->possible_strings) THEN
               wlist = field->possible_strings;
               WHILE (wlist) DO
                  IF ((strlen(new_str_vals)+strlen(wlist->word)+1) >=
                      str_size) THEN
                     char *old_str_vals = new_str_vals;
                     new_str_vals = GenAlloc(str_size+BLOCKSIZE,fnc_name);
                     *new_str_vals = '\0';
                     strcpy (new_str_vals, old_str_vals);
                     GenFree(old_str_vals,str_size);
                     str_size += BLOCKSIZE;
                   END_IF
                   STRCAT (new_str_vals, wlist->word);
                   wlist = wlist->next_word;
                END_WHILE
             END_IF
             IF (field->possible_integers) THEN
                nlist = field->possible_integers;
                WHILE (nlist) DO
                   sprintf(fl_buf,"%f",nlist->number);
                   IF ((strlen(new_num_vals)+strlen(fl_buf)+1) >=
                        num_size) THEN
                     char *old_num_vals = new_num_vals;
                     new_num_vals = GenAlloc(num_size+BLOCKSIZE,fnc_name);
                     *new_num_vals = '\0';
                     strcpy (new_num_vals, old_num_vals);
                     GenFree(old_num_vals,num_size);
                     num_size += BLOCKSIZE;
                   END_IF
                   STRCAT (new_num_vals, fl_buf);
                   nlist = nlist->next_number;
                END_WHILE
             END_IF
             IF (field->possible_floats) THEN
                floatlist = field->possible_floats;
                WHILE (floatlist) DO
                   sprintf(fl_buf,"%f",floatlist->number);
                   IF ((strlen(new_fnum_vals)+strlen(fl_buf)+1) >=
                        fnum_size) THEN
                     char *old_num_vals = new_fnum_vals;
                     new_fnum_vals = GenAlloc(fnum_size+BLOCKSIZE,fnc_name);
                     *new_fnum_vals = '\0';
                     strcpy (new_fnum_vals, old_num_vals);
                     GenFree(old_num_vals,fnum_size);
                     fnum_size += BLOCKSIZE;
                   END_IF
                   STRCAT (new_fnum_vals, fl_buf);
                   floatlist = floatlist->next_number;
                END_WHILE
             END_IF
          END_IF
          field = field->next_field;
       END_WHILE
    END_IF


    IF((MULTI EQ 0) OR (flist->next_field EQ NULL)) THEN

      fprintf(fp,"\n      (type");

      /* ================================================
       *  To remove the type, bit-wise AND the variable
       *  TYPE with the three defined values and print
       *  out the appropriate type.
       * =================================================
       */

      IF ((TYPE & 1) EQ 1)
        fprintf(fp," WORD");
      IF ((TYPE & 2) EQ 2)
        fprintf(fp," STRING");
      IF ((TYPE & 4) EQ 4)
        fprintf(fp," INTEGER");
      IF ((TYPE & 8) EQ 8)
        fprintf(fp," FLOAT");
      fprintf(fp,")");

      /* ===================================================
       *  If the variables which hold the allowed types 
       *  have a strlen greater than zero (they contain at
       *  least one value) then printout the list.
       *  LITERAL option must be turned on.
       * ====================================================
       */


      IF (strlen(new_word_vals)) 
         fprintf(fp,"\n      (allowed-words%s)",new_word_vals);
      IF (strlen(new_num_vals))
         fprintf(fp,"\n      (allowed-integers%s)",new_num_vals);
      IF (strlen(new_fnum_vals))
         fprintf(fp,"\n      (allowed-floats%s)",new_fnum_vals); 
      IF (strlen(new_str_vals)) 
         fprintf(fp,"\n      (allowed-strings%s)",new_str_vals);

      *new_word_vals = '\0';
      *new_num_vals = '\0';
      *new_fnum_vals = '\0';
      *new_word_vals = '\0';

      PrintTemplateDefaults(slot->defaults);
    END_IF

    flist = flist->next_field;
  END_WHILE

      /* ===================================================
       *  Free the memory used to hold the allowed types
       * ===================================================
       */

      GenFree(new_word_vals,word_size);
      GenFree(new_num_vals,num_size);
      GenFree(new_fnum_vals,fnum_size);
      GenFree(new_str_vals,str_size);

}

/* ============================================================ */

/*
 * --------------------------------------------------------------
 *
 *    NAME   : PrintTemplateDefaults()
 *
 *    PURPOSE : This function will print out the default values
 *              for a template when the -c option is turned on.
 * 
 *     INPUTS : Pointer to the defaults list of a template.
 *
 *    RETURNS : Nothing.
 *
 *      NOTES : None.
 *
 * --------------------------------------------------------------
 */

void PrintTemplateDefaults(def_list)
TOKEN_NODE *def_list;
{
  IF (def_list) THEN
    fprintf(fp,"\n      (default");

    while(def_list) {
      fprintf(fp," %s",def_list->token->tknword);
      def_list = def_list->next;
    }
    fprintf(fp,")");
  ELSE
    fprintf(fp,"\n      (default ?NONE)");
  END_IF
    fprintf(fp,")");
}



/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  DetermineType()
 *
 *  PURPOSE:  This function searches a list of words and tries
 *            to determine what type (NUMBER, WORD, STRING) the 
 *            list best corresponds to.
 *
 *  INPUTS :  A pointer to the list of words (WORD_LIST *) and
 *            a pointer to a counter structure for storing
 *            count information (COUNTER *).
 *
 *  RETURNS:  An integer, the type this list most likely
 *            corresponds too. Valid returns are WORD, STRING,
 *            NUMBER, WORD_OR_NUMBER, WORD_OR_STRING, VARIABLE,
 *            and STRING_OR_NUMBER.
 *
 *  NOTES  :  
 * -------------------------------------------------------------
 */

int DetermineType(list, count)
WORD_LIST *list;
COUNTER   *count;
{
   WORD_LIST  *temp;
   
   count->prime_count = 0;           /* Number of items in list */
   count->count_A     = 0;           /* Number of WORDs in list */
   count->count_B     = 0;           /* Number of STRINGs in list */
   count->count_C     = 0;           /* Number of INTEGERs in list */
   count->count_D     = 0;           /* Number of FLOATs in list */
   count->count_E     = 0;           /* Number of INSTANCEs in list */

   /*============================================*/
   /* Loop through list counting type occurences */
   /*============================================*/

   temp = list;
   WHILE (temp) DO
      count->prime_count++;

      switch (temp->type) {

         case WORD:
              count->count_A++;
              break;

         case STRING:
              count->count_B++;
              break;

         case INTEGER:
              count->count_C++;
              break;
              
         case FLOAT:
              count->count_D++;
              break;
         
         case INSTANCE:
              count->count_E++;
              break;
         case UNKNOWN:
         case GLOBAL:
              break;
         default:
              ErrorMessage(ERROR, "Unknown type in word list");
              IF(CHECK_DEBUG IS_ON) THEN
                 sprintf(MsgBuf, "\nType is: %d, word is: %.40s", temp->type, 
                                                            temp->word);
                 SendMessage(MsgBuf,NO);
              END_IF
              break;
         }

      temp = temp->next_word;
   END_WHILE
  
}

/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  OutputListItems()
 *
 *  PURPOSE:  This function prints a list of the allowed items
 *            by type. Each word of the allowed type is printed
 *            to a file.
 *
 *  INPUTS :  A pointer to the list of words (WORD_LIST *) and
 *            a flag identifying what type to print.
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  
 * -------------------------------------------------------------
 */

void OutputListItems(list, type)
WORD_LIST *list;
int        type;
{
   WORD_LIST  *temp;

   /*============================================*/
   /* Loop through list printing type occurences */
   /*============================================*/

   temp = list;
   while(temp NEQ NULL) DO
      IF(temp->type EQ type) THEN
         fprintf(fp," %s", temp->word);
      END_IF;

      temp = temp->next_word;
   END_WHILE
   
}

/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  DetermineRange()
 *
 *  PURPOSE:  This function searches a list of words and tries
 *            to determine what range the numbers represent and
 *            then outputs range info the a file.
 *
 *  INPUTS :  A pointer to the list of words (WORD_LIST *).
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  Very simplistic for now. Should eventually do 
 *            something with decimal places or number of 
 *            different values to determine whether or not a 
 *            range field is appropriate.
 * -------------------------------------------------------------
 */

/*ARGSUSED*/ /* Make lint happy */
void DetermineRange(list)
WORD_LIST *list;
{
   fprintf(fp, "\n      (range ?VARIABLE ?VARIABLE)");
}


/* ========================================================================= */
/*                    CREATE DEFEXTERNAL FUNCTIONS                           */
/* ========================================================================= */

/*
 * -------------------------------------------------------------
 *  NAME   : CreateDefextFile()
 *
 *  PURPOSE: This function creates a file with defexternal info
 *           for each of the known external functions. It makes 
 *           a 'best' guess at some of the appropriate info.
 *
 *  INPUTS : Two arguments, a pointer to the head of the
 *           external functions list (EX_FUNC *), and a pointer
 *           to the name of the file for defexternal info
 *           (char *).
 *
 *  RETURNS: Nothing.
 *
 *  NOTES  : This function assumes that the defrelation info
 *           has already been put into the same file and
 *           appends the defexternal info.
 * -------------------------------------------------------------
 */ 

void CreateDefextFile(func_list, file_name)
char      *file_name;
EX_FUNC   *func_list;
{
   /*========================*/
   /* Open defexternals file */
   /*========================*/

#if MAC_LSC
   GotoDefrelationsDirectory();
#endif

   fp = fopen(file_name, "a");
   IF(fp EQ NULL) THEN

      /*=================================================*/
      /* If file can't be opened, print error and return */
      /*=================================================*/

      sprintf(MsgBuf, "Unable to open file %.20s", file_name);
      ErrorMessage(ERROR, MsgBuf);
      return;
   END_IF
   
   /*=============================*/
   /* Loop through functions list */
   /*=============================*/

   LoopThroughFuntions(func_list);

   fprintf(fp,"\n");
   fclose(fp);
   
#if MAC_LSC
   SetDefrelationsFileCreator();
#endif
}

/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  LoopThroughFuntions()
 *
 *  PURPOSE:  This is a recursive function which processes
 *            through all the nodes of the binary tree of
 *            external functions.
 *
 *  INPUTS :  A pointer to the list of functions (EX_FUNC *)
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  Actual output of relations info is in the
 *            OutputDefexternaIInformation function.
 * -------------------------------------------------------------
 */

static void LoopThroughFuntions(node)
EX_FUNC *node;
{
   IF(node NEQ NULL) THEN
     LoopThroughFuntions(node->lft_ex_func);
     OutputDefexternaIInformation(node);
     LoopThroughFuntions(node->rht_ex_func);
   END_IF
}

/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  OutputDefexternaIInformation()
 *
 *  PURPOSE:  This function writes all the defexternal info
 *            to a file.
 *
 *  INPUTS :  A pointer to the current function (EX_FUNC *)
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  
 * -------------------------------------------------------------
 */

static void OutputDefexternaIInformation(node)
EX_FUNC *node;
{
   FIELD     *temp_fld;
   int        min = 1000;
   int        max = 0;
   register   i;
   
   IF(VERBOSE IS_ON) THEN
      sprintf(MsgBuf, "\nCreating defexternal for %.40s", node->name);
      SendMessage(MsgBuf,NO);
   END_IF
   
   fprintf(fp,"\n\n(defexternal %s", node->name);

   /*================================================================*/
   /* Search various lists to determine min and max number of fields */
   /*================================================================*/
   
   CountRRFields(node->LHS_rule_list, &min, &max);
   CountRRFields(node->RHS_rule_list, &min, &max);

   fprintf(fp,"\n   (min-number-of-args %d)", min);
   fprintf(fp,"\n   (max-number-of-args %d)", max);

   /*===================================================*/
   /* Output default Info which can't be easily derived */
   /*===================================================*/
   
   fprintf(fp,"\n   (asserts ?NONE)");
   fprintf(fp,"\n   (retracts ?NONE)");
   fprintf(fp,"\n   (return-type NUMBER)");

   /*=======================*/
   /* Process argument list */
   /*=======================*/
   
   temp_fld = node->argument_list;
   IF(temp_fld NEQ NULL) THEN

      /*=========================*/
      /* Process known arguments */
      /*=========================*/
   
      node->argument_list= SortFieldList(node->argument_list);

      temp_fld = node->argument_list;
      i = 1;
      while(temp_fld NEQ NULL) DO
         while (temp_fld->field_num > i) DO
            fprintf(fp,"\n   (argument %d", i++);
            fprintf(fp,"\n      (type ?VARIABLE)");
            fprintf(fp,"\n      (allowed-words ?VARIABLE)");
            fprintf(fp,"\n      (allowed-strings ?VARIABLE)");
            fprintf(fp,"\n      (allowed-integers ?VARIABLE)");
            fprintf(fp,"\n      (allowed-floats ?VARIABLE)");
            fprintf(fp,"\n      (range ?VARIABLE ?VARIABLE)");
            fprintf(fp,")");
         DONE

         fprintf(fp,"\n   (argument %d", temp_fld->field_num);
         ExamineFieldValues(temp_fld->lit_values);
         fprintf(fp,")");

         i        = temp_fld->field_num + 1;
         temp_fld = temp_fld->next_field;
      END_WHILE

      while(i <= max) DO
            fprintf(fp,"\n   (argument %d", i++);
            fprintf(fp,"\n      (type ?VARIABLE)");
            fprintf(fp,"\n      (allowed-words ?VARIABLE)");
            fprintf(fp,"\n      (allowed-strings ?VARIABLE)");
            fprintf(fp,"\n      (allowed-integers ?VARIABLE)");
            fprintf(fp,"\n      (allowed-integers ?VARIABLE)");
            fprintf(fp,"\n      (range ?VARIABLE ?VARIABLE)");
            fprintf(fp,")");
      DONE

   ELSE_IF(min NEQ 1000 OR max NEQ 0)

      /*===========================*/
      /* Process unknown arguments */
      /*===========================*/
   
      for(i = 1; i <= max; i++) DO
         fprintf(fp,"\n   (argument  %d", i);
         fprintf(fp,"\n      (type ?VARIABLE)");
         fprintf(fp,"\n      (allowed-words ?VARIABLE)");
         fprintf(fp,"\n      (allowed-strings ?VARIABLE)");
         fprintf(fp,"\n      (allowed-integers ?VARIABLE)");
         fprintf(fp,"\n      (allowed-floats ?VARIABLE)");
         fprintf(fp,"\n      (range ?VARIABLE ?VARIABLE)");
         fprintf(fp,")");
      END_FOR
   END_IF
         
   fprintf(fp,"\n   )");
}

