   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                     BSAVE MODULE                    */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _BSAVE_SOURCE_

#include "setup.h"

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "clipsmem.h"
#include "constant.h"
#include "symbol.h"
#include "expressn.h"
#include "evaluatn.h"
#include "router.h"
#include "utility.h"

#if DEFGENERIC_CONSTRUCT
#include "genrcbin.h"
#endif

#if DEFFUNCTION_CONSTRUCT
#include "deffnctn.h"
#endif

#include "bsave.h"
#include "bload.h"

/***************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/***************************************/

#if ANSI_COMPILER  
#if BLOAD_AND_BSAVE
   static VOID                        FindNeededFunctionsAndAtoms(void);
   static VOID                        WriteNeededFunctions(FILE *);
   static unsigned long int           FunctionBinarySize(void);
   static VOID                        BsaveAllExpressions(FILE *);
   static VOID                        MarkNeededFlags(void);
   static VOID                        MarkBuckets(void);
   static VOID                        UnmarkBuckets(void);
   static VOID                        WriteNeededSymbols(FILE *);
   static VOID                        WriteNeededFloats(FILE *);
   static VOID                        WriteNeededIntegers(FILE *);
   static VOID                        WriteBinaryHeader(FILE *);
   static VOID                        WriteBinaryFooter(FILE *);
#endif
#else
#if BLOAD_AND_BSAVE
   static VOID                        FindNeededFunctionsAndAtoms();
   static VOID                        WriteNeededFunctions();
   static unsigned long int           FunctionBinarySize();
   static VOID                        BsaveAllExpressions();
   static VOID                        MarkNeededFlags();
   static VOID                        MarkBuckets();
   static VOID                        UnmarkBuckets();
   static VOID                        WriteNeededSymbols();
   static VOID                        WriteNeededFloats();
   static VOID                        WriteNeededIntegers();
   static VOID                        WriteBinaryHeader();
   static VOID                        WriteBinaryFooter();
#endif
#endif
   
/****************************************/
/* GLOBAL INTERNAL VARIABLE DEFINITIONS */
/****************************************/

   globle unsigned long int          ExpressionCount;
   globle struct BinaryItem         *ListOfBinaryItems = NULL;

/**************************************************/
/* BsaveCommand: Handles top level bsave command. */
/**************************************************/
globle VOID BsaveCommand()
  {
#if (! RUN_TIME) && BLOAD_AND_BSAVE
   char *fileName;
   
   if (ArgCountCheck("bsave",EXACTLY,1) == -1) return;
   fileName = GetFileName("bsave",1);
   if (fileName != NULL) Bsave(fileName);
#endif
  }
  
#if BLOAD_AND_BSAVE

/********************************************************/
/* Bsave: Stores the binary representation of the CLIPS */
/*   environment to a file.                             */
/********************************************************/
globle BOOLEAN Bsave(fileName)
  char *fileName;
  {
   FILE *fp;
   struct BinaryItem *biPtr;
   char constructBuffer[CONSTRUCT_HEADER_SIZE];
    
   if (Bloaded())
     {
      PrintCLIPS(WERROR,"Cannot perform a binary save while a binary load is in effect.\n");
      return(0);
     }

   /*================*/
   /* Open the file. */
   /*================*/

   if ((fp = fopen(fileName,"wb")) == NULL)
     {
      OpenErrorMessage("bsave",fileName);
      return(0);
     }
     
   /*==================================*/
   /* Write binary header to the file. */
   /*==================================*/
   
   WriteBinaryHeader(fp);
   
   /*=============================*/
   /* Initialize count variables. */
   /*=============================*/
   
   ExpressionCount = 0;
   MarkNeededFlags();
   FindNeededFunctionsAndAtoms();
   MarkBuckets();
   
   /*=====================*/
   /* Save the functions. */
   /*=====================*/
   
   WriteNeededFunctions(fp);
   WriteNeededSymbols(fp);
   WriteNeededFloats(fp);
   WriteNeededIntegers(fp);
   
   /*===================*/
   /* Save expressions. */
   /*===================*/
   
   BsaveAllExpressions(fp);
   
   /*======================*/
   /* Save the constructs. */
   /*======================*/
   
   ExpressionCount = 0;
     
   biPtr = ListOfBinaryItems;
   while (biPtr != NULL)
     {
      if (biPtr->bsaveFunction != NULL)
        {
         strncpy(constructBuffer,biPtr->name,CONSTRUCT_HEADER_SIZE);
         GenWrite(constructBuffer,(unsigned long) CONSTRUCT_HEADER_SIZE,fp);
         (*biPtr->bsaveFunction)(fp);
        }
      biPtr = biPtr->next;
     }
   
   /*===========*/
   /* Clean up. */
   /*===========*/
   
   UnmarkBuckets();
   
   /*=================*/
   /* Close the file. */
   /*=================*/
   
   WriteBinaryFooter(fp);
   fclose(fp);
   
   /*==================================*/
   /* Return CLIPS_TRUE to indicate success. */
   /*==================================*/
   
   return(CLIPS_TRUE);
  }
  
/********************************************************************/
/* FindNeededFunctionsAndAtoms: Searches through the constructs for */
/*   the functions used in expressions. This routine also           */
/*   counts the number of expressions in use (through a global)     */
/********************************************************************/
static VOID FindNeededFunctionsAndAtoms()
  {
   struct BinaryItem *biPtr;
     
   biPtr = ListOfBinaryItems;
   while (biPtr != NULL)
     {
      if (biPtr->findFunction != NULL) (*biPtr->findFunction)();
      biPtr = biPtr->next;
     }
  }
    
/****************************************************/
/* WriteNeededFunctions: Writes the names of needed */
/*   functions to the binary save file.             */
/****************************************************/
static VOID WriteNeededFunctions(fp)
  FILE *fp;
  {
   unsigned long int space, count = 0, length;
   struct FunctionDefinition *functionList;
  
   /*==============================================*/
   /* Assign each function an index if it is used. */
   /*==============================================*/
   
   functionList = GetFunctionList();
   
   while (functionList != NULL)
     {
      if (functionList->bsaveIndex) 
        functionList->bsaveIndex = (short int) count++;
      else functionList->bsaveIndex = -1;
      functionList = functionList->next;
     }
     
   /*=======================================================*/
   /* Determine the number of function names to be written. */
   /*=======================================================*/
   
   GenWrite(&count,(unsigned long) sizeof(unsigned long int),fp);
   if (count == 0) 
     {
      GenWrite(&count,(unsigned long) sizeof(unsigned long int),fp);
      return;
     }
   
   /*=========================================================*/
   /* Determine the amount of space needed for the functions. */
   /*=========================================================*/
   
   space = FunctionBinarySize();  
   GenWrite(&space,(unsigned long) sizeof(unsigned long int),fp);
   
   /*===============================*/
   /* Write out the function names. */
   /*===============================*/
   
   functionList = GetFunctionList();
   
   while (functionList != NULL)
     {
      if (functionList->bsaveIndex >= 0) 
        { 
         length = strlen(ValueToString(functionList->callFunctionName)) + 1;
         GenWrite(ValueToString(functionList->callFunctionName),(unsigned long) length,fp);
        }
      functionList = functionList->next;
     }
  }
  
/****************************************************************/
/* FunctionBinarySize: Determines the number of bytes needed to */
/*   save all of the function names in the binary save file.    */
/****************************************************************/
static unsigned long int FunctionBinarySize()
  {
   unsigned long int size = 0;
   struct FunctionDefinition *functionList;
   
   functionList = GetFunctionList();
   
   while (functionList != NULL)
     {
      if (functionList->bsaveIndex >= 0) 
        { size += strlen(ValueToString(functionList->callFunctionName)) + 1; }
      functionList = functionList->next;
     }
  
   return(size);
  }

/************************************************************************/
/* BsaveExpression: Recursively saves an expression to the binary file. */
/************************************************************************/
globle VOID BsaveExpression(testPtr,fp)
  struct expr *testPtr;
  FILE *fp;
  {
   struct expr newTest;
   long int newIndex;
   
   while (testPtr != NULL)
     {
      ExpressionCount++;
      
      /*================*/
      /* Copy the type. */
      /*================*/
      
      newTest.type = testPtr->type;
      
      /*========================================*/
      /* Convert the arg_list slot to an index. */
      /*========================================*/
      
      if (testPtr->arg_list == NULL)
        { newTest.arg_list = (struct expr *) -1L; }
      else
        { newTest.arg_list = (struct expr *) ExpressionCount; }
      
      /*========================================*/
      /* Convert the next_arg slot to an index. */
      /*========================================*/
      
      if (testPtr->next_arg == NULL)
        { newTest.next_arg = (struct expr *) -1L; }
      else
        {
         newIndex = ExpressionCount + ExpressionSize(testPtr->arg_list);
         newTest.next_arg = (struct expr *) newIndex;
        }
        
      /*=========================*/
      /* Convert the value slot. */
      /*=========================*/
      
      switch(testPtr->type)
        {
         case FLOAT:
           newTest.value = (VOID *) ((FLOAT_HN *) testPtr->value)->bucket;
           break;
         
         case INTEGER:
           newTest.value = (VOID *) ((INTEGER_HN *) testPtr->value)->bucket;
           break;
         
         case FCALL:
           newTest.value = (VOID *) ((struct FunctionDefinition *) testPtr->value)->bsaveIndex;
           break;
           
         case GCALL:
#if DEFGENERIC_CONSTRUCT
           if (testPtr->value != NULL)
             newTest.value = (VOID *) ((GENERIC_FUNC *) testPtr->value)->bsaveIndex;
           else
#endif
             newTest.value = (VOID *) -1;
           break;

         case PCALL:
#if DEFFUNCTION_CONSTRUCT
           if (testPtr->value != NULL)
             newTest.value = (VOID *) ((struct dfunc *) testPtr->value)->bsaveIndex;
           else
#endif
             newTest.value = (VOID *) -1;
           break;

#if OBJECT_SYSTEM         
         case INSTANCE_NAME:
#endif
         case SYMBOL:
         case GBWORD:
         case STRING:
           newTest.value = (VOID *) ((SYMBOL_HN *) testPtr->value)->bucket;
           break;
           
         case SCALL_COMP_VARS:
         case SCALL_GET_VAR:
           newTest.value = testPtr->value;
           break;
           
         default:
           PrintCLIPS(WERROR,"Unidentified case\n");
           break;
        }
        
     /*===========================*/
     /* Write out the expression. */
     /*===========================*/
     
     GenWrite(&newTest,(unsigned long) sizeof(struct expr),fp);
     
     /*==========================*/
     /* Write out argument list. */
     /*==========================*/
     
     if (testPtr->arg_list != NULL)
       {
        BsaveExpression(testPtr->arg_list,fp);
       }
       
     testPtr = testPtr->next_arg;
    }
  }

/****************************************************************************/
/* MarkNeededFlags:                                                             */
/****************************************************************************/
static VOID MarkNeededFlags()
  {
   int i;
   SYMBOL_HN *symbolPtr, **symbolArray;
   FLOAT_HN *floatPtr, **floatArray;
   INTEGER_HN *integerPtr, **integerArray;
   struct FunctionDefinition *functionList;
   
   /*=================================*/
   /* Initialize the function counts. */
   /*=================================*/
   
   functionList = GetFunctionList();
   
   while (functionList != NULL)
     {
      functionList->bsaveIndex = 0;
      functionList = functionList->next;
     }
     
   /*===============================*/
   /* Initialize the symbol tables. */
   /*===============================*/
   
   symbolArray = GetSymbolTable();
      
   for (i = 0; i < SYMBOL_HASH_SIZE; i++)
     {
      symbolPtr = symbolArray[i];
      while (symbolPtr != NULL)
        {
         symbolPtr->neededSymbol = CLIPS_FALSE;
         symbolPtr = symbolPtr->next;
        }
     }
     
   floatArray = GetFloatTable();
      
   for (i = 0; i < FLOAT_HASH_SIZE; i++)
     {
      floatPtr = floatArray[i];
      while (floatPtr != NULL)
        {
         floatPtr->neededFloat = CLIPS_FALSE;
         floatPtr = floatPtr->next;
        }
     }
     
   integerArray = GetIntegerTable();
      
   for (i = 0; i < INTEGER_HASH_SIZE; i++)
     {
      integerPtr = integerArray[i];
      while (integerPtr != NULL)
        {
         integerPtr->neededInteger = CLIPS_FALSE;
         integerPtr = integerPtr->next;
        }
     }
  }
  
/****************************************************************************/
/* MarkBuckets:                                                             */
/****************************************************************************/
static VOID MarkBuckets()
  {
   unsigned int count = 0;
   int i;
   SYMBOL_HN *symbolPtr, **symbolArray;
   FLOAT_HN *floatPtr, **floatArray;
   INTEGER_HN *integerPtr, **integerArray;
   
   symbolArray = GetSymbolTable();
      
   for (i = 0; i < SYMBOL_HASH_SIZE; i++)
     {
      symbolPtr = symbolArray[i];
      while (symbolPtr != NULL)
        {
         if (symbolPtr->neededSymbol == CLIPS_TRUE) symbolPtr->bucket = count++;
         symbolPtr = symbolPtr->next;
        }
     }
     
   count = 0;
   floatArray = GetFloatTable();
      
   for (i = 0; i < FLOAT_HASH_SIZE; i++)
     {
      floatPtr = floatArray[i];
      while (floatPtr != NULL)
        {
         if (floatPtr->neededFloat == CLIPS_TRUE) floatPtr->bucket = count++;
         floatPtr = floatPtr->next;
        }
     }
     
   count = 0;
   integerArray = GetIntegerTable();
      
   for (i = 0; i < INTEGER_HASH_SIZE; i++)
     {
      integerPtr = integerArray[i];
      while (integerPtr != NULL)
        {
         if (integerPtr->neededInteger == CLIPS_TRUE) integerPtr->bucket = count++;
         integerPtr = integerPtr->next;
        }
     }
  }
  
/****************************************************************************/
/* UnmarkBuckets:                                                             */
/****************************************************************************/
static VOID UnmarkBuckets()
  {
   int i;
   SYMBOL_HN *symbolPtr, **symbolArray;
   FLOAT_HN *floatPtr, **floatArray;
   INTEGER_HN *integerPtr, **integerArray;
   
   symbolArray = GetSymbolTable();
      
   for (i = 0; i < SYMBOL_HASH_SIZE; i++)
     {
      symbolPtr = symbolArray[i];
      while (symbolPtr != NULL)
        {
         symbolPtr->bucket = i;
         symbolPtr = symbolPtr->next;
        }
     }
     
   floatArray = GetFloatTable();
      
   for (i = 0; i < FLOAT_HASH_SIZE; i++)
     {
      floatPtr = floatArray[i];
      while (floatPtr != NULL)
        {
         floatPtr->bucket = i;
         floatPtr = floatPtr->next;
        }
     }
     
   integerArray = GetIntegerTable();
      
   for (i = 0; i < INTEGER_HASH_SIZE; i++)
     {
      integerPtr = integerArray[i];
      while (integerPtr != NULL)
        {
         integerPtr->bucket = i;
         integerPtr = integerPtr->next;
        }
     }
  }

/*****************************************************/
/* BsaveAllExpressions: Writes all expression needed */
/*   for this binary image to the binary save file.  */
/*****************************************************/
static VOID BsaveAllExpressions(fp)
  FILE *fp;
  {
   struct BinaryItem *biPtr;
   
   /*==================================*/
   /* Write out number of expressions. */
   /*==================================*/
   
   GenWrite(&ExpressionCount,(unsigned long) sizeof(long int),fp);
   if (ExpressionCount == 0) return;
 
   ExpressionCount = 0;
   biPtr = ListOfBinaryItems;
   while (biPtr != NULL)
     {
      if (biPtr->expressionFunction != NULL) (*biPtr->expressionFunction)(fp);
      biPtr = biPtr->next;
     }
  }
 
/****************************************************/
/* GenWrite: Generic routine for writing to a file. */
/*   No machine specific code as of yet.            */
/****************************************************/
globle VOID GenWrite(dataPtr,size,fp)
  VOID *dataPtr;
  unsigned long size;
  FILE *fp;
  {
   if (size == 0) return;
   fwrite(dataPtr,(unsigned int) size,1,fp);
  }  

/**************************************************************/
/* MarkNeededItems: Examines an expression to determine which */
/*   items are necessary to evaluate the expression.          */
/**************************************************************/
globle VOID MarkNeededItems(testPtr)
  struct expr *testPtr;
  {   
   while (testPtr != NULL)
     {
      switch (testPtr->type)
        {
         case SYMBOL:
         case STRING:
            ((SYMBOL_HN *) testPtr->value)->neededSymbol = CLIPS_TRUE;
            break;
            
         case FLOAT:
            ((FLOAT_HN *) testPtr->value)->neededFloat = CLIPS_TRUE;
            break;
         
         case INTEGER:
            ((INTEGER_HN *) testPtr->value)->neededInteger = CLIPS_TRUE;
            break;
            
         case FCALL:
            ((struct FunctionDefinition *) testPtr->value)->bsaveIndex = CLIPS_TRUE;
            break;
        }
      if (testPtr->arg_list != NULL)
        { MarkNeededItems(testPtr->arg_list); }
        
      testPtr = testPtr->next_arg;
     }
  }

/*****************************************************************/
/* WriteNeededSymbols: Stores all of the symbols in the symbol   */
/*   table needed for this binary image in the binary save file. */
/*****************************************************************/
static VOID WriteNeededSymbols(fp)
  FILE *fp;
  {
   int i, length;
   SYMBOL_HN **symbolArray;
   SYMBOL_HN *symbolPtr;
   unsigned long int numberOfUsedSymbols = 0, size = 0;
   
   /*=================================*/
   /* Get a copy of the symbol table. */
   /*=================================*/
   
   symbolArray = GetSymbolTable();
   
   /*======================================================*/
   /* Get the number of symbols and the total string size. */
   /*======================================================*/
   
   for (i = 0; i < SYMBOL_HASH_SIZE; i++)
     {
      symbolPtr = symbolArray[i];
      while (symbolPtr != NULL)
        {
         if (symbolPtr->neededSymbol)
           {
            numberOfUsedSymbols++;
            size += strlen(symbolPtr->contents) + 1;
           }
         symbolPtr = symbolPtr->next;
        }
     }
     
   /*=============================================*/
   /* Write out the symbols and the string sizes. */
   /*=============================================*/
    
   GenWrite(&numberOfUsedSymbols,(unsigned long) sizeof(unsigned long int),fp);
   GenWrite(&size,(unsigned long) sizeof(unsigned long int),fp);
   
   for (i = 0; i < SYMBOL_HASH_SIZE; i++)
     {
      symbolPtr = symbolArray[i];
      while (symbolPtr != NULL)
        {
         if (symbolPtr->neededSymbol)
           {
            length = strlen(symbolPtr->contents) + 1;
            GenWrite(symbolPtr->contents,(long) length,fp);
           }
         symbolPtr = symbolPtr->next;
        }
     }
  }
   
/*****************************************************************/
/* WriteNeededFloats: Stores all of the floats in the float   */
/*   table needed for this binary image in the binary save file. */
/*****************************************************************/
static VOID WriteNeededFloats(fp)
  FILE *fp;
  {
   int i;
   FLOAT_HN **floatArray;
   FLOAT_HN *floatPtr;
   unsigned long int numberOfUsedFloats = 0;
   
   /*=================================*/
   /* Get a copy of the symbol table. */
   /*=================================*/
   
   floatArray = GetFloatTable();
   
   /*======================================================*/
   /* Get the number of symbols and the total string size. */
   /*======================================================*/
   
   for (i = 0; i < FLOAT_HASH_SIZE; i++)
     {
      floatPtr = floatArray[i];
      while (floatPtr != NULL)
        {
         if (floatPtr->neededFloat) numberOfUsedFloats++;
         floatPtr = floatPtr->next;
        }
     }
     
   /*======================================================*/
   /* Write out the number of floats and the float values. */
   /*======================================================*/
    
   GenWrite(&numberOfUsedFloats,(unsigned long) sizeof(unsigned long int),fp);
   
   for (i = 0 ; i < FLOAT_HASH_SIZE; i++)
     {
      floatPtr = floatArray[i];
      while (floatPtr != NULL)
        {
         if (floatPtr->neededFloat) 
           { GenWrite(&floatPtr->contents,
                      (unsigned long) sizeof(floatPtr->contents),fp); }
         floatPtr = floatPtr->next;
        }
     }
  }
  
/*****************************************************************/
/* WriteNeededIntegers: Stores all of the integers in the integer   */
/*   table needed for this binary image in the binary save file. */
/*****************************************************************/
static VOID WriteNeededIntegers(fp)
  FILE *fp;
  {
   int i;
   INTEGER_HN **integerArray;
   INTEGER_HN *integerPtr;
   unsigned long int numberOfUsedIntegers = 0;
   
   /*=================================*/
   /* Get a copy of the symbol table. */
   /*=================================*/
   
   integerArray = GetIntegerTable();
   
   /*======================================================*/
   /* Get the number of symbols and the total string size. */
   /*======================================================*/
   
   for (i = 0 ; i < INTEGER_HASH_SIZE; i++)
     {
      integerPtr = integerArray[i];
      while (integerPtr != NULL)
        {
         if (integerPtr->neededInteger) numberOfUsedIntegers++;
         integerPtr = integerPtr->next;
        }
     }
     
   /*======================================================*/
   /* Write out the number of integers and the integer values. */
   /*======================================================*/
    
   GenWrite(&numberOfUsedIntegers,(unsigned long) sizeof(unsigned long int),fp);
   
   for (i = 0 ; i < INTEGER_HASH_SIZE; i++)
     {
      integerPtr = integerArray[i];
      while (integerPtr != NULL)
        {
         if (integerPtr->neededInteger)
           {
            GenWrite(&integerPtr->contents,
                     (unsigned long) sizeof(integerPtr->contents),fp);
           }
         integerPtr = integerPtr->next;
        }
     }
  }
  
/*****************************************************************/
/* WriteBinaryHeader:               */
/*****************************************************************/
static VOID WriteBinaryHeader(fp)
  FILE *fp;
  {
   GenWrite(BinaryPrefixID,(unsigned long) strlen(BinaryPrefixID) + 1,fp);
   GenWrite(BinaryVersionID,(unsigned long) strlen(BinaryVersionID) + 1,fp);
  }
  
/*****************************************************************/
/* WriteBinaryFooter:               */
/*****************************************************************/
static VOID WriteBinaryFooter(fp)
  FILE *fp;
  {
   char footerBuffer[CONSTRUCT_HEADER_SIZE];
   
   strncpy(footerBuffer,BinaryPrefixID,CONSTRUCT_HEADER_SIZE);
   GenWrite(footerBuffer,(unsigned long) CONSTRUCT_HEADER_SIZE,fp);
  }

#endif

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
/********************/
/* AddBinaryItem:   */
/********************/
globle BOOLEAN AddBinaryItem(name,p_num,findFunction,expressionFunction,
                      bsaveFunction,bloadFunction,refreshFunction,clearFunction)
  char *name;
  int p_num;
  VOID (*findFunction)();
  VOID (*expressionFunction)();
  VOID (*bsaveFunction)();
  VOID (*bloadFunction)();
  VOID (*refreshFunction)();
  VOID (*clearFunction)();
  {
   struct BinaryItem *newPtr, *currentPtr, *lastPtr = NULL;

   newPtr = get_struct(BinaryItem);

   newPtr->name = name;
   newPtr->findFunction = findFunction;
   newPtr->expressionFunction = expressionFunction;
   newPtr->bsaveFunction = bsaveFunction;
   newPtr->bloadFunction = bloadFunction;
   newPtr->refreshFunction = refreshFunction;
   newPtr->clearFunction = clearFunction;
   newPtr->priority = p_num;
   
   if (ListOfBinaryItems == NULL)
     {
      newPtr->next = NULL;
      ListOfBinaryItems = newPtr;
      return(1);
     }
     
   currentPtr = ListOfBinaryItems;
   while ((currentPtr != NULL) ? (p_num < currentPtr->priority) : CLIPS_FALSE)
     {
      lastPtr = currentPtr; 
      currentPtr = currentPtr->next;
     }

   if (lastPtr == NULL)
     { 
      newPtr->next = ListOfBinaryItems;
      ListOfBinaryItems = newPtr;
     }
   else
     {
      newPtr->next = currentPtr;
      lastPtr->next = newPtr;
     }
   
   return(1);
  }
  
#endif



