   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*               CLIPS Version 5.10  07/17/91          */
   /*                                                     */
   /*                  CLASS FUNCTIONS MODULE             */
   /*******************************************************/

/*************************************************************/
/* Purpose: Internal class manipulation routines             */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
/* =========================================
   *****************************************
               EXTERNAL DEFINITIONS
   =========================================
   ***************************************** */
#include "setup.h"

#if OBJECT_SYSTEM

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
#include "bload.h"
#endif
#if BLOAD || BLOAD_AND_BSAVE
#include "defins.h"
#endif
#include "clipsmem.h"
#include "constant.h"
#include "constrct.h"
#include "evaluatn.h"
#include "inscom.h"
#include "insfun.h"
#include "msgcom.h"
#include "msgfun.h"
#include "router.h"
#include "scanner.h"
#include "utility.h"

#define _CLASSFUN_SOURCE_
#include "classfun.h"

/* =========================================
   *****************************************
                   CONSTANTS
   =========================================
   ***************************************** */
#define DIRECT          0
#define INHERIT         1
#define BIG_PRIME       11329

/* =========================================
   *****************************************
               MACROS AND TYPES
   =========================================
   ***************************************** */
typedef struct tilink
  {
   SLOT_DESC *desc;
   unsigned toffset;
   struct tilink *nxt,*inh;
  } TILINK;
  
typedef struct seqnode
  {
   CLASS_TYPE *cls;
   struct seqnode *nxt,*suc;
  } SEQNODE;
    
/* =========================================
   *****************************************
      INTERNALLY VISIBLE FUNCTION HEADERS
   =========================================
   ***************************************** */
#if ANSI_COMPILER

static unsigned HashClass(SYMBOL_HN *);
#if (! BLOAD_ONLY) && (! RUN_TIME)
static VOID CreateSystemClasses(void);
static CLASS_TYPE *AddSystemClass(char *,unsigned,int,CLASS_TYPE *);
static VOID AddToClassList(CLASS_TYPE *);
static CLASS_LINK *CopyClassLinks(CLASS_LINK *org);
static VOID InstallClass(CLASS_TYPE *,int);
static int DeleteClass(CLASS_TYPE *);
static VOID DeleteSublink(CLASS_TYPE *,CLASS_TYPE *);
static VOID BuildSubclassLinks(CLASS_TYPE *);
static VOID ResetCompositeSlots(CLASS_TYPE *);
static VOID FormInstanceTemplate(CLASS_TYPE *);
static TILINK *MergeSlots(TILINK *,CLASS_TYPE *,unsigned *,int,TILINK **);
static SEQNODE *BuildPartialOrders(SEQNODE *,CLASS_LINK *);
static VOID RecordPartialOrder(SEQNODE *,CLASS_TYPE *);
static VOID PrintPartialOrderLoop(SEQNODE *);
#endif

#else

static unsigned HashClass();
#if (! BLOAD_ONLY) && (! RUN_TIME)
static VOID CreateSystemClasses();
static CLASS_TYPE *AddSystemClass();
static VOID AddToClassList();
static CLASS_LINK *CopyClassLinks();
static VOID InstallClass();
static int DeleteClass();
static VOID DeleteSublink();
static VOID BuildSubclassLinks();
static VOID ResetCompositeSlots();
static VOID FormInstanceTemplate();
static TILINK *MergeSlots();
static SEQNODE *BuildPartialOrders();
static VOID RecordPartialOrder();
static VOID PrintPartialOrderLoop();
#endif
#endif

/* =========================================
   *****************************************
      EXTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */
globle CLASS_TYPE *ClassList = NULL,
globle            **ClassTable = NULL;
globle CLASS_TYPE *PrimitiveClassMap[] = { NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL };

/* =========================================
   *****************************************
      INTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */
static CLASS_TYPE *ClassListBottom = NULL;

/* =========================================
   *****************************************
          EXTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */

/***************************************************
  NAME         : SetClassList
  DESCRIPTION  : Sets global class list pointers
  INPUTS       : The top of the list
  RETURNS      : Nothing useful
  SIDE EFFECTS : Pointers set
  NOTES        : WARNING: Does not attempt to 
                 validate this operation.
 ***************************************************/
globle VOID SetClassList(top)
  CLASS_TYPE *top;
  {
   ClassList = top;
   ClassListBottom = top;
   if (top != NULL)
     {
      while (ClassListBottom->nxt_lst != NULL)
        ClassListBottom = ClassListBottom->nxt_lst;
     }
  }

/***************************************************
  NAME         : InstancesPurge
  DESCRIPTION  : Removes all instances
  INPUTS       : None
  RETURNS      : Nothing meaningful
  SIDE EFFECTS : The instance hash table is cleared
  NOTES        : None
 ***************************************************/
globle VOID InstancesPurge()
  {
   int svdepth;

   DestroyAllInstances();
   svdepth = CurrentEvaluationDepth;
   if (CurrentEvaluationDepth == 0)
     CurrentEvaluationDepth = -1;
   CleanupInstances();
   CurrentEvaluationDepth = svdepth;
  }

#if ! RUN_TIME

/***************************************************
  NAME         : InitializeClasses
  DESCRIPTION  : Allocates class hash table
                 Initializes class hash table
                   to all NULL addresses
                 Creates system classes
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Hash table initialized
  NOTES        : None
 ***************************************************/
globle VOID InitializeClasses()
  {
#if BLOAD_ONLY
   register int i;
#endif
   
   ClassTable = 
      (CLASS_TYPE **) gm2((int) (sizeof(CLASS_TYPE *) * CLASS_TABLE_HASH_SIZE));

#if BLOAD_ONLY
   ClassList = ClassListBottom = NULL;
   
   for (i = 0 ; i < CLASS_TABLE_HASH_SIZE ; i++)
     ClassTable[i] = NULL;
#else
   ReinitializeClasses();
#endif
  }

#endif

/*******************************************************************
  NAME         : FindDefclassBySymbol
  DESCRIPTION  : Looks up a specified class in the class hash table
  INPUTS       : The CLIPS symbol for the name of the class
  RETURNS      : The address of the found class, NULL otherwise
  SIDE EFFECTS : None
  NOTES        : Conflict resolution is done by chaining
 ******************************************************************/
globle CLASS_TYPE *FindDefclassBySymbol(cname)
  SYMBOL_HN *cname;
  {
   CLASS_TYPE *cls;
   
   cls = ClassTable[HashClass(cname)];
   while (cls != NULL)
     {
      if (cls->name == cname)
        return((cls->installed == 1) ? cls : NULL);
      cls = cls->nxt_hsh;
     }
   return(cls);
  }
  
/********************************************************
  NAME         : FindClassSlot
  DESCRIPTION  : Searches for a named slot in a class
  INPUTS       : 1) The class address
                 2) The symbolic slot name
  RETURNS      : Address of slot if found, NULL otherwise
  SIDE EFFECTS : None
  NOTES        : Only looks in class defn, does not
                   examine inheritance paths
 ********************************************************/
globle SLOT_DESC *FindClassSlot(cls,sname)
  CLASS_TYPE *cls;
  SYMBOL_HN *sname;
  {
   register int i;
   
   for (i = 0 ; i < cls->slot_cnt ; i++)
     {
      if (cls->slots[i].name == sname)
        return(&cls->slots[i]);
     }
   return(NULL);
  }

/***************************************************************
  NAME         : ClassExistError
  DESCRIPTION  : Prints out error message for non-existent class
  INPUTS       : 1) Name of function having the error
                 2) The name of the non-existent class
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************************/
globle VOID ClassExistError(func,cname)
  char *func,*cname;
  {
   PrintCLIPS(WERROR,"Unable to find class ");
   PrintCLIPS(WERROR,cname);
   PrintCLIPS(WERROR," in function ");
   PrintCLIPS(WERROR,func);
   PrintCLIPS(WERROR,".\n");
   SetEvaluationError(TRUE);
  }
  
/*********************************************
  NAME         : DeleteClassLinks
  DESCRIPTION  : Deallocates a class link list
  INPUTS       : The address of the list
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : None
 *********************************************/
globle VOID DeleteClassLinks(clink)
  CLASS_LINK *clink;
  {
   CLASS_LINK *ctmp;

   for (ctmp = clink ; ctmp != NULL ; ctmp = clink)
     {
      clink = clink->nxt;
      rtn_struct(class_link,ctmp);
     }
  }
  
#if ! RUN_TIME

/*******************************************************
  NAME         : PutClassInTable
  DESCRIPTION  : Inserts a class in the class hash table
  INPUTS       : The class
  RETURNS      : Nothing useful
  SIDE EFFECTS : Class inserted
  NOTES        : None
 *******************************************************/
globle VOID PutClassInTable(cls)
  CLASS_TYPE *cls;
  {
   cls->hashval = HashClass(cls->name);
   cls->nxt_hsh = ClassTable[cls->hashval];
   if (ClassTable[cls->hashval] != NULL)
     ClassTable[cls->hashval]->prv_hsh = cls;
   ClassTable[cls->hashval] = cls;
  }
  
#endif

#if BLOAD || BLOAD_AND_BSAVE

/****************************************************
  NAME         : ObjectSystemPurge
  DESCRIPTION  : Removes all classes and definstances
  INPUTS       : None
  RETURNS      : Nothing meaningful
  SIDE EFFECTS : The class hash table is cleared
  NOTES        : None
 ***************************************************/
globle VOID ObjectSystemPurge()
  {
   CLASS_TYPE *ctmp,*cptr;

   InstancesPurge();
   if (Bloaded())
     return;
   ctmp = ClassList;
   while (ctmp != NULL)
     {
      cptr = ctmp;
      ctmp = ctmp->nxt_lst;
      DeleteClass(cptr);
     }
   ClearDefinstances();
  }

#endif

#if (! BLOAD_ONLY) && (! RUN_TIME)

/***************************************************
  NAME         : ReinitializeClasses
  DESCRIPTION  : Just like InitializeClasses except
                   doesn't allocate table
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Hash table initialized
  NOTES        : None
 ***************************************************/
globle VOID ReinitializeClasses()
  {
   register int i;

   ClassList = ClassListBottom = NULL;

   for (i = 0 ; i < CLASS_TABLE_HASH_SIZE ; i++)
     ClassTable[i] = NULL; 
   CreateSystemClasses();
  }

/**************************************************************
  NAME         : NewClass
  DESCRIPTION  : Allocates and initalizes a new class structure
  INPUTS       : None
  RETURNS      : The address of the new class
  SIDE EFFECTS : None
  NOTES        : None
 **************************************************************/
globle CLASS_TYPE *NewClass()
  {
   register CLASS_TYPE *cls;
   
   cls = get_struct(cls);
   cls->installed = 0;
   cls->busy = 0;
   cls->system = 0;
   cls->abstract = 0;
   cls->primitive = 0;
   cls->primitive_type = 0;
   cls->hashval = 0;
   cls->predecessors = 0;
   cls->name = NULL;
   cls->superlink = NULL;
   cls->sublink = NULL;
   cls->precedence = NULL;
   cls->slots = NULL;
   cls->itemplate = NULL;
   cls->sordered = NULL;
   cls->islot_cnt = 0;
   cls->slot_cnt = 0;
   cls->handlers = NULL;
   cls->hordered = NULL;
   cls->handler_cnt = 0;
   cls->instances = NULL;
   cls->prv_hsh = NULL;
   cls->nxt_hsh = NULL;
   cls->prv_lst = NULL;
   cls->nxt_lst = NULL;
   cls->pp_form = NULL;
   cls->tvids = (char *) gm2((int) (sizeof(char) * TRAVERSAL_BYTES));
   return(cls);
  }

/***************************************************
  NAME         : IsClassBeingUsed
  DESCRIPTION  : Checks the busy flag of a class
                   and ALL classes that inherit from
                   it to make sure that it is not
                   in use before deletion
  INPUTS       : The class
  RETURNS      : TRUE if in use, FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : Recursively examines all subclasses
 ***************************************************/
globle int IsClassBeingUsed(cls)
  CLASS_TYPE *cls;
  {
   CLASS_LINK *sptr;
   
   if (cls->busy > 0)
     return(TRUE);
   for (sptr = cls->sublink ; sptr != NULL ; sptr = sptr->nxt)
     if (IsClassBeingUsed(sptr->cls))
       return(TRUE);
   return(FALSE);
  }
  
/***************************************************
  NAME         : ClearDefclasses
  DESCRIPTION  : Removes all classes
  INPUTS       : Flag indicating whether to remove
                  user-defined message-handlers from
                  message-handlers or not
  RETURNS      : TRUE if succesful, FALSE otherwise
  SIDE EFFECTS : The class hash table is cleared
  NOTES        : None
 ***************************************************/
globle int ClearDefclasses(udsm)
  int udsm;
  {
   CLASS_TYPE *user_classes,*ctmp;
   register unsigned i,mhdok,wtd;
   int success = TRUE;
   
#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   if (Bloaded())
     return(FALSE);
#endif   
   /* ====================================================
      Don't delete built-in system classes at head of list
      ==================================================== */
   user_classes = ClassList;
   while ((user_classes != NULL) ? (user_classes->system == 1) : FALSE)
     {
      if (udsm)
        {
         mhdok = TRUE;
         wtd = FALSE;
         for (i = 0 ; i < user_classes->handler_cnt ; i++)
           {
            if (user_classes->handlers[i].busy > 0)
              mhdok = FALSE;
            if (user_classes->handlers[i].system == 0)
              {
               user_classes->handlers[i].mark = 1;
               wtd = TRUE;
              }
           }
         if (wtd == TRUE)
           {
            if (mhdok == FALSE)
              {
               HandlerDeleteError(ValueToString(user_classes->name));
               success = FALSE;
               for (i = 0 ; i < user_classes->handler_cnt ; i++)
                 user_classes->handlers[i].mark = 0;
              }
            else
              DeallocateMarkedHandlers(user_classes);
           }
        }
      user_classes = user_classes->nxt_lst;
     }
   while (user_classes != NULL)
     {
      ctmp = user_classes;
      user_classes = user_classes->nxt_lst;
      DeleteClass(ctmp);
     }
   return(success);
  }
  
/****************************************************
  NAME         : DeleteClassUAG
  DESCRIPTION  : Deallocates a class and all its 
                 subclasses
  INPUTS       : The address of the class
  RETURNS      : 1 if successful, 0 otherwise
  SIDE EFFECTS : Removes the class from each of
                 its superclasses' subclass lists
  NOTES        : None
 ****************************************************/
globle int DeleteClassUAG(cls)
  CLASS_TYPE *cls;
  {
   CLASS_LINK *stmp;
   
   while (cls->sublink != NULL)
     {
      stmp = cls->sublink;
      DeleteClassUAG(cls->sublink->cls);
      if (cls->sublink == stmp)
        return(0);
     }
   return(DeleteClass(cls));
  }
  
/***************************************************
  NAME         : DeleteSlots
  DESCRIPTION  : Deallocates a list of slots and
                   their values
  INPUTS       : The address of the slot list
  RETURNS      : Nothing useful
  SIDE EFFECTS : The slot list is destroyed
  NOTES        : share_val field temporarily used for
                   links
 ***************************************************/
globle VOID DeleteSlots(slots)
  SLOT_DESC *slots;
  {
   SLOT_DESC *stmp;
   
   while (slots != NULL)
     {
      stmp = slots;
      slots = (SLOT_DESC *) slots->share_val;
      ReturnPackedExpression(stmp->defexp);
      rtn_struct(slot_desc,stmp);
     }
  }
  
/**************************************************************
  NAME         : NewSlot
  DESCRIPTION  : Allocates and initalizes a new slot structure
  INPUTS       : The symbolic name of the new slot
  RETURNS      : The address of the new slot
  SIDE EFFECTS : None
  NOTES        : Also adds symbols of the form get-<name> and
                   put-<name> for slot accessors
 **************************************************************/
globle SLOT_DESC *NewSlot(name)
  SYMBOL_HN *name;
  {
   SLOT_DESC *slot;
   char *buf;
   int bufsz;
   
   slot = get_struct(slot_desc);
   slot->default_specified = 0;
   slot->cardinality_specified = 0;
   slot->storage_specified = 0;
   slot->access_specified = 0;
   slot->dynamic = 0;
   slot->noinherit = 0;
   slot->nowrite = 0;
   slot->initonly = 0;
   slot->shared = 0;
   slot->multiple = 0;
   slot->override = 0;
   slot->composite = 0;
   slot->share_cnt = 0;
   slot->cls = NULL;
   slot->gethandler = NULL;
   slot->puthandler = NULL;
   slot->defexp = NULL;
   slot->share_val = NULL;
   
   slot->name = name;
   bufsz = strlen(ValueToString(name)) + 
           ((GSMP_LEN > PSMP_LEN) ? GSMP_LEN : PSMP_LEN) + 1;
   buf = (char *) gm2(bufsz);
   strcpy(buf,GSM_PREFIX);
   strcat(buf,ValueToString(name));
   slot->gethandler = (SYMBOL_HN *) AddSymbol(buf);
   slot->gethandler->relatedSymbol = name;
   
   strcpy(buf,PSM_PREFIX);
   strcat(buf,ValueToString(name));
   slot->puthandler = (SYMBOL_HN *) AddSymbol(buf);
   slot->puthandler->relatedSymbol = name;
   
   rm((VOID *) buf,bufsz);
   return(slot);
  }
  
/**********************************************************
  NAME         : InsertSlot
  DESCRIPTION  : Inserts a slot into the list of slots
  INPUTS       : 1) The current head of the slot list
                 2) The slot to be inserted
  RETURNS      : The head of the slot list
  SIDE EFFECTS : The slot is inserted if no errors,
                   otherwise the original list and the
                   new slot are destroyed
  NOTES        : share_val field temporarily used for links
 **********************************************************/
globle SLOT_DESC *InsertSlot(slist,slot)
  SLOT_DESC *slist,*slot;
  {
   SLOT_DESC *stmp,*sprv;
   
   if (slist == NULL)
     slist = slot;
   else
     {
      stmp = slist;
      sprv = NULL;
      while (stmp != NULL)
        {
         if (stmp->name == slot->name)
           {
            DeleteSlots(slot);
            DeleteSlots(slist);
            PrintCLIPS(WERROR,"Duplicate slots not allowed.\n");
            return(NULL);
           }
         sprv = stmp;
         stmp = (SLOT_DESC *) stmp->share_val;
        }
      sprv->share_val = (DATA_OBJECT *) slot;
     }
   return(slist);
  }
  
/***********************************************************************
  NAME         : PackSlots
  DESCRIPTION  : Groups class-slots into a contiguous array
                  "slots" field points to array
                  "slot_cnt" field set
  INPUTS       : 1) The class
                 2) The list of slots
  RETURNS      : Nothing useful
  SIDE EFFECTS : Temporary list deallocated, contiguous array allocated,
                   and nxt pointers linked
                 Class pointer set for slots
  NOTES        : Assumes class->slot_cnt == 0 && class->slots == NULL
 ***********************************************************************/
globle VOID PackSlots(cls,slots)
  CLASS_TYPE *cls;
  SLOT_DESC *slots;
  {
   SLOT_DESC *stmp,*sprv;
   register int i;

   for (stmp = slots ; stmp != NULL ; stmp = (SLOT_DESC *) stmp->share_val)
     {
      stmp->cls = cls;
      cls->slot_cnt++;
     }
   cls->slots = (SLOT_DESC *) gm2((int) (sizeof(SLOT_DESC) * cls->slot_cnt));
   stmp = slots;
   for (i = 0 ; i < cls->slot_cnt ; i++)
     {
      sprv = stmp;
      stmp = (SLOT_DESC *) stmp->share_val;
      CopyMemory(SLOT_DESC,1,&(cls->slots[i]),sprv);
      cls->slots[i].share_val = NULL;
      rtn_struct(slot_desc,sprv);
     }
  }
  
/***************************************************************************
  NAME         : FindPrecedenceList
  DESCRIPTION  : A complete class precedence list is obtained from the
                 list of direct superclasses as follows :
                 
                 Each class and its direct superclasses are recursively
                 entered in order to a list called the sequence table of
                 partial orders.  A class is only entered once.  The order
                 reflects a pre-order depth-first traversal of the classes,
                 and this order will be followed as closely as possible to
                 preserve the "family" heuristic when constructing the
                 class precedence list.  Each node is linked to the next
                 via the nxt pointer field.
                 
                 Attached to each node is a count indicating the number of
                 classes which must precede this class and a list of classes
                 which must succeed this class (attached via the suc field and
                 linked via nxt fields).  These predecessor counts
                 and successor lists indicate the partial orderings given
                 by the rules of multiple inheritance for the classes:
                 1) a class must precede all its superclasses, and 2) a
                 class determines the precedence of its immediate superclasses.
                 
                 For example, the following class definitions
                 
                 (defclass A (is-a USER))
                 (defclass B (is-a USER))
                 (defclass C (is-a A B))
                 
                 would give the following partial orders:
                 
                 C < A          by Rule 1
                 C < B          by Rule 1
                 A < B          by Rule 2
                 B < USER       by Rule 1
                 A < USER       by Rule 1
                 USER < OBJECT  by Rule 1
                 
                 In turn, these partial orders would be recorded in a
                 sequence table:
                 
                                     C     A      USER    OBJECT    B
                 Predecessor Count   0     1       2        1       2
                 Successor List     A,B  B,USER  OBJECT   <NIL>    USER
                 
                 To generate a precedence list for C, we pick the first
                 class with a predecessor count of 0, append it to the
                 precedence list, and decrement the counts of all its
                 successors.  We continue scanning for a 0 from where
                 we left off.  If we ever scan completely through the
                 table without finding a 0, then we know there is an
                 error.
                 
                 Shown below is the table above after each class is
                 entered onto the precedence list:
                 
                 Precedence list: C
                                     A      USER    OBJECT    B
                 Predecessor Count   0       2        1       1
                 Successor List     B,USER  OBJECT   <NIL>    USER
                 
                 Precedence list: C A
                                    USER    OBJECT    B
                 Predecessor Count   1        1       0
                 Successor List     OBJECT   <NIL>   USER
                 
                 Precedence list: C A B
                                    USER    OBJECT
                 Predecessor Count   0        1
                 Successor List     OBJECT   <NIL>
                 
                 Precedence list: C A B USER
                                    OBJECT
                 Predecessor Count   0
                 Successor List     <NIL>
                 
                 Precedence List: C A B USER OBJECT
                 
                 And since the table is now empty we are done!
  INPUTS       : 1) Pointer to old class definition, NULL if this is new
                 2) The list of direct superclasses
  RETURNS      : The address of the precedence list if successful,
                 NULL otherwise
  SIDE EFFECTS : Precedence list allocated
  NOTES        : Every class-precedence list has the class itself on it
                 (implicitly) and a built-in system class on it explicitly
                 (except for the built-in classes).
                 
                 Assumes the predecessor field for all classes is 0 before
                   entering this function!!!
                 This function also insures that the predecessor counts of
                   any classes it examines are returned to 0 before exiting.
                   
                 The precedence determination algorithm is a variation on
                 the topological sorting algorithm given in The Art of
                 Computer Programming - Vol. I (Fundamental Algorithms) by
                 Donald Knuth.
 ***************************************************************************/
globle CLASS_LINK *FindPrecedenceList(cls,supers)
  CLASS_TYPE *cls;
  CLASS_LINK *supers;
  {
   SEQNODE *seqtable = NULL,*start,*sqtmp;
   register SEQNODE *sqp,*sqprv;
   CLASS_LINK *ptop,*pbot,*ptmp;
   
   /* ====================================================================
      Build a table of partial orders representing the class relationships
      If the class already exists, add the rule 1 partial orders for the
      class first.
      ==================================================================== */
   if (cls != NULL)
     {
      seqtable = get_struct(seqnode);
      seqtable->nxt = NULL;
      seqtable->suc = NULL;
      seqtable->cls = cls;
      for (ptmp = supers ; ptmp != NULL ; ptmp = ptmp->nxt)
        RecordPartialOrder(seqtable,ptmp->cls);
     }
     
   seqtable = BuildPartialOrders(seqtable,supers);

   start = NULL;
   sqprv = NULL;
   sqp = seqtable;
   ptop = pbot = NULL;
   while (sqp != start)
     {
      /* ==============================================================
         Allow wraparound - happens when the search for a 0 node begins
         somewhere in the middle of the sequence table
         ============================================================== */
      if (sqp == NULL)
        {
         sqprv = NULL;
         sqp = seqtable;
         start = start->nxt;
        }
        
      /* =========================================================
         Search for the first class with no remaining predecessors
         ========================================================= */
      if (sqp->cls->predecessors == 0)
        {
         /* =================================================
            Decrement the predecessor count for all the
            successors of this class and delete the list.
            
            This is the variation on Knuth's algorithm which
            allows us to preserve the "family" heuristic.
            Since we will pick up scanning for 0's from
            this point, we will be able to keep "family"
            trees together, if possible. BuildPartialOrders()
            entered the classes into the sequence table
            in a pre-order depth traversal order.
            ================================================= */
         while (sqp->suc != NULL)
           {
            sqtmp = sqp->suc;
            sqp->suc = sqtmp->nxt;
            sqtmp->cls->predecessors--;
            rtn_struct(seqnode,sqtmp);
           }
           
         /* ===========================================
            Append the class to the precedence list and
            remove its entry from the sequence table
            =========================================== */
         sqtmp = sqp;
         if (sqprv == NULL)
           seqtable = sqp->nxt;
         else
           sqprv->nxt = sqp->nxt;
         sqp = sqp->nxt;
         start = sqprv;
         ptmp = get_struct(class_link);
         ptmp->cls = sqtmp->cls;
         ptmp->nxt = NULL;
         rtn_struct(seqnode,sqtmp);
         if (ptop == NULL)
           ptop = ptmp;
         else
           pbot->nxt = ptmp;
         pbot = ptmp;
        }
      else
        {
         sqprv = sqp;
         sqp = sqp->nxt;
        }
     }
   
   /* ======================================================================
      If the sequence table of partial orders is not empty and we were
      unable to find a class with no predecessors, then there is no solution!
      Print out the precedence loop in the partial orders.
      Delete the remaining sequence table and the partial precedence list.
      ====================================================================== */
   if (seqtable != NULL)
     {
      PrintPartialOrderLoop(seqtable);
      while (seqtable != NULL)
        {
         while (seqtable->suc != NULL)
           {
            sqtmp = seqtable->suc;
            seqtable->suc = sqtmp->nxt;
            rtn_struct(seqnode,sqtmp);
           }
         seqtable->cls->predecessors = 0;
         sqtmp = seqtable;
         seqtable = seqtable->nxt;
         rtn_struct(seqnode,sqtmp);
        }
      DeleteClassLinks(ptop);
      return(NULL);
     }
   return(ptop);
  }


/*****************************************************************************
  NAME         : AddClass
  DESCRIPTION  : Determines the precedence list of the new class.
                 If it is valid, the routine checks to see if the class
                 already exists.  If it does not, all the subclass
                 links are made from the class's direct superclasses,
                 and the class is inserted in the hash table.  If it
                 does, all sublclasses are deleted. An error will occur
                 if any instances of the class (direct or indirect) exist.
                 If all checks out, the old definition is replaced by the new. 
  INPUTS       : The new class description
  RETURNS      : TRUE if class added, FALSE otherwise.
  SIDE EFFECTS : The class is deleted if there is an error.
  NOTES        : No change in the class graph state will occur
                 if there were any errors.
                 Assumes class is not busy!!!
 *****************************************************************************/
globle int AddClass(cls)
  CLASS_TYPE *cls;
  {
   CLASS_TYPE *ctmp;
   HANDLER *oldHandlers = NULL;
   unsigned *oldHordered = NULL;
   unsigned oldHndCnt = 0,i;
   
   /* ===============================================================
      Adjust the facets of any composite slots to reflect inheritance
      =============================================================== */
   ResetCompositeSlots(cls);
      
   /* ===============================================
      If class does not already exist, insert and 
      form progeny links with all direct superclasses
      =============================================== */
   cls->hashval = HashClass(cls->name);
   ctmp = ClassTable[cls->hashval];
   while ((ctmp != NULL) ? (ctmp->name != cls->name) : FALSE)
     ctmp = ctmp->nxt_hsh;
   if (ctmp != NULL)
     {
      int oldcw;

      oldHandlers = ctmp->handlers;
      oldHordered = ctmp->hordered;
      oldHndCnt = ctmp->handler_cnt;
      ctmp->handlers = NULL;
      ctmp->hordered = NULL;
      ctmp->handler_cnt = 0;
      
      oldcw = GetCompilationsWatch();
      SetCompilationsWatch(OFF);
      DeleteClassUAG(ctmp);
      SetCompilationsWatch(oldcw);
     }
   PutClassInTable(cls);

   BuildSubclassLinks(cls);
   InstallClass(cls,TRUE);
   if (ClassList == NULL)
     ClassList = cls;
   else
     ClassListBottom->nxt_lst = cls;
   cls->prv_lst = ClassListBottom;
   ClassListBottom = cls;
   
      
   cls->handlers = oldHandlers;
   cls->hordered = oldHordered;
   cls->handler_cnt = oldHndCnt;
   for (i = 0 ; i < oldHndCnt ; i++)
     cls->handlers[i].cls = cls;

   FormInstanceTemplate(cls);
   for (i = 0 ; i < cls->handler_cnt ; i++)
     if (CheckHandlerAgainstSlots(cls,cls->handlers[i].name,
                                      cls->handlers[i].type) == FALSE)
       cls->handlers[i].mark = 1;
   DeallocateMarkedHandlers(cls);
        
   return(TRUE);
  }
  
/***************************************************
  NAME         : IsSystemClassName
  DESCRIPTION  : Determines if name belongs to one
                   built-in system classes :
                 standard-class, integer, float,
                 symbol, or string.
  INPUTS       : The CLIPS symbolic name
  RETURNS      : TRUE if it is one of those names,
                    FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
globle int IsSystemClassName(name)
  SYMBOL_HN *name;
  {
   CLASS_TYPE *sys;
   
   /* =======================================================
      All the built-in system classes are at the beginning of
      the class list
      ======================================================= */
   sys = ClassList;
   while ((sys != NULL) ? (sys->system == 1) : FALSE)
     {
      if (name == sys->name)
        return(TRUE);
      sys = sys->nxt_lst;
     }
   return(FALSE);
  }
  
#endif

/* =========================================
   *****************************************
          INTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */

/*******************************************************
  NAME         : HashClass
  DESCRIPTION  : Generates a hash index for a given
                 class name
  INPUTS       : The address of the class name SYMBOL_HN
  RETURNS      : The hash index value
  SIDE EFFECTS : None 
  NOTES        : Counts on the fact that the symbol
                 has already been hashed into the CLIPS
                 symbol table - uses that hash value
                 multiplied by a prime for a new hash
 *******************************************************/
static unsigned HashClass(cname)
  SYMBOL_HN *cname;
  {
   unsigned long tally;
   
   tally = ((unsigned long) cname->bucket) * BIG_PRIME;
   return((unsigned) (tally % CLASS_TABLE_HASH_SIZE));
  }

#if (! BLOAD_ONLY) && (! RUN_TIME)

/***************************************************************
  NAME         : CreateSystemClasses
  DESCRIPTION  : Creates the built-in system classes 
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : System classes inserted in the
                   class hash table
  NOTES        : The binary/load save indices for the primitive
                   types (integer, float, symbol and string) are
                   very important.  Need to be able to refer to
                   types with the same index regardless of
                   whether the object system is installed or
                   not.  Thus, the bsave/blaod indices of these
                   classes match their CLIPS integer codes.
                WARNING!!: Assumes no classes exist yet!
 ***************************************************************/
static VOID CreateSystemClasses()
  {
   CLASS_TYPE *user,*any,*primitive,*number,*lexeme;
   
   /* ========================================================
      Bsave Indices for non-primitive classes start at 5
      because: float = 0, integer = 1, symbol = 2, string = 3,
               multifield = 4, and external-address = 5.
      ======================================================== */
   any = AddSystemClass(OBJECT_CLASS_STRING,0,0,NULL);
   primitive = AddSystemClass(PRIMITIVE_TYPE_NAME,0,0,any);
   user = AddSystemClass(USER_CLASS_STRING,0,0,any);

   number = AddSystemClass(NUMBER_TYPE_NAME,0,0,primitive);
   PrimitiveClassMap[INTEGER] = AddSystemClass(INTEGER_TYPE_NAME,INTEGER,1,number);
   PrimitiveClassMap[FLOAT] = AddSystemClass(FLOAT_TYPE_NAME,FLOAT,1,number);
   lexeme = AddSystemClass(LEXEME_TYPE_NAME,0,0,primitive);
   PrimitiveClassMap[SYMBOL] = AddSystemClass(SYMBOL_TYPE_NAME,SYMBOL,1,lexeme);
   PrimitiveClassMap[STRING] = AddSystemClass(STRING_TYPE_NAME,STRING,1,lexeme);
   PrimitiveClassMap[MULTIFIELD] = AddSystemClass(MULTIFIELD_TYPE_NAME,
                                          MULTIFIELD,1,primitive);
   PrimitiveClassMap[EXTERNAL_ADDRESS] = AddSystemClass(ADDRESS_TYPE_NAME,
                                          EXTERNAL_ADDRESS,1,primitive);
   ClassList = ClassListBottom = PrimitiveClassMap[FLOAT];
   AddToClassList(PrimitiveClassMap[INTEGER]);
   AddToClassList(PrimitiveClassMap[SYMBOL]);
   AddToClassList(PrimitiveClassMap[STRING]);
   AddToClassList(PrimitiveClassMap[MULTIFIELD]);
   AddToClassList(PrimitiveClassMap[EXTERNAL_ADDRESS]);
   AddToClassList(any);
   AddToClassList(primitive);
   AddToClassList(number);
   AddToClassList(lexeme);
   AddToClassList(user);
  }
  
/*********************************************************
  NAME         : AddSystemClass
  DESCRIPTION  : Performs all necessary allocations
                   for adding a system class
  INPUTS       : 1) The name-string of the system class
                 2) The numeric code of the corresponding
                      CLIPS primitive type (if any)
                 3) A flag indicating whether this is a
                      CLIPS primitive type class or not
                 4) The address of the parent class
                    (NULL if none)
  RETURNS      : The address of the new system class
  SIDE EFFECTS : Allocations performed
  NOTES        : Assumes system-class name is unique
                 Also assumes SINGLE INHERITANCE for
                   system classes to simplify precedence
                   list determination
                 Adds classes to has table but NOT to
                  class list (this is responsibility
                  of caller)
 *********************************************************/
static CLASS_TYPE *AddSystemClass(name,type,prim,parent)
  char *name;
  unsigned type;
  int prim;
  CLASS_TYPE *parent;
  {
   CLASS_TYPE *sys,*hptr;
   CLASS_LINK *sptr,*stmp;
   
   sys = NewClass();
   sys->name = (SYMBOL_HN *) AddSymbol(name);
   sys->abstract = 1;
   sys->primitive_type = type;
   sys->primitive = prim;
   IncrementSymbolCount(sys->name);
   sys->installed = 1;
   sys->system = 1;
   sys->hashval = HashClass(sys->name);
   
   if (parent != NULL)
     {
      sys->superlink = get_struct(class_link);
      sys->superlink->nxt = NULL;
      sys->superlink->cls = parent;
      sptr = get_struct(class_link);
      sptr->cls = sys;
      sptr->nxt = NULL;
      if (parent->sublink == NULL)
        parent->sublink = sptr;
      else
        {
         stmp = parent->sublink;
         while (stmp->nxt != NULL)
           stmp = stmp->nxt;
         stmp->nxt = sptr;
        }  
      sys->precedence = (CLASS_LINK *) get_struct(class_link);
      sys->precedence->cls = parent;
      sys->precedence->nxt = CopyClassLinks(parent->precedence);
     }
     
   hptr = ClassTable[sys->hashval];
   if (hptr != NULL)
     {
      while (hptr->nxt_hsh != NULL)
        hptr = hptr->nxt_hsh;
      sys->prv_hsh = hptr;
      hptr->nxt_hsh = sys;
     }
   else
     ClassTable[sys->hashval] = sys;
   return(sys);
  }
  
/*********************************************************
  NAME         : AddToClassList
  DESCRIPTION  : Appends a class to the global class list
  INPUTS       : The class
  RETURNS      : Nothing useful
  SIDE EFFECTS : Class list update
  NOTES        : None
 *********************************************************/
static VOID AddToClassList(cls)
  CLASS_TYPE *cls;
  {
   if (ClassList == NULL)
     ClassList = cls;
   else
     ClassListBottom->nxt_lst = cls;
   cls->prv_lst = ClassListBottom;
   ClassListBottom = cls;
  }

/**********************************************
  NAME         : CopyClassLinks
  DESCRIPTION  : Allocates a new set of class
                  links which are a copy of the
                  original links
  INPUTS       : The original links
  RETURNS      : The address of the copy
  SIDE EFFECTS : New links allocated
  NOTES        : None
 **********************************************/
static CLASS_LINK *CopyClassLinks(org)
  CLASS_LINK *org;
  {
   CLASS_LINK *top = NULL,*bot = NULL, *tmp;

   while (org != NULL)
     {
      tmp = (CLASS_LINK *) get_struct(class_link);
      tmp->cls = org->cls;
      tmp->nxt = NULL;
      if (top == NULL)
        top = tmp;
      else
        bot->nxt = tmp;
      bot = tmp;
      org = org->nxt;
     }
   return(top);
  }

/***************************************************
  NAME         : InstallClass
  DESCRIPTION  : In(De)crements all symbol counts for
                 for CLIPS symbols in use by class
                 Disallows (allows) symbols to become
                 ephemeral.
  INPUTS       : Nothing useful
  RETURNS      : TRUE if succesful, FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
static VOID InstallClass(cls,set)
  CLASS_TYPE *cls;
  int set;
  {
#if ANSI_COMPILER
   int (*install)(SYMBOL_HN *),
       (*expinstall)(EXPRESSION *);
#else
   int (*install)(),
       (*expinstall)();
#endif
   SLOT_DESC *slot;
   HANDLER *hnd;
   register int i;
   
   if (set == TRUE) 
     {
      if (cls->installed)
        return;
      cls->installed = 1;
      install = PTIF IncrementSymbolCount;
      expinstall = PTIF ExpressionInstall;
     }
   else
     {
      if (! cls->installed)
        return;
      cls->installed = 0;
      install = PTIF DecrementSymbolCount;
      expinstall = PTIF ExpressionDeinstall;
     }

   (*install)(cls->name);
   for (i = 0 ; i < cls->slot_cnt ; i++)
     {
      slot = &cls->slots[i];
      (*install)(slot->name);
      (*install)(slot->gethandler);
      (*install)(slot->puthandler);
      (*expinstall)(slot->defexp);
     }
     
   /* ==================================================================
      Handler installation is handled when message-handlers are defined:
      see ParseDefmessageHandler() in MSGCOM.C
      ================================================================== */
   if (set == FALSE)
     {
      for (i = 0 ; i < cls->handler_cnt ; i++)
        {
         hnd = &cls->handlers[i];
         DecrementSymbolCount(hnd->name);
         if (hnd->actions != NULL)
           ExpressionDeinstall(hnd->actions);
        }
     }
  }
  
/*******************************************************************
  NAME         : DeleteClass
  DESCRIPTION  : Deallocates a class structure and 
                 all its fields - returns all symbols
                 in use by the class back to CLIPS
                 for ephemeral removal
                 
  INPUTS       : The address of the class
  RETURNS      : 1 if successful, 0 otherwise
  SIDE EFFECTS : None
  NOTES        : Deletes all instances of the class
                   as well
                 Assumes class has no subclasses
                 IMPORTANT WARNING!! : Assumes class
                   busy count and all instances' busy
                   counts are 0 and all handlers' busy counts are 0!
 *******************************************************************/
static int DeleteClass(cls)
  CLASS_TYPE *cls;
  {
   CLASS_LINK *stmp;
   HANDLER *hnd;
   register int i;
   
   if (IsClassBeingUsed(cls))
     {
      PrintCLIPS(WERROR,"Unable to delete class ");
      PrintCLIPS(WERROR,ValueToString(cls->name));
      PrintCLIPS(WERROR," while outstanding references to it still exist.\n");
      return(0);
     }
   if ((cls->system == 0) ? (GetCompilationsWatch() == ON) : FALSE)
     {
      PrintCLIPS(WDIALOG,"Removing class: ");
      PrintCLIPS(WDIALOG,ValueToString(cls->name));
      PrintCLIPS(WDIALOG,"\n");
     }
   /* ====================================================
      Remove all of this class's superclasses' links to it
      ==================================================== */
   for (stmp = cls->superlink ; stmp != NULL ; stmp = stmp->nxt)
     DeleteSublink(stmp->cls,cls);

   /* ====================================
      Remove the class from the hash table
      ==================================== */
   if (cls->nxt_hsh != NULL)
     cls->nxt_hsh->prv_hsh = cls->prv_hsh;
   if (cls->prv_hsh != NULL)
     cls->prv_hsh->nxt_hsh = cls->nxt_hsh;
   else
     ClassTable[cls->hashval] = cls->nxt_hsh;

   /* ===========================================
      Delete the class from the global class list
      =========================================== */
   if (cls->prv_lst != NULL)
     cls->prv_lst->nxt_lst = cls->nxt_lst;
   else
     ClassList = cls->nxt_lst;
   if (cls->nxt_lst != NULL)
     cls->nxt_lst->prv_lst = cls->prv_lst;
   else
     ClassListBottom = cls->prv_lst;

   InstallClass(cls,FALSE);
   DeleteClassLinks(cls->superlink);
   DeleteClassLinks(cls->sublink);
   DeleteClassLinks(cls->precedence);
   for (i = 0 ; i < cls->slot_cnt ; i++)
     ReturnPackedExpression(cls->slots[i].defexp);
   if (cls->slot_cnt != 0)
     rm((VOID *) cls->slots,(int) (sizeof(SLOT_DESC) * cls->slot_cnt));
   if (cls->islot_cnt != 0)
     {
      rm((VOID *) cls->itemplate,
         (int) (sizeof(SLOT_DESC *) * cls->islot_cnt));
      rm((VOID *) cls->sordered,
         (int) (sizeof(unsigned) * cls->islot_cnt));
     }
   for (i = 0 ; i < cls->handler_cnt ; i++)
     {
      hnd = &cls->handlers[i];
      if ((cls->system == 0) ? (GetCompilationsWatch() == ON) : FALSE)
        PrintAbbreviatedHandlerRemoval(hnd);
      if (hnd->actions != NULL)
        ReturnPackedExpression(hnd->actions);
      if (hnd->pp_form != NULL)
        rm((VOID *) hnd->pp_form,(int) (sizeof(char) * (strlen(hnd->pp_form)+1)));
     }
   if (cls->handler_cnt != 0)
     {
      rm((VOID *) cls->handlers,(int) (sizeof(HANDLER) * cls->handler_cnt));
      rm((VOID *) cls->hordered,(int) (sizeof(unsigned) * cls->handler_cnt));
     }
   if (cls->pp_form != NULL)
     rm((VOID *) cls->pp_form,(int) ((strlen(cls->pp_form) + 1) * sizeof(char)));
   rm((VOID *) cls->tvids,(int) (sizeof(char) * TRAVERSAL_BYTES));
   rtn_struct(cls,cls);
   return(1);
  }
  
/***************************************************
  NAME         : DeleteSublink
  DESCRIPTION  : Removes a class from another
                 class's subclass list
  INPUTS       : 1) The superclass whose subclass
                    list is to be modified
                 2) The subclass to be removed
  RETURNS      : Nothing useful
  SIDE EFFECTS : The subclass list is changed
  NOTES        : None
 ***************************************************/
static VOID DeleteSublink(sclass,cls)
  CLASS_TYPE *sclass,*cls;
  {
   CLASS_LINK *prv,*cur;
   
   prv = cur = sclass->sublink;
   while ((cur != NULL) ? (cur->cls != cls) : FALSE)
     {
      prv = cur;
      cur = cur->nxt;
     }
   if (cur != NULL)
     {
      if (prv == cur)
        sclass->sublink = cur->nxt;
      else
        prv->nxt = cur->nxt;
      rtn_struct(class_link,cur);
     }
  }  

/**************************************************************************
  NAME         : ResetCompositeSlots
  DESCRIPTION  : Composite slots are ones that get their facets
                   from more than one class.  By default, the most
                   specific class in object's precedence list specifies
                   the complete set of facets for a slot.  The composite
                   facet in a slot allows facets that are not overridden
                   by the most specific class to be obtained from other
                   classes.
                   
                 Since all superclasses are predetermined before creating
                   a new class based on them, this routine need only
                   examine the immediately next most specific class for
                   extra facets.  Even if that slot is also composite, the
                   other facets have already been filtered down.  If the
                   slot is no-inherit, the next most specific class must
                   be examined.
  INPUTS       : The new class address
  RETURNS      : Nothing useful
  SIDE EFFECTS : Any composite slots are updated to reflact facets from
                   a less specific class
  NOTES        : None
 *************************************************************************/
static VOID ResetCompositeSlots(cls)
  CLASS_TYPE *cls;
  {
   SLOT_DESC *slot,*gslot = NULL;
   CLASS_LINK *prec;
   register int i;
   
   for (i = 0 ; i < cls->slot_cnt ; i++)
     {
      slot = &cls->slots[i];
      if (slot->composite == 1)
        {
         for (prec = cls->precedence ; prec != NULL ; prec = prec->nxt)
           {
            gslot = FindClassSlot(prec->cls,slot->name);
            if ((gslot != NULL) ? (gslot->noinherit == 0) : FALSE)
              break;
           }
         if (gslot != NULL)
           {
            if (slot->default_specified == 0)
              {
               slot->default_specified = gslot->default_specified;
               slot->dynamic = gslot->dynamic;
               slot->defexp = PackExpression(gslot->defexp);
              }
            if (slot->cardinality_specified == 0)
              {
               slot->cardinality_specified = gslot->cardinality_specified;
               slot->multiple = gslot->multiple;
              }
            if (slot->storage_specified == 0)
              {
               slot->storage_specified = gslot->storage_specified;
               slot->shared = gslot->shared;
              }
            if (slot->access_specified == 0)
              {
               slot->access_specified = gslot->access_specified;
               slot->nowrite = gslot->nowrite;
               slot->initonly = gslot->initonly;
              }
           }
        }
     }
  }
  
/*******************************************************
  NAME         : BuildSubclassLinks
  DESCRIPTION  : Follows the list of superclasses
                 for a class and puts the class in
                 each of the superclasses' subclass
                 list.
  INPUTS       : The address of the class
  RETURNS      : Nothing useful
  SIDE EFFECTS : The subclass lists for every superclass
                 are modified.
  NOTES        : Assumes the superclass list is formed.
 *******************************************************/
static VOID BuildSubclassLinks(cls)
  CLASS_TYPE *cls;
  {
   CLASS_LINK *super,*sub,*sptr;
   
   for (super = cls->superlink ; super != NULL ; super = super->nxt)
     {
      sub = get_struct(class_link);
      sub->cls = cls;
      sub->nxt = NULL;
      if (super->cls->sublink == NULL)
        super->cls->sublink = sub;
      else
        {
         sptr = super->cls->sublink;
         while (sptr->nxt != NULL)
           sptr = sptr->nxt;
         sptr->nxt = sub;
        }
     }
  }

/**********************************************************
  NAME         : FormInstanceTemplate
  DESCRIPTION  : Forms a contiguous array of instance
                  slots for use in creating instances later
                 Also used in determining instance slot
                  indices a priori during handler defns
  INPUTS       : The class
  RETURNS      : Nothing useful
  SIDE EFFECTS : Contiguous array of instance slots formed
  NOTES        : None
 **********************************************************/
static VOID FormInstanceTemplate(cls)
  CLASS_TYPE *cls;
  {
   CLASS_LINK *cplist;
   TILINK *islots = NULL,*inhslots = NULL,*stmp;
   unsigned scnt = 0;
   register unsigned i;

   /* ========================
      Get direct class's slots
      ======================== */
   islots = MergeSlots(islots,cls,&scnt,DIRECT,&inhslots);
   
   /* ===================================================================
      Get all inherited slots - a more specific slot takes precedence
      over more general, i.e. the first class in the precedence list with
      a particular slot gets to specify its default value
      =================================================================== */
   cplist = cls->precedence;
   while (cplist != NULL)
     {
      islots = MergeSlots(islots,cplist->cls,&scnt,INHERIT,&inhslots);
      cplist = cplist->nxt;
     }
    
   /* ================================================================
      Allocate a contiguous array to store all the slots.
      They are in ascending order according to symbolic hash value and
        thus can be quickly looked up via binary search - this order
        is via the nxt fields and islots is the top.
      They are in inherited order from least specific to most specific
        via the inh fields and inhslots is the top.
      ================================================================ */
   cls->islot_cnt = scnt;
   if (scnt > 0)
     {
      cls->itemplate = (SLOT_DESC **) gm2((int) (scnt * sizeof(SLOT_DESC *)));
      cls->sordered = (unsigned *) gm2((int) (scnt * sizeof(unsigned)));
     }
   for (i = 0 ; i < scnt ; i++)
     {
      stmp = inhslots;
      inhslots = inhslots->inh;
      cls->itemplate[i] = stmp->desc;
      stmp->toffset = i;
     }
   for (i = 0 ; i < scnt ; i++)
     {
      stmp = islots;
      islots = islots->nxt;
      cls->sordered[i] = stmp->toffset;
      rtn_struct(tilink,stmp);
     }
  }
   
/********************************************************************
  NAME         : MergeSlots
  DESCRIPTION  : Adds non-duplicate slots to list and increments
                   slot count for the class instance template
  INPUTS       : 1) The old slot list
                 2) The address of class containing new slots
                 3) Caller's buffer for # of slots
                 4) A flag indicating whether the new list of slots
                    is from the direct parent-class or not.
                 5) Buffer for holding pointer to top of list 
                    of slots in inherited order (least specific to
                    most specific) - which can be traced by following
                    the inh pointers instead of the nxt pointers
                    (which is the list in sorted order by hash value)
  RETURNS      : The address of the new expanded list, or NULL
                   for an empty list
  SIDE EFFECTS : The list is expanded
                 Caller's slot count is adjusted.
                 Caller's buffer for top of list in inherited order
                   is modified.
  NOTES        : Lists are assumed to be sorted by CLIPS name
                 symbolic hash value (no duplicates)
 *******************************************************************/
static TILINK *MergeSlots(old,cls,scnt,src,inhtop)
  TILINK *old;
  CLASS_TYPE *cls;
  unsigned *scnt;
  int src;
  TILINK **inhtop;
  {
   TILINK *curr,*prv,*tmp;
   register int i;
   SLOT_DESC *new;
   
   /* ===============================================
      Process the slots in reverse order since we are
      pushing them onto a stack
      =============================================== */
   for (i = cls->slot_cnt - 1 ; i >= 0 ; i--)
     {
      new = &cls->slots[i];
     
      /* =================================================================
         A class can prevent it slots from being propagated to all but its
         direct instances
         ================================================================= */
      if ((new->noinherit == 0) ? TRUE : (src == DIRECT))
        {
         prv = NULL;
         curr = old;
         while ((curr != NULL) ? 
                (new->name->bucket >= curr->desc->name->bucket) : FALSE)
           {
            if (new->name == curr->desc->name)
              break;
            prv = curr;
            curr = curr->nxt;
           }
         if ((curr == NULL) ? TRUE : (new->name != curr->desc->name))
           {
            tmp = get_struct(tilink);
            tmp->desc = new;
            tmp->nxt = curr;
            if (prv == NULL)
              old = tmp;
            else
              prv->nxt = tmp;
            prv = tmp;
            tmp->inh = *inhtop;
            *inhtop = tmp;
            (*scnt)++;
           }
        }
     }
   return(old);
  }
  
/***********************************************************************************
  NAME         : BuildPartialOrders
  DESCRIPTION  : This function recursively records the Rule 1 and Rule 2
                 partial orders (see FindPrecedenceList()) for given
                 list of classes and their superclasses.
                 
                 A partial order is recorded by incrementing the
                 predecessor count of the successor class and adding
                 the successor class to the successor list of the
                 predecessor class in the sequence table.
  INPUTS       : 1) The sequence table
                 2) A list of classes which must precede one another
                    in the given order by Rule 2
  RETURNS      : The top of sequence table
  SIDE EFFECTS : The sequence table is built, e.g.:
                 
                 CLASS1 < CLASS2
                 CLASS1 < CLASS3
                 
                 would be recorded as:
  
                 SEQTABLE -> NXT     -> NXT       -> NXT ->   <NIL>
                             <CLASS1>   <CLASS2>     <CLASS3>
                             SUC        SUC          SUC
                              |          |            |
                              V          V            V
                             <CLASS2>   <NIL>        <NIL>
                             NXT
                              |
                              V
                             <CLASS3>
                             NXT
                              |
                              V
                             <NIL>
                             
                 The predecessor counts are stored with the classes themselves
                 and would be 0, 1 and 1 for CLASS1, CLASS2 and CLASS3 respectively.
  NOTES        : 
 ***********************************************************************************/
static SEQNODE *BuildPartialOrders(seqtable,supers)
  SEQNODE *seqtable;
  CLASS_LINK *supers;
  {
   register SEQNODE *sqp,*sqprv;
   register CLASS_LINK *spc;
   
   while (supers != NULL)
     {
      /* ============================================
         Append this class at the end of the sequence
         table only if it is not already present
         ============================================ */
      sqprv = NULL;
      for (sqp = seqtable ; sqp != NULL ; sqp = sqp->nxt)
        {
         if (sqp->cls == supers->cls)
           break;
         sqprv = sqp;
        }
      if (sqp == NULL)
        {
         sqp = get_struct(seqnode);
         sqp->cls = supers->cls;
         sqp->nxt = NULL;
         sqp->suc = NULL;
         if (sqprv == NULL)
           seqtable = sqp;
         else
           sqprv->nxt = sqp;
           
         /* ==============================================================
            If this is the first time we are adding the class, record the
            Rule 1 ( a class has precedence over its superclasses) partial
            orders, e.g. D <- A B C gives the following partial orders for
            the class D: D < A, D < B and D < C.
            
            Recording a partial order entails incrementing the predecessor
            count for the successor classes (B and C in the example above)
            and adding the successor classes to the successor list of the
            predecessor class (A in the example above).
         
            We must make sure we only do Rule 1 partial orders once per
            class, or the sequence table will give too much weight to them
            in the case that classes share some of the same superclasses.
            ============================================================== */
         for (spc = supers->cls->superlink ; spc != NULL ; spc = spc->nxt)
           RecordPartialOrder(sqp,spc->cls);
        
         /* =============================================================
            Recursively append all its superclasses immediately after it.
            This order will allow us to preserve the "family" heuristic
            in the precedence list.
            ============================================================ =*/
         seqtable = BuildPartialOrders(seqtable,supers->cls->superlink);
        }
        
      /* ===============================================================
         Record the Rule 2 (a class determines precedence of its direct
         superclasses) partial orders for this class, e.g. D <- A B C
         gives the following partial orders for class A: A < B and A < C
         
         Since the same class can be on more than one direct superclass
         list, we need to always do this step even if the class is
         already in the sequence table.
         =============================================================== */
      for (spc = supers->nxt ; spc != NULL ; spc = spc->nxt)
        RecordPartialOrder(sqp,spc->cls);
      
      supers = supers->nxt;
     }
   return(seqtable);
  }
  
/*****************************************************
  NAME         : RecordPartialOrder
  DESCRIPTION  : Given a predecessor sequence table
                 node and a successor class, this
                 procedure modifies the sequence
                 table to reflect the ordering.
                 
                 For example, A < B increments the
                 predecessor count for B and adds
                 a sequence node for B to the A
                 sequence node's successor list.
  INPUTS       : 1) Pointer to a sequence table node
                    which contains a pointer to the
                    predecessor class
                 2) Pointer to the successor class
  RETURNS      : Nothing useful
  SIDE EFFECTS : 1) The predecessor count of the
                    successor class is incremented
                 2) A new sequence table successor
                    node is allocated and initialized
                    with the address of the successor
                    class.  It is linked to the
                    successor list of the predecessor
                    node.
  NOTES        : None
 *****************************************************/
static VOID RecordPartialOrder(sqp,cls)
  SEQNODE *sqp;
  CLASS_TYPE *cls;
  {
   register SEQNODE *nsuc;

   cls->predecessors++;
   nsuc = get_struct(seqnode);
   nsuc->cls = cls;
   nsuc->suc = NULL;
   nsuc->nxt = sqp->suc;
   sqp->suc = nsuc;
  }
  
/**************************************************************************
  NAME         : PrintPartialOrderLoop
  DESCRIPTION  : This routine prints a conflicting loop (there may be more
                 than one) in the given sequence table of partial orders.
                 
                 The algorithm works as follows:
                 
                 Given the following class definitions,
                 
                 (defclass A (is-a USER))
                 (defclass B (is-a USER))
                 (defclass C (is-a A B))
                 (defclass D (is-a B A))
                 (defclass E (is-a C D))
                 
                 the sequence table will look as follows after as many
                 classes as possible have been entered onto the precedence
                 list:
                 
                                      A      USER    OBJECT   B
                 Predecessor Count    1       2        1      1
                 Successor List     B,USER  OBJECT   <NIL>   A,USER
                 
                 Construct a new table where each class is linked to one
                 of its predecessors.  For the example above this would be:
                 
                 Class:            A    USER    OBJECT    B
                 Predecessor:      B      A      USER     A
                 
                 This table is actually implemnted using the original
                 sequence table (see the code below for specifics).
                 Now using this table, start with the first node, and visit
                 successive nodes by following the predecessor links.  Mark
                 each node ass "visited".  When a previously visited node is
                 encountered, the loop has been found.
                 
                 In the case above, we start with A, goto B and then goto A
                 again which we have already seen.  So starting from where
                 we found the loop (A) we follow the links again, printing
                 the nodes as we go, until we're back where we started:
                 A B A.  Notice that this loop reflects the Rule 2 conflicts
                 between Class C and Class D in Class E's precedence list.
                 
  INPUTS       : The remaining sequence table of conflicting partial orders
  RETURNS      : Nothing useful
  SIDE EFFECTS : The predecessor counts and successor lists are modified to
                 implement the loop detection.
  NOTES        : This algorithm is adopted from one  given in Donald
                 Knuth's The Art of Computer Programming - Vol. I
                 (Fundamental Algorithms).
 **************************************************************************/
static VOID PrintPartialOrderLoop(seqtable)
  SEQNODE *seqtable;
  {
   register SEQNODE *sqp1,*sqp2,*prc;
   SEQNODE *stmp;
   
   /* ====================================================
      Set the predecessor count of every node to 0 so that
      this field can be used as a marker
      ==================================================== */
   for (sqp1 = seqtable ; sqp1 != NULL ; sqp1 = sqp1->nxt)
     sqp1->cls->predecessors = 0;
     
   /* =====================================================
      Mark each node in the sequence table with one of its
      predecessors.  If the class has already been marked
      (predecessor count > 0), don't bother.  This is
      accomplished by adding a node to the front of its
      successors' successor lists.
      
      When the process is finished, all nodes will have one
      predecessor chained to them by their 'suc' field.
      (If any nodes had had no predecessors, they would not
       still be in the table.)
      ===================================================== */
   for (sqp1 = seqtable ; sqp1 != NULL ; sqp1 = sqp1->nxt)
     {
      if (sqp1->cls->predecessors == 0)
        {
         prc = sqp1->suc;
         sqp1->suc = NULL;
        }
      else
        {
         prc = sqp1->suc->nxt;
         sqp1->suc->nxt = NULL;
        }
      while (prc != NULL)
        {
         for (sqp2 = seqtable ; sqp2 != NULL ; sqp2 = sqp2->nxt)
           if (sqp2->cls == prc->cls)
             break;
         if (sqp2->cls->predecessors == 0)
           {
            stmp = get_struct(seqnode);
            stmp->cls = sqp1->cls;
            stmp->suc = sqp1;
            stmp->nxt = sqp2->suc;
            sqp2->suc = stmp;
            sqp2->cls->predecessors = 1;
           }
         stmp = prc;
         prc = prc->nxt;
         rtn_struct(seqnode,stmp);
        }
     }
   
   /* =================================================
      Set the predecessor count of every node back to 0
      so that this field can be used as a marker again
      ================================================= */
   for (sqp1 = seqtable ; sqp1 != NULL ; sqp1 = sqp1->nxt)
     sqp1->cls->predecessors = 0;
     
   /* =====================================================
      Now start with the first node in the sequence table,
      and follow the predecessors links, marking the
      nodes as they are visited.  When we reach a node
      we have been before, we have found a loop!
      Follow all the marked nodes again  starting from the
      CURRENT position to print the loop.
      ===================================================== */
   sqp1 = seqtable;
   while (sqp1->cls->predecessors == 0)
     {
      sqp1->cls->predecessors = 1;
      sqp1 = sqp1->suc->suc;
     }
   
   PrintCLIPS(WERROR,"Precedence loop in superclasses:\n");
   while (sqp1->cls->predecessors == 1)
     {
      PrintCLIPS(WERROR,ValueToString(sqp1->cls->name));
      PrintCLIPS(WERROR,"\n");
      sqp1->cls->predecessors = 0;
      sqp1 = sqp1->suc->suc;
     }
   PrintCLIPS(WERROR,ValueToString(sqp1->cls->name));
   PrintCLIPS(WERROR,"\n");
  }
  
#endif

#endif

/***************************************************
  NAME         : 
  DESCRIPTION  : 
  INPUTS       : 
  RETURNS      : 
  SIDE EFFECTS : 
  NOTES        : 
 ***************************************************/


