   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*               CLIPS Version 5.10  07/17/91          */
   /*                                                     */
   /*                  DEFINSTANCES MODULE                */
   /*******************************************************/

/*************************************************************/
/* Purpose: CLIPS Kernel definstances interface commands     */
/*              and routines                                 */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
/* =========================================
   *****************************************
               EXTERNAL DEFINITIONS
   =========================================
   ***************************************** */
#include "setup.h"

#if OBJECT_SYSTEM

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
#include "bload.h"
#endif
#include "clipsmem.h"
#include "constant.h"
#include "constrct.h"
#include "engine.h"
#include "evaluatn.h"
#include "generate.h"
#include "inscom.h"
#include "router.h"
#include "scanner.h"
#include "utility.h"

#define _DEFINS_SOURCE_
#include "defins.h"

#if (! BLOAD_ONLY) && (! RUN_TIME)
extern struct token ObjectParseToken;
#endif

/* =========================================
   *****************************************
                   CONSTANTS
   =========================================
   ***************************************** */
   
/* =========================================
   *****************************************
               MACROS AND TYPES
   =========================================
   ***************************************** */

/* =========================================
   *****************************************
      INTERNALLY VISIBLE FUNCTION HEADERS
   =========================================
   ***************************************** */

#if ANSI_COMPILER
#if (! BLOAD_ONLY) && (! RUN_TIME)
static int ParseDefinstances(char *);
static SYMBOL_HN *ParseDefinstancesName(char *);
static int RemoveDefinstances(DEFINSTANCES *);
static VOID SaveDefinstances(char *);
#endif
static DEFINSTANCES *FindDefinstancesBySymbol(SYMBOL_HN *);
static VOID InitializeDefinstances(void);

#else

#if (! BLOAD_ONLY) && (! RUN_TIME)
static int ParseDefinstances();
static SYMBOL_HN *ParseDefinstancesName();
static int RemoveDefinstances();
static VOID SaveDefinstances();
#endif

static DEFINSTANCES *FindDefinstancesBySymbol();
static VOID InitializeDefinstances();

#endif

/* =========================================
   *****************************************
      EXTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */
globle DEFINSTANCES *DefinstancesList = NULL;

/* =========================================
   *****************************************
      INTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */
static DEFINSTANCES *DefinstancesListBottom = NULL;
   
/* =========================================
   *****************************************
          EXTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
   
/***************************************************
  NAME         : SetupDefinstances
  DESCRIPTION  : Adds the definstance support routines
                   to the CLIPS Kernel
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Appropriate function lists modified
  NOTES        : None
 ***************************************************/
globle VOID SetupDefinstances()
  {
#if ! RUN_TIME
#if ! BLOAD_ONLY
   AddConstruct("definstances",ParseDefinstances);
   AddClearFunction("definstances",(VOID (*)(VOID_ARG)) ClearDefinstances,0);
   DefineFunction("undefinstances",'v',PTIF CmdUndefinstances,"CmdUndefinstances");
   AddSaveFunction("definstances",PTIF SaveDefinstances,0);
#endif
#if DEBUGGING_FUNCTIONS
   DefineFunction("ppdefinstances",'v',PTIF PPDefinstances ,"PPDefinstances");
   DefineFunction("list-definstances",'v',PTIF CmdListDefinstances,"CmdListDefinstances");
#endif
#endif
   AddResetFunction("definstances",(VOID (*)()) InitializeDefinstances,0);
  }

/***************************************************
  NAME         : SetDefinstancesList
  DESCRIPTION  : Sets Definstances list pointers
  INPUTS       : The top of the list
  RETURNS      : Nothing useful
  SIDE EFFECTS : List pointers set
  NOTES        : WARNING - does not validate this
                  operation
 ***************************************************/
globle VOID SetDefinstancesList(top)
  DEFINSTANCES *top;
  {
   DefinstancesList = top;
   DefinstancesListBottom = top;
   if (top != NULL)
     {
      while (DefinstancesListBottom->nxt != NULL)
        DefinstancesListBottom = DefinstancesListBottom->nxt;
     }
  }
  
/***************************************************
  NAME         : GetDefinstancesName
  DESCRIPTION  : Gets the name of a definstances
  INPUTS       : Pointer to a definstances
  RETURNS      : Name-string of definstances
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
globle char *GetDefinstancesName(ptr)
  VOID *ptr;
  {
   return(ValueToString(((DEFINSTANCES *) ptr)->name));
  }
  
/***********************************************************
  NAME         : GetNextDefinstances
  DESCRIPTION  : Finds first or next definstances
  INPUTS       : The address of the current definstances
  RETURNS      : The address of the next definstances
                   (NULL if none)
  SIDE EFFECTS : None
  NOTES        : If ptr == NULL, the first definstances
                    is returned.
 ***********************************************************/
globle VOID *GetNextDefinstances(ptr)
  VOID *ptr;
  {
   if (ptr == NULL)
     return((VOID *) DefinstancesList);
   return((VOID *) ((DEFINSTANCES *) ptr)->nxt);
  }
  
/***************************************************
  NAME         : FindDefinstances
  DESCRIPTION  : Looks up a definstance construct
                   by name-string
  INPUTS       : The symbolic name
  RETURNS      : The definstance address, or NULL
                    if not found
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
globle VOID *FindDefinstances(name)
  char *name;
  {
   SYMBOL_HN *dsym;

   dsym = FindSymbol(name);
   if (dsym == NULL)
     return(NULL);
   return((VOID *) FindDefinstancesBySymbol(dsym));
  }
  
/***************************************************
  NAME         : IsDefinstancesDeletable
  DESCRIPTION  : Determines if a definstances
                   can be deleted
  INPUTS       : Address of the definstances
  RETURNS      : TRUE if deletable, FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
globle int IsDefinstancesDeletable(ptr)
  VOID *ptr;
  {
#if BLOAD_ONLY || RUN_TIME
   return(FALSE);
#else
#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded())
     return(FALSE);
#endif
   return((((DEFINSTANCES *) ptr)->busy == 0) ? TRUE : FALSE);
#endif
  }

/***********************************************************
  NAME         : CmdUndefinstances
  DESCRIPTION  : Removes a definstance
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Definstance deallocated
  NOTES        : CLIPS Syntax : (undefinstances <name> | *)
 ***********************************************************/
globle VOID CmdUndefinstances()
  {
   DATA_OBJECT temp;
   DEFINSTANCES *dptr;

   if (ArgCountCheck("undefinstances",EXACTLY,1) != 1)
     return;
   if (ArgTypeCheck("undefinstances",1,SYMBOL,&temp) == FALSE)
     return;
   dptr = (DEFINSTANCES *) FindDefinstances(DOToString(temp));
   if ((dptr == NULL) ? (strcmp(DOToString(temp),"*") != 0) : FALSE)
     {
      PrintCLIPS(WERROR,"Unable to find definstances ");
      PrintCLIPS(WERROR,DOToString(temp));
      PrintCLIPS(WERROR," in function undefinstances.\n");
     }
   DeleteDefinstances((VOID *) dptr);
  }

#if (! BLOAD_ONLY) && (! RUN_TIME)
    
/***************************************************
  NAME         : ClearDefinstances
  DESCRIPTION  : Removes all definstances constructs
  INPUTS       : None
  RETURNS      : TRUE if successful, FALSE otherwise
  SIDE EFFECTS : All definstances deallocated
  NOTES        : None
 ***************************************************/
globle int ClearDefinstances()
  {
   DEFINSTANCES *dptr,*dhead;
   int success = TRUE;
   
#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded())
     return(FALSE);
#endif
  dhead = DefinstancesList;
  while (dhead != NULL)
    {
     dptr = dhead;
     dhead = dhead->nxt;
     if (RemoveDefinstances(dptr) == 0)
       success = FALSE;
    }
   return(success);
  }
  
#endif

/***********************************************************
  NAME         : DeleteDefinstances
  DESCRIPTION  : Removes a definstance
  INPUTS       : Address of definstances to remove
  RETURNS      : 1 if successful, 0 otherwise
  SIDE EFFECTS : Definstance deallocated
  NOTES        : None
 ***********************************************************/
globle int DeleteDefinstances(vptr)
  VOID *vptr;
  {
   DEFINSTANCES *dptr;
   
   dptr = (DEFINSTANCES *) vptr;
#if RUN_TIME || BLOAD_ONLY
   PrintCLIPS(WERROR,"Unable to delete definstances ");
   PrintCLIPS(WERROR,(dptr != NULL) ? ValueToString(dptr->name) : "*");
   PrintCLIPS(WERROR,".\n");
   return(0);
#else

#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded())
     {
      PrintCLIPS(WERROR,"Unable to delete definstances ");
      PrintCLIPS(WERROR,(dptr != NULL) ? ValueToString(dptr->name) : "*");
      PrintCLIPS(WERROR,".\n");
      return(0);
     }
#endif
   if (dptr == NULL)
     return(ClearDefinstances());
   return(RemoveDefinstances(dptr));
#endif
  }
  
#if DEBUGGING_FUNCTIONS

/***************************************************
  NAME         : CmdListDefinstances
  DESCRIPTION  : Lists all definstances by name
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : CLIPS Syntax : (list-definstances)
 ***************************************************/
globle VOID CmdListDefinstances()
  {
   if (ArgCountCheck("list-definstances",EXACTLY,0) != 0)
     return;
   ListDefinstances();
  }

/***************************************************************
  NAME         : PPDefinstances
  DESCRIPTION  : Prints out the pretty-print form of a definstance
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : CLIPS Syntax : (ppdefinstances <name>)
 ***************************************************************/
globle VOID PPDefinstances()
  {
   DATA_OBJECT temp;
   DEFINSTANCES *dptr;

   
   if (ArgCountCheck("ppdefinstances",EXACTLY,1) != 1)
     return;
   if (ArgTypeCheck("ppdefinstances",1,SYMBOL,&temp) == FALSE)
     return;
   dptr = (DEFINSTANCES *) FindDefinstances(DOToString(temp));
   if (dptr == NULL)
     {
      PrintCLIPS(WERROR,"Unable to find definstances ");
      PrintCLIPS(WERROR,DOToString(temp));
      PrintCLIPS(WERROR," in function ppdefinstances.\n");
     }
   else if (dptr->pp_form != NULL)
     PrintCLIPS(WDISPLAY,dptr->pp_form);
  }

/*****************************************************
  NAME         : GetDefinstancesPPForm
  DESCRIPTION  : Gets a definstances pretty print form
  INPUTS       : Address of the definstances
  RETURNS      : Pretty print form
  SIDE EFFECTS : None
  NOTES        : None
 *****************************************************/
globle char *GetDefinstancesPPForm(ptr)
  VOID *ptr;
  {
   return(((DEFINSTANCES *) ptr)->pp_form);
  }

/***************************************************
  NAME         : ListDefinstances
  DESCRIPTION  : Lists all definstances by name
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
globle VOID ListDefinstances()
  {
   DEFINSTANCES *dptr;
   long count = 0;

   for (dptr = DefinstancesList ; dptr != NULL ; dptr = dptr->nxt)
     {
      count++;
      PrintCLIPS(WDISPLAY,ValueToString(dptr->name));
      PrintCLIPS(WDISPLAY,"\n");
     }
   PrintTally(WDISPLAY,count,"definstances block","definstances blocks");
  }
  
#endif
    
/* =========================================
   *****************************************
          INTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
   
#if (! BLOAD_ONLY) && (! RUN_TIME)
  
/*********************************************************************
  NAME         : ParseDefinstances
  DESCRIPTION  : 
  INPUTS       : The logical name of the input source
  RETURNS      : 
  SIDE EFFECTS : FALSE if no errors, TRUE otherwise
  NOTES        : CLIPS Syntax :
  
                 (definstances  <name> [<comment>]
                    <instance-definition>+)
                    
                 <instance-definition> ::= 
                    (<instance-name> of <class-name> <slot-override>*)
                    
                 <slot-override> ::= (<slot-name> <value-expression>*)
 *********************************************************************/
static int ParseDefinstances(read_source)
  char *read_source;
  {
   SYMBOL_HN *dname;
   VOID *mkinsfcall;
   EXPRESSION *mkinstance,*mkbot = NULL;
   DEFINSTANCES *dobj,*dold;
   
   SetPPBufferStatus(ON);
   FlushPPBuffer();          
   SetIndentDepth(3);    
   SavePPBuffer("(definstances ");

#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded())
     {
      PrintCLIPS(WERROR,"Cannot load definstances with binary load in effect.\n");
      return(TRUE);
     }
#endif
   dname = ParseDefinstancesName(read_source);
   if (dname == NULL)
     {
      SyntaxErrorMessage("definstances");
      return(TRUE);
     }
   dold = FindDefinstancesBySymbol(dname);
   if (GetPrintWhileLoading() == TRUE)
     {
      if (GetCompilationsWatch() == TRUE)
        {
         PrintCLIPS(WDIALOG,dold != NULL ? "Redefining" : "Defining");
         PrintCLIPS(WDIALOG," definstances block ");
         PrintCLIPS(WDIALOG,ValueToString(dname));
         PrintCLIPS(WDIALOG,"\n");
        }
      else
        PrintCLIPS(WDIALOG,"@");
     }
   if ((dold != NULL) ? (dold->busy > 0) : FALSE)
     {
      PrintCLIPS(WERROR,"Cannot redefine definstances ");
      PrintCLIPS(WERROR,ValueToString(dold->name));
      PrintCLIPS(WERROR," while it is in use.\n");
      return(TRUE);
     }
   dobj = get_struct(definstance);
   dobj->busy = 0;
   dobj->name = dname;
   dobj->mkinstance = NULL;
   mkinsfcall = (VOID *) FindFunction("make-instance");
   while (GetType(ObjectParseToken) == LPAREN)
     {
      mkinstance = GenConstant(KUNKNOWN,mkinsfcall);
      mkinstance = ParseInitializeInstance(mkinstance,read_source);
      if (mkinstance == NULL)
        {
         ReturnExpression(dobj->mkinstance);
         rtn_struct(definstance,dobj);
         return(TRUE);
        }
      if (ExpressionContainsVariables(mkinstance,TRUE) == TRUE)
        {
         PrintCLIPS(WERROR,"Variables cannot be accessed by definstances.\n");
         ReturnExpression(mkinstance);
         ReturnExpression(dobj->mkinstance);
         rtn_struct(definstance,dobj);
         return(TRUE);
        }
      if (mkbot == NULL)
        dobj->mkinstance = mkinstance;
      else
        GetNextArgument(mkbot) = mkinstance;
      mkbot = mkinstance;
      GetToken(read_source,&ObjectParseToken);
      PPBackup();
      PPCRAndIndent();
      SavePPBuffer(ObjectParseToken.print_rep);
     }
   if (GetType(ObjectParseToken) != RPAREN)
     {
      ReturnExpression(dobj->mkinstance);
      rtn_struct(definstance,dobj);
      SyntaxErrorMessage("definstances");
      return(TRUE);
     }
   else
     {
      if (GetConserveMemory() == FALSE)
        {
         PPBackup();
         PPBackup();
         SavePPBuffer(")\n");
         dobj->pp_form = CopyPPBuffer();
        }
      else
        dobj->pp_form = NULL;
      if (dold != NULL)
        {
         int oldcw;
         
         oldcw = GetCompilationsWatch();
         SetCompilationsWatch(OFF);
         RemoveDefinstances(dold);
         SetCompilationsWatch(oldcw);
        }
      if (DefinstancesList == NULL)
        {
         DefinstancesList = dobj;
         dobj->prv = NULL;
        }
      else
        {
         DefinstancesListBottom->nxt = dobj;
         dobj->prv = DefinstancesListBottom;
        }
      DefinstancesListBottom = dobj;
      dobj->nxt = NULL;
      mkinstance = dobj->mkinstance;
      dobj->mkinstance = PackExpression(mkinstance);
      ReturnExpression(mkinstance);
      IncrementSymbolCount(dobj->name);
      ExpressionInstall(dobj->mkinstance);
     }
   return(FALSE);
  }
  
/***********************************************************
  NAME         : ParseDefinstancesName
  DESCRIPTION  : Parses definstance name and optional comment
  INPUTS       : 1) The logical name of the input source
                 2) The address of the caller's token buffer
  RETURNS      : Address of CLIPS name symbol, or
                   NULL if there was an error
  SIDE EFFECTS : Token after name or comment is scanned
  NOTES        : Assumes "(definstances" has already
                   been scanned.
 ***********************************************************/
static SYMBOL_HN *ParseDefinstancesName(read_source)
  char *read_source;
  {
   SYMBOL_HN *dname;
   
   GetToken(read_source,&ObjectParseToken);
   if (GetType(ObjectParseToken) != SYMBOL)
     return(NULL);
   dname = (SYMBOL_HN *) GetValue(ObjectParseToken);
   
   /*================================*/
   /* Get comment too, if it exists. */
   /*================================*/

   GetToken(read_source,&ObjectParseToken);
   if (GetType(ObjectParseToken) == STRING)
     {
      PPBackup();
      SavePPBuffer(" ");
      SavePPBuffer(ObjectParseToken.print_rep);
      PPCRAndIndent();          
      GetToken(read_source,&ObjectParseToken);
     }
   else                                 
     {
      PPBackup();
      PPCRAndIndent();
      SavePPBuffer(ObjectParseToken.print_rep);
     }
   return(dname);
  }
  
/**************************************************************
  NAME         : RemoveDefinstances
  DESCRIPTION  : Deallocates and removes a definstance construct
  INPUTS       : The definstance address
  RETURNS      : 1 if successful, 0 otherwise
  SIDE EFFECTS : Existing definstance construct deleted
  NOTES        : Assumes busy count of definstance is 0
 **************************************************************/
static int RemoveDefinstances(dptr)
  DEFINSTANCES *dptr;
  {
   if (dptr->busy > 0)
     {
      PrintCLIPS(WERROR,"Unable to delete definstances ");
      PrintCLIPS(WERROR,ValueToString(dptr->name));
      PrintCLIPS(WERROR,".\n");
      return(0);
     }
   if (GetCompilationsWatch() == ON)
     {
      PrintCLIPS(WDIALOG,"Removing definstances: ");
      PrintCLIPS(WDIALOG,ValueToString(dptr->name));
      PrintCLIPS(WDIALOG,"\n");
     }
   if (dptr->nxt != NULL)
     dptr->nxt->prv = dptr->prv;
   else
     DefinstancesListBottom = dptr->prv;
   if (dptr->prv != NULL)
     dptr->prv->nxt = dptr->nxt;
   else
     DefinstancesList = dptr->nxt;
   DecrementSymbolCount(dptr->name);
   ExpressionDeinstall(dptr->mkinstance);
   ReturnPackedExpression(dptr->mkinstance);
   if (dptr->pp_form != NULL)
     rm((VOID *) dptr->pp_form,(int) ((strlen(dptr->pp_form) + 1) * sizeof(char)));
   rtn_struct(definstance,dptr);
   return(1);
  }

/***************************************************
  NAME         : SaveDefinstances
  DESCRIPTION  : Prints pretty print form of
                   definstances to specified output
  INPUTS       : The logical name of the output
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
static VOID SaveDefinstances(log_name)
  char *log_name;
  {
   DEFINSTANCES *dobj;
   
   for (dobj = DefinstancesList ; dobj != NULL ; dobj = dobj->nxt)
     {
      if (dobj->pp_form != NULL)
        {
         PrintInChunks(log_name,dobj->pp_form);
         PrintCLIPS(log_name,"\n");
        }
     }
  }
  
#endif
  
/***************************************************
  NAME         : FindDefinstancesBySymbol
  DESCRIPTION  : Looks up a definstance construct
                   by name-symbol
  INPUTS       : The symbolic name
  RETURNS      : The definstance address, or NULL
                    if not found
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
static DEFINSTANCES *FindDefinstancesBySymbol(name)
  SYMBOL_HN *name;
  {
   DEFINSTANCES *dptr;
   
   dptr = DefinstancesList;
   while (dptr != NULL)
     {
      if (dptr->name == name)
        return(dptr);
      dptr = dptr->nxt;
     }
   return(NULL);
  }
  
/***************************************************
  NAME         : InitializeDefinstances
  DESCRIPTION  : Calls EvaluateExpression for each of
                   the make-instance calls in all
                   of the definstances constructs
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : All instances in the definstances
                   are evaluated (and created if
                   there are no errors)
                 Any previously existing instances
                 are deleted first.
  NOTES        : None
 ***************************************************/
static VOID InitializeDefinstances()
  {
   DEFINSTANCES *dptr;
   EXPRESSION *exp;
   int oldec;
   DATA_OBJECT temp;

   DestroyAllInstances();
   oldec = ExecutingConstruct();
   SetExecutingConstruct(TRUE);
   for (dptr = DefinstancesList ; dptr != NULL ; dptr = dptr->nxt)
     {
      if (HaltExecution)
        break;
      dptr->busy++;
      for (exp = dptr->mkinstance ; exp != NULL ; exp = GetNextArgument(exp))
        {
         EvaluateExpression(exp,&temp);
         if (HaltExecution)
           break;
         if ((temp.type == SYMBOL) && (temp.value == (VOID *) CLIPSFalseSymbol))
           break;
        }
      dptr->busy--;
     }
   SetExecutingConstruct(oldec);
  }

#endif

/***************************************************
  NAME         : 
  DESCRIPTION  : 
  INPUTS       : 
  RETURNS      : 
  SIDE EFFECTS : 
  NOTES        : 
 ***************************************************/


