   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                 FACT MANAGER MODULE                 */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _FACTMNGR_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_

#include "setup.h"

#include "constant.h"
#include "symbol.h"
#include "clipsmem.h"
#include "expressn.h"
#include "evaluatn.h"
#include "facts.h"
#include "scanner.h"
#include "router.h" 
#include "match.h"  
#include "generate.h"
#include "build.h" 
#include "reteutil.h" 
#include "utility.h"
#include "retract.h"
#include "intrfile.h"
#include "constrct.h"
#include "parsutil.h"

#if DEFRULE_CONSTRUCT
#include "engine.h"
#include "lgcldpnd.h"
#endif

#if DEFTEMPLATE_CONSTRUCT
#include "deftempl.h"
#include "deftmfun.h" 
#endif

#if ANSI_COMPILER
   extern VOID             InitFactCommands(void);
#else
   extern VOID             InitFactCommands();
#endif

#if ANSI_COMPILER  
   static VOID                    InitializeFactHashTable(void);
   static VOID                    ResetFacts(void);
#else
   static VOID                    InitializeFactHashTable();
   static VOID                    ResetFacts();
#endif

/****************************************/
/* GLOBAL INTERNAL VARIABLE DEFINITIONS */
/****************************************/

   globle int              ChangeToFactList = CLIPS_FALSE; 

/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/

   static struct fact     *GarbageFacts = NULL;
   static struct fhash   **FactHashtable;
   static struct fact     *LastFact = NULL;
   static struct fact     *FactList = NULL;
   static long int         NextFactIndex = 0L; 
   static long int         NumberOfFacts = 0;
   static struct fact     *ListOfSegments = NULL;
   static BOOLEAN          FactDuplication = CLIPS_FALSE;
   static BOOLEAN          AssertRetractInProgress = CLIPS_FALSE;
#if DEBUGGING_FUNCTIONS
   static int              WatchFacts = OFF;
#endif
   
/************************************************************************/
/* FactCompare:  Determines if two facts are identical.  Returns       */ 
/*   CLIPS_TRUE (1) if identical, and CLIPS_FALSE (0) if not identical. */
/************************************************************************/
globle int FactCompare(fact1,fact2)
  struct fact *fact1, *fact2; 
  {
   struct element *elem1, *elem2;
   int length, i = 0;

   length = fact1->factLength;
   if (length != (int) fact2->factLength)
     { return(CLIPS_FALSE); }

   elem1 = fact1->atoms;
   elem2 = fact2->atoms;

   /*==================================================*/
   /* Compare each field of both facts until the facts */
   /* match completely or the facts mismatch.          */
   /*==================================================*/

   while (i < length)
     {
      if (elem1[i].type != elem2[i].type)
        { return(CLIPS_FALSE); }
      
      if (elem1[i].value != elem2[i].value)
        { return(CLIPS_FALSE); }

      i++;
     }
   return(CLIPS_TRUE); 
  }
  
/***************************************/
/* ListFacts:                       */
/***************************************/
globle VOID ListFacts()
  {   
   struct fact *factPtr;
   
   factPtr = (struct fact *) GetNextFact(NULL); 
   while (factPtr != NULL)
     {
      PrintFactWithIdentifier(WDISPLAY,factPtr);
      PrintCLIPS(WDISPLAY,"\n");
      factPtr = (struct fact *) GetNextFact(factPtr);
     }
  }
   
/***********************************************/
/* PrintFactWithIdentifier:  Displays a single */  
/*   fact preceded by its fact identifier.     */
/***********************************************/
globle VOID PrintFactWithIdentifier(logicalName,factPtr)
  char *logicalName;
  struct fact *factPtr;
  {
   char print_space[20];
   
   sprintf(print_space,"f-%-5ld ",factPtr->factIndex);
   PrintCLIPS(logicalName,print_space);
   PrintFact(logicalName,factPtr);
  }
  
/***********************************************/
/* PrintFact: Displays the individual elements */
/*   of a fact enclosed within paretheses.     */
/***********************************************/
globle VOID PrintFact(logicalName,factPtr)
  char *logicalName;
  struct fact *factPtr;
  {
   struct element *sublist;
   int length, i;

   PrintCLIPS(logicalName,"(");
   
#if DEFTEMPLATE_CONSTRUCT
   if (PrintTemplateFact(logicalName,factPtr) == CLIPS_TRUE) 
     {
      PrintCLIPS(logicalName,")");
      return;
     }
#endif

   sublist = factPtr->atoms;
   length = factPtr->factLength;

   for (i = 0; i < length ; i++)
     {
      PrintAtom(logicalName,sublist[i].type,sublist[i].value);
      if (i + 1 != length)
        { PrintCLIPS(logicalName," "); }
     }
     
   PrintCLIPS(logicalName,")");
  }

/*****************************************************/
/* HashFact:  Returns a hash value for a given fact. */
/*****************************************************/
globle int HashFact(fptr)
  struct fact *fptr;
  {
   int length, i = 0;
   unsigned int tvalue;
   unsigned int count = 0;
   int hashValue = 0;
   struct element *fieldPtr;
   union
     {
      double fv;
      unsigned int liv;
     } fis;

   length = fptr->factLength;
   fieldPtr = fptr->atoms;
   
    while (i < length)
     {
      if (fieldPtr[i].type == FLOAT)
        {
         fis.fv = ValueToDouble(fieldPtr[i].value);
         count += (fis.liv * (i + 29));
        } 
      else if (fieldPtr[i].type == INTEGER)
        { count += (int) (((int) ValueToLong(fieldPtr[i].value)) * (i + 29)); } 
      else if ((fieldPtr[i].type == FACT_ADDRESS) || 
#if OBJECT_SYSTEM
               (fieldPtr[i].type == INSTANCE) || 
#endif
               (fieldPtr[i].type == EXTERNAL_ADDRESS))
        { count += ((int) fieldPtr[i].value); } 
      else
       {
        tvalue = ((SYMBOL_HN *) fieldPtr[i].value)->bucket;             
        count += (tvalue * (i + 31));
       }
      i++;
     }

   hashValue = (int) (count % SIZE_FACT_HASH);
   if (hashValue < 0) hashValue = - hashValue;
   return(hashValue); 
  }

/*****************************************************************/
/* FactExists:  Determines if a fact in the the fact hash table. */
/*   Returns the fact if it is and NULL otherwise.               */
/*****************************************************************/
globle struct fact *FactExists(fptr,hashValue)
  struct fact *fptr;
  int hashValue;
  {
   struct fhash *hash_ptr;

   hash_ptr = FactHashtable[hashValue];

   while (hash_ptr != NULL)
     {
      if (FactCompare(fptr,hash_ptr->fptr) == 1) return(hash_ptr->fptr);
      hash_ptr = hash_ptr->next;
     }

   return(NULL);
  }

/*******************************************************/
/* AddHashedFact:  Adds a fact to the fact hash table. */
/*******************************************************/
globle VOID AddHashedFact(fptr,hashValue)
  struct fact *fptr;
  int hashValue;
  {
   struct fhash *newhash, *temp;

   newhash = get_struct(fhash);
   newhash->fptr = fptr;

   temp = FactHashtable[hashValue];
   FactHashtable[hashValue] = newhash;
   newhash->next = temp;
  }

/************************************************************/
/* RemoveHashedFact:  Removes a fact from the fact hash table. */
/************************************************************/
globle BOOLEAN RemoveHashedFact(fptr)
  struct fact *fptr;
  {
   int hashValue;
   struct fhash *hptr, *prev = NULL;

   hashValue = HashFact(fptr);
   
   hptr = FactHashtable[hashValue];
   while (hptr != NULL)
     {
      if (hptr->fptr == fptr)
        {
         if (prev == NULL)
           { 
            FactHashtable[hashValue] = hptr->next;
            rtn_struct(fhash,hptr);
            return(1);
           }
         else
           {
            prev->next = hptr->next;
            rtn_struct(fhash,hptr);
            return(1);
           }
        }
      prev = hptr;
      hptr = hptr->next;
     }
   return(0);
  }
  
/**********************************************************/
/* RetractFact:  Retracts a fact from the fact list given */
/*   a pointer to the fact.                               */
/**********************************************************/
globle BOOLEAN RetractFact(theFact)
  VOID *theFact;
  {
   struct fact *temp_ptr;
   int rv = -1;
#if DEFRULE_CONSTRUCT
   int ov;
#endif
   struct fact *factPtr; 
   
   if (AssertRetractInProgress)
     {
      PrintCLIPS(WDISPLAY,"\nWARNING: Facts may not be retracted during an assert or a retract\n");
      return(CLIPS_FALSE);
     }
   AssertRetractInProgress = CLIPS_TRUE;
    
   if (theFact == NULL)
     {
      RemoveAllFacts();
      AssertRetractInProgress = CLIPS_FALSE; 
      return(CLIPS_TRUE);
     }
     
   factPtr = (struct fact *) theFact; 

   while (factPtr != NULL) 
     {
      /*===================================================*/
      /* Reset the evaluation error flag since expressions */
      /* will be evaluated as part of the retract.         */
      /*===================================================*/
      
      SetEvaluationError(CLIPS_FALSE);
      
      /*======================================================*/
      /* Check to see if the fact has already been retracted. */
      /*======================================================*/

      if (factPtr->garbage) 
        {
         if (rv < 0) rv = CLIPS_FALSE;
#if DEFRULE_CONSTRUCT
         factPtr = GetNextLogicalRetraction();
#else
         factPtr = NULL;
#endif
         continue;
        }

      /*=========================================*/
      /* Show retraction if facts being watched. */
      /*=========================================*/

#if DEBUGGING_FUNCTIONS   
      if (WatchFacts == ON)
        {
         PrintCLIPS(WTRACE,"<== ");
         PrintFactWithIdentifier(WTRACE,factPtr);
         PrintCLIPS(WTRACE,"\n");
        } 
  
      if (CRSVTraceActive())
        {
         char print_space[20];
      
         PrintCLIPS("wcrsv_tr","R    ");
         sprintf(print_space,"%-5ld  ",factPtr->factIndex);
         PrintCLIPS("wcrsv_tr",print_space);
         PrintAtom("wcrsv_tr",factPtr->atoms[0].type,factPtr->atoms[0].value);        
         PrintCLIPS("wcrsv_tr","\n");
        }  
#endif
     
      ChangeToFactList = CLIPS_TRUE;
    
      /*=====================================*/
      /* Delete the fact from the fact list. */
      /*=====================================*/

#if DEFRULE_CONSTRUCT
      RemoveFactDependencies(factPtr);
      ov = SetRuleDeletions(CLIPS_FALSE);
#endif

      RemoveHashedFact(factPtr);

      if (factPtr == LastFact)
        { LastFact = factPtr->previous; }
   
      if (factPtr->previous == NULL)
        {
         /* Delete the head of the fact list. */
         FactList = FactList->next;
         if (FactList != NULL)
           { FactList->previous = NULL; }
        }
      else
        {
         /* Delete a fact other than the head of the fact list. */
         factPtr->previous->next = factPtr->next;
         if (factPtr->next != NULL)
           { factPtr->next->previous = factPtr->previous; }
        }

      temp_ptr = GarbageFacts;
      GarbageFacts = factPtr;  
      
      /* Memory bug fix change. */
      
      FactDeinstall(factPtr);
      EphemeralItemCount++;
      EphemeralItemSize += sizeof(struct fact) + (sizeof(struct element) * factPtr->factLength);
     
      /* End Memory bug fix change. */
      
      factPtr->next = temp_ptr;
      factPtr->garbage = CLIPS_TRUE;
      if (rv < 0) rv = CLIPS_TRUE;
   
      /*================================================*/
      /* Loop through the list of all the patterns that */
      /* matched the fact.                              */
      /*================================================*/

#if DEFRULE_CONSTRUCT
      NetworkRetract((struct patternMatch *) factPtr->list,factPtr->factIndex);
   
      if (ExecutingRule == NULL)
        { FlushGarbagePartialMatches(); }
     
      SetRuleDeletions(ov);
   
      factPtr = GetNextLogicalRetraction();
#else
      factPtr = NULL;
#endif
     }

   AssertRetractInProgress = CLIPS_FALSE; 
   return(rv);
  }

/*******************************************************************/
/* RemoveOldFacts:  Returns facts that have been retracted to the  */
/*   pool of available memory.  It is necessary to postpone        */
/*   returning the facts to memory because RHS actions retrieve    */
/*   their variable bindings directly from the fact data structure */
/*   and the facts may be in use in other data structures.         */
/*******************************************************************/
globle VOID RemoveOldFacts()
  {
   struct fact *factPtr, *nextPtr, *lastPtr = NULL;

   factPtr = GarbageFacts;
   while (factPtr != NULL)
     {
      nextPtr = factPtr->next;
      /* Memory Bug Fix. */
      /* if ((factPtr->depth > CurrentEvaluationDepth) && (factPtr->count == 0)) */
      if (factPtr->count == 0)
        {
         EphemeralItemCount--;
         EphemeralItemSize -= sizeof(struct fact) + (sizeof(struct element) * factPtr->factLength);
         /* FactDeinstall(factPtr); */ /* Memory Bug Fix. */
         ReturnElements(factPtr);
         if (lastPtr == NULL) GarbageFacts = nextPtr;
         else lastPtr->next = nextPtr;
        }
      else
        { lastPtr = factPtr; }
        
      factPtr = nextPtr;
     }
  }

/*********************************************************************/
/* AddFact: Places a fact onto the end of the fact list and calls    */
/*   compare to filter the fact through the pattern network. Returns */  
/*   null if the fact was already in the knowledge base, and a       */
/*   pointer to the fact if it was not in the knowledge base.        */
/*********************************************************************/
globle VOID *AddFact(theFact)
  VOID *theFact;
  {
   struct fact *tempPtr;
   int hashValue;
   int length, i;
   struct element *fieldPtr;
   struct fact *newFact;
#if DEFRULE_CONSTRUCT
   int ov;
#endif

   if (AssertRetractInProgress)
     {
      PrintCLIPS(WDISPLAY,"\nWARNING: Facts may not be asserted during an assert or a retract\n");
      return(CLIPS_FALSE);
     }
   AssertRetractInProgress = CLIPS_TRUE;
   
   /*=============================================================*/
   /* Replace invalid data types in the fact with the symbol nil. */
   /*=============================================================*/
   
   newFact = (struct fact *) theFact;
   length = newFact->factLength;
   fieldPtr = newFact->atoms;
   
   for (i = 0; i < length; i++)
     {
      if (fieldPtr[i].type == RVOID)
        {
         fieldPtr[i].type = SYMBOL;
         fieldPtr[i].value = (VOID *) AddSymbol("nil");
        } 
     }
     
   /*========================================================*/
   /* If fact assertions are being checked for duplications, */
   /* then search the fact list for a duplicate fact.        */
   /*========================================================*/

   hashValue = HashFact(newFact);
   
   if (FactDuplication == CLIPS_FALSE)
     {
      if ((tempPtr = FactExists(newFact,hashValue)) != NULL)
        {
         ReturnElements(newFact);
#if DEFRULE_CONSTRUCT
         AddLogicalDependencies(tempPtr,CLIPS_TRUE);
#endif
         AssertRetractInProgress = CLIPS_FALSE;
         return(NULL);
        }
     }
        
#if DEFRULE_CONSTRUCT
   if (AddLogicalDependencies(newFact,CLIPS_FALSE) == FALSE)
     {
      ReturnElements(newFact);
      AssertRetractInProgress = CLIPS_FALSE;
      return(NULL);
     }
#endif

   AddHashedFact(newFact,hashValue);

   /*===================================================*/
   /* Add the fact to the fact list. Set the fact index */
   /* for the fact and install the symbols used by the  */
   /* fact in the symbol table.                         */
   /*===================================================*/

   newFact->next = NULL;
   newFact->list = NULL;
   newFact->previous = LastFact;
   if (LastFact == NULL)
     { FactList = newFact; }
   else
     { LastFact->next = newFact; }
   LastFact = newFact;
   
   newFact->factIndex = NextFactIndex++;
   FactInstall(newFact);

   /*===============================================*/
   /* Indicate the addition of the fact to the fact */
   /* list if facts are being watched.              */
   /*===============================================*/

#if DEBUGGING_FUNCTIONS
   if (WatchFacts == ON)
     {
      PrintCLIPS(WTRACE,"==> ");
      PrintFactWithIdentifier(WTRACE,newFact);
      PrintCLIPS(WTRACE,"\n");
     }

   if (CRSVTraceActive())
     {
      char print_space[20];
      
      PrintCLIPS("wcrsv_tr","AS   ");
      sprintf(print_space,"%-5ld ",newFact->factIndex);
      PrintCLIPS("wcrsv_tr",print_space);
      PrintFact("wcrsv_tr",newFact);
      PrintCLIPS("wcrsv_tr","\n");
     }
#endif
     
   ChangeToFactList = CLIPS_TRUE;
   
   /*===============================*/
   /* Check for errors in the fact. */
   /*===============================*/

#if DEFTEMPLATE_CONSTRUCT   
   CheckTemplateFact(newFact);
#endif
   
   /*==============================================*/
   /* Filter the fact through the pattern network. */
   /*==============================================*/
   
#if DEFRULE_CONSTRUCT
   ov = SetRuleDeletions(CLIPS_FALSE);

   SetEvaluationError(CLIPS_FALSE);
   PatternMatch(newFact,newFact->atoms,NetworkPointer(),1,0,NULL,NULL);
   
   ForceLogicalRetractions();
   
   if (ExecutingRule == NULL) FlushGarbagePartialMatches();
   SetRuleDeletions(ov);
#endif

   AssertRetractInProgress = CLIPS_FALSE;
   return((VOID *) newFact);
  }
  
/**********************************************************************/
/* RemoveAllFacts: Loops through the fact-list and removes each fact. */
/**********************************************************************/
globle VOID RemoveAllFacts()
  {
   while (FactList != NULL)
     { RetractFact((VOID *) FactList); }
  }

/***********************************************************/
/* CreateFact:  Allocates a fact structure / element array */
/*   structure combination for the storage of a fact.      */
/***********************************************************/
globle VOID *CreateFact(size)
  int size;
  {
   struct fact *factPtr;
   struct element *atoms;
   
   if (size > 0)
     {
      atoms = (struct element *) gm3((long) sizeof(struct element) * (long) size);
      if (atoms == NULL) return(NULL);
     }
   else
     { 
      size = 0;
      atoms = NULL; 
     }

   factPtr = get_struct(fact);
   factPtr->garbage = CLIPS_FALSE;
   factPtr->factLength = size;
   factPtr->depth = CurrentEvaluationDepth;
   factPtr->factIndex = 0;
   factPtr->count = 0;
   factPtr->atoms = atoms;
   factPtr->dependents = NULL;
   factPtr->next = NULL;
   factPtr->previous = NULL;
   factPtr->list = NULL;
    
   return((VOID *) factPtr);
  }
  
/*****************************************************************/
/* ReturnElements:                                                */
/*****************************************************************/
globle VOID ReturnElements(factPtr)
  struct fact *factPtr;
  {
   if (factPtr->factLength > 0)
     { rm3(factPtr->atoms,(long) sizeof(struct element) * (long) factPtr->factLength); }
   rtn_struct(fact,factPtr);
  }
  
/**************************************************************/
/* FactInstall:  Increments all occurrences in the hash table */
/*   of symbols found in the atoms of a fact.                 */ 
/**************************************************************/
globle VOID FactInstall(newFact)
  struct fact *newFact;
  {
   NumberOfFacts++;
   SegmentInstall(newFact);
   newFact->count = 1; /* Memory Bug Fix. */
  }
  
/****************************************************************/
/* FactDeinstall:  Decrements all occurrences in the hash table */
/*   of symbols found in the atoms of a fact.                   */ 
/****************************************************************/
globle VOID FactDeinstall(newFact)
  struct fact *newFact;
  {
   NumberOfFacts--;
   SegmentDeinstall(newFact);
   /* newFact->count = 0; */ /* Memory Bug Fix. */
  }

/******************************/
/* SegmentInstall:            */
/******************************/
globle VOID SegmentInstall(newFact)
  struct fact *newFact;
  {
   int length, i;
   struct element *elem_ptr;
   
   length = newFact->factLength;
   newFact->count++;

   elem_ptr = newFact->atoms;
   for (i = 0 ; i < length ; i++)
     { AtomInstall(elem_ptr[i].type,elem_ptr[i].value); }            
  }
  
/******************************/
/* SegmentDeinstall:            */
/******************************/
globle VOID SegmentDeinstall(newFact)
  struct fact *newFact;
  {
   int length, i;
   struct element *elem_ptr;
   
   length = newFact->factLength;
   newFact->count--;

   elem_ptr = newFact->atoms;
   for (i = 0 ; i < length ; i++)
     { AtomDeinstall(elem_ptr[i].type,elem_ptr[i].value); }  
  }

/**************************************************************/
/* IncrementFactCount:  Increments the counter for the number */
/*   of needed references to a particular fact.               */
/**************************************************************/
globle VOID IncrementFactCount(factPtr)
  VOID *factPtr;
  { 
   ((struct fact *) factPtr)->count++; 
  }

/**************************************************************/
/* DecrementFactCount:  Decrements the counter for the number */
/*   of needed references to a particular fact.               */
/**************************************************************/
globle VOID DecrementFactCount(factPtr)
  VOID *factPtr;
  { 
   ((struct fact *) factPtr)->count--; 
  }
   
/*****************************************************************/
/* SetFactID:                              */ 
/*****************************************************************/
globle VOID SetFactID(value)
  FACT_ID value;
  {
   NextFactIndex = value;
  }
  
/*****************************************************************/
/* GetNextFact:                    */ 
/*****************************************************************/
globle VOID *GetNextFact(factPtr)
  VOID *factPtr;
  {
   if (factPtr == NULL)
     { return((VOID *) FactList); }
   
   if (((struct fact *) factPtr)->garbage) return(NULL);
   
   return((VOID *) ((struct fact *) factPtr)->next);
  }

/***************************************************/
/* GetFactPPForm:                                  */
/***************************************************/
globle VOID GetFactPPForm(buffer,bufferLength,theFact)
  char *buffer;
  int bufferLength;
  VOID *theFact;
  {
   OpenStringDestination("FactPPForm",buffer,bufferLength);
   PrintFactWithIdentifier("FactPPForm",(struct fact *) theFact);
   CloseStringDestination("FactPPForm");
  }
   
/***************************************/
/* GetFactIndex:                       */
/***************************************/
globle long int GetFactIndex(factPtr)
  VOID *factPtr;
  {      
   if (((struct fact *) factPtr)->garbage) return(-1L);
   
   return(((struct fact *) factPtr)->factIndex);
  }
 
/********************************************************/
/* AssertString:                                              */
/********************************************************/
globle VOID *AssertString(str)
  char *str;
  {
   struct fact *factPtr;

   factPtr = StringToFact(str);
   
   if (factPtr == NULL) return(NULL);
   
   return((VOID *) AddFact((VOID *) factPtr));
  }

/*******************************************************/
/* StringToFact:  Returns a fact structure that represents */
/*   the string sent as the argument.                  */
/*   Note:  Duplicates the functionality of assert ex- */
/*          cept it does not assert the fact.          */
/*******************************************************/
globle struct fact *StringToFact(str)
  char *str;
  {
   struct token inp_tkn;
   struct element *last_elem;
   struct fact *factPtr;
   int num_elems = 0;
   struct expr *assertArgs, *next_elem;
   int multi = CLIPS_FALSE, error = CLIPS_FALSE;

   last_elem = NULL;
 
   OpenStringSource("assert_str",str,0);
 
   assertArgs = GetRHSPattern("assert_str",&inp_tkn,&multi,
                                   &error,CLIPS_TRUE,CLIPS_FALSE,STOP);

   CloseStringSource("assert_str");
   
   if (error) 
     {
      ReturnExpression(assertArgs);
      return(NULL);
     }
   
   num_elems = ExpressionSize(assertArgs);
   
   if ((num_elems == 0) || ExpressionContainsVariables(assertArgs,CLIPS_TRUE))
     {
      ReturnExpression(assertArgs);
      return(NULL); 
     }
 
   factPtr = (struct fact *) CreateFact(num_elems);

   next_elem = assertArgs;
   last_elem = factPtr->atoms;
   num_elems = 0;

   while (next_elem != NULL)
     {   
      last_elem[num_elems].type = next_elem->type;
      last_elem[num_elems].value = next_elem->value;
      num_elems++;
      next_elem = next_elem->next_arg;
     }  
      
   ReturnExpression(assertArgs);
 
   return(factPtr);
  }

/*******************************************************/
/* StringToMultifield:  Returns a multifield structure */
/*    that represents the string sent as the argument. */
/*******************************************************/
globle struct fact *StringToMultifield(str)
  char *str;
  {
   struct token tkn;
   struct element *last_elem;
   struct fact *factPtr;
   int num_elems = 0;
   struct expr *top = NULL, *bot = NULL, *tmp, *next_elem;
 
   OpenStringSource("multifield_str",str,0);
   
   GetToken("multifield_str",&tkn);
   while (tkn.type != STOP)
     {
      if ((tkn.type == SYMBOL) || (tkn.type == STRING) ||
          (tkn.type == FLOAT) || (tkn.type == INTEGER) ||
          (tkn.type == INSTANCE_NAME))
        { tmp = GenConstant(tkn.type,tkn.value); }
      else
        { tmp = GenConstant(STRING,AddSymbol(tkn.print_rep)); }
        
      if (top == NULL) top = tmp;
      else bot->next_arg = tmp;
      
      bot = tmp;
      GetToken("multifield_str",&tkn);
     }

   CloseStringSource("multifield_str");
    
   num_elems = ExpressionSize(top);
 
   if (num_elems == 0) return(NULL);
      
   factPtr = (struct fact *) CreateMultifield(num_elems);

   next_elem = top;
   last_elem = factPtr->atoms;
   num_elems = 0;

   while (next_elem != NULL)
     {   
      last_elem[num_elems].type = next_elem->type;
      last_elem[num_elems].value = next_elem->value;
      num_elems++;
      next_elem = next_elem->next_arg;
     }  
      
   ReturnExpression(top);
 
   return(factPtr);
  }
   
/***************************************************/
/* GetFactListChanged:                              */
/***************************************************/
globle int GetFactListChanged()
  { return(ChangeToFactList); }
  
/***************************************************/
/* SetFactListChanged:                              */
/***************************************************/
globle VOID SetFactListChanged(value)
  int value;
  { 
   ChangeToFactList = value;
  }
  
/***************************************************/
/* GetFactDuplication:                              */
/***************************************************/
globle BOOLEAN GetFactDuplication()
  { return(FactDuplication); }
  
/***************************************************/
/* SetFactDuplication:                              */
/***************************************************/
globle BOOLEAN SetFactDuplication(value)
  int value;
  { 
   int ov;
   
   ov = FactDuplication;
   FactDuplication = value;
   return(ov);
  }

/***************************************************/
/* GetNumberOfFacts:                               */
/***************************************************/
globle long int GetNumberOfFacts()
  { return(NumberOfFacts); }
  
/**********************************************************************************/
/* InitializeFactHashTable: Purpose is to initialize the fact hash table to NULL. */
/**********************************************************************************/
static VOID InitializeFactHashTable()
   {
    int i;

    FactHashtable = (struct fhash **) gm2((int) sizeof (struct fhash *) * SIZE_FACT_HASH);
                
    if (FactHashtable == NULL) ExitCLIPS(1);
      
    for (i = 0; i < SIZE_FACT_HASH; i++) FactHashtable[i] = NULL;
   }
    
/***********************************************************/
/* CreateMultifield:                                       */
/***********************************************************/
globle VOID *CreateMultifield(size)
  int size;
  {
   struct fact *seg_ptr;
   
   if (size == 0) 
     {
      size = 1;
      seg_ptr = (struct fact *) CreateFact(size);
      seg_ptr->atoms[0].type = SYMBOL;
      seg_ptr->atoms[0].value = CLIPSFalseSymbol;
     }
   else
     { seg_ptr = (struct fact *) CreateFact(size); }
   seg_ptr->next = ListOfSegments;
   ListOfSegments = seg_ptr;
   
   EphemeralItemCount++;
   EphemeralItemSize += sizeof(struct fact) + (sizeof(struct element) * size);
      
   return((VOID *) seg_ptr);
  }

/***********************************************************/
/* AddToSegmentList:                                       */
/***********************************************************/
globle VOID AddToSegmentList(factPtr)
  struct fact *factPtr;
  {
   factPtr->depth = CurrentEvaluationDepth;
   factPtr->next = ListOfSegments;
   ListOfSegments = factPtr;
   
   EphemeralItemCount++;
   EphemeralItemSize += sizeof(struct fact) + (sizeof(struct element) * factPtr->factLength);
  }
  
/***********************************************************/
/* FlushSegments:                                         */
/***********************************************************/
globle VOID FlushSegments()
  {
   struct fact *segPtr, *nextPtr, *lastPtr = NULL;

   segPtr = ListOfSegments;
   while (segPtr != NULL)
     {
      nextPtr = segPtr->next;
      if ((segPtr->depth > CurrentEvaluationDepth) && (segPtr->count == 0))
        {
         EphemeralItemCount--;
         EphemeralItemSize -= sizeof(struct fact) + (sizeof(struct element) * segPtr->factLength);
         ReturnElements(segPtr);
         if (lastPtr == NULL) ListOfSegments = nextPtr;
         else lastPtr->next = nextPtr;
        }
      else
        { lastPtr = segPtr; }
        
      segPtr = nextPtr;
     }
  }
  
/*********************************************************************/
/* DuplicateSegment: Allocates a new segment and copies results from */
/*                  old value to new - NOT put on ListOfSegments!!   */
/*********************************************************************/  
globle VOID DuplicateSegment(dst,src)
  DATA_OBJECT_PTR dst, src;
  {
   dst->type = MULTIFIELD;
   dst->begin = 0;
   dst->end = src->end - src->begin;
   dst->value = (VOID *) CreateFact(dst->end + 1);
   CopyMemory(struct element,dst->end + 1,&((SEGMENT) dst->value)->atoms[0],
                                          &((SEGMENT) src->value)->atoms[src->begin]);
  }
  
/**********************************************************/
/* InitializeFacts: Initializes reset function for facts. */
/**********************************************************/
globle VOID InitializeFacts()
  {
   InitializeFactHashTable();
   AddResetFunction("facts",ResetFacts,20);
   AddClearFunction("facts",ResetFacts,0);
#if DEBUGGING_FUNCTIONS
   AddWatchItem("facts",&WatchFacts,80);
#endif
   InitFactCommands();
  }
  
/*********************************************************************/
/* ResetFacts: */
/*********************************************************************/  
static VOID ResetFacts()
  {
   /*====================================*/
   /* Initialize some global parameters. */
   /*====================================*/

   SetFactID(0L);  
#if DEFRULE_CONSTRUCT 
   SetPseudoFactIndex(-2L); 
#endif

   /*======================================*/
   /* Remove all facts from the fact list. */
   /*======================================*/
   
   RemoveAllFacts();
  } 
  
/***************************************************/
/* FindIndexedFact: Returns a pointer to a fact in */
/*   the fact list with the specified fact index.  */
/***************************************************/  
globle struct fact *FindIndexedFact(factIndexSought)
  long int factIndexSought;
  {
   struct fact *ptr;
   
   ptr = (struct fact *) GetNextFact(NULL);
   while (ptr != NULL)
     {
      if (ptr->factIndex == factIndexSought)
        { return(ptr); }
        
      ptr = (struct fact *) GetNextFact(ptr);
     }
     
   return(NULL);
  }
