   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*               CLIPS Version 5.10  07/17/91          */
   /*                                                     */
   /*                                                     */
   /*******************************************************/

/*************************************************************/
/* Purpose: CLIPS Generic Functions Internal Routines        */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
/* =========================================
   *****************************************
               EXTERNAL DEFINITIONS
   =========================================
   ***************************************** */
#include "setup.h"

#if DEFGENERIC_CONSTRUCT

#if OBJECT_SYSTEM
#include "classcom.h"
#include "classfun.h"
#include "insfun.h"
#else
#if ANSI_COMPILER
#include <string.h>
#endif
#endif

#include "clipsmem.h"
#include "constant.h"
#include "constrct.h"
#include "router.h"
#include "sysprime.h"
#include "utility.h"

#define _GENRCFUN_SOURCE_
#include "genrcfun.h"

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
#include "bload.h"
#include "bsave.h"
#endif

/* =========================================
   *****************************************
                   CONSTANTS
   =========================================
   ***************************************** */
#define HIGHER_PRECEDENCE -1
#define IDENTICAL          0
#define LOWER_PRECEDENCE   1

#define BEGIN_TRACE     ">>"
#define END_TRACE       "<<"

#define SYSTEM_NO       0
#define SYSTEM_YES      1

/* =========================================
   *****************************************
               MACROS AND TYPES
   =========================================
   ***************************************** */
typedef struct method_link
  {
   METHOD *method;
   struct method_link *nxt;
  } METHOD_LINK;
   
#define SaveBusyCount(gfunc)    (OldGenericBusySave = gfunc->busy)
#define RestoreBusyCount(gfunc) (gfunc->busy = OldGenericBusySave)

/* =========================================
   *****************************************
      INTERNALLY VISIBLE FUNCTION HEADERS
   =========================================
   ***************************************** */
#if ANSI_COMPILER

#if (! BLOAD_ONLY) && (! RUN_TIME)

static GENERIC_FUNC *NewGeneric(void);
static METHOD *AddGenericMethod(GENERIC_FUNC *,int,unsigned);
static int RestrictionsCompare(EXPRESSION *,int,int,METHOD *);
static int TypeListCompare(RESTRICTION *,RESTRICTION *);

#endif


static EXPRESSION *EvaluateGenericParameters(EXPRESSION *,int,SYMBOL_HN *);
static METHOD_LINK *FindApplicableMethods(GENERIC_FUNC *);
static VOID DestroyMethodLinks(METHOD_LINK *);

#if OBJECT_SYSTEM
static CLASS_TYPE *DetermineRestrictionClass(EXPRESSION *);
#endif

static int IsMethodApplicable(METHOD *);

#if DEBUGGING_FUNCTIONS
static VOID DisplayGenericCore(GENERIC_FUNC *,METHOD_LINK *);
static VOID TraceGeneric(char *);
static VOID TraceMethod(char *,int);
#endif

#else

#if (! BLOAD_ONLY) && (! RUN_TIME)

static GENERIC_FUNC *NewGeneric();
static METHOD *AddGenericMethod();
static int RestrictionsCompare();
static int TypeListCompare();

#endif

static EXPRESSION *EvaluateGenericParameters();
static METHOD_LINK *FindApplicableMethods();
static VOID DestroyMethodLinks();

#if OBJECT_SYSTEM
static CLASS_TYPE *DetermineRestrictionClass();
#endif

static int IsMethodApplicable();

#if DEBUGGING_FUNCTIONS
static VOID DisplayGenericCore();
static VOID TraceGeneric();
static VOID TraceMethod();
#endif

#endif
      

/* =========================================
   *****************************************
      EXTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */
globle GENERIC_FUNC *GenericList = NULL,
globle              *CurrentGeneric = NULL,
globle              *GenericListBottom = NULL;
globle METHOD *CurrentMethod = NULL;
globle EXPRESSION *GenericStackFrame = NULL;
globle int GenericStackSize = -1;

#if DEBUGGING_FUNCTIONS
globle int WatchGenerics = OFF;
globle int WatchMethods = OFF;
#endif

/* =========================================
   *****************************************
      INTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */
static METHOD_LINK *TopOfCore = NULL,*NextInCore = NULL;
#if (! RUN_TIME) && (! BLOAD_ONLY)
static int OldGenericBusySave;
#endif

/* =========================================
   *****************************************
          EXTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
#if (! BLOAD_ONLY) && (! RUN_TIME)

/************************************************************************
  NAME         : AddMethod
  DESCRIPTION  : (Re)defines a new method for a generic
                 If method already exists, deletes old information
                    before proceeding.
  INPUTS       : 1) The generic address
                 2) The old method address (can be NULL)
                 3) The old method array position (can be -1)
                 4) The method index to assign (0 if don't care)
                 5) The parameter expression-list
                    (restrictions attached to arg_list pointers)
                 6) The number of restrictions
                 7) The wildcard symbol (NULL if none)
                 8) Method actions
                 9) Method pretty-print form
  RETURNS      : The new (old) method address
  SIDE EFFECTS : Method added to (or changed in) method array for generic
                 Restrictions repacked into new method
                 Actions and pretty-print form attached
  NOTES        : Assumes if a method is being redefined, its busy
                   count is 0!!
                 IMPORTANT: Expects that FindMethodByRestrictions() has
                   previously been called to determine if this method
                   is already present or not.  Arguments #1 and #2
                   should be the values obtained from FindMethod...().
 ************************************************************************/
globle METHOD *AddMethod(gfunc,meth,mposn,mi,params,rcnt,wildcard,actions,pp_form)
  GENERIC_FUNC *gfunc;
  SYMBOL_HN *wildcard;
  METHOD *meth;
  int mposn,rcnt;
  unsigned mi;
  EXPRESSION *params,*actions;
  char *pp_form;
  {
   RESTRICTION *rptr,*rtmp;
   register int i;
#if OBJECT_SYSTEM
   register int j;
#endif
   int mai;
   
   SaveBusyCount(gfunc);
   if (meth == NULL)
     {
      mai = (mi != 0) ? FindMethodByIndex(gfunc,mi) : -1;
      if (mai == -1)
        meth = AddGenericMethod(gfunc,mposn,mi);
      else
        {
         DeleteMethodInfo(gfunc,&gfunc->methods[mai]);
         if (mai < mposn)
           {
            mposn--;
            for (i = mai+1 ; i <= mposn ; i++)
              CopyMemory(METHOD,1,&gfunc->methods[i-1],&gfunc->methods[i]);
           }
         else
           {
            for (i = mai-1 ; i >= mposn ; i--)
              CopyMemory(METHOD,1,&gfunc->methods[i+1],&gfunc->methods[i]);
           }
         meth = &gfunc->methods[mposn];
         meth->index = mi;
        }
     }
   else
     {
      ExpressionDeinstall(meth->actions);
      ReturnPackedExpression(meth->actions);
      if (meth->pp_form != NULL)
        rm((VOID *) meth->pp_form,(int) (sizeof(char) * (strlen(meth->pp_form)+1)));
     }
   meth->actions = PackExpression(actions);
   ExpressionInstall(meth->actions);
   meth->pp_form = pp_form;
   if (mposn == -1)
     {
      RestoreBusyCount(gfunc);
      return(meth);
     }
   
   meth->minRestrictions = rcnt;
   meth->maxRestrictions = (wildcard != NULL) ? -1 : rcnt;
   if (rcnt != 0)
     meth->restrictions = (RESTRICTION *) 
                          gm2((int) (sizeof(RESTRICTION) * rcnt));
   else
      meth->restrictions = NULL;
   for (i = 0 ; i < rcnt ; i++)
     {
      rptr = &meth->restrictions[i];
      rtmp = (RESTRICTION *) params->arg_list;
      rptr->query = PackExpression(rtmp->query);
      rptr->tcnt = rtmp->tcnt;
      rptr->types = rtmp->types;

      /* =====================================================
         Make sure the types-array is not deallocated when the
           temporary restriction nodes are
         ===================================================== */
      rtmp->tcnt = 0;
      rtmp->types = NULL;

      ExpressionInstall(rptr->query);
#if OBJECT_SYSTEM
      for (j = 0 ; j < rptr->tcnt ; j++)
        ((CLASS_TYPE *) rptr->types[j])->busy++;
#endif
      params = params->next_arg;
     }
   RestoreBusyCount(gfunc);
   return(meth);
  }
  
/*********************************************************
  NAME         : AddGeneric
  DESCRIPTION  : Inserts a new generic function
                   header into the generic list
  INPUTS       : 1) Symbolic name of the new generic
                 2) Caller's input buffer for flag
                      if added generic is new or not
  RETURNS      : The address of the new node, or
                   address of old node if already present
  SIDE EFFECTS : Generic header inserted
                 If the node is already present, it is
                   moved to the end of the list, otherwise
                   the new node is inserted at the end
  NOTES        : None
 *********************************************************/
globle GENERIC_FUNC *AddGeneric(name,new)
  SYMBOL_HN *name;
  int *new;
  {
   GENERIC_FUNC *gfunc;
   
   gfunc = GenericList;
   while (gfunc != NULL)
     {
      if (gfunc->name == name)
        break;  
      gfunc = gfunc->nxt;
     }
   if (gfunc != NULL)
     {
      *new = FALSE;
      if (gfunc->prv != NULL)
        gfunc->prv->nxt = gfunc->nxt;
      else
        GenericList = gfunc->nxt;
      if (gfunc->nxt != NULL)
        gfunc->nxt->prv = gfunc->prv;
      else
        GenericListBottom = gfunc->prv;
      gfunc->nxt = gfunc->prv = NULL;
     }
   else
     {
      *new = TRUE;
      gfunc = NewGeneric();
      gfunc->name = name;
      gfunc->sysfunc = FindFunction(ValueToString(name));
      IncrementSymbolCount(name);
     }
   if (GenericList == NULL)
     GenericList = gfunc;
   else
     {
      gfunc->prv = GenericListBottom;
      GenericListBottom->nxt = gfunc;
     }
   GenericListBottom = gfunc;
   return(gfunc);
  }
   
/**********************************************************
  NAME         : FindMethodByRestrictions
  DESCRIPTION  : See if a method for the specified
                   generic satsifies the given restrictions
  INPUTS       : 1) Generic function
                 2) Parameter/restriction expression list
                 3) Number of restrictions
                 4) Wildcard symbol (can be NULL)
                 5) Caller's buffer for holding array posn
                      of where to add new generic method
                      (-1 if method already present)
  RETURNS      : The address of the found method, NULL if
                    not found
  SIDE EFFECTS : Sets the caller's buffer to the index of 
                   where to place the new method, -1 if
                   already present
  NOTES        : None
 **********************************************************/
globle METHOD *FindMethodByRestrictions(gfunc,params,rcnt,wildcard,posn)
  GENERIC_FUNC *gfunc;
  EXPRESSION *params;
  SYMBOL_HN *wildcard;
  int rcnt,*posn;
  {
   register int i,pcode,cmp;
   
   pcode = (wildcard == NULL) ? rcnt : -1;
   for (i = 0 ; i < gfunc->mcnt ; i++)
     {
      cmp = RestrictionsCompare(params,rcnt,pcode,&gfunc->methods[i]);
      if (cmp == IDENTICAL)
        {
         *posn = -1;
         return(&gfunc->methods[i]);
        }
      else if (cmp == HIGHER_PRECEDENCE)
        {
         *posn = i;
         return(NULL);
        }
     }
   *posn = i;
   return(NULL);
  }

/************************************************************
  NAME         : ClearDefmethods
  DESCRIPTION  : Deletes all defmethods - generic headers
                   are left intact
  INPUTS       : None
  RETURNS      : TRUE if all methods deleted, FALSE otherwise
  SIDE EFFECTS : Defmethods deleted
  NOTES        : Clearing generic functions is done in
                   two stages
                   
                 1) Delete all methods (to clear any
                    references to other constructs)
                 2) Delete all generic headers
                 
                 This allows other constructs which
                   mutually refer to generic functions
                   to be cleared
 ************************************************************/
globle int ClearDefmethods()
  {
   register GENERIC_FUNC *gfunc;
   register int i;
   int success = TRUE;
   
#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded() == CLIPS_TRUE) return(FALSE);
#endif

   gfunc = GenericList;
   while (gfunc != NULL)
     {
      if (MethodsExecuting(gfunc) == FALSE)
        {
         for (i = 0 ; i < gfunc->mcnt ; i++)
           DeleteMethodInfo(gfunc,&gfunc->methods[i]);
         if (gfunc->mcnt != 0)
           rm((VOID *) gfunc->methods,(int) (sizeof(METHOD) * gfunc->mcnt));
         gfunc->mcnt = 0;
         gfunc->methods = NULL;
        }
      else
        success = FALSE;
      gfunc = gfunc->nxt;
     }
   return(success);
  }
  
/************************************************************
  NAME         : ClearDefgenerics
  DESCRIPTION  : Deletes all generic headers
  INPUTS       : None
  RETURNS      : TRUE if all methods deleted, FALSE otherwise
  SIDE EFFECTS : Generic headers deleted
  NOTES        : None
 ************************************************************/
globle int ClearDefgenerics()
  {
   register GENERIC_FUNC *gfunc,*gtmp;
   int success = TRUE;
   
#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded() == CLIPS_TRUE) return(FALSE);
#endif

   gfunc = GenericList;
   while (gfunc != NULL)
     {
      gtmp = gfunc;
      gfunc = gfunc->nxt;
      if (gtmp->busy > 0)
        {
         PrintCLIPS(WERROR,"Unable to delete generic function ");
         PrintCLIPS(WERROR,ValueToString(gtmp->name));
         PrintCLIPS(WERROR,".\n");
         success = FALSE;
        }
      else
        {
         if (GetCompilationsWatch())
           {
            PrintCLIPS(WDIALOG,"Removing defgeneric ");
            PrintCLIPS(WDIALOG,ValueToString(gtmp->name));
            PrintCLIPS(WDIALOG,".\n");
           }
         if (gtmp->prv == NULL)
           GenericList = gtmp->nxt;
         else
           gtmp->prv->nxt = gtmp->nxt;
         if (gtmp->nxt == NULL)
           GenericListBottom = gtmp->prv;
         else
           gtmp->nxt->prv = gtmp->prv;
         DecrementSymbolCount(gtmp->name);
         if (gtmp->pp_form != NULL)
           rm((VOID *) gtmp->pp_form,(int) (sizeof(char) * (strlen(gtmp->pp_form)+1)));
         rtn_struct(generic_func,gtmp);
        }
     }
   return(success);
  }
  
/********************************************************
  NAME         : MethodAlterError
  DESCRIPTION  : Prints out an error message reflecting
                   that a generic function's methods
                   cannot be altered while any of them
                   are executing
  INPUTS       : The name-string of the generic function
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : None
 ********************************************************/
globle VOID MethodAlterError(gname)
  char *gname;
  {
   PrintCLIPS(WERROR,"Cannot modify methods for generic function ");
   PrintCLIPS(WERROR,gname);
   PrintCLIPS(WERROR," while one is executing.\n");
  }
  
/***************************************************
  NAME         : DeleteMethodInfo
  DESCRIPTION  : Deallocates all the data associated
                  w/ a method but does not release
                  the method structure itself
  INPUTS       : 1) The generic function address
                 2) The method address
  RETURNS      : Nothing useful
  SIDE EFFECTS : Nodes deallocated
  NOTES        : None
 ***************************************************/
globle VOID DeleteMethodInfo(gfunc,meth)
  GENERIC_FUNC *gfunc;
  METHOD *meth;
  {
   register int j;
#if OBJECT_SYSTEM
   register int k;
#endif
   register RESTRICTION *rptr;
   
   if (GetCompilationsWatch())
     {
      PrintCLIPS(WDIALOG,"Removing defgeneric ");
      PrintCLIPS(WDIALOG,ValueToString(gfunc->name));
      PrintCLIPS(WDIALOG," method #");
      PrintLongInteger(WDIALOG,(long) meth->index);
      PrintCLIPS(WDIALOG,".\n");
     }
   SaveBusyCount(gfunc);
   ExpressionDeinstall(meth->actions);
   ReturnPackedExpression(meth->actions);
   if (meth->pp_form != NULL)
     rm((VOID *) meth->pp_form,(int) (sizeof(char) * (strlen(meth->pp_form)+1)));
   for (j = 0 ; j < meth->minRestrictions ; j++)
     {
      rptr = &meth->restrictions[j];
#if OBJECT_SYSTEM
      for (k = 0 ; k < rptr->tcnt ; k++)
        ((CLASS_TYPE *) rptr->types[k])->busy--;
#endif
      if (rptr->types != NULL)
        rm((VOID *) rptr->types,(int) (sizeof(VOID *) * rptr->tcnt));
      ExpressionDeinstall(rptr->query);
      ReturnPackedExpression(rptr->query);
     }
   if (meth->restrictions != NULL)
     rm((VOID *) meth->restrictions,
        (int) (sizeof(RESTRICTION) * meth->minRestrictions));
   RestoreBusyCount(gfunc);
  }
  
/***************************************************
  NAME         : MethodsExecuting
  DESCRIPTION  : Determines if any of the methods of
                   a generic function are currently
                   executing
  INPUTS       : The generic function address
  RETURNS      : TRUE if any methods are executing,
                   FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
globle int MethodsExecuting(gfunc)
  GENERIC_FUNC *gfunc;
  {
   register unsigned i;
   
   for (i = 0 ; i < gfunc->mcnt ; i++)
     if (gfunc->methods[i].busy > 0)
       return(TRUE);
   return(FALSE);
  }
  
#endif

/************************************************
  NAME         : FindDefgenericBySymbol
  DESCRIPTION  : Looks up a generic function
  INPUTS       : The name of the generic
  RETURNS      : The address of the generic, NULL
                   if not found
  SIDE EFFECTS : None
  NOTES        : None
 ************************************************/
globle GENERIC_FUNC *FindDefgenericBySymbol(gname)
  SYMBOL_HN *gname;
  {
   GENERIC_FUNC *gfunc;
   
   gfunc = GenericList;
   while ((gfunc != NULL) ? (gfunc->name != gname) : FALSE)
     gfunc = gfunc->nxt;
   return(gfunc);
  }
    
/*****************************************************
  NAME         : FindMethodByIndex
  DESCRIPTION  : Finds a generic function method of
                   specified index
  INPUTS       : 1) The generic function
                 2) The index
  RETURNS      : The position of the method in the
                   generic function's method array,
                   -1 if not found
  SIDE EFFECTS : None
  NOTES        : None
 *****************************************************/
globle int FindMethodByIndex(gfunc,index)
  GENERIC_FUNC *gfunc;
  unsigned index;
  {
   register unsigned i;
   
   for (i = 0 ; i < gfunc->mcnt ; i++)
     if (gfunc->methods[i].index == index)
       return(i);
   return(-1);
  }
  
/***********************************************************************************
  NAME         : GenericDispatch
  DESCRIPTION  : Forms a "core" framework of all the "applicable"
                  methods from a generic function for the specified
                  arguments and directly calls the first one.
                  Any of these methods could call (shadow-call)
                    which allows the next method in the core to be
                    executed.  Otherwise, only the first method
                    (i.e. the most specific) is executed.
  INPUTS       : 1) The generic function
                 2) The generic function argument expressions
                 3) The caller's result value buffer
  RETURNS      : Nothing useful
  SIDE EFFECTS : Any side-effects of evaluating the generic function arguments
                 Any side-effects of evaluating query functions on method parameter
                   restrictions when determining the core (see warning #1)
                 Any side-effects of actual execution of methods (see warning #2)
                 Caller's buffer set to the result of the generic function call
                 
                 In case of errors, the result is FALSE, otherwise it is the
                   result returned by the most specific method (which can choose
                   to ignore or return the values of more general methods)
  NOTES        : WARNING #1: Query functions on method parameter restrictions
                    should not have side-effects, for they will be evaluated even
                    for methods that don't end up in the final core.
                 WARNING #2: Side-effects of method execution should not always rely
                    on only being executed once per generic function call.  Every
                    time a method calls (shadow-call) the same next-most-specific
                    method is executed.  Thus, it is possible for a method to be
                    executed multiple times per generic function call.
                    
                 See the header on FindApplicableMethods() for details on
                   forming the core of applicable methods.
 ***********************************************************************************/
globle VOID GenericDispatch(gfunc,params,result)
  GENERIC_FUNC *gfunc;
  EXPRESSION *params;
  DATA_OBJECT *result;
  {
   EXPRESSION *oldStackFrame;
   DATA_OBJECT *oldBindList;
   GENERIC_FUNC *previousGeneric;
   METHOD *previousMethod;
   int oldStackSize,itmp,oldce,svReturnContext,svBreakContext;
   METHOD_LINK *oldCore;
   
   result->type = SYMBOL;
   result->value = (VOID *) CLIPSFalseSymbol;
   EvaluationError = FALSE;
   if (HaltExecution)
     return;
   oldce = ExecutingConstruct();
   SetExecutingConstruct(TRUE);
   previousGeneric = CurrentGeneric;
   CurrentGeneric = gfunc;
   itmp = CountArguments(params);
   CurrentEvaluationDepth++;
   oldStackFrame = GenericStackFrame;
   GenericStackFrame = EvaluateGenericParameters(params,itmp,gfunc->name);
   if (EvaluationError) 
     {
      GenericStackFrame = oldStackFrame;
      CurrentGeneric = previousGeneric;
      CurrentEvaluationDepth--;
      PeriodicCleanup(FALSE,TRUE);
      SetExecutingConstruct(oldce);
      return;
     }
   oldStackSize = GenericStackSize;
   GenericStackSize = itmp;                         
   previousMethod = CurrentMethod;
   oldBindList = BindList;
   BindList = NULL;
   
   oldCore = TopOfCore;
   gfunc->busy++;
   TopOfCore = FindApplicableMethods(gfunc);
   svBreakContext = BreakContext;
   BreakContext = CLIPS_FALSE;
   svReturnContext = ReturnContext;
   ReturnContext = CLIPS_TRUE;
   if (TopOfCore != NULL)
     {
      METHOD_LINK *oldNext;
      
      oldNext = NextInCore;
      NextInCore = TopOfCore->nxt;
      CurrentMethod = TopOfCore->method;
#if DEBUGGING_FUNCTIONS
      if (WatchGenerics)
        TraceGeneric(BEGIN_TRACE);
      if (WatchMethods)
        {
         TraceMethod(BEGIN_TRACE,SYSTEM_NO);
         EvaluateExpression(CurrentMethod->actions,result);
         TraceMethod(END_TRACE,SYSTEM_NO);
        }
      else
        EvaluateExpression(CurrentMethod->actions,result);
      if (WatchGenerics)
        TraceGeneric(END_TRACE);
#else
      EvaluateExpression(CurrentMethod->actions,result);
#endif
      DestroyMethodLinks(TopOfCore);
      NextInCore = oldNext;
     }
   else if (! EvaluationError)
     {
      if (gfunc->sysfunc != NULL)
        {
         EXPRESSION fcall;
      
         fcall.type = FCALL;
         fcall.value = (VOID *) gfunc->sysfunc;
         fcall.next_arg = NULL;
         fcall.arg_list = GenericStackFrame;
#if DEBUGGING_FUNCTIONS
         if (WatchGenerics)
           TraceGeneric(BEGIN_TRACE);
         if (WatchMethods)
           {
            TraceMethod(BEGIN_TRACE,SYSTEM_YES);
            EvaluateExpression(&fcall,result);
            TraceMethod(END_TRACE,SYSTEM_YES);
           }
         else
           EvaluateExpression(&fcall,result);
         if (WatchGenerics)
           TraceGeneric(END_TRACE);
#else
         EvaluateExpression(&fcall,result);
#endif
        }
      else
        {
         PrintCLIPS(WERROR,"No applicable methods or system function for ");
         PrintCLIPS(WERROR,ValueToString(gfunc->name));
         PrintCLIPS(WERROR,".\n");
         SetEvaluationError(TRUE);
        }
     }
   TopOfCore = oldCore;
   gfunc->busy--;
   ReturnContext = svReturnContext;
   BreakContext = svBreakContext;
   ReturnFlag = CLIPS_FALSE;
   
   if (GenericStackFrame != NULL)
     {
      for (itmp = 0 ; itmp < GenericStackSize ; itmp++)
        if (GenericStackFrame[itmp].type == MULTIFIELD)
          rtn_struct(dataObject,GenericStackFrame[itmp].value);
      rm((VOID *) GenericStackFrame,(int) (sizeof(EXPRESSION) * GenericStackSize));
     }
   GenericStackFrame = oldStackFrame;
   GenericStackSize = oldStackSize;
   if (BindList != NULL)
     ReturnValues(BindList);
   BindList = oldBindList;
   CurrentGeneric = previousGeneric;
   CurrentMethod = previousMethod;
   CurrentEvaluationDepth--;
   PropogateReturnValue(result);
   PeriodicCleanup(FALSE,TRUE);
   SetExecutingConstruct(oldce);
  }
  
#if DEBUGGING_FUNCTIONS

/*************************************************************
  NAME         : PreviewGeneric
  DESCRIPTION  : Allows the user to see a printout of all the
                   applicable methods for a particular generic
                   function call
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Any side-effects of evaluating the generic
                   function arguments
                 and evaluating query-functions to determine
                   the set of applicable methods
  NOTES        : CLIPS Syntax: (preview-generic <func> <args>)
 *************************************************************/
globle VOID PreviewGeneric()
  {
   GENERIC_FUNC *gfunc;
   EXPRESSION *oldStackFrame;
   DATA_OBJECT *oldBindList;
   GENERIC_FUNC *previousGeneric;
   METHOD *previousMethod;
   int oldStackSize,itmp,oldce;
   METHOD_LINK *oldCore;
   DATA_OBJECT temp;
   
   EvaluationError = FALSE;
   if (ArgTypeCheck("preview-generic",1,SYMBOL,&temp) == FALSE)
     return;
   gfunc = (GENERIC_FUNC *) FindDefgenericBySymbol((SYMBOL_HN *) temp.value);
   if (gfunc == NULL)
     {
      PrintCLIPS(WERROR,"Unable to find generic function ");
      PrintCLIPS(WERROR,ValueToString(temp.value));
      PrintCLIPS(WERROR," in function preview-generic.\n");
      return;
     }
   oldce = ExecutingConstruct();
   SetExecutingConstruct(TRUE);
   previousGeneric = CurrentGeneric;
   CurrentGeneric = gfunc;
   itmp = CountArguments(GetFirstArgument()->next_arg);
   CurrentEvaluationDepth++;
   oldStackFrame = GenericStackFrame;
   GenericStackFrame = EvaluateGenericParameters(GetFirstArgument()->next_arg,
                                                 itmp,gfunc->name);
   if (EvaluationError) 
     {
      GenericStackFrame = oldStackFrame;
      CurrentGeneric = previousGeneric;
      CurrentEvaluationDepth--;
      SetExecutingConstruct(oldce);
      return;
     }
   oldStackSize = GenericStackSize;
   GenericStackSize = itmp;                         
   previousMethod = CurrentMethod;
   oldBindList = BindList;
   BindList = NULL;
   
   oldCore = TopOfCore;
   gfunc->busy++;
   TopOfCore = FindApplicableMethods(gfunc);
   gfunc->busy--;
   if (! EvaluationError)
     {
      if ((TopOfCore == NULL) && (gfunc->sysfunc == NULL))
        {
         PrintCLIPS(WERROR,"No applicable methods or system function for ");
         PrintCLIPS(WERROR,ValueToString(gfunc->name));
         PrintCLIPS(WERROR,".\n");
        }
      else
        DisplayGenericCore(gfunc,TopOfCore);
     DestroyMethodLinks(TopOfCore);
    }
   TopOfCore = oldCore;
   if (GenericStackFrame != NULL)
     {
      for (itmp = 0 ; itmp < GenericStackSize ; itmp++)
        if (GenericStackFrame[itmp].type == MULTIFIELD)
          rtn_struct(dataObject,GenericStackFrame[itmp].value);
      rm((VOID *) GenericStackFrame,(int) (sizeof(EXPRESSION) * GenericStackSize));
     }
   GenericStackFrame = oldStackFrame;
   GenericStackSize = oldStackSize;
   if (BindList != NULL)
     ReturnValues(BindList);
   BindList = oldBindList;
   CurrentGeneric = previousGeneric;
   CurrentMethod = previousMethod;
   CurrentEvaluationDepth--;
   SetExecutingConstruct(oldce);
  }

/******************************************************************
  NAME         : PrintMethod
  DESCRIPTION  : Lists a brief description of methods for a method
  INPUTS       : 1) Buffer for method info
                 2) Size of buffer (not including space for '\0')
                 3) The method address
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : A terminating newline is NOT included
 ******************************************************************/
globle VOID PrintMethod(buf,buflen,meth)
  char *buf;
  int buflen;
  METHOD *meth;
  {
   register int j,k;
   register RESTRICTION *rptr;
   char numbuf[15];
   
   sprintf(numbuf,"%-2d ",meth->index);
   strncpy(buf,numbuf,buflen);
   for (j = 0 ; j < meth->minRestrictions ; j++)
     {
      rptr = &meth->restrictions[j];
      strncat(buf,"(",buflen-strlen(buf));
#if OBJECT_SYSTEM
      for (k = 0 ; k < rptr->tcnt ; k++)
        {
         strncat(buf,ValueToString(((CLASS_TYPE *) rptr->types[k])->name),
                 buflen-strlen(buf));
         if (k < (rptr->tcnt - 1))
           strncat(buf," ",buflen-strlen(buf));
        }
#else
      for (k = 0 ; k < rptr->tcnt ; k++)
        {
         if ((int) rptr->types[k] == INTEGER)
           strncat(buf,INTEGER_TYPE_NAME,buflen-strlen(buf));
         else if ((int) rptr->types[k] == FLOAT)
           strncat(buf,FLOAT_TYPE_NAME,buflen-strlen(buf));
         else if ((int) rptr->types[k] == SYMBOL)
           strncat(buf,SYMBOL_TYPE_NAME,buflen-strlen(buf));
         else if ((int) rptr->types[k] == STRING)
           strncat(buf,STRING_TYPE_NAME,buflen-strlen(buf));
         else if ((int) rptr->types[k] == MULTIFIELD)
           strncat(buf,MULTIFIELD_TYPE_NAME,buflen-strlen(buf));
         else if ((int) rptr->types[k] == EXTERNAL_ADDRESS)
           strncat(buf,ADDRESS_TYPE_NAME,buflen-strlen(buf));
         if (k < (rptr->tcnt - 1))
           strncat(buf," ",buflen-strlen(buf));
        }
#endif
      if (rptr->query != NULL)
        {
         if (rptr->tcnt != 0)
           strncat(buf," ",buflen-strlen(buf));
         strncat(buf,"<qry>",buflen-strlen(buf));
        }
      strncat(buf,")",buflen-strlen(buf));
      if (j != (meth->minRestrictions-1))
        strncat(buf," ",buflen-strlen(buf));
     }
   if (meth->maxRestrictions == -1)
     {
      if (meth->minRestrictions != 0)
        strncat(buf," ",buflen-strlen(buf));
      strncat(buf,"$?",buflen-strlen(buf));
     }
  }
    
#endif

/***************************************************
  NAME         : NextMethodP
  DESCRIPTION  : Determines if a shadowed generic
                   function method is available for
                   execution
  INPUTS       : None
  RETURNS      : TRUE if there is a method available,
                   FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : CLIPS Syntax: (next-methodp)
 ***************************************************/
globle int NextMethodP()
  {
   if (CurrentGeneric == NULL)
     return(FALSE);
   if (NextInCore != NULL)
     return(TRUE);
   if (CurrentGeneric->sysfunc != NULL)
     return(TRUE);
   return(FALSE);
  }
  
/****************************************************
  NAME         : CallNextMethod
  DESCRIPTION  : Executes the next available method
                   in the core for a generic function
  INPUTS       : Caller's buffer for the result
  RETURNS      : Nothing useful
  SIDE EFFECTS : Side effects of execution of shadow
                 EvaluationError set if no method
                   is available to execute.
  NOTES        : CLIPS Syntax: (call-next-method)
 ****************************************************/
globle VOID CallNextMethod(result)
  DATA_OBJECT *result;
  {
   result->type = SYMBOL;
   result->value = (VOID *) CLIPSFalseSymbol;
   if (HaltExecution)
     return;
   if (! NextMethodP())
     {
      PrintCLIPS(WERROR,"Shadowed methods not applicable in current context.\n");
      SetEvaluationError(TRUE);
      return;
     }
     
   if (NextInCore != NULL)
     {
      METHOD_LINK *oldNext;
      METHOD *oldMethod;
   
      oldNext = NextInCore;
      oldMethod = CurrentMethod;
      CurrentMethod = NextInCore->method;
      NextInCore = NextInCore->nxt;
#if DEBUGGING_FUNCTIONS
      if (WatchMethods)
        {
         TraceMethod(BEGIN_TRACE,SYSTEM_NO);
         EvaluateExpression(CurrentMethod->actions,result);
         TraceMethod(END_TRACE,SYSTEM_NO);
        }
      else
#endif
        EvaluateExpression(CurrentMethod->actions,result);
      NextInCore = oldNext;
      CurrentMethod = oldMethod;
     }
   else
     {
      EXPRESSION fcall;
      
      fcall.type = FCALL;
      fcall.value = (VOID *) CurrentGeneric->sysfunc;
      fcall.next_arg = NULL;
      fcall.arg_list = GenericStackFrame;
#if DEBUGGING_FUNCTIONS
      if (WatchMethods)
        {
         TraceMethod(BEGIN_TRACE,SYSTEM_YES);
         EvaluateExpression(&fcall,result);
         TraceMethod(END_TRACE,SYSTEM_YES);
        }
      else
#endif
        EvaluateExpression(&fcall,result);
     }
   ReturnFlag = CLIPS_FALSE;
  }
  
/************************************************************
  NAME         : SetGenericList
  DESCRIPTION  : Resets global generic headers list pointers
  INPUTS       : New address of top of list
  RETURNS      : Nothing useful
  SIDE EFFECTS : Global pointers changed
  NOTES        : WARNING!! : Does not check to see if current
                   pointers are != NULL
 ************************************************************/
globle VOID SetGenericList(top)
  GENERIC_FUNC *top;
  {
   GenericList = top;
   if (top == NULL)
     GenericListBottom = NULL;
   else
     {
      while (top->nxt != NULL)
        top = top->nxt;
      GenericListBottom = top;
     }
  }

/* =========================================
   *****************************************
          INTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
   
#if (! BLOAD_ONLY) && (! RUN_TIME)

/***************************************************
  NAME         : NewGeneric
  DESCRIPTION  : Allocates and initializes a new
                   generic function header
  INPUTS       : None
  RETURNS      : The address of the new generic
  SIDE EFFECTS : Generic function  header created
  NOTES        : None
 ***************************************************/
static GENERIC_FUNC *NewGeneric()
  {
   GENERIC_FUNC *ngen;
   
   ngen = get_struct(generic_func);
   ngen->name = NULL;
   ngen->busy = 0;
   ngen->new_index = 1;
   ngen->methods = NULL;
   ngen->sysfunc = NULL;
   ngen->mcnt = 0;
   ngen->pp_form = NULL;
   ngen->prv = ngen->nxt = NULL;
   return(ngen);
  }

/**********************************************************************
  NAME         : AddGenericMethod
  DESCRIPTION  : Inserts a blank method (with the method-index set)
                   into the specified position of the generic
                   method array
  INPUTS       : 1) The generic function
                 2) The index where to add the method in the array
                 3) The method user-index (0 if don't care)
  RETURNS      : The address of the new method
  SIDE EFFECTS : Fields initialized (index set) and new method inserted
                 Generic function new method-index set to specified
                   by user-index if > current new method-index
  NOTES        : None
 **********************************************************************/
static METHOD *AddGenericMethod(gfunc,mposn,mi)
  GENERIC_FUNC *gfunc;
  int mposn;
  unsigned mi;
  {
   METHOD *narr;
   register int b,e;
   
   narr = (METHOD *) gm2((int) (sizeof(METHOD) * (gfunc->mcnt+1)));
   for (b = e = 0 ; b < gfunc->mcnt ; b++ , e++)
     {
      if (b == mposn)
        e++;
      CopyMemory(METHOD,1,&narr[e],&gfunc->methods[b]);
     }
   if (mi == 0)
     narr[mposn].index = gfunc->new_index++;
   else
     {
      narr[mposn].index = mi;
      if (mi >= gfunc->new_index)
        gfunc->new_index = mi+1;
     }
   narr[mposn].busy = 0;
   narr[mposn].minRestrictions = 0;
   narr[mposn].maxRestrictions = 0;
   narr[mposn].restrictions = NULL;
   narr[mposn].actions = NULL;
   narr[mposn].pp_form = NULL;
   if (gfunc->mcnt != 0)
     rm((VOID *) gfunc->methods,(int) (sizeof(METHOD) * gfunc->mcnt));
   gfunc->mcnt++;
   gfunc->methods = narr;
   return(&narr[mposn]);
  }
  
/****************************************************************
  NAME         : RestrictionsCompare
  DESCRIPTION  : Compares the restriction-expression list
                   with an existing methods restrictions to
                   determine an ordering
  INPUTS       : 1) The parameter/restriction expression list
                 2) The number of minimum restrictions
                 3) The number of maximum restrictions (-1
                    if unlimited)
                 3) The method with which to compare restrictions
  RETURNS      : A code representing how the method restrictions
                   -1 : New restrictions have higher precedence
                    0 : New restrictions are identical
                    1 : New restrictions have lower precedence
  SIDE EFFECTS : None
  NOTES        : The new restrictions are stored in the arg_list
                   pointers of the parameter expressions
 ****************************************************************/
static int RestrictionsCompare(params,min,max,meth)
  EXPRESSION *params;
  int min,max;
  METHOD *meth;
  {
   register int i;
   register RESTRICTION *r1,*r2;
   int diff = FALSE,rtn;

   for (i = 0 ; (i < min) && (i < meth->minRestrictions) ; i++)
     {
      r1 = (RESTRICTION *) params->arg_list;
      r2 = &meth->restrictions[i];
      rtn = TypeListCompare(r1,r2);
      if (rtn != IDENTICAL)
        return(rtn);
      if ((r1->query == NULL) && (r2->query != NULL))
        return(LOWER_PRECEDENCE);
      if ((r1->query != NULL) && (r2->query == NULL))
        return(HIGHER_PRECEDENCE);
      if (IdenticalExpression(r1->query,r2->query) == FALSE)
        diff = TRUE;
      params = params->next_arg;
     }
   if ((min == meth->minRestrictions) && (max == meth->maxRestrictions))
     {
      if (diff)
        return(LOWER_PRECEDENCE);
      return(IDENTICAL);
     } 
   if (i < min)
     return(HIGHER_PRECEDENCE);
   if (max == -1)
     return(LOWER_PRECEDENCE);
   if (meth->maxRestrictions == -1)
     return(HIGHER_PRECEDENCE);
   return(LOWER_PRECEDENCE);
  }

/*****************************************************
  NAME         : TypeListCompare
  DESCRIPTION  : Determines the precedence between
                   the class lists on two restrictions
  INPUTS       : 1) Restriction address #1
                 2) Restriction address #2
  RETURNS      : -1 : r1 precedes r2
                  0 : Identical classes
                  1 : r2 precedes r1
  SIDE EFFECTS : None
  NOTES        : None
 *****************************************************/
static int TypeListCompare(r1,r2)
  RESTRICTION *r1,*r2;
  {
   register int i,diff = FALSE;
   
   if ((r1->tcnt == 0) && (r2->tcnt == 0))
     return(IDENTICAL);
   if (r1->tcnt == 0)
     return(LOWER_PRECEDENCE);
   if (r2->tcnt == 0)
     return(HIGHER_PRECEDENCE);
   for (i = 0 ; (i < r1->tcnt) && (i < r2->tcnt) ; i++)
     {
      if (r1->types[i] != r2->types[i])
        {
         diff = TRUE;
#if OBJECT_SYSTEM
         if (HasSuperclass((CLASS_TYPE *) r1->types[i],(CLASS_TYPE *) r2->types[i]))
           return(HIGHER_PRECEDENCE);
         if (HasSuperclass((CLASS_TYPE *) r2->types[i],(CLASS_TYPE *) r1->types[i]))
           return(LOWER_PRECEDENCE);
#endif         
        }
     }
   if (r1->tcnt < r2->tcnt)
     return(HIGHER_PRECEDENCE);
   if (r1->tcnt > r2->tcnt)
     return(LOWER_PRECEDENCE);
   if (diff)
     return(LOWER_PRECEDENCE);
   return(IDENTICAL);
  }
  
#endif

/***************************************************************
  NAME         : EvaluateGenericParameters
  DESCRIPTION  : Given a list of parameter expressions,
                   this function evaluates each expression
                   and stores the results in a contiguous
                   array of EXPRESSIONS.  Used in creating a new
                   GenericStackFrame for the execution of a
                   generic function
  INPUTS       : 1) The paramter expression list
                 2) The number of paramters in the list
                 3) The name of the generic function for which
                    these parameters are being evaluated
  RETURNS      : The address of the contiguous array,
                   NULL on errors
  SIDE EFFECTS : Any side-effects of the evaluation of the
                   parameter expressions
                 EXPRESSION array allocated (deallocated on errors)
  NOTES        : Stored in EXPRESSIONS instead of VALUES so 
                    system functions can also be easily called
                    with these arguments
 ***************************************************************/
static EXPRESSION *EvaluateGenericParameters(parameterList,numberOfParameters,gName)
  EXPRESSION *parameterList;
  int numberOfParameters;
  SYMBOL_HN *gName;
  {
   DATA_OBJECT temp;
   EXPRESSION *rva;
   int i = 0;
   
   if (numberOfParameters == 0) return(NULL);
   
   rva = (EXPRESSION *) gm2((int) (sizeof(EXPRESSION) * numberOfParameters));
   
   while (parameterList != NULL)
     {
      if ((EvaluateExpression(parameterList,&temp) == TRUE) ? TRUE :
          (temp.type == RVOID))
        {
         if (temp.type == RVOID)
           {
            PrintCLIPS(WERROR,"Functions without a return value are illegal as generic function arguments.\n");
            SetEvaluationError(TRUE);
           }
         PrintCLIPS(WERROR,"This error occurred while evaluating arguments ");
         PrintCLIPS(WERROR,"for the generic function ");
         PrintCLIPS(WERROR,ValueToString(gName));
         PrintCLIPS(WERROR,".\n");
         while (i > 1)
           {
            if (rva[i-1].type == MULTIFIELD)
              rtn_struct(dataObject,rva[i-1].value);
            i--;
           }
         rm((VOID *) rva,(int) (sizeof(EXPRESSION) * numberOfParameters));
         return(NULL);
        }
      rva[i].type = temp.type;
      if (temp.type != MULTIFIELD)
        rva[i].value = temp.value;
      else
        {
         rva[i].value = (VOID *) get_struct(dataObject);
         CopyMemory(DATA_OBJECT,1,rva[i].value,&temp);
        }
      parameterList = parameterList->next_arg;
      rva[i].arg_list = NULL;
      if (parameterList != NULL)
        rva[i].next_arg = &rva[i+1];
      else
        rva[i].next_arg = NULL;
      i++;
     }
   return(rva);
  }

/**************************************************************************
  NAME         : FindApplicableMethods
  DESCRIPTION  : Generates a core frame of applicable methods from
                   a generic function for a particular set of arguments.
                   
                 A method is applicable if all of its restrictions
                   are satisfied by the generic function arguments.
                   
                 When two or more methods are applicable, precedence must
                   be invoked to determine an ordering.  Here are the
                   rules:
                   
                 1) The precedence determined by the left-most restrictions
                    of the two methods has priority
                    
                 Considering the restrictions on the same argument from
                   the two different methods:
                   
                 2) The most specific class restriction has highest
                    precedence
                 3) For argument restrictions that have the
                      same class restriction (or lack thereof),
                      query-restrictions take precence
                      
  INPUTS       : The generic function
  RETURNS      : The top of the core, NULL on errors
  SIDE EFFECTS : Core allocated
                 EvaluationError set on errors
  NOTES        : Uses the globals GenericStackFrame and GenericStackSize
                 Assumes EvaluationError is FALSE when called
 **************************************************************************/
static METHOD_LINK *FindApplicableMethods(gfunc)
  GENERIC_FUNC *gfunc;
  {
   register int i;
   METHOD_LINK *gtop = NULL,*gbot = NULL,*tmp;
   
   /* ==========================================================================
      Form the core by testing all methods which belong to this generic function
      The list of methods is already in ascending order of precedence
      ========================================================================== */
   for (i = 0 ; i < gfunc->mcnt ; i++)
     {
      if (IsMethodApplicable(&gfunc->methods[i]) == TRUE)
        {
         tmp = get_struct(method_link);
         tmp->method = &gfunc->methods[i];
         tmp->method->busy++;
         tmp->nxt = NULL;
         if (gtop == NULL)
           gtop = tmp;
         else
           gbot->nxt = tmp;
         gbot = tmp;
        }
      if (EvaluationError)
        {
         DestroyMethodLinks(gtop);
         return(NULL);
        }
     }
   return(gtop);
  }
  
/*********************************************************************
  NAME         : DestroyMethodLinks
  DESCRIPTION  : Deallocates core frame for a generic function call
                   and decrements the counts on the applicable methods
  INPUTS       : The method-links of the core
  RETURNS      : Nothing useful
  SIDE EFFECTS : Counts decremented and links deallocated
  NOTES        : None
 *********************************************************************/
static VOID DestroyMethodLinks(mhead)
  METHOD_LINK *mhead;
  {
   METHOD_LINK *tmp;
   
   while (mhead != NULL)
     {
      tmp = mhead;
      tmp->method->busy--;
      rtn_struct(method_link,tmp);
      mhead = mhead->nxt;    
     } 
  }
 
#if OBJECT_SYSTEM

/***************************************************
  NAME         : DetermineRestrictionClass
  DESCRIPTION  : Finds the class of an argument in
                   the GenericStackFrame
  INPUTS       : The argument expression
  RETURNS      : The class address, NULL if error
  SIDE EFFECTS : EvaluationError set on errors
  NOTES        : None
 ***************************************************/
static CLASS_TYPE *DetermineRestrictionClass(exp)
  EXPRESSION *exp;
  {
   INSTANCE_TYPE *ins;
   CLASS_TYPE *cls;
   
   if (exp->type == INSTANCE_NAME)
     {
      ins = FindInstanceBySymbol((SYMBOL_HN *) exp->value);
      cls = (ins != NULL) ? ins->cls : NULL;
     }
   else if (exp->type == INSTANCE)
     {
      ins = (INSTANCE_TYPE *) exp->value;
      cls = (ins->garbage == 0) ? ins->cls : NULL;
     }
   else
     return(PrimitiveClassMap[exp->type]);
   if (cls == NULL)
     {
      SetEvaluationError(TRUE);
      PrintCLIPS(WERROR,"Unable to determine class of ");
      if (exp->type != MULTIFIELD)
        PrintAtom(WERROR,exp->type,exp->value);
      else
        PrintDataObject(WERROR,(DATA_OBJECT *) exp->value);
      PrintCLIPS(WERROR," in generic function ");
      PrintCLIPS(WERROR,ValueToString(CurrentGeneric->name));
      PrintCLIPS(WERROR,".\n");
     }
   return(cls);
  }
  
#endif
 
/***********************************************************************
  NAME         : IsMethodApplicable
  DESCRIPTION  : Tests to see if a method satsifies the arguments of a
                   generic function
                 A method is applicable if all its restrictions are
                   satisfied by the corresponding arguments
  INPUTS       : The method address
  RETURNS      : TRUE if method is applicable, FALSE otherwise
  SIDE EFFECTS : Any query functions are evaluated
  NOTES        : Uses globals GenericStackSize and GenericStackFrame
 ***********************************************************************/
static int IsMethodApplicable(meth)
  METHOD *meth;
  {
   DATA_OBJECT temp;
   register int i,j;
   register RESTRICTION *rp;
   VOID *type;
   
   if ((GenericStackSize < meth->minRestrictions) ||
       ((GenericStackSize > meth->minRestrictions) && (meth->maxRestrictions != -1)))
     return(FALSE);
   for (i = 0 ; i < GenericStackSize ; i++)
     {
      if (i > meth->minRestrictions-1)
        return(TRUE);
      rp = &meth->restrictions[i];
      if (rp->tcnt != 0)
        {
#if OBJECT_SYSTEM
         type = (VOID *) DetermineRestrictionClass(&GenericStackFrame[i]);
         if (type == NULL)
           return(FALSE);
         for (j = 0 ; j < rp->tcnt ; j++)
           {
            if (type == rp->types[j])
              break;
            if (HasSuperclass((CLASS_TYPE *) type,(CLASS_TYPE *) rp->types[j]))
              break;
           }
#else
         type = (VOID *) GenericStackFrame[i].type;
         for (j = 0 ; j < rp->tcnt ; j++)
           {
            if (type == rp->types[j])
              break;
           }
#endif
         if (j == rp->tcnt)
           return(FALSE);
        }
      if (rp->query != NULL)
        {
         EvaluateExpression(rp->query,&temp);
         if ((temp.type != SYMBOL) ? FALSE :
             (temp.value == (VOID *) CLIPSFalseSymbol))
           return(FALSE);
        }
     }
   return(TRUE);
  }
  
#if DEBUGGING_FUNCTIONS

/*********************************************************
  NAME         : DisplayGenericCore
  DESCRIPTION  : Prints out a description of a core
                   frame of applicable methods for
                   a particular call of a generic function
  INPUTS       : 1) The generic function
                 2) The core of appplicable methods
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : None
 *********************************************************/
static VOID DisplayGenericCore(gfunc,core)
  GENERIC_FUNC *gfunc;
  METHOD_LINK *core;
  {
   char buf[61];
   
   while (core != NULL)
     {
      PrintCLIPS(WDISPLAY,ValueToString(gfunc->name));
      PrintCLIPS(WDISPLAY," #");
      PrintMethod(buf,60,core->method);
      PrintCLIPS(WDISPLAY,buf);
      PrintCLIPS(WDISPLAY,"\n");
      core = core->nxt;
     }
   if (gfunc->sysfunc != NULL)
     {
      PrintCLIPS(WDISPLAY,ValueToString(gfunc->name));
      PrintCLIPS(WDISPLAY,":SYSTEM ($?)\n");
     }
  }
  
/**********************************************************************
  NAME         : TraceGeneric
  DESCRIPTION  : Prints out a trace of the beginning or end
                   of the execution of a generic function
  INPUTS       : A string to indicate beginning or end of execution
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : Uses the globals CurrentGeneric, GenericStackSize and 
                   GenericStackFrame for other trace info
 **********************************************************************/
static VOID TraceGeneric(tstring)
  char *tstring;
  {
   register int i;
   register EXPRESSION *exp;
   
   PrintCLIPS(WTRACE,"GNC ");
   PrintCLIPS(WTRACE,tstring);
   PrintCLIPS(WTRACE," ");
   PrintCLIPS(WTRACE,ValueToString(CurrentGeneric->name));
   PrintCLIPS(WTRACE," ");
   PrintCLIPS(WTRACE," ED:");
   PrintLongInteger(WTRACE,(long) CurrentEvaluationDepth);
   PrintCLIPS(WTRACE," (");
   for (i = 0 ; i < GenericStackSize ; i++)
     {
      exp = &GenericStackFrame[i];
      if (exp->type != MULTIFIELD)
        PrintAtom(WTRACE,exp->type,exp->value);
      else
        PrintDataObject(WTRACE,(DATA_OBJECT *) exp->value);
      if (i != GenericStackSize-1)
        PrintCLIPS(WTRACE," ");
     }
   PrintCLIPS(WTRACE,")\n");
  }
 
/**********************************************************************
  NAME         : TraceMethod
  DESCRIPTION  : Prints out a trace of the beginning or end
                   of the execution of a generic function
                   method or system function
  INPUTS       : 1) A string to indicate beginning or end of execution
                 2) A flag indicating whether the trace is for a system
                     function or not
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : Uses the globals CurrentGeneric, CurrentMethod,
                   GenericStackSize and GenericStackFrame for 
                   other trace info
 **********************************************************************/
static VOID TraceMethod(tstring,system)
  char *tstring;
  int system;
  {
   register int i;
   register EXPRESSION *exp;
   
   PrintCLIPS(WTRACE,"MTH ");
   PrintCLIPS(WTRACE,tstring);
   PrintCLIPS(WTRACE," ");
   if (system == SYSTEM_NO)
     {
      PrintCLIPS(WTRACE,ValueToString(CurrentGeneric->name));
      PrintCLIPS(WTRACE,":#");
      PrintLongInteger(WTRACE,(long) CurrentMethod->index);
     }
   else
     {
      PrintCLIPS(WTRACE,ValueToString(CurrentGeneric->name));
      PrintCLIPS(WTRACE,":SYSTEM");
     }
   PrintCLIPS(WTRACE," ");
   PrintCLIPS(WTRACE," ED:");
   PrintLongInteger(WTRACE,(long) CurrentEvaluationDepth);
   PrintCLIPS(WTRACE," (");
   for (i = 0 ; i < GenericStackSize ; i++)
     {
      exp = &GenericStackFrame[i];
      if (exp->type != MULTIFIELD)
        PrintAtom(WTRACE,exp->type,exp->value);
      else
        PrintDataObject(WTRACE,(DATA_OBJECT *) exp->value);
      if (i != GenericStackSize-1)
        PrintCLIPS(WTRACE," ");
     }
   PrintCLIPS(WTRACE,")\n");
  }
  
#endif

#endif

/***************************************************
  NAME         : 
  DESCRIPTION  : 
  INPUTS       : 
  RETURNS      : 
  SIDE EFFECTS : 
  NOTES        : 
 ***************************************************/
