   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*               CLIPS Version 5.10  07/17/91          */
   /*                                                     */
   /*                  OBJECT MESSAGE FUNCTIONS           */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
/* =========================================
   *****************************************
               EXTERNAL DEFINITIONS
   =========================================
   ***************************************** */
#include "setup.h"

#if OBJECT_SYSTEM

#include "classfun.h"
#include "clipsmem.h"
#include "constant.h"
#include "constrct.h"
#include "evaluatn.h"
#include "generate.h"
#include "insfun.h"
#include "msgcom.h"
#include "router.h"
#include "sysprime.h"
#include "utility.h"

#define _MSGFUN_SOURCE_
#include "msgfun.h"

/* =========================================
   *****************************************
                   CONSTANTS
   =========================================
   ***************************************** */
#define BEGIN_TRACE ">>"
#define END_TRACE   "<<"

#define PERFORM     1
#define PREVIEW     0

/* =========================================
   *****************************************
      INTERNALLY VISIBLE FUNCTION HEADERS
   =========================================
   ***************************************** */
#if ANSI_COMPILER
static VOID PerformMessage(DATA_OBJECT *,EXPRESSION *,SYMBOL_HN *);
static DATA_OBJECT *EvaluateMessageParameters(EXPRESSION *,int);

#if DEBUGGING_FUNCTIONS
static HANDLER_LINK *DisplayPrimaryCore(HANDLER_LINK *,int);
static VOID PrintPreviewHandler(HANDLER_LINK *,int,char *);
static VOID TraceMessage(char *,char *);
static VOID TraceHandler(char *,HANDLER_LINK *,char *);
#endif

static HANDLER_LINK *FindApplicableHandlers(CLASS_TYPE *,SYMBOL_HN *);
static int FindApplicableOfName(CLASS_TYPE *,HANDLER_LINK *[],
                                HANDLER_LINK *[],SYMBOL_HN *);
static HANDLER_LINK *JoinHandlerLinks(HANDLER_LINK *[],HANDLER_LINK *[],SYMBOL_HN *);
static VOID PrintNoHandlerError(char *);
static VOID CallHandlers(DATA_OBJECT *);
static VOID PerformImplicitHandler(DATA_OBJECT *);
static int CheckHandlerArgCount(void);
#else
static VOID PerformMessage();
static DATA_OBJECT *EvaluateMessageParameters();

#if DEBUGGING_FUNCTIONS
static HANDLER_LINK *DisplayPrimaryCore();
static VOID PrintPreviewHandler();
static VOID TraceMessage();
static VOID TraceHandler();
#endif

static HANDLER_LINK *FindApplicableHandlers();
static int FindApplicableOfName();
static HANDLER_LINK *JoinHandlerLinks();
static VOID PrintNoHandlerError();
static VOID CallHandlers();
static VOID PerformImplicitHandler();
static int CheckHandlerArgCount();
#endif
      
/* =========================================
   *****************************************
      EXTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */
globle char *hndquals[] = {"around","before","primary","after"};

#if DEBUGGING_FUNCTIONS
globle int WatchHandlers = OFF;
globle int WatchMessages = OFF;
#endif

globle DATA_OBJECT *CurrentMessageFrame = NULL;
globle int CurrentMessageSize = -1;
globle SYMBOL_HN *CurrentMessageName = NULL;

/* =========================================
   *****************************************
      INTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */
static HANDLER_LINK *TopOfCore = NULL, *CurrentCore = NULL, *NextInCore = NULL;

/* =========================================
   *****************************************
          EXTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
/*****************************************************
  NAME         : DirectMessage
  DESCRIPTION  : Plugs in given instance and 
                  performs specified message
  INPUTS       : 1) First part of message name
                    (string - can be NULL)
                 2) Second part of message name
                    (Symbol)
                 3) The instance address
                 4) Address of DATA_OBJECT buffer
                    (NULL if don't care)
                 5) Message argument expressions
  RETURNS      : Nothing useful
  SIDE EFFECTS : Side effects of message execution
  NOTES        : None
 *****************************************************/
globle VOID DirectMessage(ms1,ms2,ins,resultbuf,remargs)
  char *ms1;
  SYMBOL_HN *ms2;
  INSTANCE_TYPE *ins;
  DATA_OBJECT *resultbuf;
  EXPRESSION *remargs;
  {
   int bufsz = 0;
   char *buf = NULL;
   EXPRESSION args;
   DATA_OBJECT temp;
   SYMBOL_HN *msg;
      
   if (ms1 != NULL)
     {
      bufsz = strlen(ms1) + strlen(ValueToString(ms2)) + 1;
      buf = (char *) gm2(bufsz);
      strcpy(buf,ms1);
      strcat(buf,ValueToString(ms2));
      msg = FindSymbol(buf);
      if (msg == NULL)
        {
         PrintNoHandlerError(buf);
         rm((VOID *) buf,bufsz);
         if (resultbuf != NULL)
           {
            resultbuf->type = SYMBOL;
            resultbuf->value = (VOID *) CLIPSFalseSymbol;
           }
         return;
        }
      rm((VOID *) buf,bufsz);
     }
   else
     msg = ms2;
   if (resultbuf == NULL)
     resultbuf = &temp;
   args.next_arg = remargs;
   args.arg_list = NULL;
   args.type = INSTANCE;
   args.value = (VOID *) ins;
   PerformMessage(resultbuf,&args,msg);
  }
    
/***************************************************
  NAME         : CLIPSSendMessage
  DESCRIPTION  : C Interface for sending messages to
                  instances
  INPUTS       : 1) The data object of the instance
                 2) The message name-string
                 3) The message arguments string
                    (Constants only)
                 4) Caller's buffer for result
  RETURNS      : Nothing useful
  SIDE EFFECTS : Executes message and stores result
                   caller's buffer
  NOTES        : None
 ***************************************************/
globle VOID CLIPSSendMessage(idata,msg,args,result)
  DATA_OBJECT *idata,*result;
  char *msg,*args;
  {
   int error;
   EXPRESSION *iexp;
   SYMBOL_HN *msym;

   result->type = SYMBOL;
   result->value = (VOID *) CLIPSFalseSymbol;
   msym = FindSymbol(msg);
   if (msym == NULL)
     {
      PrintNoHandlerError(msg);
      SetEvaluationError(TRUE);
      return;
     }
   iexp = GenConstant(idata->type,idata->value);
   iexp->next_arg = ParseConstantArguments(args,&error);
   if (error == TRUE)
     {
      ReturnExpression(iexp);
      SetEvaluationError(TRUE);
      return;
     }
   PerformMessage(result,iexp,msym);
   ReturnExpression(iexp);
  }

/*****************************************************
  NAME         : DestroyHandlerLinks
  DESCRIPTION  : Iteratively deallocates handler-links
  INPUTS       : The handler-link list
  RETURNS      : Nothing useful
  SIDE EFFECTS : Deallocation of links
  NOTES        : None
 *****************************************************/
globle VOID DestroyHandlerLinks(mhead)
  HANDLER_LINK *mhead;
  {
   HANDLER_LINK *tmp;
   
   while (mhead != NULL)
     {
      tmp = mhead;
      mhead = mhead->nxt;
      if (tmp->implicit == 0)
        {
         ((HANDLER *) tmp->hnd)->busy--;
         ((HANDLER *) tmp->hnd)->cls->busy--;
        }
      rtn_struct(msg_handler_link,tmp);
     }
  }
  
#if DEBUGGING_FUNCTIONS

/********************************************************************
  NAME         : DisplayCore
  DESCRIPTION  : Gives a schematic "printout" of the
                   core framework for a message showing
                   arounds, primaries, shadows etc.
                 This routine uses recursion to print indentation
                   to indicate shadowing and where handlers begin
                   and end execution wrt one another.
  INPUTS       : 1) The remaining core
                 2) The number of handlers this (partial) core
                    shadows
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : Expects that the core was created in PREVIEW mode,
                   i.e. implicit handlers are SLOT_DESC addresses
                   (in PERFORM mode they are INSTANCE_SLOT addresses)
                 Assumes (partial) core is not empty
 ********************************************************************/
globle VOID DisplayCore(core,sdepth)
  HANDLER_LINK *core;
  int sdepth;
  {
   if (core->type == MAROUND)
     {
      PrintPreviewHandler(core,sdepth,BEGIN_TRACE);
      if (core->nxt != NULL)
        DisplayCore(core->nxt,sdepth+1);
      PrintPreviewHandler(core,sdepth,END_TRACE);
     }
   else
     {
      while ((core != NULL) ? (core->type == MBEFORE) : FALSE)
        {
         PrintPreviewHandler(core,sdepth,BEGIN_TRACE);
         PrintPreviewHandler(core,sdepth,END_TRACE);
         core = core->nxt;
        }
      if ((core != NULL) ? (core->type == MPRIMARY) : FALSE)
        core = DisplayPrimaryCore(core,sdepth);
      while ((core != NULL) ? (core->type == MAFTER) : FALSE)
        {
         PrintPreviewHandler(core,sdepth,BEGIN_TRACE);
         PrintPreviewHandler(core,sdepth,END_TRACE);
         core = core->nxt;
        }
     }
  }

/*******************************************************************
  NAME         : FindPreviewApplicableHandlers
  DESCRIPTION  : See FindApplicableHandlers
                 However, this function only examines classes rather
                   than instances for implicit slot-accessors
  INPUTS       : 1) The class address
                 2) The message name symbol
  RETURNS      : The links of applicable handlers, NULL on errors
  SIDE EFFECTS : Links are allocated for the list
  NOTES        : None
 ******************************************************************/
globle HANDLER_LINK *FindPreviewApplicableHandlers(cls,mname)
  CLASS_TYPE *cls;
  SYMBOL_HN *mname;
  {
   register int i,primary_found;
   HANDLER_LINK *tops[4],*bots[4],*tmp;
   CLASS_LINK *cp,cprv;
   SLOT_DESC *sp;
   
   for (i = MAROUND ; i <= MAFTER ; i++)
     tops[i] = bots[i] = NULL;
     
   cprv.cls = cls;
   cprv.nxt = cls->precedence;
   
   for (cp = &cprv ; cp != NULL ; cp = cp->nxt)
     {
      primary_found = FindApplicableOfName(cp->cls,tops,bots,mname);
      if (((primary_found == TRUE) || (mname->relatedSymbol == NULL)) ? FALSE :
          ((sp = FindClassSlot(cp->cls,mname->relatedSymbol)) != NULL))
        {
         tmp = get_struct(msg_handler_link);
         tmp->hnd = (VOID *) sp;
         tmp->implicit = 1;
         if (mname == sp->gethandler)
           tmp->get = 1;
         else
           tmp->get = 0;
         tmp->type = MPRIMARY;
         tmp->nxt = NULL;
         if (tops[MPRIMARY] == NULL)
           tops[MPRIMARY] = tmp;
         else
           bots[MPRIMARY]->nxt = tmp;
         bots[MPRIMARY] = tmp;
        }
     }
   return(JoinHandlerLinks(tops,bots,mname));
  }
  
#endif

#if (! BLOAD_ONLY) && (! RUN_TIME)

/*******************************************************
  NAME         : NewHandler
  DESCRIPTION  : Allocates and initializes a new handler
  INPUTS       : None
  RETURNS      : The address of the new handler
  SIDE EFFECTS : New handler allocated
  NOTES        : None
 *******************************************************/
globle HANDLER *NewHandler()
  {
   HANDLER *hnd;
   
   hnd = get_struct(msg_handler);
   hnd->system = 0;
   hnd->type = MPRIMARY;
   hnd->mark = 0;
   hnd->busy = 0;
   hnd->name = NULL;
   hnd->cls = NULL;
   hnd->minParams = 0;
   hnd->maxParams = 0;
   hnd->actions = NULL;
   hnd->pp_form = NULL;
   return(hnd);
  }
  
/*****************************************************
  NAME         : HandlersExecuting
  DESCRIPTION  : Determines if any message-handlers
                   for a class are currently executing
  INPUTS       : The class address
  RETURNS      : TRUE if any handlers are executing,
                   FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : None
 *****************************************************/
globle int HandlersExecuting(cls)
  CLASS_TYPE *cls;
  {
   register unsigned i;
   
   for (i = 0 ; i < cls->handler_cnt ; i++)
     if (cls->handlers[i].busy > 0)
       return(TRUE);
   return(FALSE);
  }
  
/*********************************************************************
  NAME         : DeleteHandler
  DESCRIPTION  : Deletes one or more message-handlers
                   from a class definition
  INPUTS       : 1) The class address
                 2) The message-handler name
                    (if this is * and there is no handler
                     called *, then the delete operations
                     will be applied to all handlers matching the type
                 3) The message-handler type
                    (if this is -1, then the delete operations will be
                     applied to all handlers matching the name
                 4) A flag saying whether to print error messages when
                     handlers are not found meeting specs
  RETURNS      : 1 if successful, 0 otherwise
  SIDE EFFECTS : Handlers deleted
  NOTES        : If any handlers for the class are
                   currently executing, this routine
                   will fail
 **********************************************************************/
globle int DeleteHandler(cls,mname,mtype,indicate_missing)
   CLASS_TYPE *cls;
   SYMBOL_HN *mname;
   int mtype,indicate_missing;
  {
   register int i;
   HANDLER *hnd;
   int found,success = 1;

   if (cls->handler_cnt == 0)
     {
      if (indicate_missing)
        {
         HandlerDeleteError(ValueToString(cls->name));
         return(0);
        }
      return(1);
     }
   if (HandlersExecuting(cls))
     {
      HandlerDeleteError(ValueToString(cls->name));
      return(0);
     }
   if (mtype == -1)
     {
      found = FALSE;
      for (i = MAROUND ; i <= MAFTER ; i++)
        {
         hnd = (HANDLER *) FindHandler(cls,mname,i,LOOKUP_HANDLER_ADDRESS);
         if (hnd != NULL)
           {
            found = TRUE;
            if (hnd->system == 0)
              hnd->mark = 1;
            else
              {
               PrintCLIPS(WERROR,"System message-handlers may not be modified.\n");
               success = 0;
              }
           }
        }
      if ((found == FALSE) ? (strcmp(ValueToString(mname),"*") == 0) : FALSE)
        {
         for (i = 0 ; i < cls->handler_cnt ; i++)
           if (cls->handlers[i].system == 0)
             cls->handlers[i].mark = 1;
        }
     }
   else
     {
      hnd = (HANDLER *) FindHandler(cls,mname,mtype,LOOKUP_HANDLER_ADDRESS);
      if (hnd == NULL)
        {
         if (strcmp(ValueToString(mname),"*") == 0)
           {
            for (i = 0 ; i < cls->handler_cnt ; i++)
              if ((cls->handlers[i].type == mtype) && 
                  (cls->handlers[i].system == 0))
                cls->handlers[i].mark = 1;
           }
         else
           {
            if (indicate_missing)
              HandlerDeleteError(ValueToString(cls->name));
            success = 0;
           }
        }
      else if (hnd->system == 0)
        hnd->mark = 1;
      else
        {
         if (indicate_missing)
           PrintCLIPS(WERROR,"System message-handlers may not be modified.\n");
         success = 0;
        }
     }     
   DeallocateMarkedHandlers(cls);
   return(success);
  }

/***************************************************
  NAME         : DeallocateMarkedHandlers
  DESCRIPTION  : Removes any handlers from a class
                   that have been previously marked
                   for deletion.
  INPUTS       : The class
  RETURNS      : Nothing useful
  SIDE EFFECTS : Marked handlers are deleted
  NOTES        : Assumes none of the handlers are
                   currently executing or have a
                   busy count != 0 for any reason
 ***************************************************/
globle VOID DeallocateMarkedHandlers(cls)
  CLASS_TYPE *cls;
  {
   unsigned count;
   HANDLER *hnd,*nhnd;
   unsigned *arr,*narr;
   register int i,j;

   for (i = 0 , count = 0 ; i < cls->handler_cnt ; i++)
     {
      hnd = &cls->handlers[i];
      if (hnd->mark == 1)
        {
         count++;
         if (GetCompilationsWatch())
           PrintAbbreviatedHandlerRemoval(hnd);
         DecrementSymbolCount(hnd->name);
         ExpressionDeinstall(hnd->actions);
         ReturnPackedExpression(hnd->actions);
         if (hnd->pp_form != NULL)
           rm((VOID *) hnd->pp_form,
              (int) (sizeof(char) * (strlen(hnd->pp_form)+1)));
        }
      else
         /* ============================================
            Use the busy field to count how many
            message-handlers are removed before this one
            ============================================ */
        hnd->busy = count;
     }
   if (count == 0)
     return;
   if (count == cls->handler_cnt)
     {
      rm((VOID *) cls->handlers,(int) (sizeof(HANDLER) * cls->handler_cnt));
      rm((VOID *) cls->hordered,(int) (sizeof(unsigned) * cls->handler_cnt));
      cls->handlers = NULL;
      cls->hordered = NULL;
      cls->handler_cnt = 0;
     }
   else
     {
      count = cls->handler_cnt - count;
      hnd = cls->handlers;
      arr = cls->hordered;
      nhnd = (HANDLER *) gm2((int) (sizeof(HANDLER) * count));
      narr = (unsigned *) gm2((int) (sizeof(unsigned) * count));
      for (i = 0 , j = 0 ; j < count ; i++)
        {
         if (hnd[arr[i]].mark == 0)
           {
            /* ==============================================================
               The offsets in the map need to be decremented by the number of
               preceding nodes which were deleted.  Use the value of the busy
               field set in the first loop.
               ============================================================== */
            narr[j] = arr[i] - hnd[arr[i]].busy;
            j++;
           }
        }
      for (i = 0 , j = 0 ; j < count ; i++)
        {
         if (hnd[i].mark == 0)
           {
            hnd[i].busy = 0;
            CopyMemory(HANDLER,1,&nhnd[j],&hnd[i]);
            j++;
           }
        }
      rm((VOID *) hnd,(int) (sizeof(HANDLER) * cls->handler_cnt));
      rm((VOID *) arr,(int) (sizeof(unsigned) * cls->handler_cnt));
      cls->handlers = nhnd;
      cls->hordered = narr;
      cls->handler_cnt = count;
     }
  }
   
/*****************************************************
  NAME         : PrintAbbreviatedHandlerRemoval
  DESCRIPTION  : Printouts a brief description of a
                   message-handler about to be removed
  INPUTS       : The handler address
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : None
 *****************************************************/
globle VOID PrintAbbreviatedHandlerRemoval(hnd)
  HANDLER *hnd;
  {
   PrintCLIPS(WDIALOG,"Removing defmessage-handler c:");
   PrintCLIPS(WDIALOG,ValueToString(hnd->cls->name));
   PrintCLIPS(WDIALOG," ");
   PrintCLIPS(WDIALOG,ValueToString(hnd->name));
   PrintCLIPS(WDIALOG," t:");
   PrintCLIPS(WDIALOG,hndquals[hnd->type]);
   PrintCLIPS(WDIALOG,"\n");
  }
  
#endif

/*****************************************************
  NAME         : HandlerType
  DESCRIPTION  : Determines type of message-handler
  INPUTS       : 1) Calling function string
                 2) String representing type
  RETURNS      : MAROUND  (0) for "around"
                 MBEFORE  (1) for "before"
                 MPRIMARY (2) for "primary"
                 MAFTER   (3) for "after"
                         (-1) for wildcards (if ok)
                 MERROR   (4) on errors
  SIDE EFFECTS : None
  NOTES        : None
 *****************************************************/
globle int HandlerType(func,str)
  char *func,*str;
  {
   register int i;
   
   for (i = MAROUND ; i <= MAFTER ; i++)
     if (strcmp(str,hndquals[i]) == 0)
       return(i);
   PrintCLIPS("werror","Unrecognized message-handler type in ");
   PrintCLIPS("werror",func);
   PrintCLIPS("werror",".\n");
   return(MERROR);
  }
  
/*****************************************************************
  NAME         : CheckCurrentMessage
  DESCRIPTION  : Makes sure that a message is available
                   and active for an internal message function
  INPUTS       : 1) The name of the function checking the message
                 2) A flag indicating whether the object must be
                      a class instance or not (it could be a CLIPS
                      primitive type)
  RETURNS      : TRUE if all OK, FALSE otherwise
  SIDE EFFECTS : EvaluationError set on errors
  NOTES        : None
 *****************************************************************/
globle int CheckCurrentMessage(func,ins_reqd)
  char *func;
  int ins_reqd;
  {
   if (CurrentMessageFrame == NULL)
     {
      PrintCLIPS(WERROR,func);
      PrintCLIPS(WERROR," may only be called from within message-handlers.\n");
      SetEvaluationError(TRUE);
      return(FALSE);
     }
   if ((ins_reqd == TRUE) ? (CurrentMessageFrame->type != INSTANCE) : FALSE)
     {
      PrintCLIPS(WERROR,func);
      PrintCLIPS(WERROR," operates only on instances.\n");
      SetEvaluationError(TRUE);
      return(FALSE);
     }
   if ((CurrentMessageFrame->type == INSTANCE) ?
       (((INSTANCE_TYPE *) CurrentMessageFrame->value)->garbage == 1) : FALSE)
     {
      StaleInstanceAddress(func);
      SetEvaluationError(TRUE);
      return(FALSE);
     }
   return(TRUE);
  }
  
/***************************************************
  NAME         : PrintCurrentMessage
  DESCRIPTION  : Prints a synopsis of currently
                   executing handler
  INPUTS       : The logical name for the output
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
globle VOID PrintCurrentMessage(log)
  char *log;
  {
   HANDLER *hnd;
   SLOT_DESC *sp;
   
   if (CurrentCore->implicit == 0)
     {
      hnd = (HANDLER *) CurrentCore->hnd;
      PrintHandler(log,ValueToString(hnd->cls->name),ValueToString(hnd->name),
                   hndquals[hnd->type]);
     }
   else
     {
      sp = ((INSTANCE_SLOT *) CurrentCore->hnd)->desc;
      PrintHandler(log,ValueToString(sp->cls->name),
                   (CurrentCore->get == 1) ? ValueToString(sp->gethandler) : 
                                             ValueToString(sp->puthandler),
                    "slot-accessor");
     }
  }
  
/***************************************************
  NAME         : PrintHandler
  DESCRIPTION  : Displays a handler synopsis
  INPUTS       : 1) Logical name of output
                 2) Class name-string
                 3) Handler name-string
                 4) Type name-string
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
globle VOID PrintHandler(log,cstr,mstr,tstr)
  char *log,*cstr,*mstr,*tstr;
  {
   PrintCLIPS(log,mstr);
   PrintCLIPS(log," ");
   PrintCLIPS(log,tstr);
   PrintCLIPS(log," in class ");
   PrintCLIPS(log,cstr);
   PrintCLIPS(log,".\n");
  }
  
/***********************************************************
  NAME         : FindHandler
  DESCRIPTION  : Uses a binary search on a class's
                   handler header array
  INPUTS       : 1) The class address
                 2) The handler symbolic name
                 3) The handler type (MPRIMARY,etc.)
                 4) The type of return value expected
                    Handler index (0) or address (1)
  RETURNS      : The address (or index) of the found handler,
                   NULL (or -1) if not found
  SIDE EFFECTS : None
  NOTES        : Assumes array is in ascending order
                   1st key: symbolic name of handler
                   2nd key: type of handler
 ***********************************************************/
globle VOID *FindHandler(cls,name,type,lookup_type)
  CLASS_TYPE *cls;
  SYMBOL_HN *name;
  unsigned type;
  int lookup_type;
  {
   register int b,i;
   HANDLER *hnd;
   unsigned *arr;
   
   if ((b = FindHandlerNameGroup(cls,name)) == -1)
     return((lookup_type == LOOKUP_HANDLER_ADDRESS) ? NULL : (VOID *) -1);
   arr = cls->hordered;
   hnd = cls->handlers;
   for (i = b ; i < cls->handler_cnt ; i++)
     {
      if (hnd[arr[i]].name != name)
        return((lookup_type == LOOKUP_HANDLER_ADDRESS) ? NULL : (VOID *) -1);
      if (hnd[arr[i]].type == type)
        return((lookup_type == LOOKUP_HANDLER_ADDRESS) ? (VOID *) &hnd[arr[i]] : 
                                                                 (VOID *) arr[i]);
     }
   return((lookup_type == LOOKUP_HANDLER_ADDRESS) ? NULL : (VOID *) -1);
  }
  
/*****************************************************
  NAME         : FindHandlerNameGroup
  DESCRIPTION  : Uses a binary search on a class's
                   handler header array
  INPUTS       : 1) The class address
                 2) The handler symbolic name
  RETURNS      : The index of the found handler group
                   -1 if not found
  SIDE EFFECTS : None
  NOTES        : Assumes array is in ascending order
                   1st key: handler name symbol bucket
 *****************************************************/
globle int FindHandlerNameGroup(cls,name)
  CLASS_TYPE *cls;
  SYMBOL_HN *name;
  {
   register int b,e,i,j;
   HANDLER *hnd;
   unsigned *arr;
   int start;
   
   if (cls->handler_cnt == 0)
     return(-1);
   hnd = cls->handlers;
   arr = cls->hordered;
   b = 0;
   e = cls->handler_cnt-1;
   start = -1;
   do 
     {
      i = (b+e)/2;
      if (name->bucket == hnd[arr[i]].name->bucket)
        {
         for (j = i ; j >= b ; j--)
           {
            if (hnd[arr[j]].name == name)
              start = j;
            if (hnd[arr[j]].name->bucket != name->bucket)
              break;
           }
         if (start != -1)
           return(start);
         for (j = i+1 ; j <= e ; j++)
           {
            if (hnd[arr[j]].name == name)
              return(j);
            if (hnd[arr[j]].name->bucket != name->bucket)
              return(-1);
           }
         return(-1);
        }
      else if (name->bucket < hnd[arr[i]].name->bucket)
        e = i-1;
      else
        b = i+1;
     }
   while (b <= e);
   return(-1);
  }
  
/***************************************************
  NAME         : HandlerDeleteError
  DESCRIPTION  : Prints out an error message when
                   handlers cannot be deleted
  INPUTS       : Name-string of the class
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
globle VOID HandlerDeleteError(cname)
  char *cname;
  {
   PrintCLIPS(WERROR,"Unable to delete message-handler(s) from class ");
   PrintCLIPS(WERROR,cname);
   PrintCLIPS(WERROR,".\n");
  }

/***********************************************************************
  NAME         : DispatchMessage
  DESCRIPTION  : Determines the applicable handler(s) and sets up the
                   core calling frame.  Then calls the core frame.
  INPUTS       : Caller's space for storing the result of the handler(s)
  RETURNS      : Nothing useful
  SIDE EFFECTS : Any side-effects caused by the execution of handlers in
                   the core framework
  NOTES        : CLIPS Syntax : (send <instance> <hnd> <args>*)
 ***********************************************************************/
globle VOID DispatchMessage(result)
  DATA_OBJECT *result;
  {
   EXPRESSION args;
   SYMBOL_HN *msg;
   
   result->type = SYMBOL;
   result->value = (VOID *) CLIPSFalseSymbol;
   if ((GetFirstArgument() == NULL) ? TRUE : (GetFirstArgument()->next_arg == NULL))
     {
      ExpectedCountError("send",AT_LEAST,2);
      return;
     }
   if (GetFirstArgument()->next_arg->type == SYMBOL)
     msg = (SYMBOL_HN *) GetFirstArgument()->next_arg->value;
   else
     {
      DATA_OBJECT temp;

      EvaluateExpression(GetFirstArgument()->next_arg,&temp);
      if (temp.type != SYMBOL)
        {
         ExpectedTypeError("send",2,"symbol");
         return;
        }
      msg = (SYMBOL_HN *) temp.value;
     }
   
   /* ============================================= 
      Get the instance or primitive for the message
      ============================================= */
   args.type = GetFirstArgument()->type;
   args.value = GetFirstArgument()->value;
   args.arg_list = GetFirstArgument()->arg_list;
   args.next_arg = GetFirstArgument()->next_arg->next_arg;
   
   PerformMessage(result,&args,msg);
  }
    
/*****************************************************
  NAME         : NextHandlerAvailable
  DESCRIPTION  : Determines if there the currently
                   executing handler can call a
                   shadowed handler
                 Used before calling call-next-handler
  INPUTS       : None
  RETURNS      : TRUE if shadow ready, FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : CLIPS Syntax: (next-handlerp)
 *****************************************************/
globle int NextHandlerAvailable()
  {
   if (ArgCountCheck("next-handlerp",EXACTLY,0) == -1)
     return(FALSE);
   if (CurrentCore == NULL)
     return(FALSE);
   if (CurrentCore->type == MAROUND)
     return((NextInCore != NULL) ? TRUE : FALSE);
   if ((CurrentCore->type == MPRIMARY) && (NextInCore != NULL))
     return((NextInCore->type == MPRIMARY) ? TRUE : FALSE);
   return(FALSE);
  }
  
/********************************************************
  NAME         : CallNextHandler
  DESCRIPTION  : This function allows around-handlers
                   to execute the rest of the core frame.
                 It also allows primary handlers
                   to execute shadowed primaries.
                   
                 The original handler arguments are
                   left intact.
  INPUTS       : The caller's result-value buffer
  RETURNS      : Nothing useful
  SIDE EFFECTS : The core frame is called and any
                   appropriate changes are made when
                   used in an around handler
                   See CallHandlers()
                 But when call-next-handler is called
                   from a primary, the same shadowed
                   primary is called over and over
                   again for repeated calls to
                   call-next-handler.
  NOTES        : CLIPS Syntax: (call-next-handler) OR
                    (override-next-handler <arg> ...)
 ********************************************************/
globle VOID CallNextHandler(result)
  DATA_OBJECT *result;
  {
   EXPRESSION args;
   int oldMessageSize = -1,itmp;
   DATA_OBJECT *oldFrame = NULL,*oldBindList;
   
   SetpType(result,SYMBOL);
   SetpValue(result,(VOID *) CLIPSFalseSymbol);
   EvaluationError = FALSE;
   if (HaltExecution)
     return;
   if (CurrentExpression->value == (VOID *) FindFunction("override-next-handler"))
     oldMessageSize = 0;
   else if (ArgCountCheck("call-next-handler",EXACTLY,0))
     return;
   if ((CurrentCore == NULL) ? TRUE :
       (CurrentCore->type == MBEFORE) || (CurrentCore->type == MAFTER))
     {
      PrintCLIPS(WERROR,"Shadowed message-handlers not applicable in current context.\n");
      SetEvaluationError(TRUE);
      return;
     }
   if (CurrentCore->type == MAROUND)
     {
      if (NextInCore == NULL)
        {
         PrintCLIPS(WERROR,"Shadowed message-handlers not applicable in current context.\n");
         SetEvaluationError(TRUE);
         return;
        }
      if (oldMessageSize != -1)
        {
         oldFrame = CurrentMessageFrame;
         args.type = SYMBOL;
         args.value = (VOID *) CLIPSTrueSymbol;
         args.next_arg = GetFirstArgument();
         args.arg_list = NULL;
         itmp = CountArguments(&args);
         CurrentMessageFrame = EvaluateMessageParameters(&args,itmp);
         if (EvaluationError)
           {
            ReturnFlag = CLIPS_FALSE;
            CurrentMessageFrame = oldFrame;
            return;
           }
         oldMessageSize = CurrentMessageSize;
         CurrentMessageSize = itmp;
         CopyMemory(DATA_OBJECT,1,&CurrentMessageFrame[0],&oldFrame[0]);
        }
      oldBindList = BindList;
      BindList = NULL;
      if (NextInCore->type == MAROUND)
        {
         HANDLER_LINK *oldNext,*oldCurrent;
         
         oldNext = NextInCore;
         oldCurrent = CurrentCore;
         CurrentCore = NextInCore;
         NextInCore = NextInCore->nxt;
#if DEBUGGING_FUNCTIONS
         if (WatchHandlers)
           {
            TraceHandler(WTRACE,CurrentCore,BEGIN_TRACE);
            if (CheckHandlerArgCount())
              EvaluateExpression(((HANDLER *) CurrentCore->hnd)->actions,result);
            TraceHandler(WTRACE,CurrentCore,END_TRACE);
           }
         else
#endif
         if (CheckHandlerArgCount())
           EvaluateExpression(((HANDLER *) CurrentCore->hnd)->actions,result);
         NextInCore = oldNext;
         CurrentCore = oldCurrent;
        }
      else
        CallHandlers(result);
     }
   else
     {
      HANDLER_LINK *oldNext,*oldCurrent;
         
      if ((NextInCore != NULL) ? (NextInCore->type != MPRIMARY) : TRUE)
        {
         PrintCLIPS(WERROR,"Shadowed message-handlers not applicable in current context.\n");
         SetEvaluationError(TRUE);
         return;
        }
      if (oldMessageSize != -1)
        {
         oldFrame = CurrentMessageFrame;
         oldMessageSize = CurrentMessageSize;
         args.type = SYMBOL;
         args.value = (VOID *) CLIPSTrueSymbol;
         args.next_arg = GetFirstArgument();
         args.arg_list = NULL;
         itmp = CountArguments(&args);
         CurrentMessageFrame = EvaluateMessageParameters(&args,itmp);
         if (EvaluationError)
           {
            ReturnFlag = CLIPS_FALSE;
            CurrentMessageFrame = oldFrame;
            return;
           }
         oldMessageSize = CurrentMessageSize;
         CurrentMessageSize = itmp;
         CopyMemory(DATA_OBJECT,1,&CurrentMessageFrame[0],&oldFrame[0]);
        }
      oldBindList = BindList;
      BindList = NULL;
      oldNext = NextInCore;
      oldCurrent = CurrentCore;
      CurrentCore = NextInCore;
      NextInCore = NextInCore->nxt;
#if DEBUGGING_FUNCTIONS
      if (WatchHandlers)
        {
         TraceHandler(WTRACE,CurrentCore,BEGIN_TRACE);
         if (oldNext->implicit == 0)
           {
            if (CheckHandlerArgCount())
              EvaluateExpression(((HANDLER *) oldNext->hnd)->actions,result);
           }
         else
           PerformImplicitHandler(result);
         TraceHandler(WTRACE,CurrentCore,END_TRACE);
        }
      else
#endif
      if (oldNext->implicit == 0)
        {
         if (CheckHandlerArgCount())
           EvaluateExpression(((HANDLER *) CurrentCore->hnd)->actions,result);
        }
      else
        PerformImplicitHandler(result);
      NextInCore = oldNext;
      CurrentCore = oldCurrent;
     }
   if (BindList != NULL)
     ReturnValues(BindList);
   BindList = oldBindList;
   if (oldMessageSize != -1)
     {
      if (CurrentMessageSize != 0)
        rm(CurrentMessageFrame,(int) (sizeof(DATA_OBJECT) * CurrentMessageSize));
      CurrentMessageFrame = oldFrame;
      CurrentMessageSize = oldMessageSize;
     }
   ReturnFlag = CLIPS_FALSE;
  }
  
/* =========================================
   *****************************************
          INTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
   
/*****************************************************
  NAME         : PerformMessage
  DESCRIPTION  : Calls core framework for a message
  INPUTS       : 1) Caller's result buffer
                 2) Message argument expressions
                    (including implicit object)
                 3) Message name
  RETURNS      : Nothing useful
  SIDE EFFECTS : Any side-effects of message execution
                    and caller's result buffer set
  NOTES        : None
 *****************************************************/
static VOID PerformMessage(result,args,mname)
  DATA_OBJECT *result;
  EXPRESSION *args;
  SYMBOL_HN *mname;
  {
   DATA_OBJECT *oldFrame,*oldBindList;
   int oldMessageSize,itmp,oldce,svReturnContext,svBreakContext;
   HANDLER_LINK *oldCore;
   CLASS_TYPE *cls = NULL;
   INSTANCE_TYPE *ins = NULL;
   SYMBOL_HN *oldName;
   
   result->type = SYMBOL;
   result->value = (VOID *) CLIPSFalseSymbol;
   EvaluationError = FALSE;
   if (HaltExecution)
     return;
   oldce = ExecutingConstruct();
   SetExecutingConstruct(TRUE);
   oldName = CurrentMessageName;
   CurrentMessageName = mname;
   CurrentEvaluationDepth++;
   oldFrame = CurrentMessageFrame;
   itmp = CountArguments(args);
   CurrentMessageFrame = EvaluateMessageParameters(args,itmp);
   if (EvaluationError)
     {
      CurrentEvaluationDepth--;
      CurrentMessageFrame = oldFrame;
      CurrentMessageName = oldName;
      PeriodicCleanup(FALSE,TRUE);
      SetExecutingConstruct(oldce);
      return;
     }
   oldMessageSize = CurrentMessageSize;
   CurrentMessageSize = itmp;
   if (CurrentMessageFrame->type == INSTANCE)
     {
      ins = (INSTANCE_TYPE *) CurrentMessageFrame->value;
      if (ins->garbage == 1)
        {
         StaleInstanceAddress("send");
         SetEvaluationError(TRUE);
        }
      else
        {
         cls = ins->cls;
         ins->busy++;
        }
     }
   else if (CurrentMessageFrame->type == INSTANCE_NAME)
     {
      ins = FindInstanceBySymbol((SYMBOL_HN *) CurrentMessageFrame->value);
      if (ins == NULL)
        {
         PrintCLIPS(WERROR,"No such instance ");
         PrintCLIPS(WERROR,ValueToString((SYMBOL_HN *) CurrentMessageFrame->value));
         PrintCLIPS(WERROR," in function send.\n");
         SetEvaluationError(TRUE);
        }
      else
        {
         CurrentMessageFrame->value = (VOID *) ins;
         CurrentMessageFrame->type = INSTANCE;
         cls = ins->cls;
         ins->busy++;
        }
     }
   else if ((cls = PrimitiveClassMap[CurrentMessageFrame->type]) == NULL)
     {
      PrintCLIPS(WERROR,"Expected a valid instance in function send.\n");
      SetEvaluationError(TRUE);
     }
   if (EvaluationError)
     {
      rm(CurrentMessageFrame,(int) (sizeof(DATA_OBJECT) * CurrentMessageSize));
      CurrentEvaluationDepth--;
      CurrentMessageFrame = oldFrame;
      CurrentMessageSize = oldMessageSize;
      CurrentMessageName = oldName;
      PeriodicCleanup(FALSE,TRUE);
      SetExecutingConstruct(oldce);
      return;
     }
   oldBindList = BindList;
   BindList = NULL;
   
   oldCore = TopOfCore;
   TopOfCore = FindApplicableHandlers(cls,mname);
   svBreakContext = BreakContext;
   BreakContext = CLIPS_FALSE;
   svReturnContext = ReturnContext;
   ReturnContext = CLIPS_TRUE;
   if (TopOfCore != NULL)
     {
      HANDLER_LINK *oldCurrent,*oldNext;

      oldCurrent = CurrentCore;
      oldNext = NextInCore;
      
      if (TopOfCore->type == MAROUND)
        {
         CurrentCore = TopOfCore;
         NextInCore = TopOfCore->nxt;
#if DEBUGGING_FUNCTIONS
         if (WatchMessages)
           TraceMessage(WTRACE,BEGIN_TRACE);
         if (WatchHandlers)
           {
            TraceHandler(WTRACE,TopOfCore,BEGIN_TRACE);
            if (CheckHandlerArgCount())
              EvaluateExpression(((HANDLER *) CurrentCore->hnd)->actions,result);
            TraceHandler(WTRACE,CurrentCore,END_TRACE);
           }
         else if (CheckHandlerArgCount())
           EvaluateExpression(((HANDLER *) CurrentCore->hnd)->actions,result);
         if (WatchMessages)
           TraceMessage(WTRACE,END_TRACE);
#else
         if (CheckHandlerArgCount())
           EvaluateExpression(((HANDLER *) CurrentCore->hnd)->actions,result);
#endif
        }
      else
        {
         CurrentCore = NULL;
         NextInCore = TopOfCore;
#if DEBUGGING_FUNCTIONS
         if (WatchMessages)
           {
            TraceMessage(WTRACE,BEGIN_TRACE);
            CallHandlers(result);
            TraceMessage(WTRACE,END_TRACE);
           }
         else
#endif
           CallHandlers(result);
        }

      DestroyHandlerLinks(TopOfCore);
      CurrentCore = oldCurrent;
      NextInCore = oldNext;
     }
   TopOfCore = oldCore;
   BreakContext = svBreakContext;
   ReturnContext = svReturnContext;
   ReturnFlag = CLIPS_FALSE;

   if (ins != NULL)
     ins->busy--;

   /* ==================================
      Restore the original calling frame
      ================================== */
   rm(CurrentMessageFrame,(int) (sizeof(DATA_OBJECT) * CurrentMessageSize));
   if (BindList != NULL)
     ReturnValues(BindList);
   CurrentMessageFrame = oldFrame;
   CurrentMessageSize = oldMessageSize;
   BindList = oldBindList;
   CurrentEvaluationDepth--;
   CurrentMessageName = oldName;
   PropogateReturnValue(result);
   PeriodicCleanup(FALSE,TRUE);
   SetExecutingConstruct(oldce);
   if (EvaluationError)
     {
      result->type = SYMBOL;
      result->value = (VOID *) CLIPSFalseSymbol;
     }
  }

/*********************************************************
  NAME         : EvaluateMessageParameters
  DESCRIPTION  : Takes an argument list and evaluates it
  INPUTS       : 1) The argument list
                 2) The number of arguments in the list
  RETURNS      : The address of a contiguous array holding
                  the final values, NULL on errors
  SIDE EFFECTS : Value array allocated
  NOTES        : None
 *********************************************************/
static DATA_OBJECT *EvaluateMessageParameters(plist,pcnt)
  EXPRESSION *plist;
  int pcnt;
  {
   DATA_OBJECT *rva;
   int i = 0;
   
   if (pcnt == 0) return(NULL);
   
   rva = (DATA_OBJECT *) gm2((int) (sizeof(DATA_OBJECT) * pcnt));
   
   while (plist != NULL)
     {
      if ((EvaluateExpression(plist,&rva[i]) == TRUE) ? TRUE :
          (rva[i].type == RVOID))
        {
         if (rva[i].type == RVOID)
           {
            PrintCLIPS(WERROR,"Functions without a return value are illegal as message arguments.\n");
            SetEvaluationError(TRUE);
           }
         PrintCLIPS(WERROR,"This error occurred while evaluating arguments ");
         PrintCLIPS(WERROR,"for the message ");
         PrintCLIPS(WERROR,ValueToString(CurrentMessageName));
         PrintCLIPS(WERROR,".\n");
         rm(rva,sizeof(DATA_OBJECT) * pcnt);
         return(NULL);
        }
      plist = plist->next_arg;
      if (plist != NULL)
        { rva[i].next = &rva[i+1]; }
      else
        { rva[i].next = NULL; }
      i++;
     }
     
   return(rva);
  }
  
#if DEBUGGING_FUNCTIONS

/********************************************************************
  NAME         : DisplayPrimaryCore
  DESCRIPTION  : Gives a schematic "printout" of the primary
                   message showing other shadowed primaries
                 This routine uses recursion to print indentation
                   to indicate shadowing and where handlers begin
                   and end execution wrt one another.
  INPUTS       : 1) The remaining core
                 2) The number of handlers this (partial) core
                    shadows
  RETURNS      : The address of the handler following the primary
                   group of handlers in the core
  SIDE EFFECTS : None
  NOTES        : Expects that the core was created in PREVIEW mode,
                   i.e. implicit handlers are SLOT_DESC addresses
                   (in PERFORM mode they are INSTANCE_SLOT addresses)
                 Assumes (partial) core is not empty
 ********************************************************************/
static HANDLER_LINK *DisplayPrimaryCore(core,pdepth)
  HANDLER_LINK *core;
  int pdepth;
  {
   register HANDLER_LINK *rtn;
   
   PrintPreviewHandler(core,pdepth,BEGIN_TRACE);
   if ((core->nxt != NULL) ? (core->nxt->type == MPRIMARY) : FALSE)
     rtn = DisplayPrimaryCore(core->nxt,pdepth+1);
   else
     rtn = core->nxt;
   PrintPreviewHandler(core,pdepth,END_TRACE);
   return(rtn);
  }
  
/***************************************************
  NAME         : PrintPreviewHandler
  DESCRIPTION  : Displays a message preview
  INPUTS       : 1) Handler-link
                 2) Number of handlers shadowed
                 3) The trace-string
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
static VOID PrintPreviewHandler(cptr,sdepth,tstr)
  HANDLER_LINK *cptr;
  int sdepth;
  char *tstr;
  {
   register int i;
   SLOT_DESC *sp;
   HANDLER *hnd;
   
   for (i = 0 ; i < sdepth ; i++)
     PrintCLIPS(WDISPLAY,"| ");
   PrintCLIPS(WDISPLAY,tstr);
   PrintCLIPS(WDISPLAY," ");
   if (cptr->implicit == 1)
     {
      sp = (SLOT_DESC *) cptr->hnd;
      PrintHandler(WDISPLAY,ValueToString(sp->cls->name),
                   (cptr->get == 1) ? ValueToString(sp->gethandler) : 
                                      ValueToString(sp->puthandler),
                    "slot-accessor");
     }
   else
     {
      hnd = (HANDLER *) cptr->hnd;
      PrintHandler(WDISPLAY,ValueToString(hnd->cls->name),ValueToString(hnd->name),
                   hndquals[hnd->type]);
     }
  }

/***********************************************************
  NAME         : TraceMessage
  DESCRIPTION  : Prints a condensed description of a
                   message and its arguments
  INPUTS       : 1) The output logical name
                 2) BEGIN_TRACE or END_TRACE string
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : Uses the global variables CurrentMessageFrame
                   and CurrentMessageName
 ***********************************************************/
static VOID TraceMessage(log,tstring)
  char *log;
  char *tstring;
  {
   register int i;
   
   PrintCLIPS(log,"MSG ");
   PrintCLIPS(log,tstring);
   PrintCLIPS(log," ");
   PrintCLIPS(log,ValueToString(CurrentMessageName));
   PrintCLIPS(log," ED:");
   PrintLongInteger(log,(long) CurrentEvaluationDepth);
   PrintCLIPS(log," (");
   if (CurrentMessageFrame[0].type == INSTANCE)
     {
      PrintCLIPS(log,"[");
      PrintCLIPS(log,ValueToString(((INSTANCE_TYPE *) CurrentMessageFrame[0].value)->name));
      PrintCLIPS(log,"]");
     }
   else
     PrintDataObject(log,&CurrentMessageFrame[0]);
   if (CurrentMessageSize > 1)
     PrintCLIPS(log," ");
   for (i = 1 ; i < CurrentMessageSize ; i++)
     {
      PrintDataObject(log,&CurrentMessageFrame[i]);
      if (i != CurrentMessageSize-1)
        PrintCLIPS(log," ");
     }
   PrintCLIPS(log,")\n");
  }
  
/***********************************************************
  NAME         : TraceHandler
  DESCRIPTION  : Prints a condensed description of a
                   message-handler and its arguments
  INPUTS       : 1) The output logical name
                 2) The handler address
                 3) BEGIN_TRACE or END_TRACE string
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : Uses the global variables CurrentMessageFrame
                   and CurrentMessageName
 ***********************************************************/
static VOID TraceHandler(log,hndl,tstring)
  char *log;
  HANDLER_LINK *hndl;
  char *tstring;
  {
   register int i;
   HANDLER *hnd;
   SLOT_DESC *sp;
   
   PrintCLIPS(log,"HND ");
   PrintCLIPS(log,tstring);
   PrintCLIPS(log," ");
   if (hndl->implicit == 0)
     {
      hnd = (HANDLER *) hndl->hnd;
      PrintHandler(WTRACE,ValueToString(hnd->cls->name),ValueToString(hnd->name),
                   hndquals[hnd->type]);
     }
   else
     {
      sp = ((INSTANCE_SLOT *) hndl->hnd)->desc;
      PrintHandler(WTRACE,ValueToString(sp->cls->name),
                   (hndl->get == 1) ? ValueToString(sp->gethandler) : 
                                      ValueToString(sp->puthandler),
                   "slot-accessor");
     }
   PrintCLIPS(log,"       ED:");
   PrintLongInteger(log,(long) CurrentEvaluationDepth);
   PrintCLIPS(log," (");
   if (CurrentMessageFrame[0].type == INSTANCE)
     {
      PrintCLIPS(log,"[");
      PrintCLIPS(log,ValueToString(((INSTANCE_TYPE *) CurrentMessageFrame[0].value)->name));
      PrintCLIPS(log,"]");
     }
   else
     PrintDataObject(log,&CurrentMessageFrame[0]);
   if (CurrentMessageSize > 1)
     PrintCLIPS(log," ");
   for (i = 1 ; i < CurrentMessageSize ; i++)
     {
      PrintDataObject(log,&CurrentMessageFrame[i]);
      if (i != CurrentMessageSize-1)
        PrintCLIPS(log," ");
     }
   PrintCLIPS(log,")\n");
  }

#endif
  
/*****************************************************************************
  NAME         : FindApplicableHandlers
  DESCRIPTION  : Given a message name, this routine forms the "core frame"
                   for the message : a list of all applicable class handlers.
                   An applicable class handler is one whose name matches
                     the message and whose class matches the instance.
                     
                   The list is in the following order :
                   
                   All around handlers (from most specific to most general)
                   All before handlers (from most specific to most general)
                   All primary handlers (from most specific to most general)
                   All after handlers (from most general to most specific)
                   
                 Implicit primary get- and put- handler links are also formed
  INPUTS       : 1) The class of the instance (or primitive) for the message
                 2) The message name
  RETURNS      : NULL if no applicable handlers or errors,
                   the list of handlers otherwise
  SIDE EFFECTS : Links are allocated for the list
  NOTES        : The instance is the first thing on the CurrentMessageFrame
                 The number of arguments is in CurrentMessageSize
 *****************************************************************************/
static HANDLER_LINK *FindApplicableHandlers(cls,mname)
  CLASS_TYPE *cls;
  SYMBOL_HN *mname;
  {
   register int i;
   HANDLER_LINK *tops[4],*bots[4],*tmp;
   CLASS_LINK *cp,cprv;
   INSTANCE_SLOT *sp = NULL;
   CLASS_TYPE *spclass;
   
   for (i = MAROUND ; i <= MAFTER ; i++)
     tops[i] = bots[i] = NULL;
     
   cprv.cls = cls;
   cprv.nxt = cls->precedence;
     
   if ((CurrentMessageFrame[0].type == INSTANCE) && 
       (mname->relatedSymbol != NULL))
     {
      sp = FindInstanceSlot((INSTANCE_TYPE *) CurrentMessageFrame[0].value,
                            mname->relatedSymbol);
      spclass = (sp != NULL) ? sp->desc->cls : NULL;
     }
   else
     spclass = NULL;
   for (cp = &cprv ; cp != NULL ; cp = cp->nxt)
     {
      if ((FindApplicableOfName(cp->cls,tops,bots,mname) == FALSE) ?
          (spclass == cp->cls) : FALSE)
        {
         tmp = get_struct(msg_handler_link);
         tmp->hnd = (VOID *) sp;
         tmp->implicit = 1;
         if (mname == sp->desc->gethandler)
           tmp->get = 1;
         else
           tmp->get = 0;
         tmp->type = MPRIMARY;
         tmp->nxt = NULL;
         if (tops[MPRIMARY] == NULL)
           tops[MPRIMARY] = tmp;
         else
           bots[MPRIMARY]->nxt = tmp;
         bots[MPRIMARY] = tmp;
        }
     }
   return(JoinHandlerLinks(tops,bots,mname));
  }
  
/*************************************************************************
  NAME         : FindApplicableOfName
  DESCRIPTION  : Groups all handlers of all types of the specified
                   class of the specified name into the applicable handler
                   list
  INPUTS       : 1) The class address
                 2-3) The tops and bottoms of the four handler type lists:
                      around, before, primary and after
                 4) The message name symbol
  RETURNS      : TRUE if any primary handlers found, FALSE otherwise
  SIDE EFFECTS : Modifies the handler lists to include applicable handlers
  NOTES        : None
 *************************************************************************/
static int FindApplicableOfName(cls,tops,bots,mname)
  CLASS_TYPE *cls;
  HANDLER_LINK *tops[4],*bots[4];
  SYMBOL_HN *mname;
  {
   register int i,e;
   HANDLER *hnd;
   unsigned *arr;
   HANDLER_LINK *tmp,*sv;
   
   i = FindHandlerNameGroup(cls,mname);
   if (i == -1)
     return(FALSE);
   e = cls->handler_cnt-1;
   hnd = cls->handlers;
   arr = cls->hordered;
   sv = bots[MPRIMARY];
   for ( ; i <= e ; i++)
     {
      if (hnd[arr[i]].name != mname)
        break;
      tmp = get_struct(msg_handler_link);
      hnd[arr[i]].busy++;
      hnd[arr[i]].cls->busy++;
      tmp->hnd = (VOID *) &hnd[arr[i]];
      tmp->implicit = 0;
      tmp->type = hnd[arr[i]].type;
      if (tops[tmp->type] == NULL)
        {
         tmp->nxt = NULL;
         tops[tmp->type] = bots[tmp->type] = tmp;
        }
      else if (tmp->type == MAFTER)
        {
         tmp->nxt = tops[tmp->type];
         tops[tmp->type] = tmp;
        }
      else
        {
         bots[tmp->type]->nxt = tmp;
         bots[tmp->type] = tmp;
         tmp->nxt = NULL;
        }
     }
   return((sv != bots[MPRIMARY]) ? TRUE : FALSE);
  }
    
/************************************************************************
  NAME         : JoinHandlerLinks
  DESCRIPTION  : 
  INPUTS       : 1-2) The tops and bottoms of the four handler type lists:
                      around, before, primary and after
                 3) The message name symbol
  RETURNS      : The top of the joined lists, NULL on errors
  SIDE EFFECTS : Links all the handler type lists together, or all the lists
                   are destroyed if there
  NOTES        : None
 ************************************************************************/
static HANDLER_LINK *JoinHandlerLinks(tops,bots,mname)
  HANDLER_LINK *tops[4],*bots[4];
  SYMBOL_HN *mname;
  {
   register int i;
   HANDLER_LINK *mlink;
   
   if (tops[MPRIMARY] == NULL)
    {
     PrintNoHandlerError(ValueToString(mname));
     for (i = MAROUND ; i <= MAFTER ; i++)
       DestroyHandlerLinks(tops[i]);
     SetEvaluationError(TRUE);
     return(NULL);
    }
   
   mlink = tops[MPRIMARY];
   if (tops[MBEFORE] != NULL)
     {
      bots[MBEFORE]->nxt = mlink;
      mlink = tops[MBEFORE];
     }
   if (tops[MAROUND] != NULL)
     {
      bots[MAROUND]->nxt = mlink;
      mlink = tops[MAROUND];
     }
   bots[MPRIMARY]->nxt = tops[MAFTER];
   return(mlink);
  }
  
/*****************************************************************
  NAME         : PrintNoHandlerError
  DESCRIPTION  : Print "No primaries found" error message for send
  INPUTS       : The name of the message
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : None
 *****************************************************************/
static VOID PrintNoHandlerError(msg)
  char *msg;
  {
   PrintCLIPS(WERROR,"No applicable primary message-handlers found for ");
   PrintCLIPS(WERROR,msg);
   PrintCLIPS(WERROR,".\n");
  }
  
/***************************************************************
  NAME         : CallHandlers
  DESCRIPTION  : Moves though the current message frame
                   for a send-message as follows :

                 Call all before handlers and ignore their
                   return values.
                 Call the first primary handler and
                   ignore the rest.  The return value
                   of the handler frame is this message's value.
                 Call all after handlers and ignore their
                   return values.
  INPUTS       : Caller's buffer for the return value of
                   the message
  RETURNS      : Nothing useful
  SIDE EFFECTS : The handlers are evaluated.
  NOTES        : IMPORTANT : The global NextInCore should be
                 pointing to the first handler to be executed.
 ***************************************************************/
static VOID CallHandlers(result)
  DATA_OBJECT *result;
  {
   HANDLER_LINK *oldCurrent,*oldNext;
   DATA_OBJECT *oldBindList;
   DATA_OBJECT temp;
   
   if (HaltExecution)
     return;
   oldBindList = BindList;
   BindList = NULL;
   oldCurrent = CurrentCore;
   oldNext = NextInCore;
   while (NextInCore->type == MBEFORE)
     {
      CurrentCore = NextInCore;
      NextInCore = NextInCore->nxt;
#if DEBUGGING_FUNCTIONS
      if (WatchHandlers)
        {
         TraceHandler(WTRACE,CurrentCore,BEGIN_TRACE);
         if (CheckHandlerArgCount())
           EvaluateExpression(((HANDLER *) CurrentCore->hnd)->actions,&temp);
         TraceHandler(WTRACE,CurrentCore,END_TRACE);
        }
      else
#endif
      if (CheckHandlerArgCount())
        EvaluateExpression(((HANDLER *) CurrentCore->hnd)->actions,&temp);
      ReturnFlag = CLIPS_FALSE;
      if (BindList != NULL)
        {
         ReturnValues(BindList);
         BindList = NULL;
        }
      if ((NextInCore == NULL) || HaltExecution)
        {
         NextInCore = oldNext;
         CurrentCore = oldCurrent;
         BindList = oldBindList;
         return;
        }
     }
   if (NextInCore->type == MPRIMARY)
     {
      CurrentCore = NextInCore;
      NextInCore = NextInCore->nxt;
#if DEBUGGING_FUNCTIONS
      if (WatchHandlers)
        {
         TraceHandler(WTRACE,CurrentCore,BEGIN_TRACE);
         if (CurrentCore->implicit == 1)
           PerformImplicitHandler(result);
         else if (CheckHandlerArgCount())
           EvaluateExpression(((HANDLER *) CurrentCore->hnd)->actions,result);
         TraceHandler(WTRACE,CurrentCore,END_TRACE);
        }
      else
#endif
      if (CurrentCore->implicit == 1)
        PerformImplicitHandler(result);
      else if (CheckHandlerArgCount())
        EvaluateExpression(((HANDLER *) CurrentCore->hnd)->actions,result);
      ReturnFlag = CLIPS_FALSE;
      if (BindList != NULL)
        {
         ReturnValues(BindList);
         BindList = NULL;
        }
      if ((NextInCore == NULL) || HaltExecution)
        {
         NextInCore = oldNext;
         CurrentCore = oldCurrent;
         BindList = oldBindList;
         return;
        }
      while (NextInCore->type == MPRIMARY)
        {
         NextInCore = NextInCore->nxt;
         if (NextInCore == NULL)
           {
            NextInCore = oldNext;
            CurrentCore = oldCurrent;
            BindList = oldBindList;
            return;
           }
        }
     }
   while (NextInCore->type == MAFTER)
     {
      CurrentCore = NextInCore;
      NextInCore = NextInCore->nxt;
#if DEBUGGING_FUNCTIONS
      if (WatchHandlers)
        {
         TraceHandler(WTRACE,CurrentCore,BEGIN_TRACE);
         if (CheckHandlerArgCount())
           EvaluateExpression(((HANDLER *) CurrentCore->hnd)->actions,&temp);
         TraceHandler(WTRACE,CurrentCore,END_TRACE);
        }
      else
#endif
      if (CheckHandlerArgCount())
        EvaluateExpression(((HANDLER *) CurrentCore->hnd)->actions,&temp);
      ReturnFlag = CLIPS_FALSE;
      if (BindList != NULL)
        {
         ReturnValues(BindList);
         BindList = NULL;
        }
      if ((NextInCore == NULL) || HaltExecution)
        {
         NextInCore = oldNext;
         CurrentCore = oldCurrent;
         BindList = oldBindList;
         return;
        }
     }
   BindList = oldBindList;
   NextInCore = oldNext;
   CurrentCore = oldCurrent;
  }
  
/**************************************************************
  NAME         : PerformImplicitHandler
  DESCRIPTION  : Used to call implicit primary handlers,
                   such as get-<slot> and put-<slot>
                 These are implicit in that they are effective
                   for all slots of a class, but they are not
                   actually defined and thus part of the core
                 These implicit primaries are only called in
                   the case where there are no other primary
                   handlers to call.
  INPUTS       : 1) The caller's value result buffer
  RETURNS      : Nothing useful
  SIDE EFFECTS : Evaluation of the handlers
  NOTES        : Uses global CurrentCore
 *************************************************************/
static VOID PerformImplicitHandler(result)
  DATA_OBJECT *result;
  {
   INSTANCE_TYPE *ins;
   INSTANCE_SLOT *sp;
   DATA_OBJECT temp;
   
   ins = (INSTANCE_TYPE *) CurrentMessageFrame->value;
   sp = (INSTANCE_SLOT *) CurrentCore->hnd;
   if (CurrentCore->get == 1)
     {
      if (CurrentMessageSize != 1)
        {
         SetEvaluationError(TRUE);
         PrintCLIPS(WERROR,"Message-handler ");
         PrintCLIPS(WERROR,GSM_PREFIX);
         PrintCLIPS(WERROR,ValueToString(sp->desc->name));
         PrintCLIPS(WERROR," primary in class ");
         PrintCLIPS(WERROR,ValueToString(ins->cls->name));
         PrintCLIPS(WERROR," expected exactly 0 argument(s).\n");
        }
      else if (*(sp->valaddr) != NULL)
        CopyDataObject(result,*(sp->valaddr));
      else
        {
         PrintCLIPS(WERROR,ValueToString(sp->desc->name));
         PrintCLIPS(WERROR," slot unbound in instance ");
         PrintCLIPS(WERROR,ValueToString(ins->name));
         PrintCLIPS(WERROR,".\n");
         SetEvaluationError(TRUE);
        }
     }
   else
     {
      if ((sp->desc->nowrite == 0) ? FALSE :
          ((sp->desc->initonly == 0) || (WithinInit == FALSE)))
        {
         if (OverrideSlotProtection == FALSE)
           {
            PrintCLIPS(WERROR,ValueToString(sp->desc->name));
            PrintCLIPS(WERROR," slot in instance ");
            PrintCLIPS(WERROR,ValueToString(ins->name));
            PrintCLIPS(WERROR,": write access denied.\n");
            SetEvaluationError(TRUE);
            return;
           }
        }
      if (CurrentMessageSize > 2)
        {
         GroupHandlerWildargs(&temp,2);
         if ((ValidSlotValue(&temp,ValueToString(CurrentMessageName)) == FALSE) ? FALSE :
             (PutSlotValue(ins,sp,&temp,TRUE) != SLOT_ERROR))
           {
            result->type = SYMBOL;
            result->value = (VOID *) CLIPSTrueSymbol;
           }
        }
      else if (CurrentMessageSize == 2)
        {
         if ((ValidSlotValue(&CurrentMessageFrame[1],
                             ValueToString(CurrentMessageName)) == FALSE) ? FALSE :
             (PutSlotValue(ins,sp,&CurrentMessageFrame[1],TRUE) != SLOT_ERROR))
           {
            result->type = SYMBOL;
            result->value = (VOID *) CLIPSTrueSymbol;
           }
        }
      else if (PutSlotValue(ins,sp,NULL,TRUE) != SLOT_ERROR)
        {
         result->type = SYMBOL;
         result->value = (VOID *) CLIPSTrueSymbol;
        }
     }
  }
  
/***************************************************************
  NAME         : CheckHandlerArgCount
  DESCRIPTION  : Verifies that the current argument
                   list satisfies the current
                   handler's parameter count restriction
  INPUTS       : None
  RETURNS      : TRUE if all OK, FALSE otherwise
  SIDE EFFECTS : EvaluationError set on errors
  NOTES        : Uses CurrentMessageSize and CurrentCore globals
 ***************************************************************/
static int CheckHandlerArgCount()
  {
   HANDLER *hnd;
   
   hnd = (HANDLER *) CurrentCore->hnd;
   if ((hnd->maxParams == -1) ? (CurrentMessageSize < hnd->minParams) :
       (CurrentMessageSize != hnd->minParams))
     {
      SetEvaluationError(TRUE);
      PrintCLIPS(WERROR,"Message-handler ");
      PrintCLIPS(WERROR,ValueToString(hnd->name));
      PrintCLIPS(WERROR," ");
      PrintCLIPS(WERROR,hndquals[hnd->type]);
      PrintCLIPS(WERROR," in class ");
      PrintCLIPS(WERROR,ValueToString(hnd->cls->name));
      PrintCLIPS(WERROR," expected ");
      PrintCLIPS(WERROR,(hnd->maxParams == -1) ? "at least " : "exactly ");
      PrintLongInteger(WERROR,(long) (hnd->minParams-1));
      PrintCLIPS(WERROR," argument(s).\n");
      return(FALSE);
     }
   return(TRUE);
  }

#endif

/***************************************************
  NAME         : 
  DESCRIPTION  : 
  INPUTS       : 
  RETURNS      : 
  SIDE EFFECTS : 
  NOTES        : 
 ***************************************************/
