   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                 SPECIAL FORM MODULE                 */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _SPCLFORM_SOURCE_

#include "setup.h"

#if (! RUN_TIME)

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "constant.h"
#include "scanner.h"
#include "expressn.h"
#include "evaluatn.h"
#include "spclform.h"
#include "clipsmem.h"
#include "symbol.h"
#include "router.h"
#include "utility.h"
#include "generate.h"
#include "parsutil.h"

#if DEFGLOBAL_CONSTRUCT
#include "defglobl.h"
#endif

#if ANSI_COMPILER  
   static struct expr            *AssertParse(struct expr *,char *);
   static struct expr            *RetractParse(struct expr *,char *);
   static struct expr            *WhileParse(struct expr *,char *);
   static struct expr            *IfParse(struct expr *,char *);
   static struct expr            *BindParse(struct expr *,char *);
   static VOID                    AddBindName(struct symbolHashNode *);
   static struct expr            *NotParse(struct expr *,char *);
   static struct expr            *MultiArgNumericParse(struct expr *,char *);
   static struct expr            *MultiArgParse(struct expr *,char *);
   static struct expr            *CheckArgListParse(struct expr *,char *,int,int,int);
#else
   static struct expr            *AssertParse();
   static struct expr            *RetractParse();
   static struct expr            *WhileParse();
   static struct expr            *IfParse();
   static struct expr            *BindParse();
   static VOID                    AddBindName();
   static struct expr            *NotParse();
   static struct expr            *MultiArgNumericParse();
   static struct expr            *MultiArgParse();
   static struct expr            *CheckArgListParse();
#endif

/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/
  
   static struct BindInfo *ListOfParsedBindNames;
   
/*****************************************************************/
/* InitializeSpecialForms:                                       */
/*****************************************************************/
globle VOID InitializeSpecialForms()
  {
   AddFunctionParser("assert",AssertParse);
   AddFunctionParser("bind",BindParse);
   AddFunctionParser("if",IfParse);
   AddFunctionParser("while",WhileParse);
   AddFunctionParser("retract",RetractParse);
   
   AddFunctionParser("+",MultiArgNumericParse);
   AddFunctionParser("*",MultiArgNumericParse);
   AddFunctionParser("-",MultiArgNumericParse);
   AddFunctionParser("/",MultiArgNumericParse);
   AddFunctionParser("div",MultiArgNumericParse);
   AddFunctionParser("<=",MultiArgNumericParse);
   AddFunctionParser(">=",MultiArgNumericParse);
   AddFunctionParser("<",MultiArgNumericParse);
   AddFunctionParser(">",MultiArgNumericParse);
   AddFunctionParser("=",MultiArgNumericParse);
   AddFunctionParser("!=",MultiArgNumericParse);
   AddFunctionParser("<>",MultiArgNumericParse);
   AddFunctionParser("min",MultiArgNumericParse);
   AddFunctionParser("max",MultiArgNumericParse);
   
   AddFunctionParser("and",MultiArgParse);
   AddFunctionParser("or",MultiArgParse);
   AddFunctionParser("not",NotParse);
   AddFunctionParser("eq",MultiArgParse);
   AddFunctionParser("neq",MultiArgParse);
  }

/**************************************************************/
/* AssertParse: Purpose is to parse the assert statement. The */
/*   parse of the statement is the return value.              */
/*   Assert syntax:                                           */
/*     (assert (<field>+))                                    */
/**************************************************************/
static struct expr *AssertParse(top,in_file)
  struct expr *top;
  char *in_file;
  {
   int error;
   struct expr *rv;
   
   SavePPBuffer(" ");
   IncrementIndentDepth(8);
   rv = BuildRHSAssert(top,in_file,CLIPS_TRUE,&error);
   DecrementIndentDepth(8);
   return(rv);
  }

/*************************************************************/
/* RetractParse: Purpose is to parse the retract statement. */
/*   The parse of the statement is the return value.         */
/*   Syntax: (retract <?-var>+ )                             */
/*************************************************************/
static struct expr *RetractParse(top,in_file)
  struct expr *top;
  char *in_file;
  {
   struct token act_tkn;
   struct expr *last_one, *next_one = NULL;
   int error_flag = CLIPS_FALSE;
   int arg_num = 1;

   last_one = NULL;

   SavePPBuffer(" ");

   GetToken(in_file,&act_tkn);
   while (act_tkn.type != RPAREN)
     {
      if (act_tkn.type == INTEGER)
        { 
         if (ValueToLong(act_tkn.value) < 0)
           {
            ExpectedTypeError("retract",arg_num,"fact-address or fact-index");
            ReturnExpression(top); 
            return(NULL);
           }
         next_one = GenConstant(INTEGER,act_tkn.value); 
        }
      else if ((act_tkn.type == BWORD) || (act_tkn.type == GBWORD))
        { next_one = GenConstant(act_tkn.type,act_tkn.value); }
      else if ((act_tkn.type == SYMBOL) ? 
               (strcmp(ValueToString(act_tkn.value),"*") == 0) : CLIPS_FALSE)
        {
         next_one = GenConstant(INTEGER,AddLong(-1L)); 
        }
      else
        { error_flag = CLIPS_TRUE; }

      if (error_flag == CLIPS_TRUE)
        {
         ExpectedTypeError("retract",arg_num,"non-negative integer or variable");
         ReturnExpression(top); 
         return(NULL);
        }

      if (last_one == NULL)
        { top->arg_list = next_one; }
      else
        { last_one->next_arg = next_one; }

      last_one = next_one;

      GetToken(in_file,&act_tkn);

      if (act_tkn.type != RPAREN) 
        {
         PPBackup();
         SavePPBuffer(" ");
         SavePPBuffer(act_tkn.print_rep);
        }
        
      arg_num++;
     }
     
    if (top->arg_list == NULL)
     {
      ExpectedCountError("retract",AT_LEAST,1);
      ReturnExpression(top);
      return(NULL);
     }

   return(top);
  }

/*********************************************************/
/* WhileParse: purpose is to parse the while statement. */
/*   The parse of the statement is the return value.     */
/*   Syntax: (while <expression> <action>+)              */
/*********************************************************/
static struct expr *WhileParse(parse,infile)
  struct expr *parse;
  char *infile;
  {
   struct token act_tkn;
   int read_first_paren;
   
   /*===============================*/
   /* Process the while expression. */
   /*===============================*/

   SavePPBuffer(" ");   

   parse->arg_list = ParseAtomOrExpression(infile);
   if (parse->arg_list == NULL) 
     { 
      ReturnExpression(parse);
      return(NULL);
     }

   /*====================================*/
   /* Process the do keyword if present. */
   /*====================================*/
   
   GetToken(infile,&act_tkn);
   if ((act_tkn.type == SYMBOL) && (strcmp(ValueToString(act_tkn.value),"do") == 0))
     { 
      read_first_paren = CLIPS_TRUE; 
      PPBackup(); 
      SavePPBuffer(" "); 
      SavePPBuffer(act_tkn.print_rep); 
      IncrementIndentDepth(3);
      PPCRAndIndent();
     }
   else if (act_tkn.type == LPAREN)
     { 
      read_first_paren = CLIPS_FALSE;
      PPBackup(); 
      IncrementIndentDepth(3);
      PPCRAndIndent();
      SavePPBuffer(act_tkn.print_rep); 
     }
   else  
     {
      SyntaxErrorMessage("while function");
      ReturnExpression(parse);
      return(NULL); 
     }
     
   /*============================*/
   /* Process the while actions. */
   /*============================*/

   parse->arg_list->next_arg = 
                 GroupActions(infile,&act_tkn,read_first_paren,NULL);

   if (parse->arg_list->next_arg == NULL) 
     {
      ReturnExpression(parse); 
      return(NULL);
     } 
     
   PPBackup();
   PPBackup();
   SavePPBuffer(act_tkn.print_rep);

   /*======================================================*/
   /* Check for the closing right parenthesis of the while. */
   /*======================================================*/

   if (act_tkn.type != RPAREN)
     {
      SyntaxErrorMessage("while function");
      ReturnExpression(parse);
      return(NULL);
     }

   DecrementIndentDepth(3);
   
   return(parse);
  }

/*********************************************************/
/* IfParse: purpose is to parse the if statement.  The  */
/*   parse of the statement is the return value.         */
/*   Syntax: (if <expression> then <action>+             */
/*               [ else <action>+ ] )                    */
/*********************************************************/
static struct expr *IfParse(top,infile)
  struct expr *top;
  char *infile;
  {   
   struct token act_tkn;
   
   /*============================*/
   /* Process the if expression. */
   /*============================*/

   SavePPBuffer(" "); 

   top->arg_list = ParseAtomOrExpression(infile);
   
   if (top->arg_list == NULL)
     {
      ReturnExpression(top);
      return(NULL);
     }

   /*========================================*/
   /* Keyword 'then' must follow expression. */
   /*========================================*/

   IncrementIndentDepth(3);
   PPCRAndIndent();

   GetToken(infile,&act_tkn);
   if ((act_tkn.type != SYMBOL) || (strcmp(ValueToString(act_tkn.value),"then") != 0))
     { 
      SyntaxErrorMessage("if function");
      ReturnExpression(top);
      return(NULL);
     }

   /*==============================*/
   /* Process the if then actions. */
   /*==============================*/
   
   PPCRAndIndent();
   top->arg_list->next_arg = GroupActions(infile,&act_tkn,CLIPS_TRUE,"else");
   
   if (top->arg_list->next_arg == NULL) 
     {
      ReturnExpression(top);
      return(NULL);
     }

   /*===========================================*/
   /* A ')' signals an if then without an else. */
   /*===========================================*/

   if (act_tkn.type == RPAREN) 
     {
      DecrementIndentDepth(3);
      PPBackup();
      PPBackup();
      SavePPBuffer(act_tkn.print_rep);
      return(top);
     }

   /*=============================================*/
   /* Keyword 'else' must follow if then actions. */
   /*=============================================*/

   if ((act_tkn.type != SYMBOL) || (strcmp(ValueToString(act_tkn.value),"else") != 0))
     { 
      SyntaxErrorMessage("if function");
      ReturnExpression(top);
      return(NULL);
     }

   /*==============================*/
   /* Process the if else actions. */
   /*==============================*/

   PPCRAndIndent();
   top->arg_list->next_arg->next_arg = GroupActions(infile,&act_tkn,CLIPS_TRUE,NULL);
      
   if (top->arg_list->next_arg->next_arg == NULL) 
     {
      ReturnExpression(top);
      return(NULL);
     }
 
   /*======================================================*/
   /* Check for the closing right parenthesis of the if. */
   /*======================================================*/

   if (act_tkn.type != RPAREN)
     {
      SyntaxErrorMessage("if function");
      ReturnExpression(top);
      return(NULL);
     }

   /*===========================================*/
   /* A ')' signals an if then without an else. */
   /*===========================================*/

   PPBackup();
   PPBackup();
   SavePPBuffer(")");   
   DecrementIndentDepth(3);  
   return(top);
  }
  
/***********************************************************/
/* BindParse: purpose is to parse the bind statement. The */
/*   parse of the statement is the return value.           */
/*   Syntax:  (bind ?var <expression>)                     */
/***********************************************************/
static struct expr *BindParse(top,infile)
  struct expr *top;
  char *infile;
  {
   struct token act_tkn;
   SYMBOL_HN *var_name;
   int error_flag;
#if DEFGLOBAL_CONSTRUCT
   int pos;
#endif

   SavePPBuffer(" "); 

   /*=============================================*/
   /* Next token must be the name of the variable */
   /* to be bound.                                */
   /*=============================================*/ 

   GetToken(infile,&act_tkn);
   if ((act_tkn.type != BWORD) && (act_tkn.type != BWORDS) && (act_tkn.type != GBWORD))
     {
      SyntaxErrorMessage("bind function");
      ReturnExpression(top);
      return(NULL);
     }

   SavePPBuffer(" "); 
   
   /*==============================*/
   /* Process the bind expression. */
   /*==============================*/

   top->arg_list = GenConstant(SYMBOL,act_tkn.value);
   var_name = (SYMBOL_HN *) act_tkn.value;
   
#if DEFGLOBAL_CONSTRUCT
   if ((act_tkn.type == GBWORD) ? ((pos = QFindDefglobal(var_name)) >= 0) : CLIPS_FALSE)
     {
      top->arg_list->type = INTEGER;
      top->arg_list->value = (VOID *) AddLong((long) pos);
     }
   else if (act_tkn.type == GBWORD)
     {
      PrintCLIPS(WERROR,"\nGlobal variable ?");
      PrintCLIPS(WERROR,ValueToString(var_name));
      PrintCLIPS(WERROR," was referenced, but is not defined\n");  
      ReturnExpression(top); 
      return(NULL);
     }
#endif

   error_flag = CLIPS_FALSE;
   top->arg_list->next_arg = ArgumentParse(infile,&error_flag);
   if (error_flag == CLIPS_TRUE) 
     {
      ReturnExpression(top); 
      return(NULL);
     }
     
   if (top->arg_list->next_arg == NULL)
     {
      SyntaxErrorMessage("bind function");
      ReturnExpression(top); 
      return(NULL);
     }
   
   /*===========================================*/
   /* Bind construct must be closed with a ')'. */
   /*===========================================*/

   GetToken(infile,&act_tkn);
   if (act_tkn.type != RPAREN)
     {  
      SyntaxErrorMessage("bind function");
      ReturnExpression(top);
      return(NULL);
     }
     
#if DEFGLOBAL_CONSTRUCT
   if (top->arg_list->type == INTEGER) return(top);
#endif

   AddBindName(var_name);
   
   return(top);
  }
  
  
/********************************************************/
/* SearchParsedBindNames:                                      */
/********************************************************/
globle BOOLEAN SearchParsedBindNames(name_sought)
  SYMBOL_HN *name_sought;
  {
   struct BindInfo *var_ptr;
   
   var_ptr = ListOfParsedBindNames;
   while (var_ptr != NULL)
     {
      if (var_ptr->name == name_sought)
        { return(CLIPS_TRUE); }
      var_ptr = var_ptr->next;
     }

   return(CLIPS_FALSE);
  }

/********************************************************/
/* GetParsedBindNames:                                      */
/********************************************************/
globle struct BindInfo *GetParsedBindNames()
  {
   return(ListOfParsedBindNames);
  }
  
/********************************************************/
/* SetParsedBindNames:                                      */
/********************************************************/
globle VOID SetParsedBindNames(newValue)
  struct BindInfo *newValue;
  {
   ListOfParsedBindNames = newValue;
  }
  
/********************************************************/
/* AddBindName:                                      */
/********************************************************/
static VOID AddBindName(new_name)
  SYMBOL_HN *new_name;
  {
   struct BindInfo *new_bind;
   
   if (SearchParsedBindNames(new_name) == CLIPS_TRUE) return;
   
   new_bind = get_struct(BindInfo);
   new_bind->name = new_name;
   new_bind->next = ListOfParsedBindNames;
   ListOfParsedBindNames = new_bind;
  }

/********************************************************/
/* ClearParsedBindNames:                                     */
/********************************************************/
globle VOID ClearParsedBindNames()
  {
   struct BindInfo *temp_bind;
   
   while (ListOfParsedBindNames != NULL)
     {
      temp_bind = ListOfParsedBindNames->next;
      rtn_struct(BindInfo,ListOfParsedBindNames);
      ListOfParsedBindNames = temp_bind;
     }
  }

/********************************************************/
/* ParsedBindNamesEmpty:                                     */
/********************************************************/
globle BOOLEAN ParsedBindNamesEmpty()
  {  
   if (ListOfParsedBindNames != NULL) return(CLIPS_FALSE);
   
   return(CLIPS_TRUE);
  }
  
/*****************************************************************/
/* MultiArgNumericParse: purpose is to parse statements that require  */
/*   at least two arguments, but may have more. Examples include */
/*   addition, multiplication, etc.                              */
/*****************************************************************/
static struct expr *MultiArgNumericParse(parse,infile)
  struct expr *parse;
  char *infile;
  {
   return(CheckArgListParse(parse,infile,AT_LEAST,2,CLIPS_TRUE));
  }
  
/*****************************************************************/
/* MultiArgParse: purpose is to parse statements that require  */
/*   at least two arguments, but may have more. Examples include */
/*   addition, multiplication, etc.                              */
/*****************************************************************/
static struct expr *MultiArgParse(parse,infile)
  struct expr *parse;
  char *infile;
  {
   return(CheckArgListParse(parse,infile,AT_LEAST,2,CLIPS_FALSE));
  }

/*****************************************************************/
/* NotParse:                                                    */
/*****************************************************************/
static struct expr *NotParse(parse,infile)
  struct expr *parse;
  char *infile;
  {
   return(CheckArgListParse(parse,infile,EXACTLY,1,CLIPS_FALSE));
  }

/********************************************************************/
/* CheckArgListParse: Purpose is to parse statements that only a */
/*   specific number of arguments and/or only numeric arguments.    */
/********************************************************************/
static struct expr *CheckArgListParse(parse,infile,check_val,exp_num,
                                        must_be_numeric)
  struct expr *parse;
  char *infile;
  int check_val, exp_num, must_be_numeric;
  {
   struct expr *arg_ptr;
   int error, count, numa;

   /*========================*/
   /* Collect the arguments. */
   /*========================*/

   if (CollectArguments(parse,infile) == NULL) return(NULL);

   /*======================*/
   /* Count the arguments. */
   /*======================*/

   arg_ptr = parse->arg_list;              
   numa = 0;
   while (arg_ptr != NULL)
     {
      numa++;
      arg_ptr = arg_ptr->next_arg;
     }

   /*=======================================*/
   /* Check for proper number of arguments. */
   /*=======================================*/

   error = 0; 
   if (check_val == EXACTLY)
     { if (numa != exp_num) error = 1; }
   else if (check_val == AT_LEAST)
     { if (numa < exp_num) error = 1; }
   else if (check_val == NO_MORE_THAN)
     { if (numa > exp_num) error = 1; }

   if (error)                               
     {
      ExpectedCountError(ValueToString(ExpressionFunctionCallName(parse)),check_val,exp_num);
      ReturnExpression(parse);
      return(NULL);
     }

   /*==============================*/
   /* Check for numeric arguments. */
   /*==============================*/

   if (must_be_numeric == CLIPS_FALSE) 
     { return(parse); }

   arg_ptr = parse->arg_list;                   
   count = 1;
   while (arg_ptr != NULL)
     {
      if (arg_ptr->type == FCALL)
        {
         if ((must_be_numeric == CLIPS_TRUE) &&
             (ExpressionFunctionType(arg_ptr) != 'f') &&
             (ExpressionFunctionType(arg_ptr) != 'd') &&
             (ExpressionFunctionType(arg_ptr) != 'i') &&
             (ExpressionFunctionType(arg_ptr) != 'l') &&
             (ExpressionFunctionType(arg_ptr) != 'u'))
           {
            ExpectedTypeError(ValueToString(ExpressionFunctionCallName(parse)),count,
                           "number or numeric function");
            ReturnExpression(parse);
            return(NULL);
           }
        }
      else if ((arg_ptr->type != FLOAT) && 
               (arg_ptr->type != INTEGER) &&
#if DEFGENERIC_CONSTRUCT
               (arg_ptr->type != GCALL) &&
#endif
#if DEFFUNCTION_CONSTRUCT
               (arg_ptr->type != PCALL) &&
#endif
               (arg_ptr->type != BWORD) &&
               (arg_ptr->type != GBWORD))
        {
         if (must_be_numeric == CLIPS_TRUE)
           {
            ExpectedTypeError(ValueToString(ExpressionFunctionCallName(parse)),count,
                              "number or numeric function");
            ReturnExpression(parse);
            return(NULL);
           }
        }
      arg_ptr = arg_ptr->next_arg;
      count++;
     }
   
   return(parse);
  }


#endif

  
