   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                   UTILITY MODULE                    */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*      Provides a set of utility functions useful to other  */
/*      modules.                                             */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
 
#define _UTILITY_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>
#include <math.h>

#include "setup.h"

#include "utility.h"
#include "clipsmem.h"
#include "constant.h"
#include "evaluatn.h"
#include "facts.h"
#include "symbol.h"
#include "scanner.h"
#include "router.h"

#if OBJECT_SYSTEM
#include "extobj.h"
#endif

#define MAX_EPHEMERAL_COUNT 1000L
#define MAX_EPHEMERAL_SIZE 10240L
#define COUNT_INCREMENT 1000L
#define SIZE_INCREMENT 10240L

struct cleanup_func
  {
   char *name;  
#if ANSI_COMPILER
   VOID (*ip)(void);
#else
   VOID (*ip)();
#endif
   int priority;
   struct cleanup_func *next;
  };
  
struct watchItem
  {
   char *name;  
   int *flag;
   int temp;
   int priority;
   struct watchItem *next;
  };
  
/***************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/***************************************/

#if ANSI_COMPILER
   static BOOLEAN                 AddCPFunction(char *,VOID (*)(),int,struct cleanup_func **);
   static BOOLEAN                 RemoveCPFunction(char *,struct cleanup_func **);
#else
   static BOOLEAN                 AddCPFunction();
   static BOOLEAN                 RemoveCPFunction();
#endif
    
/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/ 

   static struct cleanup_func   *ListOfCleanupFunctions = NULL;
   static struct cleanup_func   *ListOfPeriodicFunctions = NULL;
   static struct watchItem      *ListOfWatchItems = NULL;
   
/****************************************/
/* GLOBAL INTERNAL VARIABLE DEFINITIONS */
/****************************************/ 
    
   globle unsigned long          EphemeralItemCount = 0;
   globle unsigned long          EphemeralItemSize = 0;
   globle unsigned long          CurrentEphemeralCountMax = MAX_EPHEMERAL_COUNT;
   globle unsigned long          CurrentEphemeralSizeMax = MAX_EPHEMERAL_SIZE;
   globle BOOLEAN                PreserveEscapedCharacters = CLIPS_FALSE;
   globle BOOLEAN                AddressesToStrings = CLIPS_FALSE;

/************************************************************/
/* PrintFloat:  Controls printout of floating point numbers. */
/************************************************************/
globle VOID PrintFloat(fileid,number)
  char *fileid;
  double number;
  {
   char *new_str;

   new_str = FloatToString(number);
   PrintCLIPS(fileid,new_str);
  }
  
/***************************************************/
/* PrintLongInteger:  Controls printout of integers. */
/***************************************************/
globle VOID PrintLongInteger(fileid,number)
  char *fileid;
  long int number;
  {
   char printBuffer[32];
   
   sprintf(printBuffer,"%ld",number);
   PrintCLIPS(fileid,printBuffer);
  }
  
/***********************************************************/
/* FloatToString:  Converts number to CLIPS string format. */  
/***********************************************************/
globle char *FloatToString(number)
  double number;
  {
   char floatString[40];
   int i;
   char x;
   VOID *thePtr;
   
   sprintf(floatString,"%0.16g",number);
   
   for (i = 0; (x = floatString[i]) != '\0'; i++)
     { 
      if ((x == '.') || (x == 'e')) 
        {
         thePtr = AddSymbol(floatString);
         return(ValueToString(thePtr));
        }
     }
     
   strcat(floatString,".0");
   
   thePtr = AddSymbol(floatString);
   return(ValueToString(thePtr));
  }

/***********************************************************************/
/* LongIntegerToString:  Converts long integer to CLIPS string format. */  
/***********************************************************************/
globle char *LongIntegerToString(number)
  long number;
  {
   char buffer[30];
   VOID *thePtr;

   sprintf(buffer,"%ld",number);

   thePtr = AddSymbol(buffer);
   return(ValueToString(thePtr));
  }
  
/**********************************************************/
/* AppendToString:                                        */
/**********************************************************/
globle char *AppendToString(appendStr,oldStr,oldPos,oldMax)
  char *appendStr, *oldStr;
  int *oldPos, *oldMax;
  {
   int length;
   
   length = strlen(appendStr);
   if (length + *oldPos + 1 > *oldMax)
     { 
      oldStr = genrealloc(oldStr,*oldMax,length + *oldPos + 1); 
      *oldMax = length + *oldPos + 1;
     }
     
   if (oldStr == NULL) { return(NULL); }
     
   strcpy(&oldStr[*oldPos],appendStr);
   *oldPos += length;
   
   return(oldStr);
  }
  
/**********************************************************/
/* AppendNToString:                                       */
/**********************************************************/
globle char *AppendNToString(appendStr,oldStr,length,oldPos,oldMax)
  char *appendStr, *oldStr;
  int length;
  int *oldPos, *oldMax;
  {
   int lengthWithEOS;
   
   if (appendStr[length-1] != '\0') lengthWithEOS = length + 1;
   else lengthWithEOS = length;
   
   if (lengthWithEOS + *oldPos > *oldMax)
     { 
      oldStr = genrealloc(oldStr,*oldMax,*oldPos + lengthWithEOS); 
      *oldMax = *oldPos + lengthWithEOS;
     }
     
   if (oldStr == NULL) { return(NULL); }
     
   strncpy(&oldStr[*oldPos],appendStr,length);
   *oldPos += (lengthWithEOS - 1);
   oldStr[*oldPos] = '\0';
   
   return(oldStr);
  }
  
/**********************************************************/
/* ExpandStringWithChar:                                 */
/**********************************************************/
globle char *ExpandStringWithChar(inchar,str,pos,max,newSize)
  int inchar;
  char *str;
  int *max, *pos, newSize;
  {
   if (*pos >= (*max - 1))
     { 
      str = genrealloc(str,*max,newSize);
      *max = newSize;
     }
        
  if (inchar != '\b')
    { 
     str[*pos] = (char) inchar;
     (*pos)++;
     str[*pos] = '\0';
    }
  else
    {
     if (*pos > 0) (*pos)--;
     str[*pos] = '\0';
    }
  
   return(str);
  }

/**************************************************************/
/* GetLogicalName:             */
/**************************************************************/
globle char *GetLogicalName(position,t_str)
  int position;
  char *t_str;
  {
   char *log_id;
   DATA_OBJECT arg_ptr;
   
   RtnUnknown(position,&arg_ptr);

   if ((GetType(arg_ptr) == SYMBOL) || 
       (GetType(arg_ptr) == STRING) ||
       (GetType(arg_ptr) == INSTANCE_NAME))
     {
      log_id = ValueToString(arg_ptr.value);
      if ((strcmp(log_id,"t") == 0) || (strcmp(log_id,"T") == 0))
        { log_id = t_str; }
     }
   else if (GetType(arg_ptr) == FLOAT)
     {
      log_id = ValueToString(AddSymbol(FloatToString(DOToDouble(arg_ptr)))); 
     }
   else if (GetType(arg_ptr) == INTEGER)
     {
      log_id = ValueToString(AddSymbol(LongIntegerToString(DOToLong(arg_ptr)))); 
     }
   else
     { log_id = NULL; }
     
   return(log_id);
  } 
  
/*************************************************/
/* GetFileName:                                  */
/*************************************************/
globle char *GetFileName(fun_name,pos)
  char *fun_name;
  int pos;
  {
   DATA_OBJECT arg_ptr;

   RtnUnknown(pos,&arg_ptr);
   if ((GetType(arg_ptr) != STRING) && (GetType(arg_ptr) != SYMBOL))
     {
      ExpectedTypeError(fun_name,pos,"file name");
      return(NULL);
     }
    
   return(DOToString(arg_ptr));
  }

/*************************************************************/
/* GetConstructName:  For functions requiring a construct    */
/*   name as an argument, this function checks for the right */
/*   number of arguments and checks to see that the argument */
/*   that is the name is of type symbol.                     */
/*************************************************************/
globle char *GetConstructName(exp_args,name_pos,fun_name,type)
  int exp_args, name_pos;
  char *fun_name, *type;
  {
   int num_a; 
   DATA_OBJECT arg_ptr;

   num_a = RtnArgCount();
   if (num_a != exp_args)
     {
      ExpectedCountError(fun_name,EXACTLY,exp_args);
      return(NULL);
     }

   RtnUnknown(name_pos,&arg_ptr);

   if (GetType(arg_ptr) != SYMBOL)
     { 
      ExpectedTypeError(fun_name,name_pos,type);
      return(NULL);
     }

   return(DOToString(arg_ptr));
  } 

/***********************************************************/
/* ExpectedTypeError:                               */
/***********************************************************/
globle VOID ExpectedTypeError(fun_name,arg_num,exp_type)
  char *fun_name;
  int arg_num;
  char *exp_type;
  {
   PrintCLIPS(WERROR,"ERROR: Function ");
   PrintCLIPS(WERROR,fun_name);
   PrintCLIPS(WERROR," expected argument #");
   PrintLongInteger(WERROR,(long int) arg_num);
   PrintCLIPS(WERROR," to be of type ");
   PrintCLIPS(WERROR,exp_type);
   PrintCLIPS(WERROR,"\n");
  }

/***********************************************************/
/* SyntaxErrorMessage:                                     */
/***********************************************************/
globle VOID SyntaxErrorMessage(location)
  char *location;
  {
   PrintCLIPS(WERROR,"\nSyntax Error");
   if (location != NULL)
     {
      PrintCLIPS(WERROR,":  Check appropriate syntax for ");
      PrintCLIPS(WERROR,location);
     }
     
   PrintCLIPS(WERROR,"\n");
   SetEvaluationError(CLIPS_TRUE);
  }

/***********************************************************/
/* ExpectedCountError:                               */
/***********************************************************/
globle VOID ExpectedCountError(fun_name,check_val,exp_num)
  char *fun_name;
  int check_val, exp_num; 
  {
   PrintCLIPS(WERROR,"ERROR: ");
   PrintCLIPS(WERROR,"Function ");
   PrintCLIPS(WERROR,fun_name);
      
   if (check_val == EXACTLY)
     { PrintCLIPS(WERROR," expected exactly "); }
   else if (check_val == AT_LEAST)
     { PrintCLIPS(WERROR," expected at least "); }
   else if (check_val == NO_MORE_THAN)
     { PrintCLIPS(WERROR," expected no more than "); }

   PrintLongInteger(WERROR,(long int) exp_num);
   PrintCLIPS(WERROR," argument(s)\n");
  }
  
/******************************************************************/
/* OpenErrorMessage: Generalized error message for opening files. */ 
/******************************************************************/ 
globle VOID OpenErrorMessage(functionName,fileName)
  char *functionName, *fileName;
  {
   PrintCLIPS(WERROR,"Function ");
   PrintCLIPS(WERROR,functionName);
   PrintCLIPS(WERROR," was unable to open file ");
   PrintCLIPS(WERROR,fileName);
   PrintCLIPS(WERROR,"\n");
  }
   
/**************************************************************************/
/* CLIPSSystemError: Generalized error message for major internal errors. */
/**************************************************************************/
globle VOID CLIPSSystemError(module,errorID)
  char *module;
  int errorID;
  {
   PrintCLIPS(WERROR,"*** CLIPS SYSTEM ERROR ***\n");
   PrintCLIPS(WERROR,"ID = ");
   PrintCLIPS(WERROR,module);
   PrintLongInteger(WERROR,(long int) errorID);
   PrintCLIPS(WERROR,"\n");
   PrintCLIPS(WERROR,"CLIPS data structures are in an inconsistent or corrupted state.\n");
   PrintCLIPS(WERROR,"This error may have occurred from errors in user defined code.\n");
   PrintCLIPS(WERROR,"**************************\n");  
  }

/*************************************************************/
/* PeriodicCleanup: Returns garbage created during execution */
/*   that has not been returned to the memory pool yet. The  */
/*   cleanup is normally deferred so that an executing rule  */
/*   can still access these data structures.                 */
/*                                                           */
/* Always calls a series of functions that should be called  */
/*   periodically.  Usually used by interfaces to update     */
/*   displays.                                               */
/*************************************************************/
globle VOID PeriodicCleanup(cleanupAllDepths,useHeuristics)
  BOOLEAN cleanupAllDepths;
  BOOLEAN useHeuristics;
  {
   int oldDepth = -1;
   struct cleanup_func *cleanupPtr,*periodPtr;
   static int lastEvaluationDepth = -1;
   
   periodPtr = ListOfPeriodicFunctions;
   while (periodPtr != NULL)
     {
      (*periodPtr->ip)();
      periodPtr = periodPtr->next;
     }
   
   if (lastEvaluationDepth > CurrentEvaluationDepth)
     { 
      lastEvaluationDepth = CurrentEvaluationDepth;
      CurrentEphemeralCountMax = MAX_EPHEMERAL_COUNT;
      CurrentEphemeralSizeMax = MAX_EPHEMERAL_SIZE;
     }
     
   if (useHeuristics &&
       (EphemeralItemCount < CurrentEphemeralCountMax) && 
       (EphemeralItemSize < CurrentEphemeralSizeMax)) 
     { return; }
     
   if (cleanupAllDepths)
     {
      oldDepth = CurrentEvaluationDepth;
      CurrentEvaluationDepth = -1;
     }

   RemoveOldFacts();
   FlushSegments(); 
   
   cleanupPtr = ListOfCleanupFunctions;
   while (cleanupPtr != NULL)
     {
      (*cleanupPtr->ip)();
      cleanupPtr = cleanupPtr->next;
     }
   
   RemoveEphemeralAtoms();
   
   if (cleanupAllDepths) CurrentEvaluationDepth = oldDepth;
   
   if ((EphemeralItemCount + COUNT_INCREMENT) > CurrentEphemeralCountMax)
     { CurrentEphemeralCountMax = EphemeralItemCount + COUNT_INCREMENT; }
     
   if ((EphemeralItemSize + SIZE_INCREMENT) > CurrentEphemeralSizeMax)
     { CurrentEphemeralSizeMax = EphemeralItemSize + SIZE_INCREMENT; }
     
   /* Memory Bug Fix. */
   lastEvaluationDepth = CurrentEvaluationDepth;
  }
  
/*************************/
/* AddCleanupFunction:   */
/*************************/
globle BOOLEAN AddCleanupFunction(name,func_ptr,priority)
  char *name;
  VOID (*func_ptr)();
  int priority;
  {
   return(AddCPFunction(name,func_ptr,priority,&ListOfCleanupFunctions));
  }

/*************************/
/* AddPeriodicFunction:   */
/*************************/
globle BOOLEAN AddPeriodicFunction(name,func_ptr,priority)
  char *name;
  VOID (*func_ptr)();
  int priority;
  {
   return(AddCPFunction(name,func_ptr,priority,&ListOfPeriodicFunctions));
  }

/*************************/
/* AddCPFunction:   */
/*************************/
static BOOLEAN AddCPFunction(name,func_ptr,priority,head)
  char *name;
  VOID (*func_ptr)();
  int priority;
  struct cleanup_func **head;
  {
   struct cleanup_func *newPtr, *currentPtr, *lastPtr = NULL;

   newPtr = get_struct(cleanup_func);

   newPtr->name = name;
   newPtr->ip = func_ptr;
   newPtr->priority = priority;
   
   if (*head == NULL)
     {
      newPtr->next = NULL;
      *head = newPtr;
      return(1);
     }
     
   currentPtr = *head;
   while ((currentPtr != NULL) ? (priority < currentPtr->priority) : CLIPS_FALSE)
     {
      lastPtr = currentPtr; 
      currentPtr = currentPtr->next;
     }

   if (lastPtr == NULL)
     { 
      newPtr->next = *head;
      *head = newPtr;
     }
   else
     {
      newPtr->next = currentPtr;
      lastPtr->next = newPtr;
     }
   
   return(1);
  }
 
/****************************/
/* RemoveCleanupFunction:    */
/****************************/
globle BOOLEAN RemoveCleanupFunction(name)
  char *name;
  {
   return(RemoveCPFunction(name,&ListOfCleanupFunctions));
  }

/****************************/
/* RemovePeriodicFunction:  */
/****************************/
globle BOOLEAN RemovePeriodicFunction(name)
  char *name;
  {
   return(RemoveCPFunction(name,&ListOfPeriodicFunctions));
  }

/****************************/
/* RemoveCPFunction:    */
/****************************/
static BOOLEAN RemoveCPFunction(name,head)
  char *name;
  struct cleanup_func **head;
  {
   struct cleanup_func *currentPtr, *lastPtr;

   lastPtr = NULL;
   currentPtr = *head;
   
   while (currentPtr != NULL)
     {
      if (strcmp(name,currentPtr->name) == 0)
        {
         if (lastPtr == NULL)
           { *head = currentPtr->next; }
         else
           { lastPtr->next = currentPtr->next; }
         rtn_struct(cleanup_func,currentPtr);
         return(CLIPS_TRUE);
        }
      lastPtr = currentPtr;
      currentPtr = currentPtr->next;
     }
     
   return(CLIPS_FALSE);
  }
  
/*************************/
/* AddWatchItem:         */
/*************************/
globle BOOLEAN AddWatchItem(name,flag,priority)
  char *name;
  int *flag;
  int priority;
  {
   struct watchItem *newPtr, *currentPtr, *lastPtr = NULL;

   newPtr = get_struct(watchItem);

   newPtr->name = name;
   newPtr->flag = flag;
   newPtr->priority = priority;
   
   if (ListOfWatchItems == NULL)
     {
      newPtr->next = NULL;
      ListOfWatchItems = newPtr;
      return(1);
     }
     
   currentPtr = ListOfWatchItems;
   while ((currentPtr != NULL) ? (priority < currentPtr->priority) : CLIPS_FALSE)
     {
      lastPtr = currentPtr; 
      currentPtr = currentPtr->next;
     }

   if (lastPtr == NULL)
     { 
      newPtr->next = ListOfWatchItems;
      ListOfWatchItems = newPtr;
     }
   else
     {
      newPtr->next = currentPtr;
      lastPtr->next = newPtr;
     }
   
   return(1);
  }
  
/****************************************************************/
/* SetWatchItem:  Sets the global watch variables to on or off. */
/****************************************************************/
globle BOOLEAN SetWatchItem(item,act_value)
  char *item;
  int act_value; 
  {
   struct watchItem *wPtr;
   
   if ((act_value != ON) && (act_value != OFF)) return(0);

   wPtr = ListOfWatchItems;
   
   if (strcmp(item,"all") == 0)
     {
      while (wPtr != NULL)
        {
         *(wPtr->flag) = act_value;
         wPtr = wPtr->next;
        }
      return(1);
     }
   
   while (wPtr != NULL)
     {
      if (strcmp(item,wPtr->name) == 0)
        { 
         *(wPtr->flag) = act_value;
         return(1);
        }
        
      wPtr = wPtr->next;
     }
     
   if (act_value) ExpectedTypeError("watch",1,"watchable symbol"); 
   else ExpectedTypeError("unwatch",1,"watchable symbol"); 
   
   return(0);
  }
  
/************************************************************************/
/* SetAllWatchItems:  Sets all the global watch variables to on or off. */
/************************************************************************/
globle BOOLEAN SetAllWatchItems(act_value)
  int act_value; 
  {
   struct watchItem *wPtr;
   
   if ((act_value != ON) && (act_value != OFF))
     { return(0);  }

   wPtr = ListOfWatchItems;
   
   while (wPtr != NULL)
     {
      wPtr->temp = *(wPtr->flag);
      *(wPtr->flag) = act_value;
      wPtr = wPtr->next;
     }
   return(1);
  }

/************************************************************************/
/* RestoreAllWatchItems:  Restores the value of all watch items. Should */
/*   only be called to undo the effects of a SetAllWatchItems call.     */
/************************************************************************/
globle VOID RestoreAllWatchItems()
  {
   struct watchItem *wPtr;

   wPtr = ListOfWatchItems;
   
   while (wPtr != NULL)
     {
      *(wPtr->flag) = wPtr->temp;
      wPtr = wPtr->next;
     }
  }
  
/*************************************************************/
/* GetWatchItem:  Gets the value of a global watch variable. */
/*************************************************************/
globle int GetWatchItem(item)
  char *item;
  {
   struct watchItem *wPtr;

   wPtr = ListOfWatchItems;
   
   while (wPtr != NULL)
     {
      if (strcmp(item,wPtr->name) == 0)
        { return(*(wPtr->flag)); }
        
      wPtr = wPtr->next;
     }
     
   return(-1);
  }

/**********************************************************/
/* GetNthWatchName:  Gets the name of the nth watch item. */
/**********************************************************/
globle char *GetNthWatchName(n)
  int n;
  {
   int i = 1;
   struct watchItem *wPtr;

   wPtr = ListOfWatchItems;
   
   while ((wPtr != NULL) && (i < n))
     {
      i++;
      wPtr = wPtr->next;
     }
   
   if (wPtr == NULL) return(NULL);  
   
   return(wPtr->name);
  }
  
/************************************************************/
/* GetNthWatchValue:  Gets the value of the nth watch item. */
/************************************************************/
globle int GetNthWatchValue(n)
  int n;
  {
   int i = 1;
   struct watchItem *wPtr;

   wPtr = ListOfWatchItems;
   
   while ((wPtr != NULL) && (i < n))
     {
      i++;
      wPtr = wPtr->next;
     }
   
   if (wPtr == NULL) return(-1); 
    
   return(*(wPtr->flag));
  }
  
/************************************************/
/* PrintAtom: */
/************************************************/
globle VOID PrintAtom(logicalName,type,value)
  char *logicalName;
  int type;
  VOID *value;
  {
   char buffer[20];
   
   switch (type)
     {
      case FLOAT: 
        PrintFloat(logicalName,ValueToDouble(value));
        break;
      case INTEGER:
        PrintLongInteger(logicalName,ValueToLong(value));
        break;
      case SYMBOL:
        PrintCLIPS(logicalName,ValueToString(value));
        break;
      case STRING:
        if (PreserveEscapedCharacters)
          { PrintCLIPS(logicalName,StringPrintForm(ValueToString(value))); }
        else
          {
           PrintCLIPS(logicalName,"\"");
           PrintCLIPS(logicalName,ValueToString(value));
           PrintCLIPS(logicalName,"\""); 
          }
        break;
        
      case EXTERNAL_ADDRESS:
        if (AddressesToStrings) PrintCLIPS(logicalName,"\"");
        PrintCLIPS(logicalName,"<Pointer: ");
#if ANSI_COMPILER
        sprintf(buffer,"%p",value);
#else
        sprintf(buffer,"%lu",(unsigned long) value);
#endif
        PrintCLIPS(logicalName,buffer);
        PrintCLIPS(logicalName,">");
        if (AddressesToStrings) PrintCLIPS(logicalName,"\"");
        break;
        
#if OBJECT_SYSTEM
      case INSTANCE:
        if (AddressesToStrings) PrintCLIPS(logicalName,"\"");
        PrintCLIPS(logicalName,"<Instance: ");
#if ANSI_COMPILER
        sprintf(buffer,"%p",value);
#else
        sprintf(buffer,"%lu",(unsigned long) value);
#endif
        PrintCLIPS(logicalName,buffer);
        PrintCLIPS(logicalName,">");
        if (AddressesToStrings) PrintCLIPS(logicalName,"\"");
        break;
        
      case INSTANCE_NAME: 
        PrintCLIPS(logicalName,"[");
        PrintCLIPS(logicalName,ValueToString(value)); 
        PrintCLIPS(logicalName,"]");
        break;
#endif

      case FACT_ADDRESS:
        if (AddressesToStrings) PrintCLIPS(logicalName,"\"");
        PrintCLIPS(logicalName,"<Fact-");
        PrintLongInteger(logicalName,((struct fact *) value)->factIndex);
        PrintCLIPS(logicalName,">");
        if (AddressesToStrings) PrintCLIPS(logicalName,"\"");
     }
  }
  
/*******************************************/
/* PrintTally:   */       
/*******************************************/
globle VOID PrintTally(logicalName,count,singular,plural)
  char *logicalName;
  long count;
  char *singular, *plural;
  {
   if (count == 0) return;
   
   PrintCLIPS(logicalName,"For a total of ");
   PrintLongInteger(logicalName,count);
   PrintCLIPS(logicalName," ");
   
   if (count == 1) PrintCLIPS(logicalName,singular);
   else PrintCLIPS(logicalName,plural);
   
   PrintCLIPS(logicalName,".\n");
  }

/*******************************************/
/* AtomInstall:   */       
/*******************************************/
globle VOID AtomInstall(type,vPtr)
  int type;
  VOID *vPtr;
  {
   switch (type)
     {
      case SYMBOL:
      case STRING:
#if DEFGLOBAL_CONSTRUCT
      case GBWORD:
#endif
#if OBJECT_SYSTEM
      case INSTANCE_NAME:
#endif
        IncrementSymbolCount(vPtr);
        break;
        
      case FLOAT:
        IncrementFloatCount(vPtr);
        break;
        
      case INTEGER:
        IncrementIntegerCount(vPtr);
        break;
        
      case FACT_ADDRESS:
        IncrementFactCount(vPtr);
        break;
        
#if OBJECT_SYSTEM
      case INSTANCE:
        IncrementInstanceCount(vPtr);
        break;
#endif
     }
  }
  
/*******************************************/
/* AtomDeinstall:   */       
/*******************************************/
globle VOID AtomDeinstall(type,vPtr)
  int type;
  VOID *vPtr;
  {
   switch (type)
     {
      case SYMBOL:
      case STRING:
#if DEFGLOBAL_CONSTRUCT
      case GBWORD:
#endif
#if OBJECT_SYSTEM
      case INSTANCE_NAME:
#endif
        DecrementSymbolCount(vPtr);
        break;
        
      case FLOAT:
        DecrementFloatCount(vPtr);
        break;
        
      case INTEGER:
        DecrementIntegerCount(vPtr);
        break;
        
      case FACT_ADDRESS:
        DecrementFactCount(vPtr);
        break;
        
#if OBJECT_SYSTEM
      case INSTANCE:
        DecrementInstanceCount(vPtr);
        break;
#endif
     }
  }
  
/**************************************************************/
/* PrintInChunks:  Prints a string in chunks to accomodate  */
/*   systems which have a limit on the maximum size of a      */
/*   string which can be printed.                             */
/**************************************************************/
globle VOID PrintInChunks(logicalName,str)
  char *logicalName, *str;
  {
   char tc, *s_ptr;

   s_ptr = str;
   
   if (s_ptr == NULL) return;

   while (strlen(s_ptr) > 500)
     {
      tc = s_ptr[500];
      s_ptr[500] = EOS;
      PrintCLIPS(logicalName,s_ptr);
      s_ptr[500] = tc;
      s_ptr += 500;
     }

   PrintCLIPS(logicalName,s_ptr);
  }
  
