   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*               VARIABLE MANAGER MODULE               */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _VARIABLE_SOURCE_

#include "setup.h"

#if (! RUN_TIME) && (! BLOAD_ONLY)

#include <stdio.h>
#define _CLIPS_STDIO_

#include "constant.h"
#include "symbol.h"
#include "clipsmem.h"
#include "expressn.h"
#include "reorder.h"
#include "generate.h"
#include "router.h"
#include "utility.h"
#include "variable.h"

#if DEFTEMPLATE_CONSTRUCT
#include "deftempl.h"
#endif

/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/
  
   static struct expr_info  *AnalysisExpressions;
   static struct pat_info   *CurrentPatternInfo;

/************************************************************/
/* SetRuleInformation: */
/************************************************************/
globle VOID SetRuleInformation(ePtr)
  struct expr_info *ePtr;
  { 
   AnalysisExpressions = ePtr;
  }

/************************************************************/
/* SetVariableInformation: */
/************************************************************/
globle VOID SetVariableInformation(ePtr)
  struct pat_info *ePtr;
  { 
   CurrentPatternInfo = ePtr;
  }

/************************************************************/
/* GetVariableInformation: */
/************************************************************/
globle struct pat_info *GetVariableInformation()
  { 
   return(CurrentPatternInfo);
  }
  
/*#################################################*/
/* FUNCTIONS FOR ACCESSING AND MANAGING VARIABLES. */
/*#################################################*/

/************************************************************/
/* FindVariable: */
/*    Note: -1 for element indicates that the variable can be anywhere */
/*    in the pattern.                   */
/************************************************************/
globle struct var_info *FindVariable(name,start,pattern,element,location)
  SYMBOL_HN *name;
  int start, pattern, element, location;
  {
   struct pat_info *temp_pat;
   struct var_info *temp_var;
   
   /*========================================*/
   /* Skip patterns up to point of interest. */
   /*========================================*/
   
   temp_pat = CurrentPatternInfo;
   while ((temp_pat != NULL) ? temp_pat->position < start : CLIPS_FALSE)
     { temp_pat = temp_pat->next; }
   
   if (temp_pat == NULL) return(NULL);
   
   /*=============================================*/
   /* Look for information in preceding patterns. */
   /*=============================================*/
   
   while ((temp_pat != NULL) ? temp_pat->position < pattern : CLIPS_FALSE)
     {
      if (temp_pat->logic != NEGATIVE)
        {
         temp_var = temp_pat->variables;
         while (temp_var != NULL)
           {
            if (temp_var->name == name)
              { return(temp_var); }
            temp_var = temp_var->next;
           }
        }
      
      temp_pat = temp_pat->next;
     }
     
   if (temp_pat == NULL) return(NULL);
   
   /*=======================================*/
   /* Look for information in same pattern. */
   /*=======================================*/
      
   if ((temp_pat->logic == NEGATIVE) && (location == OUTSIDE))
     { return(NULL); }
     
   /* Should the test for element be less than? */
   temp_var = temp_pat->variables;
   while ((temp_var != NULL) ?
          ((temp_var->element < element) || 
           (element == -1) ||
           (location == OUTSIDE)) :
          CLIPS_FALSE)
     {
      if (temp_var->name == name)
        { return(temp_var); }
      temp_var = temp_var->next;
     }
   
   return(NULL);
  }
 
/*******************************************************************/
/* FlushVariableAnalysis: Frees up current variable analysis info. */
/*******************************************************************/
globle VOID FlushVariableAnalysis()
  {
   struct pat_info *temp_pat;
   struct var_info *temp_var, *var_ptr;
   
   while (CurrentPatternInfo != NULL)
     {
      temp_pat = CurrentPatternInfo->next;
      var_ptr = CurrentPatternInfo->variables;
      while (var_ptr != NULL)
        {
         temp_var = var_ptr->next;
         rtn_struct(var_info,var_ptr);
         var_ptr = temp_var;
        }
      rtn_struct(pat_info,CurrentPatternInfo);
      CurrentPatternInfo = temp_pat;
     }
  }
  
/********************************************************/
/* GetFactAddressPosition:                              */
/********************************************************/
globle int GetFactAddressPosition(name)
  SYMBOL_HN *name;
  {
   struct pat_info *pat_ptr;
   
   pat_ptr = CurrentPatternInfo;
   while (pat_ptr != NULL)
     {
      if (pat_ptr->fact_address == name)
        { return(pat_ptr->position); }
      pat_ptr = pat_ptr->next;
     }
     
   return(0);
  }
  
/********************************************************/
/* GetRelationForPattern:                               */
/********************************************************/
globle SYMBOL_HN *GetRelationForPattern(patternNumber)
  int patternNumber;
  {
   struct pat_info *pat_ptr;
   
   pat_ptr = CurrentPatternInfo;
   while ((pat_ptr != NULL) && (patternNumber > 1))
     {
      patternNumber--;
      pat_ptr = pat_ptr->next;
     }
    
   if (pat_ptr != NULL) return(pat_ptr->relation);
    
   return(NULL);
  }

/********************************************************/
/* PatternHasTemplate:                          */
/********************************************************/
globle BOOLEAN PatternHasTemplate(patternNumber)
  int patternNumber;
  {
   struct pat_info *pat_ptr;
   
   pat_ptr = CurrentPatternInfo;
   while ((pat_ptr != NULL) && (patternNumber > 1))
     {
      patternNumber--;
      pat_ptr = pat_ptr->next;
     }
    
   if (pat_ptr != NULL) 
   
#if DEFTEMPLATE_CONSTRUCT 
     {
      if (pat_ptr->relation == NULL)
        { return(CLIPS_FALSE); }
      else if (FindDeftemplate(pat_ptr->relation->contents) != NULL)
        { return(CLIPS_TRUE); }
      else
        { return(CLIPS_FALSE); }
     }
#else
     { return(CLIPS_FALSE); }
#endif 
   else
     { return(CLIPS_FALSE); }
  } 
  
  
/******************************/
/* CountJoins:                */
/******************************/
globle int CountJoins()
  {
   struct expr_info *chase_ptr;
   int count = 0;
   
   chase_ptr = AnalysisExpressions;
   while (chase_ptr != NULL)
     { 
      chase_ptr = chase_ptr->next;
      count++;
     }
     
   return(count);
  }
  
/******************************/
/* GetJoinLogic:            */
/******************************/
globle char GetJoinLogic(join_number)
  int join_number;
  {
   struct expr_info *chase_ptr;
   int count = 1;
   
   chase_ptr = AnalysisExpressions;
   while (chase_ptr != NULL)
     { 
      if (count == join_number) 
        {
         if (chase_ptr->logic == POSITIVE) return('+');
         else if (chase_ptr->logic == NEGATIVE) return('-');
        }
      chase_ptr = chase_ptr->next;
      count++;
     }
     
   return('?');
  }
 
/******************************/
/* GetPrimaryJoinExpression:            */
/******************************/
globle struct expr *GetPrimaryJoinExpression(join_number)
  int join_number;
  {
   struct expr_info *chase_ptr;
   struct expr *temp_jnt;
   int count = 1;
   
   chase_ptr = AnalysisExpressions;
   while (chase_ptr != NULL)
     { 
      if (count == join_number) 
        {
         temp_jnt = chase_ptr->jnt;
         chase_ptr->jnt = NULL;
         return(temp_jnt);
        }
      chase_ptr = chase_ptr->next;
      count++;
     }
     
   return(NULL);
  }
  
/******************************/
/* GetNotJoinExpression:            */
/******************************/
globle struct expr *GetNotJoinExpression(join_number)
  int join_number;
  {
   struct expr_info *chase_ptr;
   struct expr *temp_jnt;
   int count = 1;
   
   chase_ptr = AnalysisExpressions;
   while (chase_ptr != NULL)
     { 
      if (count == join_number) 
        {
         temp_jnt = chase_ptr->otest;
         chase_ptr->otest = NULL;
         return(temp_jnt);
        }
      chase_ptr = chase_ptr->next;
      count++;
     }
     
   return(NULL);
  }
  
/******************************/
/* GetNodeType:               */
/******************************/
globle int GetNodeType(pattern,element)
  int pattern, element;
  {
   struct expr_info *chase_ptr;
   struct pn_test *field_ptr;
   int count = 1;
   
   chase_ptr = AnalysisExpressions;
   while (chase_ptr != NULL)
     { 
      if (count == pattern) 
        {
         field_ptr = chase_ptr->pntl;
         while (field_ptr != NULL)
           {
            if (field_ptr->element == element)
              { return(field_ptr->type); }
            field_ptr = field_ptr->next;
           }
           
         return(STOP);
        }
      chase_ptr = chase_ptr->next;
      count++;
     }
     
   return(KUNKNOWN);
  }

/*********************************************/
/* CountPatternFields: Returns the number of */
/*   fields in the specified pattern.        */
/*********************************************/
globle int CountPatternFields(pattern)
  int pattern;
  {
   struct expr_info *chase_ptr;
   struct pn_test *field_ptr;
   int count = 1;
   
   chase_ptr = AnalysisExpressions;
   while (chase_ptr != NULL)
     { 
      if (count == pattern) 
        {
         field_ptr = chase_ptr->pntl;
         count = 0;
         while (field_ptr != NULL)
           {
            count++;
            field_ptr = field_ptr->next;
           }
           
         return(count);
        }
      chase_ptr = chase_ptr->next;
      count++;
     }
     
   return(0);
  }
  
/******************************/
/* GetPatternExpression:            */
/******************************/
globle struct expr *GetPatternExpression(pattern,element)
  int pattern, element;
  {
   struct expr_info *chase_ptr;
   struct pn_test *field_ptr;
   struct expr *temp_test;
   int count = 1;
   
   chase_ptr = AnalysisExpressions;
   while (chase_ptr != NULL)
     { 
      if (count == pattern) 
        {
         field_ptr = chase_ptr->pntl;
         while (field_ptr != NULL)
           {
            if (field_ptr->element == element)
              { 
               temp_test = field_ptr->pnt;
               field_ptr->pnt = NULL;
               return(temp_test);
              }
            field_ptr = field_ptr->next;
           }
           
         return(NULL);
        }
      chase_ptr = chase_ptr->next;
      count++;
     }
     
   return(NULL);
  }
  
/*******************************************************************/
/* FlushAnalysisExpressions: Frees up expressions generated during */
/*   the analysis phase of a defrule that represent expressions to */
/*   be used in the pattern and join networks.                     */
/*******************************************************************/
globle VOID FlushAnalysisExpressions()
  {
   struct expr_info *temp_expr;
   struct pn_test *temp_pn, *pnt_ptr;
   
   while (AnalysisExpressions != NULL)
     {
      temp_expr = AnalysisExpressions->next;
      
      if (AnalysisExpressions->jnt != NULL) ReturnExpression(AnalysisExpressions->jnt);
      if (AnalysisExpressions->otest != NULL) ReturnExpression(AnalysisExpressions->otest);
      
      pnt_ptr = AnalysisExpressions->pntl;
      while (pnt_ptr != NULL)
        {
         temp_pn = pnt_ptr->next;
         if (pnt_ptr->pnt != NULL) ReturnExpression(pnt_ptr->pnt);
         rtn_struct(pn_test,pnt_ptr);
         pnt_ptr = temp_pn;
        }
      
      rtn_struct(expr_info,AnalysisExpressions);
      AnalysisExpressions = temp_expr;
     }
  }
  
/************************************************************/
/* RuleComplexity: Returns the complexity of a rule for use */
/*   by the LEX and MEA conflict resolution strategies.     */
/************************************************************/
globle int RuleComplexity()
  {
   int complexity = 0;
   struct expr_info *chase_ptr;
   struct pn_test *pn_ptr;
   
   chase_ptr = AnalysisExpressions;
   while (chase_ptr != NULL)
     { 
      complexity += ExpressionComplexity(chase_ptr->jnt);
      complexity += ExpressionComplexity(chase_ptr->otest);
      pn_ptr = chase_ptr->pntl;
      while (pn_ptr != NULL)
        {
         complexity += ExpressionComplexity(pn_ptr->pnt);
         pn_ptr = pn_ptr->next;
        }
      chase_ptr = chase_ptr->next;
     }
     
   return(complexity);
  }
  
/********************************************************************/
/* ExpressionComplexity: Determines the complexity of a expression. */
/********************************************************************/
globle int ExpressionComplexity(exprPtr)
  struct expr *exprPtr;
  {
   int complexity = 0;
   
   while (exprPtr != NULL)
     {
      if (exprPtr->type == FCALL)
        {
         /*=========================================*/
         /* Logical combinations do not add to the  */
         /* complexity, but their arguments do.     */
         /*=========================================*/
         
         if ((exprPtr->value == PTR_AND) || 
                  (exprPtr->value == PTR_NOT) ||
                  (exprPtr->value == PTR_OR))
           { complexity += ExpressionComplexity(exprPtr->arg_list); } 
           
         /*=========================================*/
         /* else other function calls increase the  */
         /* complexity, but their arguments do not. */
         /*=========================================*/
         
         else
           { complexity++; } 
        }
      else if (exprPtr->type == SCALL_COMP_VARS) 
        { complexity++; }
#if DEFGENERIC_CONSTRUCT
      else if (exprPtr->type == GCALL)
        complexity++;
#endif
#if DEFFUNCTION_CONSTRUCT
      else if (exprPtr->type == PCALL)
        complexity++;
#endif
        
      exprPtr = exprPtr->next_arg;
     }
     
   return(complexity);
  }
  
#endif

